-- Various damage related stuff;
-- spawning projectiles, hurtboxes and actually hurting entities, too

freeslot("MT_PSO_HURTBOX", "sfx_cburst")
mobjinfo[MT_PSO_HURTBOX] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	flags = MF_NOGRAVITY|MF_NOCLIPHEIGHT,
	radius = 32<<FRACBITS,
	height = 64<<FRACBITS,
}

PSO.lib.getHitboxPotency = function(source, potency)

	local info = source.player and source.player.pso or source.pso
	local savewpn = info.savewpn
	
	local mult = 100
	
	if not savewpn
		if source.player
			local wp = PSO.lib.getPlayerActiveWeapon(source.player)
			if wp
				savewpn = wp
			end
		end
	end
	
	if savewpn
	and savewpn.pa_damage
		mult = savewpn.pa_damage
	end
	
	return potency*mult/100
end

-- spawns a hurtbox
PSO.lib.spawnHurtBox = function(x, y, z, w, t, source, potency, dtype, duration, rtype, condition, rate)
	
	--assert(source, "spawnHurtBox: No source to get stats from")
	
	if rtype == nil
		rtype = RT_STAGGER
	end
	
	local h = P_SpawnMobj(x, y, z, MT_PSO_HURTBOX)
	h.radius = w/2
	h.height = t
	--h.scale = source.scale
	
	if not source or not source.valid
		h.pso = {potency=1}	-- in case some maths happen
		h.flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPTHING
		return h
	end
	
	if PSO.debug
		h.sprite = SPR_THOK
		h.frame = A
		h.color = SKINCOLOR_RED
		
		/*local coords = {
			{h.x - h.radius, h.y - h.radius},
			{h.x - h.radius, h.y + h.radius},
			{h.x + h.radius, h.y - h.radius},
			{h.x + h.radius, h.y + h.radius},
		}
		
		for i = 1, #coords
			local cx = coords[i][1]
			local cy = coords[i][2]
			local t1 = P_SpawnMobj(cx, cy, h.z, MT_THOK)
			t1.frame = A
			t1.color = SKINCOLOR_RED
			t1.scale = h.scale/2
			t1.tics = duration
			t1.fuse = duration
		end*/
	end	
	
	h.target = source
	h.tics = duration or TICRATE/2
	h.fuse = duration or TICRATE/2
	h.starttime = leveltime
	
	local info = source.player and source.player.pso or source.pso
	info = $ or {}
	local savewpn = info.savewpn
	
	local mult = 100
	
	if not savewpn
		if source.player
			local wp = PSO.lib.getPlayerActiveWeapon(source.player)
			if wp
				savewpn = wp
			end
		end
	end
	
	if savewpn
		mult = savewpn.pa_damage
	end
	
	h.pso = {
		
		satk = info.satk or 1,
		ratk = info.ratk or 1,
		tatk = info.tatk or 1,
		potency = (potency or 1)*mult/100,
		damagetype = dtype or DMT_STRIKE,
		reactiontype = rtype,
		status = condition or 0,
		statuschance = rate or 0,
		savewpn = savewpn,
		origin = {source.x, source.y, source.z}
	}
	
	if info.weakness
		h.pso.element = PSO.info.elementOpposite[info.weakness]
	elseif source.pso and source.pso.element
		h.pso.element = source.pso.element
	end
	
	--print(h.pso.element)
	
	-- share table if possible
	if info.whitelist
		h.pso.whitelist = info.whitelist
	end
	
	return h
end

-- copy attack stats and saved weapon from hit1 to hi2
PSO.lib.copyWeaponProperties = function(hit1, hit2)
	
	if hit1 and hit1.target and hit1.target.valid and hit1.target.player and hit1.pso
	and hit2 and hit2.valid and hit2.pso
		hit2.pso.satk = hit1.pso.satk
		hit2.pso.ratk = hit1.pso.ratk
		hit2.pso.tatk = hit1.pso.tatk
		hit2.pso.savewpn = hit1.pso.savewpn
	end
end

local function maintain_whitelist(mo1, mo2)
	-- add this entity to our whitelist, we can't hurt it anymore after this:
	mo1.pso.whitelist = $ or {}
		
	for i = 1, #mo1.pso.whitelist
		if mo1.pso.whitelist[i] == mo2
			return false	-- we already got hurt by this!
		end
	end
	
	-- check if it was already hit by someone else and add it to our own table to make subsequent checks faster
	if mo1.pso.links
		for i = 1, #mo1.pso.links
			local hb = mo1.pso.links[i]
			if hb and hb.valid
				hb.pso.whitelist = $ or {}
				for j = 1, #hb.pso.whitelist
					if hb.pso.whitelist[j] == mo2
						table.insert(mo1.pso.whitelist, mo2)
						return false
					end
				end
			end
		end	
	end
		
	mo1.pso.whitelist[#mo1.pso.whitelist+1] = mo2	
		
	return true
end

local function ra_tacticalroll_trigger(p)
	if PSO.lib.hasSkill(p, "ra_tacticalroll")
					
		S_StartSoundAtVolume(p.mo, sfx_cburst, 127)
					
		for i = 1, 8
			local s = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_SUPERSPARK)
			s.destscale = 1
			s.momx = P_RandomRange(-32, 32)*FRACUNIT
			s.momy = P_RandomRange(-32, 32)*FRACUNIT
			s.momz = P_RandomRange(-32, 32)*FRACUNIT
			s.colorized = true
			s.color = SKINCOLOR_BLUE
		end
					
		p.pso.pp = min(p.pso.maxpp, p.pso.pp+35)
	end
end

local function fo_phantom_trigger(p)
					
	S_StartSoundAtVolume(p.mo, sfx_cburst, 127)
					
	for i = 1, 8
		local s = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_SUPERSPARK)
		s.destscale = 1
		s.momx = P_RandomRange(-32, 32)*FRACUNIT
		s.momy = P_RandomRange(-32, 32)*FRACUNIT
		s.momz = P_RandomRange(-32, 32)*FRACUNIT
		s.colorized = true
		s.color = SKINCOLOR_BLACK
	end
					
	p.pso.phantomcast = 3
end

local function hu_artsavenger_trigger(p)
	if PSO.lib.hasSkill(p, "hu_artsavenger")
	and not p.pso.artsavenger
					
		S_StartSoundAtVolume(p.mo, sfx_cburst, 127)
					
		for i = 1, 8
			local s = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_SUPERSPARK)
			s.destscale = 1
			s.momx = P_RandomRange(-32, 32)*FRACUNIT
			s.momy = P_RandomRange(-32, 32)*FRACUNIT
			s.momz = P_RandomRange(-32, 32)*FRACUNIT
			s.colorized = true
			s.color = SKINCOLOR_RED
		end
					
		p.pso.artsavenger = true
	end
end

-- collision handling for the hitbox
local function hurtbox_collide(mo1, mo2)
	
	if mo1 and mo1.valid and mo1.type == MT_PSO_HURTBOX
	and mo2 and (mo2.flags & (MF_SHOOTABLE) or mo2.player)
	and mo2.type ~= MT_DOOR_HITBOX	-- this sometimes happens for some reason and the results are catastrophic
		
		-- make it hitbox accurate.
		if mo1.sectorbased --and mo1.subsector.sector ~= mo2.subsector.sector
			-- check if any point of the player's hitbox touches us~
			local points = {
				{mo2.x - mo2.radius, mo2.y - mo2.radius},
				{mo2.x - mo2.radius, mo2.y + mo2.radius},
				{mo2.x + mo2.radius, mo2.y - mo2.radius},
				{mo2.x + mo2.radius, mo2.y + mo2.radius}
			}
			
			local ok = false
			
			for i = 1, #points
				local s = R_PointInSubsector(points[i][1], points[i][2]).sector
				if s == mo1.subsector.sector
					ok = true
					break
				end
			end
			
			if not ok
				return false
			end
		end
		
		local owner = mo1.target
		if not owner or not owner.valid	-- nope.
			return false
		end

		if mo1.pso.locks == 0	-- specifically 0, don't check for nil/0
			return false
		end
		
		if mo1.continuous
		and mo2.lasthit
		and leveltime - mo2.lasthit < mo1.continuous
			return false
		end
		
		--print("OK1")
		-- don't care
		if owner == mo2
		or owner.player and mo2.player
		or (not owner.player and not mo2.player)
			return false
		end
		
		if mo2.player
			local m2info = mo2.player.pso
			if (not mo1.pso.nododge and m2info.phantomdodge) and maintain_whitelist(mo1, mo2)
				fo_phantom_trigger(mo2.player)
				return false
			end
			
			if (not mo1.pso.nododge and m2info.dodgeroll) and maintain_whitelist(mo1, mo2)
				
				--print("playerDodge")
				PSO.lib.runHook("playerDodge", mo2.player, mo1, mo1.target)
				ra_tacticalroll_trigger(mo2.player)
				hu_artsavenger_trigger(mo2.player)
				return false
			end
		end
		
		--print("OK2")
		if mo1.z + mo1.height < mo2.z return false end	-- (too low) why do we have to do this manually....
		if mo1.z > mo2.z + mo2.height return false end	-- (too high)

		-- locks are now handled in the mobjthinker
		/*if mo1.pso.locks
			mo1.pso.locks = $-1
		elseif mo1.pso.locks == 0	-- hit the maximum number of entities
			return false
		end*/
		
		-- whitelisted item (already hit)
		if not maintain_whitelist(mo1, mo2) return end
		
		-- players specifically can only be hit once per frame.
		-- we do this check now so that we still whitelist them on contact, otherwise we'd hit them the next frame and that would defeat the point
		if mo2.player
		and mo2.hitboxcheck == leveltime
			return false	-- do not hit them.
		end
		
		mo2.hitboxcheck = leveltime
		
		if mo1.continuous
			mo2.lasthit = leveltime
		end
		
		-- ok we can get hurt
		mo1.pso.damagelist = $ or {}
		
		mo1.pso.damagelist[#mo1.pso.damagelist+1] = mo2
		--P_DamageMobj(mo2, mo1, owner)
		
		return false	-- we still have never collided, technically.
	end
end

addHook("MobjThinker", function(mo)
	
	local owner = mo.target
	local ownertarget
	
	if not owner or not owner.valid
		P_RemoveMobj(mo)
		return
	end
	
	if owner.player
		ownertarget = owner.player.pso.target
	end
	
	if mo.pso.damagelist
		
		if mo.pso.locks ~= nil
			-- if we have locks, sort the damagelist by relevancy
			if #mo.pso.damagelist > 2
				table.sort(mo.pso.damagelist,
					function(a, b)
						
						-- these two cases shouldn't happen but they do in rare instances i haven't figured out
						if a == nil or not a.valid		-- ????
							return false
						elseif b == nil or not b.valid	-- ???? (2)
							return false
						end
						
						if a == ownertarget
							return true
						end
						
						return R_PointToDist2(mo.x, mo.y, a.x, a.y) < R_PointToDist2(mo.x, mo.y, b.x, b.y)
					end
				)
			end
		end
		
		for i = 1, #mo.pso.damagelist
			if not mo.pso.damagelist[i]
			or not mo.pso.damagelist[i].valid
				continue
			end
			
			P_DamageMobj(mo.pso.damagelist[i], mo, owner)
			
			-- if we have locks, decrement the counter and stop iterating if we 
			if mo.pso.locks
				mo.pso.locks = $-1
				if not mo.pso.locks
					break
				end
			end
		end
		
		mo.pso.damagelist = nil
	end
	
	if server.showhitboxes
		local numsegs = 3
		
		local h = mo	-- lazy
		local coords = {
			{h.x - h.radius, h.y - h.radius},
			{h.x - h.radius, h.y + h.radius},
			{h.x + h.radius, h.y + h.radius},
			{h.x + h.radius, h.y - h.radius},
		}
		
		for i = 1, 2
			for j = 1, #coords
				local cx = coords[j][1]
				local cy = coords[j][2]
				local t1 = P_SpawnMobj(cx, cy, h.z + (i==2 and h.height or 0), MT_THOK)
				t1.frame = A|FF_FULLBRIGHT
				t1.color = SKINCOLOR_RED
				t1.scale = h.scale/2
				t1.tics = 2
				
				local nextj = j+1 < 5 and j+1 or 1
				local tx = coords[nextj][1]
				local ty = coords[nextj][2]
				local dir = R_PointToAngle2(cx, cy, tx, ty)
				local dist = R_PointToDist2(cx, cy, tx, ty)
				
				local sdist = dist/(numsegs+1)
				
				for k = 1, numsegs
					
					
					local sx = cx + FixedMul(sdist, cos(dir))
					local sy = cy + FixedMul(sdist, sin(dir))

					local t1 = P_SpawnMobj(sx, sy, h.z + (i==2 and h.height or 0), MT_THOK)
					t1.frame = A|FF_FULLBRIGHT
					t1.color = SKINCOLOR_RED
					t1.scale = h.scale/6
					t1.tics = 2
					
					sdist = $ + dist/(numsegs+1)
					
				end
			end
		end
	end
	
end, MT_PSO_HURTBOX)

addHook("MobjCollide", hurtbox_collide, MT_PSO_HURTBOX)
addHook("MobjMoveCollide", hurtbox_collide, MT_PSO_HURTBOX)

-- spawns a missile object from mobj with the specified offsets from it
-- speed, angle etc can be specified as well as potency and damage type (strike, ranged or technique)
PSO.lib.spawnMissile = function(source, typ, angle, speed, potency, dtype, ox, oy, oz, rtype, condition, rate)
	
	assert(source, "spawnMissile: No source to get stats and position from")
	
	local x = source.x + FixedMul(ox or 0, source.scale) 
	local y = source.y + FixedMul(oy or 0, source.scale) 
	local z = source.z + FixedMul(oz or 0, source.scale) 
	local msl = P_SpawnMobj(x, y, z, typ)
	msl.scale = source.scale
	--msl.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOBLOCKMAP)
	
	if rtype == nil
		rtype = RT_STAGGER
	end
	
	local info = source.player and source.player.pso or source.pso
	
	assert(info, "spawnMissile: No info on source")

	local mult = 100
	local savewpn = info.savewpn
	if source.player
		local wp = PSO.lib.getPlayerActiveWeapon(source.player)
		if wp
			mult = wp.pa_damage
			savewpn = wp
		end
	end
	
	msl.pso = {
		satk = info.satk or 1,
		ratk = info.ratk or 1,
		tatk = info.tatk or 1,
		potency = (potency or 1)*mult/100,
		damagetype = dtype or DMT_STRIKE,
		reactiontype = rtype,
		status = condition,
		statuschance = rate,
		savewpn = savewpn,
		origin = {source.x, source.y, source.z}
	}

	if info.weakness
		msl.pso.element = PSO.info.elementOpposite[info.weakness]
	end
	
	if angle ~= nil
		msl.angle = angle 
	else
		msl.angle = source.angle
	end
	
	if speed
		P_InstaThrust(msl, msl.angle, FixedMul(msl.scale, speed))
	end
	
	-- who it belongs to
	msl.target = source
	msl.friendly = source.player and true or false	-- in case things go apeshit.
	
	return msl
end

-- aims the missile's z momentum towards the target
-- call straight after spawning the missle
PSO.lib.aimMissile = function(missile, target)
	--assert(target and target.valid, "PSO.lib.aimMissile: no valid target")
	
	local slope = 0
	local speed = FixedHypot(missile.momx, missile.momy)
	local ignoretarget
	
	if not speed return end	-- there's nothing to redirect lol

	if missile.target and missile.target.valid and missile.target.player
	and missile.target.player.PSO_chasecam == 0
		ignoretarget = true
	
	end

	-- only aim if we wouldn't hit to begin with, or if the target is on a slope
	if not ignoretarget and target and target.valid
		
		missile.angle = R_PointToAngle2(missile.x, missile.y, target.x, target.y)
		P_InstaThrust(missile, missile.angle, speed)
		
		if missile.z + missile.height < target.z
		or missile.z > target.z + target.height
		or target.standingslope
			
			local dist = FixedHypot(missile.x - target.x, missile.y - target.y)
			local zdiff = ((target.z+(target.height/2)) - (missile.z)) -- middle of target mobj
			slope = R_PointToAngle2(0, 0, dist, zdiff)
			
			/*local dist = max(1, R_PointToDist2(missile.x, missile.y, target.x, target.y))
			local speed = FixedHypot(missile.momx, missile.momy)
			
			local tics = FixedDiv(dist, speed or 1) /FRACUNIT
		

			missile.momz = ((target.z + target.height/3) - missile.z) / (tics or 1)*/
		end
	elseif missile.target and missile.target.valid and missile.target.player
		slope = missile.target.player.aiming
		
	end
	
	if slope
		missile.momx = FixedMul($, cos(slope))
		missile.momy = FixedMul($, cos(slope))
		missile.momz = FixedMul(speed, sin(slope))
	end
	return slope
end

-- lob the missile towards a target
PSO.lib.lobMissile = function(msl, target, bspeed)
	-- kind of do things like lobshot
	
	if bspeed
		P_InstaThrust(msl, R_PointToAngle2(msl.x, msl.y, target.x, target.y), bspeed)
	end
	
	local spd = R_PointToDist2(0, 0, msl.momx, msl.momy)
	local dist = R_PointToDist2(msl.x, msl.y, target.x, target.y)
					
	// horizontal = dist / airtime;
	local airtime = dist / spd
					
	local vertical = -(P_GetMobjGravity(msl)*airtime)/2
					
	msl.momz = $ + vertical
end

PSO.info.chain_start = {
	140,	-- 1P
	120,	-- 2P
	115,	-- 3P
	110,	-- 4P+
}

PSO.info.chain_increase = {	-- 2    3    4    5    6    7    8
	{30, 30, 10, 5, 5, 2},	-- 170, 200, 210, 215, 220, 222, 224...
	{20, 20, 10, 5, 2},		-- 140, 160, 165, 167, 169...
	{15, 15, 10, 5, 2},		-- 130, 145, 155, 157, 159...
	{10, 10, 5, 3, 2},		-- 120, 130, 135, 138, 140, 142...
}

-- calculates damage
PSO.lib.damageCalc = function(mo, inflictor)
	
	-- the formula is different between players and enemies
	if inflictor and inflictor.valid and inflictor.target and inflictor.target.valid
		PSO.lib.runHook("playerPreDamageCalc", inflictor.target.player, mo, inflictor)
	end
	
	local dmt = inflictor.pso.damagetype or DMT_STRIKE	-- all damage is strike unless stated otherwise...
	local source = inflictor.target
	local minfo = mo.pso or mo.player.pso
	local sinfo
	if source
		sinfo = source.pso or source.player.pso
	end
	local infinfo = inflictor.pso
	
	local extra_mult = 100
	
	local stat = {
		[DMT_STRIKE] = "satk",
		[DMT_RANGED] = "ratk",
		[DMT_TECHNIQUE] = "tatk",
	}
	
	local atk = inflictor.pso[stat[dmt]] or 1
	local def = minfo.def or 1
	local potency = inflictor.pso.potency or 1
	
	-- add player weapon attack if applicable
	if source and source.valid and source.player
		local weapon = infinfo.savewpn or PSO.lib.getPlayerActiveWeapon(source.player)
				
		if weapon
			local wdmt = PSO.info.weaponDamageType[weapon.weapontype]
			
			if wdmt == dmt
				
				local satk
				if sinfo.levelsynch
					satk = PSO.lib.lvSynchWeapon(source.player, sinfo.levelsynch, weapon)
				end
				
				atk = $ + (satk or weapon.atk)
				--print("Add weapon atk")
			end
			
			-- increase potency on weaknesses!
			
			-- for attack element:
			if inflictor.pso.element == minfo.weakness
			or inflictor.pso.element == minfo.weakness2
				potency = $ + potency*15/100
			end
			
			-- for weapon element
			if not weapon.element	-- elementless: elem% of 15% damage bonus (up to 9% at 60%)
				local mult = max(1, (15*(weapon.elementv or 0)/100))
				potency = $ + potency*mult/100
			
			elseif weapon.element == minfo.weakness
			or weapon.element == minfo.weakness2
				local mult = 15 + (15*(weapon.elementv or 0)/100)
				potency = $ + potency*mult/100
			end
		end
		
		-- hateful aggression (HU)
		if PSO.lib.hasSkill(source.player, "hu_hatefulaggression")
			local en = mo
			if en.pso.part
				en = mo.tracer
			end
			en.hateful = true	-- play effect when damaged
			
			if en.target == source
				extra_mult = $+5
			end
		end
		
		-- weakness & brink of death!
		if sinfo.rweakness
			potency = $*3/4	-- -25%
		elseif sinfo.rbrink
			potency = $/2	-- -50%
		end
		
	-- add player armor if applicable
	elseif mo and mo.player
		
		local armor = PSO.lib.getPlayerArmor(mo.player)
		
		if armor
			local sdef
			if minfo.levelsynch
				sdef = PSO.lib.lvSynchArmour(mo.player, minfo.levelsynch, armor)
			end
			
			def = $ + (sdef or armor.def)
			
			if not armor.element	-- elementless armor, elem% of 15% DR
				local mult = (15*armor.elementv/100)
				potency = $ - potency*mult/100
				
			elseif inflictor.pso.element == armor.element
				local mult = 15 + (15*armor.elementv/100)
				potency = $ - potency*mult/100
			end
		end
	end
	
	local dmg
	if inflictor and inflictor.pso and inflictor.pso.fixedpercent
		dmg = (minfo.maxhp*inflictor.pso.potency)/100
	else
		local stat_diff = max(atk/10, (atk - (def)))
		dmg = max(atk/10, (stat_diff * potency)/100 / 3) *extra_mult/100
	end
	
	-- chain:
	if minfo.chain and minfo.chain > 1 and inflictor.pso.combobreaker
	and not source.alternate_chain_mode
		
		local playercount = 0
		for player in players.iterate
			playercount = $+1
			if playercount >= #PSO.info.chain_start
				break
			end
		end
		
		local start = PSO.info.chain_start[playercount] --or PSO.info.chain_start[#PSO.info.chain_start]
		-- check increase...
		local inc = start
		
		if minfo.chain > 2
			for i = 1, minfo.chain -2
				local val = PSO.info.chain_increase[playercount][i]
				inc = $ + (val or PSO.info.chain_increase[playercount][#PSO.info.chain_increase[playercount]])
			end
		end
		
		--print("chain mult: "..inc)
		dmg = $ * inc / 100
	end

	-- hunter arts avenger:
	if source and source.valid and source.player and source.player.pso.artsavenger_on
		dmg = $*4/3
	end
	
	-- lol variance
	local min_variance = 90
	
	-- ranger range skills
	if inflictor.pso.origin	-- origin pos
	and source and source.valid and source.player
		
		local range = FixedHypot(FixedHypot(inflictor.pso.origin[1] - mo.x, inflictor.pso.origin[2] - mo.y), inflictor.pso.origin[3] - mo.z)
		
		if range < PSO.info.CLOSERANGE*FRACUNIT
		and PSO.lib.hasSkill(source.player, "ra_zerorange")
		
		or range > PSO.info.LONGRANGE*FRACUNIT
		and PSO.lib.hasSkill(source.player, "ra_sharpshooter")
			
			-- dmg +10%
			dmg = $*11/10
			mo.rangeskill = true
		end
	end
	
	-- photon blast boosts
	if source and source.valid and source.player
		local pb = PSO.lib.usingPB(source.player)
		
		if pb == "hu_overload" or pb == "hu_berserk"
			dmg = $*3/2		-- +50% dmg
		end
	
	-- player mitigations
	elseif mo and mo.player
		local pb = PSO.lib.usingPB(mo.player)
		
		if pb == "hu_overload"
			dmg = $*7/10
		end
		
		-- Swordbreaker, -15*stack% dmg 
		if mo.player.pso.swordbreaker
			dmg = $ - dmg*mo.player.pso.swordbreaker*15/100
		end
		
		-- Hunter Physique: -5% dmg
		if PSO.lib.hasSkill(mo.player, "hu_hunterphysique")
			dmg = $*95/100
		end
	end


	-- OLD
	-- apply shifta?
	/*if inflictor.pso.shifta
	or sinfo and sinfo.shifta
		dmg = $ * (100+PSO.info.SHIFTA)/100
	end
	
	-- apply deband?
	if minfo and minfo.deband
		dmg = $ * (100-PSO.info.DEBAND)/100
	end
	
	-- apply jellen?
	if inflictor.pso.jellen
	or sinfo and sinfo.jellen
		dmg = $ * (100+PSO.info.JELLEN)/100
	end
	
	-- apply zalure?
	if minfo and minfo.zalure
		dmg = $ * (100-PSO.info.ZALURE)/100
	end*/
	
	local dta1 = {min_variance}
	local dta2 = {100}
	if source and source.player
		PSO.lib.runHook("playerVariance", source.player, mo, inflictor, dta1, dta2)
	end
	min_variance = dta1[1]
	local max_variance = dta2[1]


	local crit_rate = PSO.info.BASECRITRATE	-- base crit rate, 10%
	local crit_damage = PSO.info.BASECRITDMG
	local crit_roll = P_RandomRange(1, 100)

	-- direct contact damage?
	if inflictor.pso.shifta
	and inflictor.player
	and PSO.lib.hasSkill(inflictor.player, "fo_shiftacritical")
	
	-- projectiles
	or sinfo and sinfo.shifta
	and source and source.valid and source.player
	and PSO.lib.hasSkill(source.player, "fo_shiftacritical")
		crit_rate = $+15
	end
	
	-- player: gear
	if source and source.valid and source.player
		local p = source.player
		local armour = PSO.lib.getPlayerActiveArmour(p)
		if armour
			local def = PSO.info.items[armour.ref]
			if def.slots
				for i = 1, def.slots
						
					if armour.slots[i]
						local unit = armour.slots[i]
						crit_rate = $ + (unit.critrate or 0)
						crit_damage = $ + (unit.critdamage or 0)
					end
				end
			end
		end	
	end
	
	local arg = {crit_rate}
	local arg2 = {crit_damage}
	if source and source.valid and source.player
		PSO.lib.runHook("playerCritRate", source.player, mo, inflictor, arg)
		PSO.lib.runHook("playerCritDamage", source.player, mo, inflictor, arg2)
	end
	crit_rate = arg[1]
	crit_damage = arg2[1]
	
	local crit = crit_roll <= crit_rate
	
	-- if we crit, ignore variance.
	if crit
		dmg = $*crit_damage/100
	
	-- if we didn't, apply variance.
	else
		local variance = P_RandomRange(min_variance, max_variance)
		
		--print(variance..", "..tostring(crit))
		
		dmg = $ * variance / 100
	end
	
	local arg = {dmg, crit}
	if mo.player
		PSO.lib.runHook("playerDamagedCalc", mo.player, mo, inflictor, arg)
	end
	if inflictor and inflictor.valid and inflictor.target and inflictor.target.valid and inflictor.target.player
		PSO.lib.runHook("playerDamageCalc", inflictor.target.player, mo, inflictor, arg)
		
	end
	
	return max(1, arg[1]), arg[2]
end

freeslot("sfx_lowhp1", "sfx_lowhp2", "sfx_lowhp3")
PSO.lib.damageObject = function(mo, inflictor, damage, skipchain, crit)
	
	local info = mo.player and mo.player.pso or mo.pso
	
	-- do not damage players after a quest has been cleared
	if mo.player
	and (server.questdata and server.questdata.complete)
		return
	end
	
	-- STO-O-O-OOOP HE'S ALREADY DEAD :sob:
	if not info.hp
		return 
	end
	
	local prevhp = info.hp
	
	if mo.hateful
		local avgx = (mo.x/FRACUNIT + inflictor.x/FRACUNIT)/2 *FRACUNIT
		local avgy = (mo.y/FRACUNIT + inflictor.y/FRACUNIT)/2 *FRACUNIT
		local avgz = (mo.z/FRACUNIT + inflictor.z/FRACUNIT)/2 *FRACUNIT	
		for i = 1, 6
			local p = P_SpawnMobj(avgx, avgy, avgz, MT_THOK)
			p.color = SKINCOLOR_RED
			p.scale = $/2
			p.destscale = 1
			p.fuse = 12
			p.tics = 12
			
			p.momx = P_RandomRange(-16, 16)*FRACUNIT
			p.momy = P_RandomRange(-16, 16)*FRACUNIT
			p.momz = P_RandomRange(-16, 16)*FRACUNIT
		end
		mo.hateful = nil
	end

	if mo.rangeskill
		local avgx = (mo.x/FRACUNIT + inflictor.x/FRACUNIT)/2 *FRACUNIT
		local avgy = (mo.y/FRACUNIT + inflictor.y/FRACUNIT)/2 *FRACUNIT
		local avgz = (mo.z/FRACUNIT + inflictor.z/FRACUNIT)/2 *FRACUNIT
		local rnd = P_RandomRange(0, 1) and 1 or -1
		for i = 1, 6
			local p = P_SpawnMobj(avgx, avgy, avgz, MT_SUPERSPARK)
			p.colorized = true
			p.color = SKINCOLOR_CYAN
			p.fuse = 12
			--p.scale = FRACUNIT*4
			p.angle = R_PointToAngle(p.x, p.y) + ANG1*90 + (i&1 and ANG1*180 or 0)
			
			P_InstaThrust(p, p.angle, P_RandomRange(6, 24)*FRACUNIT)
			p.momz = P_RandomRange(6, 8)*FRACUNIT*(i&1 and -1 or 1)*rnd
		end
		mo.rangeskill = nil
	end
	
	-- do not process this if we are hitting a part
	if inflictor and inflictor.valid
		
		-- jellen/zalure
		if inflictor.pso.inflictjellen
			PSO.lib.addJellen(mo, inflictor.pso.inflictjellen)
		end
		
		if inflictor.pso.inflictzalure
			PSO.lib.addZalure(mo, inflictor.pso.inflictzalure)
		end
		
		if inflictor.pso.blight
			PSO.lib.addBlight(mo)
		end
		
		if (inflictor.player or inflictor.target and inflictor.target.valid and inflictor.target.player)
		and ((not info.part) or (info.part and not info.partlinkdamage))
		
			local p = inflictor.target.player
			
			-- dps counter
			p.pso.dps_damage = $ and $+damage or damage
			p.pso.dps_time = $ or 1
			p.pso.dps_resettimer = TICRATE*15
			
			if server.psodps
				
				server.psodamage = $ + damage
				server.psodamagetime = $ or leveltime
			end
			
			-- elemental blot
			info.blottrackers = $ or {}	-- ready those.
			
			if inflictor.pso.damagetype == DMT_TECHNIQUE and inflictor.pso.element
				
				info.blottrackers[#p] = $ or {0, 0}
				
				local tracker = info.blottrackers[#p]
				if PSO.lib.hasSkill(p, "fo_elementalblot")
					
					if tracker[2]	-- tracker already set
						if inflictor.pso.element == tracker[2]	-- elements match
							tracker[1] = $ + inflictor.pso.potency
						else	-- they don't match, detonate!
							--print("tracker detonation here.")
							info.blottrackers[#p] = {0, 0}

							
							local level = min(4, tracker[1] / PSO.info.ELEMBLOTRATE)
							if level and level > 1
								local potency = PSO.info.ELEMBLOTPB[level-1]
								
								if PSO.lib.playerHasPB(p)
									p.pso.pb = min(PSO.info.MAXPB, $ + potency)
								end
								
								for i = 1, 8
									local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
									s.momx = P_RandomRange(-32, 32)*mo.scale
									s.momy = P_RandomRange(-32, 32)*mo.scale
									s.momz = P_RandomRange(-32, 32)*mo.scale
									s.color = PSO.info.elementColours[inflictor.pso.element]
								end
							end
							
						end
						
					else	-- new one
						info.blottrackers[#p] = {tracker[1] + inflictor.pso.potency, inflictor.pso.element}
					end
					--print("trackers: "..tracker[1]..", "..tracker[2])
				end
				
			end
			
		end
	end
	
	-- dont actually deal damage in dps test mode
	if not server.psodps
	and not (server.quesdata and server.questdata.failed)	-- do not deal damage to anything if the quest is failed.
		-- minhp is the minimum amount of HP something can have, bosses use it so that they can't be phase-skipped etc.
		info.hp = max(info.minhp or 0, $ - damage)
		if info.damagetrackers
		and inflictor and inflictor.valid
			local source = inflictor.target
			if source and source.valid and source.player
				info.damagetrackers[#source.player+1] = $ + damage
				
				-- hunters generate more aggro with that skill!
				if PSO.lib.hasSkill(source.player, "hu_hatefulaggression")
					info.damagetrackers[#source.player+1] = $ + damage/2
				end
				--print("Updating dmg tracker for player "..(#source.player)..": "..info.damagetrackers[#source.player+1])
			end
		end
	end

	
	-- damage is dealt, run hooks...
	-- if either returns true, return.
	if inflictor and inflictor.valid and inflictor.target and inflictor.target.valid and inflictor.target.player or inflictor and inflictor.valid and inflictor.player
	and not skipchain	-- parts stuff.
		local p
		if inflictor.player
			p = inflictor.player
		else
			p = inflictor.target.player
		end	
			
		if PSO.lib.runHook("playerDamage", p, mo, inflictor, damage, skipchain, crit)
			return
		end
	-- generic hook
	elseif PSO.lib.runHook("objectDamage", mo, inflictor, damage, skipchain, crit)
		return
	end

	-- this is handled separatly because PvP might exist one day.
	-- although it may not have potentials, modding suppport is always nice to have.
	if mo.player
	and PSO.lib.runHook("playerDamaged", mo.player, inflictor, damage)
		return
	end
	
	if mo.player
		-- visuals
		if mo.player == displayplayer
			P_StartQuake(FRACUNIT*20, 8)
			
			local sound_thresholds = {15, 30, 50}
			local sound_refs = {sfx_lowhp3, sfx_lowhp2, sfx_lowhp1}
			local maxhp = info.maxhp - info.maxhp*(info.injury or 0)/100
			
			for i = 1, #sound_thresholds
				if prevhp*100/maxhp > sound_thresholds[i]
				and info.hp*100/maxhp <= sound_thresholds[i]
					S_StartSound(nil, sound_refs[i], mo.player)
					break
				end
			end
		end
		
		-- auto mate
		if info.hp > 0
		and not info.matesickness
		and prevhp >= info.maxhp/2
		and info.hp < info.maxhp/2
		and PSO.lib.hasSkill(mo.player, "all_automate")
			local mates = {"monomate", "dimate", "trimate"}
			for i = 1, #mates
				local mate = mates[i]
				local index = PSO.lib.findPlayerItem(mo.player, mate)
				if index
					local it = mates[i]
					local itinfo = PSO.lib.itemInfo(it)
					
					-- if we made it here then we can use it!
					PSO.lib.playerLoseItem(mo.player, index, 1)	-- use the item up
								
					if itinfo.use
						itinfo.use(mo.player)
					end
					break
				end
			end
		end
		
		-- iron will
		if not info.hp	-- dead
		and not info.ironwill
		and PSO.lib.hasSkill(mo.player, "hu_ironwill")
			info.hp = 1
			info.ironwill = true	-- used
			
			local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			s.sprite = SPR_FORC
			s.frame = A|FF_ADD
			s.destscale = FRACUNIT*8
			s.scalespeed = FRACUNIT/7
			s.fuse = 16
			s.tics = s.fuse
			
			P_FlashPal(mo.player, 1, 3)
		end
		
		-- limit break endurance
		if not info.hp
		and PSO.lib.usingPB(mo.player) == "hu_berserk"
			info.hp = 1
			mo.player.powers[pw_flashing] = TICRATE*3 -1

			local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			s.sprite = SPR_FORC
			s.colorized = true
			s.color = SKINCOLOR_RED
			s.frame = A|FF_ADD
			s.destscale = FRACUNIT*8
			s.scalespeed = FRACUNIT/7
			s.fuse = 16
			s.tics = s.fuse
			
			P_FlashPal(mo.player, 1, 3)
			
			info.pb = 0
		end
		
		-- pp absorb
		if info.hp
		and PSO.lib.hasSkill(mo.player, "fo_ppabsorb")
			local ppa = damage/3
			info.pp = min(info.maxpp, info.pp+ppa)
		end
		
		-- swordbreaker
		if info.hp
		and PSO.lib.hasSkill(mo.player, "hu_swordbreaker")
			info.swordbreaker = min(3, $ and $+1 or 1)
			info.swordbreakertimer = TICRATE*10
		end
		
	end
	
	if info.part
	and info.partlinkdamage
	and mo.tracer and mo.tracer.valid
	and not skipchain	-- no more recursion past this shit.
		--print("recursive or garbo")
		PSO.lib.damageObject(mo.tracer, inflictor, damage, true)
	end
	
	-- players: count as damage taken
	if info.quest and damage and damage > 0	-- only count positive damage, just in case.
		info.quest.damagetaken = $ + damage
	end
	
	if not skipchain
	
		local source
		if inflictor and inflictor.valid
			source = inflictor.target or inflictor
		end
		
		PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z + mo.height/2, damage, 0, mo, 0, source, 0, crit)
	end


	-- buildup stuff
	-- photon blast / hunter gale...
	
	local source
	if inflictor and inflictor.valid
	and not skipchain
		source = inflictor.target or inflictor	-- whichever
		if source and source.valid and source.player
			local p = source.player
			--print("A")
			local pb = PSO.lib.playerHasPB(p)
			

			if inflictor and inflictor.valid and inflictor.pso and inflictor.pso.potency
			
				-- photon blast buildup
				if not p.pso.usepb	-- that'd be broken.
				and pb			
					inflictor.pso.pbpenalty = $ or 1
					
					local rate = PSO.info.typeSkills[pb].photonblast_rate or 100
					if mo.pso and mo.pso.zalure
						rate = $ + PSO.info.ZALURE
					end
					
					local amount = (inflictor.pso.potency*3 / inflictor.pso.pbpenalty)*rate/100
						
					
					local hookdta = {amount}
					PSO.lib.runHook("playerPBGeneration", p, mo, inflictor, damage, hookdta)
					amount = hookdta[1]
					
					p.pso.pb = min(PSO.info.MAXPB, $ + amount)
					inflictor.pso.pbpenalty = $+2	-- so that explosions dont give instant free pb
					--print(p.pso.pb)
				end
				
				-- hunter gale
				if PSO.lib.hasSkill(p, "hu_whirlwind")
					p.pso.huntergale = $ or 0
					p.pso.huntergale = min(PSO.info.MAXHUNTERGALE, $+inflictor.pso.potency*2)
					p.pso.huntergaletimer = TICRATE*15	-- 15 seconds before it all goes away!
				end
			end
		end
	end
	
	-- if inflictor isn't a "combo breaker", build the chain combo!
	if inflictor and inflictor.valid	-- environmental damage doesn't count
	and (not inflictor.pso.combobreaker or source and source.valid and source.alternate_chain_mode)
	and not mo.player	-- only works on enemies.
	and not info.chainbreak
	and not skipchain
		
		if not info.chain
			server.chaintrackers[#server.chaintrackers+1] = mo
		end
		
		local chainfac = info.blighted and 2 or 1
		
		info.chain = $ and $+chainfac or 1
		info.maxchaintime = max(TICRATE/2, TICRATE*5 - (info.chain-1)*3/2)
		
		if server.psodps
			server.psodpschain = $ and $+chainfac or 1
		end
		
		if info.jellen	-- chain resist down
			info.maxchaintime = $ + info.maxchaintime*PSO.info.JELLEN/100
		end
		
		info.chaintime = info.maxchaintime
		
		if info.chain == 1
			info.chaintriggers = {}
		end
		
	elseif info.chain and info.chain > 1 and inflictor and inflictor.valid and (inflictor.pso.combobreaker and not source.alternate_chain_mode)
	and not info.chainbreak
	and not skipchain
		
		local chainadd = max(3, 8 - (PSO.lib.countPlayers()-1)*2)
		-- to max out chain duration you need:
		/*
			34 chain in 1p
			45 chain in 2p
			68 chain in 3p
			92 chain in 4p
		*/
		
		info.maxchainbreak = min(TICRATE*10, TICRATE*2 + info.chain*chainadd)

		if info.jellen	-- chain resist down
			info.maxchainbreak = $ + info.maxchainbreak*PSO.info.JELLEN/100
		end
		
		info.chainbreak = info.maxchainbreak
	end
	
	
	if inflictor and inflictor.valid
		
		local cbreak = info.chainbreak and inflictor.pso and (inflictor.pso.combobreaker and not source.alternate_chain_mode)
		
		local painsounds_normal = {sfx_dmpain, sfx_s3k6e, sfx_s3kb4}
		local painsounds_crit = {sfx_mspogo, sfx_cdfm14, sfx_cdfm64, sfx_kc40}
		
		--print(crit)
		local uset = crit and painsounds_crit or painsounds_normal
		
		S_StartSound(mo, uset[P_RandomRange(1, #uset)])	--, 160)
		
		-- screen shake and hitlag
		if inflictor.pso.hitlag
		and inflictor.target.player
			inflictor.target.player.pso.hitlag = inflictor.pso.hitlag
		end
		
		if inflictor.pso.screenshake
		and inflictor.target.player
		and inflictor.target.player == displayplayer
			P_StartQuake(inflictor.pso.screenshake.intensity, inflictor.pso.screenshake.time)
		end
		
		if inflictor.pso.momentummult ~= nil
			inflictor.target.momx = $*inflictor.pso.momentummult/100
			inflictor.target.momy = $*inflictor.pso.momentummult/100
		end
		
		if cbreak
			S_StartSoundAtVolume(mo, sfx_cburst, 127)
			
			-- get average coord
			local avgx = (mo.x/FRACUNIT + inflictor.x/FRACUNIT)/2 *FRACUNIT
			local avgy = (mo.y/FRACUNIT + inflictor.y/FRACUNIT)/2 *FRACUNIT
			local avgz = (mo.z/FRACUNIT + inflictor.z/FRACUNIT)/2 *FRACUNIT
			
			for i = 1, 8
				local s = P_SpawnMobj(avgx, avgy, avgz, MT_SUPERSPARK)
				s.destscale = 1
				s.momx = P_RandomRange(-32, 32)*FRACUNIT
				s.momy = P_RandomRange(-32, 32)*FRACUNIT
				s.momz = P_RandomRange(-32, 32)*FRACUNIT
				s.colorized = true
				s.color = SKINCOLOR_GOLD
			end
			
			-- chain trigger
			local source = inflictor.target or inflictor
			local p = source.player
			
			if p
				
				local playercount = 0
				for player in players.iterate
					if player and player.valid and player.mo and player.mo.valid
						playercount = $+1
					end
				end	
				
				local firstchain = info.chaintriggers[#p]
				
				-- auto turret, spawns a turret that assists you for a few seconds!
				/* MT_PSO_RANGERTURRET*/
				
				PSO.lib.runHook("playerChainBreak", p, mo, inflictor, damage, firstchain)
				
				if PSO.lib.hasSkill(p, "ra_autoturret")
					
					if not firstchain
					and not (p.pso.lastturret and leveltime - p.pso.lastturret < TICRATE*2)
						info.chaintriggers[#p] = true
						
						-- now generate PP.
						-- formula is chain - playercount, capped at 1.
						local fuse = TICRATE*5 + info.chain*5/playercount
						local firerate = TICRATE/2
						
						local turret = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z + p.mo.scale*70, MT_PSO_RANGERTURRET)
						turret.tracer = p.mo
						turret.target = mo
						turret.fuse = fuse
						turret.maxfiredelay = firerate
						turret.angle = R_PointToAngle2(turret.x, turret.y, mo.x, mo.y)
						
						turret.pso = {
							ratk = inflictor.pso.ratk
						}
						
						p.pso.lastturret = leveltime
					end
				end
				
				-- chain reactor: extend max PP on chain break
				if PSO.lib.hasSkill(p, "ra_chainreactor")
					
					if not firstchain
						info.chaintriggers[#p] = true
						
						-- now generate PP.
						-- formula is chain - playercount, capped at 1.
						local mult = max(50, 100 - (playercount-1)*5)
						local pp = max(1, info.chain - playercount)
						
						p.pso.chainreactor = $ or 0
						local prevchainreactor = p.pso.chainreactor
						p.pso.chainreactor = min(30, $+pp) 
						
						-- increase max pp here instead or recalculating everything
						p.pso.maxpp = $ + (p.pso.chainreactor - prevchainreactor) 
					end
				end
				
				-- chain trigger: recover PP on chain break
				if PSO.lib.hasSkill(p, "ra_chaintrigger")
					
					if not firstchain
						info.chaintriggers[#p] = true
						
						-- now give PP back based on chain count and playercount
						-- the formula is: 4*chaincount - 5% per player (max 50%)
						local mult = max(50, 100 - (playercount-1)*5)
						local pp = 4*info.chain*mult/100
						
						--p.pso.pp = min(p.pso.maxpp, $+pp)
						PSO.lib.healPP(p.mo, p.mo, pp)
					end
				end
				
				-- also makes use of the chain trigger table
				if PSO.lib.hasSkill(p, "hu_chainadrenaline")
					if not firstchain
						info.chaintriggers[#p] = true
												
						local mult = max(50, 100 - (playercount-1)*5)
						local duration = 2*info.chain*mult/100
						
						PSO.lib.addShifta(p.mo, duration*TICRATE)
						PSO.lib.addDeband(p.mo, duration*TICRATE)
						--PSO.lib.addZalure(p.mo, duration*TICRATE)
						--p.pso.pp = min(p.pso.maxpp, $+pp)
					end
				end
				
				-- players: deband heals on chain burst!
				if p.pso.deband
				and not firstchain
					info.chaintriggers[#p] = true
					-- base: 2% heal / chain with 1p
					local heal_amount = (2*info.chain)*p.pso.maxhp / 100
					if playercount > 1
						
						local mult = max(30, 100 - ((playercount-1)*20))
						heal_amount = max(1, mult*heal_amount/100)
						--print(heal_amount)
					end
					PSO.lib.healObject(p.mo, p.mo, heal_amount)
				end
			end
		end
		
		if inflictor.pso.ppregen
		and inflictor.target and inflictor.target.player
			
			local regen = inflictor.pso.ppregen
			
			--inflictor.target.player.pso.pp = min(inflictor.target.player.pso.maxpp, $ + inflictor.pso.ppregen)
			-- additional 20% recovery w shifta, 20% less w jellen
			if inflictor.target.player.pso.shifta
				regen = $*(100+PSO.info.SHIFTA)/100
				--inflictor.target.player.pso.pp = min(inflictor.target.player.pso.maxpp, $ + inflictor.pso.ppregen*PSO.info.SHIFTA/100)
			elseif inflictor.target.player.pso.jellen
				--inflictor.target.player.pso.pp = max(0, $ - inflictor.pso.ppregen*PSO.info.SHIFTA/100)
				regen = max(1, $*(100-PSO.info.SHIFTA)/100)
			end
			
			local arg = {regen}
			PSO.lib.runHook("playerAPPRegen", inflictor.target.player, mo, inflictor, arg)
			regen = arg[1]
			
			inflictor.target.player.pso.pp = min(inflictor.target.player.pso.maxpp, $+regen)
		end
		
		if inflictor.pso.lifesteal
		and inflictor.target and inflictor.target.valid
			PSO.lib.healObject(inflictor.target, inflictor.target, damage*inflictor.pso.lifesteal/100)
		end
		
		if not info.hp
			PSO.lib.killObject(mo, inflictor)
		else
			-- check if stagger etc must be inflicted.
			local drt = inflictor.pso.reactiontype or RT_NONE
			local weight
			
			-- weight takes away from drt
			-- so blowdown can become knockback or stagger depending on weight.
			if mo.player
				-- for players, "weight" depends on damage taken vs max hp.
				/*local percentage = damage*100/info.maxhp
				
				if percentage <= 1
					weight = 3	-- less than 1%: take no effect except blowdown as stagger
				
				elseif percentage <= 5	-- less than 5%: blowdown -> knockback, knockback -> stagger
					weight = 2
				
				else
					weight = 1
				end*/
				weight = 1	-- maybe later?
				
				-- old behaviour
				/*if PSO.lib.hasSkill(mo.player, "hu_hunterphysique")
				and not inflictor.pso.ignorephysique
					weight = 2
				end*/
				
			elseif PSO.info.enemies[mo.type]
				-- enemies have weight directly defined on them.
				weight = PSO.info.enemies[mo.type].weight or 1
			end
			
			drt = max(RT_NONE, $ - ((weight or 1)-1))
			
			-- ais cannot suffer knockback
			if mo.player and mo.player.pso.ais
				drt = min($, RT_STAGGER)
			end
			
			if drt == RT_STAGGER
			and mo.player
			and PSO.lib.hasSkill(mo.player, "hu_hunterphysique")
				drt = RT_NONE
				mo.player.powers[pw_flashing] = TICRATE
			end
			
			local arg = {drt}
			
			if not inflictor.pso.ignorephysique
				if mo.player
					PSO.lib.runHook("playerReaction", mo.player, mo, inflictor, arg)
				elseif inflictor.target and inflictor.target.valid and inflictor.target.player
					PSO.lib.runHook("enemyReaction", inflictor.target.player, mo, inflictor, arg)
				end
			end
			
			drt = arg[1]
			
			if PSO.info.enemies[mo.type]
				if PSO.info.enemies[mo.type].damage
					local run, res = pcall(PSO.info.enemies[mo.type].damage, mo, inflictor, damage, drt)
					
					if not run
						print("\x82".."WARNING: ".."\x80".."damageObject: Enemy "..mo.type.." damage function: "..res)
					end
				end
			end
			
			-- statuses!
			if inflictor.pso.status
			and P_RandomRange(1, 100) <= inflictor.pso.statuschance
			and not info.boss
			and not (info.weight != nil and info.weight >= 4)
				
				PSO.lib.inflictStatus(mo, inflictor.pso.status, inflictor.pso.statusduration)
			end
			
			-- superarmor / enemies with deband are immune to reactions
			if (not (info.superarmor or (PSO.info.enemies[mo.type] and info.deband))
			or inflictor.pso.ignorephysique)	-- ignore superarmor status
			and not info.jumppad	-- jump pad should ignore all reactions because it would break horribly otherwise!!
			
				-- if a player got hit by any stagger, stop action
				if drt > RT_NONE and mo.player
				and mo.player.pso.action	-- only reset the action if it's necessary
					local p = mo.player
					PSO.lib.resetPlayerAction(p)
				end
				
				-- now apply the effects
				if drt == RT_STAGGER
					--print("A")
					P_InstaThrust(mo, 0, 0)
					info.knockback = 0
					info.blowdown = 0
					info.stagger = 1
				
				elseif drt == RT_KNOCKBACK
					local strength = inflictor.pso.reactionstrength or 40
					info.stagger = 0
					info.blowdown = 0
					info.knockback = 1
					
					local source = inflictor.target
					if not source or not source.valid
						source = inflictor
					end
					
					local ang = inflictor.angle
					if inflictor.pso.reactionangle ~= nil
						ang = inflictor.pso.reactionangle
					elseif inflictor.pso.reactionradialangle
						ang = R_PointToAngle2(inflictor.x, inflictor.y, mo.x, mo.y)				
						/*if source.player
							ang = source.player.drawangle
						end*/
					end
					--mo.angle = ang + ANG1*180
					P_InstaThrust(mo, 0, 0)
					P_InstaThrust(mo, ang, strength*source.scale)
					
				elseif drt == RT_BLOWDOWN
					local strength = inflictor.pso.reactionstrength or 8
					info.stagger = 0
					info.knockback = 0
					info.blowdown = 1
					
					mo.flags = $ & ~(MF_NOCLIP|MF_NOGRAVITY)
					
					local source = inflictor.target
					if not source or not source.valid
						source = inflictor
					end

					local ang = inflictor.angle
					if inflictor.pso.reactionangle ~= nil
						ang = inflictor.pso.reactionangle
					elseif inflictor.pso.reactionradialangle
						ang = R_PointToAngle2(inflictor.x, inflictor.y, mo.x, mo.y)
					end		
					
					if not mo.player
						mo.angle = ang + ANG1*180
					else
						mo.player.drawangle = ang + ANG1*180
					end
						
					P_InstaThrust(mo, ang, strength*mo.scale)
					mo.momz = strength*mo.scale
				end
			elseif drt > RT_NONE
			and mo.player
				-- if you have superarmour, you're entitled to 1 second of flashtics
				mo.player.powers[pw_flashing] = TICRATE
			end
		end
	else
		S_StartSoundAtVolume(mo, sfx_dmpain, 127)
		if not info.hp
			PSO.lib.killObject(mo, nil)
		end
	end
end

COM_AddCommand("psod_hurtme", function(p, arg)
	
	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	PSO.lib.damageObject(p.mo, nil, tonumber(arg))
end)

PSO.lib.killObject = function(mo, inflictor)
	
	if mo.flags & MF_ENEMY
		local num
		
		local exp_val = mo.pso.exp
		if consoleplayer and consoleplayer.pso.ais
		or PSO.lib.arenaQuestActive()
			exp_val = 0
		end
		
		if mo.pso.exp and not mo.pso.part	-- parts don't give exp.
		and not mo.pso.nopdrops
			num = PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z, exp_val, 1)
		end
		
		-- parts don't drop items.
		if not mo.pso.part
		and not mo.pso.nodrops
			
			-- players who kill enemies in ais cannot get drops UNLESS the enemy is a boss.
			if not (inflictor and inflictor.target and inflictor.target.valid
			and inflictor.target.player
			and inflictor.target.player.pso.ais and not mo.pso.boss)
			
				PSO.lib.enemyDropItems(mo)
			end
		end
		
		-- if the object had parts, kill them too.
		if mo.parts
			for i = 1, #mo.parts
				if mo.parts[i] and mo.parts[i].valid
					--print("kill part "..i)
					mo.parts[i].pso.hp = 0
					PSO.lib.killObject(mo.parts[i])
				end
			end
		end

		if mo.pso.blighted
			mo.pso.blighted = $-1
			
			if not mo.pso.blighted
				if server.blighted
					for i = 1, #server.blighted
						if server.blighted[i] == mo
							table.remove(server.blighted, i)
							break
						end
					end
				end
			end
			
		end
		
		for p in players.iterate do
			local pmo = p.mo
			
			if not pmo or not pmo.valid continue end
			
			if R_PointToDist2(pmo.x, pmo.y, mo.x, mo.y) < PSO.info.expdist*pmo.scale
				
				if not mo.pso.part and not mo.pso.nodrops
					PSO.lib.givePlayerExp(p, mo.pso.exp)

					-- count as a kill!
					if p.pso.quest
						p.pso.quest.kills[mo.type] = $ and $+1 or 1
					end	
				end
				
				-- parts shouldn't count as kills for these skills
				if not mo.pso.part
					if PSO.lib.hasSkill(p, "hu_reapingscythe")
						PSO.lib.healObject(p.mo, p.mo, p.pso.maxhp/10)
					elseif PSO.lib.hasSkill(p, "fo_photonscythe")
						p.pso.pp = min(p.pso.maxpp, $ + p.pso.maxpp/10)
					end
				end
				
			elseif p == consoleplayer
			and num
				num.extravalue2 = 1	-- don't display locally
			end
		end
		
		P_KillMobj(mo)
	
	elseif mo.player
		
		mo.player.pso.deadtimer = 1
		PSO.lib.resetPlayerAction(mo.player)
		PSO.lib.removeStatus(mo)
		PSO.lib.closeMenu(mo.player)	-- close any actve menu
	end
end

-- heals an object, if possible.
PSO.lib.healObject = function(mo, inflictor, heal, silent)
	local info = mo.player and mo.player.pso or mo.pso
	
	-- zalure reduces healing
	if mo.player 
		if info.zalure
			heal = max(1, $/2)
		end
		
		-- hook
		local arg = {heal}
		PSO.lib.runHook("playerHeal", mo.player, inflictor, arg)
		heal = arg[1]
	end
	
	if heal	-- don't do anything if this was reduced to 0.
		
		info.hp = min(info.maxhp, $ + heal)
		
		local source
		if inflictor and inflictor.valid
			source = inflictor.target or inflictor
		end
		
		PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z + mo.height/2, heal, 0, mo, 1, source)
	end
end

PSO.lib.revivePlayer = function(p, pinflictor, amount)
	PSO.lib.healObject(p.mo, pinflictor, amount or p.pso.maxhp/2)
	p.pso.deadtimer = 0
	p.mo.state = S_PLAY_STND
	p.powers[pw_flashing] = TICRATE*3 -1
							
	PSO.lib.closeMenu(p, true)
	
	S_StartSound(p.mo, sfx_heal)
end

PSO.lib.damagePP = function(mo, inflictor, dmg)
	if not mo.player return end	-- this shouldn't happen.
	
	mo.player.pso.pp = max(0, $-dmg)
	
	local source
	if inflictor and inflictor.valid
		source = inflictor.target or inflictor
	end
	
	PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z + mo.height/2, dmg, 0, mo, 0, source, 1)
end

PSO.lib.healPP = function(mo, inflictor, heal, silent)
	if not mo.player return end	-- this shouldn't happen.
	
	mo.player.pso.pp = min(mo.player.pso.maxpp, $+heal)
	
	local source
	if inflictor and inflictor.valid
		source = inflictor.target or inflictor
	end
	
	
	if not silent
		PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z + mo.height/2, heal, 0, mo, 1, source, 1)
		S_StartSound(mo, sfx_heal)
	end
end

-- remove these trackers if we change maps
addHook("MapChange", do
	server.dmgnums = nil
	server.chaintrackers = nil
end)

-- handles damage numbers and chains
local function handleDamageNumbers()
	
	server.dmgnums = $ or {}
	server.chaintrackers = $ or {}
	
	local i = #server.dmgnums
	while i
		if not server.dmgnums[i]
		or not server.dmgnums[i].valid
			table.remove(server.dmgnums, i)
		else
			server.dmgnums[i].momz = $*90/100
		end
		i = $-1
	end

	i = #server.chaintrackers
	while i
		if not server.chaintrackers[i]
		or not server.chaintrackers[i].valid
		or not server.chaintrackers[i].pso.chain
			table.remove(server.chaintrackers, i)
		else
			local m = server.chaintrackers[i]
			if m and m.pso.chain
				
				if m.pso.chainbreak
					m.pso.chainbreak = $-1
				else
					m.pso.chaintime = $-1
				end	
					
				if not m.pso.chaintime
				or m.pso.chainbreak == 0
					m.pso.chain = nil
					m.pso.chaintime = nil	
					m.pso.chainbreak = nil
					table.remove(server.chaintrackers, i)
				end
			end
		end
		i = $-1
	end
	
end
PSO.lib.registerFrameFunc(handleDamageNumbers)


PSO.lib.spawnDamageNumber = function(x, y, z, num, xp, target, heal, source, pp, crit)
	
	local distfac = FRACUNIT
	if displayplayer and displayplayer.valid
		local dist = R_PointToDist(x, y)
		distfac = FixedSqrt(dist/160)	-- let's try this
	end
	
	--print(distfac)
	
	local d = P_SpawnMobj(x, y, z, MT_THOK)
	d.state = S_INVISIBLE
	d.tics = TICRATE*2
	d.momz = FixedMul(distfac, FRACUNIT*8)
	d.target = target
	d.source = source
	
	-- iterate thru damage numbers and move them away if they're going to overlap...
	/*for i = 1, #server.dmgnums
		local dn = server.dmgnums[i]
		
		if dn and dn.valid
			
			local dist = R_PointToDist2(d.x, d.y, dn.x, dn.y)
			if dist < FRACUNIT*32
			and abs(d.z - dn.z) < FRACUNIT*64
				local a = R_PointToAngle(d.x, d.y) + ANG1*90 + (i & 1)*ANG1*180	-- use local camera angle...
				local nx = dn.x + FixedMul(96*cos(a), distfac)
				local ny = dn.y + FixedMul(96*sin(a), distfac)
				P_SetOrigin(d, nx, ny, d.z)
				break
			end
		end
	end*/
	
	-- lol
	d.angle = num
	d.extravalue1 = xp or 0		-- hack
	d.extravalue2 = heal or 0	-- hack (1)
	d.cusval = pp or 0	-- hack (2)
	d.cvmem = crit and 1 or 0	-- critical hits
	
	server.dmgnums[#server.dmgnums+1] = d
	
	return d
end

addHook("ShouldDamage", function(mo, inflictor, source, damagetype)
	
	local minfo = mo.pso
	if not minfo and mo.player
		minfo = mo.player.pso
	end
	
	if (minfo) and inflictor and inflictor.valid and inflictor.pso
		
		if inflictor.target == mo
			return false
		end
		
		--non-players and players don't damage among each other
		if (inflictor.target and inflictor.target.player and mo.player)
		or (inflictor.target and not inflictor.target.player and not mo.player)
			return false
		end
		
		-- hurtboxes have their own damage handling.
		if inflictor.flags & MF_MISSILE|MF_PAIN and inflictor.type ~= MT_PSO_HURTBOX
			if not maintain_whitelist(inflictor, mo) return false end
		end
		
		-- players don't take damage from direct enemy contact.
		if mo.player
				
			if inflictor.flags & MF_ENEMY
				-- push me out
				--local ang = R_PointToAngle2(inflictor.x, inflictor.y, mo.x, mo.y)
				--P_InstaThrust(mo, ang, 16*FRACUNIT)
				return false
			end
			
			-- other damage sources.
			if inflictor.type ~= MT_PSO_HURTBOX
				if (not inflictor.pso.nododge and mo.player.pso.dodgeroll)
					--print("playerDodge")
					PSO.lib.runHook("playerDodge", mo.player, inflictor, inflictor.target)
					ra_tacticalroll_trigger(mo.player)
					hu_artsavenger_trigger(mo.player)
					return false
				elseif (not inflictor.pso.nododge and mo.player.pso.phantomdodge)
					fo_phantom_trigger(mo.player)
					return false
				end
			end
			
			if mo.player.powers[pw_flashing]	-- hurt
			or mo.player.pso.iframes	-- iframes
			or mo.player.pso.jumppad	-- jumping
			or not mo.player.pso.type_initialized	-- new joiner?'
			or (server.questdata and server.questdata.complete)	-- quest complete
			--or minfo.dodgeroll	-- dodging
				return false
			end
		end
		
		-- entities with nil hp use their normal damage function/hooks whatever
		if minfo.hp == nil
			--print("Do damage")
			return true
		end
		
		-- used mostly by enemies
		if minfo.invincible
			return false
		end

	
		-- mitigations (guard points)
		if minfo.mitigations
			-- there can be multiple
			-- do not use hitboxes as reference, use the spawner if that's the case.
			local ref = inflictor
			if not (inflictor.flags & MF_MISSILE)
				ref = source
			end	
			
			if ref and ref.valid
				for i = 1, #minfo.mitigations
					local mit = minfo.mitigations[i]
					
					local a = R_PointToAngle2(mo.x, mo.y, ref.x, ref.y)
					local a1 = mit[1]
					local a2 = mit[2]
					
					if PSO.lib.angBetween(a, a1, a2)
						
						-- sfx: s3k7b
						S_StartSound(mo, sfx_s3k7b)
						PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z + mo.height/2, 0, 0, mo, 0, source)
						
						-- mitigation hook
						if mo.player
							PSO.lib.runHook("playerBlock", mo.player, inflictor, inflictor.target)
						end
						
						return false
					end
				end
			end
		end
		
		-- hunter arts avenger
		if minfo.avengerguard and PSO.lib.hasSkill(mo.player, "hu_artsavenger")
			S_StartSound(mo, sfx_s3k7b)
			S_StartSoundAtVolume(mo, sfx_cburst, 127)
			PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z + mo.height/2, 0, 0, mo, 0, source)		
			minfo.artsavenger_on = true

			for i = 1, 32
				local p = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
				p.colorized = true
				p.color = SKINCOLOR_CRIMSON
					
				p.momx = P_RandomRange(-16, 16)*FRACUNIT
				p.momy = P_RandomRange(-16, 16)*FRACUNIT
				p.momz = P_RandomRange(-16, 16)*FRACUNIT
			end
			
			minfo.hitlag = TICRATE/3
			minfo.avengerhitlag = true
			minfo.iframes = TICRATE*3/2	-- LENGHTY iframes...
			
			if mo.player == displayplayer
				P_StartQuake(FRACUNIT*20, TICRATE/5)
			end
			
			return false
		end
		
		-- ais reflect shield, only used by players...
		if minfo.aisreflect and inflictor and inflictor.valid	-- needs to have an inflictor.
			
			if not minfo.aisreflectend
				
				local avgx = (mo.x/FRACUNIT + inflictor.x/FRACUNIT)/2 *FRACUNIT
				local avgy = (mo.y/FRACUNIT + inflictor.y/FRACUNIT)/2 *FRACUNIT
				local avgz = (mo.z/FRACUNIT + inflictor.z/FRACUNIT)/2 *FRACUNIT	
				for i = 1, 6
					local p = P_SpawnMobj(avgx, avgy, avgz, MT_SUPERSPARK)
					p.colorized = true
					p.color = SKINCOLOR_MASTER
					
					p.momx = P_RandomRange(-16, 16)*FRACUNIT
					p.momy = P_RandomRange(-16, 16)*FRACUNIT
					p.momz = P_RandomRange(-16, 16)*FRACUNIT
				end
				
				S_StartSound(mo, sfx_cdfm41)
				minfo.aisreflect[#minfo.aisreflect+1] = inflictor.target and inflictor.target or inflictor
				--minfo.pp = max(0, $ - 25)
			end
			return false	-- no damage calc etc.
		end
		
		local dmg, crit = PSO.lib.damageCalc(mo, inflictor)
		--print(crit)
		--PSO.lib.dprint("Should deal "..dmg.." damage")
		PSO.lib.damageObject(mo, inflictor, dmg, nil, crit)
		return false
	else
		-- don't damage
		return false
	end
end)

-- players: ignore missiles from friendly sources
local function player_missile_collide(mo1, mo2)
	if mo1 and mo1.valid and mo1.player
	and mo2 and mo2.valid and mo2.flags & MF_MISSILE
		
		if mo2.target and mo2.target.valid and mo2.target.player
			return false
		end
	end
end

addHook("MobjCollide", player_missile_collide, MT_PLAYER)
addHook("MobjMoveCollide", player_missile_collide, MT_PLAYER)