-- Status condition behaviours for both players and enemies
freeslot("sfx_shiftp", "sfx_debp", "sfx_jelp", "sfx_zalp")

-- for blind status
local sectorbrightness = {}
local mapweather
local mapsky

addHook("MapChange", do
	sectorbrightness = {}
	mapweather = 0
	mapsky = 0
end)

addHook("MapLoad", do
	for s in sectors.iterate
		sectorbrightness[#s] = s.lightlevel
	end
end)

PSO.info.statusConditions = {
	
	[COND_BURN] = {
		
		name = "Burn",
		tip = "Losing HP over time.",
		
		duration = TICRATE*10,
		
		startfunc = function(mo)
			S_StartSound(mo, sfx_s3k70)
		end,
		
		runfunc = function(mo, timer)
			
			local info
			if mo.player
				info = mo.player.pso
			else
				info = mo.pso
			end
			
			-- visuals
			local steam = P_SpawnMobj(mo.x+P_RandomRange(-45, 45)*mo.scale, mo.y+P_RandomRange(-45, 45)*mo.scale, mo.z+P_RandomRange(5, 30)*mo.scale, MT_SMOKE)
			P_SetObjectMomZ(steam, P_RandomRange(1, 3)*mo.scale)

			if not (leveltime%2)
				local fire = P_SpawnMobj(mo.x+P_RandomRange(-45, 45)*mo.scale, mo.y+P_RandomRange(-45, 45)*FRACUNIT, mo.z+P_RandomRange(5, 30)*mo.scale, MT_THOK)
				fire.sprite = SPR_FPRT
				fire.frame = $ & ~FF_TRANSMASK
				fire.frame = $|FF_FULLBRIGHT|TR_TRANS20
				fire.scale = mo.scale*5/2
				fire.momz = P_RandomRange(1, 3)*mo.scale
				fire.scalespeed = mo.scale/12
				fire.destscale = 1
				fire.tics = TICRATE
			end		
			
			
			if timer % (TICRATE) == 0
				
				-- deal damage
				
				local maxdmg = info.maxhp/25
				if mo.pso
					maxdmg = info.maxhp/50
				end
				
				local dmg = min(info.hp-1, maxdmg)
				
				if dmg and dmg > 0
					PSO.lib.damageObject(mo, nil, dmg)
				end
			end
		end,
		
	},

	[COND_FREEZE] = {

		name = "Freeze",
		tip = "Cannot move or use attacks.",		
		duration = TICRATE*5,
		
		startfunc = function(mo)
			S_StartSound(mo, sfx_s3k80)
			mo.freezeframe = mo.frame
			mo.freezesprite2 = mo.sprite2
			mo.freezeangle = mo.angle
			mo.colorized = true
			mo.freezecolor = mo.color
			mo.color = SKINCOLOR_CYAN
			
			if mo.pso	-- enemy
				mo.savedetectrange = mo.pso.detectrange
				mo.pso.detectrange = 1	-- no.
				for i = 1, 32
					mo.pso.damagetrackers[i] = 0
				end
				mo.target = nil			-- lmao.
			end
			
		end,
		
		runfunc = function(mo, timer)
			
			local info
			if mo.player
				info = mo.player.pso
				mo.player.cmd.angleturn = mo.freezeangle/FRACUNIT
				mo.player.drawangle = mo.freezeangle
				mo.player.normalspeed = 0
			else
				info = mo.pso
				mo.angle = mo.freezeangle
				mo.target = nil
				for i = 1, 32
					mo.pso.damagetrackers[i] = 0
				end				
			end
			
			P_InstaThrust(mo, 0, 0)
			mo.momz = 0
			
			mo.sprite2 = mo.freezesprite2
			mo.frame = mo.freezeframe
			if mo.tics > 0
				mo.tics = $+1
			end
			
		end,
		
		endfunc = function(mo)
			mo.colorized = false
			mo.color = mo.freezecolor
			mo.freezeframe = nil
			mo.freezecolor = nil
			mo.freezeangle = nil
			
			if mo.savedetectrange
				mo.pso.detectrange = mo.savedetectrange
			end
		end,
		
	},

	[COND_SHOCK] = {

		name = "Shock",
		tip = "Cannot use attacks.",
		etip = "Movement slowed.",
		duration = TICRATE*10,
		
		startfunc = function(mo)
			S_StartSound(mo, sfx_s3k7c)
			mo.colorized = true
			mo.shockcolor = mo.color
			mo.color = SKINCOLOR_YELLOW
			
		end,
		
		runfunc = function(mo, timer)
			if leveltime & 1
				local f = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
				f.sprite = SPR_ELEC
				f.frame = P_RandomRange(0, 4)
				f.tics = 1
				f.colorized = true
				f.color = SKINCOLOR_YELLOW
				f.scale = $*2/3
				
			end	
			
			if leveltime%(TICRATE*2) == 0
				S_StartSound(mo, sfx_s3k7c)
			end
			
			if mo.pso
				mo.momx = $/2
				mo.momy = $/2
			end
		end,
		
		endfunc = function(mo)
			mo.colorized = false
			mo.color = mo.shockcolor
			mo.shockcolor = nil
		end,
	},

	[COND_POISON] = {

		name = "Poison",
		tip = "Losing HP over time.",		
		duration = TICRATE*10,
		
		startfunc = function(mo)
			S_StartSound(mo, sfx_s3k95)
		end,
		
		runfunc = function(mo, timer)
			
			local info
			if mo.player
				info = mo.player.pso
			else
				info = mo.pso
			end
			
			-- visuals
			for i = 1,2
				local psn = P_SpawnMobj(mo.x+P_RandomRange(-15,15)*FRACUNIT, mo.y+P_RandomRange(-15,15)*FRACUNIT, mo.z+P_RandomRange(0,15)*FRACUNIT, MT_THOK)
				psn.sprite = SPR_THOK
				psn.frame = A
				psn.color = SKINCOLOR_PURPLE
				psn.scale = FRACUNIT/2
				psn.destscale = FRACUNIT/99
				psn.momz = P_RandomRange(5, 12)*FRACUNIT
				psn.tics = TICRATE
			end
			
			
			if timer % (TICRATE) == 0
				
				-- deal damage
				
				local dmg = min(info.hp-1, info.maxhp/25)
				
				if dmg and dmg > 0
					PSO.lib.damageObject(mo, nil, dmg)
				end
			end
		end,
		
	},

	[COND_CONFUSE] = {

		name = "Confusion",
		tip = "Movement direction shuffled.",
		etip = "Enemy movement halted.",
		duration = TICRATE*10,
		
		startfunc = function(mo)
			S_StartSound(mo, sfx_cdfm21)
			mo.colorized = true
			mo.confcolor = mo.color
			mo.color = SKINCOLOR_WHITE
			
		end,
		
		runfunc = function(mo, timer)
			-- stops enemy movement, players have a different behaviour
			if not mo.player
				mo.momx = 0
				mo.momy = 0
			end

			local an = leveltime * ANG1 * 8
			for i = 1, 2 do
				local dx = mo.x + 32*cos(an)
				local dy = mo.y + 32*sin(an)
				local dz = mo.z + mo.height + 6*sin(an + leveltime*ANG1*4)
				local s = P_SpawnMobj(dx, dy, dz, MT_THOK)
				s.state = S_SPRK1
				s.destscale = 0
				s.frame = A|FF_FULLBRIGHT
				--s.tics = 1
				if (leveltime%2)
					s.fuse = 2
				end
				an = $+ ANG1*180
			end
			
		end,
		
		endfunc = function(mo)
			mo.colorized = false
			mo.color = mo.confcolor
			mo.confcolor = nil
		end,
	},
	
	[COND_BLIND] = {

		name = "Blind",
		tip = "Vision impaired, cannot lock on.",
		etip = "Enemy movement halted.",
		
		duration = TICRATE*10,
		
		startfunc = function(mo)
			S_StartSound(mo, sfx_s3ka1)
			
			if mo.player == consoleplayer
			
				mapweather = curWeather
				mapsky = levelskynum
				--print(mapsky)
				P_SwitchWeather(PRECIP_NONE, consoleplayer)
				P_SetupLevelSky(45, consoleplayer)
				
				for s in sectors.iterate
					--sectorbrightness[#s] = s.lightlevel
					s.lightlevel = $/3
				end
			end
			
		end,
		
		runfunc = function(mo, timer)
			-- stops enemy movement, players have a different behaviour
			if not mo.player
				mo.momx = 0
				mo.momy = 0
			end
			
		end,
		
		endfunc = function(mo)
			if mo.player == consoleplayer and sectorbrightness
				if mo.y > PSO.info.LOBBY_YPOS*FRACUNIT
					P_SwitchWeather(mapweather, consoleplayer)
					P_SetupLevelSky(mapsky, consoleplayer)
				end
				for s in sectors.iterate
					local bright = sectorbrightness[#s]
					if bright
						s.lightlevel = bright
					end
				end
			end
		end,
	},	

	[COND_CORRUPTION] = {

		name = "uyfg*&v6b^5e",
		tip = "yutgfvV67uikv6g7iv65VC786uygfv&65#$",
		
		duration = TICRATE*20,
		
		startfunc = function(mo)
			S_StartSound(mo, sfx_s3k7c)			
		end,
		
		runfunc = function(mo, timer)
			if leveltime%2 == 0
					
				local t = P_SpawnMobj(mo.x + P_RandomRange(-32, 32)*mo.scale, mo.y + P_RandomRange(-32, 32)*mo.scale, mo.z, MT_THOK)
				t.tics = TICRATE
				t.fuse = TICRATE
				t.sprite = SPR_NTHK
				t.frame = E|FF_ADD|TR_TRANS20|FF_FULLBRIGHT
				t.color = SKINCOLOR_MASTER
				t.destscale = 1
				t.momz = P_RandomRange(2, 5)*mo.scale
			end
			
		end,
	},	
	
	-- the following statuses aren't real statuses and just exist for their tips to be disaplayed
	[COND_SHIFTA] = {
		name = "Shifta",
		tip = "PP recovery up.",
		etip = "Movement & Attack speed up.",
	},
	
	[COND_DEBAND] = {
		name = "Deband",
		tip = "Heal on chain detonation.",
		etip = "Stagger immunity.",
	},
	
	[COND_JELLEN] = {
		name = "Jellen",
		tip = "PP recovery down.",
		etip =  "Chain duration up.",
	},
	
	[COND_ZALURE] = {
		name = "Zalure",
		tip = "HP recovery halved.",
		etip = "Photon Blast generation up.",
	},
	
	[COND_MATESICKNESS] = {
		name = "Mate Sickness",
		tip = "Cannot use -mate items.",
	},
	
	[COND_LEVELSYNCH] = {
		name = "Level Synch",
		tip = "Stats synched to quest level.",
	},

	[COND_WEAKNESS] = {
		name = "Reverser Weakness",
		tip = "Damage -25%.",
	},

	[COND_BRINKOFDEATH] = {
		name = "Reverser Weakness II",
		tip = "Damage -50%.",
	},
	
}

PSO.lib.removeStatus = function(mo)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	if not info.condition
		return
	end
	
	local dta = PSO.info.statusConditions[info.condition]
	
	if dta
		
		if dta.endfunc
			dta.endfunc(mo)
		end
		
		info.condition = nil
		info.conditiontimer = nil
	end
end

PSO.lib.inflictStatus = function(mo, cond, duration)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	-- status is already applied
	if info.condition == cond
		return
	end
	
	-- player is using an ais
	if info.ais
		return
	end

	PSO.lib.removeStatus(mo)
	
	local dta = PSO.info.statusConditions[cond]
	
	if dta
		
		if dta.startfunc
			dta.startfunc(mo)
		end
		
		info.condition = cond
		info.conditiontimer = duration or dta.duration
	end
end

COM_AddCommand("psod_status", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	local cond = tonumber(arg)
	PSO.lib.inflictStatus(p.mo, cond)
	
end)

PSO.lib.addShifta = function(mo, duration)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	if not info.shifta or info.shifta < TICRATE*180
		info.shifta = min(TICRATE*180, $ and $+duration or duration)
	end
	info.jellen = 0
	
end

PSO.lib.addDeband = function(mo, duration)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	if not info.deband or info.deband < TICRATE*180
		info.deband = min(TICRATE*180, $ and $+duration or duration)
	end
	info.zalure = 0
	
	if mo.player
		if PSO.lib.hasSkill(mo.player, "fo_debandtoughness")
		and not info.debandtoughness
			info.debandtoughness = info.maxhp/10
			info.maxhp = $ + info.debandtoughness
			info.hp = $ + info.debandtoughness
		end
	end
	
end

PSO.lib.addJellen = function(mo, duration)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	info.jellen = min(TICRATE*180, $ and $+duration or duration)
	info.shifta = 0
	
	-- make the effect also go to the source for enemies.
	if info.part
		local source = mo.tracer
		PSO.lib.addJellen(source, duration)
	end
	
end

PSO.lib.addZalure = function(mo, duration)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	info.zalure = min(TICRATE*180, $ and $+duration or duration)
	info.deband = 0
	
	-- remove toughness if it had been applied.
	if info.debandtoughness
		info.maxhp = $ - info.debandtoughness
		info.hp = min($, info.maxhp)
		info.debandtoughness = nil
	end

	-- make the effect also go to the source for enemies.
	if info.part
		local source = mo.tracer
		PSO.lib.addZalure(source, duration)
	end
end

PSO.lib.addInjury = function(mo, amount, duration, force)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	if force
	or not info.injury or info.injury <= amount
		info.injury = amount
		info.injuryduration = duration
	end
end

PSO.lib.addBlight = function(mo)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	info.blighted = TICRATE*20
	
	if server.blighted
		for i = 1, #server.blighted
			if server.blighted[i] == mo
				return
			end
		end
		
		server.blighted[#server.blighted+1] = mo
		
	else
		server.blighted = {mo}
	end
end

local sdjz_visualdelay = TICRATE*2
local sdjz_dist = 48

local function SDJZvisuals(mo)
	
	local info
	local p
	
	if mo.player
		info = mo.player.pso
		p = mo.player
	else
		info = mo.pso
	end
	
	local atktimer = (leveltime)
	local deftimer = (leveltime- TICRATE/5)
	
	
	if atktimer % sdjz_visualdelay == 0 and (info.shifta or info.jellen)
		
		for i = 1, 8
			
			local a = P_RandomRange(0, 359)*ANG1
			local sx = mo.x+ sdjz_dist*cos(a)
			local sy = mo.y+ sdjz_dist*sin(a)
			
			local spark = P_SpawnMobj(sx, sy, mo.z, MT_SUPERSPARK)
			spark.tics = $ + P_RandomRange(0, 4)
			spark.colorized = true
			spark.scale = mo.scale/2
			spark.momx = $ + mo.momx
			spark.momy = $ + mo.momy
			spark.renderflags = $|RF_FULLBRIGHT
			spark.flags2 = $|MF2_DONTDRAW
			spark.fuse = TICRATE/2
			
			if info.shifta
				if p == displayplayer
					S_StartSoundAtVolume(mo, sfx_shiftp, 128)
				end
				spark.z = mo.z + mo.height/4
				spark.momz = mo.scale*P_RandomRange(6, 12)
				spark.color = SKINCOLOR_RED
				spark.flags2 = $ & ~MF2_DONTDRAW
				
			elseif info.jellen
				
				if p == displayplayer
					S_StartSoundAtVolume(mo, sfx_jelp, 128)
				end
				spark.z = mo.z + mo.height
				spark.momz = mo.scale*P_RandomRange(-12, -6)
				spark.color = SKINCOLOR_CRIMSON
				spark.flags2 = $ & ~MF2_DONTDRAW
			end
		end
		
	elseif deftimer % sdjz_visualdelay == 0 and (info.deband or info.zalure)

		for i = 1, 8
			
			local a = P_RandomRange(0, 359)*ANG1
			local sx = mo.x+ sdjz_dist*cos(a)
			local sy = mo.y+ sdjz_dist*sin(a)
			
			local spark = P_SpawnMobj(sx, sy, mo.z, MT_SUPERSPARK)
			spark.tics = $ + P_RandomRange(0, 4)
			spark.colorized = true
			spark.scale = mo.scale/2
			spark.momx = $ + mo.momx
			spark.momy = $ + mo.momy
			spark.renderflags = $|RF_FULLBRIGHT
			spark.flags2 = $|MF2_DONTDRAW
			spark.fuse = TICRATE/2
			
			if info.deband
				if p == displayplayer
					S_StartSoundAtVolume(mo, sfx_debp, 128)
				end
				spark.z = mo.z + mo.height/4
				spark.momz = mo.scale*P_RandomRange(6, 12)
				spark.color = SKINCOLOR_BLUE
				spark.flags2 = $ & ~MF2_DONTDRAW
				
			elseif info.zalure
				
				if p == displayplayer
					S_StartSoundAtVolume(mo, sfx_zalp, 128)
				end
				spark.z = mo.z + mo.height
				spark.momz = mo.scale*P_RandomRange(-12, -6)
				spark.color = SKINCOLOR_DUSK
				spark.flags2 = $ & ~MF2_DONTDRAW
			end
		end	
	end
end

PSO.lib.handleStatus = function(mo)
	
	local info
	if mo.player
		info = mo.player.pso
	else
		info = mo.pso
	end
	
	if info.condition
		
		local dta = PSO.info.statusConditions[info.condition]
		
		if dta.runfunc
			dta.runfunc(mo, info.conditiontimer)
		end
		
		info.conditiontimer = $-1
		
		if not info.conditiontimer
			PSO.lib.removeStatus(mo)			
		end
		
	end
	
	-- stackable buffs/debuffs
	-- we'll handle these separatly since it doesn't matter too much...
	
	if info.matesickness
		info.matesickness = $-1
	end
	
	-- these statuses do not tick down if you're dead mind you!!
	if info.rweakness
	and info.hp
		info.rweakness = $-1
	end
	
	if info.rbrink
	and info.hp
		info.rbrink = $-1
	end
	
	if info.shifta
		info.shifta = $-1
	end
	
	if info.deband
		info.deband = $-1

		-- remove toughness if it had been applied.		
		
		if info.debandtoughness
			if not info.deband
			or (mo.player and not PSO.lib.hasSkill(mo.player, "fo_debandtoughness"))	-- skill switching can be problematic.
			
				info.maxhp = $ - info.debandtoughness
				info.hp = min($, info.maxhp)
				info.debandtoughness = nil
			end
		end
	end
	
	if info.jellen
		info.jellen = $-1
	end
	
	if info.zalure
		info.zalure = $-1
	end
	
	if info.shifta or info.jellen or info.deband or info.zalure
		SDJZvisuals(mo)
	end
	
	if info.injury and info.injuryduration
		if info.injuryduration > 0
			info.injuryduration = $-1
			if info.injuryduration <= 0
				info.injury = nil
				info.injuryduration = nil
			end
		end
	end
	
	if info.blighted
		info.blighted = $-1
		
		if not info.blighted
			if server.blighted
				for i = 1, #server.blighted
					if server.blighted[i] == mo
						table.remove(server.blighted, i)
						break
					end
				end
			end
		end
		
	end
end

-- effect for confusion:
addHook("PreThinkFrame", do
	for p in players.iterate do
		if p and p.valid and p.pso and p.pso.condition == COND_CONFUSE
			p.cmd.forwardmove = $ * (-1)
			p.cmd.sidemove = $ * (-1)	
		end
	end
end)

COM_AddCommand("psod_shifta", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	local duration = tonumber(arg)*TICRATE
	duration = $ > 0 and $ or TICRATE*90
	
	PSO.lib.addShifta(p.mo, duration)
end)

COM_AddCommand("psod_deband", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local duration = tonumber(arg)*TICRATE
	duration = $ > 0 and $ or TICRATE*90
	
	PSO.lib.addDeband(p.mo, duration)
end)

COM_AddCommand("psod_jellen", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local duration = tonumber(arg)*TICRATE
	duration = $ > 0 and $ or TICRATE*90
	
	PSO.lib.addJellen(p.mo, duration)
end)

COM_AddCommand("psod_zalure", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local duration = tonumber(arg)*TICRATE
	duration = $ > 0 and $ or TICRATE*90
	
	PSO.lib.addZalure(p.mo, duration)
end)

COM_AddCommand("psod_weakness", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local duration = tonumber(arg)*TICRATE
	duration = $ > 0 and $ or TICRATE*90
	
	p.pso.rweakness = $ and $+duration or duration
end)

COM_AddCommand("psod_brink", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local duration = tonumber(arg)*TICRATE
	duration = $ > 0 and $ or TICRATE*90
	
	p.pso.rbrink = $ and $+duration or duration
end)