-- basic linear interpolation. Percent must be between 0 and FRACUNIT


PSO.lib.lerp = function(start, finish, percent)
	return start + FixedMul(finish-start, FRACUNIT-percent)
end

-- for some reason we don't have arctan in srb2 so define it here lol?
rawset(_G, "atan", function(x)
  return asin(FixedDiv(x, FixedSqrt(FRACUNIT + FixedMul(x, x))))
end)

-- adds \n characters to a string to word wrap it properly.
PSO.lib.wordWrap = function(v, str, width, flags, display)
	
	local w = 0
	local nstr = {}
	local fstr = ""
	
	local lastspace
	local lastw
	
	for i = 1, #str
		local s = str:sub(i, i)
		w = $ + v.stringWidth(s, flags, display)
		
		if s == " "
			lastspace = i
			lastw = w
		end
		
		nstr[#nstr+1] = s
		
		if w > width
			
			if lastspace
				nstr[lastspace] = "\n"
				w = $ - lastw
			else
				nstr[#nstr+1] = "\n"	-- yikes.
				w = 0
			end
			
			lastspace = nil
			lastw = nil
		end
	end
	
	for i = 1, #nstr
		fstr = $..nstr[i]
	end
	return fstr
end

-- for large numbers, so that 3000000 for instance is converted into a much more readable 3,000,000
PSO.lib.numCommas = function(n)
  return tostring(n):reverse():gsub("(%d%d%d)","%1,"):gsub(",(%-?)$","%1"):reverse()
end

-- yes good idea lets push uint32 to int32, itll be ok they said
-- NO IT'S NOT FUCKING OK YOU ABSOLUTE DUMBASS
PSO.lib.angToInt = function(a)
    if not a return 0 end
    a = $/ANG1
    if a < 0
        a = $+360
    end
    return a
end

PSO.lib.angDiff = function(a1, a2)
	a1 = PSO.lib.angToInt(a1)
	a2 = PSO.lib.angToInt(a2)
	
	local a = a2 - a1
	return (a + 180) % 360 - 180	
end

-- this always assumes a1 < a2 on the trig circle
PSO.lib.angBetween = function(comp, a1, a2)
	
	comp = PSO.lib.angToInt(comp)
	a1 = PSO.lib.angToInt(a1)
	a2 = PSO.lib.angToInt(a2)
	
	a2 = (a2-a1) < 0 and (a2-a1)+360 or (a2-a1)
	comp = (comp-a1) < 0 and (comp-a1)+360 or (comp-a1)
	return comp < a2
end	

PSO.lib.countPlayers = function()
	local count = 0
	for p in players.iterate
		if p.mo and p.mo.valid
			count = $+1
		end
	end
	return count
end

-- spline garbage, returns a position along a path of points.
-- This takes a fixed_t between 0 and FRACUNIT
-- ...And gives us the point on a spline.
-- Basically, fixed math hell
PSO.lib.pointOnSpline = function(path, frac)
	
	local p0, p1, p2, p3
	p1 = frac/FRACUNIT +1	-- +1 because lua is not 0-indexed
	
	-- out of bounds check
	if p1 == #path
		p2 = p1
	else	
		p2 = p1 + 1
	end	
	
	-- another out of bounds check
	if p2 == #path
		p3 = p2
	else	
		p3 = p2 + 1
	end

	-- this one is safe
	p0 = max(1, p1 - 1)
	
	frac = $ - FixedFloor($)
	
	-- cache frac^2 and frac^3
	local fsq = FixedMul(frac, frac)
	local fcb = FixedMul(fsq, frac)
	
	-- equations:
	local q1 = -fcb + FixedMul(2<<FRACBITS, fsq) - frac
	local q2 = FixedMul(3<<FRACBITS, fcb) - FixedMul(5<<FRACBITS, fsq) + 2<<FRACBITS
	local q3 = -FixedMul(3<<FRACBITS, fcb) + FixedMul(4<<FRACBITS, fsq) + frac
	local q4 = fcb - fsq
	
	-- now get the points from the dark magic above:
	--print(p0, p1, p2, p3)
	
	local tx = FixedMul(FRACUNIT>>1, FixedMul(path[p0].x, q1) + FixedMul(path[p1].x, q2) + FixedMul(path[p2].x, q3) + FixedMul(path[p3].x, q4)) 
	local ty = FixedMul(FRACUNIT>>1, FixedMul(path[p0].y, q1) + FixedMul(path[p1].y, q2) + FixedMul(path[p2].y, q3) + FixedMul(path[p3].y, q4))
	local tz = FixedMul(FRACUNIT>>1, FixedMul(path[p0].z, q1) + FixedMul(path[p1].z, q2) + FixedMul(path[p2].z, q3) + FixedMul(path[p3].z, q4)) 
	
	return {x = tx, y = ty, z = tz}
end

-- makes a distinct copy of a table and any sub table it may have
-- if norec is true, it will not recursively call itself and keep the original adresses of sub-tables in the copied table.
PSO.lib.copyTable = function(t, norec)
	
	if type(t) ~= "table"
		return t
	end
	
	local new = {}
	for k,v in pairs(t)
		if not norec and type(v) == "table"
			v = PSO.lib.copyTable(v)
		end
		
		new[k] = v
	end
	return new
end

PSO.lib.mergeTables = function(t1, t2)
	for i = 1, #t2
		t1[#t1+1] = t2[i]
	end
	return t1
end

PSO.lib.mergeTablesNoDupes = function(t1, t2)
	for i = 1, #t2
		local val = t2[i]
		local dupe
		
		for j = 1, #t1
			if val == t1[j]
				dupe = true
				break
			end
		end
		
		if not dupe
			t1[#t1+1] = val
		end
	end
	
	return t1
end

-- These are used for player save serialization
-- compacts a sparse table
PSO.lib.compactPow2Table = function(t)
	local new = {}
	for i = 0, 31
		local pow = 2^i
		if t[pow] ~= nil
			if type(t[pow]) == "table"
				new[i+1] = PSO.lib.copyTable(t[pow])
			else
				new[i+1] = t[pow]
			end
		end
	end
	
	return new
end

PSO.lib.sparsePow2Table = function(t)
	local new = {}
	
	for i = 1, #t
		new[2^(i-1)] = t[i]
	end
	return new
end

-- converts int keys of a sparse table into strings so that it may be serialized into json
PSO.lib.sparseStrTable = function(t)
	local new = {}
	
	for k,v in pairs(t)
		if type(v) == "table"
			v = PSO.lib.sparseStrTable(v)
		end
		
		new[tostring(k)] = v
	end
	return new
end

PSO.lib.sparseIntTable = function(t)
	local new = {}
	
	for k,v in pairs(t)
		if type(v) == "table"
			v = PSO.lib.sparseIntTable(v)
		end
		
		new[tonumber(k)] = v
	end
	return new
end

-- sets spritexscale and spriteyscale in accordance to the floor sprite slope of mo
PSO.lib.fixSplatScale = function(mo)
	local slopedir = mo.floorspriteslope.xydirection
	local slopeangle = mo.floorspriteslope.zangle 
	local rollangle = mo.rollangle - slopedir
	
	-- what a happy fucking mess
	local minscale = cos(slopeangle)
	
	mo.spriteyscale = FRACUNIT - FixedMul(abs(cos(rollangle)), FRACUNIT-minscale)
	mo.spritexscale = FRACUNIT - FixedMul(abs(sin(rollangle)), FRACUNIT-minscale)
	
	--print((rollangle/ANG1)..": "..mo.spritexscale..", "..mo.spriteyscale)
end

PSO.lib.spawnLine = function(source, dest, steps, sf, tics)

	local xstep = (dest.x -source.x)/steps
	local ystep = (dest.y - source.y)/steps
	local zstep = ((dest.z + dest.height/2) - (source.z + source.height/2))/steps
			
	for i = 1, steps-1
				
		local sx = source.x + xstep*i
		local sy = source.y + ystep*i
		local sz = source.z + source.height/2 + zstep*i
				
		local s = P_SpawnMobj(sx, sy, sz, MT_THOK)
		
		if sf
			sf(s)
		else
			s.sprite = SPR_HOOP
			s.frame = A|FF_FULLBRIGHT|FF_TRANS50
			s.tics = tics or 2
		end
	end
end

-- special object with MF_SOLID that can ONLY collide with gargoyles...
-- useful in gyrrey glardia for instance.
freeslot("MT_PSO_GARGOYLETESTER")
mobjinfo[MT_PSO_GARGOYLETESTER] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	height = 32<<FRACBITS,
	radius = 32<<FRACBITS,
	flags = MF_SOLID|MF_NOGRAVITY,
}

local function tester_collide(mo1, mo2)
	if mo1 and mo1.valid
	and mo2 and mo2.valid
	and mo2.type == MT_PSO_GARGOYLETESTER
		
		if mo1.type ~= MT_GARGOYLE
			return false	-- no
		elseif mo1.z+mo1.height >= mo2.z
		and mo1.z <= mo2.z + mo2.height
			return true
		end
	end
end

addHook("MobjCollide", tester_collide, MT_PSO_GARGOYLETESTER)
addHook("MobjMoveCollide", tester_collide, MT_PSO_GARGOYLETESTER)

local sightsteps = 32
PSO.lib.quickSightCheck = function(origin, dest)
	local tester = P_SpawnMobj(origin.x, origin.y, origin.z, MT_GARGOYLE)
	
	local diffx = dest.x - origin.x
	local diffy = dest.y - origin.y
	local diffz = dest.z - origin.z
	
	for i = 1, sightsteps-1
		
		local nextx = tester.x + diffx/sightsteps
		local nexty = tester.y + diffy/sightsteps
		local nextz = tester.z + diffz/sightsteps
		
		if not P_TryMove(tester, nextx, nexty, true)
			local coords = {tester.x, tester.y, tester.z}
			P_RemoveMobj(tester)
			return false, coords[1], coords[2], coords[3]
		end
		tester.z = nextz	-- uuh yeah we kinda have to
		
	end
	
	P_RemoveMobj(tester)
	return true
end

-- this assumes a centered sphere.
-- deprecated, does not work properly.
/*PSO.lib.pointOnSphere = function(pitch, yaw, hangle, distance)

	-- calculate value based solely on pitch
	local init_xoff = FixedMul(cos(pitch), FixedMul(cos(hangle), distance))
	local init_yoff = FixedMul(cos(pitch), FixedMul(sin(hangle), distance))
	local zoff = FixedMul(sin(pitch), FixedMul(sin(hangle), distance))
        
	-- now transform with the yaw
	local xoff = FixedMul(cos(yaw), init_xoff) - FixedMul(sin(yaw), init_yoff)
	local yoff = FixedMul(sin(yaw), init_xoff) + FixedMul(cos(yaw), init_yoff)
	
	return xoff, yoff, zoff
end*/

-- Positions a point on a centered sphere assuming a roll, pitch, yaw and radius.
PSO.lib.pointOnSphere = function(roll, pitch, yaw, radius)

    -- init point. For SRB2, intiialize X = radius. Normally we would use Z.
    local x, y, z = radius, 0, 0

    -- yaw
    local x_yaw = FixedMul(x, cos(yaw)) - FixedMul(y, sin(yaw))
    local y_yaw = FixedMul(x, sin(yaw)) + FixedMul(y, cos(yaw))
    local z_yaw = z

   -- pitch
    local x_pitch = FixedMul(x_yaw, cos(pitch)) + FixedMul(z_yaw, sin(pitch))
    local y_pitch = y_yaw
    local z_pitch = FixedMul(-x_yaw, sin(pitch)) + FixedMul(z_yaw, cos(pitch))

   -- roll
    local x_roll = x_pitch
    local y_roll = FixedMul(y_pitch, cos(roll)) - FixedMul(z_pitch, sin(roll))
    local z_roll = FixedMul(y_pitch, sin(roll)) + FixedMul(z_pitch, cos(roll))
	
    return x_roll, y_roll, z_roll
end

-- positions a point on a centered sphere assuming a horizontal (X/Y) angle and a vertical (Z) angle as well as a radius.
PSO.lib.anglesOnSphere = function(hangle, vangle, radius)
    -- Calculate the Cartesian coordinates
	
    local x = FixedMul(FixedMul(radius, cos(vangle)), cos(hangle))
	local y = FixedMul(FixedMul(radius, cos(vangle)), sin(hangle))
    local z = FixedMul(radius, sin(vangle))
    
    return x, y, z
end
