-- Blue Crawla
/*
	Drives towards players.
	Once they get close enough, charge for 1/4th of a second and lunge forward at higher speed, ending in lag
	
	HARD MODE: Throws projectiles if you stay out of range
	
	CORRUPTED: The projectile becomes a slow moving Megid that continuously deals damage and has a high chance of poisonning the player.
*/

-- simplified version of megid
freeslot("MT_PSO_MEGID_ENEMY")
mobjinfo[MT_PSO_MEGID_ENEMY] = {
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 16<<FRACBITS,
	height = 16<<FRACBITS,
	flags = MF_NOGRAVITY|MF_MISSILE,
}

addHook("MobjThinker", function(mo)
	
	mo.sprite = SPR_MEGD
	mo.frame = (leveltime%32 /2)|TR_TRANS20|FF_SUBTRACT|FF_FULLBRIGHT
	mo.color = SKINCOLOR_MASTER
	
	-- krilling myself
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	if leveltime & 1
		local g = P_SpawnGhostMobj(mo)
		g.color = SKINCOLOR_BLUE
		g.momx = P_RandomRange(-2, 2)*mo.scale
		g.momy = P_RandomRange(-2, 2)*mo.scale
		g.momz = P_RandomRange(-4, 0)*mo.scale
		g.destscale = 1
		g.fuse = 12
	end
	
	
	if not mo.passive
		if leveltime% (TICRATE/2) == 0
			local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z-mo.height, 64<<FRACBITS, 64<<FRACBITS, mo.target, mo.pso.potency, DMT_TECHNIQUE, TICRATE/2, RT_NONE, COND_CORRUPTION, 5)
			hb.pso.combobreaker = mo.pso.combobreaker
			hb.pso.element = mo.pso.element
			hb.pso.potency = mo.pso.potency
		end
	end	
end, MT_PSO_MEGID_ENEMY)

addHook("MobjCollide", function(mo1, mo2) return false end, MT_PSO_MEGID_ENEMY)
addHook("MobjMoveCollide", function(mo1, mo2) return false end, MT_PSO_MEGID_ENEMY)

PSO.info.enemies[MT_BLUECRAWLA] = {
		
		name = "Blue Crawla",
		sprite = SPR_POSS,
		frame = A,
		painframe = A,
		
		bhp = 200,
		satk = 30,
		ratk = 20,
		tatk = 20,
		def = 25,
		exp = 5,
		weight = 1,
		
		
		drops = {
			
			-- Lv 1-10
			{100, 1, 9, 1, 15},				-- meseta (100% mult) 2/5 (Lv1-10)
			{"sword", 1, 9, 1, 100},		-- sword: 	1/100 (Lv1-10)
			{"rifle", 1, 9, 1, 100},		-- rifle: 	1/100 (Lv1-10)
			{"rod", 1, 9, 1, 100},			-- rod: 	1/100 (Lv1-10)
			{"lockgun", 1, 9, 1, 100},
			{"tcsm_crawla1", 1, 200, 1, 140},
			
			{"monomate", 1, 9, 1, 25},
			
			-- Lv10-19
			
			{200, 10, 19, 1, 15},
			{"gigush", 10, 19, 1, 100},
			{"beam", 10, 19, 1, 100},
			{"cane", 10, 19, 1, 100},
			{"monomate", 10, 19, 1, 25},
			{"autogun", 10, 19, 1, 100},

			-- Lv20-29
			{300, 20, 29, 1, 15},
			{"breaker", 20, 29, 1, 150},
			{"doublegladius", 20, 29, 1, 150},
			{"laser", 20, 29, 1, 150},
			{"canon", 20, 29, 1, 150},
			{"ulrod", 20, 29, 1, 150},
			{"peacekeeper", 20, 29, 1, 100},
			{"dimate", 20, 29, 1, 25},

			-- Lv30-50
			{500, 30, 50, 1, 15},
			{"peacekeeper", 30, 50, 1, 50},
			{"alva peacekeeper", 30, 50, 1, 100},
			{"vita peacekeeper", 30, 50, 1, 150},
			{"poss shooter", 30, 50, 1, 250},
			{"tcsm_crawla3", 30, 50, 1, 250},
			{"rostig frame", 30, 50, 1, 60},
			{"fehler frame", 30, 50, 1, 200},
			{"hart frame", 30, 50, 1, 600},
			{"dimate", 30, 50, 1, 35},
			
			-- Lv51-70
			{550, 51, 200,1, 15},
			{"vita peacekeeper", 51, 79,1, 60},
			{"poss shooter", 51, 200,1, 150},
			{"tcsm_crawla3", 51, 79,1, 150},
			{"baston frame", 51, 79,1, 200},
			{"makel frame", 51, 79, 1, 800},
			{"crimson frame", 51, 79, 1, 800},
			{"yasminkov 2000h", 51, 79, 1, 400},
			{"sarcueid", 51, 79, 1, 400},
			{"bravace", 51, 79, 1, 700},
			{"varista", 51, 79, 1, 700},			
			{"rostig frame", 51, 79,1, 30},
			{"fehler frame", 51, 79,1, 80},
			{"dimate", 51, 79 ,1, 35},
			
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_BADNIK,
		
		weakness = ELEM_FIRE,
		weakness2 = ELEM_ELEC,
		
		reset =	function(mo)
			mo.actiontimer = 0	-- reset action if we get staggered etc.
			mo.outranged = 0
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			
			local attack_potency = 250
			local move_speed = mo.scale*3/2 * (diff >= DIFF_HARD and 2 or 1)
			local attack_range = mo.scale*80
			local attack_charge = TICRATE/2
			local attack_delay = TICRATE
			local attack_speed = mo.scale*14
			local anim_speed = 3 
			
			local shoot_wait = TICRATE*5
			local shoot_charge = TICRATE*3/4
			local shoot_knockback = TICRATE*3/4
			
			
			if mo.pso.shifta
				move_speed = $*12/10	-- +20% movement speed w shifta
			end
			
			PSO.lib.enemySetCorruption(mo)
			--mo.pso.corrupted = true
			
			PSO.lib.enemyManageAggro(mo)
			
			mo.sprite = SPR_POSS
			mo.frame = A
			
			if mo.target and mo.target.valid
				
				if not mo.actiontimer
				and not mo.shooting
					A_FaceTarget(mo)
					P_InstaThrust(mo, mo.angle, move_speed)
					
					mo.frame = (leveltime%(anim_speed*6)) /anim_speed
					
					--print(mo)
					if R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) < attack_range
						mo.actiontimer = 1
						mo.outranged = 0
						S_StartSound(mo, sfx_spndsh)
					elseif diff >= DIFF_HARD
						mo.outranged = $ and $+1 or P_RandomRange(1, TICRATE)
						if mo.outranged > shoot_wait
							mo.shooting = 1
							mo.outranged = 0
						end
					end
				elseif mo.shooting
					
					mo.shooting = $+1
					
					if mo.shooting < shoot_charge
						A_FaceTarget(mo)
						P_InstaThrust(mo, 0, 0)
					elseif mo.shooting == shoot_charge
						
						P_InstaThrust(mo, mo.angle, -FRACUNIT*16)
						S_StartSound(mo, sfx_trfire)
						
						if not mo.pso.corrupted
							local msl = PSO.lib.spawnMissile(mo, MT_JETTBULLET, mo.angle, 30*FRACUNIT, attack_potency, DMT_RANGED, 0, 0, mo.height/2, RT_NONE)
							PSO.lib.aimMissile(msl, mo.target)	
						else
							local msl = PSO.lib.spawnMissile(mo, MT_PSO_MEGID_ENEMY, mo.angle, 4*FRACUNIT, attack_potency*2/3, DMT_TECHNIQUE, 0, 0, mo.height/2, RT_NONE)
							PSO.lib.aimMissile(msl, mo.target)
							msl.pso.element = ELEM_DARK
							msl.fuse = TICRATE*10
						end
						
						for i = 1, 8
							local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
							d.angle = mo.angle + P_RandomRange(-30, 30)*ANG1
							P_InstaThrust(d, d.angle, P_RandomRange(2, 16)*FRACUNIT)
						end
						
					elseif mo.shooting == shoot_charge+shoot_knockback
						mo.shooting = nil
					end
					
				elseif mo.actiontimer
					mo.actiontimer = $+1
					
					if mo.actiontimer == attack_charge
						P_InstaThrust(mo, mo.angle, attack_speed)
						S_StartSound(mo, sfx_zoom)
						
						local hx = mo.x + 64*cos(mo.angle)
						local hy = mo.y + 64*sin(mo.angle)
						
						local hb = PSO.lib.spawnHurtBox(hx, hy, mo.z, 80<<FRACBITS, 40<<FRACBITS, mo, attack_potency, DMT_STRIKE, 3, RT_STAGGER)
						hb.angle = mo.angle
						P_InstaThrust(hb, hb.angle, attack_speed)
						
					elseif mo.actiontimer > attack_charge
					and mo.actiontimer < attack_delay
						mo.momx = $*98/100
						mo.momy = $*98/100
					
					elseif mo.actiontimer >= attack_delay
						mo.actiontimer = 0
					end
				end
				
				if PSO.lib.enemyPitCheck(mo)
					P_InstaThrust(mo, 0, 0)
				end
				
			end
		end,	
}

--print("Best test: "..PSO.lib.findBestEnemyDrop(MT_BLUECRAWLA, 40))