freeslot("MT_PSO_BUFFPOD")

-- Buff pod
/*
	Gives buffs to enemies.
	
	Gives buffs every 10 or so seconds to enemies. Buff depends on flags.
	Hard mode: Immediately gives buffs to enemies after spawning and then reapplies it periodically.
	
	default: grants shifta to all enemies within the room
	extra: grants deband instead of shifta
	special: grants both
	ambush: turns the buffs towards enemies to debuffs towards players
	
	
*/

mobjinfo[MT_PSO_BUFFPOD] = {
	doomednum = 2700,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	height = 48<<FRACBITS,
	radius = 32<<FRACBITS,
	flags = MF_NOGRAVITY|MF_ENEMY|MF_SPECIAL|MF_SHOOTABLE,
}

PSO.info.enemies[MT_PSO_BUFFPOD] = {
		
		name = "S-POD",
		sprite = SPR_GWLR,
		frame = A,
		painframe = A,
		
		bhp = 120,
		satk = 10,
		ratk = 10,
		tatk = 35,
		def = 25,
		exp = 4,
		weight = 4,
		
		
		drops = {
			{300, 1, 29, 1, 15},
			{"breaker", 1, 29, 1, 150},
			{"doublegladius", 1, 29, 1, 150},
			{"laser", 1, 29, 1, 150},
			{"canon", 1, 29, 1, 150},
			{"ulrod", 1, 29, 1, 150},
			{"dimate", 1, 29, 1, 25},
			{"tcsm_unidus", 1, 200, 1, 75},

			-- Lv30-50
			{650, 30, 50, 1, 15},
			{"scepter", 30, 50, 1, 50},
			{"alva scepter", 30, 50, 1, 75},
			{"vita scepter", 30, 50, 1, 100},
			{"asterod", 30, 50, 1, 200},
			{"rostig frame", 30, 50, 1, 30},
			{"fehler frame", 30, 50, 1, 100},
			{"permafrost frame", 51, 200,1, 250},
			{"hart frame", 30, 50, 1, 300},
			{"digrinder", 30, 50, 1, 25},
			{"dimate", 30, 50, 1, 35},	

			{700, 30, 50, 1, 15},
			{"vita scepter", 51, 200,1, 50},
			{"crimson frame", 51, 79, 1, 700},
			{"windmill", 51, 79, 1, 200},
			{"hexigraph", 51, 79, 1, 200},
			{"serdote", 51, 79, 1, 400},
			{"red processor", 51, 79, 1, 400},
			{"clarita visas", 51, 79, 1, 200},
			{"battle verge", 51, 79, 1, 200},
			{"brave hammer", 51, 79, 1, 400},
			{"rose quartz", 51, 79, 1, 400},			
			{"rostig frame", 51, 200,1, 30},
			{"fehler frame", 51, 200,1, 80},
			{"dimate", 51, 200,1, 35},	
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_BADNIK,
		weakness = ELEM_ICE,
		weakness2 = ELEM_FIRE,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			local range = 2048
			local buff_delay = diff >= DIFF_HARD and 5 or 10
			local buff_duration = diff >= DIFF_HARD and 40 or 20
			local targetz = mo.floorz + mo.scale*32
			local rate = FRACUNIT*8/10
			local c1, c2

			mo.flags = $ & ~(MF_FLOAT|MF_NOGRAVITY)
			mo.colorized = true
			mo.sprite = SPR_GWLR
			mo.frame = A
			
			if not mo.styleroll
				if mo.z < targetz
					mo.momz = $ + rate
				elseif mo.momz < 0
					mo.momz = min(-mo.scale, $ + rate*2)
				end
			end
			
			local spawnpoint = mo.spawnpoint
			
			-- initialization
			if not mo.extravalue1
				
				mo.atkbuff = true
				
				if mo.options
					
					if mo.options & 1
						mo.atkbuff = false
						mo.defbuff = true
					end
					
					if mo.options & 4
						mo.atkbuff = true
						mo.defbuff = true
					end
					
					if mo.options & 8
						mo.reversebuff = true
					end
				end
				mo.extravalue1 = 1
				
				mo.extravalue2 = TICRATE*buff_delay/2
			end
			
			-- spark colours...
			if mo.reversedebuff
				if mo.atkbuff and mo.defbuff
					c1 = SKINCOLOR_CRIMSON
					c2 = SKINCOLOR_DUSK
					mo.pso.name = "Jallure Pod"
				elseif mo.atkbuff
					c1 = SKINCOLOR_CRIMSON
					mo.pso.name = "Jellen Pod"
				elseif mo.defbuff
					c1 = SKINCOLOR_DUSK
					mo.pso.name = "Zalure Pod"
				end
			else
				if mo.atkbuff and mo.defbuff
					c1 = SKINCOLOR_RED
					c2 = SKINCOLOR_BLUE
					mo.pso.name = "Shiftaband Pod"
				elseif mo.atkbuff
					c1 = SKINCOLOR_RED
					mo.pso.name = "Shifta Pod"
				elseif mo.defbuff
					c1 = SKINCOLOR_BLUE
					mo.pso.name = "Deband Pod"
				end
			end
			
			mo.color = c1
			
			-- runtime
			mo.extravalue2 = $-1
			
			if mo.extravalue2 < TICRATE*3
				for i = 1, 2
					local t = P_SpawnMobj(mo.x + P_RandomRange(-128, 128)*mo.scale, mo.y + P_RandomRange(-128, 128)*mo.scale, mo.z + P_RandomRange(-128, 128)*mo.scale, MT_THOK)
					t.tics = 10
					t.color = i & 1 and c2 or c1
					t.frame = A|FF_ADD
					t.scale = mo.scale
					t.destscale = 1
					t.scalespeed = mo.scale/12		
					t.momx = (mo.x - t.x)/10
					t.momy = (mo.y - t.y)/10
					t.momz = (mo.z - t.z)/10				
				end
			end
			
			if not mo.extravalue2
				
				mo.extravalue2 = TICRATE*buff_delay
				
				
				-- debuff nearby players
				if mo.reversebuff
					S_StartSound(mo, sfx_kc59)
					
					for p in players.iterate do
						local pmo = p.mo
						
						if pmo and pmo.valid
						and R_PointToDist2(pmo.x, pmo.y, mo.x, mo.y) < range*mo.scale	
							
							if mo.atkbuff
								PSO.lib.addJellen(pmo, TICRATE*buff_duration)
							end
							if mo.defbuff
								PSO.lib.addZalure(pmo, TICRATE*buff_duration)
							end
						end
						
					end
				
				-- buff nearby enemies
				else
					
					if mo.atkbuff == true and mo.defbuff
						S_StartSound(mo, sfx_shifta)
						S_StartSound(mo, sfx_deband)
					elseif mo.atkbuff
						S_StartSound(mo, sfx_shifta)
					elseif mo.defbuff
						S_StartSound(mo, sfx_deband)
					end
					
					local room = server.psorooms[mo.roomid]
					if room and room.objects
						for k,v in pairs(room.objects)
							if v and v.valid and v.flags & MF_ENEMY and v ~= mo
							and v.pso and v.pso.hp
								
								if mo.atkbuff
									PSO.lib.addShifta(v, TICRATE*buff_duration)
								end
								
								if mo.defbuff
									PSO.lib.addDeband(v, TICRATE*buff_duration)
								end
							end
						end
					end
				end

				
				for i = 1, 16
					local ang = (360/16)*i*ANG1
					
					local s = P_SpawnMobj(mo.x, mo.y, mo.floorz, MT_SUPERSPARK)
					s.color = i & 1 and c2 or c1
					s.colorized = true
					
					P_InstaThrust(s, ang, FRACUNIT*32)
				end
				
			end
		end
}