-- enemy definitions
-- all stats are at level 1
-- enemies will automatically scale up with enemy level / difficulty

-- dummy state for enemies to go in so we won't be bothered by external behaviours.
freeslot("S_PSO_ENEMYDUMMY", "SPR_INDC")
states[S_PSO_ENEMYDUMMY] = {SPR_NULL, A, -1, nil, 0, 0, S_PSO_ENEMYDUMMY}

-- enemy "categories"
local enemy_attributes = {
	"ATTR_BADNIK",
	"ATTR_HARBINGER",
	"ATTR_ABERRATION",
	"ATTR_SUPERNATURAL",
}

for i = 1, #enemy_attributes
	rawset(_G, enemy_attributes[i], i)
end

PSO.info.attributeNames = {
	"Badnik",
	"Harbinger",
	"Aberration",
	"Supernatural",
}


PSO.info.enemies = {}
PSO.info.enemyHPScaling = {
	[DIFF_NORMAL] = 100,
	[DIFF_HARD] = 130,	-- +30% HP in hard (enemy damage is FINE, they just are a bit too squishy!)
	[DIFF_VERYHARD] = 140,	-- we'll see how that ends up working lol.
}


-- build lookup table for stats scaling:
local stat_scaling_res = 100
local stat_scaling_start = 108
local stat_scaling = {stat_scaling_res}

for i = 2, 255
	local start = stat_scaling[i-1] * stat_scaling_start
	stat_scaling[i] = stat_scaling[i-1] * stat_scaling_start / min(stat_scaling_start-1, (stat_scaling_res + (i/10) + (i/30)))
	--print(stat_scaling[i])
	--print("scale ("..i.."): "..stat_scaling[i])
end

PSO.info.enemyStatScaling = stat_scaling
PSO.info.enemyStatScalingRes = stat_scaling_res

-- mapload: give stats to all the enemies
PSO.lib.setupEnemies = function()
	--print("setupEnemies")
	-- count players at setup time...
	local numplayers = 0
	for p in players.iterate do
		numplayers = $+1
	end
	
	for m in mobjs.iterate()	-- why does this one need ()
		
		if m and m.valid and m.flags & MF_ENEMY
			PSO.lib.setupEnemy(m, numplayers)
		end
	end
end

PSO.lib.setupEnemy = function(m, numplayers)
	if not numplayers
		numplayers = 0
		for p in players.iterate do
			if not p.bot
				numplayers = $+1
			end
		end	
	end
	
	m.pso = {
			
		maxhp = 100,
		hp = 100,
		satk = 20,
		ratk = 20,
		tatk = 20,
		def = 15,
		exp = 3,
		weakness = ELEM_FIRE,
	}
			
	local lv = 1

	-- get level if applicable
	if server.questdata
				
		local q = PSO.info.quests[server.questdata.section].quests[server.questdata.index]
		lv = q.levels[server.questdata.difficulty] or server.pso.level
	end
	
	if server.psoenemylv	-- debug cheats
		lv = min(255, server.psoenemylv)
	end
			
	-- now apply the base stats and level scaling!
	-- we will also scale HP specifically with playercount
	
	if PSO.info.enemies[m.type]
		local info = PSO.info.enemies[m.type]
				
		m.pso.maxhp = info.bhp *stat_scaling[lv]/stat_scaling_res
		m.pso.basemaxhp = m.pso.maxhp	-- used for meseta drops calculation
		
		if numplayers > 1
			if info.boss
				m.pso.maxhp = $ + m.pso.maxhp*numplayers/2
			else
				m.pso.maxhp = $ + m.pso.maxhp*numplayers/3
			end
		end
		
		-- now enemy HP scaling...
		local diff = PSO.lib.questDifficulty()
		local scaling = not PSO.lib.arenaQuestActive() and PSO.info.enemyHPScaling[diff] or 100
		--print("HP pre scaling: "..m.pso.maxhp)
		m.pso.maxhp = $*scaling/100
		--print("HP post scaling: "..m.pso.maxhp)
		
		-- track how much damage we take, this will be used to gauge aggro...
		m.pso.damagetrackers = {}
		for i = 1, 32
			m.pso.damagetrackers[i] = 0
		end
		
		m.pso.hp = m.pso.maxhp
		m.pso.satk = info.satk *stat_scaling[lv]/stat_scaling_res
		m.pso.ratk = info.ratk *stat_scaling[lv]/stat_scaling_res
		m.pso.tatk = info.tatk *stat_scaling[lv]/stat_scaling_res
		m.pso.def = (info.def *stat_scaling[lv]/stat_scaling_res)/3
		m.pso.weight = info.weight or 1
		m.pso.exp = (info.exp or $) *stat_scaling[lv]/stat_scaling_res
		m.pso.detectrange = info.range or 4096
		
		if (info.corruptionchance ~= nil)
			m.pso.corruptionchance = info.corruptionchance
		else
			m.pso.corruptionchance = 25	-- default corruption chance is 25%
		end
		
		m.pso.weakness = info.weakness or $
		m.pso.weakness2 = info.weakness2 or nil	-- this can be nil and it's okay if it is.
		m.pso.attribute = info.attribute or ATTR_SUPERNATURAL	-- if we have no attribute just pretend it's a supernatural or some shit lol
		m.pso.level = lv
		m.pso.boss = info.boss
		m.pso.waitclear = info.waitclear
		
		-- this prevents death pit collision. yep.
		if info.boss
			m.flags2 = $|MF2_BOSSDEAD
		end
		
		-- validate drops!
		if info.drops
			for i = 1, #info.drops
				local d = info.drops[i]
				
				local name = d[1]
				
				if type(name) ~= "number"
					
					if not PSO.info.items[name]
						print("\x82".."WARNING: ".."\x80".."Enemy "..info.name..": Invalid drop \'"..name.."\'")
					end
				end
			end
		end
	end
end

PSO.lib.spawnEnemyFromEnemy = function(ref, x, y, z, newtype)
	local mo = P_SpawnMobj(x, y, z, newtype)
	mo.flags = MF_NOCLIP|MF_NOGRAVITY
	PSO.lib.setupEnemy(mo)
	
	-- add that enemy in the same room as the ref
	table.insert(server.psorooms[ref.roomid].waves[ref.waveid], mo)
	
	local tin = P_SpawnMobj(mo.x, mo.y, mo.z, MT_PSO_TELEPORTIN)
	tin.target = mo
	
	return mo
end

/*
	Enemy stats:
		
	name:
		Enemy display name
	
	sprite:
		SPR_ sprite to use by default. This can be overwritten at runtime but is used by some default functions.
	
	frame:
		frame to use by default. This can be overwritten at runtime but is used by some default functions.
		
	painframe:
		frame used by default functions for stagger, knockback and blowdown.
		Will fallback to 'frame' if it doesn't exist. Can also be left nil if you overwrite all behaviours.
	
	bhp:
		Enemy max HP at Lv1
	
	satk:
		Enemy striking attack at Lv1
	
	ratk:
		Enemy ranged attack at Lv1
		
	tatk:
		Enemy technique attack at Lv1
	
	def:
		Enemy def at Lv1
	
	exp:
		Enemy EXP at Lv1
		
	weakness:
		Enemy weakness
		
	drops:
		Enemy drop table
		{
			{"item", min. level, max. level, ratelow, ratehigh},
				^ "item" drops if enemy is between min. level and max. level at a rate of ratelow/ratehigh
				
			{"item2", .....},
			{...},
			...
		}
		
	weight:
		Enemy weight class, determines if an enemy can be staggered, knocked back and blown down
		1: Lightweight:
			- Stagger: Yes
			- Knockback: Yes
			- Blowdown: Yes
	
		2: Normal:
			- Stagger: No
			- Knockback -> Stagger
			- Blowdown -> Knockback
			
		3: Heavyweight:
			- Stagger: No
			- Knockback: No
			- Blowdown -> Stagger
		
		4: Boss:
			- Stagger: No
			- Knockback: No
			- Blowdown: No
		
		
	thinker: function(mobj_t mo)
		Function ran each frame, calling mo as the enemy argument.
		This function does not run during stagger, knockback or blowdown.
	
	stagger: function(mobj_t mo)
		Function to call when enemy is staggered. 
		Enemy stagger stops when the function returns true.
		If nil, uses a default behaviour.
		
	knockback: function(mobj_t mo)
		Function to call when enemy is knocked back.
		Enemy knockback ends when the function returns true.
		If nil, uses a default behaviour.
		
	blowdown: function(mobj_t mo)
		Function to call when the enemy is blown down.
		Enemy blowdown ends when the function returns true.
		If nil, uses a default behaviour.
*/	

-- drop items after killing an enemy.
PSO.lib.enemyDropItems = function(mo)
	
	if mo.nodrops return end	-- used by gyrrey glardia specifically for the first clear lol.
	
	local info = PSO.info.enemies[mo.type]
	local boxdrops
	local boxes
	local drops
	
	if mo.options & 1	-- drop items from boxes exclusively
		--print("mo is a boss")
		-- check if we CAN
		local room = server.psorooms[mo.roomid]
		
		if room and room.boxes and room.boxes[room.wavenum] and #room.boxes[room.wavenum]
			boxes = room.boxes[room.wavenum]
			--print("boxes "..tostring(boxes))
		end
		
		-- then just drop nothing lol
		if not boxes
			return
		end
	end
	
	local soundflag	-- used to know if we should mute boxes that spawn
	if info and info.drops
		if boxes
			for i = 1, #boxes
				local box = boxes[i]
				if not (box.options & 4)
					box.drops = PSO.lib.calcDrops(info.drops, mo.pso.level)
				end
				--box.state = S_PSO_CRATE	-- this will make them appear
				--print(box.drops)
				local tpin = P_SpawnMobj(box.x, box.y, box.z, MT_PSO_TELEPORTIN)
				tpin.target = box
				tpin.nopush = true
				tpin.mutesound = soundflag
				soundflag = true
			end
		else
			drops = PSO.lib.calcDrops(info.drops, mo.pso.level)
		end
	end
	
	-- iterate drops and check for meseta amount, if so, make sure to scale it apropriately.
	if drops
		for k,v in pairs(drops)
			if type(v) == "number"	-- meseta
				--local mult = FixedMul( FixedSqrt((mo.pso.level or 1)<<FRACBITS), FixedSqrt((mo.pso.maxhp or 1)<<FRACBITS) /FRACUNIT
				local mult = max(1, (mo.pso.basemaxhp)/8)
				local meseta = (mult/3)*v/100
				meseta = $ + P_RandomRange(-meseta/10, meseta/10)	-- variance
				PSO.lib.dprint(meseta)
				drops[k] = meseta
			end
		end
		
		local pos = {mo.x, mo.y, mo.z}
		if mo.pitdeath
			pos = mo.pso.pitcoords
		end
		
		local drop = P_SpawnMobj(pos[1], pos[2], pos[3] + mo.height/2, MT_PSO_DROP)
		local ang = P_RandomRange(0, 359)*ANG1
		--P_InstaThrust(drop, ang, P_RandomRange(0, 4)*drop.scale)
		drop.momz = drop.scale*8
		drop.psoitems = drops
		
		if consoleplayer
			local mydrop = drop.psoitems[#consoleplayer+1]
			if mydrop
			and type(mydrop) ~= "number"
				local minfo = PSO.lib.itemInfo(mydrop)
				if minfo.rarity >= 7
					PSO.lib.startRareDropAnim(consoleplayer, minfo.rarity)
				end
			end
		end
	end
end

-- finds the best available drop for that particular enemy type
PSO.lib.findBestEnemyDrops = function(etype, level, itype)
	
	--local maxrarity = 0
	local bestitems = {}
	local info = PSO.info.enemies[etype]
	
	if info and info.drops
		for i = 1, #info.drops
			local it = info.drops[i]
			
			if level >= (it[2] or 1)
			and level <= (it[3] or 9999)
			and type(it[1]) ~= "number"	-- not meseta lol
				local iteminfo = PSO.lib.itemInfo(it[1])				
				if iteminfo and iteminfo.rarity and (itype == nil or iteminfo.type == itype)
					bestitems[#bestitems+1] = it[1]
				end
			end
		end
	end
	
	-- now sort
	table.sort(bestitems, function(a, b)
							local info1 = PSO.lib.itemInfo(a)
							local info2 = PSO.lib.itemInfo(b)
							return info1.rarity > info2.rarity
						end)
	
	return bestitems
end

-- Manage the enemy's target depending on aggro.
-- Aggro changes depending on damage dealt
PSO.lib.enemyManageAggro = function(mo)
	
	local bestdist = mo.pso.detectrange*mo.scale
	
	if not mo.target or not mo.target.valid
	or R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) > bestdist	-- blud is too far!
	or (mo.target.player and mo.target.player.pso and mo.target.player.pso.hp <= 0)	-- blud is dead
	or (leveltime % (TICRATE*5) == 0 and not mo.pso.targetlock)
		
		local prevtarget = mo.target
		
		local bestdamagetarget
		local bestdamage = 0
		
		local closesttarget
		
		local tester = P_SpawnMobj(mo.x, mo.y, mo.z+FRACUNIT*256, MT_THOK)	-- some enemies need this
		
		local notarget = (not mo.target or not mo.target.valid)
		for p in players.iterate do
			if p.mo and p.mo.valid
			and p.pso
			and p.pso.hp
			and p.pso.type_initialized	-- don't attempt to target new joiners, they're in limbo.
			and R_PointToDist2(mo.x, mo.y, p.mo.x, p.mo.y) < mo.pso.detectrange*mo.scale
			and (not notarget or P_CheckSight(mo, p.mo) or P_CheckSight(tester, p.mo))
				
				local dist = R_PointToDist2(mo.x, mo.y, p.mo.x, p.mo.y)
				if dist < bestdist
				and (PSO.info.enemies[mo.type] and PSO.info.enemies[mo.type].skipsight or P_CheckSight(mo, p.mo))
					closesttarget = p.mo
					bestdist = dist
				end
				
				if mo.pso.damagetrackers[#p+1] > bestdamage
					bestdamage = mo.pso.damagetrackers[#p+1]
					bestdamagetarget = p.mo
				end
			end
		end
		
		P_RemoveMobj(tester)
		
		mo.target = nil
		
		-- pick a new target, priotirizing targets with the highest total damage dealt on us!
		mo.target = bestdamagetarget or closesttarget
		
		if not mo.target and prevtarget and R_PointToDist2(mo.x, mo.y, prevtarget.x, prevtarget.y) < 4096*mo.scale
			mo.target = prevtarget
		end
	end
end

local function downSparkAnim(mo)
	local an = leveltime * ANG1 * 8
	for i = 1, 2 do
		local dx = mo.x + 64*cos(an)
		local dy = mo.y + 64*sin(an)
		local dz = mo.z + mo.height + 12*sin(an + leveltime*ANG1*4)
		local s = P_SpawnMobj(dx, dy, dz, MT_THOK)
		s.state = S_SPRK1
		s.scale = FRACUNIT*2
		s.destscale = 0
		s.frame = A|FF_FULLBRIGHT
		s.tics = 1
		if (leveltime%2)
			s.fuse = 2
		end
		an = $+ ANG1*180
	end
end

PSO.lib.enemyStaggerDefault = function(mo)
	
	local info = PSO.info.enemies[mo.type]
	local stagger_duration = info.stagger_duration or TICRATE/3
	
	/*if mo.pso.zalure
		stagger_duration = $ + stagger_duration*PSO.info.ZALURE/100
		downSparkAnim(mo)
	end*/
	
	mo.rollangle = 0
	mo.sprite = info.sprite
	if info.painframe ~= nil
		mo.frame = info.painframe
	else
		mo.frame = info.frame or A
	end
	
	if mo.pso.stagger == 1
		mo.momz = mo.scale*2
	end
	
	mo.pso.stagger = $+1
	if mo.pso.stagger > stagger_duration
		mo.pso.stagger = 0
	end
end

PSO.lib.enemyKnockbackDefault = function(mo)
	
	local info = PSO.info.enemies[mo.type]
	
	mo.rollangle = 0
	
	if not P_IsObjectOnGround(mo)
		mo.momx = FixedMul(mo.momx, mo.friction)
		mo.momy = FixedMul(mo.momy, mo.friction)
	else
		local rad = mo.radius/FRACUNIT
		local d = P_SpawnMobj(mo.x + P_RandomRange(-rad, rad)*FRACUNIT, mo.y + P_RandomRange(-rad, rad)*FRACUNIT, mo.z, MT_SPINDUST)
		d.momz = P_RandomRange(1, 4)*mo.scale
	end
	
	mo.momx = FixedMul(mo.momx, mo.friction)
	mo.momy = FixedMul(mo.momy, mo.friction)
	
	if FixedHypot(mo.momx, mo.momy) < mo.scale
		--print(mo.pso.knockback)
		--if mo.pso.zalure	-- zalure has enemies get fucked over
		--	mo.pso.knockback = max(0, $-1)
		--	downSparkAnim(mo)
		--else
		mo.pso.knockback = 0
		--end
	else
		mo.pso.knockback = min($+1, info.stagger_duration*3/2 or TICRATE*3/4)
	end
end

PSO.lib.enemyBlowdownDefault = function(mo)
	
	local info = PSO.info.enemies[mo.type]
	
	if mo.pso.blowdown == 1
		mo.z = $+1
	end
	
	if not P_IsObjectOnGround(mo)
		mo.blowdownmomz = mo.momz
		mo.rollangle = $ + ANG1*16
		mo.z = $ + mo.momz
		mo.pso.blowdown = min($+1, info.stagger_duration*3/2 or TICRATE*3/4)
	else	--elseif mo.pso.blowdown > 4
		
		if FixedHypot(mo.momx, mo.momy) < mo.scale*4
			/*if mo.pso.zalure
				mo.pso.blowdown = $-1
				if mo.pso.blowdown == 1
					mo.pso.blowdown = 0
				end
				downSparkAnim(mo)
			else*/
			mo.pso.blowdown = 0
			--end
			mo.rollangle = 0
		else
		
			mo.momz = -mo.blowdownmomz/3
			mo.momx = $/2
			mo.momy = $/2
			
			for i = 1, 8
				local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
				d.angle = (360/8)*i*ANG1
				P_InstaThrust(d, d.angle, mo.blowdownmomz*3)
				d.momz = -mo.blowdownmomz/3
			end	
		end
	end
end

-- calls enemy thinkers and other things

local status_funcs = {
	{"stagger", PSO.lib.enemyStaggerDefault},
	{"knockback", PSO.lib.enemyKnockbackDefault},
	{"blowdown", PSO.lib.enemyBlowdownDefault},
}

PSO.lib.enemyThinker = function(mo)
	
	local info = PSO.info.enemies[mo.type]
	if info and mo.pso.hp and mo.health
		
		local ran_reset
		mo.flags = $ & ~MF_SOLID
		--mo.z = mo.z
		
		-- do nothing.
		if server.psodps
			mo.sprite = info.sprite
			mo.frame = info.sprite
			mo.momx = 0
			mo.momy = 0
			mo.momz = 0
			return
		end
		
		if not mo.pso.pitcoords
			mo.pso.pitcoords = {mo.x, mo.y, mo.z}
		end
		
		-- check for death pit FOFs
		-- for some reason these arent normally accounted for.
		--if P_MobjTouchingSectorSpecial(mo, 1, 7)
		local mysec = mo.subsector.sector
		
		for rover in mysec.ffloors()
			
			if not info.boss and not info.part
			and mo.z <= rover.topheight
			and mo.z >= rover.bottomheight
			and GetSecSpecial(rover.sector.special, 1) == 7	-- death pit
				P_KillMobj(mo, nil, nil, DMG_DEATHPIT)
				return
			end
		end
		
		for i = 1, #status_funcs
			local s = status_funcs[i]
			if mo.pso[s[1]]
				
				if not ran_reset
					if info.reset
						local run, res = pcall(info.reset, mo)

						if not run
							print("\x82".."WARNING".."\x80"..": enemyThinker (reset): "..res) 
						end
					end
				end
				
				local fn = info[s[1]] or s[2]
				local run, res = pcall(fn, mo)
					
				if not run
					print("\x82".."WARNING".."\x80"..": enemyThinker ("..s[1].."): "..res) 
				end

				PSO.lib.handleStatus(mo)
				if mo.pso.corrupted
					PSO.lib.enemyCorruptionVisual(mo)
					if mo.parts
						for i = 1, #mo.parts
							if mo.parts[i] and mo.parts[i].valid
								PSO.lib.enemyCorruptionVisual(mo.parts[i])
							end
						end
					end
				end
				
				return
			end
		end
		
			
		if info.thinker
			local run, res = pcall(info.thinker, mo)
				
			if not run
				print("\x82".."WARNING".."\x80"..": enemyThinker: "..res) 
			end
		end
		
		PSO.lib.handleStatus(mo)
			
		if mo.pso.corrupted
			PSO.lib.enemyCorruptionVisual(mo)
			if mo.parts
				for i = 1, #mo.parts
					if mo.parts[i] and mo.parts[i].valid
						PSO.lib.enemyCorruptionVisual(mo.parts[i])
					end
				end
			end
		end
		
		-- bosses and music: play music for players that get near us!
		if info.boss and info.bossmusic
			
			local myroom = mo.roomid
			local room = server.psorooms[myroom]
			
			if room
				
				-- this is mainly for black flow...
				if server.questdata and server.questdata.arenatimer
					server.questdata.arenatimer_on = true
				end
				
				mo.playersmusic = $ or {}
				for p in players.iterate do
					
					if p.mo and p.mo.valid and p.pso
					
						if not mo.playersmusic[p.mo]
							if p.pso.hp
							and not p.pso.teleport	-- medbay tp
							and PSO.lib.objectInRoom(room, p.mo)
								
								if not (mo.options & 4)	-- no boss music.
									if mapheaderinfo[gamemap].bossrush
										S_ChangeMusic(info.bossrushmusic, true, p)
									else
										S_ChangeMusic(info.bossmusic, true, p)
									end
								end
								mo.playersmusic[p.mo] = true
								p.pso.inboss = true
								
								-- start the emergency!
								if info.disaster
									PSO.lib.startEmergency(p, "EMERGENCY CODE", "(DISASTER)", sfx_emrgy, sfx_cdisa, TICRATE*3/2)
								else
									PSO.lib.startEmergency(p, "EMERGENCY CODE", "(DUEL)", sfx_emrgy, sfx_cduel, TICRATE*3/2)
								end
							end
						else
							
							-- reset the stuff by checking for teleporter, not being dead.
							if p.pso.teleport and p.pso.teleport < 5	--p.pso.hp <= 0
								mo.playersmusic[p.mo] = nil
							end
						end
					end
				end
				
				-- check if players have died or smth
				
			end
		end
	end
end

-- enemies falling in pits...
addHook("ShouldDamage", function(mo, i, s, dtype)
	if mo and mo.valid and PSO.info.enemies[mo.type]
		if (dtype == DMG_DEATHPIT or P_CheckDeathPitCollide(mo)) and not mo.pitdeath
			if mo.pso.part or mo.pso.boss
				return false
			end
		end
	end
end)

addHook("MobjDeath", function(mo, i, s, dtype)
	
	--print("A")
	if PSO.info.enemies[mo.type]
		if (dtype == DMG_DEATHPIT or P_CheckDeathPitCollide(mo)) and not mo.pitdeath	--t == DMG_DEATHPIT
			if mo.pso.part or mo.pso.boss
				return true
			end
			
			mo.pitdeath = true
			PSO.lib.damageObject(mo, nil, max(mo.pso.maxhp, 999999999))
			mo.z = $+FRACUNIT
			mo.flags = $|MF_NOGRAVITY
			
			mo.momz = 0
			mo.health = 0
		
			return false
		end
		
		if PSO.info.enemies[mo.type].deathfunc
			PSO.info.enemies[mo.type].deathfunc(mo)
		end
	end
end)

-- part creator:
PSO.lib.enemyCreatePart = function(source, x, y, z, parttype)
	
	local part = P_SpawnMobj(x, y, z, parttype)
	part.pso = {}
	
	for k,v in pairs(source.pso)
		part.pso[k] = v
	end
	
	if PSO.info.enemies[parttype]
	and PSO.info.enemies[parttype].part
		--print("B")
		part.pso.maxhp = max(1, source.pso.maxhp * PSO.info.enemies[parttype].bhp / 100)
		part.pso.hp = part.pso.maxhp
		part.pso.exp = 0
		part.pso.part = true
		if PSO.info.enemies[parttype].partlinkdamage
			--print("links damage")
			part.pso.partlinkdamage = true
		end
	end
	
	part.tracer = source
	
	source.parts = $ or {}
	source.parts[#source.parts+1] = part
	
	return part
end

PSO.lib.positionInvalid = function(mo, tx, ty)
	local blocked
	local function blockcheck(mo, mo2)
		if mo2 ~= mo and mo2 and mo2.valid and mo2.flags & MF_SOLID
		and mo2.type ~= MT_PLAYER	-- ramming this is always fine :)
		and not (mo2.flags & MF_ENEMY) 
								
			if tx >= mo2.x - mo2.radius
			and tx <= mo2.x + mo2.radius
			and ty >= mo2.y - mo2.radius
			and ty <= mo2.y + mo2.radius
			and mo.z >= mo2.z
			and mo.z <= mo2.z+mo.height
				blocked = true
			end
		end	
	end
						
	local x1 = mo.x - 512*FRACUNIT
	local x2 = mo.x + 512*FRACUNIT
	local y1 = mo.y - 512*FRACUNIT
	local y2 = mo.y + 512*FRACUNIT
						
	searchBlockmap("objects", blockcheck, mo, x1, x2, y1, y2)
	
	return blocked
end

PSO.lib.enemyMovementInvalid = function(mo)
	
	local spd = FixedHypot(mo.momx, mo.momy)
	local a = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	if not spd return false end
	
	local bspd = spd
	
	for i = 1, 4
		
		local tx = mo.x + FixedMul(spd, cos(a))
		local ty = mo.y + FixedMul(spd, sin(a))
		
		if PSO.lib.positionInvalid(mo, tx, ty)
			return true
		end
		spd = $+bspd
	end
	
	-- if we made it here the movement wasn't blocked.
	return false
end

-- checks if the enemy's movement would lead to their untimely demise......
PSO.lib.enemyPitCheck = function(mo)
	
	local nextx = mo.x+mo.momx
	local nexty = mo.y+mo.momy
	
	local sec = R_PointInSubsector(nextx, nexty).sector
	
	if sec.damagetype >= 7 and sec.damagetype <= 8	-- SD_DEATHPIT/SD_DEATHPITNOTILT
		return true
	end
	
	for rover in sec.ffloors()
		if rover.topheight < mo.z
		and rover.sector.damagetype >= 7 and rover.sector.damagetype <= 8
			return true
		end
	end
end

-- enemy push hitbox: pushes the player following momentum or pushes them away
freeslot("MT_PSO_PUSHBOX")
mobjinfo[MT_PSO_PUSHBOX] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 16<<FRACBITS,
	height = 32<<FRACBITS,
	flags = MF_NOGRAVITY|MF_SOLID,
}

local reactivation_delay = 5
addHook("MobjThinker", function(mo)
	
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	P_MoveOrigin(mo, mo.target.x, mo.target.y, mo.target.z)
	mo.momx = mo.target.momx
	mo.momy = mo.target.momy
	mo.momz = mo.target.momz
	
	mo.radius = mo.target.radius	--*11/10
	mo.height = mo.target.height
	
	if mo.target.disablepushbox
		mo.delay = reactivation_delay
		mo.flags = $ & ~MF_SOLID
	elseif mo.delay
		mo.delay = $-1
		if not mo.delay
			mo.flags = $|MF_SOLID
		end
	end
	
	--print(mo.radius/FRACUNIT)
	
	--P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	
	mo.tics = 2
end, MT_PSO_PUSHBOX)

local function pushbox_collide(mo1, mo2)
	
	if not mo1.target or not mo1.target.valid return false end
	if mo1.target.disablepushbox return false end
	if not mo2 or not mo2.valid return end
	
	if mo2.type != MT_PLAYER return false end
	if mo2.snapped == leveltime return false end	-- player got moved already by a snapping hitbox
	
	if mo2.player.pso.blowdown or mo2.player.pso.stagger or mo2.player.pso.knockback return false end
end

addHook("MobjCollide", pushbox_collide, MT_PSO_PUSHBOX)
addHook("MobjMoveCollide", pushbox_collide, MT_PSO_PUSHBOX)

local function pushbox_blocked(mo1, mo2)
	if not mo2 or not mo2.valid return end	-- line block
	
	-- push mo2 the fuck out
	mo2.momx = mo1.momx
	mo2.momy = mo2.momy
	return true
end

addHook("MobjMoveBlocked", pushbox_blocked, MT_PSO_PUSHBOX)

-- find out if the hitboxes are intersecting
local function isColliding(mo, mo2)
	return (mo.x + mo.radius > mo2.x - mo.radius or mo.x - mo.radius < mo2.x + mo2.radius)
	and (mo.y + mo.radius > mo2.y - mo.radius or mo.y - mo.radius < mo2.y + mo2.radius)
end

-- find out which point of the hitbox we're closest to, and snap to it
local function snapOnHitbox(mo, mo2)
	
	local ld = {
		
		/*
			   3
			 _ _ _
			|     |
		  2 |     | 4
			|_ _ _|
			   1
		*/
		
		-- X1, 				X2, 				Y1, 				Y2
		{mo2.x - mo2.radius, mo2.x + mo2.radius, mo2.y - mo2.radius, mo2.y - mo2.radius},
		{mo2.x - mo2.radius, mo2.x - mo2.radius, mo2.y - mo2.radius, mo2.y + mo2.radius},
		{mo2.x - mo2.radius, mo2.x + mo2.radius, mo2.y + mo2.radius, mo2.y + mo2.radius},
		{mo2.x + mo2.radius, mo2.x + mo2.radius, mo2.y - mo2.radius, mo2.y + mo2.radius}
	}
	
	local invalidsides = {}
	local validtries = 0
	
	while validtries < 4
		validtries = $+1
		local points = {}
		local closestdist = mo.radius*4
		local closestindex = 0
		
		for i = 1, #ld
			-- we cheay by subtracting momx/momy because we want to be pushed away in most cases
			local x, y = P_ClosestPointOnLine(mo.x - mo.momx*2, mo.y - mo.momy*2, ld[i][1], ld[i][3], ld[i][2], ld[i][4])
			points[i] = {x, y}
			
			local dist = R_PointToDist2(mo.x, mo.y, points[i][1], points[i][2])
			if dist < closestdist
			and not (invalidsides[i])
				closestdist = dist
				closestindex = i
			end
		end
		
		-- and now we're gonna have to snap the player to a location
		local tx = mo.x
		local ty = mo.y
		
		if closestindex == 1	-- bottom
			tx = mo.x
			ty = mo2.y - mo2.radius - mo.radius - mo.momy
		elseif closestindex == 2	--left
			tx = mo2.x - mo2.radius - mo.radius - mo.momx
			ty = mo.y
		elseif closestindex == 3	-- top
			tx = mo.x
			ty = mo2.y + mo2.radius + mo.radius - mo.momy
		elseif closestindex == 4	-- right
			tx = mo2.x + mo2.radius + mo.radius - mo.momx
			ty = mo.y
		end
		
		-- test our position...
		-- our position must be valid in the sense that the player SHOULD be able to get to it.
		-- if any of them end up "inside the ground" or otherwise outside boundaries, we're cooked.
		
		mo.flags = $|MF_NOCLIPTHING	-- we consider ourselves intangible to other things...
		
		if not P_TryMove(mo, tx+mo2.momx, ty+mo2.momy, true)
			--print("INVALID")
			mo.flags = $ & ~MF_NOCLIPTHING
			invalidsides[i] = true
			continue
		end
		
		-- make sure to stop these momentums so we don't continuously run into it.
		if closestindex & 1
			mo.momy = 0
		else
			mo.momx = 0
		end
		
		mo.flags = $ & ~MF_NOCLIPTHING
		P_MoveOrigin(mo, tx+mo2.momx, ty+mo2.momy, mo.z)
		break
	end
end	

/*
local function testGargoyle(mo, mo2)
	if mo and mo.valid and mo.type == MT_GARGOYLE and mo.testing
		
		if mo2 and mo2.valid and (mo2.flags & MF_ENEMY or mo2.type == MT_PLAYER)
			return false
		end
	end
end

addHook("MobjCollide", testGargoyle, MT_GARGOYLE)
addHook("MobjMoveCollide", testGargoyle, MT_GARGOYLE)
*/

local function player_enemyTopCollide(mo, mo2)
	
	if mo2 and mo2.valid and mo2.flags & MF_ENEMY
	and mo.z+mo.height >= mo2.z	
	and mo.z <= mo2.z+(mo2.height+FRACUNIT*8)	
	and mo.lastsnap ~= leveltime
		mo.lastsnap = leveltime
		snapOnHitbox(mo, mo2)
		return false
		
		/*if (mo.snapped ~= leveltime)	-- and (mo.snapcount == nil or mo.snapcount < 5))
		and not mo.delay	-- pushbox shenanigans
		and mo.z >= mo2.z+mo2.height/2	-- standing on it?
		and mo.z <= mo2.z+mo2.height*3/2
		and (mo.momz <= 0 or (mo.momz == 0 and mo.z >= mo2.z+mo2.height))
		and isColliding(mo, mo2)
		and not mo.player.pso.stagger
		and not mo.player.pso.knockback
		and not mo.player.pso.blowdown
		-- falling or standing above the hitbox
			--P_InstaThrust(mo, mo2.angle, FRACUNIT*16)
			snapOnHitbox(mo, mo2)
			
			/*if mo.snapped ~= leveltime
				mo.snapcount = 0
			elseif mo.snapcount
				mo.snapcount = $+1
			else
				mo.snapcount = 1
			end*/
			
			/*mo.snapped = leveltime
			return false
		end*/
	end
end

/*
addHook("MobjCollide", player_enemyTopCollide, MT_PLAYER)
addHook("MobjMoveCollide", player_enemyTopCollide, MT_PLAYER)
*/

-- enemy collision is now handled in player code...

-- corruption activation

PSO.lib.enemySetCorruption = function(mo)
	
	if not mo.checkcorruption
		
		local diff = PSO.lib.questDifficulty()
		

		if mo.options & 8	-- always spawn corrupted
		or (diff == DIFF_VERYHARD and (P_RandomRange(0, 100) < (mo.pso.corruptionchance or 0)))
			mo.pso.corrupted = true
		end
		
		mo.checkcorruption = true
	end
end

-- corrupttion visuals
PSO.lib.enemyCorruptionVisual = function(mo)
	
	if not mo.corruptionoverlay
		mo.corruptionoverlay = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		P_SetScale(mo.corruptionoverlay, mo.scale, true)
	end
	
	local o = mo.corruptionoverlay
	
	o.tics = 2
	o.fuse = -1
	o.sprite = mo.sprite
	o.frame = mo.frame & FF_FRAMEMASK
	o.angle = mo.angle
	o.rollangle = mo.rollangle
	o.scale = mo.scale
	o.spritexscale = mo.spritexscale
	o.spriteyscale = mo.spriteyscale
	
	if mo.flags2 & MF2_DONTDRAW
		o.flags2 = $|MF2_DONTDRAW
	else
		o.flags2 = $ & ~MF2_DONTDRAW
	end
	
	
	local a = R_PointToAngle(mo.x, mo.y)
	local x = mo.x - cos(a)
	local y = mo.y - sin(a)
	
	local flash = (4 + 3*cos(leveltime*ANG1*10) /FRACUNIT) << FF_TRANSSHIFT
	o.frame = $ | flash | FF_FULLBRIGHT
	o.colorized = true
	o.color = SKINCOLOR_MASTER
	
	P_SetOrigin(o, x, y, mo.z)
	o.momx = mo.momx
	o.momy = mo.momy
	o.momz = mo.momz
	
	if leveltime & 1 and not (mo.flags2 & MF2_DONTDRAW)
		local sr = mo.radius/FRACUNIT
		local sx = mo.x + P_RandomRange(-sr, sr)*FRACUNIT
		local sy = mo.y + P_RandomRange(-sr, sr)*FRACUNIT
		
		local s = P_SpawnMobj(sx, sy, mo.z, MT_THOK)
		s.color = SKINCOLOR_MASTER
		s.tics = -1
		s.fuse = 10
		s.momz = mo.scale*P_RandomRange(3, 8)
		s.scale = mo.scale/2
		s.destscale = 1
		
		-- interp being a piece of shit as usual.
		local g = P_SpawnGhostMobj(o)
		P_SetOrigin(g, o.x, o.y, o.z)
		P_SetScale(g, o.scale, true)
	end
end

-- show enemy markers on players...

local markerTypes = {
	"MARKER_NOBLOCK",
	"MARKER_SIGHT",
	"MARKER_EXPAND",
	"MARKER_STACK",
	"MARKER_KNOCKBACK",
	"MARKER_USEBLOCK",
}

for i = 1, #markerTypes 
	rawset(_G, markerTypes[i], i)
end

local markerColours = {
	SKINCOLOR_RED,
	SKINCOLOR_PURPLE,
	SKINCOLOR_GREEN,
	SKINCOLOR_YELLOW,
	SKINCOLOR_PINK,
	SKINCOLOR_RED,
}
local markerFrames = {B, C, A, D, F, I}

PSO.lib.attackMarker = function(x, y, z, t, s, onetic)
	
	s = $ or FRACUNIT/3
	
	local m = P_SpawnMobj(x,y,z, MT_THOK)
	m.sprite = SPR_INDC
	m.frame = markerFrames[t]|(t ~= MARKER_USEBLOCK and FF_ADD or 0)|FF_FULLBRIGHT
	m.color = markerColours[t]
	m.scale = s
	
	if leveltime%TICRATE == 0
		local g = P_SpawnGhostMobj(m)
		g.destscale = s*8
		g.scalespeed = s/12
	end
	
	m.tics = onetic and 1 or 2
end

-- knockback marker
PSO.lib.enemyKnockbackMarker = function(mo, origin, tics)
	local angle = R_PointToAngle2(origin.x, origin.y, mo.x, mo.y)
	
	--PSO.lib.attackMarker(mo.x, mo.y, mo.z + FRACUNIT*128, MARKER_KNOCKBACK, FRACUNIT*2/3, tics == 1)
	
	-- knockback arrows
	local dist = (leveltime*2) % 64
	for i = 1, 8
		
		local x = mo.x + dist*cos(angle)
		local y = mo.y + dist*sin(angle)
		local t = P_SpawnMobj(x, y, mo.z + FRACUNIT*32, MT_THOK)
		t.sprite = SPR_INDC
		t.frame = G|FF_ADD
		t.renderflags = $ |(RF_FLOORSPRITE|RF_NOSPLATBILLBOARD|RF_NOSPLATROLLANGLE)
		t.rollangle = angle + ANG1*90
		t.scale = $*3/2
		t.tics = tics or 2
		t.color = SKINCOLOR_PINK
		P_InstaThrust(t, angle, 2<<FRACBITS)
		
		if mo.player and mo.player ~= consoleplayer
			t.flags2 = $|MF2_DONTDRAW
		end
		
		dist = $ + 64
	end
end

-- knockback "burst" visual to put around the source...
PSO.lib.enemyKnockbackBurstVisual = function(origin, strength)
	
	strength = $ or 32
	
	for i = 1, 16
		
		local a = FixedAngle((FRACUNIT*22 + FRACUNIT/2)*i)
		
		local t = P_SpawnMobj(origin.x, origin.y, origin.z + FRACUNIT*32, MT_THOK)
		t.sprite = SPR_INDC
		t.frame = G|FF_ADD
		t.renderflags = $ |(RF_FLOORSPRITE|RF_NOSPLATBILLBOARD|RF_NOSPLATROLLANGLE)
		t.rollangle = a + ANG1*90
		t.scale = $*3/2
		t.tics = 16
		t.color = SKINCOLOR_PINK
		P_InstaThrust(t, a, strength<<FRACBITS)		
	end
end

-- sight marker
PSO.lib.enemySightMarker = function(source, dest, altfn, twotics)
	
	local sightfn = altfn and altfn or P_CheckSight
	
	local res, x, y, z = sightfn(source, dest)
	
	if res
		
		if leveltime & 1
			local s = P_SpawnMobj(dest.x, dest.y, dest.z, MT_THOK)
			s.sprite = SPR_TARG
			s.frame = (leveltime%14)/2 |FF_FULLBRIGHT
			s.tics = twotics and 2 or 1
		end
		
		--PSO.lib.attackMarker(dest.x, dest.y, dest.z + FRACUNIT*128, MARKER_SIGHT, FRACUNIT*2/3, not twotics)
				
		-- spawn a line between us and the target to help us know where harm is coming from
		local fn = function(s)
			s.sprite = SPR_HOOP
			s.frame = A|FF_FULLBRIGHT|FF_TRANS50
			s.tics = twotics and 2 or 1
		end
		PSO.lib.spawnLine(source, dest, 24, fn, twotics and 2 or 1)
		
	elseif x ~= nil and y ~= nil and z ~= nil
	
		local def = P_SpawnMobj(x, y, z, MT_THOK)
		if leveltime & 1
			local s = P_SpawnMobj(x, y, z, MT_THOK)
			s.sprite = SPR_TARG
			s.frame = (leveltime%14)/2 |FF_FULLBRIGHT
			s.tics = 1
		end
		PSO.lib.spawnLine(source, def, 24, nil, twotics and 2 or 1)
		
		P_RemoveMobj(def)
		
	end
end

-- stack marker object
-- this will attack every player in sight who's not in the radius after a set time.
freeslot("MT_PSO_AOEVISUAL", "MT_PSO_STACKMARKER")

mobjinfo[MT_PSO_AOEVISUAL] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	radius = 32<<FRACBITS,
	height = 32<<FRACBITS,
	spawnhealth = 1000,
	flags = MF_NOGRAVITY|MF_NOCLIP
}

mobjinfo[MT_PSO_STACKMARKER] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	radius = 32<<FRACBITS,
	height = 32<<FRACBITS,
	spawnhealth = 1000,
	flags = MF_NOGRAVITY|MF_NOCLIP
}

-- initializes and spawns an AOE marker
-- if potency is left nil, will deal no damage and be purely visual
PSO.lib.spawnAOEMarker = function(source, x, y, z, radius, delay, colour, potency, damagetype, reaction, strength)
	local mark = P_SpawnMobj(x, y, z, MT_PSO_AOEVISUAL)
	mark.target = source
	mark.radius = radius
	mark.delay = delay
	mark.potency = potency
	mark.reaction = reaction or RT_STAGGER
	mark.reactionstrength = strength or 1
	mark.damagetype = damagetype
	
	mark.parts = {}
	mark.time = 0
	mark.interp = 0
	
	for i = 1, 64
		local part = P_SpawnMobj(x, y, z, MT_THOK)
		part.sprite = SPR_INDC
		part.frame = E|FF_FULLBRIGHT|FF_PAPERSPRITE|FF_ADD
		part.tics = -1
		part.color = colour
		part.fuse = delay
		
		mark.parts[i] = part
	end
	
	return mark
end

addHook("MobjThinker", function(mark)
	mark.time = $+1
	mark.interp = $+1
	local tpfunc = mark.interp == 1 and P_SetOrigin or P_MoveOrigin
	
	-- circle handling...
	
	if mark.time < mark.delay-1
	
		for i = 1, #mark.parts
			
			local part = mark.parts[i]
			local radius = mark.radius
			local tflag = 0
			
			if mark.time < 10
				
				radius = mark.radius + 32*(10-mark.time)
				tflag = (10-mark.time)<<FF_TRANSSHIFT
			end
			
			part.frame = ($ & FF_FRAMEMASK)|FF_PAPERSPRITE|FF_FULLBRIGHT|FF_ADD|tflag
			
			local a = mark.angle + FixedAngle(FixedDiv(FRACUNIT*360, FRACUNIT*64))*i
			local x = mark.x + radius*cos(a)
			local y = mark.y + radius*sin(a)
			tpfunc(part, x, y, mark.z + FRACUNIT*4)
			part.angle = a + ANG1*90
		end
	
		mark.angle = $ + ANG1
	elseif mark.time == mark.delay
		-- attack any player in sight
		
		if mark.potency	-- only deal damage if potency was set
			for p in players.iterate
				local pmo = p.mo
				if not pmo or not pmo.valid continue end
				
				if R_PointToDist2(pmo.x, pmo.y, mark.x, mark.y)/FRACUNIT <= mark.radius
				and P_CheckSight(pmo, mark)
					local hb = PSO.lib.spawnHurtBox(pmo.x, pmo.y, pmo.z, 48<<FRACBITS, 48<<FRACBITS, mark.target, mark.potency, mark.damagetype, 2, mark.reaction)
					hb.pso.reactionstrength = mark.reactionstrength
					hb.angle = R_PointToAngle2(mark.x, mark.y, pmo.x, pmo.y)
					hb.pso.noblock = true
				end
			end
		end
		
		P_RemoveMobj(mark)
		return
	end
	
end, MT_PSO_AOEVISUAL)

-- spawns the stack marker and initializes everything
PSO.lib.spawnStackMarker = function(source, x, y, z, radius, delay, potency, damagetype, reaction, strength)
	
	local mark = P_SpawnMobj(x, y, z, MT_PSO_STACKMARKER)
	mark.target = source
	mark.radius = radius
	mark.delay = delay
	mark.potency = potency or 100
	mark.reaction = reaction or RT_STAGGER
	mark.reactionstrength = strength or 1
	mark.damagetype = damagetype
	
	mark.time = 0
	mark.interp = 0
	
	mark.aoe = PSO.lib.spawnAOEMarker(source, x, y, z, radius, delay, SKINCOLOR_YELLOW)
	
	return mark
end

-- Oh hi Mark
addHook("MobjThinker", function(mark)
	mark.time = $+1
	mark.interp = $+1
	local tpfunc = mark.interp == 1 and P_SetOrigin or P_MoveOrigin
	
	--PSO.lib.attackMarker(mark.x, mark.y, mark.z + FRACUNIT*128, MARKER_STACK, FRACUNIT)
	
	local arrowz = 16*cos(ANG1*mark.time*5)
	
	local arrow = P_SpawnMobj(mark.x, mark.y, mark.z + FRACUNIT*64 + arrowz, MT_THOK)
	arrow.sprite = SPR_INDC
	arrow.scale = $*3
	arrow.frame = H|FF_PAPERSPRITE|FF_ADD|FF_FULLBRIGHT
	arrow.color = SKINCOLOR_YELLOW
	arrow.tics = 2
	
	local twirl = 0
	if mark.time%72 < 36
		twirl = (mark.time%72)*ANG1*10
	end
	
	arrow.angle = R_PointToAngle(arrow.x, arrow.y) + ANG1*90 + twirl
	
	-- circle handling...
	
	if mark.time < mark.delay-1
		
		if mark.aoe and mark.aoe.valid
			tpfunc(mark.aoe, mark.x, mark.y, mark.z)
		end
		
		-- make it pulsate outwards...
		local ghosttime = mark.time % TICRATE
			
			
		local radius = mark.radius + ghosttime*32
		local delay = 10 - (ghosttime*10/TICRATE)
		
		local mark2 = PSO.lib.spawnAOEMarker(source, mark.x, mark.y, mark.z, radius, TICRATE+delay, SKINCOLOR_YELLOW)
		mark2.time = TICRATE	-- a cheat...
		mark2.tics = 2
		for i = 1, #mark2.parts
			mark2.parts[i].fuse = 10
			--mark2.parts[i].tics = 2
		end
		
	elseif mark.time == mark.delay
		-- attack any player in sight
		for p in players.iterate
			local pmo = p.mo
			if not pmo or not pmo.valid continue end
			
			if R_PointToDist2(pmo.x, pmo.y, mark.x, mark.y)/FRACUNIT > mark.radius
			and P_CheckSight(pmo, mark)
				local hb = PSO.lib.spawnHurtBox(pmo.x, pmo.y, pmo.z, 48<<FRACBITS, 48<<FRACBITS, mark.target, mark.potency, mark.damagetype, 2, mark.reaction)
				hb.pso.reactionstrength = mark.reactionstrength
				hb.angle = R_PointToAngle2(mark.x, mark.y, pmo.x, pmo.y)
				hb.pso.noblock = true
			end
		end
		
		P_RemoveMobj(mark)
		return
	end
	
end, MT_PSO_STACKMARKER)