-- .executor, Capital / Mines area boss.
/*

*/

freeslot("MT_PSO_EXECUTOR", "SPR_PEXE", "S_PSO_EXECUTOR_DIE")
freeslot("MT_PSO_EXECUTOR_TURRET", "MT_PSO_EXECUTOR_PROJECTILE")

local painframes = {A, E, F, A, G, H, A, E, H, F}
function A_ExecutorDie(mo)

	local index = ((leveltime/2) % #painframes) +1
	mo.frame = (painframes[index] or A) | FF_FULLBRIGHT | FF_ADD | TR_TRANS20
	
	mo.spritexscale = P_RandomRange(1, 8)*FRACUNIT/2
	mo.spriteyscale = P_RandomRange(1, 8)*FRACUNIT/2
	
	if not mo.cusval
		mo.pain = nil
		mo.aiming = nil
		mo.aicd = nil
		mo.acid_up = nil
		mo.flags = $|MF_NOBLOCKMAP|MF_NOCLIPTHING
		mo.flags = $ & ~MF_SHOOTABLE
		
		-- remove currently active enemies
		if mo.activeenemies
			for i = 1, #mo.activeenemies
				if mo.activeenemies[i] and mo.activeenemies[i].valid
					P_RemoveMobj(mo.activeenemies[i])
				end
			end
		end
		mo.activeenemies = nil
		
		P_LinedefExecute(5001, mo)
	end
	
	mo.cusval = $+1
	if mo.cusval > TICRATE*4
		P_RemoveMobj(mo)
	end
end

states[S_PSO_EXECUTOR_DIE] = {SPR_PEXE, A, 1, A_ExecutorDie, 0, 0, S_PSO_EXECUTOR_DIE}

mobjinfo[MT_PSO_EXECUTOR] = {
	doomednum = 2555,
	spawnhealth = 1000,
	deathstate = S_PSO_EXECUTOR_DIE,
	spawnstate = S_THOK,
	height = 120<<FRACBITS,
	radius = 50<<FRACBITS,
	flags = MF_ENEMY|MF_SHOOTABLE|MF_NOGRAVITY,
}

mobjinfo[MT_PSO_EXECUTOR_TURRET] = {
	doomednum = -1,
	spawnhealth = 1000,
	spawnstate = S_THOK,
	deathstate = S_XPLD1,
	height = 96<<FRACBITS,
	radius = 24<<FRACBITS,
	flags = 0,
}

mobjinfo[MT_PSO_EXECUTOR_PROJECTILE] = {
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 24<<FRACBITS,
	height = 64<<FRACBITS,
	flags = MF_NOGRAVITY|MF_MISSILE,
}

local projectile_speed = 12*FRACUNIT
local projectile_accel = FRACUNIT*2/3

-- basic chase thinker for projectile
addHook("MobjThinker", function(mo)
	
	local target = mo.tracer
	if not target or not target.valid
		P_RemoveMobj(mo)
		return
	end
	
	local dist = P_AproxDistance(P_AproxDistance(target.x - mo.x, target.y - mo.y), target.z - mo.z)
	
	-- normalize vakues
	local dx = FixedDiv(target.x - mo.x, dist)
	local dy = FixedDiv(target.y - mo.y, dist)
	local dz = FixedDiv((target.z + target.height*6/10) - mo.z, dist)
	
	-- the values we want:
	local desired_momx = FixedMul(dx, projectile_speed)
	local desired_momy = FixedMul(dy, projectile_speed)
	local desired_momz = FixedMul(dz, projectile_speed)
	
	-- how much we'd have to change by
	local steeringx = desired_momx - mo.momx
	local steeringy = desired_momy - mo.momy
	local steeringz = desired_momz - mo.momz
	
	-- cap acceleration
	if abs(steeringx) > projectile_accel
		steeringx = (steeringx > 0 and 1 or -1) * projectile_accel
	end

	if abs(steeringy) > projectile_accel
		steeringy = (steeringy > 0 and 1 or -1) * projectile_accel
	end
	
	if abs(steeringz) > projectile_accel
		steeringz = (steeringz > 0 and 1 or -1) * projectile_accel
	end	
	
	-- apply
	mo.momx = $ + steeringx
	mo.momy = $ + steeringy
	mo.momz = $ + steeringz

	
	if leveltime%4 == 0
		local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		s.tics = TICRATE
		s.fuse = TICRATE
		s.sprite = SPR_NTHK
		s.frame = E|FF_ADD|TR_TRANS20|FF_FULLBRIGHT
		s.color = SKINCOLOR_RED
		s.colorized = true
		s.momx = P_RandomRange(-3, 3)*mo.scale
		s.momy = P_RandomRange(-3, 3)*mo.scale
		s.momz = P_RandomRange(-3, 3)*mo.scale
	end
	
	--print(dist)
	if mo.tracer and mo.tracer.valid and mo.tracer.type == MT_THOK
	and dist < mo.scale*48
		P_RemoveMobj(mo)
	end
	
end, MT_PSO_EXECUTOR_PROJECTILE)

local EXE_NEUTRAL = 0
local EXE_PAIN = 1
local EXE_TURRET = 2
local EXE_ACIDRISE = 3
local EXE_SPAWNENEMIES = 4
local EXE_MOVE = 5
local EXE_PHASESHIFT = 6
local EXE_BEAM = 7

-- link our relevant spawn / teleport points
local function executor_linkPoints(mo)
	
	-- find "taunt" point
	mo.turretpoints = {}
	mo.tppoints = {}
	mo.aises = {}
	
	for m in mobjs.iterate() do
		
		if m and m.valid
			
			if m.type == MT_BOSSFLYPOINT 
				if m.mtangle == 1
					mo.tauntpoint = m
				
				elseif m.mtangle >= 2 and m.mtangle <= 4
					-- turret points
					mo.turretpoints[m.mtangle-1] = m
				end
			
			elseif m.type == MT_BOSS9GATHERPOINT
		
				local t = P_SpawnMobj(m.x, m.y, m.z, MT_THOK)
				t.tics = -1
				t.sprite = SPR_BRAK
				t.frame = N
				t.angle = m.angle
				t.scale = $*2
				
				if m.extrainfo
					local sightmarker = P_SpawnMobj(m.x + 128*cos(m.angle), m.y + 128*sin(m.angle), m.z + mo.scale*192, MT_THOK)
					sightmarker.state = S_INVISIBLE
					sightmarker.tics = -1
					sightmarker.fuse = -1
					mo.aises[#mo.aises+1] = sightmarker
				end
				
			
			elseif m.type == MT_EGGTRAP
				mo.tppoints[#mo.tppoints+1] = m
			end
		end
	end
end

local idleframes = {A, B, C, A, D, B, A, D, B}
local function executor_idle(mo)
	
	local index = ((leveltime/2) % #idleframes) +1
	--print(index)
	
	mo.frame = (idleframes[index] or A) | FF_FULLBRIGHT | FF_ADD | TR_TRANS20
	
	if mo.executorstate == EXE_PAIN
		mo.executorstate = EXE_NEUTRAL
	end
end

-- basically sets stagger
local function executor_doPain(mo)
	mo.pso.stagger = 1
	mo.paincd = TICRATE
	mo.executorstate = EXE_PAIN
end

-- checks whether we need to be in pain
-- (we won't use conventional reactions bc we only want staggeer.)
local function executor_checkPain(mo)
	
	if mo.prevhp
	and not mo.executorstate
	and not mo.paincd
		
		if mo.prevhp > mo.pso.hp
			executor_doPain(mo)
		end
	end
	mo.prevhp = mo.pso.hp
	
	if mo.paincd
		mo.paincd = $-1
	end
end

local function executor_pain(mo)
	
	PSO.lib.enemyStaggerDefault(mo)
	mo.momz = 0
	
	local index = ((leveltime/2) % #painframes) +1
	mo.frame = (painframes[index] or A) | FF_FULLBRIGHT | FF_ADD | TR_TRANS20
end


-- fires homing projectiles to target
local function executor_fireMissiles(mo)
	
	local num_missiles = 3
	local missile_potency = 300
	local diff = PSO.lib.questDifficulty()
	
	for i = 1, num_missiles
		
		local msl = PSO.lib.spawnMissile(mo, MT_PSO_EXECUTOR_PROJECTILE, mo.angle, 0, missile_potency, DMT_RANGED, 0, 0, mo.height*3/2, RT_STAGGER)
		msl.tracer = mo.target
		msl.momx = P_RandomRange(-16, 16)*mo.scale
		msl.momy = P_RandomRange(-16, 16)*mo.scale
		msl.momz = P_RandomRange(16, 24)*mo.scale
		msl.fuse = TICRATE*4
		
		if diff >= DIFF_HARD
			msl.pso.inflictzalure = TICRATE*20
			msl.pso.inflictjellen = TICRATE*20
			msl.fuse = TICRATE*6
		end
	end
	
	S_StartSound(mo, sfx_s3k5e)
end

local function executor_prepTeleport(mo, tx, ty, tz, ta)
	
	mo.teleport = 1
	mo.teleporttarget = {tx, ty, tz, ta}
	mo.teleportstate = 1
	S_StartSound(mo, sfx_s3kb6)
end

local function executor_doTeleport(mo)
	
	if mo.teleport
		
		mo.spriteyscale = mo.teleport*FRACUNIT
		mo.spritexscale = FRACUNIT/mo.teleport

		if mo.teleportstate == 1
			mo.teleport = $+1
			mo.colorized = true
			mo.color = SKINCOLOR_WHITE
			mo.renderflags = $|RF_FULLBRIGHT
		else
			mo.teleport = $-1
			
			if not mo.teleport
				mo.colorized = false
				mo.renderflags = $ & ~RF_FULLBRIGHT
			end
		end
		
		if mo.teleport == 6

			P_SetOrigin(mo, mo.teleporttarget[1], mo.teleporttarget[2], mo.teleporttarget[3])
			mo.teleportstate = 2
			mo.angle = mo.teleporttarget[4]
		end	
	end
end

-- moves to a different teleport point.
-- 2nd argument says whether we should launch a missile after moving.
local function executor_prepMove(mo, msl)
	
	if msl
		mo.movemissile = 8
	end
	
	local pt = mo.tppoints[P_RandomRange(1, #mo.tppoints)]
	executor_prepTeleport(mo, pt.x, pt.y, pt.z, mo.angle)
	
	mo.executorstate = EXE_MOVE
	
end

local function executor_Move(mo)
	
	if mo.executorstate == EXE_MOVE
		
		if mo.movemissile
			mo.movemissile = $-1
			if not mo.movemissile
				executor_fireMissiles(mo)
			end
		elseif not mo.teleport	-- make sure we're done teleporting
			mo.executorstate = EXE_NEUTRAL
		end
	end
end

local function executor_prepTurret(mo)
	
	mo.turret = 1
	
	executor_prepTeleport(mo, mo.tauntpoint.x, mo.tauntpoint.y, mo.tauntpoint.z, mo.angle)
	mo.executorstate = EXE_TURRET
end

local function executor_openAISGates(mo)
	
	P_LinedefExecute(5002, mo)
end

local function executor_doTurret(mo)
	
	local diff = PSO.lib.questDifficulty()
	
	if mo.executorstate == EXE_TURRET
		
		mo.turret = $+1
		
		if mo.turret > 4
			
			local warnframes = {A, K}
			local index = ((leveltime/10) % #warnframes) +1
			mo.frame = (warnframes[index] or A) | FF_FULLBRIGHT | FF_ADD | TR_TRANS20			
		end
		
		
		if mo.turret == TICRATE*2
			
			local index = P_RandomRange(1, #mo.turretpoints)
			local tx = mo.turretpoints[index].x
			local ty = mo.turretpoints[index].y
			local tz = mo.turretpoints[index].z
			local t = P_SpawnMobj(tx, ty, tz, MT_PSO_EXECUTOR_TURRET)
			t.target = mo
			t.fuse = TICRATE*(diff >= DIFF_HARD and 8 or 5)
			
		end
		
		if mo.turret == TICRATE*3
			executor_prepMove(mo)
			mo.turret = nil
		end
	end
end

-- thinker for the turret itself.
local function executor_turretThinker(t)
	
	t.sprite = SPR_TURR
	t.frame = I | FF_ADD | TR_TRANS30 | FF_FULLBRIGHT
	t.colorized = true
	t.color = SKINCOLOR_RED
	t.tics = -1
	
	t.alivetime = $ and $+1 or 1
	
	if t.alivetime == 1
		S_StartSound(t, sfx_gravch)
	end
	
	-- appearing animation
	local stretchtime = max(1, 6 - t.alivetime)
	t.spriteyscale = 2*stretchtime*FRACUNIT
	t.spritexscale = 2*FRACUNIT/stretchtime
	
	local delay = TICRATE*3/4
	local interval = TICRATE/4
	local attack_potency = 150
	
	if t.alivetime > delay
		local shoot = t.alivetime-delay
		
		if shoot % interval == 0

			S_StartSound(t, sfx_trfire)
			local msl = PSO.lib.spawnMissile(t.target, MT_JETTBULLET, t.angle, 30*FRACUNIT, attack_potency, DMT_RANGED, 0, 0, 0, RT_NONE)
			P_SetOrigin(msl, t.x, t.y, t.z + 24*t.scale)
			msl.flags = MF_MISSILE|MF_NOGRAVITY|MF_BOUNCE
			msl.state = S_FORC1
			msl.colorized = true
			msl.color = SKINCOLOR_RED
			msl.spritexscale = $/2
			msl.spriteyscale = $/2
			msl.fuse = TICRATE*6
			msl.renderflags = $|RF_FULLBRIGHT
			
			t.angle = $ + ANG1*45
			
		end
	end
	
	if t.fuse < 6
	and t.fuse >= 0
		
		local tt = 6 - t.fuse
		t.spriteyscale = 2*tt*FRACUNIT
		t.spritexscale = 2*FRACUNIT/tt
	
	end
end

addHook("MobjThinker", executor_turretThinker, MT_PSO_EXECUTOR_TURRET)

local function executor_prepAcid(mo)
	
	mo.acid = 1
	mo.acid_up = 1
	
	executor_prepTeleport(mo, mo.tauntpoint.x, mo.tauntpoint.y, mo.tauntpoint.z, mo.angle)
	mo.executorstate = EXE_ACIDRISE
end


local function executor_lowerAcid(mo)
	P_LinedefExecute(5001, mo)
end	

local function executor_doAcid(mo)
	
	local acid_duration = TICRATE*40
	
	if mo.executorstate == EXE_ACIDRISE
		
		mo.acid = $+1
		
		if mo.acid > 4
			
			local warnframes = {A, M}
			local index = ((leveltime/10) % #warnframes) +1
			mo.frame = (warnframes[index] or A) | FF_FULLBRIGHT | FF_ADD | TR_TRANS20			
		end
		
		
		if mo.acid == TICRATE*2
			P_LinedefExecute(5000, mo)
		end
		
		if mo.acid == TICRATE*3
			executor_prepMove(mo)
			mo.acid = nil
		end
	end
	
	if mo.acid_up
		mo.acid_up = $+1
		
		if mo.acid_up > acid_duration
			executor_lowerAcid(mo)
			mo.acid_up = nil
		end
		
	end
end

local function executor_prepSpawn(mo)
	
	mo.spawn = 1
	
	executor_prepTeleport(mo, mo.tauntpoint.x, mo.tauntpoint.y, mo.tauntpoint.z, mo.angle)
	mo.executorstate = EXE_SPAWNENEMIES
end

local possibleEnemies = {MT_SPINCUSHION, MT_REDCRAWLA, MT_UNIDUS, MT_JETTGUNNER, MT_GOLDBUZZ}
local possibleEnemies_Acid = {MT_UNIDUS, MT_JETTGUNNER, MT_JETTBOMBER, MT_GOLDBUZZ}

local function executor_doSpawns(mo)
	
	if mo.executorstate == EXE_SPAWNENEMIES
		
		mo.spawn = $+1
		
		if mo.spawn > 4
			
			local warnframes = {A, L}
			local index = ((leveltime/10) % #warnframes) +1
			mo.frame = (warnframes[index] or A) | FF_FULLBRIGHT | FF_ADD | TR_TRANS20			
		end
		
		
		if mo.spawn == TICRATE*2
			
			local numenemies = 2
			local points = PSO.lib.copyTable(mo.tppoints)
			
			for i = 1, numenemies
				
				if not #points break end	-- contingency
				
				local ptindex = P_RandomRange(1, #points)
				local pt = points[ptindex]
				
				-- pick from a different table of enemies if there is currently acid
				local etable = mo.acid_up and possibleEnemies_Acid or possibleEnemies
				local etype = etable[P_RandomRange(1, #etable)]
				
				-- spawn a goon and add it to our table
				local en = PSO.lib.spawnEnemyFromEnemy(mo, pt.x, pt.y, pt.z, etype)
				
				if mo.activeenemies
					mo.activeenemies[#mo.activeenemies+1] = en
				else
					mo.activeenemies = {en}
				end
				en.detonationtimer = TICRATE*15 *numenemies	-- todo: adapt
				if mo.aiming
					en.detonationtimer = $*3/2
				end
				
				en.pso.nodrops = true	-- these enemies do not give rewards on kill.
				
				-- make sure that point can't be used anymore
				table.remove(points, ptindex)
			end
		end

		if mo.spawn == TICRATE*3
			executor_prepMove(mo)
			mo.spawn = nil
		end
	end
	
	-- handle enemies detonation
	if mo.activeenemies and #mo.activeenemies
		
		local i = #mo.activeenemies
		while i
			
			local en = mo.activeenemies[i]
			
			if en and en.valid
				
				if en.pso and en.pso.hp == 0 and en.health
					P_KillMobj(en)
					en.health = 0
					return
				end
				
				if leveltime%2 == 0
					
					local t = P_SpawnMobj(en.x + P_RandomRange(-32, 32)*en.scale, en.y + P_RandomRange(-32, 32)*en.scale, en.z, MT_THOK)
					t.tics = TICRATE
					t.fuse = TICRATE
					t.sprite = SPR_NTHK
					t.frame = E|FF_ADD|TR_TRANS20|FF_FULLBRIGHT
					t.color = SKINCOLOR_RED
					t.destscale = 1
					t.momz = P_RandomRange(2, 5)*en.scale
				end
				
				en.detonationtimer = $-1
				
				local flashtics = max(2, en.detonationtimer/20)
				
				if (leveltime/2) % flashtics == 0
					en.colorized = true
					en.color = SKINCOLOR_RED
					en.frame = $|FF_FULLBRIGHT
				else
					en.colorized = false
					en.frame = $ & ~FF_FULLBRIGHT
				end
				
				if en.detonationtimer == 0
					
					local fail_potency = 700
					-- damage all players with heavy damage
					for p in players.iterate do
						
						local pmo = p.mo
						if not pmo or not pmo.valid continue end
						
						local hb = PSO.lib.spawnHurtBox(pmo.x, pmo.y, pmo.z, 128<<FRACBITS, 40<<FRACBITS, mo, fail_potency, DMT_STRIKE, 2, RT_BLOWDOWN)
						hb.pso.radialangle = true
						hb.pso.nododge = true	-- cannot be blocked
					end
					
					for j = 1, 16

						local t = P_SpawnMobj(en.x, en.y, en.z, MT_THOK)
						t.tics = TICRATE
						t.fuse = TICRATE
						t.sprite = SPR_NTHK
						t.frame = E|FF_ADD|TR_TRANS20|FF_FULLBRIGHT
						t.color = SKINCOLOR_RED
						t.destscale = 1
						t.momx = P_RandomRange(-24, 24)*en.scale
						t.momy = P_RandomRange(-24, 24)*en.scale
						t.momz = P_RandomRange(-24, 24)*en.scale
					end
					P_KillMobj(en)
					
				end
				
			else
				table.remove(mo.activeenemies, i)
			end
			
			i = $-1
		end
	end
end


local function executor_prepPhase2(mo)
	
	mo.phase2 = 1
	mo.phase2_t = true
	
	executor_prepTeleport(mo, mo.tauntpoint.x, mo.tauntpoint.y, mo.tauntpoint.z, mo.angle)
	mo.executorstate = EXE_PHASESHIFT
	
	mo.flags = $|MF_NOCLIPTHING|MF_NOBLOCKMAP
	mo.flags= $ & ~MF_SHOOTABLE
	
	P_StartQuake(FRACUNIT*40, TICRATE*4)
	S_ChangeMusic("BOS4_2", true)
end

local function executor_doPhaseTransition(mo)
	
	if mo.executorstate == EXE_PHASESHIFT
		
		mo.phase2 = $+1
		
		if mo.phase2 > 4
			
			local warnframes = {A, I}
			local index = ((leveltime/10) % #warnframes) +1
			mo.frame = (warnframes[index] or A) | FF_FULLBRIGHT | FF_ADD | TR_TRANS20			
		end
		
		if mo.phase2 < TICRATE*4
		and leveltime%5 == 0
			S_StartSound(nil, sfx_s3k6f)
		end
		
		if mo.phase2 == TICRATE
			executor_openAISGates(mo)
		end
		
		if mo.phase2 == TICRATE*4
			mo.flags = $ & ~MF_NOCLIPTHING
			mo.flags = $ & ~MF_NOBLOCKMAP
			mo.flags = $|MF_SHOOTABLE
			
			executor_prepMove(mo)
			mo.phase2 = nil
		end
		
	end
end


local function executor_prepAiming(mo)
	
	mo.aiming = 1
	mo.aimingcount = 0
	mo.pso.targetlock = true
end

local function executor_doAiming(mo)
	
	local beam_potency = 550
	
	if not mo.aiming return end

	mo.aiming = $+1
	
	local tt = mo.aiming - TICRATE
	local aimtime = TICRATE*5
	--print(tt)
	
	if not mo.target return end
	
	if tt == 0 and not mo.aimingfrom
		
		mo.aimingfrom = P_RandomRange(1, #mo.aises)
		S_StartSound(nil, sfx_trpowr)
	
	elseif tt < aimtime
		
		local ais = mo.aises[mo.aimingfrom]
		if ais and ais.valid and mo.target and mo.target.valid
			PSO.lib.enemySightMarker(ais, mo.target)

			local t = P_SpawnMobj(ais.x + P_RandomRange(-128, 128)*mo.scale, ais.y + P_RandomRange(-128, 128)*mo.scale, ais.z + P_RandomRange(-128, 128)*mo.scale, MT_THOK)
			t.tics = 10
			t.color = SKINCOLOR_EMERALD
			t.frame = A|FF_ADD|FF_FULLBRIGHT
			t.scale = mo.scale
			t.destscale = 1
			t.scalespeed = mo.scale/12
								
			t.momx = (ais.x - t.x)/10
			t.momy = (ais.y - t.y)/10
			t.momz = (ais.z - t.z)/10
		end
		
	elseif tt == aimtime
		
		local pmo = mo.target
		local ais = mo.aises[mo.aimingfrom]
		
		local fn = function(s)
		
			if P_CheckSight(s, ais)
				s.sprite = SPR_THOK
				s.frame = FF_FULLBRIGHT
				s.color = SKINCOLOR_EMERALD
				s.destscale = 1
				s.scalespeed = FRACUNIT/12
				s.fuse = 12
				s.tics = 12
			else
				s.state = S_INVISIBLE
			end
		end
			
		S_StartSound(ais, sfx_s3k54)
		S_StartSound(pmo, sfx_s3k54)
		
		PSO.lib.spawnLine(ais, pmo, 128, fn)
		
		if P_CheckSight(ais, pmo)
			local hb = PSO.lib.spawnHurtBox(pmo.x, pmo.y, pmo.z, 80<<FRACBITS, 40<<FRACBITS, mo, beam_potency, DMT_RANGED, 3, RT_BLOWDOWN)
			hb.angle = R_PointToAngle2(ais.x, ais.y, pmo.x, pmo.y)	
			hb.pso.nododge = true
		end
	
	elseif tt == aimtime + TICRATE
		mo.aiming = TICRATE -1
		mo.aimingcount = $+1
		
		if mo.aimingcount > 1
			mo.aiming = nil
			mo.aimingfrom = nil
			mo.pso.targetlock = nil
		else
			S_StartSound(nil, sfx_trpowr)
		end
	end
end

PSO.info.enemies[MT_PSO_EXECUTOR] = {
		
		name = ".executor",
		sprite = SPR_PEXE,
		frame = A,
		painframe = A,
		
		bhp = 5000,
		satk = 35,
		ratk = 35,
		tatk = 35,
		def = 25,
		exp = 300,
		weight = 4,

		boss = true,		-- enemy is a major boss!
		bossmusic = "BOS4_1",		
		
		
		drops = {

			-- lv30+
			{"orakl sword", 1, 50, 1, 300},
			{"orakl doublesaber", 1, 50, 1, 300},
			{"orakl soaringblade", 1, 50, 1, 300},
			{"orakl partizan", 1, 50, 1, 300},
			{"orakl launcher", 1, 50, 1, 300},
			{"orakl tcsm", 1, 50, 1, 300},
			{"orakl rifle", 1, 50, 1, 300},
			{"orakl shot", 1, 50, 1, 300},
			{"orakl orphan", 1, 50, 1, 300},
			{"orakl rod", 1, 50, 1, 300},
			{"orakl wand", 1, 50, 1, 300},
			{"fehler frame", 1, 50, 5, 50},
			{"hart frame", 1, 50, 1, 50},
			{"digrinder", 1, 50, 5, 50},
			{"trigrinder", 1, 50, 4, 50},
			
			{"unit_hpsatk_2", 1, 50, 1, 20},
			{"unit_hpratk_2", 1, 50, 1, 20},
			{"unit_hptatk_2", 1, 50, 1, 20},
			{"unit_rainbow_2", 1, 50, 1, 20},
			
			-- Lv51+
			
			{"abnormal shard", 51, 79, 1, 4000},
			
			{"orakl sword", 51, 200,1, 150},
			{"orakl doublesaber", 51, 200,1, 150},
			{"orakl soaringblade", 51, 200,1, 150},
			{"orakl partizan", 51, 200,1, 150},
			{"orakl launcher", 51, 200,1, 150},
			{"orakl tcsm", 51, 200,1, 150},
			{"orakl rifle", 51, 200,1, 150},
			{"orakl shot", 51, 200,1, 150},
			{"orakl orphan", 51, 200,1, 150},
			{"orakl rod", 51, 200,1, 150},
			{"orakl wand", 51, 200,1, 150},
			
			{"makel frame", 51, 79,1, 150},
			{"aberrant frame", 51, 79, 1, 200},
			{"flow frame", 51, 79,1, 60},
			{"steel hearts", 51, 79,1, 700},
			{"seal case", 51, 79,1, 700},
			{"zanba", 51, 79, 1, 1000},
			{"last survivor", 51, 79, 1, 1000},
			{"chainsawd", 51, 79, 1, 4000},
			{"dragon slayer", 51, 79, 1, 4000},
			{"monkey king bar", 51, 79,1, 700},
			{"girasole", 51, 79,1, 700},
			{"d. saber legacy", 51, 79, 1, 1000},
			{"eternal heat", 51, 79, 1, 1000},
			{"vivienne", 51, 79, 1, 4000},
			{"double cannon", 51, 79, 1, 4000},			
			{"basilleon", 51, 79,1, 700},
			{"longinus", 51, 79,1, 700},
			{"meteor lancer", 51, 79, 1, 1000},
			{"brionac", 51, 79, 1, 1000},
			{"vjaya", 51, 79, 1, 4000},
			{"soul eater", 51, 79, 1, 4000},			
			{"sarblight", 51, 79,1, 700},
			{"forblight", 51, 79,1, 700},	
			{"grand cross", 51, 79, 1, 1000},
			{"red saber", 51, 79, 1, 1000},
			{"elysion", 51, 79, 1, 4000},
			{"axeon", 51, 79, 1, 4000},
			{"full cylinder", 51, 79,1, 700},
			{"assassin", 51, 79,1, 700},
			{"yasminkov 3000r", 51, 79, 1, 1000},
			{"strauss", 51, 79, 1, 1000},
			{"firearms", 51, 79, 1, 4000},
			{"frozen shooter", 51, 79, 1, 4000},			
			{"wide smack", 51, 79,1, 700},
			{"heimdall", 51, 79,1, 700},
			{"barett", 51, 79, 1, 1000},
			{"crush bullet", 51, 79, 1, 1000},
			{"spread needle", 51, 79, 1, 4000},
			{"final impact", 51, 79, 1, 4000},			
			{"photon launcher", 51, 79,1, 700},
			{"saint kilda", 51, 79,1, 700},
			{"charge howzer", 51, 79, 1, 1000},
			{"gur bazga", 51, 79, 1, 1000},
			{"sspm launcher", 51, 79, 1, 4000},
			{"inferno bazooka", 51, 79, 1, 4000},			
			{"yasminkov 2000h", 51, 79,1, 700},
			{"sarcueid", 51, 79,1, 700},
			{"varista", 51, 79, 1, 1000},
			{"bravace", 51, 79, 1, 1000},
			{"pandora extreme", 51, 79, 1, 4000},
			{"ruby bullet", 51, 79, 1, 4000},			
			{"clarita visas", 51, 79,1, 700},
			{"red processor", 51, 79,1, 700},
			{"battle verge", 51, 79, 1, 1000},
			{"brave hammer", 51, 79, 1, 1000},
			{"siren hammer", 51, 79, 1, 4000},
			{"demonic fork", 51, 79, 1, 4000},			
			{"windmill", 51, 79,1, 700},
			{"hexigraph", 51, 79,1, 700},
			{"serdote", 51, 79, 1, 1000},
			{"rose quartz", 51, 79, 1, 1000},
			{"rykros staff", 51, 79, 1, 4000},
			{"evil curst", 51, 79, 1, 4000},			
			{"tcsm_crawla4", 51, 79,1, 700},
			
			{"unit_hpsatk_3", 51, 200,1, 50},
			{"unit_hpratk_3", 51, 200,1, 50},
			{"unit_hptatk_3", 51, 200,1, 50},
			{"unit_ppsatk_3", 51, 200,1, 50},
			{"unit_ppratk_3", 51, 200,1, 50},
			{"unit_pptatk_3", 51, 200,1, 50},			
			{"unit_power_3", 51, 200,1, 50},
			{"unit_rainbow_3", 51, 200,1, 50},
				
			{1000, nil, nil, 1, 1},
		},
		
		stagger_duration = TICRATE/2,
		stagger = executor_pain,
		
		attribute = ATTR_BADNIK,
		weakness = ELEM_ELEC,
		weakness2 = ELEM_LIGHT,
		
		reset =	function(mo)
			mo.actiontimer = 0	-- reset action if we get staggered etc.
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			local maxmoves = 2
			
			PSO.lib.enemyManageAggro(mo)
			
			mo.color = SKINCOLOR_RED
			
			mo.spritexscale = mo.scale*2
			mo.spriteyscale = mo.scale*2
			mo.sprite = SPR_PEXE
			
			if not mo.turretpoints
				executor_linkPoints(mo)
				mo.executorstate = EXE_NEUTRAL
				
				mo.acttimer = TICRATE*6
				mo.movetimes = 0
			end
			
			if not mo.acttimer
				
				if mo.activeenemies and #mo.activeenemies
				or mo.aiming
					mo.acttimer = TICRATE*12
				else
					mo.acttimer = TICRATE*5
				end
			end
			
			executor_idle(mo)
			executor_checkPain(mo)
			
			if mo.executorstate == EXE_NEUTRAL
			and mo.target and mo.target.valid
				mo.acttimer = max(0, $-1)
			end
			
			if not mo.acttimer
			and mo.executorstate <= EXE_PAIN
			and mo.target and mo.target.valid
				
				--executor_prepTurret(mo)
				--executor_prepAcid(mo)
				--executor_prepSpawn(mo)
				--executor_fireMissiles(mo)
				
				if mo.pso.hp < (mo.pso.maxhp*6)/10
				and not mo.phase2_t
					executor_prepPhase2(mo)
				
					return
				elseif mo.phase2_t and not mo.open_shot
					executor_prepAiming(mo)
					mo.open_shot = true
				end
				
				if mo.movetimes < maxmoves
					executor_prepMove(mo, true)
					mo.movetimes = $+1
				else
				
					mo.movetimes = 0
					
					local actions = {executor_prepTurret, executor_prepAcid}
					if mo.acid_up
						table.remove(actions, #actions)
					end
					
					if not mo.activeenemies or not #mo.activeenemies
						table.insert(actions, executor_prepSpawn)
					end
					
					local rnd = P_RandomRange(1, #actions)
					actions[rnd](mo)
				end
				
				/*if not mo.phase2_t
					executor_prepPhase2(mo)
				
				elseif not mo.begin_shot
					executor_prepAiming(mo)
					mo.begin_shot = true
				else
					executor_prepMove(mo, true)
				end*/
			end
			
			-- handle basic movement
			executor_Move(mo)
			
			-- handle turret attack
			executor_doTurret(mo)
			
			-- handle teleportation
			executor_doTeleport(mo)
			
			-- handle acid
			executor_doAcid(mo)
			
			-- handle enemy support waves
			if mo.target and mo.target.valid
				executor_doSpawns(mo)
			end
			
			-- phase 2 transition
			executor_doPhaseTransition(mo)
			
			-- AIS aiming
			executor_doAiming(mo)
			
			if mo.phase2_t
				if mo.aiming == nil
					mo.aimingcd = $ and $+1 or 1
				end
				
				if mo.aimingcd and mo.aimingcd >= TICRATE*15
					executor_prepAiming(mo)
					mo.aimingcd = nil
				end
			end
		end
}