/*
	Gylir Cannon
	Winds up and begins shooting an infinite barrage until defeated
*/

freeslot("MT_PSO_GYLIR_CANNON", "SPR_HRBG", "SPR_HRB2", "S_PSO_HRBG_DIE")


mobjinfo[MT_PSO_GYLIR_CANNON] = {
	doomednum = 2701,
	spawnstate = S_INVISIBLE,
	deathstate = S_PSO_HRBG_DIE,
	spawnhealth = 1000,
	radius = 48<<FRACBITS,
	height = 40<<FRACBITS,
	flags = MF_ENEMY|MF_ENEMY|MF_NOGRAVITY|MF_SHOOTABLE,
}

function A_HarbingerDeath(mo)
	for i = 1, 16
		
		local t = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		t.color = i & 1 and SKINCOLOR_WHITE or SKINCOLOR_MASTER
		t.fuse = 11
		t.destscale = 1
		t.momx = P_RandomRange(-32, 32)*FRACUNIT
		t.momy = P_RandomRange(-32, 32)*FRACUNIT
		t.momz = P_RandomRange(-32, 32)*FRACUNIT
		
	end
	
	S_StartSound(mo, sfx_cdfm28)
	S_StartSound(mo, sfx_cdfm36)
end

states[S_PSO_HRBG_DIE] = {SPR_NULL, A, TICRATE, A_HarbingerDeath, 0, 0, S_NULL}

local function spawn_elements(mo)

	local diff = PSO.lib.questDifficulty()
	local spr = diff >= DIFF_HARD and SPR_HRB2 or SPR_HRBG
	local bright = (diff >= DIFF_HARD and FF_FULLBRIGHT or 0)
	
	mo.elements = {}
	for i = 1, 4
		local p = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		p.sprite = spr
		p.frame = B|FF_PAPERSPRITE|bright
		p.tics = -1
		mo.elements[#mo.elements+1] = p
	end
	
	local halo = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	halo.sprite = SPR_BFLW
	halo.frame = D|FF_ADD|FF_PAPERSPRITE|FF_FULLBRIGHT
	halo.color = SKINCOLOR_MASTER
	halo.tics = -1
	mo.halo = halo
	
	if diff >= DIFF_HARD
		mo.halo.flags2 = $|MF2_DONTDRAW
	end
	
	mo.v_angle = 0
	mo.v_dist = 12
	mo.shoottimer = 0
end

local function update_elements(mo)

	local basedist = 16
	local sx = mo.x + FixedMul(mo.scale, cos(mo.angle)*basedist)
	local sy = mo.y + FixedMul(mo.scale, sin(mo.angle)*basedist)

	local h_an = mo.angle + ANG1*90
	for i = 1, #mo.elements

		local v_an = mo.v_angle + ANG1*90*(i-1)

		-- honestly? I hate maths.
		local pos = FixedMul(mo.scale, sin(v_an)*mo.v_dist)
		local tx = sx + FixedMul(cos(h_an), pos)
		local ty = sy + FixedMul(sin(h_an), pos)
		local tz = mo.z + FixedMul(mo.scale, mo.v_dist*cos(v_an))

		P_MoveOrigin(mo.elements[i], tx, ty, tz)
		mo.elements[i].rollangle = -(v_an - ANG1*90)
		mo.elements[i].angle = mo.angle + ANG1*90
		mo.elements[i].tics = 2
		
		--print(tostring(v_an / ANG1).." | "..tostring(mo.elements[i].rollangle/ANG1))
	end	
	
	local bx = mo.x - FixedMul(mo.scale, cos(mo.angle)*basedist)
	local by = mo.y - FixedMul(mo.scale, sin(mo.angle)*basedist)
	P_MoveOrigin(mo.halo, bx, by, mo.z)
	mo.halo.angle = mo.angle + ANG1*90
	mo.halo.rollangle = $ + ANG1
	mo.halo.tics = 2
end
	
PSO.info.enemies[MT_PSO_GYLIR_CANNON] = {
		
		name = "Gylir Cannon",
		sprite = SPR_HRBG,
		frame = A,
		painframe = A,
		
		bhp = 200,
		satk = 20,
		ratk = 30,
		tatk = 20,
		def = 25,
		exp = 11,
		weight = 4,
		
		
		drops = {

			-- Lv1-50
			{650, 1, 200, 1, 15},
			{"beamgun", 1, 200, 1, 30},
			{"alva beamgun", 1, 200, 1, 45},
			{"vita beamgun", 1, 200, 1, 60},
			{"orakl orphan", 1, 200, 1, 75},
			{"rostig frame", 1, 200, 1, 20},		
			{"fehler frame", 1, 200, 1, 50},
			{"hart frame", 1, 200, 1, 75},
			{"monomate", 1, 50, 1, 35},
			{"dimate", 1, 200, 1, 35},
			{"trimate", 1, 200, 1, 35},

			{"baston frame", 51, 79,1, 200},
			{"makel frame", 51, 79, 1, 800},
			{"crimson frame", 51, 79, 1, 800},
			{"yasminkov 2000h", 51, 79, 1, 400},
			{"sarcueid", 51, 79, 1, 400},
			{"bravace", 51, 79, 1, 700},
			{"varista", 51, 79, 1, 700},			
			{"rostig frame", 51, 79,1, 30},
			{"fehler frame", 51, 79,1, 80},
			
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_HARBINGER,
		weakness = ELEM_DARK,
		weakness2 = ELEM_FIRE,
		
		reset =	function(mo)
			if not mo.pso.blowdown
				mo.flags = $|MF_NOGRAVITY
				mo.momz = 0
			end
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			local attack_potency = 70
			
			local shoot_start = TICRATE*3/2
			local shoot_windup = TICRATE*2

			PSO.lib.enemyManageAggro(mo)
			PSO.lib.enemySetCorruption(mo)

			mo.momz = 0

			if not mo.elements
				spawn_elements(mo)
			end
			
			if mo.prevhp and mo.pso.hp < mo.prevhp
				mo.refocus = TICRATE*2/3
				mo.shoottimer = 0
				mo.v_dist = 48
				mo.v_angle = 0
			end
			
			if mo.refocus
				
				mo.refocus = $-1
				
				mo.v_dist = leveltime & 1 and 48 or 32
				
				if mo.refocus == 0
					mo.v_dist = 12
				end
			end
			
			if mo.target and mo.target.valid and not mo.refocus
				
				mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
				
				mo.shoottimer = $+1
				
				local targetshoot = shoot_start + shoot_windup
				
				if mo.shoottimer > targetshoot
					
					mo.v_angle = $ + ANG1*24
					mo.v_dist = leveltime%4 < 2 and 24 or 7
					
					if mo.shoottimer%10 == 0

						local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, mo.angle, 45*FRACUNIT, attack_potency, DMT_RANGED, 0, 0, 0, RT_NONE)
						PSO.lib.aimMissile(msl, mo.target)
							
						msl.color = SKINCOLOR_RED
						msl.scale = $/4
						
						S_StartSound(mo, sfx_trfire)
					end
					
				elseif mo.shoottimer > shoot_start
					
					mo.v_dist = 24
					
					if mo.shoottimer == shoot_start+1
						S_StartSound(mo, sfx_s3k5a)
					end
					
					if mo.shoottimer > shoot_start+ TICRATE/3
						
						if mo.shoottimer == shoot_start + TICRATE/3 +1
							S_StartSound(mo, sfx_s3k60)
						end
						
						mo.v_angle = $ + ANG1*mo.shoottimer/8
					end
					
				end
			end
			
			update_elements(mo)
			
			mo.prevhp = mo.pso.hp
			mo.flags = $ & ~(MF_FLOAT|MF_NOGRAVITY)
		end,
}