/*
	Gylir Drill
	Moves around, winds up and drills on floor towards the player, leaves damaging light particles
*/

freeslot("MT_PSO_GYLIR_DRILL")


mobjinfo[MT_PSO_GYLIR_DRILL] = {
	doomednum = 2702,
	spawnstate = S_INVISIBLE,
	deathstate = S_PSO_HRBG_DIE,
	spawnhealth = 1000,
	radius = 48<<FRACBITS,
	height = 40<<FRACBITS,
	flags = MF_ENEMY|MF_ENEMY|MF_SHOOTABLE,
}

local function spawn_elements(mo)

	local diff = PSO.lib.questDifficulty()
	local spr = diff >= DIFF_HARD and SPR_HRB2 or SPR_HRBG
	local bright = (diff >= DIFF_HARD and FF_FULLBRIGHT or 0)
	
	mo.elements = {}
	for i = 1, 4
		local p = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		p.sprite = spr
		p.frame = C|bright
		p.renderflags = $ |(RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD|RF_NOSPLATROLLANGLE)
		P_CreateFloorSpriteSlope(p)	-- used for later
		p.tics = -1
		mo.elements[#mo.elements+1] = p
	end
	
	local halo = P_SpawnMobj(mo.x, mo.y, mo.z + FRACUNIT*4, MT_THOK)
	halo.sprite = SPR_BFLW
	halo.frame = D|FF_ADD
	halo.color = SKINCOLOR_MASTER
	halo.renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD|RF_FULLBRIGHT
	halo.tics = -1
	mo.halo = halo
	
	if diff >= DIFF_HARD
		mo.halo.flags2 = $|MF2_DONTDRAW
	end
	
	local crystal = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	crystal.sprite = spr
	crystal.frame = A|bright
	crystal.tics = -1
	mo.crystal = crystal
	
	mo.h_angle = 0
	mo.h_speed = ANG2
	mo.drilltimer = 0
end

local function update_elements(mo)
	
	local basedist = 20
	
	for i = 1, #mo.elements 
		
		local a = mo.h_angle + (i-1)*ANG1*90
		local x = mo.x + basedist*cos(a)
		local y = mo.y + basedist*sin(a)
		local z = mo.z + 128*mo.scale
		
		local p = mo.elements[i]
		P_MoveOrigin(p, x, y, z)
		p.angle = a
		p.spriteyscale = FRACUNIT/4
		p.tics = 2

		local origin = {x=x, y=y, z=z}
		local xydir = a
		local za = ANG1*80
		
		p.floorspriteslope.o = origin
		p.floorspriteslope.xydirection = xydir
		p.floorspriteslope.zangle = za
		
	end
	
	P_MoveOrigin(mo.halo, mo.x, mo.y, mo.z)
	mo.halo.rollangle = $ + ANG2
	mo.halo.tics = 2
	
	P_MoveOrigin(mo.crystal, mo.x, mo.y, mo.z + mo.scale*140)
	mo.crystal.tics = 2
	
	mo.h_angle = $ - mo.h_speed
	
	/*local basedist = 16
	local sx = mo.x + FixedMul(mo.scale, cos(mo.angle)*basedist)
	local sy = mo.y + FixedMul(mo.scale, sin(mo.angle)*basedist)

	local h_an = mo.angle + ANG1*90
	for i = 1, #mo.elements

		local v_an = mo.v_angle + ANG1*90*(i-1)

		-- honestly? I hate maths.
		local pos = FixedMul(mo.scale, sin(v_an)*mo.v_dist)
		local tx = sx + FixedMul(cos(h_an), pos)
		local ty = sy + FixedMul(sin(h_an), pos)
		local tz = mo.z + FixedMul(mo.scale, mo.v_dist*cos(v_an))

		P_MoveOrigin(mo.elements[i], tx, ty, tz)
		mo.elements[i].rollangle = -(v_an - ANG1*90)
		mo.elements[i].angle = mo.angle + ANG1*90
		mo.elements[i].tics = 2
		
		--print(tostring(v_an / ANG1).." | "..tostring(mo.elements[i].rollangle/ANG1))
	end	
	
	local bx = mo.x - FixedMul(mo.scale, cos(mo.angle)*basedist)
	local by = mo.y - FixedMul(mo.scale, sin(mo.angle)*basedist)
	P_MoveOrigin(mo.halo, bx, by, mo.z)
	mo.halo.angle = mo.angle + ANG1*90
	mo.halo.rollangle = $ + ANG1
	mo.halo.tics = 2*/
end
	
PSO.info.enemies[MT_PSO_GYLIR_DRILL] = {
		
		name = "Gylir Drill",
		sprite = SPR_NULL,
		frame = A,
		painframe = A,
		
		bhp = 270,
		satk = 30,
		ratk = 20,
		tatk = 30,
		def = 25,
		exp = 12,
		weight = 4,
		
		
		drops = {

			-- Lv1-50
			{650, 1, 200, 1, 15},
			{"crest", 1, 200, 1, 30},
			{"alva crest", 1, 200, 1, 45},
			{"vita crest", 1, 200, 1, 60},
			{"orakl rod", 1, 200, 1, 75},
			{"rostig frame", 1, 200, 1, 20},		
			{"fehler frame", 1, 200, 1, 50},
			{"hart frame", 1, 200, 1, 75},
			{"monomate", 1, 50, 1, 35},
			{"dimate", 1, 200, 1, 35},
			{"trimate", 1, 200, 1, 35},

			{"baston frame", 51, 79,1, 200},
			{"makel frame", 51, 79, 1, 800},
			{"crimson frame", 51, 79, 1, 500},
			{"red processor", 51, 79, 1, 200},
			{"clarita visas", 51, 79, 1, 200},
			{"brave hammer", 51, 79, 1, 500},
			{"battle verge", 51, 79, 1, 500},			
			{"rostig frame", 51, 79,1, 30},
			{"fehler frame", 51, 79,1, 80},
			
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_HARBINGER,
		weakness = ELEM_DARK,
		weakness2 = ELEM_ICE,
		
		reset =	function(mo)
			if not mo.pso.blowdown
				mo.flags = $|MF_NOGRAVITY
				mo.momz = 0
			end
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			local attack_potency = 50
			
			local drill_start = TICRATE*4
			local drill_windup = TICRATE*2
			local attack_potency = 250
			local grants_potency = 100

			PSO.lib.enemyManageAggro(mo)
			PSO.lib.enemySetCorruption(mo)

			if not mo.elements
				spawn_elements(mo)
			end
			
			if mo.target and mo.target.valid
				
				mo.drilltimer = $+1
				
				local drill_target = drill_start + drill_windup
				
				if mo.drilltimer > drill_target
					
					local t = mo.drilltimer - drill_target
					
					if t == 1
						mo.momz = -mo.scale*32
						mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
					end
					
					if P_IsObjectOnGround(mo)
						
						mo.groundtimer = $+1
						
						if mo.groundtimer == 10
							P_InstaThrust(mo, mo.angle, FRACUNIT*85)
							
							local hx = mo.x + 64*cos(mo.angle)
							local hy = mo.y + 64*sin(mo.angle)
							local hb = PSO.lib.spawnHurtBox(hx, hy, mo.z, 80<<FRACBITS, 120<<FRACBITS, mo, attack_potency, DMT_STRIKE, TICRATE/3, RT_BLOWDOWN)
							mo.hitbox = hb
							hb.angle = mo.angle
							
						elseif mo.groundtimer > 10
							local hx = mo.x + 64*cos(mo.angle)
							local hy = mo.y + 64*sin(mo.angle)
							
							if mo.hitbox and mo.hitbox.valid
								P_SetOrigin(mo.hitbox, hx, hy, mo.z)
							end
							
							if FixedHypot(mo.momx, mo.momy) < FRACUNIT*5
								mo.h_speed = max(ANG1*2, $ - ANG1)
								
								if mo.h_speed <= ANG1*2
									mo.drilltimer = 0
									mo.groundtimer = 0
								end
								
							elseif leveltime%4 == 0
								local msl = PSO.lib.spawnMissile(mo, MT_PSO_GRANTS, 0, 0, grants_potency, DMT_TECHNIQUE, 0, 0, mo.height/2, RT_STAGGER)
								S_StartSound(msl, sfx_s243)
							end
							
						end
					end
					
				elseif mo.drilltimer > drill_start
					
					mo.h_speed = $ + ANG1
					mo.groundtimer = 0
					
				else
					--mo.momz = 0
					mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
					
					P_InstaThrust(mo, mo.angle, FRACUNIT/2)
				end
				
			else
				P_InstaThrust(mo, 0, 0)
				mo.momz = 0

			end	
			
			update_elements(mo)

			if PSO.lib.enemyPitCheck(mo)
				P_InstaThrust(mo, 0, 0)
			end
			
			mo.flags = $ & ~(MF_FLOAT|MF_NOGRAVITY)
		end,
}