/*
	Gylir Shredder
	Perpetually follows its target and deals damage in a short radius around itself.
*/

freeslot("MT_PSO_GYLIR_SHREDDER")


mobjinfo[MT_PSO_GYLIR_SHREDDER] = {
	doomednum = 2706,
	spawnstate = S_INVISIBLE,
	deathstate = S_PSO_HRBG_DIE,
	spawnhealth = 1000,
	radius = 48<<FRACBITS,
	height = 40<<FRACBITS,
	flags = MF_ENEMY|MF_ENEMY|MF_SHOOTABLE,
}

local function spawn_elements(mo)

	local diff = PSO.lib.questDifficulty()
	local spr = diff >= DIFF_HARD and SPR_HRB2 or SPR_HRBG
	local bright = (diff >= DIFF_HARD and FF_FULLBRIGHT or 0)
	
	mo.elements = {}
	for i = 1, 4
		local p = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		p.sprite = spr
		p.frame = T|bright
		p.tics = -1
		p.scale = mo.scale*3/2
		p.spritexoffset = FRACUNIT*25
		--p.spritexscale = $*2

		p.renderflags = $ |(RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD|RF_NOSPLATROLLANGLE)
		P_CreateFloorSpriteSlope(p)
		mo.elements[#mo.elements+1] = p
	end
	
	local halo = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	halo.sprite = SPR_BFLW
	halo.frame = D|FF_ADD|FF_FULLBRIGHT
	halo.renderflags = $ |(RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD|RF_NOSPLATROLLANGLE)
	halo.color = SKINCOLOR_MASTER
	halo.tics = -1
	mo.halo = halo
	
	if diff >= DIFF_HARD
		mo.halo.flags2 = $|MF2_DONTDRAW
	end
end

local function update_elements(mo)

	local basedist = 48
	local a = leveltime*ANG1*20
	local a2 = leveltime*ANG1

	for i = 1, #mo.elements

		-- honestly? I hate maths.
		local ea = a + i*ANG1*90
		local tx = mo.x + FixedMul(mo.scale, basedist*cos(ea))
		local ty = mo.y + FixedMul(mo.scale, basedist*sin(ea))
		local tz = mo.z

		P_MoveOrigin(mo.elements[i], tx, ty, tz)
		--mo.elements[i].rollangle = ea
		mo.elements[i].angle = ea
		mo.elements[i].tics = 2
		mo.elements[i].floorspriteslope.o = {x=mo.x, y=mo.y, z=mo.z}
		mo.elements[i].floorspriteslope.xydirection = a2
		mo.elements[i].floorspriteslope.zangle = ANG1*7
		mo.elements[i].scale = mo.scale
		
		local g = P_SpawnGhostMobj(mo.elements[i])
		g.tics = 3
		P_CreateFloorSpriteSlope(g)
		g.floorspriteslope.o = {x=mo.x, y=mo.y, z=mo.z}
		g.floorspriteslope.xydirection = a2
		g.floorspriteslope.zangle = ANG1*7
	end	
	
	P_MoveOrigin(mo.halo, mo.x, mo.y, mo.z + mo.scale*32)
	mo.halo.scale = mo.scale
	mo.halo.rollangle = $ + ANG1
	mo.halo.tics = 2
end
	
PSO.info.enemies[MT_PSO_GYLIR_SHREDDER] = {
		
		name = "Gylir Shredder",
		sprite = SPR_HRBG,
		frame = A,
		painframe = A,
		
		bhp = 150,
		satk = 30,
		ratk = 20,
		tatk = 20,
		def = 25,
		exp = 11,
		weight = 4,
		
		
		drops = {

			-- Lv1-50
			{650, 1, 200, 1, 15},
			{"beamgun", 1, 200, 1, 30},
			{"alva beamgun", 1, 200, 1, 45},
			{"vita beamgun", 1, 200, 1, 60},
			{"orakl orphan", 1, 200, 1, 75},
			{"rostig frame", 1, 200, 1, 20},		
			{"fehler frame", 1, 200, 1, 50},
			{"hart frame", 1, 200, 1, 75},
			{"monomate", 1, 50, 1, 35},
			{"dimate", 1, 200, 1, 35},
			{"trimate", 1, 200, 1, 35},

			{"baston frame", 51, 79,1, 200},
			{"makel frame", 51, 79, 1, 800},
			{"crimson frame", 51, 79, 1, 800},
			{"yasminkov 2000h", 51, 79, 1, 400},
			{"sarcueid", 51, 79, 1, 400},
			{"bravace", 51, 79, 1, 700},
			{"varista", 51, 79, 1, 700},			
			{"rostig frame", 51, 79,1, 30},
			{"fehler frame", 51, 79,1, 80},
			
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_HARBINGER,
		weakness = ELEM_DARK,
		weakness2 = ELEM_ELEC,
		
		reset =	function(mo)
			if not mo.pso.blowdown
				mo.flags = $|MF_NOGRAVITY
				mo.momz = 0
			end
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			local movespeed = 4*FRACUNIT
			local attack_potency = 200
			local targetz = mo.floorz + mo.scale*32
			local rate = FRACUNIT*8/10

			if mo.spawnz == nil
				mo.spawnz = mo.z 
			end
			
			if mo.spawnz - targetz > FRACUNIT*320
				targetz = mo.spawnz + mo.scale*32
			end
			PSO.lib.enemyManageAggro(mo)
			
			mo.flags = $ & ~(MF_FLOAT|MF_NOGRAVITY)
				
			if mo.z < targetz
				mo.momz = $ + rate
			elseif mo.momz < 0
				mo.momz = min(-mo.scale, $ + rate*2)
			end

			if not mo.elements
				spawn_elements(mo)
			end
			
			if mo.target and mo.target.valid
				A_FaceTarget(mo)
				P_InstaThrust(mo, mo.angle, movespeed)
				
				if PSO.lib.enemyPitCheck(mo)
					P_InstaThrust(mo, 0, 0)
				end
			end
			
			update_elements(mo)
			
			if leveltime%10 == 0
				local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z-mo.height, 192<<FRACBITS, 64<<FRACBITS, mo, attack_potency, DMT_MELEE, 3, RT_NONE)
			end
		end,
}