/*
	Gyrrey Glardia
	3 parts, main body (Dark Falz), and the 2 "pawns".
	
*/

local GRID_SIZE = 5
local GRID_TAG = 400
local OFF_GRID_OFFSET = 160<<FRACBITS

local GRID_VISUAL_WIDTH = 12
local GRID_ATTACK_VISUAL = TICRATE/2	-- delay remaining that causes visual to show up

freeslot("MT_PSO_GYRREY_GLARDIA_MAIN", "MT_PSO_GYRREY_GLARDIA_PAWN", "S_PSO_GYRREY_GLARDIA_DEATH", "SPR_DAFA", "SPR_DFLA", "SPR_LITW", "SPR_LENS", "SPR_DFCH")

mobjinfo[MT_PSO_GYRREY_GLARDIA_MAIN] = {
	doomednum = 2557,
	spawnstate = S_INVISIBLE,
	deathstate = S_PSO_GYRREY_GLARDIA_DEATH,
	spawnhealth = 1000,
	radius = 32<<FRACBITS,
	height = 32<<FRACBITS,
	flags = MF_NOGRAVITY|MF_NOBLOCKMAP|MF_ENEMY|MF_NOCLIPTHING,	-- untargettable
}

mobjinfo[MT_PSO_GYRREY_GLARDIA_PAWN] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	deathstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 80<<FRACBITS,
	height = 127<<FRACBITS,
	flags = MF_NOGRAVITY|MF_ENEMY|MF_SHOOTABLE,	-- untargettable
}

-- dark falz body spawner...
PSO.lib.DF_spawn = function(mo, scale)
	
	mo.scale = scale or FRACUNIT*4	-- adjust
	mo.shoulder_angle = -90
	mo.shoulder_angle_target = -90
	
	mo.bobamp = 15
	mo.hbobamp = 15
	
	mo.target_bobamp = 15
	mo.target_hbobamp = 15
	
	mo.basez = mo.z
	
	-- torso
	mo.sprite = SPR_DAFA
	mo.frame = A
	
	mo.layer = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	mo.layer.sprite = SPR_DFLA
	mo.layer.frame = A
	mo.layer.scale = mo.scale
	mo.layer.tics = -1

	mo.hlayer = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	mo.hlayer.sprite = SPR_DFLA
	mo.hlayer.frame = A
	mo.hlayer.scale = mo.scale
	mo.hlayer.tics = -1

	-- back halo
	local halo = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	halo.sprite = SPR_BFLW
	halo.frame = D|FF_ADD|FF_PAPERSPRITE|FF_FULLBRIGHT
	halo.color = SKINCOLOR_MASTER
	halo.tics = -1
	halo.scale = mo.scale*5/2
	mo.halo = halo	

	local armhalo = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	armhalo.sprite = SPR_BFLW
	armhalo.frame = D|FF_ADD|FF_PAPERSPRITE|FF_FULLBRIGHT
	armhalo.color = SKINCOLOR_MASTER
	armhalo.tics = -1
	armhalo.scale = mo.scale
	mo.armhalo = armhalo
	
	-- head
	mo.head_height = 140*FRACUNIT
	
	local h = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	h.tics = 2
	h.sprite = SPR_DAFA
	h.frame = F
	h.angle = mo.angle
	h.scale = mo.scale
	mo.head = h
	
	-- spine
	local spinelen = 4
	mo.spine = {}
	
	for i = 1, spinelen
		local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		s.tics = 2
		s.sprite = SPR_DAFA
		s.scale = mo.scale
		s.angle = mo.angle
		s.frame = i == spinelen and E or D	-- final segment is the base.
		mo.spine[#mo.spine+1] = s
	end
	
	-- arm, elbow, forearm, hand
	local a = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	a.tics = 2
	a.sprite = SPR_DAFA
	a.frame = I|FF_PAPERSPRITE
	a.angle = mo.angle
	a.scale = mo.scale
	mo.arm = a
	
	mo.arm_hangle = 0
	mo.arm_vangle = 0

	a = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	a.tics = 2
	a.sprite = SPR_DAFA
	a.frame = J
	a.angle = mo.angle
	a.scale = mo.scale
	mo.elbow = a

	a = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	a.tics = 2
	a.sprite = SPR_DAFA
	a.frame = K|FF_PAPERSPRITE
	a.angle = mo.angle
	a.scale = mo.scale
	mo.forearm = a
	
	mo.farm_hangle = 0
	mo.farm_vangle = 0
	
	mo.wrist_hangle = 0
	mo.wrist_vangle = 0	-- hand

	a = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	a.tics = 2
	a.sprite = SPR_DAFA
	a.frame = L
	a.angle = mo.angle
	a.scale = mo.scale
	mo.hand = a
	
	mo.init = true	-- we initialized innit?
end

PSO.lib.DF_remnantParticles = function(x, y, z, range, scale)
	range = $/FRACUNIT
	x = $ + P_RandomRange(-range, range)*FRACUNIT
	y = $ + P_RandomRange(-range, range)*FRACUNIT
	z = $ + P_RandomRange(0, range)*FRACUNIT
	
	local s = P_SpawnMobj(x,y,z, MT_THOK)
	s.color = SKINCOLOR_BLACK
	s.frame = FF_SUBTRACT
	s.tics = 20
	s.fuse = 20
	s.scale = scale or FRACUNIT
	s.destscale = 1
	
	s.momz = P_RandomRange(4, 8)*s.scale
	s.momx = P_RandomRange(-4, 4)*s.scale
	s.momy = P_RandomRange(-4, 4)*s.scale
end

PSO.lib.DF_updateParts = function(mo)
	
	--if mo.nodraw then return end	-- we're invisible.
	
	local diff = PSO.lib.questDifficulty()
	local movefunc = mo.initmove and P_MoveOrigin or P_SetOrigin
	local basedist = 64
	
	local subflag = diff >= DIFF_HARD and FF_SUBTRACT or 0
	
	if mo.bobamp > mo.target_bobamp
		mo.bobamp = $-1
	elseif mo.bobamp < mo.target_bobamp
		mo.bobamp = $+1
	end

	if mo.hbobamp > mo.target_hbobamp
		mo.hbobamp = $-1
	elseif mo.hbobamp < mo.target_hbobamp
		mo.hbobamp = $+1
	end
	
	mo.z = mo.basez + mo.bobamp*cos(leveltime*ANG1*2)
	--mo.tics = -1
	
	mo.frame = $|subflag
	-- layer on chest
	if mo.frame & FF_FRAMEMASK == A
		mo.layer.flags2 = $ & ~MF2_DONTDRAW
		
		P_MoveOrigin(mo.layer, mo.x + 2*cos(mo.angle), mo.y + 2*sin(mo.angle), mo.z)
		-- frames go from A to N
		
		local maxframe = N
		local framedelay = 6
		local duration = (maxframe+1)*framedelay
		
		local frame = (leveltime/framedelay) % maxframe
		--print(frame)
		local translucency = 5 - ((4*cos(((360/maxframe)*frame)*ANG1))/FRACUNIT) --<< FF_TRANSSHIFT
		--print(translucency)
		translucency = $ << FF_TRANSSHIFT
		
		mo.layer.frame = frame|translucency|FF_FULLBRIGHT
		mo.layer.tics = -1
	elseif mo.layer and mo.layer.valid
		mo.layer.flags2 = $|MF2_DONTDRAW
	end

	-- halo
	local bx = mo.x - FixedMul(mo.scale, cos(mo.angle)*basedist)
	local by = mo.y - FixedMul(mo.scale, sin(mo.angle)*basedist)
	P_MoveOrigin(mo.halo, bx, by, mo.z + mo.scale*64)
	mo.halo.angle = mo.angle + ANG1*90
	mo.halo.rollangle = $ + ANG1
	mo.halo.tics = -1	
	
	local h = mo.head
	h.tics = 3
	local x = mo.x + 16*cos(mo.angle)
	local y = mo.y + 16*sin(mo.angle)
	local z = mo.z + FixedMul(mo.head_height, mo.scale) + mo.hbobamp*cos(leveltime*ANG1*4)
	movefunc(h, x, y, z)
	h.frame = $|subflag
	
	if diff >= DIFF_HARD
		PSO.lib.DF_remnantParticles(h.x, h.y, h.z, 384*FRACUNIT, FRACUNIT*2)
	end
	
	if mo.neck_base
		mo.neck_base.z = mo.z+FixedMul(mo.head_height, mo.scale)
	else
		mo.neck_base = {x=mo.x, y=mo.y, z=mo.z+FixedMul(mo.head_height, mo.scale)}
	end
	
	-- eye opening animation!
	if mo.openeye
		mo.openeye = $+1
		
		if mo.openeye == TICRATE/4
			h.frame = G
		elseif mo.openeye == TICRATE/2
			h.frame = H
			
			-- lens flare!
			local lx = x + 16*cos(mo.angle)
			local ly = y + 16*sin(mo.angle)
			local lz = z
			for i = 1, 2
				
				local l = P_SpawnMobj(lx, ly, lz, MT_THOK)
				l.sprite = SPR_LENS
				l.frame = A|FF_FULLBRIGHT|FF_ADD
				l.color = SKINCOLOR_MASTER
				l.scale = mo.scale
				l.destscale = mo.scale*5*i
				l.momz = mo.scale*4
				l.tics = 10*i
				l.fuse = l.tics
				l.scalespeed = mo.scale/6
			end
			S_StartSound(nil, sfx_s1c3)
			
		elseif mo.openeye > TICRATE/2
			
			-- get an overlay
			local lx = x + cos(mo.angle)
			local ly = y + sin(mo.angle)
			local lz = z
			
			local o = P_SpawnMobj(lx, ly, lz, MT_THOK)
			o.sprite = SPR_DFLA
			o.scale = mo.scale
			o.frame = O|FF_FULLBRIGHT
			o.tics = 2
			
		end
	elseif mo.closeeye
		mo.closeeye = $+1
		h.frame = G
		
		if mo.closeeye > TICRATE/2
			mo.closeeye = nil
			h.frame = F
		end
		
	end
	
	for i = 1, #mo.spine
		
		local s = mo.spine[i]
		s.tics = 3
		local dist = 20 - (i-1)*4
		local x = mo.x - dist*cos(mo.angle)
		local y = mo.y - dist*sin(mo.angle)
		
		local a = leveltime*ANG1 + ANG1*16*i
		
		local z = mo.z + mo.scale*8 - (i*16*mo.scale) + cos(a + ANGLE_22h*i)*14
		if i == #mo.spine
			z = $ - mo.scale*32
		end
		
		s.frame = $|subflag
		
		movefunc(s, x, y, z)
	end
	
	-- arm and stuff
	
	-- pivot for the shoulder
	local shoulder_dist = 150
	local shoulder_zdist = mo.shoulder_zdist or 50
	--local shoulder_angle = -90
	
	if (mo.frame & FF_FRAMEMASK) == B
		mo.shoulder_angle_target = -30
	elseif (mo.frame & FF_FRAMEMASK) == C
		mo.shoulder_angle_target = -110
	else
		mo.shoulder_angle_target = -90
	end
	
	if mo.shoulder_angle > mo.shoulder_angle_target
		mo.shoulder_angle = max(mo.shoulder_angle_target, $-8)
	elseif mo.shoulder_angle < mo.shoulder_angle_target
		mo.shoulder_angle = min(mo.shoulder_angle_target, $+8)
	end
	--print(mo.shoulder_angle)
	
	local sx = mo.head.x + FixedMul(mo.scale*shoulder_dist, cos(mo.angle +ANG1*mo.shoulder_angle))
	local sy = mo.head.y + FixedMul(mo.scale*shoulder_dist, sin(mo.angle +ANG1*mo.shoulder_angle))
	local sz = mo.head.z + - mo.scale*shoulder_zdist
	
	local pivot = {x=sx, y=sy, z=sz}
	
	local arm = mo.arm
	local arm_len = 120*FRACUNIT
	arm.tics = -1
	
	-- arm halo here
	P_MoveOrigin(mo.armhalo, sx, sy, sz)
	mo.armhalo.angle = mo.arm_hangle + ANG1*90
	mo.armhalo.rollangle = $-ANG1*2
	mo.armhalo.tics = 3
	
	-- arm location...
	local ax, ay, az = PSO.lib.anglesOnSphere(mo.arm_hangle, mo.arm_vangle, arm_len/2)
	
	P_MoveOrigin(arm, pivot.x + FixedMul(ax, mo.scale), pivot.y + FixedMul(ay, mo.scale), pivot.z + FixedMul(az, mo.scale))
	arm.angle = mo.arm_hangle
	arm.rollangle = mo.arm_vangle
	arm.frame = $|subflag
	
	local elbow = mo.elbow
	elbow.tics = -1
	local ax, ay, az = PSO.lib.anglesOnSphere(mo.arm_hangle, mo.arm_vangle, arm_len*9/10)
	P_MoveOrigin(elbow, pivot.x + FixedMul(ax, mo.scale), pivot.y + FixedMul(ay, mo.scale), pivot.z + FixedMul(az, mo.scale))
	elbow.frame = $|subflag
	
	-- hard mode: effect between arm and elbow...
	if diff >= DIFF_HARD
		for i = 1, 4
			local ax, ay, az = PSO.lib.anglesOnSphere(mo.arm_hangle, mo.arm_vangle, arm_len*i/5)
			PSO.lib.DF_remnantParticles(pivot.x + FixedMul(ax, mo.scale), pivot.y + FixedMul(ay, mo.scale), pivot.z + FixedMul(az, mo.scale), 128*FRACUNIT, FRACUNIT)	
		end
	end
	
	local forearm = mo.forearm
	local farm_len = 150
	forearm.tics = -1
	local ax, ay, az = PSO.lib.anglesOnSphere(mo.farm_hangle, mo.farm_vangle, arm_len/2 + arm_len/10)
	P_MoveOrigin(forearm, elbow.x + FixedMul(ax, mo.scale), elbow.y + FixedMul(ay, mo.scale), elbow.z + FixedMul(az, mo.scale))
	forearm.angle = mo.farm_hangle
	forearm.rollangle = mo.farm_vangle
	forearm.frame = $|subflag

	-- hard mode: effect between arm and elbow...
	if diff >= DIFF_HARD
		for i = 1, 4
			local ax, ay, az = PSO.lib.anglesOnSphere(mo.farm_hangle, mo.farm_vangle, arm_len*i/5)
			PSO.lib.DF_remnantParticles(elbow.x + FixedMul(ax, mo.scale), elbow.y + FixedMul(ay, mo.scale), elbow.z + FixedMul(az, mo.scale), 128*FRACUNIT, FRACUNIT)	
		end
	end
	
	-- wrist, kinda.
	local wristx, wristy, wristz = PSO.lib.anglesOnSphere(mo.farm_hangle, mo.farm_vangle, arm_len + arm_len/8)
	wristx = FixedMul($, mo.scale)+elbow.x
	wristy = FixedMul($, mo.scale)+elbow.y
	wristz = FixedMul($, mo.scale)+elbow.z
	
	if mo.wrist
		mo.wrist.x, mo.wrist.y, mo.wrist.z = wristx, wristy, wristz
	else
		mo.wrist = {x=wristx, y=wristy, z=wristz}	-- not a real object but do keep updating that for chains.
	end
		
	local hand = mo.hand
	hand.tics = -1
	local ax, ay, az = PSO.lib.anglesOnSphere(mo.wrist_hangle, mo.wrist_vangle, arm_len/7)
	if not mo.nodraw
		P_MoveOrigin(hand, wristx+FixedMul(ay, mo.scale), wristy+FixedMul(ay, mo.scale), wristz+FixedMul(az, mo.scale))
		hand.rollangle = mo.wrist_vangle
	end
	hand.frame = $|subflag
	if diff >= DIFF_HARD
		PSO.lib.DF_remnantParticles(hand.x, hand.y, hand.z, 384*FRACUNIT, FRACUNIT*2)
	end


	-- hand layer
	P_MoveOrigin(mo.hlayer, mo.hand.x + 2*cos(mo.angle), mo.hand.y + 2*sin(mo.angle), mo.hand.z)
	mo.hlayer.rollangle = mo.hand.rollangle
	mo.hlayer.scale = mo.hand.scale
	mo.hlayer.angle = mo.hand.angle
	mo.hlayer.frame = (mo.hand.frame & FF_FRAMEMASK)+4
	
	-- turn invisible for some specific hand frames.
	if (mo.hand.frame & FF_FRAMEMASK >= R)
		mo.hlayer.sprite = SPR_NULL
		mo.hlayer.frame = A
	else
		mo.hlayer.sprite = SPR_DFLA
	end
	
	-- frames go from A to N
		
	local maxframe = N
	local framedelay = 6
	local duration = (maxframe+1)*framedelay
	local frame = (leveltime/framedelay) % maxframe
	--print(frame)
	local translucency = 5 - ((4*cos(((360/maxframe)*frame)*ANG1))/FRACUNIT) --<< FF_TRANSSHIFT
	--print(translucency)
	translucency = $ << FF_TRANSSHIFT
		
	mo.hlayer.frame = $|translucency|FF_FULLBRIGHT
	mo.hlayer.tics = 3	
	
	mo.initmove = true
	mo.nodraw = (mo.flags2 & MF2_DONTDRAW)
end

-- opens DF's eye and has a glow...
PSO.lib.DF_openEye = function(mo)
	
	if not mo.openeye
		mo.openeye = 1
	end
end

-- closes it.
PSO.lib.DF_closeEye = function(mo)
	if mo.openeye
		mo.closeeye = 1
		mo.openeye = 0
	end
end

local function makeGrid(mo, starttag, gridsize)
	-- first of all get all the sectors... in order.
	
	local endtag = starttag + (gridsize^2) -1
	local grid = {}
	
	for s in sectors.iterate do
		if s and s.tag >= starttag and s.tag <= endtag
			grid[#grid+1] = s
		end
	end
	
	-- make sure we properly order this mess!!
	table.sort(grid, 
		function(a, b)
			return a.tag < b.tag
		end)
	
	-- now, prepare all the waypoints
	local refline = grid[1].lines[0]
	local sectorsize = R_PointToDist2(refline.v1.x, refline.v1.y, refline.v2.x, refline.v2.y)
	local z = grid[1].floorheight
	
	local gridpoints = {}
	local grid_top = {}
	local grid_left = {}
	local grid_right = {}
	local grid_bottom = {}
	
	for i = 1, #grid
		-- get the middle of the sector
		local addx = 0
		local addy = 0
		local processed_verts = {}
		local j = 0
		local s = grid[i]
		local nadd = 0
		
		while s.lines[j]
			local l = s.lines[j]
			
			local verts = {l.v1, l.v2}
			
			for k = 1, #verts
				if not processed_verts[verts[k]]
					processed_verts[verts[k]] = true	-- use key
					addx = $ + verts[k].x/FRACUNIT
					addy = $ + verts[k].y/FRACUNIT
					nadd = $+1
				end
			end
			
			
			j = $+1
		end
		
		addx = ($/nadd)*FRACUNIT
		addy = ($/nadd)*FRACUNIT
		
		local test = P_SpawnMobj(addx, addy, z, MT_THOK)
		test.color = SKINCOLOR_RED
		test.state = S_INVISIBLE
		test.tics = -1
		test.extravalue1 = i
		gridpoints[#gridpoints+1] = test
		
		-- now to know where we are for the other teleport points
		if (i - gridsize) <= 0	-- top row
			local x = addx
			local y = addy + sectorsize/2 + OFF_GRID_OFFSET
			local s = P_SpawnMobj(x, y, z+FRACUNIT*0, MT_THOK)
			s.state = S_INVISIBLE
			s.angle = ANG1*270
			s.tics = -1
			grid_top[#grid_top+1] = s
		end
		
		if (i-1) % gridsize == 0		-- left column
			local y = addy
			local x = addx - sectorsize/2 - OFF_GRID_OFFSET
			local s = P_SpawnMobj(x, y, z+FRACUNIT*0, MT_THOK)
			s.state = S_INVISIBLE
			s.tics = -1
			grid_left[#grid_left+1] = s
		end
		
		if (i-1 - gridsize +1) % gridsize == 0	-- right column
			local y = addy
			local x = addx + sectorsize/2 + OFF_GRID_OFFSET
			local s = P_SpawnMobj(x, y, z+FRACUNIT*0, MT_THOK)
			s.state = S_INVISIBLE
			s.tics = -1
			s.angle = ANG1*180
			grid_right[#grid_right+1] = s
		end
		
		if (i-1 + gridsize) > (gridsize^2)-1	-- bottom row
			local x = addx
			local y = addy - sectorsize/2 - OFF_GRID_OFFSET
			local s = P_SpawnMobj(x, y, z+FRACUNIT*0, MT_THOK)
			s.state = S_INVISIBLE
			s.tics = -1
			s.angle = ANG1*90
			grid_bottom[#grid_bottom+1] = s
		end
	end
	
	-- aaaand save all the grid data we might care about!
	mo.grid = grid
	mo.gridpoints = gridpoints
	mo.grid_top = grid_top
	mo.grid_left = grid_left
	mo.grid_right = grid_right
	mo.grid_bottom = grid_bottom
	mo.grid_size = sectorsize
	
	mo.grid_actions = {}
	
end

-- says which grid index test is closest to.
local function closestPointOnGrid(mo, test)
	
	local closest = 4096*FRACUNIT
	local index = 0
	
	for i = 1, #mo.gridpoints
		local p = mo.gridpoints[i]
		local dist = R_PointToDist2(test.x, test.y, p.x, p.y)
		if dist < closest
			closest = dist
			index = p.extravalue1
		end
	end
	
	return index
	
end

-- gives the "X" (1-5) coord of the tile on the grid...
local function gridX(index)
	return ((index-1) % GRID_SIZE) +1
end

-- same with "Y"
local function gridY(index)
	return ((index-1) / GRID_SIZE) +1
end

-- says which row the player is on!
local function whichRow(mo, pmo)
	
	return gridY(closestPointOnGrid(mo, pmo))
end

-- says which column the player is on!
local function whichColumn(mo, pmo)
	
	return gridX(closestPointOnGrid(mo, pmo))
end

local function setGridAttack(mo, index, delay, info)
	
	-- index: # index in grid where the attack takes place
	-- delay: tics before the attack takes place
	-- info: {type=MT_enum, angle=angle_t, potency=int, reaction=RT_enum, strength=int}
	
	if mo.grid[index]
		mo.grid_actions[index] = {
			delay = delay,
			angle = info.angle,
			potency = info.potency,
			reaction = info.reaction,
			strength = info.strength,
			duration = info.duration,
			damagetype = info.damagetype,
			quake = info.quake,	-- used for some floor effects
		}
	end
end

-- start an attack on a row
-- angle sets knockback direction (these are intended to be for laser-type attacks)
local function setRowAttack(mo, row, delay, potency, angle, duration)
	
	-- reminder: rows go from top to bottom
	-- 1
	-- 2
	-- 3
	-- 4
	-- 5
	
	local startindex = 1 + (row-1)*GRID_SIZE
	for i = startindex, startindex + GRID_SIZE-1
		setGridAttack(mo, i, delay, {angle=angle, potency=potency, reaction=RT_BLOWDOWN, duration=duration})
	end
end

local function setColumnAttack(mo, column, delay, potency, angle, duration)
	
	-- columns go 1, 2, 3, 4, 5
	local startindex = column	-- this one is easy
	for i = 1, GRID_SIZE
		local index = startindex + (i-1)*GRID_SIZE
		setGridAttack(mo, index, delay, {angle=angle, potency=potency, reaction=RT_BLOWDOWN, duration=duration})
	end
	
end


local function setRadiusAttack(mo, index, delay, potency, angle)
	
	setGridAttack(mo, index, delay, {angle=angle, potency=potency, reaction=RT_BLOWDOWN, quake=true})
	
	-- now to surrounding tiles
	
	-- do it manually because i'm very lazy...
	local tiles = {
		
	{-GRID_SIZE -1, -GRID_SIZE, -GRID_SIZE+1,
	-1, 							1, 
	GRID_SIZE-1, GRID_SIZE, GRID_SIZE+1},
	
	/*{
	-2*GRID_SIZE -2, -2*GRID_SIZE-1, -2*GRID_SIZE, -2*GRID_SIZE+1, -2*GRID_SIZE+2,
	-GRID_SIZE-2, 													-GRID_SIZE+2,
	-2,																2,
	GRID_SIZE-2,													GRID_SIZE+2,
	2*GRID_SIZE -2,	2*GRID_SIZE-1, 2*GRID_SIZE, 2*GRID_SIZE+1,	2*GRID_SIZE+2},
	*/
	}
	
	local startX = gridX(index)
	local startY = gridY(index)
	
	for i = 1, #tiles

		local newdelay = delay + TICRATE*3/4*i
		
		-- top
		for j = 1, #tiles[i]
			local nindex = index + tiles[i][j]
			local newX = gridX(nindex)
			local newY = gridY(nindex)
			
			if nindex > 0 and nindex < (GRID_SIZE^2)-1 and abs(startX-newX) <= i and abs(startY-newY) <= i 
				local p = mo.gridpoints[nindex]
				local ang = R_PointToAngle2(mo.gridpoints[index].x, mo.gridpoints[index].y, p.x, p.y)
				setGridAttack(mo, nindex, newdelay, {potency=potency*3/4, reaction=RT_BLOWDOWN, angle=ang, quake=true})
			end
		end
	end
end

local function handleGridAttacks(mo)
	
	if not mo.grid_actions return end
	
	-- we are not guaranteed to have them all set up and happen in order lol.
	for k,v in pairs(mo.grid_actions)
		
		if v.delay
			v.delay = $-1
			--print(v.delay)
			
			if v.delay <= GRID_ATTACK_VISUAL
				
				--print("A")
				
				local sec = mo.grid[k]

				local ang = v.delay*180/GRID_ATTACK_VISUAL
				--print(ang)
				local xscale = (mo.grid_size / GRID_VISUAL_WIDTH)
				local yscale = max(FRACUNIT/8, sin(ang*ANG1))
				local tr = 10 - (v.delay*7/GRID_ATTACK_VISUAL)
				
				for i = 0, 3
					local l = sec.lines[i]	-- we assume they're a square......
					local avgx = ((l.v1.x/FRACUNIT + l.v2.x/FRACUNIT)/2) *FRACUNIT
					local avgy = ((l.v1.y/FRACUNIT + l.v2.y/FRACUNIT)/2) *FRACUNIT 
					
					--P_SpawnMobj(avgx, avgy, sec.floorheight, MT_THOK)
					
					local vis = P_SpawnMobj(avgx, avgy, sec.floorheight, MT_THOK)
					vis.sprite = SPR_LITW
					vis.frame = A|FF_PAPERSPRITE|FF_FULLBRIGHT|(tr<<FF_TRANSSHIFT)
					vis.angle = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
					vis.spritexscale = xscale
					vis.colorized = true
					vis.color = SKINCOLOR_MASTER
					vis.spriteyscale = yscale*4
					
					
					--vis.spriteyscale = max(1, FRACUNIT - cos((v.delay*360/GRID_ATTACK_VISUAL)*ANG1))
					
					vis.tics = 2
				end
			end
			
			if v.delay == 3
				local p = mo.gridpoints[k]
				
				local potency = v.potency or 100
				local damagetype = v.damagetype or DMT_STRIKE
				local duration = v.duration or 3
				local reaction = v.reaction or RT_STAGGER
				
				local hb = PSO.lib.spawnHurtBox(p.x, p.y, p.floorz, 512<<FRACBITS, 512<<FRACBITS, mo, potency, damagetype, duration, reaction)
				hb.angle = v.angle
				hb.sectorbased = true	-- only hurts things in the same sector as itself.
				
				if v.quake
					S_StartSound(p, sfx_pstop)
					for i = 1, 32
						local x = p.x + P_RandomRange(-256, 256)*FRACUNIT
						local y = p.y + P_RandomRange(-256, 256)*FRACUNIT
						local z = p.z
						
						local s = P_SpawnMobj(x, y, z, MT_SPINDUST)
						s.momz = P_RandomRange(4, 16)*FRACUNIT
						s.scale = P_RandomRange(2, 4)*FRACUNIT
					end
					mo.grid[k].floorheight = $+96*FRACUNIT
					v.quaked = 8
				end
			end
			
		end

		if v.quaked
			mo.grid[k].floorheight = $-12*FRACUNIT
			v.quaked = $-1
		end
	end
end


-- df eye laser attack...

PSO.info.enemies[MT_PSO_GYRREY_GLARDIA_PAWN] = {
		
		name = "Gyrrey Glardia",
		sprite = SPR_GSNL,
		frame = A,
		painframe = A,
		
		part = true,			-- enemy is a part of another enemy
		partlinkdamage = true,	-- damage taken by the part is also taken by the main body
		
		bhp = 1000,	--% HP of the source object that spawned it.
		satk = 50,
		ratk = 40,
		tatk = 1,
		def = 30,
		exp = 0,
		weight = 4,
}

local function spawn_elements(mo)
	
	mo.elements = {}
	for i = 1, 4
		local p = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		p.sprite = SPR_HRBG
		p.frame = C
		p.renderflags = $ |(RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD|RF_NOSPLATROLLANGLE)
		P_CreateFloorSpriteSlope(p)	-- used for later
		p.tics = -1
		p.scale = mo.scale
		mo.elements[#mo.elements+1] = p
	end
	
	local halo = P_SpawnMobj(mo.x, mo.y, mo.z + FRACUNIT*4, MT_THOK)
	halo.sprite = SPR_BFLW
	halo.scale = mo.scale
	halo.frame = D|FF_ADD
	halo.color = SKINCOLOR_MASTER
	halo.renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD|RF_FULLBRIGHT
	halo.tics = -1
	mo.halo = halo
	
	mo.front = {}
	for i = 1, 2
		local front = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		front.sprite = SPR_HRBG
		front.frame = ((i&1) and Q or R) | FF_PAPERSPRITE
		front.scale = mo.scale
		mo.front[#mo.front+1] = front
	end
	
	mo.rects = {}
	for i = 1, 4
		local rect = P_SpawnMobj(mo.x, mo.y, mo.z + FRACUNIT*4, MT_THOK)
		rect.sprite = SPR_HRBG
		rect.scale = mo.scale
		rect.frame = S
		rect.renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD
		rect.tics = 3
		mo.rects[#mo.rects+1] = rect
	end
	
	mo.r_angle = 0
	mo.h_angle = 0
end

local function update_elements(mo, shake)
	
	local basedist = 20
	if not mo.drill
		mo.h_angle = mo.angle
	end
	
	for i = 1, #mo.elements 
		
		local a = mo.h_angle + (i-1)*ANG1*90 + ANG1*45
		local x = mo.x + FixedMul(mo.scale, basedist*cos(a))
		local y = mo.y + FixedMul(mo.scale, basedist*sin(a))
		local z = mo.z + 128*mo.scale
		
		local p = mo.elements[i]
		if not p or not p.valid continue end
		
		P_MoveOrigin(p, x + (shake and P_RandomRange(-6, 6)*FRACUNIT or 0), y+(shake and P_RandomRange(-6, 6)*FRACUNIT or 0), z+(shake and P_RandomRange(-6, 6)*FRACUNIT or 0))
		p.angle = a
		p.spriteyscale = FRACUNIT/4
		p.tics = -1

		local origin = {x=x, y=y, z=z}
		local xydir = a
		local za = ANG1*80
		
		p.floorspriteslope.o = origin
		p.floorspriteslope.xydirection = xydir
		p.floorspriteslope.zangle = za
		
	end
	
	if not mo.halo or not mo.halo.valid return end
	
	P_MoveOrigin(mo.halo, mo.x+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0), mo.y+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0), mo.z + mo.scale*140+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0))
	mo.halo.rollangle = $ + ANG2
	mo.halo.tics = -1
	
	for i = 1, #mo.front
		local front = mo.front[i]
		if not front or not front.valid continue end

		local a = mo.angle
		
		local x = mo.x + FixedMul(mo.scale, 48*cos(a))
		local y = mo.y + FixedMul(mo.scale, 48*sin(a))
		
		P_MoveOrigin(front, x+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0), y+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0), mo.z + mo.scale*80+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0))
		front.angle = mo.angle + (i==1 and 1 or -1)*130*ANG1 + ANG1*90
		front.tics = -1
	end
	
	for i = 1, #mo.rects
		
		local rect = mo.rects[i]
		if not rect or not rect.valid continue end
		local a = mo.r_angle + ANG1*90*(i-1)
		
		local x = mo.x + FixedMul(mo.scale, 24*cos(a))
		local y = mo.y + FixedMul(mo.scale, 24*sin(a))
		P_MoveOrigin(rect, x+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0), y+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0), mo.z + mo.scale*16+ (shake and P_RandomRange(-6, 6)*FRACUNIT or 0))
		rect.rollangle = a
		rect.tics = -1
		
	end

	mo.r_angle = $ - ANG1*2
	
end

local function spawnPawns(mo)
	
	mo.pawns = {}
	
	for i = 1, 2
		
		local s = mo.grid_top[i==1 and 2 or 4]
		local pawn = PSO.lib.enemyCreatePart(mo, s.x, s.y, s.z, MT_PSO_GYRREY_GLARDIA_PAWN)
		pawn.scale = FRACUNIT*3
		pawn.angle = mo.angle
		pawn.current = s
		pawn.prev = s
		spawn_elements(pawn)
		
		mo.pawns[#mo.pawns+1] = pawn
	end
	
end

-- now the actual functions to get attacks n shit happening from our pawns...!
local function movePawn(pawn, dest)
	pawn.dest = dest
	pawn.prev = pawn.current
	pawn.movetime = 1
	S_StartSound(nil, sfx_s3k93)
end

local function handlePawnMovement(pawn)
	if pawn.movetime
		
		if not pawn.dest or not pawn.dest.valid	-- happens?
			pawn.dest = nil
			pawn.movetime = nil
			return
		end
		
		if pawn.movetime < TICRATE*3/4
			pawn.momz = $ + FRACUNIT*4
			pawn.flags = $ | MF_NOCLIPHEIGHT
		elseif pawn.movetime == TICRATE*3/4
			local zmom = pawn.momz
			P_SetOrigin(pawn, pawn.dest.x, pawn.dest.y, pawn.z)
			pawn.momz = -zmom
			pawn.angle = pawn.dest.angle
			--print("A")
		elseif pawn.movetime < 2*(TICRATE*3/4)
			if pawn.drill	-- don't slow down when we drill, duh!
				pawn.momz = -FRACUNIT*48
				pawn.flags = $ & ~MF_NOCLIPHEIGHT
			else
				pawn.momz = (pawn.dest.z - pawn.z)/6
			end
		else
			pawn.current = pawn.dest
			pawn.momz = 0
			pawn.movetime = 0
			return
		end
		pawn.movetime = $+1		
	end
end


local DF_LASER_DELAY = TICRATE*4
local DF_LASER_POTENCY = 500

-- dir: 0 = left, 1 = right
local function DFRowLaser(mo, row, dir)
	
	local ctable = dir and mo.grid_right or mo.grid_left
	local otable = dir and mo.grid_left or mo.grid_right
	setRowAttack(mo, row, DF_LASER_DELAY, DF_LASER_POTENCY, ctable[row].angle)
	
	mo.startlasertarget = ctable[row]
	mo.endlasertarget = otable[row]
	mo.laser = DF_LASER_DELAY
end

local function DFColumnLaser(mo, col, dir)
	
	local ctable = dir and mo.grid_bottom or mo.grid_top
	local otable = dir and mo.grid_top or mo.grid_bottom
	setColumnAttack(mo, col, DF_LASER_DELAY, DF_LASER_POTENCY, ctable[col].angle)

	mo.startlasertarget = ctable[col]
	mo.endlasertarget = otable[col]	
	mo.laser = DF_LASER_DELAY
end

local function DFHandleLaser(mo)
	
	if mo.laser
		
		if mo.laser == DF_LASER_DELAY
			PSO.lib.DF_openEye(mo)
		end
		
		local head_x = mo.head.x + 8*cos(mo.angle)
		local head_y = mo.head.y + 8*sin(mo.angle)
		local head_z = mo.head.z + 8*mo.scale
		
		if mo.laser > DF_LASER_DELAY/4
			if mo.laser < DF_LASER_DELAY*8/10
				local s = P_SpawnMobj(head_x + P_RandomRange(-256, 256)*FRACUNIT, head_y + P_RandomRange(-256, 256)*mo.scale, head_z + P_RandomRange(-256, 256)*mo.scale, MT_THOK)
				s.color = SKINCOLOR_MASTER
				s.frame = FF_ADD|FF_FULLBRIGHT
				s.tics = 10
				s.scale = mo.scale*3/4
				s.destscale = 1
				
				s.momx = (head_x - s.x)/10
				s.momy = (head_y - s.y)/10
				s.momz = (head_z - s.z)/10
			end
		elseif mo.laser < (DF_LASER_DELAY/3)
			
			
			local timer = ((DF_LASER_DELAY/3) - mo.laser)	-11	-- the math is wrong so theres a magic number because its 2:30 am.
			-- let us not delude yourself into thinking this is temporary.
			
			local t = (timer*FRACUNIT)/(DF_LASER_DELAY/6)
			
			if t < FRACUNIT
				local start_x = mo.startlasertarget.x
				local start_y = mo.startlasertarget.y
				local start_z = mo.startlasertarget.z

				local end_x = mo.endlasertarget.x
				local end_y = mo.endlasertarget.y
				local end_z = mo.endlasertarget.z
				
				local fx = ease.linear(t, start_x, end_x)
				local fy = ease.linear(t, start_y, end_y)
				local fz = ease.linear(t, start_z, end_z)
				
				local startp = P_SpawnMobj(head_x, head_y, head_z, MT_THOK)
				startp.state = S_INVISIBLE
				startp.tics = 2

				local endp = P_SpawnMobj(fx, fy, fz, MT_THOK)
				endp.state = S_INVISIBLE
				endp.tics = 2
				
				local fn = function(s)
			
					s.sprite = SPR_THOK
					s.frame = FF_FULLBRIGHT
					s.color = SKINCOLOR_MASTER
					s.destscale = 1
					s.scalespeed = FRACUNIT/12
					s.fuse = 6
					s.tics = 6
				end
					
				PSO.lib.spawnLine(startp, endp, 128, fn)
			end
		end
		
		mo.laser = $-1
	end
	
end

local DF_TRACK_LASER_DELAY = TICRATE*6
local DF_TRACK_LASER_POTENCY = 600

local function DFTrackLaser(mo)
	
	mo.tracklaser = DF_TRACK_LASER_DELAY
	mo.pso.targetlock = true
	
end

-- tracking laser
-- layer must hide behind other enemies for cover
local function DFHandleTrackLaser(mo)
	
	if mo.tracklaser
		if mo.tracklaser == DF_TRACK_LASER_DELAY
			PSO.lib.DF_openEye(mo)
		end
		
		local head_x = mo.head.x + 8*cos(mo.angle)
		local head_y = mo.head.y + 8*sin(mo.angle)
		local head_z = mo.head.z + 8*mo.scale
		
		if mo.tracklaser > 10
			if mo.tracklaser < DF_TRACK_LASER_DELAY*9/10
				local s = P_SpawnMobj(head_x + P_RandomRange(-256, 256)*FRACUNIT, head_y + P_RandomRange(-256, 256)*mo.scale, head_z + P_RandomRange(-256, 256)*mo.scale, MT_THOK)
				s.color = SKINCOLOR_MASTER
				s.frame = FF_ADD|FF_FULLBRIGHT
				s.tics = 10
				s.scale = mo.scale*3/4
				s.destscale = 1
				
				s.momx = (head_x - s.x)/10
				s.momy = (head_y - s.y)/10
				s.momz = (head_z - s.z)/10
			end
		end
		
		if mo.tracklaser > 1 and mo.tracklaser < DF_TRACK_LASER_DELAY*9/10
			
			for i = 1, #mo.pawns
				local pawn = mo.pawns[i]
				local check = P_SpawnMobj(pawn.x, pawn.y, pawn.z, MT_PSO_GARGOYLETESTER)
				check.radius = pawn.radius*9/10
				check.height = pawn.height
				check.tics = 2
			end
			
			PSO.lib.enemySightMarker(mo.head, mo.target, PSO.lib.quickSightCheck)
			
			if mo.tracklaser == (DF_TRACK_LASER_DELAY*9/10) -1
				S_StartSound(nil, sfx_s3k9d)
			end
			
		end
		
		if mo.tracklaser == 1	-- firing
			
			local seen = true
			local blocked
			
			local fn = function(s)
			
				if PSO.lib.quickSightCheck(s, mo.head)
					s.sprite = SPR_THOK
					s.frame = FF_FULLBRIGHT
					s.color = SKINCOLOR_EMERALD
					s.destscale = 1
					s.scalespeed = FRACUNIT/12
					s.fuse = 12
					s.tics = 12
				else
					s.state = S_INVISIBLE
					seen = nil
					
					if not blocked	-- find closest puppet
						local x1 = s.x-256*FRACUNIT
						local x2 = s.x+256*FRACUNIT
						local y1 = s.y-256*FRACUNIT
						local y2 = s.y+256*FRACUNIT
						
						local sfn = function(search, found)
							
							if found and found.valid and found.type == MT_PSO_GYRREY_GLARDIA_PAWN
								blocked = found
								return
							end
						end	
						
						searchBlockmap("objects", sfn, s, x1, x2, y1, y2)
					end
					
				end
			end
				
			S_StartSound(nil, sfx_s3k54)
			S_StartSound(nil, sfx_s3k54)
			
			PSO.lib.spawnLine(mo.head, mo.target, 128, fn)
			
			if blocked	-- a pawn blocked it, damage it!
				-- have the enemy do it, a player cannot ever do damage by itself.
				PSO.lib.damageObject(blocked, mo, mo.pso.maxhp/30)
			else
						-- the player didn't hide!
				local hb = PSO.lib.spawnHurtBox(mo.target.x, mo.target.y, mo.target.z, 80<<FRACBITS, 40<<FRACBITS, mo, DF_TRACK_LASER_POTENCY, DMT_RANGED, 3, RT_BLOWDOWN)
				hb.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)	
				hb.pso.nododge = true
			end
			
			mo.pso.targetlock = true
		
		end
		
		mo.tracklaser = $-1
	end
end

local LASER_DELAY = TICRATE*5 -2
local LASER_POTENCY = 500

-- dir: 0 = left, 1 = right
local function pawnRowLaser(mo, pawn, row, dir)
	
	row = max(1, $)
	row = min(GRID_SIZE, $)
	
	local ctable = dir and mo.grid_right or mo.grid_left
	movePawn(pawn, ctable[row])
	setRowAttack(mo, row, LASER_DELAY, LASER_POTENCY, ctable[row].angle)
	
	pawn.laser = LASER_DELAY
end

local function pawnColumnLaser(mo, pawn, col, dir)

	col = max(1, $)
	col = min(GRID_SIZE, $)
	
	local ctable = dir and mo.grid_bottom or mo.grid_top
	movePawn(pawn, ctable[col])
	setColumnAttack(mo, col, LASER_DELAY, LASER_POTENCY, ctable[col].angle)
	
	pawn.laser = LASER_DELAY
end

local DRILL_DELAY = TICRATE*3
local DRILL_POTENCY = 500

local function pawnDrill(mo, pawn, index)
	movePawn(pawn, mo.gridpoints[index])
	setRadiusAttack(mo, index, DRILL_DELAY, DRILL_POTENCY, P_RandomRange(0, 3)*90*ANG1)
	S_StartSound(nil, sfx_spin)
	pawn.flags = $|MF_NOCLIPTHING|MF_NOCLIP
	
	pawn.drill = 1
end

local function updatePawns(mo, shake)
	for i = 1, #mo.pawns
		local pawn = mo.pawns[i]
		if pawn and pawn.valid
			update_elements(pawn, shake)
			
			-- don't do anything else in that case.
			if shake 
				continue
			end
			
			handlePawnMovement(pawn)
			
			if pawn.drill
				
				if pawn.drill <= DRILL_DELAY*3
					pawn.h_angle = $ + min(ANG1*8, pawn.drill*ANG1)
				elseif pawn.drill <= DRILL_DELAY*3 +8
					local drill = pawn.drill - DRILL_DELAY*3
					pawn.h_angle = $ + max(0, 8-drill)*ANG1
				end	
				
				if pawn.drill > TICRATE*2
					pawn.momz = -48*FRACUNIT
				end
				
				if pawn.z <= pawn.floorz and pawn.drill < DRILL_DELAY*3
					P_StartQuake(FRACUNIT*30, 4)
					
					local s = P_SpawnMobj(pawn.x, pawn.y, pawn.z, MT_SUPERSPARK)
					s.momx = P_RandomRange(-16, 16)*FRACUNIT
					s.momy = P_RandomRange(-16, 16)*FRACUNIT
					s.momz = P_RandomRange(8, 16)*FRACUNIT	
					
					pawn.flags = $ & ~MF_NOCLIPTHING
				end
				
				if pawn.drill < TICRATE*3
				and leveltime & 1
					local s = P_SpawnMobj(pawn.dest.x, pawn.dest.y, pawn.dest.z, MT_THOK)
					s.sprite = SPR_TARG
					s.frame = (leveltime%14)/2 |FF_FULLBRIGHT
					s.tics = 1
					s.scale = $*2
				elseif pawn.drill > TICRATE*7
					pawn.drill = nil
					pawn.flags = $ & ~(MF_NOCLIPTHING|MF_NOCLIP)
					return
				end
				
				pawn.drill = $+1
			end
			
			if pawn.laser
				
				pawn.laser = $-1
				
				if pawn.laser == LASER_DELAY*3/4
					
					local x = pawn.x + 96*cos(pawn.angle)
					local y = pawn.y + 96*sin(pawn.angle)
					local z = pawn.z + 64*pawn.scale
					
					local s = P_SpawnMobj(x, y, z, MT_THOK)
					s.sprite = SPR_NTHK
					s.frame = A|FF_FULLBRIGHT|FF_ADD
					s.color = SKINCOLOR_TAN
					s.tics = pawn.laser
					s.destscale = FRACUNIT*4
					s.scalespeed = FRACUNIT/48
					s.momz = FRACUNIT

					S_StartSound(nil, sfx_s3k73)
					S_StartSound(nil, sfx_bchrg)
					
				end
				
				if pawn.laser < LASER_DELAY*7/10
					local x = pawn.x + 96*cos(pawn.angle)
					local y = pawn.y + 96*sin(pawn.angle)
					local z = pawn.z + 64*pawn.scale
					
					local s = P_SpawnMobj(x + P_RandomRange(-192, 192)*FRACUNIT, y + P_RandomRange(-192, 192)*FRACUNIT, z + P_RandomRange(-192, 192)*FRACUNIT, MT_THOK)
					s.sprite = SPR_NTHK
					s.frame = A|FF_FULLBRIGHT|FF_ADD
					s.color = SKINCOLOR_TAN
					s.tics = 8
					s.destscale = 1
					s.momx = -(s.x - x)/8
					s.momy = -(s.y - y)/8
					s.momz = -(s.z - z)/8
					
					if leveltime&1
						local s =  P_SpawnMobj(x + P_RandomRange(-192, 192)*FRACUNIT, y + P_RandomRange(-192, 192)*FRACUNIT, z + P_RandomRange(-192, -64)*FRACUNIT, MT_THOK)
						s.sprite = SPR_NTHK
						s.frame = A|FF_FULLBRIGHT|FF_ADD|FF_PAPERSPRITE
						s.color = SKINCOLOR_TAN
						s.angle = pawn.angle
						s.tics = TICRATE*2
						s.scale = FRACUNIT/2
						s.spritexscale = FRACUNIT*6
						s.spriteyscale = FRACUNIT/2
						P_InstaThrust(s, pawn.angle, FRACUNIT*65)
					end
				end
				
				if pawn.laser == 0
					
					S_StartSound(nil, sfx_bshot)
					
					for i = 1, 128
						local dist = 96 + i*32
						local x = pawn.x + dist*cos(pawn.angle)
						local y = pawn.y + dist*sin(pawn.angle)
						local z = pawn.z + 96*pawn.scale
						
						if i == 1
							for j = 1, 3
								local d = P_SpawnMobj(x, y, z, MT_THOK)
								d.sprite = SPR_NTHK
								d.frame = B|FF_FULLBRIGHT|FF_ADD|FF_PAPERSPRITE
								d.color = SKINCOLOR_TAN
								d.angle = mo.angle + ANG1*90
								d.scale = FRACUNIT*6
								d.detscale = FRACUNIT*j*8
								d.scalespeed = FRACUNIT*j
								d.tics = j*10
								d.fuse = d.tics
								P_InstaThrust(d, d.angle, FRACUNIT*i)
							end
						end
						
						local s = P_SpawnMobj(x, y, z, MT_THOK)
						s.sprite = SPR_NTHK
						s.frame = A|FF_FULLBRIGHT|FF_ADD
						s.color = SKINCOLOR_TAN
						s.tics = 16 +(i-1)/8
						s.fuse = 16 +(i-1)/8
						s.scale = max(1, FRACUNIT - i*(FRACUNIT/128))
						s.destscale = FRACUNIT*6
						s.momz = FRACUNIT
						s.scalespeed = FRACUNIT
					end
				end
			end
		end	
	end
end

local function spawnChains(mo)
	-- first 2 chains are active, remaining are dead...
	mo.chains = {}
	local bases = {mo.neck_base, mo.elbow, mo.spine[3], mo.wrist}
	
	-- some target points for the 2 active chains
	local tx1 = mo.x + 1024*cos(mo.angle - ANG1*90)
	local ty1 = mo.y + 1024*sin(mo.angle - ANG1*90)
	local tz1 = mo.z + 2048*FRACUNIT

	local tx2 = mo.x + 1024*cos(mo.angle + ANG1*90)
	local ty2 = mo.y + 1024*sin(mo.angle + ANG1*90)
	local tz2 = mo.z + 2048*FRACUNIT

	local targets = {
		{x=tx1, y=ty1, z=tz1},
		{x=tx2, y=ty2, z=tz2},
	}
	
	for i = 1, 4
		mo.chains[i] = {
			active = i <3,
			base = P_SpawnMobj(mo.x + 8*cos(mo.angle), mo.y + 8*sin(mo.angle), mo.z, MT_THOK),
			basecoord = bases[i],
			target = targets[i],
			links = {}
		}
		--print(mo.chains[i].base)
		
		mo.chains[i].base.sprite = SPR_DFCH
		mo.chains[i].base.frame = D
		mo.chains[i].base.tics = -1
		mo.chains[i].base.scale = mo.scale
		if mo.chains[i].active
			mo.chains[i].base.frame = A|FF_FULLBRIGHT
		end
		
		-- spawn links
		local numlinks = mo.chains[i].active and 32 or 3
		for j = 1, numlinks
			local link = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			link.sprite = SPR_DFCH
			link.frame = ((mo.chains[i].active and C or F) - (j&1))|FF_PAPERSPRITE
			link.scale = mo.scale
			if mo.chains[i].active
				link.frame = $|FF_FULLBRIGHT
			elseif j == numlinks
				link.frame = G	-- broken link.
			end
				
			link.tics = -1
			
			mo.chains[i].links[j] = link
			
		end
	end
end	

local function handleChains(mo)
	
	local link_dist = 16*mo.scale	-- tent.
	
	for i = 1, #mo.chains
		local dta = mo.chains[i]
		P_MoveOrigin(dta.base, dta.basecoord.x, dta.basecoord.y, dta.basecoord.z)		
		if dta.active

			if mo.deathtimer == TICRATE*8
				P_SpawnGhostMobj(dta.base)
				dta.base.frame = ($+3) & ~FF_FULLBRIGHT
			elseif mo.deathtimer == TICRATE*11
				dta.base.sprite = SPR_NULL
				dta.base.frame = A
				S_StartSound(nil, sfx_bkpoof)
				
				for p in players.iterate
					P_FlashPal(p, 1, 3)
				end
			end
			
			-- link the... well, links, lol.
			
			-- start by getting angh and angv
			local angh = R_PointToAngle2(dta.base.x, dta.base.y, dta.target.x, dta.target.y)
			local dist = R_PointToDist2(dta.base.x, dta.base.y, dta.target.x, dta.target.y)
			local distz = dta.target.z - dta.base.z
			local angv = atan(FixedDiv(distz, dist))
			
			for j = 1, #dta.links
				
				local link = dta.links[j]
				
				-- break anim
				if mo.deathtimer and mo.deathtimer >= TICRATE*11
					
					if not link or not link.valid
						continue
					end
					
					if mo.deathtimer == TICRATE*11
						link.momx = P_RandomRange(-32, 32)*FRACUNIT
						link.momy = P_RandomRange(-32, 32)*FRACUNIT
						link.momz = P_RandomRange(-32, 32)*FRACUNIT
						link.fuse = TICRATE*5
						link.flags = $ & ~MF_NOGRAVITY
					end
					link.rollangle = $+ANG1*10
					
					continue
				end
				
				-- normal behaviour
				local ax, ay, az = PSO.lib.anglesOnSphere(angh, angv, (link_dist/2)*j)
				local cdist = i%2 == 0 and 16 or 0
				P_MoveOrigin(link, dta.base.x+FixedMul(mo.scale, ax) + cdist*cos(mo.angle), dta.base.y+FixedMul(mo.scale, ay) + cdist*sin(mo.angle), dta.base.z+FixedMul(mo.scale, az))
				link.rollangle = angv < 0 and InvAngle(angv) or angv
				link.angle = angh
				
				-- losing light + about to break
				if mo.deathtimer
					if mo.deathtimer == TICRATE*8
						P_SpawnGhostMobj(link)
						link.frame = ($+3) & ~FF_FULLBRIGHT
					elseif mo.deathtimer > TICRATE*9
						P_MoveOrigin(link, link.x + P_RandomRange(-16, 16)*FRACUNIT, link.y + P_RandomRange(-16, 16)*FRACUNIT, link.z + P_RandomRange(-16, 16)*FRACUNIT)
					end
				end
			end
			
		else
			local dir = i&1 and 1 or -1
			local angh = mo.angle + ANG1*120*dir
			local angv = -ANG1*30 + (FixedAngle(cos(leveltime*ANG2)*5))
			
			for j = 1, #dta.links
				
				local link = dta.links[j]
				local prev = dta.links[j-1]
				
				if not link or not link.valid
					continue
				end
				
				if not prev
					prev = dta.base
				end
				
				local cdist = i%2 == 0 and 16 or 0
				
				local ax, ay, az = PSO.lib.anglesOnSphere(angh, angv, (link_dist/2))
				P_MoveOrigin(link, prev.x+FixedMul(mo.scale, ax) + cdist*cos(mo.angle), prev.y+FixedMul(mo.scale, ay) + cdist*sin(mo.angle), prev.z+FixedMul(mo.scale, az))
				link.rollangle = angv
				link.angle = angh
				angv = $-ANG1*20 + (FixedAngle(cos(leveltime*ANG2)*5))
			end
		end
		
	end
end

local function armAnimation(mo)
	mo.arm_hangle = mo.angle - ANG1*90
	
	if mo.deathtimer and mo.deathtimer > TICRATE*11
		mo.arm_vangle = -ANG1*50 + FixedAngle(cos(leveltime*ANG1)*10)
	else
		mo.arm_vangle = ANG1*5
	end
	
	mo.farm_hangle = mo.arm_hangle
	mo.farm_vangle = -90*ANG1 + FixedAngle(sin(leveltime*ANG1)*10)
	mo.wrist_hangle = 0
	mo.wrist_vangle = -90*ANG1 - FixedAngle(sin(leveltime*ANG1)*5)
end


local function prepSuction(mo, where)
	mo.suction = 1
	mo.suction_targetindex = where or closestPointOnGrid(mo, mo.target)
	mo.hand.frame = M
	
	setGridAttack(mo, mo.suction_targetindex, TICRATE*11/2, {angle=mo.angle, potency=500, reaction=RT_BLOWDOWN, duration=4})
	
	-- lens flare!
	local lx = mo.hand.x + 16*cos(mo.angle)
	local ly = mo.hand.y + 16*sin(mo.angle)
	local lz = mo.hand.z
	for i = 1, 2
				
		local l = P_SpawnMobj(lx, ly, lz, MT_THOK)
		l.sprite = SPR_LENS
		l.frame = A|FF_FULLBRIGHT|FF_ADD
		l.color = SKINCOLOR_MASTER
		l.scale = mo.scale
		l.destscale = mo.scale*5*i
		l.momz = mo.scale*4
		l.tics = 10*i
		l.fuse = l.tics
		l.scalespeed = mo.scale/6
	end
	S_StartSound(nil, sfx_s1c9)
end

local function DFHandleSuction(mo)
	if mo.suction
		
		mo.suction = $+1
		
		local target = mo.gridpoints[mo.suction_targetindex]
		
		if mo.suction == TICRATE/2
			S_StartSound(nil, sfx_s3k93)
		end
		
		if mo.suction < TICRATE*3
			
			for i = 1, 8
				
				local x = target.x + P_RandomRange(-2048, 2048)*FRACUNIT
				local y = target.y + P_RandomRange(-2048, 2048)*FRACUNIT
				local z = target.z + P_RandomRange(0, 512)*FRACUNIT
				
				local t = P_SpawnMobj(x, y, z, MT_THOK)
				t.frame = A|FF_FULLBRIGHT
				t.color = SKINCOLOR_MASTER
				t.momx = (target.x-x)/14
				t.momy = (target.y-y)/14
				t.momz = (target.z-z)/14
				t.tics = 14
				t.fuse = 14
				t.scale = FRACUNIT*2
				t.destscale = 0
			end
		end
		
		if mo.suction < TICRATE*9/2
		and leveltime & 1
			local b = P_SpawnMobj(target.x, target.y, target.z + FRACUNIT*32, MT_THOK)
			b.scale = FRACUNIT*2 - (mo.suction*FRACUNIT /100)
			b.frame = FF_FULLBRIGHT
			b.color = SKINCOLOR_MASTER
			b.tics = 2
		
		elseif mo.suction == TICRATE*11/2
			S_StartSound(nil, sfx_bkpoof)
			for i = 1, 32
				local b = P_SpawnMobj(target.x, target.y, target.z + FRACUNIT*32, MT_THOK)
				b.scale = FRACUNIT
				b.frame = FF_FULLBRIGHT
				b.color = SKINCOLOR_MASTER
				b.momx = P_RandomRange(-32, 32)*FRACUNIT
				b.momy = P_RandomRange(-32, 32)*FRACUNIT
				b.momz = P_RandomRange(-32, 32)*FRACUNIT
				b.destscale = 1
				b.tics = TICRATE
				b.fuse = TICRATE
			end
		end
		
		-- hand anim
		if mo.suction > TICRATE*4
		and mo.suction < (TICRATE*11/2 -3)
			local x = P_RandomRange(-16, 16)*FRACUNIT
			local y = P_RandomRange(-16, 16)*FRACUNIT
			local z = P_RandomRange(-16, 16)*FRACUNIT
			P_MoveOrigin(mo.hand, mo.hand.x+x, mo.hand.y+y, mo.hand.z+z)	-- shake hand!
		elseif mo.suction == TICRATE*9/2 -3
		or mo.suction == TICRATE*9/2
			mo.hand.frame = $+1	-- animate the hand~
		end
		
		-- gathering particles
		if mo.suction > TICRATE*3/4
		and mo.suction < TICRATE*4
			for p in players.iterate
				local pmo = p.mo
				if P_CheckSight(mo, pmo)	-- bad hack.....
					
					-- absolute thrust.
					local a = R_PointToAngle2(pmo.x, pmo.y, target.x, target.y)
					local dist = R_PointToDist2(pmo.x, pmo.y, target.x, target.y)
					P_InstaThrust(pmo, a, min(dist, FRACUNIT*32))
				end
			end
		elseif mo.suction > TICRATE*6
			mo.hand.frame = L
			mo.suction = nil
		end
	end
end

function A_GyrreyDeath(mo)
	armAnimation(mo)	-- basic standstill arm animation
	PSO.lib.DF_updateParts(mo)
	updatePawns(mo, true)
	
	
	mo.deathtimer = $ and $+1 or 1
	
	if mo.deathtimer == 1
		S_FadeOutStopMusic(2000)
		for p in players.iterate
			p.powers[pw_nocontrol] = 32767
		end
	end
	
	--if mo.deathtimer < TICRATE*15
		handleChains(mo)
	--end
	--print(mo.deathtimer)
	
	for i = 1, #mo.pawns
	
		local pawn = mo.pawns[i]
		
		if leveltime%2 == 0
		and mo.deathtimer < TICRATE*5
			
			local boom = P_SpawnMobj(pawn.x + P_RandomRange(-128, 128)*FRACUNIT, pawn.y + P_RandomRange(-128, 128)*FRACUNIT, pawn.z + P_RandomRange(0, 256)*FRACUNIT, MT_SONIC3KBOSSEXPLODE)
			boom.colorized = true
			boom.color = SKINCOLOR_MASTER
			boom.renderflags = $|RF_FULLBRIGHT
			boom.tics = $ + P_RandomRange(0, 4)
			boom.scale = FRACUNIT*3 + P_RandomRange(-FRACUNIT/3, FRACUNIT/3)
			
			S_StartSound(pawn, sfx_s3kb4)
		end
		
		pawn.momz = 4*FRACUNIT - mo.deathtimer*FRACUNIT/16
		P_InstaThrust(pawn, pawn.angle, -FRACUNIT*4)
		--print(pawn)
	end
	
	if mo.deathtimer == TICRATE*15
		PSO.lib.DF_openEye(mo)
	end
	
	if mo.deathtimer >= TICRATE*16
		local tt = mo.deathtimer-(TICRATE*16)
		
		if tt < TICRATE*8
			P_StartQuake(FRACUNIT*(tt), 5)
		end

		local chest_x = mo.x + 320*cos(mo.angle + ANG1*65)
		local chest_y = mo.y + 320*sin(mo.angle + ANG1*65)
		local chest_z = mo.z + 32*mo.scale
			
		local s = P_SpawnMobj(chest_x + P_RandomRange(-256, 256)*FRACUNIT, chest_y + P_RandomRange(-256, 256)*mo.scale, chest_z + P_RandomRange(-256, 256)*mo.scale, MT_THOK)
		s.color = SKINCOLOR_MASTER
		s.frame = FF_ADD|FF_FULLBRIGHT
		s.tics = 10
		s.scale = mo.scale*3/4
		s.destscale = 1
			
		s.momx = (chest_x - s.x)/10
		s.momy = (chest_y - s.y)/10
		s.momz = (chest_z - s.z)/10	
		
		if tt == 1
			S_StartSound(nil, sfx_bechrg)
		end
		
		if leveltime%20 == 0
			s = P_SpawnMobj(chest_x, chest_y, chest_z, MT_THOK)
			s.sprite = SPR_NTHK
			s.color = SKINCOLOR_MASTER
			s.frame = B|FF_ADD|FF_FULLBRIGHT
			s.tics = 30
			s.fuse = 30
			s.scale = FRACUNIT*100
			s.momz = -FRACUNIT*130
			s.destscale = 1
			s.scalespeed = FRACUNIT*4
		end
		
		if leveltime%5 == 0
			S_StartSound(nil, sfx_s3k6f)
		end
		
		if mo.deathtimer % TICRATE*4 == 0
		and tt < TICRATE*8
			S_StartSound(nil, sfx_befall)
			for p in players.iterate do
				local pmo = p.mo
				if pmo and pmo.valid and P_CheckSight(mo, pmo)
					pmo.momz = FRACUNIT*7
					pmo.state = S_PLAY_PAIN
				end
			end
		end
		
		-- fade screen
		if tt > TICRATE*5
		and tt < TICRATE*8
			for p in players.iterate do
				local pmo = p.mo
				if pmo and pmo.valid and P_CheckSight(mo, pmo)
					pmo.fadecolor = 0
					if tt%6 == 0
						pmo.fadestrength = min(10, $ and $+1 or 1)
					end
				end	
			end
		end
		
		if tt > TICRATE*8	-- fade away...
			for p in players.iterate do
				local pmo = p.mo
				if pmo and pmo.valid
					pmo.fadecolor = 0
					if tt%6 == 0
						pmo.fadestrength = max(0, $-1)
					end
				end	
			end
		end
		
		-- okay now this is messy
		if tt == TICRATE*8
			
			if server.psoprogress < 16	-- first clear
				
				local diff = PSO.lib.questDifficulty()
				PSO.lib.acceptQuest(1, 17, 11, PSO.lib.questDifficulty(), server.questdata.levelsynch and true or false, true)
				PSO.lib.unlockQuest(1, 17, diff)
				-- consolation prize and the likes
				
				local tp = 500*diff
				local mst = 40000*diff
				
				for p in players.iterate
					p.pso.typepoints[TYPE_HUNTER] = $+tp
					p.pso.typepoints[TYPE_RANGER] = $+tp
					p.pso.typepoints[TYPE_FORCE] = $+tp
					p.pso.meseta = $+mst

					PSO.lib.playerPushNotification(p, "Got "..tp.."TP\nfor quest clear.", 2)
					PSO.lib.playerPushNotification(p, "Got "..mst.."Meseta\nfor quest clear.", 2)
				end
				G_ExitLevel()
			else	-- other clears
				-- first, teleport us very far away...
				P_SetOrigin(mo, mo.x+FRACUNIT*13000, mo.y, mo.z+FRACUNIT*8000)
				--handleChains(mo)
			end
		elseif tt == TICRATE*13

			for p in players.iterate
				p.powers[pw_nocontrol] = 0
			end
			
			mo.nodrops = false
			-- ok NOW YOU CAN DROP SHIT
			PSO.lib.enemyDropItems(mo)
			-- ok the pinata is open it's useless now
			P_RemoveMobj(mo)	-- kill yourself
			return
		end
	end
end

states[S_PSO_GYRREY_GLARDIA_DEATH] = {SPR_DAFA, A, 1, A_GyrreyDeath, 0, 0, S_PSO_GYRREY_GLARDIA_DEATH}

local INVINC_TIMER = TICRATE*30

local function pawnInvincHandler(mo)
	local diff = PSO.lib.questDifficulty()
	
	if diff >= DIFF_HARD --or 1
		
		-- which pawn starts invincible is random
		if not mo.invinc
			mo.invinc = P_RandomRange(1, 2)
			mo.invinctimer = INVINC_TIMER
			mo.pawns[mo.invinc].pso.def = 9999999
		end
		
		mo.invinctimer = $-1
		
		local pawn = mo.pawns[mo.invinc]
		local func = function(a)
						a.frame = $|FF_ADD|FF_FULLBRIGHT
					end
					
		if mo.invinctimer == 0
		or mo.invinctimer < TICRATE*5 and leveltime%10 < 5
			func =	function(a)
						a.frame = $ & ~(FF_ADD|FF_FULLBRIGHT)
					end
		end
		
		for i = 1, #pawn.elements
			local e = pawn.elements[i]
			func(e)
		end

		for i = 1, #pawn.rects
			local e = pawn.rects[i]
			func(e)
		end

		for i = 1, #pawn.front
			local e = pawn.front[i]
			func(e)
		end
		
		if mo.invinctimer == 0	-- swap it out!
			mo.invinc = mo.invinc == 1 and 2 or 1
			pawn.pso.def = mo.pawns[mo.invinc].pso.def
			mo.pawns[mo.invinc].pso.def = 9999999
			
			mo.invinctimer = INVINC_TIMER
			S_StartSound(nil, sfx_appear)
		end
		
	end
end

PSO.info.enemies[MT_PSO_GYRREY_GLARDIA_MAIN] = {
		
		name = "Gyrrey Glardia",
		sprite = SPR_NULL,
		frame = A,
		painframe = A,
		boss = true,		-- enemy is a major boss!
		bossmusic = "BOSS6",
		
		bhp = 8000,
		satk = 35,
		ratk = 35,
		tatk = 35,
		def = 25,
		exp = 500,
		weight = 4,
		
		drops = {
			
			{"alva pine", 1, 35, 1, 100},
			{"alva claymore", 1, 35, 1, 100},
			{"vita pine", 1, 35, 1, 100},
			{"vita claymore", 1, 35, 1, 100},
			
			{"alva doubliss", 1, 35, 1, 100},
			{"alva doublecutter", 1, 35, 1, 100},	
			{"vita doubliss", 1, 35, 1, 100},
			{"vita doublecutter", 1, 35, 1, 100},
			
			{"alva shooter", 1, 35, 1, 100},
			{"alva burster", 1, 35, 1, 100},				
			{"vita shooter", 1, 35, 1, 100},
			{"vita burster", 1, 35, 1, 100},
				
			{"alva shot", 1, 35, 1, 100},
			{"alva concentrator", 1, 35, 1, 100},	
			{"vita shot", 1, 35, 1, 100},
			{"vita concentrator", 1, 35, 1, 100},
			
			{"alva crest", 1, 35, 1, 100},
			{"alva crescent", 1, 35, 1, 100},
			{"vita crest", 1, 35, 1, 100},
			{"vita crescent", 1, 35, 1, 100},
				
			{"alva claw", 1, 35, 1, 100},
			{"alva cutter", 1, 35, 1, 100},			
			{"vita claw", 1, 35, 1, 100},
			{"vita cutter", 1, 35, 1, 100},
			
			{"alva beamgun", 1, 35, 1, 100},
			{"alva peacekeeper", 1, 35, 1, 100},			
			{"vita beamgun", 1, 35, 1, 100},
			{"vita peacekeeper", 1, 35, 1, 100},
					
			{"alva hammer", 1, 35, 1, 100},
			{"alva scepter", 1, 35, 1, 100},
			{"vita hammer", 1, 35, 1, 100},
			{"vita scepter", 1, 35, 1, 100},
			
			{"tcsm_scrap", 1, 35, 1, 35},
			
			{"unit_hp_2", 1, 35, 2, 50},
			{"unit_pp_2", 1, 35, 2, 50},
			{"unit_satk_2", 1, 35, 2, 50},
			{"unit_tatk_2", 1, 35, 2, 50},
			{"unit_ratk_2", 1, 35, 2, 50},

			{"digrinder", 1, 35, 1, 5},


			-- Lv36-55
			{"vita pine", 36, 55, 1, 100},
			{"vita claymore", 36, 55, 1, 100},
			{"vita doubliss", 36, 55, 1, 100},
			{"vita doublecutter", 36, 55, 1, 100},				
			{"vita shooter", 36, 55, 1, 100},
			{"vita burster", 36, 55, 1, 100},						
			{"vita shot", 36, 55, 1, 100},
			{"vita concentrator", 36, 55, 1, 100},			
			{"vita crest", 36, 55, 1, 100},
			{"vita crescent", 36, 55, 1, 100},			
			{"vita claw", 36, 55, 1, 100},
			{"vita cutter", 36, 55, 1, 100},					
			{"vita beamgun", 36, 55, 1, 100},
			{"vita peacekeeper", 36, 55, 1, 100},						
			{"vita hammer", 36, 55, 1, 100},
			{"vita scepter", 36, 55, 1, 100},
			{"vita joust", 36, 50, 1, 100},
			{"vita drivel", 36, 50, 1, 100},
			{"vita fork", 36, 50, 1, 100},
			{"vita grenade", 36, 50, 1, 100},
			
			{"unit_ppsatk_2", 36, 55, 2, 50},
			{"unit_pptatk_2", 36, 55, 2, 50},
			{"unit_ppratk_2", 36, 55, 2, 50},
			{"unit_rainbow_2", 36, 55, 2, 50},
			{"unit_satk_3", 36, 50, 1, 50},
			{"unit_ratk_3", 36, 50, 1, 50},
			{"unit_tatk_3", 36, 50, 1, 50},			
			
			{"flow frame", 36, 200,1, 100},
			{"scape doll", 36, 200,1, 40},
			
			{"orakl sword", 36, 50, 1, 200},
			{"orakl doublesaber", 36, 50, 1, 200},
			{"orakl soaringblade", 36, 50, 1, 200},
			{"orakl partizan", 36, 50, 1, 200},
			{"orakl launcher", 36, 50, 1, 200},
			{"orakl tcsm", 36, 50, 1, 200},
			{"orakl rifle", 36, 50, 1, 200},
			{"orakl shot", 36, 50, 1, 200},
			{"orakl orphan", 36, 50, 1, 200},
			{"orakl rod", 36, 50, 1, 200},
			{"orakl wand", 36, 50, 1, 200},
			{"digrinder", 36, 55, 6, 50},

			-- Lv51+
			{"abnormal shard", 51, 79, 1, 500},
			
			{"orakl sword", 51, 200,1, 150},
			{"orakl doublesaber", 51, 200,1, 150},
			{"orakl soaringblade", 51, 200,1, 150},
			{"orakl partizan", 51, 200,1, 150},
			{"orakl launcher", 51, 200,1, 150},
			{"orakl tcsm", 51, 200,1, 150},
			{"orakl rifle", 51, 200,1, 150},
			{"orakl shot", 51, 200,1, 150},
			{"orakl orphan", 51, 200,1, 150},
			{"orakl rod", 51, 200,1, 150},
			{"orakl wand", 51, 200,1, 150},
			
			{"makel frame", 51, 79,1, 150},
			{"aberrant frame", 51, 79, 1, 200},
			{"flow frame", 51, 79,1, 60},
			{"steel hearts", 51, 79,1, 400},
			{"seal case", 51, 79,1, 400},
			{"zanba", 51, 79, 1, 600},
			{"last survivor", 51, 79, 1, 600},
			{"chainsawd", 51, 79, 1, 2500},
			{"dragon slayer", 51, 79, 1, 2500},
			{"monkey king bar", 51, 79,1, 400},
			{"girasole", 51, 79,1, 400},
			{"d. saber legacy", 51, 79, 1, 600},
			{"eternal heat", 51, 79, 1, 600},
			{"vivienne", 51, 79, 1, 2500},
			{"double cannon", 51, 79, 1, 2500},			
			{"basilleon", 51, 79,1, 400},
			{"longinus", 51, 79,1, 400},
			{"meteor lancer", 51, 79, 1, 600},
			{"brionac", 51, 79, 1, 600},
			{"vjaya", 51, 79, 1, 2500},
			{"soul eater", 51, 79, 1, 2500},			
			{"sarblight", 51, 79,1, 400},
			{"forblight", 51, 79,1, 400},	
			{"grand cross", 51, 79, 1, 600},
			{"red saber", 51, 79, 1, 600},
			{"elysion", 51, 79, 1, 2500},
			{"axeon", 51, 79, 1, 2500},
			{"full cylinder", 51, 79,1, 400},
			{"assassin", 51, 79,1, 400},
			{"yasminkov 3000r", 51, 79, 1, 600},
			{"strauss", 51, 79, 1, 600},
			{"firearms", 51, 79, 1, 2500},
			{"frozen shooter", 51, 79, 1, 2500},			
			{"wide smack", 51, 79,1, 400},
			{"heimdall", 51, 79,1, 400},
			{"barett", 51, 79, 1, 600},
			{"crush bullet", 51, 79, 1, 600},
			{"spread needle", 51, 79, 1, 2500},
			{"final impact", 51, 79, 1, 2500},			
			{"photon launcher", 51, 79,1, 400},
			{"saint kilda", 51, 79,1, 400},
			{"charge howzer", 51, 79, 1, 600},
			{"gur bazga", 51, 79, 1, 600},
			{"sspm launcher", 51, 79, 1, 2500},
			{"inferno bazooka", 51, 79, 1, 2500},			
			{"yasminkov 2000h", 51, 79,1, 400},
			{"sarcueid", 51, 79,1, 400},
			{"varista", 51, 79, 1, 600},
			{"bravace", 51, 79, 1, 600},
			{"pandora extreme", 51, 79, 1, 2500},
			{"ruby bullet", 51, 79, 1, 2500},			
			{"clarita visas", 51, 79,1, 400},
			{"red processor", 51, 79,1, 400},
			{"battle verge", 51, 79, 1, 600},
			{"brave hammer", 51, 79, 1, 600},
			{"siren hammer", 51, 79, 1, 2500},
			{"demonic fork", 51, 79, 1, 2500},			
			{"windmill", 51, 79,1, 400},
			{"hexigraph", 51, 79,1, 400},
			{"serdote", 51, 79, 1, 600},
			{"rose quartz", 51, 79, 1, 600},
			{"rykros staff", 51, 79, 1, 2500},
			{"evil curst", 51, 79, 1, 2500},			
			{"tcsm_crawla4", 51, 79,1, 400},
			{"tcsm_omochao", 51, 79, 1, 600},
			
			{"unit_hpsatk_3", 51, 200,1, 150},
			{"unit_hpratk_3", 51, 200,1, 150},
			{"unit_hptatk_3", 51, 200,1, 150},
			{"unit_ppsatk_3", 51, 200,1, 150},
			{"unit_ppratk_3", 51, 200,1, 150},
			{"unit_pptatk_3", 51, 200,1, 150},			
			{"unit_power_3", 51, 200,1, 150},
			{"unit_rainbow_3", 51, 200,1, 150},
			
			{1000, nil, nil, 1, 1},
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_ABERRATION,
		weakness = ELEM_DARK,
		weakness2 = ELEM_LIGHT,
		
		reset =	function(mo)
			if not mo.pso.blowdown
				mo.flags = $|MF_NOGRAVITY
				mo.momz = 0
			end
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()

			PSO.lib.enemyManageAggro(mo)
			mo.nodrops = true	-- I promise it makes sense.
			
			if not mo.init
				PSO.lib.DF_spawn(mo)
				makeGrid(mo, GRID_TAG, GRID_SIZE)
				spawnPawns(mo)
				
				mo.attacktimer = 0
				
				--setGridAttack(mo, 3, TICRATE*2, {})
				--setColumnAttack(mo, 1, TICRATE*2, 100, 1)
				--setRadiusAttack(mo, 3+5, TICRATE*2, {})
			end
			armAnimation(mo)	-- basic standstill arm animation
			PSO.lib.DF_updateParts(mo)
			updatePawns(mo)
			handleGridAttacks(mo)
			DFHandleLaser(mo)
			DFHandleTrackLaser(mo)
			DFHandleSuction(mo)
			
			pawnInvincHandler(mo)
			
			if not mo.chains
				spawnChains(mo)
			end
			handleChains(mo)
			
				-- test
				--print(PSO.lib.quickSightCheck (mo, players[0].mo))
			
			if mo.target and mo.target.valid
				mo.attacktimer = $+1
				local t = mo.attacktimer
				
				--print(whichColumn(mo, mo.target))
				--mo.firstattack = true	-- debug
				--mo.phase_attack = true	-- debug
				
				if not mo.firstattack	-- always open with the same attack
					
					if t == TICRATE*5
						-- the pawn furthest from us does a laser on the row we stand on
						-- compare pawn distances
						local dist1 = R_PointToDist2(mo.pawns[1].x, mo.pawns[1].y, mo.target.x, mo.target.y)
						local dist2 = R_PointToDist2(mo.pawns[2].x, mo.pawns[2].y, mo.target.x, mo.target.y)
						local pawn = dist2<dist1 and mo.pawns[1] or mo.pawns[2]
						
						local playerrow = whichRow(mo, mo.target)
						pawnRowLaser(mo, pawn, playerrow, dist2<dist1)	-- closest to player, let's be nice haha.
					elseif t == TICRATE*10
						mo.firstattack = true
						mo.attacktimer = 0
					end
					
					return
				end
				
				-- phase transition attack
				if ((mo.pso.hp < (mo.pso.maxhp*65/100)))
				and not mo.phase_attack
					
					if t == TICRATE*4
						prepSuction(mo, 23)
					end
				
					if t == TICRATE*7
						pawnRowLaser(mo, mo.pawns[1], 1, 0)
					elseif t == TICRATE*9
						pawnRowLaser(mo, mo.pawns[2], 2, 1)
					elseif t == TICRATE*12
						pawnRowLaser(mo, mo.pawns[1], 3, 0)
					elseif t == TICRATE*14
						pawnRowLaser(mo, mo.pawns[2], 4, 1)			
					elseif t == TICRATE*17
						pawnRowLaser(mo, mo.pawns[1], 5, 0)				
					elseif t == TICRATE*22
						pawnDrill(mo, mo.pawns[2], P_RandomRange(1, 5))
					elseif t == TICRATE*26
						DFTrackLaser(mo)
					elseif t == TICRATE*38
						movePawn(mo.pawns[1], mo.grid_top[2])
						movePawn(mo.pawns[2], mo.grid_top[4])
						mo.phase_attack = true
						mo.attacktimer = 0
					end
					return
				end
				
				mo.attack_wait = TICRATE*5
				
				-- and now we just do random things hehe
				if t == mo.attack_wait
					mo.attackstate = P_RandomRange(1, 4)
					--print("new state: "..mo.attackstate)
				elseif t < mo.attack_wait
					mo.attackstate = 0	-- impossible.
					PSO.lib.DF_closeEye(mo)
				end
				
				local tt = t -mo.attack_wait
				
				if not mo.phase_attack	-- pre-pinch
				or not mo.pinch_attack	-- pinch alternates between normal and pinch attacks.
					
					-- attack states handled here.
					if mo.attackstate == 1	-- + shaped laser
						if tt == 1
							local playerrow = whichRow(mo, mo.target)
							local playercol = whichColumn(mo, mo.target)
							
							pawnRowLaser(mo, mo.pawns[1], playerrow, P_RandomRange(0,1))
							pawnColumnLaser(mo, mo.pawns[2], playercol, P_RandomRange(0,1))
						elseif tt == TICRATE*5
							mo.attacktimer = 0
							mo.pinch_attack = true
							return
						end
					elseif mo.attackstate == 2	-- column laser targetted on player
						if tt == 1
							local playercol = whichColumn(mo, mo.target)
							
							pawnColumnLaser(mo, mo.pawns[2], playercol, P_RandomRange(0,1))
							local nextcol = playercol + (P_RandomRange(0,1) and 1 or -1)
							if nextcol <= 0
								nextcol = $+2
							elseif nextcol > GRID_SIZE
								nextcol = $-2
							end
							pawnColumnLaser(mo, mo.pawns[1], nextcol, P_RandomRange(0,1))
								
							
						elseif tt == TICRATE*5
							mo.attacktimer = 0
							mo.pinch_attack = true
							return
						end					
					elseif mo.attackstate == 3	-- row laser targetted on player
						if tt == 1
							local playerrow = whichRow(mo, mo.target)
							
							pawnRowLaser(mo, mo.pawns[2], playerrow, P_RandomRange(0,1))
							local nextrow = playerrow + (P_RandomRange(0,1) and 1 or -1)
							if nextrow <= 0
								nextrow = $+2
							elseif nextrow > GRID_SIZE
								nextrow = $-2
							end
							pawnRowLaser(mo, mo.pawns[1], nextrow, P_RandomRange(0,1))
								
							
						elseif tt == TICRATE*5
							mo.attacktimer = 0
							mo.pinch_attack = true
							return
						end					
					elseif mo.attackstate == 4	-- drill. Closest spawn goes up and drills where the player is.
						if tt == 1
							local dist1 = R_PointToDist2(mo.pawns[1].x, mo.pawns[1].y, mo.target.x, mo.target.y)
							local dist2 = R_PointToDist2(mo.pawns[2].x, mo.pawns[2].y, mo.target.x, mo.target.y)
							local pawn = dist2<dist1 and mo.pawns[2] or mo.pawns[1]
							local dest = closestPointOnGrid(mo, mo.target)
							mo.pawndrill_remember = pawn
							
							pawnDrill(mo, pawn, dest)
							
						elseif tt >= TICRATE*8
							movePawn(mo.pawndrill_remember, mo.pawndrill_remember.prev)
							mo.attacktimer = 0
							mo.pinch_attack = true
							return
						end					
					end
				else
					
					-- post pinch attacks
					
					-- sucks player to an index and then has 2 lasers.
					-- pattern is random based on position
					if mo.attackstate == 1
						--pick a random index...
						
						if tt == 1
							
							local id = P_RandomRange(1, GRID_SIZE*GRID_SIZE)
							local row = ((id-1) / GRID_SIZE) +1
							local col = ((id-1) % GRID_SIZE) +1
							
							prepSuction(mo, id)
							
							-- if the player is sucked to an edge, both things go in the same direction
							if col == 1 or col == GRID_SIZE
							or row == 1 or row == GRID_SIZE
								-- the two go in the same direction.
								col = $ + (P_RandomRange(0,1) and 1 or -1)
								if col <= 0
									col = $+2
								elseif col > GRID_SIZE
									col = $-2
								end
								
								row = $ + (P_RandomRange(0,1) and 1 or -1)
								if row <= 0
									row = $+2
								elseif row > GRID_SIZE
									row = $-2
								end
								
								--print(row, col)
								
								if P_RandomRange(0,1)
									pawnColumnLaser(mo, mo.pawns[2], col, P_RandomRange(0,1))
									local nextcol = col + (P_RandomRange(0,1) and 1 or -1)
									if nextcol <= 0
										nextcol = $+2
									elseif nextcol > GRID_SIZE
										nextcol = $-2
									end
									pawnColumnLaser(mo, mo.pawns[1], nextcol, P_RandomRange(0,1))						
								else
									pawnRowLaser(mo, mo.pawns[2], row, P_RandomRange(0,1))
									local nextrow = row + (P_RandomRange(0,1) and 1 or -1)
									if nextrow <= 0
										nextrow = $+2
									elseif nextrow > GRID_SIZE
										nextrow = $-2
									end
									pawnRowLaser(mo, mo.pawns[1], nextrow, P_RandomRange(0,1))						
								end
							
							-- otherwise, random + shape laser somewhere.
							else	-- cross pattern.
								local tcol = col + P_RandomRange(0,1) and -1 or 1
								local trow = row + P_RandomRange(0,1) and -1 or 1
								pawnColumnLaser(mo, mo.pawns[2], tcol, P_RandomRange(0,1))
								pawnRowLaser(mo, mo.pawns[1],trow, P_RandomRange(0,1))	
							end
						elseif tt == 6*TICRATE
							
							mo.attacktimer = 0
							mo.pinch_attack = nil
							return
						end
					
					-- blankets the arena in 3 columns/rows of lasers.
					elseif mo.attackstate == 2
						if tt == 1
							
							if P_RandomRange(0,1)	-- column
								local col = P_RandomRange(1, 3)
								pawnColumnLaser(mo, mo.pawns[1], col+1, P_RandomRange(0,1))
								pawnColumnLaser(mo, mo.pawns[2], col+2, P_RandomRange(0,1))
								DFColumnLaser(mo, col, P_RandomRange(0,1))
							else	-- row
								local row = P_RandomRange(1, 3)
								pawnRowLaser(mo, mo.pawns[1], row+1, P_RandomRange(0,1))
								pawnRowLaser(mo, mo.pawns[2], row+2, P_RandomRange(0,1))
								DFRowLaser(mo, row, P_RandomRange(0,1))						
							end
							
						elseif tt == TICRATE*5
							mo.pinch_attack = nil
							mo.attacktimer = 0
							return
						end
					
					-- one pawn drills to either side of the arena, the other lasers right in front.
					-- df does a sight attack
					elseif mo.attackstate == 3
						if tt == 1
							
							local row = P_RandomRange(1, GRID_SIZE-2)
							local col = P_RandomRange(0,1) and 1 or GRID_SIZE
							local index = col + (row-1)*GRID_SIZE
							
							pawnDrill(mo, mo.pawns[1], index)
							pawnRowLaser(mo, mo.pawns[2], row+1, col==1)
							
						elseif tt == TICRATE*3
							DFTrackLaser(mo)
						elseif tt == TICRATE*10
							movePawn(mo.pawns[1], mo.pawns[1].prev)
							mo.attacktimer = 0
							mo.pinch_attack = nil
							return
						end
					
					-- blanket the whole arena
					elseif mo.attackstate == 4
						if tt == 1
							mo.attack4_dir1 = P_RandomRange(0,1)	-- row/column
							mo.attack4_dir2 = P_RandomRange(0,1)	-- ascending/descending
							mo.attack4_ccount = mo.attack4_dir2 and 5 or 1	-- column/row count
							mo.attack4_fcount = 1	-- function counts, to know which function to call	
						end
						
						if tt > 1
						and tt % (TICRATE*3/2) == 0
						and mo.attack4_fcount < 6
							
							local basefunc1 = mo.attack4_dir1 and pawnColumnLaser or pawnRowLaser
							local basefunc2 = mo.attack4_dir1 and DFColumnLaser or DFRowLaser
							
							--print("Data: fcount: "..mo.attack4_fcount..", ccount: "..mo.attack4_ccount)
							
							if mo.attack4_fcount & 1
								basefunc1(mo, mo.pawns[mo.attack4_fcount%3 and 1 or 2], mo.attack4_ccount, mo.attack4_fcount%3 and 1 or 0)
								--print("bfunc1()")
							else
								basefunc2(mo, mo.attack4_ccount, 0)
								--print("bfunc2()")
							end
							
							mo.attack4_fcount = $+1
							if mo.attack4_dir2
								mo.attack4_ccount = $-1
							else
								mo.attack4_ccount = $+1
							end
						end
						
						if tt == TICRATE*10
							mo.attacktimer = 0
							mo.pinch_attack = nil
							return
						end
					end
				end
			end
			
			-- below is just test stuff
			/*
			if t == TICRATE*2
				--movePawn(mo.pawns[1], mo.grid_left[1])
				--pawnColumnLaser(mo, mo.pawns[1], 1, 0)
				--pawnDrill(mo, mo.pawns[1], 7)
				
				--PSO.lib.DF_openEye(mo)
				--DFRowLaser(mo, 1, 0)
				--DFTrackLaser(mo)
				prepSuction(mo, 20)
			end
			--PSO.lib.enemySightMarker(mo.head, mo.target, PSO.lib.quickSightCheck)
			/*
			/*elseif t == TICRATE*3 + TICRATE/2
				pawnColumnLaser(mo, mo.pawns[2], 2, 0)
			elseif t == TICRATE*5
				pawnColumnLaser(mo, mo.pawns[1], 3, 0)
			elseif t == TICRATE*6 + TICRATE/2
				pawnColumnLaser(mo, mo.pawns[2], 4, 0)			
			elseif t == TICRATE*8
				pawnColumnLaser(mo, mo.pawns[1], 5, 0)				
			end*/
		end,
}