-- Heavyweapons Crawla

/*
	Moves towards player
	
	Can dash towards player, turning back after the dash to shoot projectiles.
	If far away, will launch exploding missiles
	
	HARD MODE:
	Dashes twice in a row, throws missiles instead of projectiles after dash
	If far away, throws more exploding missiles
	
	CORRUPTED:
	Dashing leaves a Megid trail, can teleport between dashes
*/

freeslot("MT_PSO_HEAVYWEAPONS_CRAWLA", "SPR_ORS3")
mobjinfo[MT_PSO_HEAVYWEAPONS_CRAWLA] = {
	doomednum = 2705,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 48<<FRACBITS,
	height = 80<<FRACBITS,
	flags = MF_ENEMY|MF_SHOOTABLE,
}

-- shoots a projectile at mo's target, ptype is the MT_ constant and 'dir' says whether we use the left or right cannon
local function shootProjectile(mo, ptype, dir, potency, speed, reaction)
	local offh = 64
	local offz = 128
	local dir = mo.angle + ANG1*90 * (dir and -1 or 1)
	
	local x = offh*cos(dir)
	local y = offh*sin(dir)
	local z = offz*FRACUNIT
	
	local msl = PSO.lib.spawnMissile(mo, ptype, mo.angle, speed, potency, DMT_RANGED, x, y, offz, reaction or RT_NONE)
	PSO.lib.aimMissile(msl, mo.target)
end

PSO.info.enemies[MT_PSO_HEAVYWEAPONS_CRAWLA] = {
		
		name = "Heavyarms Crawla",
		sprite = SPR_ORS3,
		frame = A,
		painframe = A,
		
		bhp = 700,
		satk = 35,
		ratk = 30,
		tatk = 20,
		def = 25,
		exp = 40,
		weight = 4,
		corruptionchance = 100,
		
		drops = {
			-- Lv1-10
			{"calibur", 1, 9, 2, 50},
			{"burst", 1, 9, 2, 50},
			{"cane", 1, 9, 2, 50},		
			
			{"doublesaber", 1, 9, 8, 50},
			{"shotgun", 1, 9, 8, 50},
			
			{"gigush", 1, 9, 2, 50},
			{"beam", 1, 9, 2, 50},
			{"baton", 1, 9, 2, 50},	
			{"bomber", 1, 9, 2, 50},
			{"pike", 1, 9, 2, 50},
			
			{"unit_satk_1", 1, 9, 2, 50},
			{"unit_ratk_1", 1, 9, 2, 50},
			{"unit_tatk_1", 1, 9, 2, 50},

			-- Lv30-50
			{"vita pine", 30, 50, 2, 50},
			{"vita shooter", 30, 50, 2, 50},
			{"vita crest", 30, 50, 2, 50},	
			{"vita drivel", 30, 50, 2, 50},
			{"vita joust", 30, 50, 2, 50},
			
			{"unit_power_2", 30, 50, 2, 50},
			{"unit_rainbow_1", 30, 50, 2, 50},
			{"unit_hppp_2", 30, 50, 2, 50},
			
			--Lv10-29 units
			{"unit_pp_2", 10, 29, 5, 50},
			{"unit_hp_2", 10, 29, 5, 50},
			{"unit_satk_2", 10, 29, 5, 50},
			{"unit_ratk_2", 10, 29, 5, 50},
			{"unit_tatk_2", 10, 29, 5, 50},			
			
			{"heimdall", 51, 79, 1, 80},
			{"wide smack", 51, 79, 1, 80},

			{"barett", 51, 79, 1, 120},
			{"crush bullet", 51, 79, 1, 120},

			{"spread needle", 51, 79, 1, 320},
			{"final impact", 51, 79, 1, 320},
			
			{"vita pine", 51, 79,2, 50},
			{"vita shooter", 51, 79,2, 50},
			{"vita crest", 51, 79,2, 50},	
			{"vita drivel", 51, 79,2, 50},
			{"vita joust", 51, 79,2, 50},
			
			{"makel frame", 51, 79,1, 100},
			{"hart frame", 51, 79,1, 60},
			{"baston frame", 51, 79,1, 100},
			
			{"unit_power_3", 51, 79,2, 50},
			{"unit_rainbow_3", 51, 79,2, 50},
			{"unit_hppp_3", 51, 79,2, 50},
			
			{1000, nil, nil, 1, 5},
			{"dimate", nil, nil, 1, 5},
		},
		
		attribute = ATTR_BADNIK,
		weakness = ELEM_ICE,
		weakness2 = ELEM_FIRE,

		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			
			local attack_potency = 400
			local missile_potency = 200
			
			local hardmode = diff >= DIFF_HARD
			local move_speed = mo.scale*(hardmode and 4 or 3)
			local attack_range = mo.scale*356
			local attack_charge = diff >= DIFF_HARD and TICRATE or TICRATE*3/2
			local attack_duration = TICRATE
			local turnaround = attack_duration/2
			local attack_delay = TICRATE
			local attack_speed = mo.scale*100
			local anim_speed = 2
			
			local num_missiles = hardmode and 10 or 4
			local waittime = TICRATE*5
			
			local melee_dist = FRACUNIT*768
			
			
			if mo.pso.shifta
				move_speed = $*12/10	-- +20% movement speed w shifta
			end
			
			PSO.lib.enemySetCorruption(mo)
			--mo.pso.corrupted = true
			
			PSO.lib.enemyManageAggro(mo)
			
			mo.sprite = SPR_ORS3
			mo.frame = A
			
			if mo.target and mo.target.valid
				
				if mo.shooting
					
					local shooting_duration = 10 + num_missiles*5  +5
					mo.shooting = $+1
					
					if mo.shooting >= 10
					and mo.shooting < shooting_duration -5
						if mo.shooting % 5 == 0
							shootProjectile(mo, MT_TORPEDO, mo.shooting%10 == 0, 350, FRACUNIT*(hardmode and 50 or 35), RT_STAGGER)
							S_StartSound(mo, sfx_trfire)
						end
					end
					
					if mo.shooting >= shooting_duration
						mo.shooting = nil
					end
					
					return
				end
				
				if mo.actiontimer
				
					mo.actiontimer = $+1
					
					if mo.actiontimer < attack_charge
						if mo.actiontimer % 10 == 0
							S_StartSound(mo, sfx_spndsh)
						end
						
						mo.teleported = false
						
						local dust = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
						dust.angle = mo.angle + ANG1*180 + P_RandomRange(-30, 30)*ANG1
						P_InstaThrust(dust, dust.angle, P_RandomRange(8, 16)*FRACUNIT)
						dust.momz = P_RandomRange(4, 8)*FRACUNIT
						
					
					elseif mo.actiontimer == attack_charge
						mo.superarmor = true
						P_InstaThrust(mo, mo.angle, attack_speed)
						S_StartSound(mo, sfx_zoom)

						local hx = mo.x + 64*cos(mo.angle)
						local hy = mo.y + 64*sin(mo.angle)
						
						local hb = PSO.lib.spawnHurtBox(hx, hy, mo.z, 100<<FRACBITS, 70<<FRACBITS, mo, attack_potency, DMT_STRIKE, attack_duration, RT_KNOCKBACK)
						mo.hitbox = hb
						hb.angle = mo.angle
						hb.pso.reactionstrength = 80
					
					end
					if mo.actiontimer >= attack_charge
					and mo.actiontimer < attack_charge + attack_duration
						if mo.hitbox and mo.hitbox.valid
							local a = R_PointToAngle2(0, 0, mo.momx, mo.momy)
							local hx = mo.x + 64*cos(a)
							local hy = mo.y + 64*sin(a)
							
							P_MoveOrigin(mo.hitbox, hx, hy, mo.z)
						end
						
						if mo.actiontimer >= attack_charge+turnaround
							mo.angle = $ + (180/turnaround)*ANG1
							if leveltime % 5 == 0
								S_StartSound(mo, sfx_skid)
							end

							local dust = P_SpawnMobj(mo.x + P_RandomRange(-128, 128)*FRACUNIT, mo.y + P_RandomRange(-128, 128)*FRACUNIT, mo.z, MT_SPINDUST)
							dust.momz = P_RandomRange(4, 8)*FRACUNIT
						end
						
						P_SpawnGhostMobj(mo)
					end
					
					if mo.actiontimer >= attack_charge + attack_duration
						local t = mo.actiontimer - attack_charge - attack_duration
						
						if t == 1
							A_FaceTarget(mo)
							shootProjectile(mo, MT_JETTBULLET, 1, 200, FRACUNIT*50, RT_NONE)
							S_StartSound(mo, sfx_trfire)
						elseif t == 6
							shootProjectile(mo, MT_JETTBULLET, 0, 200, FRACUNIT*50, RT_NONE)
							S_StartSound(mo, sfx_trfire)
						elseif t == 8
							mo.attackcount = $+1
							
							if not hardmode
							or mo.attackcount > 1
								mo.actiontimer = nil
							else
								mo.actiontimer = 1
							end
						end
					end
					
					return
				end
				
				if not mo.actiontimer
				and not mo.shooting
					A_FaceTarget(mo)
					P_InstaThrust(mo, mo.angle, move_speed)
					
					mo.frame = (leveltime%(anim_speed*4)) /anim_speed
					mo.attackcount = 0
					mo.waittimer = $ and $+1 or 1
					
					
					if mo.waittimer >= waittime
						mo.waittimer = 0
						
						if R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) > melee_dist
							mo.shooting = 1
						else
							mo.actiontimer = 1
						end
					end
					
				end
			end
		end,

}