-- Jetty Syn Gunner
/*
	Floats and tries to maintain a constant distance to the player
	Every so often, performs an evasive manoeuver and shoots 3 projectiles.
	
	Hard mode: Waits less time before meanoeuvering, Shoots 6 projectiles
*/

PSO.info.enemies[MT_JETTGUNNER] = {
		
		name = "J.Syn Gunner",
		sprite = SPR_JETG,
		frame = A,
		painframe = A,
		
		bhp = 225,
		satk = 20,
		ratk = 30,
		tatk = 20,
		def = 25,
		exp = 11,
		weight = 1,
		
		
		drops = {

			-- Lv1-50
			{650, 1, 50, 1, 15},
			{"beamgun", 1, 50, 1, 30},
			{"alva beamgun", 1, 50, 1, 45},
			{"vita beamgun", 1, 50, 1, 60},
			{"rostig frame", 1, 50, 1, 20},		
			{"aileron frame", 1, 200, 1, 50},
			{"hart frame", 1, 50, 1, 205},
			{"tcsm_gunner", 1, 200, 1, 50},
			{"dimate", 1, 200, 1, 35},
			
			{850, 51, 79, 1, 15},
			{"baston frame", 51, 79,1, 200},
			{"makel frame", 51, 79, 1, 800},
			{"crimson frame", 51, 79, 1, 800},
			{"yasminkov 2000h", 51, 79, 1, 400},
			{"sarcueid", 51, 79, 1, 400},
			{"bravace", 51, 79, 1, 700},
			{"varista", 51, 79, 1, 700},
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_BADNIK,
		weakness = ELEM_ELEC,
		weakness2 = ELEM_EARTH,
		
		reset =	function(mo)
			if not mo.pso.blowdown
				mo.flags = $|MF_NOGRAVITY
				mo.momz = 0
			end
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			local attack_potency = 50
			local targetz = mo.floorz + mo.scale*32
			local rate = FRACUNIT*8/10

			local melee_dist = FRACUNIT*512
			local charge_visual_speed = 10
			local speed_closein = FRACUNIT*10
			
			local waittime = diff >= DIFF_HARD and TICRATE*5/2 or TICRATE*4
			local shotdelay = diff >= DIFF_HARD and 1 or 2

			if mo.spawnz == nil
				mo.spawnz = mo.z 
			end
			
			if mo.spawnz - targetz > FRACUNIT*320
				targetz = mo.spawnz + mo.scale*32
			end

			PSO.lib.enemyManageAggro(mo)
			PSO.lib.enemySetCorruption(mo)
			
			mo.flags = $ & ~(MF_FLOAT|MF_NOGRAVITY)
			
			if not mo.styleroll
				if mo.z < targetz
					mo.momz = $ + rate
				elseif mo.momz < 0
					mo.momz = min(-mo.scale, $ + rate*2)
				end
			end
			
			if mo.target and mo.target.valid
				-- do different actions depending on how far the player is
				
				if mo.styleroll
					
					mo.styleroll = $+1
					
					if mo.styleroll < 16
						mo.rollangle = $ - ANG1*10*mo.styledir
					elseif mo.styleroll >= 16 and mo.styleroll <= 22
						mo.rollangle = $ - ANG1*2*mo.styledir
						
						if mo.styleroll < 20
						and mo.styleroll % shotdelay == 0

							S_StartSound(mo, sfx_trfire)
							local a = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
							local msl = PSO.lib.spawnMissile(mo, MT_JETTBULLET, a, 30*FRACUNIT, attack_potency, DMT_RANGED, 0, 0, mo.height/2, RT_NONE)
							PSO.lib.aimMissile(msl, mo.target)
							msl.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOBLOCKMAP)
							
						end
						
					else
						mo.rollangle = $ - ANG1*10*mo.styledir
					end
					
					if leveltime & 1
						P_SpawnGhostMobj(mo)
					end
					
					if mo.styleroll > TICRATE*3/2
					or P_IsObjectOnGround(mo)
						mo.styleroll = nil
						mo.rollangle = 0
						mo.waittimer = -TICRATE*2
						mo.superarmor = nil
					end
					
				else
					
					mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
					
					local dist = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y)
					
					local movespeed = (dist - melee_dist*3/4) / 96
					if movespeed < 0	-- make it faster when backpedalling
						movespeed = $*2
					end
					-- calculate movespeed based on distance to player
					-- clamp it between the target speeds
					movespeed = min($, speed_closein)
					movespeed = max($, -speed_closein)
					
					P_InstaThrust(mo, mo.angle, movespeed)	
					
					if dist < melee_dist*2
						mo.waittimer = $ and $+1 or 1
						
						if mo.waittimer > waittime
							mo.styleroll = 1
							mo.styledir = P_RandomRange(0, 1) and -1 or 1
							
							local a = mo.angle + ANG1*45 * (mo.styledir)
							P_InstaThrust(mo, a, 5*FRACUNIT)
							mo.momz = 9*FRACUNIT
							mo.superarmor = true	-- cancelling from here would be problematic
						end
					end
				end
			end

			if PSO.lib.enemyPitCheck(mo)
				P_InstaThrust(mo, 0, 0)
			end
			
			mo.sprite = SPR_JETG
			mo.frame = leveltime%4 /2
		end,
}