-- Pop Hat
/*
	Behaves the same as vanilla but the projectile explodes into an AOE.
	Projectile has a 100% freeze chance, AOE freeze chance varies on difficulty
*/

local aoe_potency = 300
local aoe_freeze_chance = 25
			
freeslot("S_PSO_POPSHOTEXPLODE")
function A_PSOPopShotExplode(actor)
	
	if not actor.pso or actor.exploded return end
	for i = 1, 16
		
		local s = P_SpawnMobj(actor.x, actor.y, actor.z, MT_POPSHOT_TRAIL)
		s.angle = (360/16)*i*ANG1
		P_InstaThrust(s, s.angle, FRACUNIT*32)
	end

	if actor.target and actor.target.valid and actor.target.pso and actor.target.pso.corrupted
		local attack_potency = 200
							
		for i = 1, 16
			local msl = PSO.lib.spawnMissile(actor, MT_PSO_MEGID_ENEMY, actor.angle, 4*FRACUNIT, attack_potency, DMT_TECHNIQUE, 0, 0, actor.height/2, RT_NONE)
			P_InstaThrust(msl, 360/16*i*ANG1, FRACUNIT*4)
			msl.scale = $*2/3
			msl.fuse = TICRATE*10
		end
	end
	
	if actor.target and actor.target.valid and actor.target.player
		local h  = PSO.lib.spawnHurtBox(actor.x, actor.y, actor.z - FRACUNIT*48, 600<<FRACBITS, 192<<FRACBITS, actor.target, actor.pso.potency, DMT_RANGED, 4)
		if h.pso.whitelist
			actor.pso.whitelist = h.pso.whitelist
		else
			if not actor.pso.whitelist
				actor.pso.whitelist = {}
			end
			h.pso.whitelist = actor.pso.whitelist
		end
		h.pso.potency = actor.pso.potency
		h.pso.reactionradialangle = true
		h.pso.inflictjellen = actor.pso.inflictjellen
		h.pso.inflictzalure = actor.pso.inflictzalure
		h.pso.blight = actor.pso.blight
		
	else
		PSO.lib.spawnHurtBox(actor.x, actor.y, actor.z - FRACUNIT*48, 192<<FRACBITS, 192<<FRACBITS, actor, aoe_potency, DMT_RANGED, 4, RT_NONE, COND_FREEZE, aoe_freeze_chance)
	end
	
	actor.exploded = true
end

states[S_PSO_POPSHOTEXPLODE] = {SPR_NULL, A, 1, A_PSOPopShotExplode, 0, 0, 	S_SONIC3KBOSSEXPLOSION1}
mobjinfo[MT_POPSHOT].deathstate = S_PSO_POPSHOTEXPLODE

PSO.info.enemies[MT_POPHAT] = {
		
		name = "Pop Hat",
		sprite = SPR_POPH,
		frame = A,
		painframe = A,
		
		bhp = 200,
		satk = 10,
		ratk = 30,
		tatk = 10,
		def = 25,
		exp = 7,
		weight = 4,
		
		drops = {

			-- Lv1-29
			{300, 1, 29, 1, 15},
			{"breaker", 1, 29, 1, 150},
			{"doublegladius", 1, 29, 1, 150},
			{"laser", 1, 29, 1, 150},
			{"canon", 1, 29, 1, 150},
			{"ulrod", 1, 29, 1, 150},
			{"drivel", 1, 29, 1, 100},
			{"alva drivel", 1, 29, 1, 200},
			{"icicle frame", 1, 29, 1, 25},
			{"dimate", 1, 29, 1, 25},

			-- Lv30-50
			{650, 30, 50, 1, 15},
			{"concentrator", 30, 50, 1, 50},
			{"alva concentrator", 30, 50, 1, 75},
			{"vita concentrator", 30, 50, 1, 100},
			{"concentrator", 30, 50, 1, 50},
			{"alva drivel", 30, 50, 1, 75},
			{"vita drivel", 30, 50, 1, 100},
			{"digrinder", 30, 50, 1, 25},
			{"icicle frame", 30, 50, 1, 30},
			{"rostig frame", 30, 50, 1, 30},
			{"fehler frame", 30, 50, 1, 100},			
			{"dimate", 30, 50, 1, 35},
			{"hatted cannon", 30, 50, 1, 200},
			{"permafrost frame", 30, 50, 1, 200},
			{"hart frame", 30, 50, 1, 85},

			-- Lv51-79
			-- permanent drop
			{"hatted cannon", 51, 200,1, 100},
			
			{700, 51, 79, 1, 15},
			{"vita drivel", 51, 79,1, 50},
			{"permafrost frame", 51, 79,1, 100},
			{"rostig frame", 51, 79,1, 30},
			{"fehler frame", 51, 79,1, 80},
			{"photon launcher", 51, 79,1, 250},
			{"saint kilda", 51, 79,1, 250},
			{"charge howzer", 51, 79, 1, 450},
			{"gur bazga", 51, 79, 1, 450},			
			{"hart frame", 51, 79,1, 50},
			{"baston frame", 51, 79,1, 80},
			{"dimate", 51, 79,1, 35},
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_BADNIK,
		weakness = ELEM_FIRE,
		weakness2 = ELEM_DARK,
		
		reset =	function(mo)
			mo.actiontimer = 0	-- reset action if we get staggered etc.
		end,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			local attack_potency = 450
			local freeze_chance = 100
			local max_dist = 1536
			
			local shoot_delay = diff >= DIFF_HARD and TICRATE*4 or TICRATE*7
			
			PSO.lib.enemyManageAggro(mo)
			PSO.lib.enemySetCorruption(mo)
			
			mo.sprite = SPR_POPH
			mo.frame = A
			
			if mo.target and mo.target.valid
				
				mo.waittimer = $ and $+1 or 1
				
				if mo.waittimer > shoot_delay - 10
					mo.frame = B
				end
				
				if R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) > max_dist*FRACUNIT
					mo.waittimer = 0
				end
				
				if mo.waittimer == shoot_delay
					
					mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
					local spd = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) / 60
					local msl = PSO.lib.spawnMissile(mo, MT_POPSHOT, mo.angle, spd, attack_potency, DMT_RANGED, 0, 0, FRACUNIT*32, RT_NONE, COND_FREEZE, freeze_chance)
					PSO.lib.lobMissile(msl, mo.target)
					mo.waittimer = 0
				end
			end
		end,	
}