-- Spincushion
/*
	Drives towards players.
	After a while, turns red and begins charging towards the player, bouncing into walls for 6 seconds.
	Can still be attacked in that state but touching the enemy inflicts heavy damage.
	
	Hard mode: Flips upside down and is able to track the player
*/

local spinframes = {G, I, H, I}
mobjinfo[MT_SPINCUSHION].height = 36*FRACUNIT

PSO.info.enemies[MT_SPINCUSHION] = {
		
		name = "Spincushion",
		sprite = SPR_SHRP,
		frame = A,
		painframe = A,
		
		bhp = 300,
		satk = 35,
		ratk = 20,
		tatk = 20,
		def = 25,
		exp = 12,
		weight = 1,
		
		
		drops = {
			-- Lv 1-10
			{100, 1, 9, 1, 15},				-- meseta (100% mult) 2/5 (Lv1-10)
			{"sword", 1, 9, 1, 100},		-- sword: 	1/100 (Lv1-10)
			{"rifle", 1, 9, 1, 100},		-- rifle: 	1/100 (Lv1-10)
			{"rod", 1, 9, 1, 100},			-- rod: 	1/100 (Lv1-10)
			{"lockgun", 1, 9, 1, 50},
			
			{"monomate", 1, 9, 1, 25},	-- monomate: 1/5 (Lv1-10)
			
			-- Lv10-19
			
			{200, 10, 19, 1, 15},
			{"gigush", 10, 19, 1, 100},
			{"beam", 10, 19, 1, 100},
			{"cane", 10, 19, 1, 100},
			{"monomate", 10, 19, 1, 25},
			{"autogun", 10, 19, 1, 50},

			-- Lv20-29
			{300, 20, 29, 1, 15},
			{"breaker", 20, 29, 1, 150},
			{"doublegladius", 20, 29, 1, 150},
			{"laser", 20, 29, 1, 150},
			{"canon", 20, 29, 1, 150},
			{"ulrod", 20, 29, 1, 150},
			{"peacekeeper", 20, 29, 1, 50},
			{"dimate", 20, 29, 1, 25},

			-- Lv30-50
			{650, 30, 50, 1, 15},
			{"claw", 30, 50, 1, 50},
			{"alva claw", 30, 50, 1, 75},
			{"vita claw", 30, 50, 1, 100},
			{"rostig frame", 30, 50, 1, 30},
			{"fehler frame", 30, 50, 1, 100},			
			{"dimate", 30, 50, 1, 35},
			
			-- Lv51-79
			{850, 51, 79, 1, 15},
			{"baston frame", 51, 79,1, 80},
			{"makel frame", 51, 79, 1, 700},
			{"crimson frame", 51, 79, 1, 700},			
			{"vita claw", 51, 79,1, 50},
			{"sarblight", 51, 79,1, 450},
			{"forblight", 51, 79,1, 450},
			{"red saber", 51, 79, 1, 650},
			{"grand cross", 51, 79, 1, 650},			
			{"permafrost frame", 51, 79,1, 100},
			{"rostig frame", 51, 79,1, 30},
			{"fehler frame", 51, 79,1, 80},			
			{"dimate", 51, 79,1, 35},	
			
		},
		
		stagger_duration = TICRATE/2,
		
		attribute = ATTR_BADNIK,
		weakness = ELEM_ICE,
		weakness2 = ELEM_LIGHT,
		
		thinker = function(mo)
			
			local diff = PSO.lib.questDifficulty()
			
			local attack_potency = 350
			local move_speed = mo.scale*(diff >= DIFF_HARD and 3 or 2)
			local attack_delay = TICRATE*(diff >= DIFF_HARD and 4 or 5)
			local attack_speed = (diff >= DIFF_HARD and 24 or 18)*mo.scale
			local attack_duration = TICRATE*6
			local anim_speed = 2
			
			PSO.lib.enemyManageAggro(mo)
			PSO.lib.enemySetCorruption(mo)
			
			mo.sprite = SPR_SHRP
			mo.frame = A
			
			if mo.target and mo.target.valid
				
				if not mo.attacking
					if P_IsObjectOnGround(mo)
						A_FaceTarget(mo)
						P_InstaThrust(mo, mo.angle, move_speed)
					end
						
					mo.frame = (leveltime%(anim_speed*4)) /anim_speed
						
					mo.attacktimer = $ and $+1 or 1
					
					if mo.attacktimer > attack_delay
						mo.attacking = 1
						S_StartSound(mo, sfx_s3kd8s)
					end
					
				else
					
					if mo.attacking < 10
						mo.frame = mo.attacking/2
					elseif mo.attacking == 10
						mo.momz = FRACUNIT*4
					else
						
						if diff >= DIFF_HARD
						and mo.attacking < 15
							mo.rollangle = $ + ANG1*45
						end
						
						mo.frame = spinframes[leveltime%4 +1]
					end
					
					if mo.attacking == TICRATE*2/3
						A_FaceTarget(mo)
						P_InstaThrust(mo, mo.angle, attack_speed)
						S_StartSound(mo, sfx_shrpgo)
					elseif mo.attacking > TICRATE*2/3
						mo.friction = FRACUNIT
						mo.flags = $|MF_BOUNCE
						
						-- redirect momentum in hard...
						if diff >= DIFF_HARD
							P_Thrust(mo, R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y), FRACUNIT*3/2)
							local newa = R_PointToAngle2(0, 0, mo.momx, mo.momy)
							P_InstaThrust(mo, newa, attack_speed)
							
							-- visual
							local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
							s.color = SKINCOLOR_ORANGE
							s.colorized = true
							s.momz = P_RandomRange(2, 4)*FRACUNIT
							s.momx = P_RandomRange(-3, 3)*FRACUNIT
							s.momy = P_RandomRange(-3, 3)*FRACUNIT
						end
						
						if leveltime & 1
							local a = R_PointToAngle2(0, 0, mo.momx, mo.momy)
							local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 128<<FRACBITS, 40<<FRACBITS, mo, attack_potency, DMT_STRIKE, 2, RT_BLOWDOWN)
							hb.angle = a				
						end
						
						mo.pso.superarmor = true
					end
					
					mo.attacking = $+1
					
					if mo.attacking >= attack_duration
						mo.momz = FRACUNIT*4
						P_InstaThrust(mo, 0, 0)
						mo.attacking = nil
						A_FaceTarget(mo)
						mo.attacktimer = -TICRATE
						mo.rollangle = 0
						
						mo.pso.superarmor = nil
						mo.flags = $ & ~MF_BOUNCE
					end
				end
			end

			if PSO.lib.enemyPitCheck(mo)
				P_InstaThrust(mo, 0, 0)
			end
			
		end,
		
}
