-- Item crates
-- Drop items depending on map & difficulty.
-- The crate changes colour depending on whether it contains a consumable, weapon, unit or a rare item.

-- crate drop tables per map / difficulty:
PSO.info.crateDropTables = {}

freeslot("MT_PSO_CRATE", "S_PSO_CRATE_WAIT", "S_PSO_CRATE", "SPR_PCRT")

mobjinfo[MT_PSO_CRATE] = {
	doomednum = 2505,
	spawnstate = S_PSO_CRATE_WAIT,
	spawnhealth = 1000,
	flags = MF_NOCLIPTHING,
	radius = 64<<FRACBITS,
	height = 128<<FRACBITS,
}	

-- sets up the crate, visual and drops.
function A_SetupCrate(mo)
	
	-- box alredy set up, update it!
	if mo.hnext	
		local curr = mo.hnext
		while curr
			curr.momx = mo.momx
			curr.momy = mo.momy
			curr.momz = mo.momz
			
			curr = curr.hnext
		end
		return
	end
	
	--PSO.lib.dprint("A_SetupCrate()...")
	local boxsize = FRACUNIT*3/4
	
	if not mo.pso
		mo.pso = {name = "Item crate"}
	end
	
	mo.flags = MF_SHOOTABLE
	
	-- only compute drops if they don't come premade (this is the case with boss crates)
	if not mo.drops
		local difficulty = PSO.lib.questDifficulty() or DIFF_NORMAL
		
		-- should we use the map or quest drop table?
		
		local quest
		if server.questdata
			quest = PSO.info.quests[server.questdata.section].quests[server.questdata.index]
		end
		
		if quest and quest.drops and mo.options & 4	-- allow fallback			
			local droptable_diff = quest.drops[difficulty]
			mo.drops = PSO.lib.calcDrops(droptable_diff) or {}		
		else
		
			-- start by figuring out what drops we're giving out to every player
			local droptable_map = PSO.info.crateDropTables[gamemap]
			assert(droptable_map, "No crate drop table for this map!")
			local droptable_diff = droptable_map[difficulty]
			assert(droptable_diff, "No crate drop table for this difficulty!")
			
			-- precompute the drop table
			-- this lets us assign the colour of the box for each player!
			mo.drops = PSO.lib.calcDrops(droptable_diff) or {}
		end
	end
	
	-- get the frame
	local frame = A
	local pdrop = consoleplayer and mo.drops[#consoleplayer+1] or mo.drops[1]
	
	if pdrop and type(pdrop) ~= "number"	-- meseta: just use frame A
		local info = PSO.lib.itemInfo(pdrop)
		
		if info.rarity and info.rarity >= 7
			frame = D
		elseif info.type == IT_WEAPON
			frame = B
		elseif info.type == IT_ARMOR
		or info.type == IT_UNIT
			frame = C
		else
			frame = A	-- consumables / other
		end
	end
	
	-- spawn the crate's sides
	local curr = mo
	local dist = 64*boxsize / FRACUNIT
	local ang = mo.angle
	
	for i = 1, 4
		local x = mo.x + dist*cos(ang)
		local y = mo.y + dist*sin(ang)
		local s = P_SpawnMobj(x, y, mo.z, MT_THOK)
		s.tics = -1
		s.fuse = -1
		s.sprite = SPR_PCRT
		s.frame = frame|FF_PAPERSPRITE
		s.scale = boxsize
		s.angle = ang+ANG1*90
		ang = s.angle
		
		curr.hnext = s
		curr = s
	end
end

states[S_PSO_CRATE_WAIT] = {SPR_NULL, A, 4, nil, 0, 0, S_PSO_CRATE}
states[S_PSO_CRATE] = {SPR_NULL, A, 1, A_SetupCrate, 0, 0, S_PSO_CRATE}


-- destroying the crate...
addHook("ShouldDamage", function(mo, inflictor)
	
	-- no.
	if not mo.pso
		return false
	end
	
	if inflictor and inflictor.valid and inflictor.target and inflictor.target.player
		-- get rid of the sides of the box
		local curr = mo.hnext
		while curr
				
			local ang = curr.angle - ANG1*90
			P_InstaThrust(curr, ang, P_RandomRange(10, 24)*curr.scale)
			curr.momz = P_RandomRange(6, 12)*curr.scale
			curr.fuse = TICRATE*3/4
			curr.tics = TICRATE*3/4
			curr.flags = MF_NOCLIPHEIGHT
				
			curr = curr.hnext
		end
			
		-- drop the item
		
		local has_drops = false
		for k,v in pairs(mo.drops or {})
			has_drops = true
			if type(v) == "number"	-- add variance to meseta
				mo.drops[k] = v + P_RandomRange(-v/10, v/10)
			end
		end
		
		if has_drops
			local drop = P_SpawnMobj(mo.x, mo.y, mo.z + mo.height/2, MT_PSO_DROP)
			drop.psoitems = mo.drops
			--drop.momx = P_RandomRange(-3, 3)*mo.scale
			--drop.momy = P_RandomRange(-3, 3)*mo.scale
			drop.momz = mo.scale*6
			
			if consoleplayer
				local mydrop = mo.drops[#consoleplayer+1]
				if mydrop
				and type(mydrop) ~= "number"
					local minfo = PSO.lib.itemInfo(mydrop)
					if minfo.rarity >= 7
						PSO.lib.startRareDropAnim(consoleplayer, minfo.rarity)
					end
				end
			end
			
		end
		
		local soundhelper = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		soundhelper.state = S_INVISIBLE
		soundhelper.tics = TICRATE
		S_StartSound(soundhelper, sfx_pop)
		
		P_RemoveMobj(mo)
		return
	end
end, MT_PSO_CRATE)
