-- Drop box item

freeslot("sfx_pikup", "sfx_full")
for i = 0, 4
	freeslot("sfx_rdrp"..i)
end

freeslot("MT_PSO_DROP", "S_PSO_DROP", "S_PSO_GOLDDROP", "S_PSO_RAINBOWDROP", "SPR_RDRP", "SPR_RDRB", "SPR_MSTA", "S_PSO_MESETA")
mobjinfo[MT_PSO_DROP] = {
	doomednum = -1,
	spawnstate = S_PSO_DROP,
	radius = 16 << FRACBITS,
	height = 32 << FRACBITS,
	spawnhealth = 1000,
	flags = MF_RUNSPAWNFUNC,
}
states[S_PSO_DROP] = {SPR_RDRP, FF_ANIMATE|FF_FULLBRIGHT, -1, nil, 5, 4, S_PSO_DROP}
states[S_PSO_MESETA] = {SPR_MSTA, FF_ANIMATE|FF_FULLBRIGHT, -1, nil, 5, 4, S_PSO_MESETA}
states[S_PSO_GOLDDROP] = {SPR_RDRP, O|FF_ANIMATE|FF_FULLBRIGHT, -1, nil, 5, 4, S_PSO_GOLDDROP}
states[S_PSO_RAINBOWDROP] = {SPR_RDRP, G|FF_ANIMATE|FF_FULLBRIGHT, -1, nil, 5, 4, S_PSO_RAINBOWDROP}

freeslot("S_PSO_RDRAURA")
states[S_PSO_RDRAURA] = {SPR_RDRB, FF_ANIMATE|FF_ADD|FF_FULLBRIGHT, ((TICRATE/2) *4), nil, 3, TICRATE/2, S_NULL}

local USE_WEAPONDROPMODEL = true

local weapon_col = {
	[WPT_SWORD] = SKINCOLOR_ORANGE,
	[WPT_DOUBLESABER] = SKINCOLOR_ORANGE,
	[WPT_RIFLE] = SKINCOLOR_SAPPHIRE,
	[WPT_SHOTGUN] = SKINCOLOR_SAPPHIRE,
	[WPT_ROD] = SKINCOLOR_YELLOW,
	[WPT_SABER] = SKINCOLOR_ORANGE,
	[WPT_HANDGUN] = SKINCOLOR_SAPPHIRE,
	[WPT_WAND] = SKINCOLOR_YELLOW,
	[WPT_SPEAR] = SKINCOLOR_ORANGE,
	[WPT_TWINSABERS] = SKINCOLOR_ORANGE,
	[WPT_LAUNCHER] = SKINCOLOR_BLUE,
	[WPT_TWINHANDGUNS] = SKINCOLOR_BLUE,
	[WPT_TECHMAG] = SKINCOLOR_YELLOW,
}

local weapon_dropangle = {
	[WPT_SWORD] = ANG1*90,
	[WPT_DOUBLESABER] = ANG1*90,
	[WPT_RIFLE] = ANG1*90,
	[WPT_SHOTGUN] = ANG1*90,
	[WPT_ROD] = ANG1*90,
	[WPT_SABER] = ANG1*90,
	[WPT_HANDGUN] = ANG1*90,
	[WPT_WAND] = ANG1*90,
	[WPT_SPEAR] = ANG1*90,
	[WPT_TWINSABERS] = ANG1*90,
	[WPT_LAUNCHER] = ANG1*90,
	[WPT_TWINHANDGUNS] = ANG1*90,
	[WPT_TECHMAG] = 0,
}

local weapon_offsets = {
	[WPT_SWORD] = -48,
	[WPT_DOUBLESABER] = -48,
	[WPT_RIFLE] = -24,
	[WPT_SHOTGUN] = -24,
	[WPT_ROD] = -48,
	[WPT_SABER] = -32,
	[WPT_HANDGUN] = -12,
	[WPT_WAND] = -32,
	[WPT_SPEAR] = -48,
	[WPT_TWINSABERS] = -32,
	[WPT_LAUNCHER] = -48,
	[WPT_TWINHANDGUNS] = -12,
	[WPT_TECHMAG] = -12,
}


-- returns what drops should be obtained by every player for any given roll in a table.
PSO.lib.calcDrops = function(droptable, level)
	
	-- find the max rate:
	local drops = {}	-- drops per player
	local maxrate = 0
	local curr_odd = 1
	local final_droptable = {}
	
	-- pass 1: figure out the largest number in our drop table
	for i = 1, #droptable
		local drop = droptable[i]
		
		if not level
		or (not drop[2] or level >= drop[2]) and (not drop[3] or level <= drop[3])
			maxrate = max($, drop[5])
		end
	end

	-- pass 2: build the usable drop table using that max rate
	for i = 1, #droptable

		local drop = droptable[i]

		if not level
		or (not drop[2] or level >= drop[2]) and (not drop[3] or level <= drop[3])
				
			local odd = drop[4] * maxrate / drop[5]
				
			curr_odd = $ + odd	
			final_droptable[#final_droptable+1] = {drop[1], curr_odd}	
			--print("Table["..i.."]: {"..drop[1]..", "..curr_odd.."}")
		end
	end
	
	final_droptable[#final_droptable+1] = {nil, maxrate}
	--print("Table["..#final_droptable.."]: {nil, "..maxrate.."}")
	
	-- pass 3: and now we just pick random shit in the table!
	for i = 1, #players
		local num = P_RandomRange(1, maxrate)
		--print("player "..(#p+1)..": "..num)
		-- figure out which item we got!
		for j = 1, #final_droptable-1
			
			local drop = final_droptable[j]
			if num <= drop[2]
				--print("Get "..drop[1])
				if type(drop[1]) == "number"	-- meseta
					drops[i] = drop[1]
				else
					drops[i] = PSO.lib.makeItem(drop[1])	-- item
					local drop = drops[i]
					
					if drop.type == IT_WEAPON and drop.rarity >= 7 and P_RandomRange(0, 1)
						drop.unidentified = true
					end
					
				end
				break
			end
		end	
	end
	return drops
end

-- make a player pick up a drop!
-- if possible, that is.
PSO.lib.playerPickUpDrop = function(p, drop)
	
	if drop.collected
		if drop.collected & (2^(#p)) 
			return	-- player already collected that drop!!
		end
	end
	
	if not drop.psoitem and not drop.psoitems
		return  -- there's.... nothing.
	end
	
	local item = drop.psoitem
	if drop.psoitems
		item = drop.psoitems[#p+1]
	end
	if not item return end	-- there's nothing for you, buddy
	
	local info, take
	
	if type(item) ~= "number"	-- numbers are for meseta
		info = PSO.lib.itemInfo(item)
		if info == PSO.info.items["noref"]
			take = false
		else
			take = PSO.lib.playerGetItem(p, item, drop.psoqty)
		end
		
		if take
			local txt = PSO.lib.getItemName(item)
			if take and type(take) == "number" and take > 1
				txt = $.." x"..take
			end
			
			local rarity = info.rarity
			
			if type(item) == "table"
			and item.unidentified
				rarity = 0
			end
			
			PSO.lib.addDropText(p, rarity, txt)
			
			-- notification if it's a rare item
			if info.rarity and info.rarity >= 7 and drop.psoitems	-- don't show for items intentionally dropped on the floor.
				local name
				if not rarity
					name = PSO.lib.getItemName(item)
				else
					name = info.rarity.."* "..info.name
				end
				
				for p2 in players.iterate do
					if p2 and p2.mo and p2.mo.valid and p2.pso
						PSO.lib.playerPushNotification(p2, "\x82"..p.name.."\x80 found\n\x82"..name, 2)
					end
				end
			end
		end
	else
		take = true
		-- add the meseta to the player's count!
		p.pso.meseta = $+item
		PSO.lib.addDropText(p, -1, PSO.lib.numCommas(item))
		
	end
		
	local globaldrop = (drop.psoitem)	-- if this isn't set then it's a local drop
	
	if not take
		S_StartSound(nil, sfx_full, p)
		return 	-- player couldn't take the item
	else
		
		PSO.lib.runHook("playerPickupItem", p, item, globaldrop)
		
		if type(take) == "number" and drop.psoqty and take < drop.psoqty
			-- took a certain amount
			drop.psoqty = $ - take
		else	-- we took everything
			
			for i = 1, 8
				local s = P_SpawnMobj(drop.x, drop.y, drop.z, MT_SUPERSPARK)
				s.momx = P_RandomRange(-16, 16)*drop.scale
				s.momy = P_RandomRange(-16, 16)*drop.scale
				s.momz = P_RandomRange(-16, 16)*drop.scale
				s.destscale = 1
				s.scalespeed = s.scale/12
				
				-- don't show them to other players!
				if p ~= consoleplayer
				and not globaldrop
					s.flags2 = $|MF2_DONTDRAW
				end
			end
			
			if globaldrop
				P_RemoveMobj(drop)
			else
				-- has everyone taken us?
				-- this player collected us!
				drop.collected = $ | (2^(#p))
				
				local numplayers = 0
				local numcollected = 0
				for pl in players.iterate do
					if drop.collected & (2^(#pl))
					or drop.psoitems and not drop.psoitems[#pl+1]	-- null item also counts as "collected"
						numcollected = $+1
					end
					numplayers = $+1
				end
				
				PSO.lib.dprint(numcollected.."/"..numplayers)
				if numcollected >= numplayers
					PSO.lib.dprint("collected by everyone, vanish")
					P_RemoveMobj(drop)
				end
			end
		end
		
		S_StartSound(p.mo, sfx_pikup, not globaldrop and p or nil)
		return true
	end
end

-- drop thinker handler
-- we will not use MobjThinker anymore because srb2 fucking sucks and it lags some PCs.


-- mo.psoitem: item ref for global drops
-- mo.psoqty: # of a given item (may be nil) can only be set for global drops.
-- mo.psoitems[]: item ref PER PLAYER


--addHook("MobjThinker", function(mo)
local function dropThinker(mo)

	-- first hit the ground...
	
	local rarity = 1
	local typ = 1
	local col = SKINCOLOR_GREEN
	local mapobjectscale = FRACUNIT
	
	local dropsprite
	local dropframe
	local dropangle
	local soffs
	local mypnum = consoleplayer and (#consoleplayer+1) or 1
	local myp = players[mypnum-1]
	local addflag = FF_PAPERSPRITE
	local spritescale = FRACUNIT
	
	
	mo.collected = $ or 0
	
	local itref = mo.psoitem
	if mo.psoitems
		itref = mo.psoitems[mypnum]
		--print(itref)
		--print(itref.ref)
	end

	local dontdraw = 0
	local meseta = 0

	if not itref or mo.collected & (2^(mypnum-1))
		dontdraw = MF2_DONTDRAW
	else
		dontdraw = 0
		-- may happen if you lose connection and rejoin mid game
		if mo.flags2 & MF2_DONTDRAW
			mo.flags2 = $ & ~MF2_DONTDRAW
		end
	end
	
	if itref and type(itref) == "number"
		meseta = MF2_DONTDRAW
		itref = nil
	end
	
	if itref
		
		local info
		local id = true
		
		if type(itref) == "string"
			info = PSO.info.items[itref]
		else
			info = PSO.info.items[itref.ref]
			if itref.unidentified
				id = false
			end
		end
		
		rarity = info and info.rarity or 1
		typ = info and info.type or IT_CONSUMABLE
		
		if rarity < 7
			
			if typ == IT_WEAPON
				col = SKINCOLOR_ORANGE
				
				if USE_WEAPONDROPMODEL
				and id
					col = weapon_col[info.weapontype]
				end
				
			elseif typ == IT_ARMOR
			or typ == IT_UNIT
				col = USE_WEAPONDROPMODEL and SKINCOLOR_RASPBERRY or SKINCOLOR_BLUE
			
			elseif info
				col = SKINCOLOR_GREEN
			else
				col = SKINCOLOR_NONE
			end
		
		elseif rarity < 10
			col = SKINCOLOR_RED
		
		elseif rarity < 13
			col = SKINCOLOR_YELLOW
		else
			col = SKINCOLOR_CARBON
		end
		
		if USE_WEAPONDROPMODEL
		and typ == IT_WEAPON
		and id
			
			dropsprite = info.sprite
			dropframe = info.frame
			dropangle = weapon_dropangle[info.weapontype]
			soffs = weapon_offsets[info.weapontype]
			spritescale = info.spritescale or $
			
			mo.spritexscale = spritescale
			mo.spriteyscale = spritescale
			
			if info.weapontype == WPT_TECHMAG
				addflag = 0
			end
		end
	end
	
	if (P_IsObjectOnGround(mo)) and not mo.extravalue2
	
		P_InstaThrust(mo, 0, 0)
		mo.extravalue2 = 1
		mo.scale = 1
		mo.destscale = mapobjectscale
		
		mo.scalespeed = mapobjectscale/4
		mo.flags2 = $ & ~MF2_DONTDRAW
		mo.flags = $ & ~MF_NOCLIPTHING
		mo.rollangle = dropangle or 0
		
		if meseta
			mo.state = S_PSO_MESETA
		
		elseif rarity >= 10 and rarity <= 12
			mo.state = S_PSO_GOLDDROP
		elseif rarity >= 13
			mo.state = S_PSO_RAINBOWDROP
		end
		if not dontdraw
			S_StartSound(mo, sfx_rdrp0)
		end


		-- drop circle aura
		local ang = leveltime*ANG1*2
		local m = mo
		for i = 1, 6 do
			local x = mo.x + FixedMul(16*mapobjectscale, cos(ang))
			local y = mo.y + FixedMul(16*mapobjectscale, sin(ang))
			local z = mo.z

			local a = P_SpawnMobj(x, y, z, MT_THOK)
			--print("A")
			a.flags = $|MF_SCENERY
			a.sprite = SPR_RDRP
			a.frame = M|FF_ADD|FF_PAPERSPRITE|FF_FULLBRIGHT
			a.scale = mapobjectscale/2
			a.tics = -1
			a.colorized = true
			a.spritexscale = $*12/10
			-- colour:
			a.color = col

			a.angle = ang + ANG1*90


			-- build hnext chain
			m.hnext = a
			m = a
			ang = $ + (360/6)*ANG1
		end
		
		mo.rainbowaura = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		mo.rainbowaura.state = S_INVISIBLE
	elseif mo.extravalue2
	and dropsprite
		
		mo.sprite = dropsprite
		mo.frame = dropframe|addflag
		mo.angle = leveltime*ANG1*2
		
		mo.rollangle = dropangle
		mo.spriteyoffset = -soffs*FRACUNIT
		
		/*if R_PointToAngle(mo.x, mo.y, mo.z) - mo.angle > 0
		and render_hack == "software"
			mo.rollangle = -dropangle	-- this is retarded
		end*/
		
	end
	
	mo.flags2 = $|dontdraw
	
	local spin = mo.hnext
	local ang = leveltime*ANG1*2
	while spin
		
		local x = mo.x + FixedMul(16*mapobjectscale, cos(ang))
		local y = mo.y + FixedMul(16*mapobjectscale, sin(ang))
		local z = mo.z		
		
		P_MoveOrigin(spin, x, y, z)
		spin.angle = ang + ANG1*90
		
		-- for mid-game rejoining
		if not dontdraw
			spin.flags2 = $ & ~MF2_DONTDRAW
		end
		
		spin.flags2 = $|dontdraw|meseta
		spin.tics = 2
		
		ang = $ + (360/6)*ANG1
		spin = spin.hnext
	end

	-- while in the air, move twice as fast
	local gravity = P_GetMobjGravity(mo)
	if mo.extravalue2 == 0	-- in the air
		
		for i = 1, 3
			mo.z = $+mo.momz
			P_TryMove(mo, mo.x+mo.momx, mo.y+mo.momy, true)
			--P_MoveOrigin(mo, mo.x+mo.momx, mo.y+mo.momy, mo.z+mo.momz)
			mo.momz = $ + gravity
			mo.flags2 = $|MF2_DONTDRAW
			local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
			s.flags = $|MF_SCENERY
			s.flags2 = $|dontdraw
		end	
	elseif mo.extravalue2 == 1	-- on the ground ready to be picked up!
		
		-- respawning:
		if mo.cvmem
			mo.cvmem = $-1
			if not mo.cvmem
				--mo.flags = $ | MF_SPECIAL
				mo.flags2 = $ & ~MF2_DONTDRAW
			else
				return
			end	
		end
		
		if mo.fuse == 2 and mo.hnext
			-- remove hnext list
			local m = mo.hnext
			while m
				local buf = m
				m = buf.hnext
				P_RemoveMobj(buf)
			end
		end
		
		if not meseta
			if mo.state != S_PSO_GOLDDROP and not dropsprite
				mo.colorized = true
			else
				mo.colorized = false
			end
		end
		
		
		-- 10* sparkles
		if leveltime % 20 == 0
		and (netgame or (rarity >= 10 and rarity <= 12))
			local spark = P_SpawnMobj(mo.x + P_RandomRange(-32, 32)*mo.scale, mo.y + P_RandomRange(-32, 32)*mo.scale, mo.z, MT_SUPERSPARK)
			spark.momz = mo.scale*3
			spark.tics = $*3
			spark.flags2 = $|dontdraw
			spark.flags = $|MF_SCENERY
			
			if rarity < 10 or rarity > 12
				spark.flags2 = $|MF2_DONTDRAW
			end
		end
		
		if rarity <= 12
			mo.color = col
			
			if meseta and mo.sprite ~= states[S_PSO_MESETA].sprite
			and not dropsprite	-- what the fuck?
				mo.state = S_PSO_MESETA
				mo.colorized = false
			
			elseif not meseta and mo.sprite ~= states[S_PSO_DROP].sprite
			and not dropsprite	-- let it do whatever it wants
				mo.state = S_PSO_DROP
			end
		end	-- 13+
		
		-- don't spawn all that in singleplayer if we can help it.
		if netgame or rarity >= 13
			
			-- local synch
			if rarity >= 13
				if mo.state ~= S_PSO_RAINBOWDROP
				and not dropsprite
				
					mo.state = S_PSO_RAINBOWDROP
				end
			end
			
			local rainbow = {
				SKINCOLOR_PINK,
				SKINCOLOR_PEACHY,
				SKINCOLOR_SUNSET,
				SKINCOLOR_YELLOW,
				SKINCOLOR_LIME,
				SKINCOLOR_MINT,
				SKINCOLOR_AQUA,
				SKINCOLOR_SKY,
				SKINCOLOR_PASTEL,
				SKINCOLOR_RASPBERRY
			}

			local colour = rainbow[leveltime % (10*(#rainbow)) / 10 +1]
			local tcolour = rainbow[(leveltime+10) % (10*(#rainbow)) / 10 +1]
			local trans = (9 - (leveltime % 10)) << V_ALPHASHIFT

			
			local t = mo.rainbowaura --P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			if t and t.valid
				t.sprite = mo.sprite
				t.tics = 3
				t.frame = mo.frame | (trans << FF_TRANSSHIFT) | FF_FULLBRIGHT
				t.scale = mo.scale
				t.flags2 = mo.flags2
				t.angle = mo.angle
				t.colorized = true
				t.color = tcolour
				t.spriteyoffset = mo.spriteyoffset
				t.rollangle = mo.rollangle
				t.flags2 = $|dontdraw
				
				if rarity < 13
					t.flags2 = $|MF2_DONTDRAW
				else
					mo.color = colour
				end
			end

			if leveltime % 12 == 0
				local spark = P_SpawnMobj(mo.x + P_RandomRange(-32, 32)*mo.scale, mo.y + P_RandomRange(-32, 32)*mo.scale, mo.z, MT_SUPERSPARK)
				spark.momz = mo.scale*6
				spark.tics = $*3
				spark.flags = $|MF_SCENERY

				local aura = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
				aura.momz = 8*mapobjectscale
				aura.scale = mo.scale/2
				aura.state = S_PSO_RDRAURA
				aura.colorized = true
				aura.color = mo.color
				aura.scalespeed = mapobjectscale/(TICRATE*2)
				aura.flags2 = $|dontdraw
				
				if rarity < 13
					spark.flags2 = $|MF2_DONTDRAW
					aura.flags2 = $|MF2_DONTDRAW
				end

				if rarity < 15
					aura.destscale = 1
				else
					aura.scalespeed = mapobjectscale*2
				end
			end

			--if rarity >= 15
			if leveltime % 4 == 0
				local spark = P_SpawnMobj(mo.x + P_RandomRange(-12, 12)*mo.scale, mo.y + P_RandomRange(-12, 12)*mo.scale, mo.z, MT_SUPERSPARK)
				spark.momz = mo.scale*16
				spark.scale = $*3/2
				spark.flags2 = $|dontdraw
				spark.flags = $|MF_SCENERY
				
				if rarity < 15
					spark.flags2 = $|MF2_DONTDRAW
				end
			end
		end
	end
end	--, MT_PSO_DROP)

addHook("MapLoad", do
	if gamestate == GS_LEVEL
	and server
		server.dropthinkers = {}
	end
end)

addHook("MapChange", do
	if gamestate == GS_TITLESCREEN or not server return end
	server.dropthinkers = nil
end)

addHook("MobjSpawn", function(mo)
	if gamestate == GS_LEVEL 
	and server
		server.dropthinkers[#server.dropthinkers+1] = mo
	end
end, MT_PSO_DROP)

local function runDropThinkers()
	if gamestate ~= GS_LEVEL return end
	if not server or not server.dropthinkers return end

	local i = #server.dropthinkers
	
	while i
		local mo = server.dropthinkers[i]
		if not mo or not mo.valid
			table.remove(server.dropthinkers, i)
		else
			dropThinker(mo)
		end
		
		i = $-1
	end
end

PSO.lib.registerFrameFunc(runDropThinkers)

-- debug command
COM_AddCommand("psod_drop", function(p, item, unidentified)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local drop = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_PSO_DROP)
	
	if tonumber(item)
		drop.psoitem = tonumber(item)
	else
		local selected = PSO.lib.makeItem(item, nil, unidentified and true or nil)
		
		drop.psoitem = selected	-- global drop
	end
end)

