freeslot("SPR_LSTR")
freeslot("MT_PSO_LASERTRAP", "MT_PSO_LASERTRAP_START", "MT_PSO_LASERTRAP_END")

mobjinfo[MT_PSO_LASERTRAP] = {
	doomednum = 2530,
	spawnstate = S_INVISIBLE,
	height = 30<<FRACBITS,
	radius = 16<<FRACBITS,
	flags = MF_NOGRAVITY,
}

mobjinfo[MT_PSO_LASERTRAP_START] = {
	doomednum = 2531,
	spawnstate = S_INVISIBLE,
	height = 30<<FRACBITS,
	radius = 16<<FRACBITS,
	flags = MF_NOGRAVITY,
}	

mobjinfo[MT_PSO_LASERTRAP_END] = {
	doomednum = 2532,
	spawnstate = S_INVISIBLE,
	height = 30<<FRACBITS,
	radius = 16<<FRACBITS,
	flags = MF_NOGRAVITY,
}	

local function linkStartAndEnd(mo)
	
	local a = mo.mtangle
	for m in mobjs.iterate() do
		if m and m.valid and m.mtangle == a
			
			if m.type == MT_PSO_LASERTRAP_START
				mo.target = m
				m.tics = -1
				
			elseif m.type == MT_PSO_LASERTRAP_END
				mo.tracer = m
				m.tics = -1
			end
		end
		
		-- stop iterating if we have everything set up
		if mo.target and mo.tracer
			break
		end
	end	
	
	if not mo.target and not mo.tracer
		print("\x82".."WARNING: ".."\x80"..": Laser trap with ID "..a.." has no corresponding start or end point and will be removed.")
		
		P_RemoveMobj(mo)
		return
	end
	
	-- uh simply go with that and don't move.
	if not mo.target
		mo.target = mo.tracer
		mo.tracer = nil
	end
	
	-- cache parameters...
	mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
	mo.baseangle = mo.angle
	mo.dist = R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) / FRACUNIT
	
	mo.maxdelay = mo.target.mtz		-- delay before moving in tics.
	
	if mo.tracer
		mo.movespeed = mo.tracer.mtz 	-- move speed in ang per tics
		mo.angdir = mo.tracer.options & 1 and -1 or 1	-- move direction
		mo.targetangle = R_PointToAngle2(mo.x, mo.y, mo.tracer.x, mo.tracer.y)	-- angle to reach
		mo.neverstop = mo.tracer.options & 4	-- never stops spinning!
		
	else
		mo.movespeed = 0
		mo.angdir = 0
		mo.targetangle = mo.angle
	end
	
	mo.delay = 0
	mo.travelangle = 0
	mo.tics = -1
end

local function spawnVisuals(mo)
	
	local hprev = mo
	for i = 1, 2
		local t = P_SpawnMobj(mo.x, mo.y, mo.z + i*8*FRACUNIT, MT_THOK)
		t.sprite = SPR_LSTR
		t.frame = A
		t.color = SKINCOLOR_RED
		t.flags2 = $ | MF2_SPLAT
		t.renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD
		t.tics = -1
		t.rollangle = 0
		t.angle = mo.angle
		
		hprev.hprev = t
		hprev = t
	end
	
	-- spawn it anywhere, we're gonna have to move it anyway...
	mo.hnext = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	mo.hnext.sprite = SPR_LSTR
	mo.hnext.frame = B|FF_PAPERSPRITE|FF_ADD|FF_FULLBRIGHT
	mo.hnext.tics = -1
	mo.hnext.color = SKINCOLOR_RED
	mo.hnext.spritexscale = (mo.dist/2)*FRACUNIT
end

local function updateVisuals(mo)
	
	local hprev = mo.hprev
	while hprev
		hprev.angle = mo.angle + ANG1*90
		hprev = hprev.hprev
		
		mo.color = mo.extravalue2 and SKINCOLOR_EMERALD or SKINCOLOR_RED
		
	end
	
	local halfdist = mo.dist/2
	local hx = mo.x + halfdist*cos(mo.angle)
	local hy = mo.y + halfdist*sin(mo.angle)
	
	P_MoveOrigin(mo.hnext, hx, hy, mo.z+FRACUNIT*8)
	mo.hnext.angle = mo.angle
	if leveltime & 1 or mo.extravalue2
		mo.hnext.flags2 = $|MF2_DONTDRAW
	else
		mo.hnext.flags2 = $ & ~MF2_DONTDRAW
	end
end

local function spawnHurtboxes(mo)
	local hb_width = 32
	local num_hitboxes = (mo.dist / hb_width) +1
	
	mo.hitboxes = {}
	
	for i = 1, num_hitboxes
		
		local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, hb_width<<FRACBITS, 64<<FRACBITS, mo, 10, DMT_STRIKE, -1, RT_NONE)
		hb.pso.fixedpercent = true
		hb.pso.links = mo.hitboxes	-- set the reference, the table will build over time
		mo.hitboxes[#mo.hitboxes+1] = hb
	end
end

local function updateHurtboxes(mo)
	
	local dist = 0
	local hb_width = 32
	
	for i = 1, #mo.hitboxes
		
		local hx = mo.x + dist*cos(mo.angle)
		local hy = mo.y + dist*sin(mo.angle)
		P_SetOrigin(mo.hitboxes[i], hx, hy, mo.z)
		
		dist = $ + hb_width
	end
end

local function cleanWhitelist(mo)

	for i = 1, #mo.hitboxes
		mo.hitboxes[i].pso.whitelist = {}
	end
end


local function moveLaser(mo)
	
	if not mo.tracer return end	-- we have no destination
	
	--print(mo.maxdelay)
	if mo.delay < mo.maxdelay
		mo.delay = $+1
	else
		
		-- start moving...
		if not mo.travelangle	-- towards tracer
			mo.angle = $ + mo.movespeed*ANG1*mo.angdir
			
			local myang = PSO.lib.angToInt(mo.angle)
			local tang = PSO.lib.angToInt(mo.targetangle)
			
			--print(myang..", "..tang..", "..mo.movespeed..", "..mo.movedir)
			
			local diff = abs(myang - tang)
			
			if not mo.neverstop
				if diff <= mo.movespeed
				or (diff <= 360 and diff >= 360 - mo.movespeed)	-- we have to account for this kinda mess too unfortunately lol
					--print("A")
					mo.angle = mo.targetangle
					mo.delay = 0
					mo.travelangle = 1
					cleanWhitelist(mo)	-- reset hitbox leniency
				end
			elseif leveltime%15 == 0
				cleanWhitelist(mo)
			end
		else
			mo.angle = $ - mo.movespeed*ANG1*mo.angdir
			
			local myang = PSO.lib.angToInt(mo.angle)
			local tang = PSO.lib.angToInt(mo.baseangle)
			
			local diff = abs(myang - tang)
			
			if diff <= mo.movespeed
			or diff <= 360 and diff >= 360 - mo.movespeed	-- we have to account for this kinda mess too unfortunately lol
				--print("B")
				mo.angle = mo.baseangle
				mo.delay = 0
				mo.travelangle = 0
				cleanWhitelist(mo)	-- reset hitbox leniency
			end		
		end
	end
end

local function disableLaser(mo)
	
	for i = 1, #mo.hitboxes do
		mo.hitboxes[i].fuse = 1
	end
	mo.hitboxes = nil
end

local function reEnableLaser(mo)
	
	mo.extravalue2 = 0
	spawnHurtboxes(mo)
	updateVisuals(mo)
	updateHurtboxes(mo)
end

addHook("MobjThinker", function(mo)
	
	if not mo.extravalue1
		
		linkStartAndEnd(mo)
		spawnVisuals(mo)
		spawnHurtboxes(mo)
		
		mo.pso = {}	-- this is necessary to avoid bugs
		
		mo.extravalue1 = 1
	end
	
	if mo.extravalue2	-- delete our laser
		
		if mo.hitboxes
			disableLaser(mo)
			updateVisuals(mo)
		end
		
		return
	elseif not mo.hitboxes
		reEnableLaser(mo)	-- instantly rebuild our laser!!
	end
	
	moveLaser(mo)
	updateVisuals(mo)
	updateHurtboxes(mo)
	
end, MT_PSO_LASERTRAP)