-- Here we define Hooks for modding. Trust me, you're going to love this.
-- Taken from SRB2P. Credits to Spectra for initial editing

/*
	"sus is a bad meme"
		- Spectra
						*/


rawset(_G,"hookslist", {

	"playerThink",	-- runs each in the player thinker
	-- function(player_t p)
	
	"playerReset",	-- runs when the player is reset on map switch or after death.
	-- function(player_t p)
	
	"playerAPPRegen",	-- runs when active PP regen takes place (when attacks connect)
	
	"playerDamage",	-- runs when a player deals damage. Returning true overwrites behaviour.
	-- function(player_t p, mobj_t inflictor)
	
	"playerPreDamageCalc",	-- runs at the beginning of damage calc, allows changing the type of the attack before calculations take place for instance.
	-- function(player_t p, mobj_t target, mobj_t inflictor)
	
	"playerDamageCalc",	-- runs when the player's dealt damage is calculated
	-- function(player_t p, mobj_t target, mobj_t inflictor, {int dmg})
	-- modify the table directly to ensure it is passed to subsequent hook calls.
	
	"playerDamaged",	-- runs when the player gets damaged. Returning true overwrites behaviour
	
	"playerReaction",	-- reaction when player is damaged
	"enemyReaction",	-- reaction when enemy is damaged

	"playerHeal",	-- runs when a player gets healed.
	-- function(player_t, mobj_t inflictor, {int heal})
	-- healing amount can be modified with heal[1]
	
	"playerVariance",
	
	"playerCritRate",
	"playerCritDamage",
	
	"playerDamagedCalc",	-- runs when the player's dealt damage is calculated
	-- function(player_t p, mobj_t target, mobj_t inflictor)
	-- modify the table directly to ensure it is passed to subsequent hook calls.
	
	"objectDamage",		-- runs when something deals damage.
	
	"playerChainBreak",	-- runs on chain break.
	-- function(player_t, mobj_t target, mobj_t inflictor, {int dmg}, {boolean chainbroken})
	-- chainbroken is false if the player has no capitalized on chain break yet (condition for skills like chain trigger)
	
	"playerPBGeneration",	-- runs on PB generation.
	
	"playerAction",	-- runs when the player is performing the designated action
	-- function(player_t p)
	-- returning true ends the action.
	
	"playerActionFinish",	-- runs after the player finishes FULLY performing an action, before variables are reset.
	
	"playerActionCancel",	-- runs if the player cancels an action by stepdodging.
	
	"playerActionReset",	-- runs when playerResetAction is called, but only if there was an action to reset.
	
	"playerActionOverride",	-- runs before action, returning true overrides action behaviour.
	
	"playerTechnique",	-- runs when casting a technique.
	"playerCastTime",	-- runs frame 1 of rod/wand cast.
	
	"switchingWeapons",	-- runs when the player is about to switch weapons.
	-- function(player_t p)
	-- for potentials, use to reset variables after switching!
	
	"switchedWeapons",	-- runns when the player has finished switching weapons.
	-- function(player_t p, item_t newweapon)
	-- for potentials, use to set variables when switched to!

	"PPCost",	-- runs when PP cost are determined
	-- function(player_t p, {int pp})
	-- modify the table directly to ensure it is passed to subsequent hook calls.
	-- does not return anything

	"attackSpeed",	-- runs when attack speed is determined
	-- function(player_t, {int speed})
	-- modify the table directly to ensure it is passed to subsequent hook calls.
	-- does not return anything
	
	"playerPickupItem",	-- runs when player picks an item up.
	-- function(player_t, string/table item, boolean dropped by player?}
	
	"playerCalcStats",	-- runs when player stats are calculated.
	-- function(player_t, int level, table stats)
	
	"playerDodge",		-- runs when an attack is dodged
	-- function(player_t, mobj_t inflictor, mobj_t source)	(you can infer that source == inflictor.target)
	
	"playerBlock",	-- runs when an attack is blocked
	-- function(player_t, mobj_t inflictor, mobj_t source) (ditto)
	
	"playerHUD",	-- runs in the hud hook alongside the player status hud.
	-- function (drawer_t, player_t)
})

-- List of hooks that can overwrite default behaviours
rawset(_G, "overwritingHooks", {
	["playerActionOverride"] = 1,
	["playerDamage"] = 1,
	["playerDamaged"] = 1,
	["objectDamage"] = 1,
})

PSO.info.hooks = {}
for i = 1, #hookslist
	PSO.info.hooks[hookslist[i]] = {}	-- table set
end

-- Format of hooks table:
-- PSO.info.hooks[hookname] = {function1, function2, function3, [...], functionn}
-- A table of functions essentially

/********************************************
	Functions - PSO.lib.ddHook, PSO.lib.runHook
********************************************/

-- This function adds a hook to be ran by PSO.lib.runHook
-- For your safety, keep this outside of other hooks, SRB2i based or Vanilla
-- Make sure "hookname" is one of the hooks in "hookslist" above.
-- Example Usage:

/*SRB2P_addHook("BattleThinker", function(btl)
	if (leveltime%35) then return end
	
	print("A")
end)*/
-- When used by PSO.lib.runHook (explained below), all hooks of this type will be ran,
-- in order of addition

-- Enjoy :)
PSO.lib.addHook = function(hookname, func)
	-- check if hookname is valid.
	local ishookvalid
	for i = 1, #hookslist do
		if hookslist[i] == hookname
			ishookvalid = 1
			break
		end
	end

	assert(ishookvalid, "PSO.lib.addHook: Invalid hook name '"..hookname.."'.")

	-- make sure we have a function
	assert(func and type(func) == "function", "PSO.lib.addHook: In hook '"..hookname.."': Invalid argument #2, expected function, got different type.")

	--print("Added a hook type of '"..hookname.."' to SRB2i")

	-- Now we can proceed.
	table.insert(PSO.info.hooks[hookname], func)
end

PSO.lib.runHook = function(hookname, ...)
	if PSO.info.hooks[hookname]
		local numhooks = #PSO.info.hooks[hookname]
		if not numhooks return end	-- Do nothing
		
--		local returnedValues = {}

		-- Iterate over all the functions and run them
		for i = 1, numhooks do
			local args = {...} -- help whats a "..."... uhh... a table?
			local run, res = pcall(PSO.info.hooks[hookname][i], unpack(args))
			
			if not run
				print("\x82".."WARNING: Hook '"..hookname.."' #"..i..": ".."\x80"..res)
				continue
			end
			
			if overwritingHooks[hookname] and res ~= nil
				return res	-- Returning non-nil CAN overwrite certain behaviours
			end
		end
	end
end