freeslot("sfx_login")
-- how many characters we send per ticks? there is an upper limit of around 70-ish characters that can fit in the console buffer.
local CHUNK_SIZE = 60

-- arbitrary encoding for " characters as those cannot be sent thru the console.
-- this is in case internal items ever use ' for their naming convention.
local CONVERT_QUOTE = "@"

-- time to wait before we send the "finish" confirmation.
local CHUNK_WAIT = TICRATE

-- local variable spaghetti.
local data_send_buff = nil
local data_send_order = 1
local data_finish_wait = TICRATE	-- wait 1 full second before finishing
local local_data_loaded = nil		-- quest unlock data loaded
local backup_done = nil

-- begins sending the data
COM_AddCommand("__&begin", function(p)
	
	print("Attempting to begin data transfer from player "..(p.name))
	if p == consoleplayer
		local f = io.openlocal(PSO.info.soloSave, "r")
		
		if f
			data_send_buff = f:read()
			--print("Data to send:")
			--print(data_send_buff)
			data_finish_wait = CHUNK_WAIT
			data_send_order = 1
		end
	end
	p.send_maxorder = 1
end)

COM_AddCommand("s", function(p, order, dta)
	if tonumber(order)
		
		if server.clientbuffers[#p] and server.clientbuffers[#p][1]
			--print("Recived chunk #"..order.." from player "..p.name)
			--print(dta)
			server.clientbuffers[#p][2][tonumber(order)] = dta
			p.send_maxorder = max($, tonumber(order)) 
		else
			print("Blocked attempt to send chunk #"..order.." from player "..p.name..": no permission to write!")
		end
	end
end)

COM_AddCommand("__&finished", function(p)
	
	if server.clientbuffers[#p] and server.clientbuffers[#p][1]
		server.clientbuffers[#p][1] = false
		
		if not p.send_maxorder or p.send_maxorder <= 1	-- the player sent...... nothing.
			-- this means they have no savefile to speak of.
			PSO.lib.openMenu(p, "type_set")
			p.loadsuccess = true	-- consider loading succesful regardless.
			p.psologin = true		-- we consider this player logged in, they're setting their type.
			PSO.lib.playerLoadPrefs(p)
		else
			
			-- tally everything up
			--print("Finished receiving file chunks from player "..p.name)
			for i = 1, p.send_maxorder
				local s = server.clientbuffers[#p][2][i]
				if not s	-- !?
					print("ERR: chunk "..i.." / "..p.send_maxorder.." is missing")	-- TODO
					print("Requeuing player "..p.name.." and reattempting later.")
					server.clientbuffers[i] = {false, {}, ""}
					table.remove(server.send_queue, 1)
					table.insert(server.send_queue, p)
					return
				else
					server.clientbuffers[#p][3] = $..(s:gsub(CONVERT_QUOTE, "\""))
				end
			end
			
			--print(server.clientbuffers[#p][3])		
			
			local dta
			local run, err = pcall(function()
				dta = json.decode(server.clientbuffers[#p][3])
			end)
			
			if not run
				print("\x82".."WARNING: ".."\x80".."Player "..p.name..": Failure to read save file, data may be corrupt.")
				print(err)
				PSO.lib.openMenu(p, "type_set")
				p.loadsuccess = true	-- consider loading succesful regardless.
			else
				-- remove potential level synch before deserializing.
				p.pso.levelsynch = nil
				p.pso.levelsynch_low = nil
				PSO.lib.deserializePlayer(p, dta)
				p.psologinfinish = TICRATE
			end
			p.psologin = true
			PSO.lib.playerLoadPrefs(p)
		end
	end
end)

local function client_saveExists()
	return io.openlocal(PSO.info.soloSave, "r")
end

-- loads the client's save file for singleplayer.
local function client_localLoad(p)
	-- this is local so io.openlocal is fine
	local f = client_saveExists()
	
	-- if the file doesn't exist then we have nothing to load
	if f
		local rdta = f:read()
		local dta
		
		local run = pcall(function()
			dta = json.decode(rdta)
		end)		
		
		p.loadsuccess = true
		
		if not run
			print("\x82".."WARNING: ".."\x80".."Savefile data cannot be read, file may be corrupted.")
			PSO.lib.openMenu(p, "type_set")
			p.loadsuccess = false	-- we failed to load.
		else
			PSO.lib.deserializePlayer(p, dta)
		end
		
		-- if we get to that point, we're loading a blank save.
		if not p.pso.type_initialized
			PSO.lib.openMenu(p, "type_set")
			p.loadsuccess = true	--(rdta == PSO.info.EMPTYSAVE)
		end
		
		f:close()
	else
		-- no save? Have the player set their Type!
		PSO.lib.openMenu(p, "type_set")
		p.loadsuccess = true	-- consider loading succesful
	end
	
	p.psologin = true
	PSO.lib.playerLoadPrefs(p)
end

local function server_loadHost(f)

	if f
		local rdta = f:read()
		local dta
		
		local run = pcall(function()
			dta = json.decode(rdta)
		end)		
		
		server.loadsuccess = true
		
		if not run
			print("\x82".."WARNING: ".."\x80".."Save data cannot be read, file may be corrupted.")
			PSO.lib.openMenu(server, "type_set")
			server.loadsuccess = false	-- we failed to load.
		else
			PSO.lib.deserializePlayer(server, dta)
		end
		
		-- if we get to that point, we're loading a blank save.
		if not server.pso.type_initialized
			PSO.lib.openMenu(server, "type_set")
			server.loadsuccess = true	--(rdta == PSO.info.EMPTYSAVE)
		end
	else
		-- no save? Have the player set their Type!
		PSO.lib.openMenu(server, "type_set")
		server.loadsuccess = true	-- consider loading succesful
	end
	server.psologin = true
	PSO.lib.playerLoadPrefs(server)
end

PSO.lib.localSave = function(p)
	if not p.pso return end	-- player is not ready.
	if PSO.lib.arenaQuestActive() return end	-- do not save while arena quests are ongoing.
	
	assert(p.psologin, "\x82".."Safety trigger: ".."\x80".."Login process still ongoing, data cannot yet be saved.")
	assert(p.loadsuccess, "\x82".."Safety trigger: ".."\x80".."Save data was not loaded succesfully and will not be overwritten as a result. Consider restarting the game or swapping your save file for backed up data.")
	
	local dta = PSO.lib.serializePlayer(p)
	
	local f = io.openlocal(PSO.info.soloSave, "w")
	assert(f, "Couldn't create Savefile, please check your permissions to /luafiles and available disk space")
		
	f:write(dta)
	f:close()
	print("\x83".."Data saved succesfully.")
end

local backup_readme = {
	"This folder contains automatic backups of your save data.",
	"To restore save data from a backup, copy the file you would like to use as '"..PSO.info.soloSave.."'",
	"",
	"The contents of this folder are not erased automatically, please do so at your convenience.",
}

-- backs up the save.
-- for security reasons, this function is not exposed so that it cannot be replaced by other addons
local function backupSave()
	local folder = PSO.info.backupFolder
	local f = client_saveExists()
	
	if f
		
		-- build the file name from the current date.
		
		local name = "pso_backup_"..os.date("%Y-%m-%d_%H_%M_%S")..".sav2"
		--name = "testbackup.sav2"
		print("Backing up save data as '"..name.."'")
		
		local f2 = io.openlocal(folder..name, "w")
		
		assert(f2, "Cannot create backup. Please check your permissions to "..folder.." and available disk space")
		
		local r = f:read()
		-- copy data
		while r
			f2:write(r)
			r = f:read()
		end
		
		f2:close()
		print("\x83".."Backup succesful.")
		
		-- make a readme in the folder too if it does not yet exist
		f2 = io.openlocal(folder.."README.txt", "r")
		if f2
			f2:close()
			return
		else
			f2 = io.openlocal(folder.."README.txt", "w")
			if f2
				for i = 1, #backup_readme
					f2:write(backup_readme[i].."\n")
				end
				f2:close()
			end
		end
		
	else
		print("No save data to back up, continuing...")
		-- otherwise, no data to back up, do nothing
	end
	
end

-- saves progress data using the local buffer.
PSO.lib.server_saveData = function()
	
	local f = io.openlocal(PSO.info.soloData, "w")
	
	assert(f, "Cannot open data file, check your permissions to luafiles and available disk space")
	
	local dta = PSO.lib.sparseStrTable(PSO.lib.getLocalUnlocks())
	dta["version"] = 2	-- append this.
	dta["progress"] = PSO.lib.getLocalQuestProgress() or 0	-- and that
	f:write(json.encode(dta))
	f:close()
end

-- data deletion
PSO.lib.clientNukeData = function(p)	
	local f = io.openlocal(PSO.info.soloData, "w")
	assert(f, "Couldn't create Savefile, please check your permissions to /luafiles and available disk space")
	f:write(PSO.info.EMPTYSAVE)
	f:close()
	--COM_BufInsertText(p, "exitgame") (this crashes the game)
end

PSO.lib.clientNukeSave = function(p)
	if netgame return end	-- no
	
	local f = io.openlocal(PSO.info.soloSave, "w")
	assert(f, "Couldn't create Savefile, please check your permissions to /luafiles and available disk space")
	f:write(PSO.info.EMPTYSAVE)
	f:close()
	--COM_BufInsertText(p, "exitgame") (this crashes the game)
end

-- basically for every map except vr, add them to [1]
-- vr is gone.
local function ver1_dataCompat(d)
	local newtabl = {{}}
	for k,v in pairs(d)
		if tonumber(k) > 1
			if tonumber(k) ~= 6	-- VR map
				for kk, vv in ipairs(d[k])
					table.insert(newtabl[1], vv)
				end
			end
		end
	end
	
	return newtabl
end

local function readProgress(f)
	if f
		local dta = f:read()
		
		if dta == PSO.info.EMPTYSAVE return end	-- nothing to read
		
		local run = pcall(function()
			dta = json.decode(dta)
		end)
		
		if not dta return end
		
		-- ver.1 compat
		local ver1 = not dta.version
		
		if not run
			print("\x82".."WARNING: ".."Server data may be corrupted and cannot be read. Quest unlock progress will not be loaded.")
		else
			
			-- read progress data if we have any
			local progress = dta["progress"] or 0
			
			-- we do some quick fuckshit to ensure we only have ints...
			local newdta = {}
			for k,v in pairs(dta)
				if k ~= "version"
				and k ~= "progress"
					newdta[k] = v
				end
			end
			
			dta = newdta
			
			dta = PSO.lib.sparseIntTable(dta)
			
			if ver1
				dta = ver1_dataCompat(dta)
			end
			
			return dta, progress
		end
	end
end

-- progress data loading.
local function server_readData(f)
	server.psounlocks, server.psoprogress = readProgress(f)
	-- fallback
	server.psounlocks = $ or {}
	server.psoprogress = $ or PSO.lib.serverProgressValue()	-- if 0, doublecheck
	server.init_data = 2
	--print("serv")
	--print(server.psoprogress)
	
	-- otherwise, nothing to read or do
end

local function local_readData()
	local f = io.openlocal(PSO.info.soloData, "r")
	if not f return end	-- file is not created, just skip!
	
	local loc = PSO.lib.getLocalQuestProgress()
	local prog = 0
	loc, prog = readProgress(f)
	loc = $ or {}
	PSO.lib.setLocalUnlocks(loc)
	prog = $ or PSO.lib.localProgressValue()	-- if 0, doublecheck
	PSO.lib.setLocalQuestProgress(prog)

	-- set the sky in the lobby if we have progressed enough
	if gamemap == PSO.info.LOBBY_MAP and PSO.lib.getLocalQuestProgress() >= 16
		P_SetupLevelSky(67, consoleplayer)	-- this hack sucks
	end
	
	--print("loc")
	--print(prog)
	
	local_data_loaded = true
	
	f:close()
	
end

-- handles sending our save data once allowed to by the server
local function handleDataSending()
	if server.clientbuffers[#consoleplayer]
	and server.clientbuffers[#consoleplayer][1]
	and data_send_buff
		local slen = #data_send_buff
		local sub_start = CHUNK_SIZE*(data_send_order-1)
		local sub_end = CHUNK_SIZE*(data_send_order) -1
			
		if sub_start > slen
				
			if data_finish_wait
				data_finish_wait = $-1
			end
				
			if data_finish_wait <= 0
				COM_BufInsertText(consoleplayer, "__&finished")
				data_finish_wait = CHUNK_WAIT
				data_send_buff = nil	-- finished sending.
			end
		else
				
			--print("sending from "..sub_start.." to "..sub_end)
			local chunk = data_send_buff:sub(sub_start, sub_end):gsub("\"", CONVERT_QUOTE)
			--print("Sending chunk: "..chunk)
			COM_BufInsertText(consoleplayer, "s "..data_send_order.." \""..chunk.."\"")
				
			data_send_order = $+1	-- yeah this is hacky but no worries about it, what matters is what the client sends.
		end
	end
end

-- reset client variables when not in a game.
local resetneeded
addHook("ThinkFrame", do
	if gamestate == GS_TITLESCREEN
		if resetneeded
			resetneeded = nil
			data_send_buff = nil
			data_send_order = 1
			data_finish_wait = TICRATE
			local_data_loaded = nil
			backup_done = nil
		end
	else
		resetneeded = true
	end
end)

-- server/client data handling.
local function loadSave()
	
	-- load the progress data...
	if leveltime < 4 return end	-- let things load in properly.
	
	-- initialize progress data server wide.
	if not server.init_data
		io.open(PSO.info.soloData, "r", server_readData)
		server.init_data = 1
		return
	elseif server.init_data < 2
		/*if server.init_data == 2
			server.init_wait = $ and $+1 or 1
			if server.init_wait >= TICRATE*5
				server.init_data = 3
			end
		end*/
		return
	end

	if not local_data_loaded
		local_readData()
	end
	
	if not netgame
		
		-- not in a netgame? just load local save	
		if not server.psologin
		and server.pso	-- wait for this to load
			--CONS_Printf(server, "Loading local save file...")
			client_localLoad(server)			
		end
		
	else
		
		-- mute music while loading
		if consoleplayer
		and (not consoleplayer.psologin or (consoleplayer.psologinfinish and consoleplayer.psologinfinish > TICRATE-10))
			if consoleplayer ~= server
				
				if not S_IdPlaying(sfx_login)
					S_StartSound(nil, sfx_login, consoleplayer)
				end
				
				S_SetInternalMusicVolume(0, consoleplayer)
			end
			
			-- backup my save before the server does anything unbelievably stupid and makes it explode...!!!!!
			if not backup_done
				backupSave()
				backup_done = true
			end
		end
		
		-- order we process requests at.
		-- send queue:
		-- {player_t p, bool in_progress}
		if not server.send_queue
			server.send_queue = {}
		end
		
		-- table that handles save file sending
		-- general pipeline:
		/*
			server initializes send buffer for player
				
			specified player opens their file.
			file is sent in chunks via the 's <order> <chunk>' command
				the chunks are saved in order in the client buffer
				
			specified player finishes sending with the '__&finished' command
				this prompts the server to concatenate all of the saved values in the buffer.
		*/		
		if not server.clientbuffers
			server.clientbuffers = {}
			for i = 0, 31
				server.clientbuffers[i] = {false, {}, ""}
				-- {can write? {chunks}, "final string"}
			end
		end
		
		for p in players.iterate
			
			if not p.pso continue end	-- skip uninitialized players.
			
			if not p.psologin
				-- player requires attention!
				-- did I already request attention?
				
				local dupe
				for i = 1, #server.send_queue
					if server.send_queue[i][1] == p
						dupe = true
						break
					end
				end
				
				-- I didn't!
				if not dupe
					--print("Added player "..p.name.." to data send queue...")
					server.send_queue[#server.send_queue+1] = {p, false}
				end	
			end
		end
		
		-- handle our wait list...
		if server.send_queue[1]
			local qdta = server.send_queue[1]
			local p = qdta[1]
			if not p or not p.valid	-- make sure to check for any client buffer that might remain open when it shouldn't be!
				for i = 0, 31
					if not players[i] or not players[i].valid
						server.clientbuffers[i] = {false, {}, ""}
					end
				end
				table.remove(server.send_queue, 1)
				return
			end
			
			if not qdta[2]	-- we haven't started taking care of them yet.
				
				--print("Processing player "..p.name)
				-- the player in question is the host, we'll use io.open for that one lol.
				if p == server
					io.open(PSO.info.soloSave, "r", server_loadHost)
				else
					-- open communication with that player!
					--print("Ready to receive data from player "..p.name)
					server.clientbuffers[#p] = {true, {}, ""}	-- ready the buffer and allow the player to send data to it.
					
					if p == consoleplayer
						-- get ready to answer...
						-- so, like, do I even have a savefile?
						local f = client_saveExists()
					
						if not f
							-- send a __&finished command, we have nothing *to* send.
							COM_BufInsertText(p, "__&finished")
						else
							f:close()
							COM_BufInsertText(p, "__&begin")	-- start sending data.
						end
					end
				end
				
				-- process has been started
				qdta[2] = true
			else
			
				-- we are in the process of taking care of that player.
				local success	-- set to true when we have succesfully logged this player in
				
				if p == server	-- we're waiitng for that player to be logged in
					success = p.psologin
				else
					if p == consoleplayer
						handleDataSending()	-- send our file to the server...
					end
					success = p.psologin
				end
				
				if success
					print("Player "..p.name.." succesfully loaded!")
					table.remove(server.send_queue, 1)
					return
				end
			end
		end
	end
end

PSO.lib.registerFrameFunc(loadSave)

-- saving the game
addHook("PlayerQuit", function(p)
	if p.pso
	and p == consoleplayer
	and p.psologin	-- make sure we are succesfully logged in
	and not (p.pso and p.pso.nukesave)
		print("Player quit, saving game data...")
		PSO.lib.localSave(p)
	end
end)

addHook("GameQuit", function()
	
	if not server return end		-- title screen etc
	if not consoleplayer return end	-- fuck off dedicated servs
	
	if consoleplayer
	and consoleplayer.psologin	-- make sure we are sucecssfully logged in
	and not (consoleplayer.pso and consoleplayer.pso.nukesave)
		print("Game quit, saving game data...")
		PSO.lib.localSave(consoleplayer)
	end
end)

-- when maps change, also save
addHook("MapChange", function()

	if not server return end		-- title screen etc
	if not consoleplayer return end	-- fuck off dedicated servs
	
	if consoleplayer
	and consoleplayer.psologin	-- make sure we are sucecssfully logged in
	and not (consoleplayer.pso and consoleplayer.pso.nukesave)
		print("Map change, saving game data")
		PSO.lib.localSave(consoleplayer)
	end
end)



-- the HUD for loading is defined here because we want to access local variables.
-- it's a bit of a mess but this is how it is

local function drawLoadingCircle(v, rng_offset, timer_offset)
	
	local timer = ((timer_offset or 0)+leveltime) % TICRATE
	local invt = TICRATE-timer
	local tiltangle = ANGLE_90 + FixedAngle(20*cos(leveltime*ANG2))
	--print(tiltangle/ANG1)
	
	local startx = 160<<FRACBITS
	local starty = 200<<FRACBITS
	--local dist = 100 - max(0, ((invt-TICRATE/2)^2)/3)
	local dist = ease.outcubic(FRACUNIT*timer/TICRATE, 0, 100*FRACUNIT)/FRACUNIT
	--print("d: "..dist)
	--print("e: "..ease.outcubic(FRACUNIT*timer/TICRATE, 0, 100*FRACUNIT)/FRACUNIT)
	
	local center_x = startx + dist*cos(tiltangle) --160<<FRACBITS
	local center_y = starty - dist*sin(tiltangle) --+ (invt^2)/8)<<FRACBITS
	--print(dist*sin(tiltangle))
	local width = 4 + (timer^2)/5
	
	local base_offset = rng_offset
	
	local ang = rng_offset*ANG1*16 + leveltime*ANG2
	local pp = v.cachePatch("H_WFX01")
	
	-- visuals
	--v.drawScaled(startx, starty, FRACUNIT, v.cachePatch("THOKA0"))
	--v.drawScaled(center_x, center_y, FRACUNIT, v.cachePatch("THOKA0"))
	tiltangle = FixedAngle(-20*cos(leveltime*ANG2))
	
	for i = 1, 12
		
		local x = center_x + width*cos(ang)
		local y = center_y + (width*sin(ang)*((timer+(TICRATE/2))) / TICRATE)
		
		local xrel = x - center_x
		local yrel = y - center_y
		
		local x_rot = FixedMul(xrel, cos(tiltangle)) - FixedMul(yrel, sin(tiltangle))
		local y_rot = FixedMul(xrel, sin(tiltangle)) + FixedMul(yrel, cos(tiltangle))
		
		x = center_x+x_rot
		y = center_y+y_rot
		
		local basescale = (FRACUNIT/8 + rng_offset*FRACUNIT/20) * timer*3 / TICRATE
		
		local tflag
		if timer < 30
			tflag = 0
		else
			local tt = timer-30
			tflag = (tt*2)
		end
		
		--local tflag = (10 * timer / TICRATE)
		
		if tflag < 10
			v.drawScaled(x, y, basescale, pp, V_ADD|(tflag << V_ALPHASHIFT), v.getColormap(TC_DEFAULT, SKINCOLOR_GALAXY)) 
		end
		
		rng_offset = ($+base_offset*2)%20
		ang = $ + ANG1*14 + rng_offset*ANG1*4
	end
end

PSO.lib.HU_dataTransfer = function(v, p, c)
	
	-- do not display this for anyone other than the client.
	if p ~= consoleplayer return end
	
	v.drawFill()	-- black screen
	
	drawLoadingCircle(v, 15)
	drawLoadingCircle(v, 5, TICRATE/2)
	
	if p.psologinfinish and p.psologinfinish > TICRATE -10	-- we are going through!!
		local fadestr = TICRATE - p.psologinfinish
		
		-- omg ignition like ignite infinity!!??!?!//1/1/!
		v.drawString(160, 20, "IGNITION COMPLETE!!", (leveltime % 4 < 2 and V_YELLOWMAP or V_ORANGEMAP), "thin-center")
		v.fadeScreen(0, fadestr)
		
		return
	end
		
	-- tell the player what's happening while they're loading....
	local login_queue = server.send_queue

	if login_queue[1]
		local login_p = login_queue[1][1]
		if login_p != p
			-- the server is taking care of someone else.
			-- get my position in the queue
			local mypos = "NaN"
			for i = 1, #login_queue
				if login_queue[i][1] == p
					mypos = i
					break
				end
			end
			
			v.drawString(160, 20, "OTHER PLAYER(S) LOGGING IN...", V_YELLOWMAP|V_ALLOWLOWERCASE, "thin-center")
			v.drawString(160, 30, "\x82".."POSITION IN QUEUE: ".."\x80"..mypos, V_ALLOWLOWERCASE, "thin-center")
		
		else
			-- 's my turn, woo.
			--server.clientbuffers[i] = {false, {}, ""}
			local mybuffer = server.clientbuffers[#p]
			
			if not mybuffer[1]	-- not allowed to send data yet
				v.drawString(160, 20, "WAITING FOR SERVER TO ALLOW TRANSFER...", V_YELLOWMAP|V_ALLOWLOWERCASE, "thin-center")
			else
				if data_send_buff
					v.drawString(160, 20, "SENDING CLIENT INFO...", V_YELLOWMAP|V_ALLOWLOWERCASE, "thin-center")
					
					
					-- calculate chunk length...
					local slen = #data_send_buff
					local numchunks = (slen / CHUNK_SIZE) +1
						
					v.drawString(160, 30, "\x82".."SENT CHUNK(S): ".."\x80"..(p.send_maxorder or "NaN").."\x82".."/".."\x80"..numchunks, V_ALLOWLOWERCASE, "thin-center")
				end
			end
		end
	end
	
	--v.drawString(160, 30, "SENDING CLIENT INFO...", V_YELLOWMAP, "center")
end