-- Handgun PAs

PSO.info.photonArts["handgun_normal"] = {
		name = "Shot",
		desc = "A quick draw from a trusty handgun that allows shooting while retaining your freedom of movement",
		cost = 1,
		normal = true,
		potency = 100,
		speed = 100,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_HGUN1",
		combo = TICRATE/2,		-- can be reused after 17 frames on the timer!
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[HAND_LEFT]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle + ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = -ANG1*10	-- lat. angle
					
					if timer == 1
						
						if combo
							--pos[5] = FixedDiv(mo.height, mo.scale)/2
							pos[6] = 0
							
						else	-- we've just started that attack
							local z = FixedDiv(mo.height, mo.scale)
							pos[4] = z
								
							-- set angle
							pos[6] = ANG1*90 -- vert. angle
						end
					end
					
					if timer < 5
					and not combo
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					
					if timer == 9
						pos[6] = ANG1*30
						
						-- fire the bullet!
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, p.drawangle, 65*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.scale = $/3
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end	
						msl.pso.potency = potency	-- normals shouldn't benefit from PA%
						p.pso.cancancel = 5

						-- yeah not the best way but oh well
						if PSO.lib.hasSkill(p, "ra_zalureshot")
							msl.pso.inflictzalure = TICRATE*4
						end
	
						if PSO.lib.hasSkill(p, "ra_jellenshot")
							msl.pso.inflictjellen = TICRATE*4
						end	
												
						PSO.lib.aimMissile(msl, p.pso.target)

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, sfx_turhit)
						
						p.pso.cancancel = 5

						-- make the missile fade after it gets out of the range.
						local range = wep.range or 512
						msl.fuse = FixedDiv(range*FRACUNIT, max(1, FixedHypot(msl.momx, msl.momy)))/FRACUNIT

						msl.pso.blight = p.pso.blightround
						p.pso.blightround = nil
						
					elseif timer > 9 and timer < 13
						pos[6] = $ - ANG1*10
					end
					
					if timer == TICRATE
						return true
					end
				end,
}

freeslot("MT_PSO_RICOCHET")
mobjinfo[MT_PSO_RICOCHET] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	deathstate = S_NULL,
	xdeathstate = S_NULL,
	radius = 48<<FRACBITS,
	height = 24<<FRACBITS,
	flags = MF_MISSILE|MF_NOGRAVITY,
}

-- collision handler for ricochet
local check_radius = 512

local function ricochetCollide(mo, mo2)
	
	if mo and mo.valid and mo.type == MT_PSO_RICOCHET
	and mo2 and mo2.valid and mo2.flags & MF_ENEMY
		
		mo.whitelist = $ or {}
		
		mo.whitelist[mo2] = true
		
		-- find the closest enemy to ricochet towards!
		local x1 = mo.x - check_radius*FRACUNIT
		local x2 = mo.x + check_radius*FRACUNIT
		local y1 = mo.y - check_radius*FRACUNIT
		local y2 = mo.y + check_radius*FRACUNIT
		local maxdist = check_radius*FRACUNIT
		local foundtarget
					
		local function searchTarget(ref, found)
			if ref != found
			and found and found.valid and found.type ~= MT_PLAYER
							
				if found.flags & MF_SHOOTABLE
				and not mo.whitelist[found]
					
					local dist = R_PointToDist2(mo.x, mo.y, found.x, found.y)
					
					if dist < maxdist
						foundtarget = found
						maxdist = dist
					end
				end
			end
		end
					
		searchBlockmap("objects", searchTarget, mo, x1, x2, y1, y2)
		
		if foundtarget and foundtarget.valid
			
			P_SpawnMobj(foundtarget.x, foundtarget.y, foundtarget.z, MT_THOK)
			
			-- redirect missile.
			mo.fuse = $ + TICRATE	-- extend lifetime
			local speed = 64
			
			mo.angle = R_PointToAngle2(mo.x, mo.y, foundtarget.x, foundtarget.y)
			P_InstaThrust(mo, mo.angle, speed*mo.scale)
			
			local zdiff = foundtarget.z - mo.z
			
			if abs(zdiff) > 64*mo.scale
				local dist = R_PointToDist2(mo.x, mo.y, foundtarget.x, foundtarget.y)
				
				local tt = dist / speed
				
				mo.momz = zdiff / tt
			end
		end
		
		return false
	end
end

addHook("MobjCollide", ricochetCollide, MT_PSO_RICOCHET)
addHook("MobjMoveCollide", ricochetCollide, MT_PSO_RICOCHET)

addHook("MobjThinker", function(mo)
	
	mo.friction = FRACUNIT	-- no slowdown
	if not mo.hurtbox
		mo.hurtbox = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 96<<FRACBITS, 64<<FRACBITS, mo.target, mo.pso.potency, DMT_RANGED, TICRATE*999)
		mo.hurtbox.pso.combobreaker = true
	else
		mo.hurtbox.fuse = 3
		P_MoveOrigin(mo.hurtbox, mo.x, mo.y, mo.z)
	end
	
	if leveltime%5 == 0
		local t = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		t.sprite = SPR_NTHK
		t.frame = B
		t.fuse = 15
		t.tics = 10
		t.scale = mo.scale
		t.destscale = mo.scale*8
		t.scalespeed = mo.scale/4
		t.color = SKINCOLOR_WHITE
	end
	
	mo.coloorized = true
	mo.color = SKINCOLOR_WHITE
	
	A_RifleShotVisual(mo)
end, MT_PSO_RICOCHET)


PSO.info.photonArts["handgun_pa_ricochet"] = {
		name = "Ricochet",
		desc = "A bullet that bounces off of enemies it hits towards another nearby target.",
		cost = 15,
		potency = 380,
		speed = 100,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_HGUN3",
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[HAND_LEFT]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle + ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = -ANG1*10	-- lat. angle
					
					if timer == 1
							
						local z = FixedDiv(mo.height, mo.scale)
						pos[4] = z
								
						-- set angle
						pos[6] = ANG1*90 -- vert. angle
					end
					
					if timer < 5
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					end
					
					if timer < 10
						pos[6] = $ + ANG1*45
					end
					
					
					if timer == 11
						pos[6] = ANG1*30
						
						-- fire the bullet!
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_RICOCHET, p.drawangle, 65*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.scale = $/3
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end
												
						PSO.lib.aimMissile(msl, p.pso.target)

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, sfx_turhit)

						-- make the missile fade after it gets out of the range.
						local range = wep.range or 512
						msl.fuse = FixedDiv(range*FRACUNIT, max(1, FixedHypot(msl.momx, msl.momy)))/FRACUNIT
						
					elseif timer > 11 and timer < 15
						pos[6] = $ - ANG1*10
					end
					
					if timer == TICRATE
						return true
					end
				end,
}

freeslot("MT_PSO_HOMINGSHELL")
mobjinfo[MT_PSO_HOMINGSHELL] = {
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 16<<FRACBITS,
	height = 16<<FRACBITS,
	flags = MF_NOGRAVITY|MF_MISSILE,
	deathstate = S_PSO_RIFLESHOT_EXPLODE,
	xdeathstate = S_PSO_RIFLESHOT_EXPLODE,	
}
local hshell_speed = 24*FRACUNIT
local hshell_accel = FRACUNIT*2

-- basic chase thinker for illgrants
addHook("MobjThinker", function(mo)
	
	if mo.tracer and mo.tracer.valid
		local target = mo.tracer
		local dist = P_AproxDistance(P_AproxDistance(target.x - mo.x, target.y - mo.y), target.z - mo.z)
		
		-- normalize vakues
		local dx = FixedDiv(target.x - mo.x, dist)
		local dy = FixedDiv(target.y - mo.y, dist)
		local dz = FixedDiv(target.z - mo.z, dist)
		
		-- the values we want:
		local desired_momx = FixedMul(dx, hshell_speed)
		local desired_momy = FixedMul(dy, hshell_speed)
		local desired_momz = FixedMul(dz, hshell_speed)
		
		-- how much we'd have to change by
		local steeringx = desired_momx - mo.momx
		local steeringy = desired_momy - mo.momy
		local steeringz = desired_momz - mo.momz
		
		-- cap acceleration
		if abs(steeringx) > hshell_accel
			steeringx = (steeringx > 0 and 1 or -1) * hshell_accel
		end

		if abs(steeringy) > hshell_accel
			steeringy = (steeringy > 0 and 1 or -1) * hshell_accel
		end
		
		if abs(steeringz) > hshell_accel
			steeringz = (steeringz > 0 and 1 or -1) * hshell_accel
		end	
		
		-- apply
		mo.momx = $ + steeringx
		mo.momy = $ + steeringy
		mo.momz = $ + steeringz
	end
	
	A_RifleShotVisual(mo)
	
end, MT_PSO_HOMINGSHELL)

PSO.info.photonArts["handgun_pa_homingshell"] = {
		name = "Homing Shell",
		desc = "Fires a heat seeking photon bullet that explodes on impact with the target.",
		cost = 15,
		potency = 350,
		speed = 100,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_HGUN4",
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[HAND_LEFT]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle + ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = -ANG1*10	-- lat. angle
					
					if timer == 1
							
						local z = FixedDiv(mo.height, mo.scale)
						pos[4] = z
								
						-- set angle
						pos[6] = ANG1*90 -- vert. angle
					end
					
					if timer < 5
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					end
					
					if timer < 10
						pos[6] = $ + ANG1*45
					end
					
					
					if timer == 11
						pos[6] = ANG1*30
						
						-- fire the bullet!
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_HOMINGSHELL, p.drawangle, 35*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.pso.combobreaker = true
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, sfx_turhit)

						msl.fuse = TICRATE*3
						
						if p.pso.target
							msl.tracer = p.pso.target
						end

						local s = P_SpawnMobj(mo.x + pos[2], mo.y + pos[3], mo.z + 64*mo.scale, MT_THOK)
						s.sprite = SPR_NTHK
						s.frame = B|FF_FULLBRIGHT
						s.color = winfo.effectcolor or PSO.info.elementColours[wep.element]
						s.fuse = 10
						s.tics = 10
						s.scale = 1
						s.destscale = mo.scale*10
						s.scalespeed = mo.scale/4
						
					elseif timer > 11 and timer < 15
						pos[6] = $ - ANG1*10
					end
					
					if timer == TICRATE
						return true
					end
				end,
}

PSO.info.photonArts["handgun_pa_charge"] = {
		name = "Power Bullet",
		desc = "Fire a powerful bullet. Charge and release to power the attack up and make it explode on impact",
		cost = 15,
		potency = 380,
		speed = 100,
		
		icon = "H_HGUN2",
		
		use = 	function(p, timer, hold, input)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					local chargethreshold = PSO.lib.hasSkill(p, "ra_quickload") and TICRATE*2/3 or TICRATE
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle + ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = -ANG1*10	-- lat. angle
					
					if timer == 1
						local z = FixedDiv(mo.height, mo.scale)
						pos[4] = z
							
						-- set angle
						pos[6] = ANG1*90 -- vert. angle
					end
					
					if timer < 5
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					if timer >= 5
					and timer < chargethreshold*3/4

						if timer == 5
							S_StartSound(mo, sfx_s3ka3)
						end
						
						-- show the charge...
						local cang = p.drawangle + ANG1*5
						local cx = 50*cos(cang)
						local cy = 50*sin(cang)
						
						if leveltime & 1
							local t = P_SpawnMobj(mo.x + FixedMul(cx, mo.scale), mo.y + FixedMul(cy, mo.scale), mo.z + mo.height*2/3, MT_THOK)
							t.tics = 2
							t.fuse = 2
							t.scale = mo.scale/3
							t.frame = A
							t.color = PSO.info.elementColours[wep.element]
							
							local t2 = P_SpawnMobj(t.x + P_RandomRange(-64, 64)*mo.scale, t.y + P_RandomRange(-64, 64)*mo.scale, t.z + P_RandomRange(-64, 64)*mo.scale, MT_THOK)
							t2.scale = mo.scale/4
							t2.destscale = mo.scale/8
							t2.scalespeed = mo.scale/12
							t2.frame = A
							t2.color = PSO.info.elementColours[wep.element]
							
							t2.momx = (t.x - t2.x)/10
							t2.momy = (t.y - t2.y)/10
							t2.momz = (t.z - t2.z)/10
							
							t2.fuse = 10
							t2.tics = 10
							
						end
					end
					
					if timer >= chargethreshold/3
					and timer < TICRATE*20
					and not hold
						-- just let it go.
						p.pso.actiontimer = timer < chargethreshold and TICRATE*20 or TICRATE*30	-- first firing state.
						pos[6] = timer < chargethreshold and ANG1*40 or ANG1*90
						
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, p.drawangle, 70*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.pso.combobreaker = true	-- charge shots are always combo breakers
						msl.extravalue1 = 1
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end	
						
						PSO.lib.aimMissile(msl, p.pso.target)

						local range = wep.range or 512
						msl.fuse = FixedDiv(range*FRACUNIT, max(1, FixedHypot(msl.momx, msl.momy)))/FRACUNIT

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, sfx_s3k83)						

						if timer >= chargethreshold*3/4
							msl.pso.potency = $*152/100	-- eq to 200 pot at base level
							msl.extravalue2 = 1
							
							P_InstaThrust(mo, p.drawangle, -10*FRACUNIT)
							p.skidtime = 20
							mo.saveang = p.drawangle
							p.pso.actionspeed = 0	-- remove all cmd prompts.
							p.drawangle = mo.angle
							mo.state = S_PLAY_WALK

							if p == displayplayer
								P_StartQuake(FRACUNIT*20, 5)
							end

							local da = p.drawangle - ANG1*20
							for i = 1, 8
								local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
								P_InstaThrust(d, da, FRACUNIT*P_RandomRange(96, 128))
								d.momz = P_RandomRange(0, 4)*mo.scale
								d.scale = mo.scale*2
								d.renderflags = $|RF_PAPERSPRITE
								d.angle = da + ANG1*90
								d.tics = $ + P_RandomRange(1, 6)
								d.destscale = mo.scale*3
									
								da = $ + (40/8)*ANG1
							end	
							
						else
							msl.scale = $ / 2
						end
					end
					
					-- if we keep holding we'll eventually reach full charge!
					if timer == chargethreshold*3/4
						S_StopSound(mo)
						local g = P_SpawnGhostMobj(mo)
						g.destscale = mo.scale*4
						g.scalespeed = mo.scale/12
						S_StartSound(mo, sfx_s3ka8)
					end
					
					if timer >= TICRATE*20
					and timer < TICRATE*20 + 5
						pos[6] = $ - ANG1*10
					elseif timer == TICRATE*21
						return true
					end
					
					
					if timer >= TICRATE*30
					and timer < TICRATE*31
						p.drawangle = mo.saveang
						p.cmd.angleturn = mo.saveang/FRACUNIT
					end	
					
					if timer >= TICRATE*30
					and timer < TICRATE*30 + 10
						pos[6] = $ - ANG1*10
					elseif timer == TICRATE*31
						P_InstaThrust(mo, 0, 0)
						p.drawangle = mo.saveang
						return true
					end
					
					-- hack to prevent us from reaching firing states.
					if p.pso.actiontimer > TICRATE*10
					and p.pso.actiontimer < TICRATE*20	-- if we're past here, don't do it.
						p.pso.actiontimer = TICRATE*10
					end
					
					
					/*if timer == 9
						pos[6] = -ANG1*30
						
						-- fire the bullet!
						PSO.lib.spawnMissile(mo, MT_ROCKET, p.drawangle, 40*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], pos[4])
						
					elseif timer > 9 and timer < 13
						pos[6] = $ + ANG1*10
					end
					
					if timer == TICRATE/2
						return true
					end*/
				end,
		
}