-- Rifle photon arts

freeslot("MT_PSO_RIFLESHOT")
freeslot("S_PSO_RIFLESHOT", "S_PSO_RIFLESHOT_EXPLODE")
mobjinfo[MT_PSO_RIFLESHOT] = {
	doomednum = -1,
	spawnstate = S_PSO_RIFLESHOT,
	deathstate = S_PSO_RIFLESHOT_EXPLODE,
	xdeathstate = S_PSO_RIFLESHOT_EXPLODE,
	height = 64*FRACUNIT,
	radius = 32*FRACUNIT,
	flags = MF_NOGRAVITY|MF_MISSILE,	--|MF_NOCLIPTHING,	-- noclipthing is used for specific purposes.
	spawnhealth = 1000,
}

function A_RifleShotVisual(actor, var1, var2)
	
	local limit = actor.railgun and 256 or 1
	
	if actor.railgun
		actor.flags = $ | MF_NOCLIPTHING
		actor.encheck = {}
	end
	
	for i = 1, limit
	
		actor.renderflags = $|RF_FULLBRIGHT
		if actor.extravalue1
			
			-- spawn some particle trail...
			local x = actor.x + P_RandomRange(-32, 32)*actor.scale
			local y = actor.y + P_RandomRange(-32, 32)*actor.scale
			local z = actor.z + P_RandomRange(-32, 32)*actor.scale
			
			local trail = P_SpawnMobj(x, y, z, MT_SUPERSPARK)
			trail.scale = actor.scale
			trail.tics = 10
			trail.fuse = 10
			trail.renderflags = $|RF_FULLBRIGHT
			trail.color = actor.color
		end
		
		if actor.extravalue2
		and leveltime & 1
			
			local trail = P_SpawnMobj(actor.x, actor.y, actor.z, MT_THOK)
			trail.color = SKINCOLOR_WHITE
			trail.scale = actor.scale*3/2
			trail.tics = 2
			trail.fuse = 2
			trail.frame = A|FF_FULLBRIGHT
			
		end
		
		if actor.prevpos
			for i = 0, 5
				local x = PSO.lib.lerp(actor.prevpos[1], actor.x, FRACUNIT/5 *i)
				local y = PSO.lib.lerp(actor.prevpos[2], actor.y, FRACUNIT/5 *i)
				local z = PSO.lib.lerp(actor.prevpos[3], actor.z, FRACUNIT/5 *i)
				
				local trail = P_SpawnMobj(x, y, z, MT_THOK)
				trail.scale = actor.scale*3/4
				trail.destscale = 1
				trail.scalespeed = actor.scale/12
				trail.tics = 10
				trail.fuse = 10
				trail.frame = A|TR_TRANS50|FF_FULLBRIGHT
				trail.color = actor.color
			end
		end
		
		actor.prevpos = {actor.x, actor.y, actor.z}
		
		if actor.railgun
			if P_TryMove(actor, actor.x+actor.momx, actor.y+actor.momy, true)
				P_SetOrigin(actor, actor.x+actor.momx, actor.y+actor.momy, actor.z+actor.momz*2)
				
				-- check if we are colliding with something currently.
				local function blockmap_look(mo, found)
					if found and found.valid
					and found.flags & MF_SHOOTABLE
					/*and mo.x+mo.radius >= found.x-found.radius	-- boundary checks
					and mo.x-mo.radius <= found.x+found.radius
					and mo.y+mo.radius >= found.y-found.radius
					and mo.y-mo.radius <= found.y+found.radius*/
					and mo.z + mo.height >= found.z
					and mo.z <= found.z + found.height
					and not actor.encheck[found]
						
						-- if we are, spawn an identical missile
						local msl = PSO.lib.spawnMissile(mo.target, MT_PSO_RIFLESHOT, 0, 0, mo.pso.potency, DMT_RANGED, 0, 0, mo.target.height/2)
						msl.extravalue1 = mo.extravalue1
						msl.pso.combobreaker = mo.pso.combobreaker
						msl.angle = mo.angle
						msl.pso.potency = mo.pso.potency	-- careful not to double down on PA%
						msl.pso.blight = mo.pso.blight
						msl.pso.inflictjellen = mo.pso.inflictjellen
						msl.pso.inflictzalure = mo.pso.inflictzalure
						msl.color = mo.color
						msl.target = mo.target
						msl.momx = FRACUNIT	-- this is to ensure they collide
						P_SetOrigin(msl, found.x, found.y, found.z)
						msl.fuse = 2
						actor.encheck[found] = true
						--P_KillMobj(msl)
					end
				end
				
				searchBlockmap("objects", blockmap_look, actor)
			else
				--print("what")
				--P_SpawnMobj(actor.x, actor.y, actor.z, MT_THOK)
				P_KillMobj(actor)
				return
			end
		end
		
	end
end

function A_RifleShotExplode(actor, var1, var2)
	
	if not actor.extravalue2
		for i = 1, 8
			local t = P_SpawnMobj(actor.x, actor.y, actor.z, MT_THOK)
			t.color = actor.color
			t.scale = actor.scale*3/4
			t.renderflags = $|RF_FULLBRIGHT
			t.destscale = 1
			t.scalespeed = actor.scale/12
			t.momx = P_RandomRange(-16, 16)*actor.scale
			t.momy = P_RandomRange(-16, 16)*actor.scale
			t.momz = P_RandomRange(-16, 16)*actor.scale
			t.frame = A|TR_TRANS10
			t.fuse = 20
			t.tics = 20
		end
	else
		for i = 1, 24
			local t = P_SpawnMobj(actor.x, actor.y, actor.z, MT_SUPERSPARK)
			t.color = i & 1 and SKINCOLOR_WHITE or actor.color
			t.scale = actor.scale
			t.renderflags = $|RF_FULLBRIGHT
			t.destscale = 1
			t.scalespeed = actor.scale/12
			t.momx = P_RandomRange(-16, 16)*actor.scale
			t.momy = P_RandomRange(-16, 16)*actor.scale
			t.momz = P_RandomRange(-16, 16)*actor.scale
			t.frame = A|TR_TRANS10
			t.fuse = 20
			t.tics = 20
		end	

		local t = P_SpawnMobj(actor.x, actor.y, actor.z, MT_THOK)
		t.sprite = SPR_NTHK
		t.frame = B|FF_FULLBRIGHT|FF_PAPERSPRITE
		t.color = actor.color
		t.destscale = actor.scale*8
		t.scalespeed = actor.scale/2
		t.tics = 10
		t.fuse = 10
		t.momz = actor.scale*10
		t.angle = actor.angle + ANG1*90
		
		if not actor.cusval
			-- spawn an explosion hitbox:
			local h = PSO.lib.spawnHurtBox(actor.x, actor.y, actor.z, 200<<FRACBITS, 64<<FRACBITS, actor.target, actor.pso.potency/5, DMT_RANGED, 4)
			PSO.lib.copyWeaponProperties(actor, h)
			if h.pso.whitelist
				actor.pso.whitelist = h.pso.whitelist
			else
				if not actor.pso.whitelist
					actor.pso.whitelist = {}
				end
				h.pso.whitelist = actor.pso.whitelist
			end
			h.pso.combobreaker = true
			h.pso.potency = actor.pso.potency/5
		
			-- search for the enemy we hit:
			local enemies = {}
			searchBlockmap("objects", 
				function(ref, found)
					
					if found and found.valid and found.flags & MF_SHOOTABLE
						enemies[#enemies+1] = found
						
					end
				end,
			actor)
			
			table.sort(enemies, 
				function(a, b)
					return R_PointToDist2(actor.x, actor.y, a.x, a.y) < R_PointToDist2(actor.x, actor.y, b.x, b.y)
				end
			)
			
			if enemies[1] and enemies[1].valid
				PSO.lib.dprint(enemies[1])
				h.pso.whitelist[#h.pso.whitelist+1] = enemies[1]
			end
		end
	end
end

states[S_PSO_RIFLESHOT] 		= {SPR_THOK, A|TR_TRANS30, 1, A_RifleShotVisual, 0, 0, S_PSO_RIFLESHOT}
states[S_PSO_RIFLESHOT_EXPLODE] = {SPR_THOK, A|TR_TRANS30, 1, A_RifleShotExplode, 0, 0, S_NULL}

freeslot("MT_PSO_REVOLTAIM")
mobjinfo[MT_PSO_REVOLTAIM] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	flags = MF_NOGRAVITY,
	radius = FRACUNIT*24,
	height = FRACUNIT*48,
}

addHook("MobjThinker", function(mo)
	
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	if leveltime%4 == 0
		
		local x = mo.x + P_RandomRange(-48, 48)*mo.scale
		local y = mo.y + P_RandomRange(-48, 48)*mo.scale
		local z = mo.z + P_RandomRange(0, 48)*mo.scale
		
		local e = P_SpawnMobj(x, y, z, MT_BOSSEXPLODE)
		e.destscale = $*2
		e.momz = mo.scale*3
		
		local t = P_SpawnMobj(x, y, z, MT_THOK)
		t.sprite = SPR_NTHK
		t.frame = B|FF_FULLBRIGHT
		t.color = mo.color
		t.destscale = mo.scale*4
		t.scalespeed = mo.scale/4
		t.tics = 10
		t.fuse = 10
		t.momz = mo.scale*4
		
		S_StartSound(e, sfx_s3k4e)
	end
	mo.shadowscale = FRACUNIT
	
	mo.extravalue1 = $+1
	
	if mo.extravalue1 % 8 == 0
		mo.extravalue2 = $+1
		local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z-mo.scale*32, 64<<FRACBITS, 64<<FRACBITS, mo.target, mo.pso.potency/3, DMT_RANGED, 3)
		PSO.lib.copyWeaponProperties(mo, hb)
		hb.pso.combobreaker = true
		hb.pso.potency = mo.pso.potency/3
		
		if mo.extravalue2 == 1
			local t = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			t.sprite = SPR_NTHK
			t.frame = B|FF_FULLBRIGHT
			t.color = mo.color
			t.destscale = mo.scale*8
			t.scalespeed = mo.scale/2
			t.tics = 14
			t.fuse = 14
			t.momz = mo.scale*6	
			
			local p = consoleplayer
			if mo.target and mo.target.player == p
			and R_PointToDist2(mo.x, mo.y, mo.target.x, mo.target.y) < mo.scale*1024
				P_StartQuake(FRACUNIT*30, 10)
			end
			
		end
		
		if mo.extravalue2 >= 3
			P_RemoveMobj(mo)
		end
	end
end, MT_PSO_REVOLTAIM)

freeslot("MT_PSO_SKYFIRESALVO")
mobjinfo[MT_PSO_SKYFIRESALVO] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	flags = MF_NOGRAVITY,
	radius = FRACUNIT*24,
	height = FRACUNIT*48,
}

addHook("MobjThinker", function(mo)
	
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	local x = mo.x + P_RandomRange(-300, 300)*mo.scale
	local y = mo.y + P_RandomRange(-300, 300)*mo.scale
	
	if leveltime & 1
		local r = P_SpawnMobj(x, y, mo.z, MT_THOK)
		r.sprite = SPR_RAIN
		r.frame = A|TR_TRANS20|FF_FULLBRIGHT|FF_ADD
		r.color = mo.color
		r.colorized = true
		r.destscale = mo.scale*8
		r.fuse = 10
		r.tics = 10
		r.spriteyscale = $*12
		r.spritexscale = $*3
		r.momz = -mo.scale*4
	end
	
	-- keep the way it felt the same as before
	if mo.extravalue1 % 5 == 0

		local t = P_SpawnMobj(x, y, mo.z, MT_THOK)
		t.sprite = SPR_NTHK
		t.frame = B|FF_FULLBRIGHT
		t.color = mo.color
		t.destscale = mo.scale*8
		t.scalespeed = mo.scale/2
		t.tics = 14
		t.fuse = 14
		t.momz = mo.scale*6	
		
		S_StartSound(t, sfx_s3k82)
		S_StartSound(t, sfx_turhit)
		
		if R_PointToDist(t.x, t.y) < FRACUNIT*1024
			P_StartQuake(FRACUNIT*4, 4)
		end	
	end
	
	-- hit once every 12 tics
	if mo.extravalue1 % 12 == 0
		mo.extravalue2 = $+1
		local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z-mo.scale*32, 700<<FRACBITS, 255<<FRACBITS, mo.target, mo.pso.potency/6, DMT_RANGED, 3)
		PSO.lib.copyWeaponProperties(mo, hb)
		hb.pso.combobreaker = true
		hb.pso.potency = mo.pso.potency/10
					
		if mo.extravalue2 >= 6
			P_RemoveMobj(mo)
			return
		end
	end	
	
	mo.extravalue1 = $+1
	
end, MT_PSO_SKYFIRESALVO)


PSO.info.photonArts["rifle_pa_skyfiresalvo"] = {
		name = "Skyfire Salvo",
		desc = "Fire bullets in the air before raining them down around the target.",
		cost = 20,
		potency = 850,
		speed = 10,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_RSHOT5",
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle - ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = ANG1*10	-- lat. angle
					
					if timer == 1
						
						if combo
							pos[5] = FixedDiv(mo.height, mo.scale)/2
							pos[6] = 0
							
						else	-- we've just started that attack
							local z = FixedDiv(mo.height, mo.scale)
							pos[4] = z
								
							-- set angle
							pos[6] = ANG1*90 -- vert. angle
						end
						
						mo.lockangle = p.drawangle
					end
					
					p.drawangle = mo.lockangle
					
					if timer < 5
						pos[6] = $ - ANG1*2
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					end
					
					
					if timer >= 9
					and timer % 2 == 0
					and timer < TICRATE*3/4
						local shot = P_SpawnMobj(mo.x + pos[2], mo.y + pos[3], mo.z + pos[4], MT_PSO_RIFLESHOT)
						shot.color = winfo.effectcolor or PSO.info.elementColours[wep.element]
						shot.momz = mo.scale*32
						shot.momx = P_RandomRange(-8, 8)*mo.scale
						shot.momy = P_RandomRange(-8, 8)*mo.scale
						shot.scale = $/3
						shot.fuse = TICRATE
						shot.flags = MF_NOGRAVITY|MF_NOBLOCKMAP

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, sfx_turhit)
						
						if p == displayplayer
							P_StartQuake(FRACUNIT*2, 2)
						end

						-- small shockwave...
						local t = P_SpawnMobj(shot.x, shot.y, shot.z + mo.scale*32, MT_THOK)
						t.sprite = SPR_NTHK
						t.frame = B|FF_FULLBRIGHT
						t.color = shot.color
						t.scale = mo.scale/8
						t.destscale = mo.scale*3/2
						t.scalespeed = mo.scale/8
						t.tics = 10
						t.fuse = 10
						t.momz = mo.scale*4				
						
					
					elseif timer >= TICRATE*3/4
					and timer < TICRATE*3/2
						
						local t = timer-(TICRATE*3/4)
						local ang = p.drawangle - ANG1*30 - t*ANG1*4
						local x = 24*cos(ang)
						local y = 24*sin(ang)
						
						pos[2] = x
						pos[3] = y
						pos[5] = ANG1*10 - t*ANG1*4	-- lat. angle
						pos[6] = ANG1*90 + t*ANG1*45
					end
					
					if timer == TICRATE*3/4
						
						p.pso.cancancel = TICRATE/4
						local t = p.pso.target
						
						if not t or not t.valid
							local x = mo.x + 512*cos(p.drawangle)
							local y = mo.y + 512*sin(p.drawangle)
							
							t = P_SpawnMobj(x, y, mo.z, MT_THOK)
							t.state = S_INVISIBLE
							t.fuse = 2
						end
						
						if mo.skyfire and mo.skyfire.valid
							P_RemoveMobj(mo.skyfire)
						end
						
						local r = PSO.lib.spawnMissile(mo, MT_PSO_SKYFIRESALVO, 0, 0, potency, DMT_RANGED, 0, 0, 0)
						P_SetOrigin(r, t.x, t.y, t.z)
						r.color = winfo.effectcolor or PSO.info.elementColours[wep.element]
						r.tics = -1
						r.fuse = -1
						mo.skyfire = r
						
					elseif timer == TICRATE*3/2
						return true
					end
					
				end,
}

PSO.info.photonArts["rifle_pa_revoltaim"] = {
		name = "Revolt Aim",
		desc = "Sends a targetted explosion. Hold the button to increase the distance and let go to unleash the attack.",
		cost = 25,
		potency = 950,
		speed = 10,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_RSHOT6",
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle - ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = ANG1*10	-- lat. angle
					
					if timer == 1
						
						if combo
							pos[5] = FixedDiv(mo.height, mo.scale)/2
							pos[6] = 0
							
						else	-- we've just started that attack
							local z = FixedDiv(mo.height, mo.scale)
							pos[4] = z
								
							-- set angle
							pos[6] = ANG1*90 -- vert. angle
						end
						
						mo.lockangle = p.drawangle
					end
					
					p.drawangle = mo.lockangle
					
					if timer < 5
					and not combo
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					
					if timer == 9
						-- spawn the reticle
						mo.reticle = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
						mo.reticledist = 64
						
						S_StartSound(mo, sfx_s3k60)
					end
					
					if timer >= 9
					and (hold or timer <= TICRATE*3/4)
					and timer < TICRATE*3/2
					and mo.reticle and mo.reticle.valid and P_CheckSight(mo, mo.reticle)
						
						local sx = mo.x + mo.reticledist*cos(p.drawangle)
						local sy = mo.y + mo.reticledist*sin(p.drawangle)
						local r = mo.reticle
						
						P_MoveOrigin(r, sx, sy, mo.z + mo.scale*16)
						r.tics = 2
						r.sprite = SPR_TARG
						r.frame = (leveltime%14)/2 |FF_FULLBRIGHT
						r.colorized = true
						r.color = winfo.effectcolor or PSO.info.elementColours[wep.element]
						
						if leveltime & 1
							r.flags2 = $|MF2_DONTDRAW
						else
							r.flags2 = $ & ~MF2_DONTDRAW
						end
						
						if p ~= consoleplayer
							r.flags2 = $ |MF2_DONTDRAW
						end
						
						mo.reticledist = $ + 24
						
					elseif timer > 9
						
						if mo.reticle and mo.reticle.valid
							local msl = PSO.lib.spawnMissile(mo, MT_PSO_REVOLTAIM, 0, 0, potency, DMT_RANGED, 0, 0, 0)
							P_SetOrigin(msl, mo.reticle.x, mo.reticle.y, mo.reticle.z)
							msl.color = winfo.effectcolor or PSO.info.elementColours[wep.element]
							
							local t = P_SpawnMobj(mo.reticle.x, mo.reticle.y, mo.reticle.z, MT_THOK)
							t.sprite = SPR_NTHK
							t.frame = B|FF_FULLBRIGHT
							t.scale = $*4
							t.destscale = 1
							t.fuse = 10
							t.tics = 10
							t.color = msl.color
							t.scalespeed = mo.scale/2
							
							S_StartSound(mo, sfx_s3k83)

							for i = 1, 8
								local bullet = P_SpawnMobj(mo.x + pos[2], mo.y + pos[3], mo.z + pos[4], MT_PSO_RIFLESHOT)
								bullet.flags = MF_NOGRAVITY|MF_NOCLIPTHING
								bullet.angle = p.drawangle + ANG1*P_RandomRange(-10, 10)
								P_InstaThrust(bullet, bullet.angle, P_RandomRange(48, 76)*FRACUNIT)
								bullet.momz = P_RandomRange(-6, 6)*FRACUNIT
								bullet.scale = $/8
								bullet.fuse = P_RandomRange(3, 9)
								if winfo.effectcolor
									bullet.color = winfo.effectcolor
								else
									bullet.color = PSO.info.elementColours[wep.element]
								end	
							end
							
							--P_InstaThrust(mo, p.drawangle, -FRACUNIT*12)
							mo.savet = timer
							--mo.state = S_PLAY_WALK
							
							p.pso.cancancel = 5
						end
						pos[6] = $ + ANG1*45
						
						if timer == mo.savet + TICRATE/2
							return true
						end
					end
					
					/*elseif timer > 9 and timer < 13
						pos[6] = $ - ANG1*10
					end
					
					if timer == TICRATE
						return true
					end*/
				end,
}

PSO.info.photonArts["rifle_normal"] = {
		name = "Shot",
		desc = "A fast moving photon bullet that deals damage on contact.",
		normal = true,
		cost = 2,
		potency = 140,
		speed = 30,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_RSHOT1",
		combo = TICRATE/2,		-- can be reused after 17 frames on the timer!
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle - ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = ANG1*10	-- lat. angle
					
					if timer == 1
						
						if combo
							--pos[4] = FixedDiv(mo.height, mo.scale)*3/4
							pos[6] = 0
							
						else	-- we've just started that attack
							local z = FixedDiv(mo.height, mo.scale)
							pos[4] = z
								
							-- set angle
							pos[6] = ANG1*90 -- vert. angle
						end
					end
					
					if timer < 5
					and not combo
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					
					if timer == 9
						pos[6] = ANG1*30
						
						-- fire the bullet!
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, p.drawangle, 85*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.scale = $/3
						
						-- yeah not the best way but oh well
						if PSO.lib.hasSkill(p, "ra_zalureshot")
							msl.pso.inflictzalure = TICRATE*5
						end
	
						if PSO.lib.hasSkill(p, "ra_jellenshot")
							msl.pso.inflictjellen = TICRATE*5
						end	
	
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end	

						-- small shockwave...
						local t = P_SpawnMobj(msl.x, msl.y, msl.z, MT_THOK)
						t.sprite = SPR_NTHK
						t.frame = B|FF_FULLBRIGHT|FF_PAPERSPRITE
						t.color = msl.color
						t.scale = mo.scale/8
						t.destscale = mo.scale*3/2
						t.scalespeed = mo.scale/8
						t.tics = 10
						t.fuse = 10
						t.momz = mo.scale*4
						t.angle = msl.angle + ANG1*90
						t.momx = msl.momx/8 + mo.momx
						t.momy = msl.momy/8 + mo.momy
							
						msl.pso.potency = potency	-- normals shouldn't benefit from PA%
						p.pso.cancancel = 6
												
						PSO.lib.aimMissile(msl, p.pso.target)

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, sfx_turhit)

						-- make the missile fade after it gets out of the range.
						local range = wep.range or 512
						msl.fuse = FixedDiv(range*FRACUNIT, max(1, FixedHypot(msl.momx, msl.momy)))/FRACUNIT
						
						msl.pso.blight = p.pso.blightround
						p.pso.blightround = nil
						
						-- Railgun!
						if PSO.lib.activeWeaponPotential(p, "only my railgun")
							msl.railgun = true
						end
						
						if p == displayplayer
							P_StartQuake(FRACUNIT*3/2, 4)
						end
						
					elseif timer > 9 and timer < 13
						pos[6] = $ - ANG1*10
					end
					
					if timer == TICRATE
						return true
					end
				end,
}

PSO.info.photonArts["rifle_pa_slideshot"] = {
		name = "Slide Shot",
		desc = "Evade attacks with a quick jump followed by a quick volley of shots.",
		cost = 25,
		potency = 550,
		speed = 100,		-- speed % multiplier. (hack in this case because we'll set it to 0 mid-action)
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_RSHOT4",
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle - ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = ANG1*10	-- lat. angle
					
					if timer == 1
						
						-- propel us depending on our forwardmove/sidemove
						local defaultang = ANG1*180
						local ang = mo.angle
						
						if abs(p.cmd.sidemove) < 25
							ang = mo.angle + defaultang
						else
							if p.cmd.sidemove >= 25
								ang = mo.angle - ANG1*90
							elseif p.cmd.sidemove <= -25
								ang = mo.angle + ANG1*90
							end
						end
						
						mo.z = $+2
						mo.momz = FRACUNIT*4
						mo.state = S_PLAY_FALL
						P_InstaThrust(mo, ang, FRACUNIT*48)
						p.pso.actionspeed = 0
						p.pso.iframes = TICRATE/3
						
						local z = FixedDiv(mo.height, mo.scale)
						pos[4] = z
								
						-- set angle
						pos[6] = ANG1*90 -- vert. angle
					end
					
					if timer < 5
					and not combo
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					if not P_IsObjectOnGround(mo)
						if leveltime & 1
							P_SpawnGhostMobj(mo)
						end
						mo.momx = FixedMul(mo.momx, mo.friction)
						mo.momy = FixedMul(mo.momy, mo.friction)
					end
					
					for i = 1, 3
					
						if timer == 6 + (i-1)*5
							pos[6] = ANG1*30
							
							-- fire the bullet!
							local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, p.drawangle, 85*FRACUNIT, potency/3, DMT_RANGED, pos[2], pos[3], mo.scale*30)
							msl.scale = $/3
							if winfo.effectcolor
								msl.color = winfo.effectcolor
							else
								msl.color = PSO.info.elementColours[wep.element]
							end	
							msl.pso.potency = potency/3	-- normals shouldn't benefit from PA%
							msl.pso.combobreaker = true
													
							PSO.lib.aimMissile(msl, p.pso.target)

							if winfo.sound
								S_StartSound(mo, winfo.sound)
							end
							S_StartSound(mo, sfx_turhit)
							
							if p == displayplayer
								P_StartQuake(FRACUNIT*4, 3)
							end

							-- small shockwave...
							local t = P_SpawnMobj(msl.x, msl.y, msl.z, MT_THOK)
							t.sprite = SPR_NTHK
							t.frame = B|FF_FULLBRIGHT|FF_PAPERSPRITE
							t.color = msl.color
							t.scale = mo.scale/8
							t.destscale = mo.scale*3/2
							t.scalespeed = mo.scale/8
							t.tics = 10
							t.fuse = 10
							t.momz = mo.scale*4
							t.angle = msl.angle + ANG1*90
							t.momx = msl.momx/8 + mo.momx
							t.momy = msl.momy/8 + mo.momy

							-- make the missile fade after it gets out of the range.
							local range = wep.range or 512
							msl.fuse = FixedDiv(range*FRACUNIT, max(1, FixedHypot(msl.momx, msl.momy)))/FRACUNIT

							-- Railgun!
							if PSO.lib.activeWeaponPotential(p, "only my railgun")
								msl.railgun = true
							end	
							
						elseif timer > 6 + (i-1)*5 and timer < 10 + (i-1)*5
							pos[6] = $ - ANG1*10
						end
					end
					
					if timer == TICRATE-5
						return true
					end
				end,
}

-- PIERCINGROUND
freeslot("MT_PIERCINGROUND", "S_PIERCINGROUND", "SPR_SAIR")
mobjinfo[MT_PIERCINGROUND] = {
	doomednum = -1,
	spawnstate = S_PIERCINGROUND,
	deathstate = S_NULL,
	xdeathstate = S_NULL,
	radius = 48<<FRACBITS,
	height = 24<<FRACBITS,
	flags = MF_MISSILE|MF_NOGRAVITY,
}

-- disable collision for barta
addHook("MobjCollide", function(mo1, mo2) return false end, MT_PIERCINGROUND)
addHook("MobjMoveCollide", function(mo1, mo2) return false end, MT_PIERCINGROUND)

function A_PiercingRound(mo)
	
	mo.friction = FRACUNIT	-- no slowdown
	if not mo.hurtbox
		mo.hurtbox = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 96<<FRACBITS, 64<<FRACBITS, mo.target, mo.pso.potency, DMT_RANGED, 10)
		mo.hurtbox.pso.combobreaker = true
		mo.hurtbox.pso.potency = mo.pso.potency
	else
		mo.hurtbox.fuse = 3
		mo.hurtbox.tics = 3
		P_MoveOrigin(mo.hurtbox, mo.x, mo.y, mo.z)
	end
	
	if leveltime%5 == 0
		local t = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		t.sprite = SPR_SAIR
		t.frame = A
		t.fuse = 15
		t.tics = 10
		t.scale = mo.scale
		t.destscale = mo.scale*8
		t.scalespeed = mo.scale/4
		t.color = SKINCOLOR_GREY
	end
	
	mo.coloorized = true
	mo.color = SKINCOLOR_GREY
	
	A_RifleShotVisual(mo)
end

states[S_PIERCINGROUND] = {SPR_TRLS, A, 1, A_PiercingRound, 0, 0, S_PIERCINGROUND}

PSO.info.photonArts["rifle_pa_piercinground"] = {
		name = "Piercing Round",
		desc = "Fires a sharp metal shell that can penetrate targets",
		cost = 15,
		potency = 340,
		speed = 30,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_RSHOT3",
		combo = TICRATE/2,		-- can be reused after 17 frames on the timer!
		
		-- p: player
		-- timer: how long the animation has played for
		-- hold: if true, button is being held down, if not, it's been released
		-- input: timer on how long the button has been held for
		-- combo: non-zero if the attack was performed from a combo.
					-- useful to create attacks that are faster/different if used repeatedly.
		
		-- return true to end the animation
		
		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle - ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					pos[2] = x
					pos[3] = y
					pos[5] = ANG1*10	-- lat. angle
					
					if timer == 1
						
						if combo
							pos[5] = FixedDiv(mo.height, mo.scale)/2
							pos[6] = 0
							
						else	-- we've just started that attack
							local z = FixedDiv(mo.height, mo.scale)
							pos[4] = z
								
							-- set angle
							pos[6] = ANG1*90 -- vert. angle
						end
					end
					
					if timer < 5
					and not combo
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					
					if timer == 9
						pos[6] = ANG1*30
						p.pso.cancancel = 6
						
						-- fire the bullet!
						local obj = PSO.lib.activeWeaponPotential(p, "only my railgun") and MT_PSO_RIFLESHOT or MT_PIERCINGROUND
						
						local msl = PSO.lib.spawnMissile(mo, MT_PIERCINGROUND, p.drawangle, 75*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.scale = $/3
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end	
						msl.pso.potency = potency	-- normals shouldn't benefit from PA%
												
						PSO.lib.aimMissile(msl, p.pso.target)

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, P_RandomRange(0, 1) and sfx_rail1 or sfx_rail2)

						-- make the missile fade after it gets out of the range.
						local range = wep.range or 512
						msl.fuse = FixedDiv(range*FRACUNIT, max(1, FixedHypot(msl.momx, msl.momy)))/FRACUNIT
						
						-- Railgun!
						if PSO.lib.activeWeaponPotential(p, "only my railgun")
							msl.railgun = true
							msl.pso.combobreaker = true
						end						
						
					elseif timer > 9 and timer < 13
						pos[6] = $ - ANG1*10
					end
					
					if timer == TICRATE
						return true
					end
				end,
}

PSO.info.photonArts["rifle_charge"] = {
		name = "Charge Shot",
		desc = "Fire a powerful bullet. Charge and release to power the attack up and make it explode on impact",
		cost = 20,
		potency = 500,
		speed = 100,
		
		icon = "H_RSHOT2",
		
		use = 	function(p, timer, hold, input)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local wep = PSO.lib.getPlayerActiveWeapon(p)
					local winfo = PSO.lib.itemInfo(wep)
					--print(timer)
					-- set new weapon pos:
					local ang = p.drawangle - ANG1*30
					local x = 24*cos(ang)
					local y = 24*sin(ang)
					
					local chargethreshold = PSO.lib.hasSkill(p, "ra_quickload") and TICRATE*2/3 or TICRATE
					
					pos[2] = x
					pos[3] = y
					pos[5] = ANG1*10	-- lat. angle
					
					if timer == 1
						local z = FixedDiv(mo.height, mo.scale)
						pos[4] = z
							
						-- set angle
						pos[6] = ANG1*90 -- vert. angle
					end
					
					if timer < 5
						pos[6] = $ - ANG1*20
						pos[4] = $ - FixedDiv(mo.height, mo.scale)/10
					elseif timer == 5
						pos[6] = 0
					end
					
					if timer >= 5
					and timer < chargethreshold
						
						-- show the charge...
						local cang = p.drawangle - ANG1*5
						local cx = 50*cos(cang)
						local cy = 50*sin(cang)
						
						if timer == 5
							S_StartSound(mo, sfx_s3ka3)
						end
						
						if leveltime & 1
							local t = P_SpawnMobj(mo.x + FixedMul(cx, mo.scale), mo.y + FixedMul(cy, mo.scale), mo.z + mo.height*2/3, MT_THOK)
							t.tics = 2
							t.fuse = 2
							t.scale = mo.scale/3
							t.frame = A
							t.color = winfo.effectcolor or PSO.info.elementColours[wep.element]
							
							local t2 = P_SpawnMobj(t.x + P_RandomRange(-64, 64)*mo.scale, t.y + P_RandomRange(-64, 64)*mo.scale, t.z + P_RandomRange(-64, 64)*mo.scale, MT_THOK)
							t2.scale = mo.scale/4
							t2.destscale = mo.scale/8
							t2.scalespeed = mo.scale/12
							t2.frame = A
							t2.color = winfo.effectcolor or PSO.info.elementColours[wep.element]
							
							t2.momx = (t.x - t2.x)/10
							t2.momy = (t.y - t2.y)/10
							t2.momz = (t.z - t2.z)/10
							
							t2.fuse = 10
							t2.tics = 10
							
						end
					end
					
					if timer >= chargethreshold*2/3
					and timer < TICRATE*20
					and not hold
						-- just let it go.
						S_StopSound(mo)
						p.pso.actiontimer = timer < chargethreshold and TICRATE*20 or TICRATE*30	-- first firing state.
						pos[6] = timer < chargethreshold and ANG1*40 or ANG1*90
						
						local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, p.drawangle, 90*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], mo.scale*30)
						msl.pso.combobreaker = true	-- charge shots are always combo breakers
						msl.extravalue1 = 1
						if winfo.effectcolor
							msl.color = winfo.effectcolor
						else
							msl.color = PSO.info.elementColours[wep.element]
						end	
						
						PSO.lib.aimMissile(msl, p.pso.target)
						
						p.pso.cancancel = TICRATE/4
						
						local range = wep.range or 512
						msl.fuse = FixedDiv(range*FRACUNIT, max(1, FixedHypot(msl.momx, msl.momy)))/FRACUNIT

						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end
						S_StartSound(mo, sfx_s3k83)
						S_StartSound(mo, sfx_turhit)

						-- Railgun!
						if PSO.lib.activeWeaponPotential(p, "only my railgun")
							msl.railgun = true
						end

						-- small shockwave...
						local t = P_SpawnMobj(msl.x, msl.y, msl.z, MT_THOK)
						t.sprite = SPR_NTHK
						t.frame = B|FF_FULLBRIGHT|FF_PAPERSPRITE
						t.color = msl.color
						t.scale = mo.scale/8
						t.destscale = mo.scale*3/2
						t.scalespeed = mo.scale/8
						t.tics = 10
						t.fuse = 10
						t.momz = mo.scale*4
						t.angle = msl.angle + ANG1*90
						t.momx = msl.momx/8 + mo.momx
						t.momy = msl.momy/8 + mo.momy
						
						if p == displayplayer
							P_StartQuake(FRACUNIT*6, 5)
						end

						if timer >= chargethreshold
							msl.pso.potency = $*140/100
							msl.extravalue2 = 1
							
							t.destscale = mo.scale*8
							t.scalespeed = mo.scale/4
							t.momx = 0
							t.momy = 0
							
							P_InstaThrust(mo, p.drawangle, -10*FRACUNIT)
							p.skidtime = 20
							mo.saveang = p.drawangle
							p.pso.actionspeed = 2	-- (keep some actually because otherwise dodgecancel acts very weird)
							p.drawangle = mo.angle
							mo.state = S_PLAY_WALK

							if p == displayplayer
								P_StartQuake(FRACUNIT*20, 5)
							end

							local da = p.drawangle - ANG1*20
							for i = 1, 8
								local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
								P_InstaThrust(d, da, FRACUNIT*P_RandomRange(96, 128))
								d.momz = P_RandomRange(0, 4)*mo.scale
								d.scale = mo.scale*2
								d.renderflags = $|RF_PAPERSPRITE
								d.angle = da + ANG1*90
								d.tics = $ + P_RandomRange(1, 6)
								d.destscale = mo.scale*3
									
								da = $ + (40/8)*ANG1
							end		
							
						else
							msl.scale = $ / 2
						end
					end
					
					-- if we keep holding we'll eventually reach full charge!
					if timer == chargethreshold
						S_StopSound(mo)
						local g = P_SpawnGhostMobj(mo)
						g.destscale = mo.scale*4
						g.scalespeed = mo.scale/12
						S_StartSound(mo, sfx_s3ka8)
					end
					
					if timer >= TICRATE*20
					and timer < TICRATE*20 + 5
						pos[6] = $ - ANG1*10
					elseif timer == TICRATE*21
						return true
					end
					
					
					if timer >= TICRATE*30
					and timer < TICRATE*31
					and mo.saveang
						p.drawangle = mo.saveang
						p.cmd.angleturn = mo.saveang/FRACUNIT
					end	
					
					if timer >= TICRATE*30
					and timer < TICRATE*30 + 10
						pos[6] = $ - ANG1*10
					elseif timer == TICRATE*31
						P_InstaThrust(mo, 0, 0)
						p.drawangle = mo.saveang
						return true
					end
					
					-- hack to prevent us from reaching firing states.
					if p.pso.actiontimer > TICRATE*10
					and p.pso.actiontimer < TICRATE*20	-- if we're past here, don't do it.
						p.pso.actiontimer = TICRATE*10
					end
					
					
					/*if timer == 9
						pos[6] = -ANG1*30
						
						-- fire the bullet!
						PSO.lib.spawnMissile(mo, MT_ROCKET, p.drawangle, 40*FRACUNIT, potency, DMT_RANGED, pos[2], pos[3], pos[4])
						
					elseif timer > 9 and timer < 13
						pos[6] = $ + ANG1*10
					end
					
					if timer == TICRATE/2
						return true
					end*/
				end,
		
}