-- Sword photon arts

local function spear_trail(mo, pos, prevspearpos)
	
	local wep = PSO.lib.getPlayerActiveWeapon(mo.player)
	local winfo = PSO.lib.itemInfo(wep)
	
	for i = 0, 5
							
		local x = PSO.lib.lerp(mo.x + FixedMul(prevspearpos[1], mo.scale), mo.x + FixedMul(pos[2], mo.scale), FRACUNIT/5 *i)
		local y = PSO.lib.lerp(mo.y + FixedMul(prevspearpos[2], mo.scale), mo.y + FixedMul(pos[3], mo.scale), FRACUNIT/5 *i)
		local z = PSO.lib.lerp(mo.z + FixedMul(prevspearpos[3], mo.scale), mo.z + FixedMul(pos[4], mo.scale), FRACUNIT/5 *i)
		local a = PSO.lib.lerp(prevspearpos[4], pos[6], FRACUNIT/5 *i)
							
		local trail = P_SpawnMobj(x, y, z - mo.scale*8, MT_THOK)
		
		if pos[7] and pos[7].use
			P_CreateFloorSpriteSlope(trail)
						
			trail.flags2 = $ | MF2_SPLAT
			trail.renderflags = $|RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD
			trail.floorspriteslope.o = pos[7].o
			trail.floorspriteslope.xydirection = pos[7].xy
			trail.floorspriteslope.zangle = pos[7].za
						
			trail.sprite = SPR_STRL
			trail.frame = C|TR_TRANS30
			
			trail.angle = 0	
			trail.rollangle = a

		else
			trail.sprite = SPR_STRL
			trail.frame = A|TR_TRANS30|FF_PAPERSPRITE
			trail.angle = mo.player.drawangle + pos[5]
			trail.rollangle = a
		end
		
		trail.tics = 10
		trail.fuse = 10						
		trail.color = winfo.effectcolor or SKINCOLOR_CYAN
	end
end

PSO.info.photonArts["spear_normal"] = {
		name = "Attack",
		desc = "A piercing attack that can be chained up to 3 times. Restores PP on hit.",
		cost = 0,
		normal = true,
		potency = 210,
		speed = 20,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		combo = TICRATE*20,	-- hack for the 3-hit combo
		icon = "H_SPATK1",

		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local winfo = PSO.lib.itemInfo(PSO.lib.getPlayerActiveWeapon(p))
					
					-- set pos floorsprite

					local prevspearpos = {
						pos[2],
						pos[3],
						pos[4],
						pos[6],	-- rollangle
					}
					
					if timer == 1
						mo.hb = {}
					end
					
					if combo < 2
						
						pos[7] = {use = false}
						
						local ang = p.drawangle - ANG1*90
						local x = 16*cos(ang)
						local y = 16*sin(ang)
						
						-- further dist for the skewer:
						x = x + (min(6, timer)*cos(p.drawangle)*16)
						y = y + (min(6, timer)*sin(p.drawangle)*16)

						pos[2] = x
						pos[3] = y
						pos[4] = mo.height/2
						pos[5] = ANG1*10
						pos[6] = 0

						-- damage box:
						if timer == TICRATE/8
							--(x, y, z, w, t, source, potency, dtype, duration)
							local links = {}
							
							for i = 1, 2
								local fx = x + cos(p.drawangle)*60*i
								local fy = y + sin(p.drawangle)*60*i
								
								local hb = PSO.lib.spawnHurtBox(mo.x+fx, mo.y+fy, mo.z - mo.scale*64, 180<<FRACBITS, 140<<FRACBITS, mo, potency, DMT_STRIKE, 8)
								hb.pso.potency = potency	-- normals shouldn't benefit from PA%
								hb.pso.ppregen = 6
								hb.pso.reactionangle = p.drawangle
								links[#links+1] = hb
								hb.pso.links = links
								hb.pso.hitlag = 4
								hb.pso.screenshake = {intensity = FRACUNIT*4, time=4}								
							end
							
							p.pso.cancancel = 4
							
							if winfo.sound
								S_StartSound(mo, winfo.sound)
							end
						end
						
						if timer >= TICRATE/2
						and timer < TICRATE*20
							p.pso.actiontimer = TICRATE*20
						end

						if timer >= TICRATE*20 + TICRATE/3
							return true
						end
					elseif combo == 2
						
						if timer < TICRATE/4
							
							if timer%3 == 0
								if winfo.sound
									S_StartSound(mo, winfo.sound)
								end
							end
							
							local t = min(timer, TICRATE/4)
							local a = p.drawangle - ANG1*140 + t*ANG1*5
							local x = 64*cos(a)
							local y = 64*sin(a)
							
							pos[2] = x
							pos[3] = y
							pos[4] = mo.height - FRACUNIT*t
							
							pos[5] = 0
							pos[6] = timer*ANG1*65
							
							pos[7] = {
								use = true,
								o = {x = mo.x+x, y = mo.y+y, z = mo.z + pos[4]},
								xy = p.drawangle + a+ ANG1*45,
								za = ANG1*30,
							}				
						
						else
							
							local t = timer - (TICRATE/4)
							local a = p.drawangle + min(t, 3)*ANG1*30
							
							local ang = p.drawangle - ANG1*90
							local x = 32*cos(ang)
							local y = 32*sin(ang)
							
							-- further dist for the skewer:
							x = x + (min(4, t)*cos(p.drawangle)*24)
							y = y + (min(4, t)*sin(p.drawangle)*24)
							
							pos[2] = x
							pos[3] = y
							pos[4] = mo.height/2
							pos[5] = 0
							pos[6] = a
							
							pos[7].o = {x = mo.x+pos[2], y = mo.y+pos[3], z = mo.z+pos[4]}
							pos[7].xy = p.drawangle + ANG1*90

							local fx = x --+ cos(p.drawangle)*50*i
							local fy = y --+ sin(p.drawangle)*50*i
							
							if t == 1
								if winfo.sound
									S_StartSound(mo, winfo.sound)
								end
								
								local links = {}
								local hb
								
								mo.hb = {}
								for i = 1, 2
									
									hb = PSO.lib.spawnHurtBox(mo.x+fx, mo.y+fy, mo.z - mo.scale*64, 180<<FRACBITS, 140<<FRACBITS, mo, potency*2, DMT_STRIKE, 8, RT_KNOCKBACK)
									hb.pso.potency = potency*2	-- normals shouldn't benefit from PA%
									hb.pso.ppregen = 14
									hb.pso.reactionangle = p.drawangle
									hb.pso.reactionstrength = 50
									links[#links+1] = hb
									hb.pso.links = links
									hb.pso.hitlag = 4
									hb.pso.screenshake = {intensity = FRACUNIT*8, time=4}
									mo.hb[i] = hb
								end
							
							elseif t == 2
								p.pso.cancancel = 4
								
								fx = x + 16*cos(p.drawangle)*24
								fy = y + 16*sin(p.drawangle)*24
								
								local efc = P_SpawnMobj(mo.x+fx, mo.y+fy, mo.z + mo.height/2, MT_THOK)
								efc.color = winfo.effectcolor or SKINCOLOR_CYAN
								efc.sprite = SPR_ECRC
								efc.frame = C|TR_TRANS40|FF_PAPERSPRITE
								efc.angle = p.drawangle + ANG1*90
								efc.scale = mo.scale
								efc.destscale = mo.scale*10
								efc.scalespeed = mo.scale/4
								efc.fuse = 20
								efc.tics = 20
								
								
								if P_IsObjectOnGround(mo)
									mo.speargrounded = true
									P_InstaThrust(mo, p.drawangle, FRACUNIT*48)
									mo.state = S_PLAY_WALK
								end
								
								if p == displayplayer
									P_StartQuake(FRACUNIT*20, 4)
								end
							end
							
							mo.momx = $*90/100
							mo.momy = $*90/100
							
							if mo.speargrounded
							and not P_TryMove(mo, mo.x+mo.momx, mo.y+mo.momy, false)
								P_InstaThrust(mo, 0, 0)
							end	

							if mo.hb
								for i = 1, 2
									local hb = mo.hb[i]
									if hb and hb.valid
										if p.cmd.forwardmove > 0
											hb.pso.momentummult = nil
										else
											hb.pso.momentummult = 0
										end
										P_SetOrigin(hb, hb.x+mo.momx*2, hb.y+mo.momy*2, hb.z+mo.momz*2)
									end
								end
							end						
							
							if timer >= TICRATE*3/4
								return true
							end
							
						end			
					end
					
					if timer > 1
						spear_trail(mo, pos, prevspearpos)
					end
				end,
}

freeslot("MT_PSO_SPEARSCYTHE", "SPR_SLAS")
mobjinfo[MT_PSO_SPEARSCYTHE] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	deathstate = S_NULL,
	xdeathstate = S_NULL,
	radius = 5<<FRACBITS,
	height = 10<<FRACBITS,
	flags = MF_NOGRAVITY,
}

addHook("MobjThinker", function(mo)
	
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	mo.sprite = SPR_SLAS
	mo.frame = A|TR_TRANS20|FF_ADD|FF_FULLBRIGHT
	
	if leveltime & 1
		mo.flags2 = $ | MF2_DONTDRAW
	else
		mo.flags2 = $ & ~MF2_DONTDRAW
	end
	
	if mo.extravalue1 == 0
		
		for i = 1, 3
			local g = P_SpawnGhostMobj(mo)
			g.destscale = mo.scale*(i+1)
			g.destscsle = (mo.scale/4)*i
			g.fuse = 10
		end
		
		for i = 1, 32
			local t = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			t.color = mo.color
			t.frame = $|FF_ADD
			t.fuse = 10
			t.tics = t.fuse
			t.scale = mo.scale/5
			local a = R_PointToAngle(mo.x, mo.y) + ANG1*90
			local z = -1
			if i & 1
				a = $ + ANG1*180
				z = 1
			end	
			
			local spd = P_RandomRange(7, 32)*FRACUNIT
			P_InstaThrust(t, a, spd)
			t.momz = spd*z + P_RandomRange(-spd/10/FRACUNIT, spd/10/FRACUNIT)*FRACUNIT
		end
		
		local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z - mo.scale*64, 320<<FRACBITS, 350<<FRACBITS, mo.target, mo.pso.potency, DMT_STRIKE, 8)
		PSO.lib.copyWeaponProperties(mo, hb)
		hb.pso.combobreaker = true	
		hb.pso.potency = mo.pso.potency
		
		local p = mo.target.player
		if p and p.valid and p == consoleplayer
		and R_PointToDist2(mo.x, mo.y, p.mo.x, p.mo.y) < 1024*FRACUNIT
			P_StartQuake(30*FRACUNIT, 10)
		end
		
		local soundm = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		soundm.state = S_INVISIBLE
		soundm.tics = TICRATE*2
		S_StartSound(soundm, P_RandomRange(0, 1) and sfx_rail1 or sfx_rail2)
		
		P_RemoveMobj(mo)
		return
	end 
	
	mo.extravalue1 = $-1
end, MT_PSO_SPEARSCYTHE)

PSO.info.photonArts["spear_pa_bandersnatch"] = {
		name = "Bandersnatch",
		desc = "Two quick swings followed by a quick backwards step to disengage from the enemy.",
		cost = 15,
		potency = 600,
		speed = 0,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		icon = "H_SPATK5",

		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local winfo = PSO.lib.itemInfo(PSO.lib.getPlayerActiveWeapon(p))
					
					-- set pos floorsprite
					
					-- for interpolation.
					local prevswordpos = {
						pos[2],
						pos[3],
						pos[4],
						pos[6],	-- rollangle
					}
					
					if timer == 1
						p.pso.avengerguard = 1
					end
					
					-- slash 1:
					if timer < TICRATE/4
						pos[7] = {
							use = true,
							o = {x = mo.x, y = mo.y, z = mo.z + mo.height/2},	-- the center of the slope is the center of the player.
							xy = p.drawangle + ANG1*90,	-- tilt sideways
							za = ANG1*10,				-- tilt by 30 degrees
						}
						
						local ang

						if mo.swordangle == nil
							mo.swordangle = 0
						end
						
						if timer < 3
							mo.swordangle = -ANG1*120
						elseif timer < TICRATE/2
							local nt = timer-3
							mo.swordangle = $ + max(0, ANG1*(40 - (nt-2)*5))
						end
						
						local ang = p.drawangle + mo.swordangle
						local x = 64 * cos(ang)
						local y = 64 * sin(ang)
						
						pos[2] = x
						pos[3] = y
						pos[4] = mo.height/2
						pos[5] = 0
						pos[6] = ang + ANG1*90
						
						-- damage box:
						if timer == TICRATE/6
							local x = mo.x + FixedMul(mo.scale, 100*cos(p.drawangle))
							local y = mo.y + FixedMul(mo.scale, 100*sin(p.drawangle))
							
							--(x, y, z, w, t, source, potency, dtype, duration)
							local hb = PSO.lib.spawnHurtBox(x, y, mo.z - mo.scale*64, 140<<FRACBITS, 140<<FRACBITS, mo, potency/2, DMT_STRIKE, 8)
							hb.angle = p.drawangle
							hb.pso.combobreaker = true
							hb.pso.hitlag = 4
							hb.pso.screenshake = {intensity = FRACUNIT*4, time=4}
							
							if winfo.sound
								S_StartSound(mo, winfo.sound)
							end
						end
					
					-- slash 2:
					elseif timer < TICRATE/2
						local t = timer-(TICRATE/4)
						pos[7] = {
							use = true,
							o = {x = mo.x, y = mo.y, z = mo.z + mo.height/2},	-- the center of the slope is the center of the player.
							xy = p.drawangle - ANG1*90,	-- tilt sideways
							za = ANG1*60,				-- tilt by 30 degrees
						}
						
						local ang

						if mo.swordangle == nil
							mo.swordangle = 0
						end
						
						if t < 3
							mo.swordangle = ANG1*120
						elseif t < TICRATE/2
							local nt = t-3
							mo.swordangle = $ - max(0, ANG1*(40 - (nt-2)*5))
						end
						
						local ang = p.drawangle + mo.swordangle
						local x = 64 * cos(ang)
						local y = 64 * sin(ang)
						
						pos[2] = x
						pos[3] = y
						pos[4] = mo.height/2
						pos[5] = 0
						pos[6] = ang + ANG1*90

						-- damage box:
						if t == TICRATE/6
							local x = mo.x + FixedMul(mo.scale, 100*cos(p.drawangle))
							local y = mo.y + FixedMul(mo.scale, 100*sin(p.drawangle))
							
							--(x, y, z, w, t, source, potency, dtype, duration)
							local hb = PSO.lib.spawnHurtBox(x, y, mo.z - mo.scale*64, 140<<FRACBITS, 140<<FRACBITS, mo, potency/2, DMT_STRIKE, 8)
							hb.pso.locks = 4
							hb.angle = p.drawangle
							hb.pso.combobreaker = true
							
							if winfo.sound
								S_StartSound(mo, winfo.sound)
							end
						end
						
						if t == TICRATE/4
							P_InstaThrust(mo, p.drawangle, -FRACUNIT*20)
							p.skidtime = 20
							mo.saveang = p.drawangle
							mo.z = $+1
							mo.momz = mo.scale*2
							mo.state = S_PLAY_FALL
						end
					else
						
						p.drawangle = mo.saveang
						
						if not P_IsObjectOnGround(mo)
						and leveltime & 1
							local g = P_SpawnGhostMobj(mo)
							g.colorized = true
						end
						
						if FixedHypot(mo.momx, mo.momy) < mo.scale*4
						and timer > TICRATE*3/4
							return true
						end
					end
					
					--P_SpawnMobj(mo.x+pos[2], mo.y+pos[3], mo.z+pos[4], MT_THOK)
					if timer > 1
						spear_trail(mo, pos, prevswordpos)
					end
					
				end,
}

PSO.info.photonArts["spear_pa_scythecleave"] = {
		name = "Scythe Cleave",
		desc = "A powerful slash utilizing the weapon's length. Hold to charge and perform an additional attack.",
		cost = 35,
		potency = 1200,
		speed = 30,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_SPATK6",

		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local winfo = PSO.lib.itemInfo(PSO.lib.getPlayerActiveWeapon(p))


					local prevspearpos = {
						pos[2],
						pos[3],
						pos[4],
						pos[6],	-- rollangle
					}
					
					-- set pos floorsprite
					
					if hold
					or timer < TICRATE/4
						
						if timer == 1
							S_StartSound(mo, sfx_s3k8e)
						end
						
						if timer%4 == 0
							if winfo.sound
								S_StartSound(mo, winfo.sound)
							end
						end
						
						-- charge the attack
						local t = min(timer, TICRATE*3/4 -1)
						local a = p.drawangle - ANG1*20 - t*ANG1*4
						local x = 32*cos(a)
						local y = 32*sin(a)
							
						pos[2] = x
						pos[3] = y
						pos[4] = mo.height - FRACUNIT*t
							
						pos[5] = 0
						pos[6] = p.drawangle + t*ANG1*65 + ANG1*180
							
						pos[7] = {
							use = true,
							o = {x = mo.x+x, y = mo.y+y, z = mo.z + pos[4]},
							xy = a+ ANG1*45,
							za = ANG1*30,
						}								
						
						p.pso.actiontimer = min(TICRATE*3/4, $)
						
						if timer == TICRATE*3/4 -1

							S_StopSound(mo)
							local g = P_SpawnGhostMobj(mo)
							g.destscale = mo.scale*4
							g.scalespeed = mo.scale/12
							S_StartSound(mo, sfx_s3ka8)
						end
						
					elseif timer < TICRATE*5
						
						S_StopSound(mo)
						mo.swordangle = 0
						
						if p.pso.actiontimer >= TICRATE*3/4	-- full charge
							mo.cleavemode = 1
							mo.superarmor = true
							
						else	-- partial charge
							mo.cleavemode = nil
						end	
						
						p.pso.actiontimer = TICRATE*5
						p.pso.avengerguard = 1
						--return true
					else
						
						local t = timer - TICRATE*5
						pos[7] = {
							use = true,
							o = {x = mo.x, y = mo.y, z = mo.z + mo.height/2},	-- the center of the slope is the center of the player.
							xy = p.drawangle - ANG1*90,	-- tilt sideways
							za = ANG1*50,				-- tilt by 30 degrees
						}
						
						local ang

						if mo.swordangle == nil
							mo.swordangle = 0
						end
						
						if not mo.cleavemode
							if t < 3
								mo.swordangle = -ANG1*160
							elseif t < TICRATE/2
								local nt = t-3
								mo.swordangle = $ + max(0, ANG1*(30 - (nt-2)*4))
							end

							if t == TICRATE/4
								local x = mo.x + FixedMul(mo.scale, 100*cos(p.drawangle))
								local y = mo.y + FixedMul(mo.scale, 100*sin(p.drawangle))
								
								--(x, y, z, w, t, source, potency, dtype, duration)
								local hb = PSO.lib.spawnHurtBox(x, y, mo.z - mo.scale*64, 252<<FRACBITS, 140<<FRACBITS, mo, potency/3, DMT_STRIKE, 8)
								hb.angle = p.drawangle
								hb.pso.combobreaker = true
								hb.pso.hitlag = 4
								hb.pso.screenshake = {intensity = FRACUNIT*4, time=4}
								
							end
							
						else

							if t < 3
								mo.swordangle = -ANG1*160
							elseif t < TICRATE/4
								mo.swordangle = $ + ANG1
							elseif t < TICRATE/2
								local nt = t-TICRATE/4
							
								mo.swordangle = $ + max(0, ANG1*(40 - (nt-2)*8))
							end

							if t == TICRATE/4
								local x = mo.x + FixedMul(mo.scale, 100*cos(p.drawangle))
								local y = mo.y + FixedMul(mo.scale, 100*sin(p.drawangle))
								
								--(x, y, z, w, t, source, potency, dtype, duration)
								local hb = PSO.lib.spawnHurtBox(x, y, mo.z - mo.scale*64, 152<<FRACBITS, 140<<FRACBITS, mo, potency/3, DMT_STRIKE, 8)
								hb.angle = p.drawangle
								hb.pso.combobreaker = true
								hb.pso.hitlag = 5
								hb.pso.screenshake = {intensity = FRACUNIT*10, time=4}
								
								local trail = PSO.lib.spawnMissile(mo, MT_PSO_SPEARSCYTHE, p.drawangle, 0, potency*2/3, DMT_STRIKE, 0, 0, 0)
								P_SetOrigin(trail, hb.x, hb.y, hb.z + mo.scale*32)
								trail.extravalue1 = TICRATE*3/2
								trail.scale = $*5/2
								trail.color = winfo.effectcolor or SKINCOLOR_CYAN

								if winfo.sound
									S_StartSound(mo, winfo.sound)
								end
								S_StartSound(mo, sfx_zoom)
								
								if p == consoleplayer
									P_StartQuake(FRACUNIT*40, 4)
								end
							end
						end
						
						local ang = p.drawangle + mo.swordangle
						local x = 64 * cos(ang)
						local y = 64 * sin(ang)
						
						pos[2] = x
						pos[3] = y
						pos[4] = mo.height/2
						pos[5] = 0
						pos[6] = ang + ANG1*90
						
						if t > TICRATE*3/4
							return true
						end
					end	

					if timer > 1
						spear_trail(mo, pos, prevspearpos)
					end
					
				end,
}

PSO.info.photonArts["spear_pa_raptorthrow"] = {
		name = "Raptor Throw",
		desc = "Throw the spear, stabbing enemies from a distance. Charge to increase damage and throw distance.",
		cost = 35,
		potency = 1050,
		speed = 0,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		icon = "H_SPATK4",

		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local winfo = PSO.lib.itemInfo(PSO.lib.getPlayerActiveWeapon(p))


					local prevspearpos = {
						pos[2],
						pos[3],
						pos[4],
						pos[6],	-- rollangle
					}
						
					if timer < TICRATE/2
					or hold
						pos[7] = {use = false}
						
						if timer == 1
							mo.throwcharge = nil
						end
						
						local ang = p.drawangle - ANG1*90
						local x = 16*cos(ang)
						local y = 16*sin(ang)

						pos[2] = x
						pos[3] = y
						pos[4] = mo.height/2
						pos[5] = ANG1*10
						pos[6] = min(TICRATE, timer)*ANG1*63	

						if timer%5 == 0
						and timer < TICRATE
							if winfo.sound
								S_StartSound(mo, winfo.sound)
							end
						end
						
						if timer == TICRATE
							S_StopSound(mo)
							local g = P_SpawnGhostMobj(mo)
							g.destscale = mo.scale*4
							g.scalespeed = mo.scale/12
							S_StartSound(mo, sfx_s3ka8)
							mo.throwcharge = true
						end
						
						-- cap actiontimer until we release...
						p.pso.actiontimer = min(TICRATE+2, $)
						
					else
						
						p.pso.actiontimer = max($, TICRATE*2)	-- set the timer to 2s if it's lower
						local t = p.pso.actiontimer - (TICRATE*2)
						
						if t == 1
							p.pso.avengerguard = 1
						end
						
						local ang = p.drawangle - ANG1*90
						local x = 16*cos(ang)
						local y = 16*sin(ang)
						
						-- swing backwards
						local dist = -64*sin(t*ANG1*22)
						
						x = $ + FixedMul(dist, cos(p.drawangle))
						y = $ + FixedMul(dist, sin(p.drawangle))
						
						pos[2] = x
						pos[3] = y
						--pos[4] = mo.height/2
						pos[6] = 0
						
						if t == 8
							
							-- hide the spear
							pos[4] = -9999*FRACUNIT
							local spear = P_SpawnMobj(mo.x + pos[2], mo.y + pos[3], mo.z+mo.height/2, MT_THOK)
							spear.sprite = pos[1].sprite
							spear.frame = pos[1].frame
							spear.color = pos[1].color
							spear.angle = p.drawangle
							spear.target = mo
							
							mo.spear = spear
							

							
							--P_TeleportMove(mo, spear.x, spear.y, spear.z)
							
							local spd = mo.throwcharge and 75 or 40
							local tics = TICRATE/2
							local num = mo.throwcharge and 2 or 1
							local pot = mo.throwcharge and potency or potency/2

							local hb = PSO.lib.spawnHurtBox(spear.x, spear.y, spear.z - spear.scale*64, 100<<FRACBITS, 140<<FRACBITS, mo, pot, DMT_STRIKE, tics, RT_KNOCKBACK)
							hb.pso.reactionangle = p.drawangle
							hb.pso.reactionstrength = 50
							hb.pso.combobreaker = true
							mo.hitbox = hb
							
							spear.tics = tics
							
							P_InstaThrust(spear, spear.angle, spd*FRACUNIT)
							local slope = PSO.lib.aimMissile(spear, p.pso.target)
							
							if slope
								spear.rollangle = slope
							end
							
							if mo.throwcharge
								P_InstaThrust(mo, p.drawangle, FRACUNIT*6)
								mo.state = S_PLAY_WALK
								
								if p == consoleplayer
									P_StartQuake(FRACUNIT*32, 8)
								end
							end
							

							for i = 1, num
									
								local efc = P_SpawnMobj(spear.x, spear.y, spear.z, MT_THOK)
								efc.color = winfo.effectcolor or SKINCOLOR_CYAN
								efc.sprite = SPR_ECRC
								efc.frame = C|TR_TRANS40|FF_PAPERSPRITE
								efc.angle = spear.angle + ANG1*90
								efc.scale = mo.scale
								efc.destscale = (mo.scale*10) *i
								efc.scalespeed = (mo.scale/4) *i
								efc.fuse = 20
								efc.tics = 20
							end	
						
						elseif t < 8 + TICRATE/2
						
							local spear = mo.spear
							
							if mo.spear and mo.spear.valid
							and mo.hitbox and mo.hitbox.valid
								
								P_MoveOrigin(mo.hitbox, mo.spear.x, mo.spear.y, mo.spear.z)
								
								local i = mo.throwcharge and 2 or 1
								if leveltime%5 == 0
									local efc = P_SpawnMobj(mo.spear.x, mo.spear.y, mo.spear.z + FRACUNIT*8, MT_THOK)
									efc.color = winfo.effectcolor or SKINCOLOR_CYAN
									efc.sprite = SPR_ECRC
									efc.frame = C|TR_TRANS40|FF_PAPERSPRITE
									efc.angle = spear.angle + ANG1*90
									efc.scale = mo.scale
									efc.destscale = (mo.scale*10) *i
									efc.scalespeed = (mo.scale/4) *i
									efc.fuse = 20
									efc.tics = 20								
								end
							end
						
							--return true
						elseif t == TICRATE
							return true
						end
						
					end


					if timer > 1
						spear_trail(mo, pos, prevspearpos)
					end
					
				end,
}


PSO.info.photonArts["spear_pa_millionskewer"] = {
		name = "Million Skewer",
		desc = "Hold to keep stabbing while expanding PP. Holding long enough casts an extra attack on release.",
		cost = 10,
		potency = 700,
		speed = 0,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		icon = "H_SPATK3",

		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local winfo = PSO.lib.itemInfo(PSO.lib.getPlayerActiveWeapon(p))


					local prevspearpos = {
						pos[2],
						pos[3],
						pos[4],
						pos[6],	-- rollangle
					}
					
					if timer == 1
						p.pso.avengerguard = 1
					end
					
					if hold 
					or timer < TICRATE/2
					
						mo.releasehold = nil
						
						if timer%5 == 0
							p.pso.pp = $-1
							
							if not p.pso.pp
								return true
							end
						end
						
						if timer < TICRATE/4
							pos[7] = {use = false}
							
							local ang = p.drawangle - ANG1*90
							local x = 16*cos(ang)
							local y = 16*sin(ang)

							pos[2] = x
							pos[3] = y
							pos[4] = mo.height/2
							pos[5] = ANG1*10
							pos[6] = timer*ANG1*80	

							if timer%3 == 0
								if winfo.sound
									S_StartSound(mo, winfo.sound)
								end
							end
							
						else
							local t = timer - TICRATE/4
							
							local spd = TICRATE/4
							if t > TICRATE*3/2
								spd = TICRATE/6
								mo.releasehold = timer
							end

								
							if timer == TICRATE*3/2
								S_StopSound(mo)
								local g = P_SpawnGhostMobj(mo)
								g.destscale = mo.scale*4
								g.scalespeed = mo.scale/12
								S_StartSound(mo, sfx_s3ka8)
							end
							
							local thrust = t % (spd)
							
							if thrust == 0
								mo.thrust_a = p.drawangle + P_RandomRange(-20, 20)*ANG1
								mo.thrust_a2 = P_RandomRange(0, 20)*ANG1
								
								if winfo.sound
									S_StartSound(mo, winfo.sound)
								end
								
								-- hurtbox
								
								local fx = 160*cos(p.drawangle)
								local fy = 160*sin(p.drawangle)
								local pot = potency/6
								
								if t > TICRATE*3/2
									pot = potency/7
								end
								
								local hb = PSO.lib.spawnHurtBox(mo.x+fx, mo.y+fy, mo.z - mo.scale*64, 230<<FRACBITS, 140<<FRACBITS, mo, pot, DMT_STRIKE, 3)
								hb.pso.reactionangle = p.drawangle
								hb.pso.reactionstrength = 50
								hb.pso.combobreaker = true
								hb.pso.hitlag = 2
								hb.pso.screenshake = {intensity = FRACUNIT*4, time=2}
								
								-- spawn visual...
								if t > TICRATE*3/2
									local ang = mo.thrust_a - ANG1*90
									local x = 16*cos(ang)
									local y = 16*sin(ang)
									local z = mo.height
									
									-- further dist for the skewer:
									x = $ + FixedMul( cos(mo.thrust_a2), (4 * cos(mo.thrust_a)*128))
									y = $ + FixedMul( cos(mo.thrust_a2), (4 * sin(mo.thrust_a)*128))
									z = $ + FixedMul( sin(mo.thrust_a2), (4 * 128*FRACUNIT))								

									local efc = P_SpawnMobj(mo.x+x, mo.y+y, mo.z+z, MT_THOK)
									efc.color = winfo.effectcolor or SKINCOLOR_CYAN
									efc.sprite = SPR_ECRC
									efc.frame = C|TR_TRANS40|FF_PAPERSPRITE
									efc.angle = mo.thrust_a + ANG1*90
									efc.scale = mo.scale
									efc.destscale = mo.scale*3
									efc.scalespeed = mo.scale/8
									efc.fuse = 10
									efc.tics = 10
								
								end
								
							end

							pos[7] = {use = false}
							
							local ang = mo.thrust_a - ANG1*90
							local x = 16*cos(ang)
							local y = 16*sin(ang)
							local z = mo.height
							
							-- further dist for the skewer:
							x = $ + FixedMul( cos(mo.thrust_a2), (min(4, thrust) * cos(mo.thrust_a)*64))
							y = $ + FixedMul( cos(mo.thrust_a2), (min(4, thrust) * sin(mo.thrust_a)*64))
							z = $ + FixedMul( sin(mo.thrust_a2), (min(4, thrust) * 64*FRACUNIT))
							
							pos[2] = x
							pos[3] = y
							pos[4] = z
							pos[5] = mo.thrust_a - p.drawangle
							pos[6] = mo.thrust_a2							
						end
					
					elseif mo.releasehold
					
						local t = timer - mo.releasehold

						if t == 1
							p.pso.avengerguard = 1
						end
						
						if t < TICRATE/4
							
							if t%3 == 0
								if winfo.sound
									S_StartSound(mo, winfo.sound)
								end
							end
							
							local t = min(t, TICRATE/4)
							local a = p.drawangle - ANG1*140 + t*ANG1*5
							local x = 64*cos(a)
							local y = 64*sin(a)
							
							pos[2] = x
							pos[3] = y
							pos[4] = mo.height - FRACUNIT*t
							
							pos[5] = 0
							pos[6] = t*ANG1*65
							
							pos[7] = {
								use = true,
								o = {x = mo.x+x, y = mo.y+y, z = mo.z + pos[4]},
								xy = p.drawangle + a+ ANG1*45,
								za = ANG1*30,
							}				
								
						else
								
							local t2 = t - (TICRATE/4)
							local a = p.drawangle + min(t2, 3)*ANG1*30
							
							local ang = p.drawangle - ANG1*90
							local x = 32*cos(ang)
							local y = 32*sin(ang)
							
							-- further dist for the skewer:
							x = x + (min(4, t2)*cos(p.drawangle)*24)
							y = y + (min(4, t2)*sin(p.drawangle)*24)
							
							pos[2] = x
							pos[3] = y
							pos[4] = mo.height/2
							pos[5] = 0
							pos[6] = a
							
							pos[7].o = {x = mo.x+pos[2], y = mo.y+pos[3], z = mo.z+pos[4]}
							pos[7].xy = p.drawangle + ANG1*90

							
							if t2 == 1
								if winfo.sound
									S_StartSound(mo, winfo.sound)
								end
								
								local fx = x + cos(p.drawangle)*220
								local fy = y + sin(p.drawangle)*220
								
								local hb = PSO.lib.spawnHurtBox(mo.x+fx, mo.y+fy, mo.z - mo.scale*64, 230<<FRACBITS, 140<<FRACBITS, mo, potency, DMT_STRIKE, 8, RT_BLOWDOWN)
								hb.pso.reactionangle = p.drawangle
								hb.pso.reactionstrength = 10
								hb.pso.combobreaker = true
								hb.pso.hitlag = 6
								hb.pso.screenshake = {intensity = FRACUNIT*9, time=5}
								if not (p.cmd.forwardmove > 0)
									hb.pso.momentummult = 0
								end

								
								fx = x + 16*cos(p.drawangle)*24
								fy = y + 16*sin(p.drawangle)*24
								
								local efc = P_SpawnMobj(mo.x+fx, mo.y+fy, hb.z+mo.height/2, MT_THOK)
								efc.color = winfo.effectcolor or SKINCOLOR_CYAN
								efc.sprite = SPR_ECRC
								efc.frame = C|TR_TRANS40|FF_PAPERSPRITE
								efc.angle = p.drawangle + ANG1*90
								efc.scale = mo.scale
								efc.destscale = mo.scale*10
								efc.scalespeed = mo.scale/4
								efc.fuse = 20
								efc.tics = 20

								if p == consoleplayer
									P_StartQuake(FRACUNIT*32, 8)
								end								
								
								if P_IsObjectOnGround(mo)
									P_InstaThrust(mo, p.drawangle, FRACUNIT*48)
									mo.state = S_PLAY_WALK			
								end
							end
							
							if not P_TryMove(mo, mo.x+mo.momx, mo.y+mo.momy, false)
								P_InstaThrust(mo, 0, 0)
							end
							
							mo.momx = $*90/100
							mo.momy = $*90/100
							
							if t2 >= TICRATE*3/4
								mo.releasehold = nil
								return true
							end
						end						
						
					
					else
						return true
					end

					if timer > 1
						spear_trail(mo, pos, prevspearpos)
					end
					
				end,
}

PSO.info.photonArts["spear_pa_assaultbuster"] = {
		name = "Assault Buster",
		desc = "Charge forward with your spear and pierce through enemies. Hold to charge to increase the damage and range. Hold forward to keep going after hitting an enemy instead of stopping.",
		cost = 35,
		potency = 900,
		speed = 30,		-- speed % multiplier.
						-- 100: full speed
						-- 0: stand still
		
		icon = "H_SPATK2",

		use = 	function(p, timer, hold, input, combo)
					
					local mo = p.mo
					-- rifle is always weapon slot 1
					local pos = p.pso.weaponpos[1]
					local potency = PSO.info.photonArts[p.pso.action].potency
					local winfo = PSO.lib.itemInfo(PSO.lib.getPlayerActiveWeapon(p))


					local prevspearpos = {
						pos[2],
						pos[3],
						pos[4],
						pos[6],	-- rollangle
					}
					
					-- set pos floorsprite
					
					if hold
					or timer < TICRATE/4
						
						if timer == 1
							S_StartSound(mo, sfx_s3k8e)
							mo.releasehold = 0
						end
						
						if timer%4 == 0
							if winfo.sound
								S_StartSound(mo, winfo.sound)
							end
						end
						
						-- charge the attack
						local t = min(timer, TICRATE*3/4 -1)
						local a = p.drawangle - ANG1*20 - t*ANG1*4
						local x = 32*cos(a)
						local y = 32*sin(a)
							
						pos[2] = x
						pos[3] = y
						pos[4] = mo.height - FRACUNIT*t
							
						pos[5] = 0
						pos[6] = p.drawangle + t*ANG1*65 + ANG1*180
							
						pos[7] = {
							use = true,
							o = {x = mo.x+x, y = mo.y+y, z = mo.z + pos[4]},
							xy = a+ ANG1*45,
							za = ANG1*30,
						}								
						
						p.pso.actiontimer = min(TICRATE*3/4, $)
						
						if timer == TICRATE*3/4 -1

							S_StopSound(mo)
							local g = P_SpawnGhostMobj(mo)
							g.destscale = mo.scale*4
							g.scalespeed = mo.scale/12
							S_StartSound(mo, sfx_s3ka8)
						end
						
					elseif timer < TICRATE*5
						
						S_StopSound(mo)
						
						pos[6] = p.drawangle - ANG1*50
						local reactiontype = RT_KNOCKBACK
						local strength = 60
						local pot = potency
						
						if mo.releasehold < 7
							
							if not mo.releasehold
								p.pso.avengerguard = 1
							end
							
							local t = min(timer, TICRATE*3/4 -1)
							local a = p.drawangle - ANG1*20 - t*ANG1*4
							local x = 32*cos(a)
							local y = 32*sin(a)
								
							pos[2] = x
							pos[3] = y
							pos[4] = mo.height - FRACUNIT*t
								
							pos[5] = 0
							pos[6] = $ - ANG1*96*mo.releasehold
								
							pos[7] = {
								use = true,
								o = {x = mo.x+x, y = mo.y+y, z = mo.z + pos[4]},
								xy = a+ ANG1*45,
								za = ANG1*30,
							}	
						
							mo.releasehold = $+1
							spear_trail(mo, pos, prevspearpos)
							
							return
						end
						
						local hitlag = 3
						local str = FRACUNIT*4
						
						if p.pso.actiontimer >= TICRATE*3/4	-- full charge
							
							--P_InstaThrust(mo, p.drawangle, FRACUNIT*70)
							mo.savethrust = 70
							reactiontype = RT_BLOWDOWN
							strength = 7
							hitlag = 6
							str = FRACUNIT*8
							
						else	-- partial charge
							
							--P_InstaThrust(mo, p.drawangle, FRACUNIT*40)
							mo.savethrust = 40
							pot = $/2
						end

						local ang = p.drawangle - ANG1*90
						local x = 16*cos(ang)
						local y = 16*sin(ang)
						
						-- further dist for the skewer:
						x = x + 128*cos(p.drawangle)
						y = y + 128*sin(p.drawangle)
						
						local hb = PSO.lib.spawnHurtBox(mo.x+x, mo.y+y, mo.z - mo.scale*64, 255<<FRACBITS, 200<<FRACBITS, mo, pot, DMT_STRIKE, TICRATE, reactiontype)
						hb.pso.combobreaker = true
						hb.pso.reactionangle = p.drawangle
						hb.pso.reactionstrength = strength
						hb.pso.hitlag = hitlag
						hb.pso.screenshake = {intensity = str, time=4}			
						/*hb.sprite = SPR_THOK
						hb.frame = A
						hb.color = SKINCOLOR_RED*/
						
						mo.hb = hb
						
						p.pso.actiontimer = TICRATE*5
						
						S_StartSound(mo, sfx_s3k81)
						if winfo.sound
							S_StartSound(mo, winfo.sound)
						end						
						--return true
					else
						mo.momx = $*95/100
						mo.momy = $*95/100

						pos[7] = {use = false}
						
						local ang = p.drawangle - ANG1*90
						local x = 16*cos(ang)
						local y = 16*sin(ang)
						
						-- further dist for the skewer:
						x = x + 128*cos(p.drawangle)
						y = y + 128*sin(p.drawangle)
						
						if timer == TICRATE*5 +2
							local fx = x + 16*cos(p.drawangle)*24
							local fy = y + 16*sin(p.drawangle)*24
							
							local num = (mo.savethrust > 60 and 2 or 1)
							--print(num)
							
							for i = 1, num
								
								local efc = P_SpawnMobj(mo.x+fx, mo.y+fy, mo.z+mo.height/2, MT_THOK)
								efc.color = winfo.effectcolor or SKINCOLOR_CYAN
								efc.sprite = SPR_ECRC
								efc.frame = C|TR_TRANS40|FF_PAPERSPRITE
								efc.angle = p.drawangle + ANG1*90
								efc.scale = mo.scale
								efc.destscale = (mo.scale*10) *i
								efc.scalespeed = (mo.scale/4) *i
								efc.fuse = 20
								efc.tics = 20
							end

							if p == displayplayer
							and num > 1
								P_StartQuake(FRACUNIT*32, 8)
							end	
							
							mo.speargrounded = P_IsObjectOnGround(mo)
							P_InstaThrust(mo, p.drawangle, mo.savethrust*FRACUNIT)
							p.drawangle = mo.angle
							mo.state = S_PLAY_WALK							
						end

						pos[2] = x
						pos[3] = y
						pos[4] = mo.height/2
						pos[5] = ANG1*10
						pos[6] = 0
						
						-- stop if we're about to fall off
						-- but only if we initiated the pa on the ground
						if mo.speargrounded
						and not P_TryMove(mo, mo.x+mo.momx, mo.y+mo.momy, false)
							P_InstaThrust(mo, 0, 0)
						end	
						
						if FixedHypot(mo.momx, mo.momy) < FRACUNIT*5
						and p.pso.actiontimer > TICRATE*5 + TICRATE/2
							if mo.hb and mo.hb.valid
								P_RemoveMobj(mo.hb)
								mo.hb = nil
							end
							return true
						end
						
						if mo.hb and mo.hb.valid
							if p.cmd.forwardmove > 0
								mo.hb.pso.momentummult = nil
							else
								mo.hb.pso.momentummult = 0
							end
							
							P_SetOrigin(mo.hb, mo.hb.x+mo.momx*2, mo.hb.y+mo.momy*2, mo.hb.z+mo.momz*2)
						end
					end	

					if timer > 1
						spear_trail(mo, pos, prevspearpos)
					end
					
				end,
}