
freeslot("MT_DIGA")
mobjinfo[MT_DIGA] = {
	doomednum = -1,
	spawnstate = S_ROCKCRUMBLEA,
	spawnhealth = 1000,
	radius = 32<<FRACBITS,
	height = 128<<FRACBITS,
	flags = MF_MISSILE,
}

addHook("MobjThinker", function(mo)
	if leveltime & 1
		P_SpawnGhostMobj(mo)
	end
end, MT_DIGA)



PSO.info.photonArts["diga"] = {
		-- PA is a technique
		technique = true,
		name = "Diga",
		desc = "Earth Technique. Sends a slow, powerful rock flying towards the enemy.",
		cost = 16,
		potency = 560,
		icon = "H_TDIGA",
		casttime = TICRATE*2/3,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_rocks2,
		
		--aim = true,				-- tech can be aimed up/down
		reaction = RT_BLOWDOWN,
		element = ELEM_EARTH,
		object = MT_DIGA,		-- object to spawn
		speed = 20,				-- projectile speed.
		actionspeed = 40,
		spawndist = 32,			-- distance to spawn from the player if spawntarget is false
		spawntarget = false,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.
		spawnheight = 0,		-- height % to spawn the object at
		
		spawnfunc = function(msl)
			
			local mo = msl.target
			msl.color = SKINCOLOR_ORANGE
			msl.colorized = true
			msl.scale = $*3/2
			
			if mo and mo.valid and mo.player
				
				local p = mo.player
				if p.pso.target and p.pso.target.valid
					PSO.lib.lobMissile(msl, p.pso.target)
				else
					msl.momz = FRACUNIT*12
				end
			end
			
		end,
		
}


PSO.info.photonArts["gidiga"] = {
		-- PA is a technique
		technique = true,
		name = "Gidiga",
		desc = "Earth Technique. Creates a localized earthquake around the user, dealing high damage and sending nearby enemies flying.",
		cost = 27,
		potency = 450,
		icon = "H_TGDIGA",
		casttime = TICRATE*3/4,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_rocks4,
		
		aim = true,				-- tech can be aimed up/down
		reaction = RT_BLOWDOWN,
		element = ELEM_EARTH,
		object = MT_THOK,
		speed = 20,				-- projectile speed.
		actionspeed = 10,
		spawndist = 0,			-- distance to spawn from the player if spawntarget is false
		spawntarget = false,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.
		spawnheight = 0,		-- height % to spawn the object at
		
		spawnfunc = function(msl)
			
			local mo = msl.target
			local range = 512
			
			if mo and mo.valid and mo.player
				
				if mo.player == consoleplayer
					P_StartQuake(FRACUNIT*30, 8)
				end
				
				for i = 1, 64
						
					local x = mo.x + P_RandomRange(-range, range)*mo.scale
					local y = mo.y + P_RandomRange(-range, range)*mo.scale
					
					local t = P_SpawnMobj(x, y, mo.z, MT_THOK)
					t.state = S_ROCKCRUMBLEA
					t.fuse = P_RandomRange(10, 20)
					t.scale = mo.scale/3
					
					t.momz = P_RandomRange(2, 16)*mo.scale
					t.momx = P_RandomRange(-2, 2)*mo.scale
					t.momy = P_RandomRange(-2, 2)*mo.scale
					
					t.color = SKINCOLOR_ORANGE
					t.colorized = true
				end
				
				
				local x1 = mo.x - range*FRACUNIT
				local x2 = mo.x + range*FRACUNIT
				local y1 = mo.y - range*FRACUNIT
				local y2 = mo.y + range*FRACUNIT
				
				local fn = function(mo, found)
					
					if found and found.valid
					and R_PointToDist2(mo.x, mo.y, found.x, found.y) < range*FRACUNIT
					and found.flags & MF_SHOOTABLE
						
						local hb = PSO.lib.spawnHurtBox(found.x, found.y, found.z, 32<<FRACBITS, 180<<FRACBITS, msl.target, msl.pso.potency, DMT_TECHNIQUE, 4, RT_BLOWDOWN)
						hb.pso.combobreaker = msl.pso.combobreaker
						hb.pso.angle = R_PointToAngle2(mo.x, mo.y, found.x, found.y)
						hb.pso.element = ELEM_EARTH
						hb.pso.potency = msl.pso.potency
					end
				end
				
				searchBlockmap("objects", fn, mo, x1, x2, y1, y2)
			end
			
			P_RemoveMobj(msl)
		end,
		
}


PSO.info.photonArts["radiga"] = {
		-- PA is a technique
		technique = true,
		name = "Radiga",
		desc = "Earth Technique. Lifts rocks from below the target to pierce them, dealing high damage and sending them flying.",
		cost = 19,
		potency = 500,
		icon = "H_TRDIGA",
		casttime = TICRATE*3/4,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_rocks3,
		
		aim = true,				-- tech can be aimed up/down
		reaction = RT_BLOWDOWN,
		element = ELEM_EARTH,
		object = MT_THOK,
		actionspeed = 30,
		spawndist = 96,			-- distance to spawn from the player if spawntarget is false
		spawntarget = true,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.
		spawnheight = 50,		-- height % to spawn the object at
		
		spawnfunc = function(msl, radius)
			
			local mo = msl.target
			
			for i = 1, 8
				
				local x = msl.x + P_RandomRange(-48, 48)*msl.scale
				local y = msl.y + P_RandomRange(-48, 48)*msl.scale
				
				local r = P_SpawnMobj(x, y, msl.z, MT_THOK)
				r.state = S_ROCKCRUMBLEA
				r.fuse = P_RandomRange(20, 35)
				r.scale = 1
				r.destscale = msl.scale
				r.scalespeed = msl.scale / P_RandomRange(3, 8)
					
				r.momz = P_RandomRange(4, 16)*FRACUNIT
				
				r.spriteyscale = FRACUNIT*2 + P_RandomRange(0, FRACUNIT-7)
				
				r.color = SKINCOLOR_ORANGE
				r.colorized = true
			end

			local hb = PSO.lib.spawnHurtBox(msl.x, msl.y, msl.z, (radius or 32)<<FRACBITS, 180<<FRACBITS, msl.target, msl.pso.potency, DMT_TECHNIQUE, 4, RT_BLOWDOWN)
			hb.pso.combobreaker = msl.pso.combobreaker
			hb.pso.element = ELEM_EARTH
			hb.angle = R_PointToAngle2(mo.x, mo.y, msl.x, msl.y)
			hb.pso.potency = msl.pso.potency
			
			P_RemoveMobj(msl)
		end,
		
}

freeslot("MT_PSO_NOSDIGA")
mobjinfo[MT_PSO_NOSDIGA] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	deathstate = S_NULL,
	xdeathstate = S_NULL,
	radius = 48<<FRACBITS,
	height = 48<<FRACBITS,
	flags = MF_MISSILE,
}

addHook("MobjCollide", function(mo1, mo2) return false end, MT_PSO_NOSDIGA)
addHook("MobjMoveCollide", function(mo1, mo2) return false end, MT_PSO_NOSDIGA)

addHook("MobjThinker", function(mo)
	
	mo.friction = FRACUNIT	-- no slowdown
	if mo.target and mo.target.valid 
		if not mo.hurtbox
			mo.hurtbox = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 96<<FRACBITS, 64<<FRACBITS, mo.target, mo.pso.potency, DMT_TECHNIQUE, 10, RT_BLOWDOWN)
			mo.hurtbox.pso.combobreaker = mo.pso.combobreaker
			mo.hurtbox.pso.element = mo.pso.element
			mo.hurtbox.angle = mo.angle
			mo.hurtbox.pso.potency = mo.pso.potency
			mo.hurtbox.pso.pierceref = mo.hurtbox
		else
			mo.hurtbox.fuse = 3
			mo.hurtbox.tics = 10
			P_MoveOrigin(mo.hurtbox, mo.x, mo.y, mo.z)
		end
	end
	
	for i = 1, 4
				
		local x = mo.x + P_RandomRange(-48, 48)*mo.scale
		local y = mo.y + P_RandomRange(-48, 48)*mo.scale
				
		local r = P_SpawnMobj(x, y, mo.z, MT_THOK)
		r.flags = $ & ~MF_NOGRAVITY
		r.state = S_ROCKCRUMBLEA
		r.fuse = P_RandomRange(15, 23)
		r.scale = FRACUNIT
					
		r.momz = P_RandomRange(3, 7)*mo.scale
		
		r.spriteyscale = FRACUNIT*2 + P_RandomRange(0, FRACUNIT-7)
		
		r.spritexscale = $/2
		r.spriteyscale = $/2
		
		r.color = SKINCOLOR_ORANGE
		r.colorized = true
	end
end, MT_PSO_NOSDIGA)

PSO.info.photonArts["nosdiga"] = {
		-- PA is a technique
		technique = true,
		name = "Nosdiga",
		desc = "Earth Technique. Sends a shockwave forward that sends enemies flying in its wake.",
		cost = 16,
		potency = 510,
		icon = "H_TNDIGA",
		casttime = TICRATE*3/4,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_rocks2,
		
		--aim = true,				-- tech can be aimed up/down
		reaction = RT_BLOWDOWN,
		element = ELEM_EARTH,
		object = MT_PSO_NOSDIGA,		-- object to spawn
		speed = 20,				-- projectile speed.
		actionspeed = 40,
		spawndist = 32,			-- distance to spawn from the player if spawntarget is false
		spawntarget = false,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.
		spawnheight = 0,		-- height % to spawn the object at
		
}

freeslot("MT_PSO_DIVERSE")
mobjinfo[MT_PSO_DIVERSE] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	deathstate = S_NULL,
	xdeathstate = S_NULL,
	radius = 16<<FRACBITS,
	height = 16<<FRACBITS,
	flags = MF_NOGRAVITY,
}

local diverse_range = 600
local diverse_duration = TICRATE*9
local diverse_tick = TICRATE*2
local diverse_heal = 10

addHook("MobjThinker", function(mo)

	local mot = mo.target
	local mop = mot.player
			
	if not mop return end
			
	local range = diverse_range
	if PSO.lib.hasSkill(mop, "fo_territoryburst")
		range = $*2
	end
	
	local x1 = mo.x - range*FRACUNIT
	local x2 = mo.x + range*FRACUNIT
	local y1 = mo.y - range*FRACUNIT
	local y2 = mo.y + range*FRACUNIT
	
	mo.extravalue1 = $+1
	
	local targets = {}
	
	if mo.extravalue1 % diverse_tick == 0
		
		local n_enemies = 0
		
		local function diverseCountEnemies(ref, found)
			if ref != found
			and found and found.valid	
			and found.flags & MF_ENEMY
			and PSO.info.enemies[found.type]
			and abs(mo.z - found.z) < 512*FRACUNIT
			and R_PointToDist2(mo.x, mo.y, found.x, found.y) <= range*mo.scale
			and P_CheckSight(mo, found)	
			
				n_enemies = $+1
			end
		end
						
		searchBlockmap("objects", diverseCountEnemies, mo, x1, x2, y1, y2)	
		
		-- now check for every player in range
		if n_enemies
			for p in players.iterate do
				if p and p.mo and p.mo.valid
					
					if R_PointToDist2(p.mo.x, p.mo.y, mo.x, mo.y) < range*mo.scale
						PSO.lib.healPP(p.mo, p.mo, diverse_heal*n_enemies)
						
					end
				end
			end
		end
	end
	
	-- visuals
	for i = 1, 4
		local a = P_RandomRange(0, 359)*ANG1
		local xp = mo.x + range*cos(a)
		local yp = mo.y + range*sin(a)
		
		local s = P_SpawnMobj(xp, yp, mo.z, MT_THOK)
		
		if i & 1
			
			s.colorized = true
			s.color = SKINCOLOR_LAVENDER
			s.tics = 14
			s.frame = A
			s.destscale = 1
		else
			
			s.state = S_ROCKCRUMBLEA
			s.colorized = true
			s.color = SKINCOLOR_ORANGE
			s.scale = $*2/3
			s.fuse = 14
			s.tics = 14
		end
			
		s.angle = a + ANG1*180
		P_InstaThrust(s, s.angle, (range/16)*FRACUNIT)		
	end
	
	if leveltime%16 == 0
		S_StartSound(mo, sfx_s3k72)
	end
	
	if mo.extravalue1 > diverse_duration
		P_RemoveMobj(mo)
	end
	
end, MT_PSO_DIVERSE)

PSO.info.photonArts["diverse"] = {
		-- PA is a technique
		technique = true,
		name = "Diverse",
		desc = "Earth Technique. Periodically regenerates 10PP to every player within the field for each enemy also within the field.",
		cost = 4,
		potency = diverse_heal,
		icon = "H_TDVERS",
		casttime = TICRATE*3/4,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_rocks2,
		
		element = ELEM_EARTH,
		object = MT_PSO_DIVERSE,		-- object to spawn
		speed = 0,				-- projectile speed.
		actionspeed = 40,
		spawndist = 0,

		spawnfunc = function(msl)
			-- make sure razonde cannot have more than 1 instance per player
			local mo = msl.target
			
			if mo.lastdiverse
			and mo.lastdiverse.valid
				P_RemoveMobj(mo.lastdiverse)	-- don't stack!!
			end
			
			mo.lastdiverse = msl
			msl.momx = 0
			msl.momy = 0
			msl.momz = 0
		end,
		
		
}