freeslot("SPR_LZI1", "S_ZONDE")
states[S_ZONDE] = {SPR_LZI1, FF_ANIMATE, (E+1)*3, nil, E, 3}

freeslot("MT_RAZONDE")
mobjinfo[MT_RAZONDE] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	deathstate = S_NULL,
	xdeathstate = S_NULL,
	radius = 1, 
	height = 1,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPTHING,
}

local razonde_dist = 512
addHook("MobjThinker", function(mo)
	
	mo.momz = 0
	P_InstaThrust(mo, 0, 0)
	
	if not mo.extravalue1
		mo.tics = TICRATE*3
		mo.fuse = mo.tics
		mo.extravalue1 = 1
	end
	
	if (mo.tics % ((TICRATE*3)/4)) == 0
		
		for i = 1, 16
			
			local ang = (360/16)*ANG1*i
			local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
			s.color = SKINCOLOR_GOLD
			s.colorized = true
			
			P_InstaThrust(s, ang, FRACUNIT*48)
		end
		
		local x1 = mo.x - razonde_dist*FRACUNIT
		local x2 = mo.x + razonde_dist*FRACUNIT
		local y1 = mo.y - razonde_dist*FRACUNIT
		local y2 = mo.y + razonde_dist*FRACUNIT
		
		local search = function(mo, found)
			if found and found.flags & MF_SHOOTABLE and found.type ~= MT_PLAYER
			and R_PointToDist2(mo.x, mo.y, found.x, found.y) < razonde_dist*FRACUNIT
			and abs(mo.z - found.z) < 255*FRACUNIT
				
				local hb = PSO.lib.spawnHurtBox(found.x, found.y, found.z, 96<<FRACBITS, 64<<FRACBITS, mo.target, mo.pso.potency/4, DMT_TECHNIQUE, 2)
				PSO.lib.copyWeaponProperties(mo, hb)
				hb.pso.potency = mo.pso.potency/4	-- potency was already applied.
				hb.pso.combobreaker = mo.pso.combobreaker
				hb.pso.element = mo.pso.element
				hb.pso.locks = 1
			end
		end
		
		S_StartSound(mo, sfx_zonde)
		searchBlockmap("objects", search, mo, x1, x2, y1, y2)
	end
	
	for i = 1, 2
		local ang = ANG1*mo.tics*7 + (ANG1*180*i)
		local x = mo.x + razonde_dist*cos(ang)
		local y = mo.y + razonde_dist*sin(ang)
		
		local s = P_SpawnMobj(x + P_RandomRange(-32, 32)*mo.scale, y + P_RandomRange(-32, 32)*mo.scale, mo.z + P_RandomRange(-32, 32)*mo.scale, MT_SUPERSPARK)
		s.color = SKINCOLOR_GOLD
		s.colorized = true
	end
	
end, MT_RAZONDE)

PSO.info.photonArts["zonde"] = {
		-- PA is a technique
		technique = true,
		name = "Zonde",
		desc = "Electric Technique. Summons a thunderbolt on the targetted enemy.",
		cost = 9,
		potency = 260,
		actionspeed = 30,
		icon = "H_TZONDE",
		casttime = TICRATE/3,
		sound = sfx_zonde,
		
		spawntarget = true,
		element = ELEM_ELEC,
		spawndist = 48,
		speed = 30,
		
		object = MT_THOK,
		
		spawnfunc = function(msl)
			local zonde = P_SpawnMobj(msl.x, msl.y, msl.z, MT_THOK)
			zonde.state = S_ZONDE
			zonde.color = SKINCOLOR_GOLD
			local hb = PSO.lib.spawnHurtBox(zonde.x, zonde.y, zonde.z, 32<<FRACBITS, 180<<FRACBITS, msl.target, msl.pso.potency, DMT_TECHNIQUE, 4)
			hb.pso.combobreaker = msl.pso.combobreaker
			hb.pso.element = ELEM_ELEC

			P_RemoveMobj(msl)

			return zonde
		end,
}

PSO.info.photonArts["gizonde"] = {
		-- PA is a technique
		technique = true,
		name = "Gizonde",
		desc = "Electric Technique. Electrifies the target with a current that spreads to up to 5 nearby enemies, damaging all of them.",
		cost = 18,
		potency = 220,
		actionspeed = 30,
		icon = "H_TGZOND",
		casttime = TICRATE/3,
		sound = sfx_zonde,
		
		spawntarget = true,
		element = ELEM_ELEC,
		spawndist = 48,
		speed = 30,
		
		object = MT_THOK,
		
		spawnfunc = function(msl)
			
			local curr_target = msl.target.player.pso.target
			local check_radius = 256
			msl.targets = {}
			
			if curr_target
				
				while curr_target and #msl.targets < 5
					msl.targets[#msl.targets+1] = curr_target

					local hb = PSO.lib.spawnHurtBox(curr_target.x, curr_target.y, curr_target.z, 32<<FRACBITS, 180<<FRACBITS, msl.target, msl.pso.potency, DMT_TECHNIQUE, 4)
					hb.pso.combobreaker = msl.pso.combobreaker
					hb.pso.potency = msl.pso.potency
					hb.pso.locks = 1
					hb.pso.element = ELEM_ELEC

					for i = 1, 8
						local s = P_SpawnMobj(curr_target.x, curr_target.y, curr_target.z, MT_SUPERSPARK)
						s.colorized = true
						s.scale = $*3/2
						s.color = SKINCOLOR_GOLD
						s.momx = P_RandomRange(-32, 32)*FRACUNIT
						s.momy = P_RandomRange(-32, 32)*FRACUNIT
						s.momz = P_RandomRange(-32, 32)*FRACUNIT
					end
					
					local newtarget
					local x1 = curr_target.x - check_radius*FRACUNIT
					local x2 = curr_target.x + check_radius*FRACUNIT
					local y1 = curr_target.y - check_radius*FRACUNIT
					local y2 = curr_target.y + check_radius*FRACUNIT
					
					local targets = {}
					
					local function searchTarget(ref, found)
						if ref != found
						and found and found.valid and found.type ~= MT_PLAYER
							
							if found.flags & MF_SHOOTABLE
								
								local dupe
								for i = 1, #msl.targets
									--print(tostring(msl.targets[i]).."/"..tostring(found))
									if msl.targets[i] == found
										dupe = true
										break
									end
								end
								
								if not dupe
									--print("add: "..tostring(found))
									targets[#targets+1] = found
								end
							end
						end
					end
					
					searchBlockmap("objects", searchTarget, curr_target, x1, x2, y1, y2)
					
					-- sort the target list by distance
					table.sort(targets, function(a, b)
						return R_PointToDist2(curr_target.x, curr_target.y, a.x, a.y) < R_PointToDist2(curr_target.x, curr_target.y, b.x, b.y)
					end)
					
					-- if we have no target to begin with this will be nil and end the loop.
					local newtarget = targets[1]
					if newtarget and newtarget.valid
						-- spawn a chain between curr_target and the newtarget
						
						--local dist = R_PointToDist2(curr_target.x, curr_target.y, newtarget.x, newtarget.y)
						for i = 1, 5
							
							local x = PSO.lib.lerp(curr_target.x, newtarget.x, FRACUNIT/5 *i)
							local y = PSO.lib.lerp(curr_target.y, newtarget.y, FRACUNIT/5 *i)
							local z = PSO.lib.lerp(curr_target.z, newtarget.z, FRACUNIT/5 *i)
							
							local s = P_SpawnMobj(x, y, z, MT_SUPERSPARK)
							s.colorized = true
							s.color = SKINCOLOR_GOLD
							s.momx = P_RandomRange(-4, 4)*FRACUNIT
							s.momy = P_RandomRange(-4, 4)*FRACUNIT
							s.momz = P_RandomRange(-4, 4)*FRACUNIT
						end
					end
					
					curr_target = newtarget
				end
			else
				for i = 1, 8
					local s = P_SpawnMobj(msl.x, msl.y, msl.z, MT_SUPERSPARK)
					s.colorized = true
					s.color = SKINCOLOR_GOLD
					s.momx = P_RandomRange(-32, 32)*FRACUNIT
					s.momy = P_RandomRange(-32, 32)*FRACUNIT
					s.momz = P_RandomRange(-32, 32)*FRACUNIT
				end
			end

			P_RemoveMobj(msl)
		end,
}

PSO.info.photonArts["razonde"] = {
		-- PA is a technique
		technique = true,
		name = "Razonde",
		desc = "Electric Technique. Summons a field that periodically damages enemies over time.",
		cost = 25,
		potency = 440,
		actionspeed = 30,
		icon = "H_TRZOND",
		casttime = TICRATE/3,
		sound = sfx_zonde,
		
		spawntarget = true,
		element = ELEM_ELEC,
		spawndist = 0,
		speed = 0,
		
		object = MT_RAZONDE,
		
		spawnfunc = function(msl)
			-- make sure razonde cannot have more than 1 instance per player
			local mo = msl.target
			
			if mo.lastrazonde
			and mo.lastrazonde.valid
				P_RemoveMobj(mo.lastrazonde)	-- don't stack!!
			end
			
			mo.lastrazonde = msl
			msl.momx = 0
			msl.momy = 0
			msl.momz = 0
		end,
}

freeslot("SPR_KLIT", "SPR_NZND")

PSO.info.photonArts["nazonde"] = {
		-- PA is a technique
		technique = true,
		name = "Nazonde",
		desc = "Electric Technique. Hold to summon a field that continuously damages nearby enemies while draining PP.",
		cost = 15,
		potency = 160,
		actionspeed = 100,
		icon = "H_TNZOND",
		casttime = TICRATE/3,
		--sound = sfx_zonde,
		
		element = ELEM_ELEC,
		spawndist = 0,
		speed = 0,

		held = true,
		heldmin = TICRATE,			-- min time the tech will be "held" for regardless of user input
		helddelay = TICRATE/3 +2,	-- how much time between each cast?
		helddrain = 2,			-- 2 pp
		helddrainf = 3,			-- per 3 frames
		
		object = MT_THOK,
		
		heldfunc = function(mo)
			if leveltime%10 == 0
				S_StartSound(mo, sfx_cdfm34)
			end

			for i = 1, P_RandomRange(1, 4)
				local a = P_SpawnMobj(mo.x, mo.y, mo.z +mo.height/4, MT_THOK)
				a.sprite = SPR_KLIT
				a.colorized = true
				a.color = SKINCOLOR_GOLD
				a.scale = mo.scale/2
				a.frame = P_RandomRange(0, P) |FF_FULLBRIGHT|TR_TRANS40|FF_ADD
				a.tics = 1
				a.momx, a.momy, a.momz = mo.momx, mo.momy, mo.momz
			end	
			
			local n = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			n.sprite = SPR_NZND
			n.frame = (leveltime%26)/2 | TR_TRANS80 | FF_FULLBRIGHT | FF_ADD
			n.scale = FRACUNIT*4
			n.tics = 1
			
			if leveltime%3 == 0
				for i = 1, 2
					local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
					s.momx = P_RandomRange(-32, 32)*FRACUNIT
					s.momy = P_RandomRange(-32, 32)*FRACUNIT
					s.momz = P_RandomRange(-32, 32)*FRACUNIT
					s.colorized = true
					s.color = SKINCOLOR_GOLD
				end
			end
			
		end,
		
		spawnfunc = function(msl)
			-- make sure razonde cannot have more than 1 instance per player
			if msl and msl.valid
				local hb = PSO.lib.spawnHurtBox(msl.x, msl.y, msl.z, 600<<FRACBITS, 84<<FRACBITS, msl.target, msl.pso.potency, DMT_TECHNIQUE, 2)
				hb.pso.combobreaker = msl.pso.combobreaker
				hb.pso.potency = msl.pso.potency
				hb.pso.element = msl.pso.element			
				P_RemoveMobj(msl)
			end
		end,
}

freeslot("MT_PSO_ZONDEEL")
mobjinfo[MT_PSO_ZONDEEL] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 8<<FRACBITS,
	height = 32<<FRACBITS,
	flags = MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_NOCLIP,
}

local zondeel_range = 680
local zondeel_strength = 2
local zondeel_immunity = TICRATE*10

addHook("MobjThinker", function(mo)
	
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	if leveltime%10 == 0
		S_StartSound(mo, sfx_cdfm34)
	end

	-- sucking effect
	local newtarget
	local x1 = mo.x - zondeel_range*FRACUNIT
	local x2 = mo.x + zondeel_range*FRACUNIT
	local y1 = mo.y - zondeel_range*FRACUNIT
	local y2 = mo.y + zondeel_range*FRACUNIT
					
	local targets = {}
	
	if mo.tics > TICRATE*3/2
		local function pullTarget(ref, found)
			if ref != found
			and found and found.valid	
			and found.flags & MF_ENEMY
			and PSO.info.enemies[found.type]
			and abs(mo.z - found.z) < 512*FRACUNIT
			and P_CheckSight(mo, found)
			and not (found.zondeeld and leveltime - found.zondeeld < zondeel_immunity)
			
				local info = PSO.info.enemies[found.type]
				local weight = max(1, info.weight)
				
				if weight < 4	-- weight of 4 is immovable
					
					local a = R_PointToAngle2(found.x, found.y, mo.x, mo.y)
					found.pso.stagger = (info.stagger_duration or TICRATE/3) -1
					P_Thrust(found, a, mo.pso.potency*mo.scale / weight / 100)
					
					if mo.tics == TICRATE*3/2 +1
						found.zondeeld = leveltime
					end
					
				end
			end
		end
						
		searchBlockmap("objects", pullTarget, mo, x1, x2, y1, y2)
	end

	-- visuals
	for i = 1, 2
		
		local a = P_RandomRange(0, 359)*ANG1
		local xp = mo.x + zondeel_range*cos(a)
		local yp = mo.y + zondeel_range*sin(a)
		
		local s = P_SpawnMobj(xp, yp, mo.z, MT_SUPERSPARK)
		s.colorized = true
		s.color = SKINCOLOR_GOLD
		s.angle = a + ANG1*180
		P_InstaThrust(s, s.angle, (zondeel_range/10)*FRACUNIT)
	end
	
	if leveltime % 4 == 0
		local a = P_SpawnMobj(mo.x, mo.y, mo.z +mo.height/4, MT_THOK)
		a.sprite = SPR_KLIT
		a.colorized = true
		a.color = SKINCOLOR_GOLD
		a.scale = mo.scale/2
		a.frame = P_RandomRange(0, P) |FF_FULLBRIGHT|TR_TRANS40|FF_ADD
		a.tics = 10
		a.fuse = 10
		a.momx, a.momy, a.momz = mo.momx, mo.momy, mo.momz
	end
	
end, MT_PSO_ZONDEEL)

PSO.info.photonArts["zondeel"] = {
		-- PA is a technique
		technique = true,
		name = "Zondeel",
		desc = "Electric Technique. Deploys an electromagnetic field that sucks enemies inwards based on their weight.",
		cost = 30,
		potency = 150,
		actionspeed = 30,
		icon = "H_TZDEEL",
		casttime = TICRATE/3,
		sound = sfx_zonde,
		
		element = ELEM_ELEC,
		spawndist = 0,
		speed = 0,
		
		object = MT_PSO_ZONDEEL,
		
		spawnfunc = function(msl)
			-- make sure razonde cannot have more than 1 instance per player
			local mo = msl.target
			
			if mo.zondeel
			and mo.zondeel.valid
				P_RemoveMobj(mo.zondeel)	-- don't stack!!
			end
			
			mo.zondeel = msl
			msl.tics = TICRATE*3
		end,
}