freeslot("MT_FOIE", "S_FOIE", "S_FOIEDEATH")

freeslot("sfx_foie", "sfx_barta", "sfx_zonde", "sfx_shifta", "sfx_deband")

mobjinfo[MT_FOIE] = {
	doomednum = -1,
	spawnhealth = 1000,
	spawnstate = S_FOIE,
	deathstate = S_FOIEDEATH,
	xdeathstate = S_FOIEDEATH,
	radius = 24<<FRACBITS,
	height = 48<<FRACBITS,	
	flags = MF_NOGRAVITY|MF_MISSILE,
}

function A_FoieVisual(mo)
	mo.sprite = SPR_PUMA
	mo.spritexacale = FRACUNIT*3/2
	mo.spriteyscale = FRACUNIT*3/2
	
	if leveltime & 1
		mo.flags2 = $|MF2_DONTDRAW
	else
		mo.flags2 = $ & ~MF2_DONTDRAW
	end
	
	if leveltime & 1
		local g = P_SpawnGhostMobj(mo)
		g.destscale = 1
		
		local smoke = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SMOKE)
		smoke.scale = mo.scale
		smoke.momz = P_RandomRange(1, 4)*FRACUNIT
		smoke.momx = P_RandomRange(-4, 4)*FRACUNIT
		smoke.momy = P_RandomRange(-4, 4)*FRACUNIT
	end
	
	local e = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SONIC3KBOSSEXPLODE)
	e.scale = mo.scale*2
	e.destscale = 1
	e.scalespeed = mo.scale/24
	e.momx = P_RandomRange(-8, 8)*mo.scale
	e.momy = P_RandomRange(-8, 8)*mo.scale
	e.momz = P_RandomRange(-8, 8)*mo.scale
end

function A_FoieDeath(mo)
	for i = 1, 16
		local e = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SONIC3KBOSSEXPLODE)
		e.scale = mo.scale*3/2
		e.destscale = 1
		e.scalespeed = mo.scale/12
		e.momx = P_RandomRange(-32, 32)*mo.scale
		e.momy = P_RandomRange(-32, 32)*mo.scale
		e.momz = P_RandomRange(-32, 32)*mo.scale
	end

	-- spawn an explosion hitbox:
	local h = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 200<<FRACBITS, 64<<FRACBITS, mo.target, mo.pso.potency, DMT_TECHNIQUE, 4, RT_KNOCKBACK)
	PSO.lib.copyWeaponProperties(mo, h)
	if h.pso.whitelist
		mo.pso.whitelist = h.pso.whitelist
	else
		if not mo.pso.whitelist
			mo.pso.whitelist = {}
		end
		h.pso.whitelist = mo.pso.whitelist
	end
	h.pso.combobreaker = mo.pso.combobreaker
	h.pso.element = mo.pso.element
	h.pso.potency = mo.pso.potency
		
	-- search for the enemy we hit:
	local enemies = {}
	searchBlockmap("objects", 
		function(ref, found)
				
			if found and found.valid and found.flags & MF_SHOOTABLE
				enemies[#enemies+1] = found
					
			end
		end,
	mo)
		
	table.sort(enemies, 
		function(a, b)
			return R_PointToDist2(mo.x, mo.y, a.x, a.y) < R_PointToDist2(mo.x, mo.y, b.x, b.y)
		end
	)
		
	if enemies[1] and enemies[1].valid
		h.pso.whitelist[#h.pso.whitelist+1] = enemies[1]
	end
	
	local dp = displayplayer
	if dp and dp.valid and R_PointToDist2(dp.mo.x, dp.mo.y, mo.x, mo.y) < 1024*FRACUNIT
		P_StartQuake(FRACUNIT*10, 4)
	end
	
end

states[S_FOIE] = {SPR_NULL, A, 1, A_FoieVisual, 0, 0, S_FOIE}
states[S_FOIEDEATH] = {SPR_NULL, A, 1, A_FoieDeath, 0, 0, S_NULL}

PSO.info.photonArts["foie"] = {
		-- PA is a technique
		technique = true,
		name = "Foie",
		desc = "Fire Technique. Launches a fireball that explodes on impact with an enemy",
		cost = 14,
		potency = 490,
		icon = "H_TFOIE",
		casttime = TICRATE*3/4 -5,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_foie,
		
		aim = true,				-- tech can be aimed up/down
		reaction = RT_KNOCKBACK, -- foie does knockback
		element = ELEM_FIRE,
		object = MT_FOIE,		-- object to spawn
		speed = 40,				-- projectile speed.
		actionspeed = 40,
		spawndist = 32,			-- distance to spawn from the player if spawntarget is false
		spawntarget = false,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.
		spawnheight = 50,		-- height % to spawn the object at
		
}

PSO.info.photonArts["gifoie"] = {
		-- PA is a technique
		technique = true,
		name = "Gifoie",
		desc = "Fire Technique. Creates a ring of fire around the caster.",
		cost = 22,
		actionspeed = 20,
		potency = 420,
		icon = "H_TGFOIE",
		casttime = TICRATE*3/4 +1,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_foie,

		reaction = RT_KNOCKBACK, -- foie does knockback
		element = ELEM_FIRE,
		spawndist = 0,			-- distance to spawn from the player if spawntarget is false
		spawntarget = false,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.

		object = MT_THOK,
		
		spawnfunc = function(msl)
			for i = 1, 24
				local ang = (360/24)*i*ANG1
				local e = P_SpawnMobj(msl.x, msl.y, msl.z, MT_SONIC3KBOSSEXPLODE)
				e.scale = msl.scale*3/2
				P_InstaThrust(e, ang, FRACUNIT*48)
			end

			for i = 1, 16
				local ang = (360/16)*i*ANG1
				local e = P_SpawnMobj(msl.x, msl.y, msl.z, MT_SONIC3KBOSSEXPLODE)
				e.scale = msl.scale
				e.destscale = 1
				P_InstaThrust(e, ang, FRACUNIT*(16 + P_RandomRange(-4, 4)))
			end

			for i = 1, 16
				local ang = (360/16)*i*ANG1
				local e = P_SpawnMobj(msl.x, msl.y, msl.z, MT_SONIC3KBOSSEXPLODE)
				e.scale = msl.scale
				e.destscale = msl.scale*2
				e.momz = P_RandomRange(1, 4)*FRACUNIT
				P_InstaThrust(e, ang, FRACUNIT*(16 + P_RandomRange(-4, 4)))
			end

			-- spawn an explosion hitbox:
			local h = PSO.lib.spawnHurtBox(msl.x, msl.y, msl.z, 455<<FRACBITS, 128<<FRACBITS, msl.target, msl.pso.potency, DMT_TECHNIQUE, 4)
			h.pso.whitelist = $ or {}
			h.pso.reactionradialangle = true
			h.pso.element = ELEM_FIRE
			h.pso.combobreaker = msl.pso.combobreaker

			if msl.target.player == displayplayer
				P_StartQuake(FRACUNIT*10, 4)
			end

			P_RemoveMobj(msl)
		end,
	
}

PSO.info.photonArts["rafoie"] = {
		-- PA is a technique
		technique = true,
		name = "Rafoie",
		desc = "Fire Technique. Creates a flame explosion at the target's location",
		cost = 16,
		actionspeed = 35,
		potency = 410,
		icon = "H_TRFOIE",
		casttime = TICRATE*2/3,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_foie,
		
		reaction = RT_KNOCKBACK, -- foie does knockback
		element = ELEM_FIRE,
		spawndist = 128,		-- distance to spawn from the player if spawntarget is false
		spawntarget = true,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.
		spawnheight = 50,		-- height % to spawn the object at

		object = MT_THOK,
		
		spawnfunc = function(msl)
			for i = 1, 24
				local e = P_SpawnMobj(msl.x, msl.y, msl.z, MT_SONIC3KBOSSEXPLODE)
				e.scale = msl.scale*3/2
				e.destscale = 1
				e.scalespeed = msl.scale/12
				e.momx = P_RandomRange(-32, 32)*msl.scale
				e.momy = P_RandomRange(-32, 32)*msl.scale
				e.momz = P_RandomRange(-32, 32)*msl.scale
			end

			for i = 1, 10
				local e = P_SpawnMobj(msl.x + P_RandomRange(-64, 64)*msl.scale, msl.y + P_RandomRange(-64, 64)*msl.scale, msl.z, MT_SONIC3KBOSSEXPLODE)
				e.scale = msl.scale*3/2
				e.destscale = 1
				e.scalespeed = msl.scale/12
				e.momx = P_RandomRange(-6, 6)*msl.scale
				e.momy = P_RandomRange(-6, 6)*msl.scale
				e.momz = P_RandomRange(24, 32)*msl.scale
			end

			-- spawn an explosion hitbox:
			local h = PSO.lib.spawnHurtBox(msl.x, msl.y, msl.z, 400<<FRACBITS, 64<<FRACBITS, msl.target, msl.pso.potency, DMT_TECHNIQUE, 4, RT_KNOCKBACK)
			h.angle = msl.angle
			h.pso.element = ELEM_FIRE
			h.pso.combobreaker = msl.pso.combobreaker
			h.pso.potency = msl.pso.potency

			if msl.target.player == displayplayer
				P_StartQuake(FRACUNIT*10, 4)
			end

			P_RemoveMobj(msl)
		end,
	
}

freeslot("MT_PSO_SAFOIE")
mobjinfo[MT_PSO_SAFOIE] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	radius = 32<<FRACBITS,
	height = 48<<FRACBITS,
	spawnhealth = 1000,
	flags = MF_NOGRAVITY,
}

local safoie_ang = 60
local safoie_time = TICRATE/3

addHook("MobjThinker", function(mo)
	
	local ang_move = (2*safoie_ang) / safoie_time
	
	if not mo.extravalue1
		mo.angle = mo.target.angle - ANG1*safoie_ang
		mo.extravalue1 = 1
		mo.fuse = safoie_time+1
		mo.tx = mo.target.x
		mo.ty = mo.target.y
		
		mo.dist = R_PointToDist2(mo.x, mo.y, mo.tx, mo.ty)
	end
	
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	local x = mo.tx + FixedMul(mo.dist, cos(mo.angle))
	local y = mo.ty + FixedMul(mo.dist, sin(mo.angle))
	P_MoveOrigin(mo, x, y, mo.z)
	
	mo.extravalue1 = $+1
	mo.angle = $ + ang_move*ANG1
	
	if mo.extravalue1 == 2
	or mo.extravalue1 == safoie_time/2
	or mo.extravalue1 == safoie_time-1
		local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, 256<<FRACBITS, 256<<FRACBITS, mo.target, mo.pso.potency/3, DMT_TECHNIQUE, 2, RT_STAGGER)
		PSO.lib.copyWeaponProperties(mo, hb)
		hb.pso.combobreaker = mo.pso.combobreaker
		hb.pso.element = mo.pso.element
		hb.pso.potency = mo.pso.potency/3
		hb.pso.hitlag = 3
		hb.pso.screenshake = {intensity = FRACUNIT*8, time=5}
	end
	
	-- visuals
	A_FoieVisual(mo)
end, MT_PSO_SAFOIE)

PSO.info.photonArts["safoie"] = {
		-- PA is a technique
		technique = true,
		name = "Safoie",
		desc = "Fire Technique. Creates a wave of flames in front of the user.",
		cost = 20,
		potency = 550,
		icon = "H_TSFOIE",
		casttime = TICRATE/2,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_foie,
		
		aim = true,				-- tech can be aimed up/down
		reaction = RT_NONE, -- foie does knockback
		element = ELEM_FIRE,
		object = MT_PSO_SAFOIE,		-- object to spawn
		speed = 40,				-- projectile speed.
		actionspeed = 40,
		spawndist = 160,		-- distance to spawn from the player if spawntarget is false
		spawntarget = false,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.
		spawnheight = 0,		-- height % to spawn the object at
		
}

freeslot("MT_PSO_FOVERSE")
mobjinfo[MT_PSO_FOVERSE] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	height = 32*FRACUNIT,
	radius = 32*FRACUNIT,
	flags = MF_NOGRAVITY,
}

local foverse_dist = 48
local foverse_duration = TICRATE*30
local foverse_hit_rate = TICRATE*2/3
local foverse_range = 384*FRACUNIT

addHook("MobjThinker", function(mo)
	
	mo.fuse = -1
	mo.tics = -1
	
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	local ang = leveltime*4*ANG1
	if mo.extravalue1 < (foverse_duration - TICRATE*3) or leveltime & 1
		for i = 1, 2

			
			local fx = mo.x + foverse_dist*cos(ang)
			local fy = mo.y + foverse_dist*sin(ang)
			
			local f = P_SpawnMobj(fx, fy, mo.z + FRACUNIT*8, MT_THOK)
			f.scale = FRACUNIT*3/4
			f.sprite = SPR_DFLM
			f.frame = H | FF_TRANS20 | FF_FULLBRIGHT
			f.destscale = 1
			f.tics = 12
			f.momx = P_RandomRange(-3, 3)*FRACUNIT
			f.momy = P_RandomRange(-3, 3)*FRACUNIT
			f.momz = P_RandomRange(-2, 2)*FRACUNIT
			
			if leveltime%3 == 0
				local f = P_SpawnMobj(fx, fy, mo.z + FRACUNIT*P_RandomRange(8, 32), MT_THOK)
				f.sprite = SPR_FLAM
				f.frame = P_RandomRange(I, L) | FF_TRANS20 | FF_FULLBRIGHT
				f.destscale = 1
				f.scalespeed = FRACUNIT/80
				f.tics = 12
				f.momx = P_RandomRange(-1, 1)*FRACUNIT
				f.momy = P_RandomRange(-1, 1)*FRACUNIT
				f.momz = P_RandomRange(0, 3)*FRACUNIT
			end
			
			if leveltime%10 == 0
				local s = P_SpawnMobj(fx, fy, mo.z + FRACUNIT*8, MT_SMOKE)
				s.momx = P_RandomRange(-1, 1)*FRACUNIT
				s.momy = P_RandomRange(-1, 1)*FRACUNIT
				s.momz = P_RandomRange(-1, 1)*FRACUNIT	
				s.destscale = FRACUNIT*2
			end
			
			ang = $ + ANG1*180
		end
	end
	
	-- teleport to target
	P_SetOrigin(mo, mo.target.x, mo.target.y, mo.target.z)
	
	mo.extravalue1 = $+1
	
	if mo.extravalue1 % foverse_hit_rate == 0
		local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z, foverse_range, 256<<FRACBITS, mo.target, mo.pso.potency, DMT_TECHNIQUE, 2, RT_NONE)
		PSO.lib.copyWeaponProperties(mo, hb)
		hb.pso.combobreaker = mo.pso.combobreaker
		hb.pso.element = mo.pso.element
		hb.pso.potency = mo.pso.potency
	end
	
	if mo.extravalue1 > (mo.extravalue2 or foverse_duration)
		P_RemoveMobj(mo)
		return
	end
	
end, MT_PSO_FOVERSE)

PSO.info.photonArts["foverse"] = {
		-- PA is a technique
		technique = true,
		name = "Foverse",
		desc = "Fire Technique. Summon flames that orbit you for 20s, dealing damage to nearby enemies.",
		cost = 20,
		potency = 40,
		icon = "H_TFVERS",
		casttime = TICRATE/2,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_foie,
		
		aim = true,				-- tech can be aimed up/down
		reaction = RT_NONE, -- foie does knockback
		element = ELEM_FIRE,
		object = MT_PSO_FOVERSE,		-- object to spawn
		speed = 40,				-- projectile speed.
		actionspeed = 40,
		spawndist = 0,		-- distance to spawn from the player if spawntarget is false
		spawntarget = false,	-- if true, spawns ontop of the target, or somewhere in front of the player if no target is detected.
		spawnheight = 0,		-- height % to spawn the object at
		
		spawnfunc = function(msl)
			
			if msl and msl.valid
			and msl.target and msl.target.valid
				
				-- remove existing foverse if necessary
				if msl.target.foverse
				and msl.target.foverse.valid
					P_RemoveMobj(msl.target.foverse)
				end
				
				msl.target.foverse = msl
				
			end
		end,
}


PSO.info.photonArts["shifta"] = {
		-- PA is a technique
		technique = true,
		name = "Shifta",
		desc = "Fire Technique. Boosts natural & offensive PP recovery by 20% for 1:30 for the caster and nearby allies.",
		cost = 25,
		potency = 100+PSO.info.SHIFTA,
		icon = "H_TSHIFT",
		casttime = TICRATE*3/4,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_shifta,
		
		element = ELEM_FIRE,
		spawndist = 0,
		speed = 30,
		
		object = MT_THOK,
		
		spawnfunc = function(msl)
			-- make sure razonde cannot have more than 1 instance per player
			local mo = msl.target
			local mop = mo.player
			
			if not mop return end
			
			local range = 768
			if PSO.lib.hasSkill(mop, "fo_territoryburst")
				range = $*3
			end
			
			for i = 1, 16
				local ang = (360/16)*i*ANG1
				
				local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
				s.color = SKINCOLOR_RED
				s.colorized = true
				
				P_InstaThrust(s, ang, FRACUNIT*32)
			end
			
			
			-- gib shifta
			local duration = TICRATE*90
			if PSO.lib.hasSkill(mop, "fo_supportbitch")
				duration = $*2
			end
			
			for p in players.iterate do
				
				if p and p.mo and p.mo.valid
				and R_PointToDist2(mo.x, mo.y, p.mo.x, p.mo.y) < range*mo.scale
					PSO.lib.addShifta(p.mo, duration)
				end
			end			
			
			P_RemoveMobj(msl)
		end,
}

PSO.info.photonArts["jellen"] = {
		-- PA is a technique
		technique = true,
		name = "Jellen",
		desc = "Fire Technique. Decreases nearby enemies' resistance to Chains, increasing their duration by 15% for 1:30.",
		cost = 20,
		potency = 100+PSO.info.JELLEN,
		icon = "H_TJEL",
		casttime = TICRATE/2,	-- how long it takes before the tech weapon swings to cast.
		
		sound = sfx_kc59,
		
		element = ELEM_FIRE,
		spawndist = 0,
		speed = 30,
		
		object = MT_THOK,
		
		spawnfunc = function(msl)
			-- make sure razonde cannot have more than 1 instance per player
			local mo = msl.target
			local mop = mo.player
			
			if not mop return end
			
			local range = 768
			if PSO.lib.hasSkill(mop, "fo_territoryburst")
				range = $*3
			end
			
			for i = 1, 16
				local ang = (360/16)*i*ANG1
				
				local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
				s.color = SKINCOLOR_CRIMSON
				s.colorized = true
				
				P_InstaThrust(s, ang, FRACUNIT*32)
			end
			
			
			-- gib jellen
			local duration = TICRATE*90
			/*if PSO.lib.hasSkill(mop, "fo_supportbitch")
				duration = $*2
			end*/

			local x1 = mo.x - range*FRACUNIT
			local x2 = mo.x + range*FRACUNIT
			local y1 = mo.y - range*FRACUNIT
			local y2 = mo.y + range*FRACUNIT
				
			local fn = function(mo, found)
					
				if found and found.valid
				and R_PointToDist2(mo.x, mo.y, found.x, found.y) < range*FRACUNIT
				and found.flags & MF_ENEMY
						
					PSO.lib.addJellen(found, duration)
				end
			end
				
			searchBlockmap("objects", fn, mo, x1, x2, y1, y2)		
			
			P_RemoveMobj(msl)
		end,	
}