-- some player related constants and stats
freeslot("sfx_lvup", "sfx_rnkup", "sfx_rnkup2", "sfx_rank1", "sfx_rank2", "sfx_rank3", "sfx_rank4", "sfx_mopen")
freeslot("sfx_pblast", "sfx_hulb", "sfx_raoc", "sfx_fopf", "sfx_rabr")
freeslot("SPR_RAOC", "SPR_SIGI", "SPR_HUWW")

PSO.info.dodgeroll_cost = 0

rawset(_G, "loc_camera", {})
hud.add(function(v, p, c)
	if loc_camera and c
		loc_camera = c
	end
end)

PSO.info.typestats = {
	
	[TYPE_HUNTER] = {
		
		hp = 250,
		pp = 85,
		
		satk = 20,
		ratk = 17,
		tatk = 15,
		def = 10,
		ppregen = 18,	-- 1 pp / 18f
		
		startingweapon = "sword",
		startingweaponr = "saber",	-- starting right h weapon
		startingweaponl = "handgun",		-- starting left h weapon
		
		-- weapon proficiency:
		/*
			0: Cannot equip
			1: 1-3*
			2: 4-6*
			3: 7-9*
			4: 10*-15*
		*/	
		weapons = {
			[WPT_SWORD] = 2,
			[WPT_DOUBLESABER] = 2,
			[WPT_RIFLE] = 0,
			[WPT_SHOTGUN] = 0,
			[WPT_ROD] = 0,
			[WPT_SABER] = 2,
			[WPT_HANDGUN] = 1,
			[WPT_WAND] = 0,
			[WPT_SPEAR] = 2,
			[WPT_TWINSABERS] = 2,
			[WPT_LAUNCHER] = 0,
			[WPT_TWINHANDGUNS] = 0,
			[WPT_TECHMAG] = 0,
		}
	},

	[TYPE_RANGER] = {
		
		hp = 225,
		pp = 100,
		
		satk = 15,
		ratk = 20,
		tatk = 17,
		def = 10,
		ppregen = 6,	-- 1 pp / 7f
		
		startingweapon = "rifle",
		startingweaponr = "saber",	-- starting right h weapon
		startingweaponl = "handgun",		-- starting left h weapon
		
		weapons = {
			[WPT_SWORD] = 0,
			[WPT_DOUBLESABER] = 0,
			[WPT_RIFLE] = 2,
			[WPT_SHOTGUN] = 2,
			[WPT_ROD] = 0,
			[WPT_SABER] = 1,
			[WPT_HANDGUN] = 2,
			[WPT_WAND] = 0,
			[WPT_SPEAR] = 0,
			[WPT_TWINSABERS] = 0,
			[WPT_LAUNCHER] = 2,
			[WPT_TWINHANDGUNS] = 2,
			[WPT_TECHMAG] = 0,			
		}
	},

	[TYPE_FORCE] = {
		
		hp = 200,
		pp = 115,
		
		satk = 17,
		ratk = 15,
		tatk = 20,
		def = 10,
		ppregen = 4,	-- 1 pp / 5f
		
		startingweapon = "rod",
		startingweaponr = "wand",	-- starting right h weapon
		startingweaponl = "handgun",		-- starting left h weapon
		
		weapons = {
			[WPT_SWORD] = 0,
			[WPT_DOUBLESABER] = 0,
			[WPT_RIFLE] = 0,
			[WPT_SHOTGUN] = 0,
			[WPT_ROD] = 2,
			[WPT_SABER] = 0,
			[WPT_HANDGUN] = 1,
			[WPT_WAND] = 2,		
			[WPT_SPEAR] = 0,
			[WPT_TWINSABERS] = 0,
			[WPT_LAUNCHER] = 0,
			[WPT_TWINHANDGUNS] = 0,
			[WPT_TECHMAG] = 2,			
		}
	},
	
	-- used for AIS stat calculations
	["ais"] = {
		
		hp = 450,
		pp = 200,
		satk = 35,
		ratk = 35,
		tatk = 35,
		def = 2,
	},
}

PSO.lib.initPlayerType = function(p, typ)

	local info = PSO.info.typestats[typ]
		
	local wep = PSO.lib.makeItem(info.startingweapon, {ELEM_ICE, 10})
	local wep_r = PSO.lib.makeItem(info.startingweaponr, {ELEM_ICE, 10})
	local wep_l = PSO.lib.makeItem(info.startingweaponl, {ELEM_ICE, 10})
	
	wep.equipped = true	-- cheat a little
	wep_r.equipped = true
	wep_l.equipped = true
	
	local armor = PSO.lib.makeItem("frame", {ELEM_ICE, 10})
	armor.equipped = true	-- cheat again
		
	-- defined like a shizo to test inventory sorting
	local inventory = {
		
		{wep},
		{wep_l},
		{wep_r},

		{armor},
		
		{"monomate", 5},
		{"dimate", 5},
		{"trimate", 2},
		{"moon atomizer", 1},
	}
		
	local palette = {
			
		-- [1]: consumables
		-- [2]: weapons: table because of dual wield
		-- [3]: armors
		{"monomate", "dimate", "trimate", "moon atomizer"},
		{{wep}, {wep_l, wep_r}},
		{armor},
	}
	
	p.pso.inventory = inventory
	p.pso.palette.layout = palette
end

-- EXP tables...
PSO.info.exptable = {50}
PSO.info.expdist = 4096 -- you have to be within 4096 units of something dying to get exp from it

-- the rest of the exp table
local total = 0
for i = 2, 201
	total = $+(PSO.info.exptable[i-1])
	local mult = 100 + FixedDiv(90<<FRACBITS, FixedSqrt((i-1)<<FRACBITS)*2)/FRACUNIT
	PSO.info.exptable[i] = PSO.info.exptable[i-1] *mult/100
	--print("Lv "..i..": "..PSO.info.exptable[i].." ("..mult..")")
	--print("<tr><th>"..i.."</th><td>"..PSO.info.exptable[i].."</td><td>"..total.."</td></tr>")
end

PSO.lib.givePlayerExp = function(p, exp)
	
	-- ais cumulates exp
	if p.pso.ais
		p.pso.aisexp = $ and $ + exp or exp
		return
	end
	
	-- same for arena quests, but only a portion of it.
	if PSO.lib.arenaQuestActive()
		p.pso.arenaexp = $ and $ + (exp*6/10) or (exp*6/10)
		return
	end
	
	/*if p.pso.level >= PSO.info.LEVELCAP
		p.pso.exp = 0
		return
	end*/
	
	local needxp = PSO.info.exptable[p.pso.level]
	p.pso.exp = $ + exp
	
	--print("Got "..exp.." EXP")
	
	local nosound = false
	
	local statsbefore = {
		p.pso.level,
		p.pso.maxhp,
		p.pso.maxpp,
		p.pso.satk,
		p.pso.ratk,
		p.pso.tatk,
		p.pso.def,
	}
	
	local prevlevel = p.pso.level
	
	while p.pso.exp >= PSO.info.exptable[p.pso.level]
		p.pso.exp = $ - PSO.info.exptable[p.pso.level]
		PSO.lib.levelUpPlayer(p, nosound)
		nosound = true	-- protect your ears <3
	end
	
	if nosound	-- this does mean we've levelled up!
		
		if prevlevel >= PSO.info.LEVELCAP
			PSO.lib.playerPushNotification(p, "Overlevel bonus\nreceived", 2)
		else
			
			PSO.lib.playerPushNotification(p, "\x82".."Level up! \x80"..statsbefore[1].." -> "..p.pso.level)
			PSO.lib.playerPushNotification(p, "\x82".."HP: \x80"..statsbefore[2].." -> "..p.pso.maxhp)
			PSO.lib.playerPushNotification(p, "\x82".."PP: \x80"..statsbefore[3].." -> "..p.pso.maxpp)
			PSO.lib.playerPushNotification(p, "\x82".."S. ATK: \x80"..statsbefore[4].." -> "..p.pso.satk)
			PSO.lib.playerPushNotification(p, "\x82".."R. ATK: \x80"..statsbefore[5].." -> "..p.pso.ratk)
			PSO.lib.playerPushNotification(p, "\x82".."T. ATK: \x80"..statsbefore[6].." -> "..p.pso.tatk)
			PSO.lib.playerPushNotification(p, "\x82".."DEF: \x80"..statsbefore[7].." -> "..p.pso.def)
		end
	end
end

COM_AddCommand("psod_getexp", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local expn = tonumber(arg) or 0
	PSO.lib.givePlayerExp(p, arg)
end)

COM_AddCommand("psod_getmeseta", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local meseta = tonumber(arg) or 0
	p.pso.meseta = $+meseta
end)

COM_AddCommand("psod_gettp", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local tp = tonumber(arg) or 0
	p.pso.typepoints[p.pso.type] = $+tp
end)

-- calculates the player's stats for a given type & level
-- if units is true, also counts unit stats
-- if ignoresynch is true, ignores level synch.
PSO.lib.playerPreviewStats = function(p, type, level, units, ignoresynch)
	local stats = {}
	
	if not p.pso.ais
		
		local synch = p.pso.levelsynch
		if synch
			level = synch
		end
		
		local typinfo = PSO.info.typestats[type]
		stats.maxhp = typinfo.hp + (typinfo.hp/50)*(level-1)
		stats.maxpp = typinfo.pp + (level-1)
		
		stats.satk = typinfo.satk + (typinfo.satk)*(level-1) /4
		stats.ratk = typinfo.ratk + (typinfo.ratk)*(level-1) /4
		stats.tatk = typinfo.tatk + (typinfo.tatk)*(level-1) /4
		stats.def = typinfo.def + (typinfo.def)*(level-1) /4
		stats.ppregen = 1000		-- resolution
		stats.actionspd = 100	-- resolution
		
		if units
			local armour = PSO.lib.getPlayerActiveArmour(p)
			
			if armour
				local def = PSO.info.items[armour.ref]
				local ehp = (def.hp or 0)
				local epp = (def.pp or 0)
				if synch
					local b, ahp, app = PSO.lib.lvSynchArmour(p, synch, armour)
					ehp = ahp or 0
					epp = app or 0
				end
				
				stats.maxhp = $ + ehp
				stats.maxpp = $ + epp
				
				if def.slots
					
					-- necessary to count these for lv synch
					local numslots = 0
					for i = 1, def.slots
						if armour.slots[i]
							numslots = $+1
						end
					end
					
					for i = 1, def.slots
						
						if armour.slots[i]
							local unit = armour.slots[i]
							local unitsynch
							if synch
								unitsynch = PSO.lib.lvSynchUnit(p, synch, unit, numslots)
							end
							
							stats.maxhp = $ + (unitsynch and unitsynch.hp or unit.hp or 0)
							stats.maxpp = $ + (unitsynch and unitsynch.pp or unit.pp or 0)
							stats.satk = $ + (unitsynch and unitsynch.satk or unit.satk or 0)
							stats.ratk = $ + (unitsynch and unitsynch.ratk or unit.ratk or 0)
							stats.tatk = $ + (unitsynch and unitsynch.tatk or unit.tatk or 0)
							stats.ppregen = max(1, $ + 1000*(unit.ppregen or 0)/100)
							stats.actionspd = max(1, $+(unit.actionspd or 0))
						end
					end
				end
			end
		end
		
		stats.maxhp = max($, 1)
		stats.maxpp = max($, 1)
		stats.satk = max($, 1)
		stats.ratk = max($, 1)
		stats.tatk = max($, 1)
		stats.def = max($, 1)
		
	else
		
		-- ais level is the quest's enemy level
		if server.questdata
					
			local q = PSO.info.quests[server.questdata.section].quests[server.questdata.index]
			level = q.levels[server.questdata.difficulty] or server.pso.level
		end
		
		if server.psoenemylv	-- debug cheats
			level = min(255, server.psoenemylv)
		end
		
		local info = PSO.info.typestats.ais
		
		stats.maxhp = info.hp *PSO.info.enemyStatScaling[level]/PSO.info.enemyStatScalingRes
		stats.maxpp = info.pp
		
		stats.satk = info.satk *PSO.info.enemyStatScaling[level]/PSO.info.enemyStatScalingRes
		stats.ratk = info.ratk *PSO.info.enemyStatScaling[level]/PSO.info.enemyStatScalingRes
		stats.tatk = info.tatk *PSO.info.enemyStatScaling[level]/PSO.info.enemyStatScalingRes
		
		stats.def = info.def *PSO.info.enemyStatScaling[level]/PSO.info.enemyStatScalingRes
		stats.ppregen = 1000
	end
	
	return stats
end

-- sets the player's stats for a given level using playerPreviewStats
PSO.lib.playerCalcStats = function(p, level, ignoresynch)
	level = $ or p.pso.level
	local stats = PSO.lib.playerPreviewStats(p, p.pso.type, level, true, ignoresynch)
	
	PSO.lib.removeDebandToughness(p)	-- remove deband toughness.
	
	for k, v in pairs(stats)
		p.pso[k] = max(1, v)
	end
	
	if not p.pso.ais
		PSO.lib.reapplySkills(p)
		
		p.pso.maxpp = $ + (p.pso.chainreactor or 0)
	end

	if PSO.lib.hasSkill(p, "fo_debandtoughness")
	and p.pso.deband
		PSO.lib.addDebandToughness(p)
	end
	
	PSO.lib.runHook("playerCalcStats", p, level, stats)
	
	-- make sure we don't keep extra hp/pp if we've taken units off
	p.pso.hp = min($, p.pso.maxhp)
	p.pso.pp = min($, p.pso.maxpp)
end

-- changes the player's type.
-- if unequip is true, also clears the weapon palettes
PSO.lib.playerChangeType = function(p, newtype, skipunequip)
	p.pso.type = newtype
	local typinfo = PSO.info.typestats[p.pso.type]

	--PSO.lib.removeDebandToughness(p)	-- do this here or we're gonna get some nasty issues.
	PSO.lib.playerCalcStats(p, p.pso.level)
	
	p.pso.hp = min($, p.pso.maxhp)
	p.pso.pp = min($, p.pso.maxpp)
	
	-- unequip all your shit
	if not skipunequip
		for i = 1, PSO.info.MAXPALSIZE
			
			if p.pso.palette.layout[2][i]
				for j = 1, 2
					if p.pso.palette.layout[2][i][j]
						p.pso.palette.layout[2][i][j].equipped = nil
					end
				end
			end
		end
		
		p.pso.palette.aweaponnum = 1
		p.pso.palette.weaponnum = 1
		
		p.pso.forceupdateweapon = true
		p.pso.palette.layout[2] = {}
		for i = 1, PSO.info.MAXPALSIZE
			p.pso.palette.layout[2][i] = nil
		end
		
		p.pso.palette.layout[2][1] = {}
	end

	/*-- do deband calculation
	if PSO.lib.hasSkill(p, "fo_debandtoughness")
	and p.pso.deband
		PSO.lib.addDebandToughness(p)
	end*/
	
end

-- checks to see if "it" is equipped. If it is equipped, it is forcefully unequipped.
-- if this is an armor, all units are unequipped from it.
-- if it is a unit, it is unequipped from all armors it is on.
PSO.lib.playerUnequipRef = function(p, it)
	
	-- for normal item, check item palette and remove from there.
	if it.type != IT_UNIT
	
		for i = 1, PSO.info.MAXPALSIZE
			for j = 2, 3
				if p.pso.palette.layout[j][i]
					for k = 1, 2
						if p.pso.palette.layout[j][i][k] == it
							p.pso.palette.layout[j][i][k].equipped = nil
							p.pso.palette.layout[j][i][k] = nil
							PSO.lib.playerFixWeaponPalette(p)
							p.pso.forceupdateweapon = true
						end
					end
				end
			end
		end
	
	else	-- for units, we go through the inventory instead and search for armors.
		for i = 1, #p.pso.inventory
			
			local cit = p.pso.inventory[i][1]
			if type(cit) == "table" and cit.type == IT_ARMOR
				
				local info = PSO.lib.itemInfo(cit)
				for j = 1, info.slots
					if cit.slots[j] == it
						cit.slots[j].equippedarmours = 0
						cit.slots[j].equipped = false
						cit.slots[j] = nil
					end
				end
			end
		end
	end
end

-- returns true/false, and a reason
-- if true, reason will always be 0 (OK)
-- if false, reason will either be 1 (level too low)
-- or 2, (proficiency too low)
PSO.lib.canEquip = function(p, item)
	
	local info = PSO.lib.itemInfo(item)
	local req = item.equiplevel or info.req or 1
	
	if item.unidentified
		return false
	elseif p.pso.level < req
		return false, 1
	elseif info.type == IT_WEAPON
		-- weapons check for proficiency:
		local proficiency = p.pso.weaponproficiencies[p.pso.type][info.weapontype]
		
		if (proficiency or 0) < (max(0, info.rarity -1)/3) +1
			return false, 2
		end
	end
	
	return true, 0
end


PSO.lib.levelUpPlayer = function(p, silent)
	
	--print("Level up!")

	if not silent
		S_StartSound(nil, sfx_lvup, p)
	end
	
	if p.pso.level >= PSO.info.LEVELCAP
		if not PSO.lib.playerGetItem(p, PSO.info.OVERLEVELREWARD, PSO.info.OVERLEVELQTY)
			-- bruh.
			if not PSO.lib.playerAddItemToTable(p, PSO.info.OVERLEVELREWARD, PSO.info.OVERLEVELQTY, "storage", PSO.lib.playerSortStorage, PSO.info.STORAGESIZE, true)
				-- BRO CLEAR YOUR FUCKING INVENTORY
				
				PSO.lib.playerPushNotification(p, "HOW IN THE FUCK\nIS YOUR STORAGE FULL!??!?!?!??!", 2)
				local drop = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_PSO_DROP)
				drop.psoitem = PSO.info.OVERLEVELREWARD
				drop.psoqty = PSO.info.OVERLEVELQTY
				
			else
				PSO.lib.playerPushNotification(p, "Inventory full!\nSent to storage.", 2)
			end
		end
		return	
	end
	
	local typinfo = PSO.info.typestats[p.pso.type]
	p.pso.level = $+1
	
	local maxsp = 4 + (p.pso.level/10)
	
	if p.pso.maxskillpoints < maxsp
		p.pso.maxskillpoints = maxsp
		PSO.lib.playerPushNotification(p, "Maximum Skill Points\nincreased to "..maxsp, 2)
	end
	
	-- recalculate stats!
	PSO.lib.playerCalcStats(p, p.pso.level)
	
	p.pso.hp = p.pso.maxhp
	p.pso.pp = p.pso.maxpp
end

-- debug function that changes the player's level
PSO.lib.setPlayerLevel = function(p, level)
	p.pso.level = level
	PSO.lib.playerCalcStats(p, p.pso.level)
	
	p.pso.hp = min($, p.pso.maxhp)
	p.pso.pp = min($, p.pso.maxpp)
	p.pso.maxskillpoints = 4 + (min(PSO.info.LEVELCAP, p.pso.level)/10)
end


COM_AddCommand("psod_setlevel", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	local lv = tonumber(arg) or 0
	if not lv return end
	
	PSO.lib.setPlayerLevel(p, lv)
	p.pso.exp = 0
end)

/*print("TEST EVAL")
print(nil or false)	-> false
print(false or nil) -> nil*/

PSO.lib.itemSort = function(a, b)
	
	local it1 = a[1]
	local it2 = b[1]
	
	local info1 = PSO.lib.itemInfo(it1)
	local info2 = PSO.lib.itemInfo(it2)
	
	if info1.type ~= info2.type
		return (info1.type or 1) < (info2.type or 1)
	end
	
	--print(tostring(itname(it1))..", "..tostring(itname(it2)))
	-- both items are weapons/armour
	if type(it1) == "table" and type(it2) == "table"
		
		if info1.type == info2.type
			
			-- comparing weapons!
			if info1.type == IT_WEAPON
			
				if info1.weapontype ~= info2.weapontype
					return (PSO.info.weaponToOrder[info1.weapontype] or 0) < (PSO.info.weaponToOrder[info2.weapontype] or 0)
				elseif info1.weapontype == info2.weapontype
					if (it1.unidentified or false) ~= (it2.unidentified or false)
						return it2.unidentified	-- make unidentified gear have lower priority
					elseif info1.atk ~= info2.atk
						return (info1.atk or 1) > (info2.atk or 1)
					elseif info1.rarity ~= info2.rarity
						return (info1.rarity or 1) > (info2.rarity or 1)
					elseif (it1.equipped or false) ~= (it2.equipped or false)
						return it1.equipped
					elseif it1.name ~= it2.name
						return info1.name < info2.name
					elseif it1.timestamp ~= it2.timestamp
						return (it1.timestamp or INT32_MAX) < (it2.timestamp or INT32_MAX)
					else	-- fallback, somehow?!
						return info1.name < info2.name
					end
				end
			
			-- comparing armours!
			elseif info1.type == IT_ARMOR

				if info1.def ~= info2.def
					return (info1.def or 1) > (info2.def or 1)
				elseif info1.rarity ~= info2.rarity
					return (info1.rarity or 1) > (info2.rarity or 1)
				elseif (it1.equipped or false) ~= (it2.equipped or false)
					return it1.equipped
				elseif it1.name ~= it2.name
					return it1.name < it2.name
				elseif it1.timestamp ~= it2.timestamp
					return (it1.timestamp or INT32_MAX) < (it2.timestamp or INT32_MAX)
				else
					return info1.name < info2.name
				end
			
			elseif info1.type == IT_UNIT
				if info1.rarity ~= info2.rarity
					return (info1.rarity or 1) > (info2.rarity or 1)
				elseif (it1.equipped or false) ~= (it2.equipped or false)
					return it1.equipped
				elseif it1.name ~= it2.name
					return it1.name < it2.name
				elseif it1.timestamp ~= it2.timestamp
					return (it1.timestamp or INT32_MAX) < (it2.timestamp or INT32_MAX)
				else
					return info1.name < info2.name	-- nothing yet.
				end
			end
			
		else
			return info1.name < info2.name	-- yes this works as last resort.
		end
		
	-- userdata vs string reference, userdata always wins.	
	-- impossible case
	/*elseif type(it1) ~= type(it2)
		return type(it1) == "table"	-- take whichever is userdata*/
	
	-- string reference (not weapon/armor)
	elseif type(it1) == "string" and type(it2) == "string"
		
		if info1.itemtype ~= info2.itemtype
			return info1.itemtype < info2.itemtype
		elseif info1.sortpriority ~= info2.sortpriority
			return (info1.sortpriority or INT32_MAX) < (info2.sortpriorirty or INT32_MAX)
		elseif info1.patype ~= info2.patype
			
			if info1.patype == nil
				return true
			elseif info2.patype == nil
				return false
			end
			
			return (PSO.info.weaponToOrder[info1.patype] or 999) < (PSO.info.weaponToOrder[info2.patype] or 999)
		elseif info1.panum ~= info2.panum
			return (info1.panum or 0) < (info2.panum or 0)			
		elseif info1.rarity ~= info2.rarity
			return (info1.rarity or 1) < (info2.rarity or 1)
		
		else
			
			if info1.rarity ~= info2.rarity
				return info1.rarity > info2.rarity
			else
				return info1.name < info2.name	-- yes this works
			end
		end	
	end
end

-- sorts the player's inventory and sets the variables telling where each item category begins.
-- this uses PSO.lib.itemSort
PSO.lib.playerSortItems = function(p, sortwhat)
	--print("sort")
	sortwhat = $ or "inventory"
	table.sort(p.pso[sortwhat], PSO.lib.itemSort)
	p.pso[sortwhat.."_sections"] = {1, 0, 0, #p.pso[sortwhat]+1}
	
	local i = #p.pso[sortwhat]
	
	while i
		
		local it = p.pso[sortwhat][i][1]
		local info = PSO.lib.itemInfo(it)
		
		-- we start armours at the first thing that isn't a weapon!
		if info.type == IT_ARMOR or info.type == IT_UNIT
			p.pso[sortwhat.."_sections"][2] = i
		end
		
		-- we're looking for something that's nor an armor, nor a unit!
		if info.type > IT_UNIT
			p.pso[sortwhat.."_sections"][3] = i
		end
		
		i = $-1
	end
	
	i = 3
	-- go over sections and give them the max possible thing if they haven't been set...
	while i > 1
		p.pso[sortwhat.."_sections"][i] = $ or p.pso[sortwhat.."_sections"][i+1]
		i = $-1
	end
end

-- implement this sorting method to inventory, storage and shop
PSO.lib.playerSortInventory = function(p)
	PSO.lib.playerSortItems(p, "inventory")
end

PSO.lib.playerSortStorage = function(p)
	PSO.lib.playerSortItems(p, "storage")
end

PSO.lib.playerSortShop = function(p)
	PSO.lib.playerSortItems(p, "shop")
end


-- add items to a table, stacks items that can be stacked.
PSO.lib.playerAddItemToTable = function(p, item, qty, where, sortfunc, limit, ignorecarrylimit)
	local info = PSO.lib.itemInfo(item)
	qty = $ or 1
	
	if info.type == IT_WEAPON
	or info.type == IT_ARMOR
	or info.type == IT_UNIT	-- non stackable...
		
		if #p.pso[where] >= limit
			return false
		end
		
		p.pso[where][#p.pso[where]+1] = {item, 1}
		sortfunc(p)
		return true
		
	else	-- stackable
		-- search for an instance of this item...
		for i = 1, #p.pso[where]
			if p.pso[where][i][1] == item
				
				if not ignorecarrylimit	-- ie: storage doesn't check for carry limit
				and info.carrylimit
					local curr = p.pso[where][i][2]
					
					-- nope nope nope.
					if curr >= info.carrylimit
						return false
					end
					
					p.pso[where][i][2] = $+qty
					p.pso[where][i][2] = min($, info.carrylimit)
					
					-- return how many were taken (inventory doesn't need sorting)
					return min(qty, info.carrylimit - curr)
				else
					p.pso[where][i][2] = $+qty
				end
				--PSO.lib.playerSortInventory(p)
				return true
			end
		end
		
		-- if we made it here we haven't found an instance
		if #p.pso[where] >= limit
			return false
		end
		
		p.pso[where][#p.pso[where]+1] = {item, qty or 1}
		
		-- check if we CAN hold that many
		if not ignorecarrylimit
		and info.carrylimit
		and p.pso[where][#p.pso[where]][2] > info.carrylimit
			p.pso[where][#p.pso[where]][2] = min($, info.carrylimit)
			return info.carrylimit
		end
		
		--print("A")
		sortfunc(p)
		return true
	end
end

PSO.lib.playerGetItem = function(p, item, qty)
	if PSO.lib.playerAddItemToTable(p, item, qty, "inventory", PSO.lib.playerSortInventory, PSO.info.INVENTORYSIZE)
		if type(item) == "table"
			item.timestamp = $ or p.pso.playtime	-- used for sorting
		end
		return true
	end
end


PSO.lib.playerRemoveItemFromTable = function(p, slot, qty, where, sortfunc)
	
	qty = $ or 1
	local it = p.pso[where][slot][1]
	if type(it) == "table"
		-- some sanity checks:
		if it.equipped return false end	-- cannot remove equipped item.
	end
	
	p.pso[where][slot][2] = $ or 1	-- if somehow the qty had never been set
	p.pso[where][slot][2] = max($ - qty, 0)
	
	if not p.pso[where][slot][2]
		table.remove(p.pso[where], slot)
		sortfunc(p)
	end
	return true
end

PSO.lib.playerLoseItem = function(p, slot, qty)
	return PSO.lib.playerRemoveItemFromTable(p, slot, qty, "inventory", PSO.lib.playerSortInventory)
end


-- sends an item from the player's inventory to the storage
PSO.lib.playerSendToStorage = function(p, itslot, qty)
	
	local item = p.pso.inventory[itslot][1]
	-- try to add it to the storage, if it fails, don't deplete it from the inventory
	local res = PSO.lib.playerAddItemToTable(p, item, qty, "storage", PSO.lib.playerSortStorage, PSO.info.STORAGESIZE, true)	
	-- failure
	if not res
		--print("\x82".."WARNING: ".."\x80".."Couldn't send item to storage!")
		return false
	end
	
	if type(res) == "number"
		return PSO.lib.playerLoseItem(p, itslot, res)
	else
		-- if we're here, also lose the items here.
		return PSO.lib.playerLoseItem(p, itslot, qty)
	end
end

-- retrieves an item from the player's storage and adds it to their inventory
PSO.lib.playerRetrieveFromStorage = function(p, itslot, qty)
	local item = p.pso.storage[itslot][1]
	
	-- try to add it to their inventory, if it fails, don't deplete it from the storage
	local res = PSO.lib.playerGetItem(p, item, qty)
	-- failure
	if not res
		return false
	end
	
	--print(res)
	if type(res) == "number"
		return PSO.lib.playerRemoveItemFromTable(p, itslot, res, "storage", PSO.lib.playerSortStorage)
	else
		return PSO.lib.playerRemoveItemFromTable(p, itslot, qty, "storage", PSO.lib.playerSortStorage)
	end
end

-- returns the number of the inventory slot the desired item is in, or 0 if it isn't there
PSO.lib.findPlayerItemInTable = function(p, item, where)
	
	local loosematch = 0
	
	for i = 1, #p.pso[where]
		local it = p.pso[where][i][1]
		
		if type(it) == "table" and it.ref == item
		and not loosematch
			loosematch = i
		end
		
		if it == item
			return i
		end
	end
	return loosematch
end

-- does the same but actually finds all instances of the item in the specified table
-- useful for weapon exchanges.
PSO.lib.findItemInstancesInTable = function(p, item, where)
	local matches = {}
	
	for i = 1, #p.pso[where]
		local it = p.pso[where][i][1]
		
		if type(it) == "table" and it.ref == item
		and not loosematch
			matches[#matches+1] = i
		end
		
		if it == item
			matches[#matches+1] = i
		end
	end
	return matches
end

-- does the same but find all of the items of a given rarity and type
PSO.lib.findRarityTypeInTable = function(p, rarity, typ, where)
	
	local matches = {}
	
	for i = 1, #p.pso[where]
		local it = p.pso[where][i][1]
		
		local info = PSO.lib.itemInfo(it)
		if info.type == typ
		and info.rarity == rarity
			matches[#matches+1] = i
		end
	end
	
	return matches
end

PSO.lib.findPlayerUnidentifiedWeapons = function(p)
	
	local ret = {}
	for i = 1, #p.pso.inventory
		
		local it = p.pso.inventory[i][1]
		if type(it) == "table"
			if it.unidentified
				ret[#ret+1] = {it, 1}
			elseif it.type ~= IT_WEAPON
				break	-- don't continue the iteration, if this isnt a weapon then nothing else going forward is.
			end
		end
	end
	return ret
end

-- unclean fixer for units that somehow show as equipped when they shouldn't be
PSO.lib.playerFixUnits = function(p)
	
	local armours = {}
	local units = {}
	
	-- make a list of armours and units
	for i = 1, #p.pso.inventory
		local it = p.pso.inventory[i][1]
		if it and it.type == IT_ARMOR
		and PSO.lib.itemInfo(it).slots	-- this shit better have slots.
			armours[#armours+1] = it
		elseif it and it.type == IT_UNIT and it.equipped	-- only check for units that say they're equipped
			units[#units+1] = it
		end
	end
	
	-- check if the units are equipped
	for i = 1, #armours
		local a = armours[i]
		for j = 1, PSO.lib.itemInfo(a).slots
			if a.slots[j]
				local k = #units
				while k
					if a.slots[j] == units[k]
						table.remove(units, k)	-- it's equipped so it doesn't need to be here anymore.
					end
					k = $-1
				end
			end
		end
	end
	
	-- every unit left at this point is unequipped and must be killed
	for i = 1, #units
		local u = units[i]
		u.equipped = false
		u.equippedarmours = 0	-- liar.
	end
end

PSO.lib.findPlayerItem = function(p, item)
	return PSO.lib.findPlayerItemInTable(p, item, "inventory")
end

PSO.lib.findPlayerStorageItem = function(p, item)
	return PSO.lib.findPlayerItemInTable(p, item, "storage")
end

PSO.lib.getPlayerActiveArmour = function(p)
	if p.pso.ais	-- ais has no armor equipped
		return nil
	end
	
	local pal = p.pso.palette.layout[3][p.pso.palette.armornum]
	return pal
end

-- sets trailing palette slots to nil
PSO.lib.playerFixWeaponPalette = function(p)
	local firstwep
	
	local pal = p.pso.palette.layout[2]
	
	for i = 1, PSO.info.MAXPALSIZE
		if not PSO.lib.slotHasWeapon(pal[i])
			pal[i] = nil
		end
	end
	
	-- if we set that to nil, don't do it.
	pal[p.pso.palette.aweaponnum] = $ or {}
end

-- hack for AIS
local aisweapon = {
	weapontype = WPT_SWORD,
	atk = 0,
	sprite = SPR_NULL,
	frame = A,
	ref = "ais",	-- see sword definitions
	pa_damage = 100,
	pa_cost = 100,
	range = 2048,
	actions = {"ais_shot", "ais_sword", "ais_tblast", "ais_shield"}
}

-- get the weapon we're displaying the actions of.
PSO.lib.getPlayerHUDWeapon = function(p)
	
	if p.pso.ais
		return aisweapon
	end
	
	local slot = 1
	local pal = p.pso.palette.layout[2][p.pso.palette.aweaponnum]
	local onehanded = false
	
	for i = 1, 2
		if PSO.lib.slotIsOneHanded(pal)
			onehanded = true
		end
	end
		
	if onehanded and (not p.pso.backpalette or not pal[1])
	and pal[2]	
		slot = 2	-- by default, use right handed weapon
	end
	
	return pal and pal[slot] or nil
end

-- get the player's active weapon.
-- similar to HUD weapon but is locked to the action the player is currently doing
PSO.lib.getPlayerActiveWeapon = function(p)
	
	if p.pso.ais
		-- for ais, hack things in!
		return aisweapon
	end
	
	local slot = 1
	local pal = p.pso.palette.layout[2][p.pso.palette.aweaponnum]
	local onehanded = false
	
	if p.pso.actionslot
		slot = p.pso.actionslot
	else
		for i = 1, 2
			if PSO.lib.slotIsOneHanded(pal)
				onehanded = true
			end
		end
		
		if onehanded and (not p.pso.backpalette or not pal[1])
		and pal[2]	
			slot = 2	-- by default, use right handed weapon
		end
	end
	
	return pal and pal[slot] or nil
end

-- returns true if the selected weapon has the right potential
PSO.lib.weaponPotential = function(w, p)
	return PSO.lib.itemInfo(w).potential == p
end

-- returns true if the player's active weapon has the potential
PSO.lib.activeWeaponPotential = function(p, pot)
	local w = PSO.lib.getPlayerActiveWeapon(p)
	if w
		return PSO.lib.weaponPotential(w, pot)
	end
end

-- returns true if the player's passive weapon has the potential (only works for 2h)
PSO.lib.passiveWeaponPotential = function(p, pot)
	if p.pso.ais
		-- for ais, hack things in!
		return false
	end
	
	local slot = 2
	local pal = p.pso.palette.layout[2][p.pso.palette.aweaponnum]
	local onehanded = false
	
	if p.pso.actionslot
		slot = p.pso.actionslot == 1 and 2 or 1
	else
		for i = 1, 2
			if PSO.lib.slotIsOneHanded(pal)
				onehanded = true
			end
		end
		
		if onehanded and (not p.pso.backpalette or not pal[1])
		and pal[2]	
			slot = 1	-- by default, use right handed weapon
		end
	end
	
	local w = pal and pal[slot] or nil	
	
	if w
		return PSO.lib.weaponPotential(w, pot)
	end
end

-- returns true if any weapon on the currently selected palette has the potential
PSO.lib.activeOrPassivePotential = function(p, pot)
	return PSO.lib.activeWeaponPotential(p, pot) or PSO.lib.passiveWeaponPotential(p, pot)
end

-- returns true if any weapon equipped on the player's palette has the potential
PSO.lib.paletteWeaponPotential = function(p, pot)
	if p.pso.ais
		-- for ais, hack things in!
		return false
	end
	
	local slot = 1
	local pal = p.pso.palette.layout[2]	--[p.pso.palette.aweaponnum]
	local onehanded = false
	
	for i = 1, PSO.info.MAXPALSIZE
		local wpal = pal[i]
		
		if wpal
			for j = 1, 2
				local w = wpal[j]
				if w and PSO.lib.weaponPotential(w, pot)
					return true
				end
			end
		end
	end
end

PSO.lib.slotIsOneHanded = function(slot)
	return slot and (slot[2] or (slot[1] and PSO.info.weaponHand[slot[1].weapontype] ~= HAND_BOTH))
end

-- returns true if the slot has at least 1 weapon
PSO.lib.slotHasWeapon = function(slot)
	return slot and (slot[1] or slot[2])
end

PSO.lib.getPlayerArmor = function(p)
	if p.pso.ais return nil end
	return p.pso.palette.layout[3][p.pso.palette.armornum]
end

PSO.lib.resetPlayerAction = function(p)
	
	if p.pso.action
		--p.pso.artsavenger = nil	-- HU skill
		p.pso.artsavenger_on = nil	-- HU skill	
		
		PSO.lib.runHook("playerActionReset", p)
	end
	
	-- 
	if p.pso.ais
		p.pso.aislockon = {}	-- targetting
		p.pso.aisreflect = nil	-- shield	
	end
	
	-- fo skills
	if p.pso.taction
		local tech = PSO.info.photonArts[p.pso.taction]
		local elem = tech.element
		p.pso.lastelement = elem
	end
	
	p.pso.action = nil
	p.pso.actionslot = nil
	p.pso.taction = nil
	p.pso.actionbutton = nil
	p.pso.actionweapon = nil
	p.pso.actiontimer = 0
	p.pso.actionheld = false
	p.pso.actionhold = 0
	p.pso.actionnum = 0
	p.pso.actionspeed = nil
	p.pso.actioncombo = 0	
	
	p.pso.superarmor = nil
	p.pso.springbuffer = nil
	p.pso.mitigations = nil
end

-- resets player actions etc.
-- for use when teleporting / between maps.
PSO.lib.softResetPlayer = function(p)
	
	p.pso.teleport = nil
	p.pso.teleporttarget = nil
	p.pso.quickpalette = nil
	
	p.pso.jumppad = nil
	p.pflags = $ & ~(PF_NOCLIP)
	
	if p.mo and p.mo.valid
		p.thrustfactor = skins[p.mo.skin].thrustfactor
	end
	
	--PSO.lib.removeStatus(p)
	PSO.lib.resetPlayerAction(p)
	--PSO.lib.playerResetDPS(p)
	
	p.pso.pittime = nil
	p.pso.pitcoords = nil
	p.pso.backuppitcoords = nil
end

-- fully resets player stuff.
-- intended for use bwteen maps, revival, etc...
PSO.lib.resetPlayer = function(p)
	
	PSO.lib.playerDeactivateAIS(p, false, true)	-- this cannot go in mapchange or the game SHITS ITSELF
	PSO.lib.softResetPlayer(p)
	PSO.lib.removeStatus(p)
	PSO.lib.playerResetDPS(p)
	
	p.pso.pottimer = 0
	p.pso.levelsynch = nil	-- remove level synch
	
	p.pso.pb = 0
	p.pso.usepb = 0
	
	p.pso.injury = 0
	p.pso.huntergale = 0
	p.pso.huntergaletimer = 0
	p.pso.hunterswordbreaker = 0
	p.pso.artsavenger = 0
	
	p.pso.lastelement = nil
	p.pso.chainreactor = nil
	p.pso.lastturret = nil
	
	p.pso.iframes = nil
	p.pso.inboss = nil
	
	p.pso.credits = nil
	
	PSO.lib.playerCalcStats(p, p.pso.level)
	
	if gamestate == GS_LEVEL
		p.mo.colorized = false
	end
	
	PSO.lib.runHook("playerReset", p)
end

local function calculateQuestScore(p)
	if server.questdata
	and p.pso.quest
	and gamemap == server.questdata.map	
		local q = PSO.lib.questInfo()	--PSO.info.quests[gamemap][server.questdata.index]
		local total = 0
		local totalgrade = 0
		local diff = PSO.lib.questDifficulty()
		
		p.pso.quest.grades = {}
		
		if q.grading
			
			for i = 1, #q.grading do
				local playerval = 0
				local grade = 0
				local g = q.grading[i]	
				local gtype = g[1]
				
				p.pso.quest.grades[i] = {}	-- {player_val, score_val}
				
				if gtype == QG_TIME
					
					local diffmult = 100
					
					if diff >= DIFF_VERYHARD
						diffmult = 160
					elseif diff >= DIFF_HARD
						diffmult = 130
					end
					
					local full = g[2]
					local low = g[3]
					local mintime = g[4]*diffmult/100
					local maxtime = g[5]*diffmult/100
					
					if p.pso.quest.time < mintime
						grade = full
					elseif p.pso.quest.time > maxtime
						grade = low
					else
						local normal = FixedDiv((p.pso.quest.time - mintime)*FRACUNIT, (maxtime - mintime)*FRACUNIT)
						-- and then we use black magic also known as more maths:
						grade = (normal*low + (FRACUNIT - normal)*full) /FRACUNIT
					end
					
					total = $ + full
					playerval = p.pso.quest.time
					
					--print("TIME: "..grade.."/"..full)
					
				elseif gtype == QG_KILLS
					
					local full = g[2]
					local low = 0
					local minkills = 0
					local maxkills = g[3]
					local checktype = g[4]
					
					-- count kills
					local kills = 0
					for k,v in pairs(p.pso.kills)
						if not checktype or checktype == k
							kills = $ + v
						end
					end
					
					if kills > maxkills
						grade = full
					elseif not kills
						grade = low
					else
						local normal = FixedDiv((kills - minkills)*FRACUNIT, (maxkills - minkills)*FRACUNIT)
						-- and then we use black magic also known as more maths:
						grade = (normal*low + (FRACUNIT - normal)*full) /FRACUNIT
					end
					
					total = $ + full
					playerval = kills
					
					--print("KILLS: "..grade.."/"..full)
				
				elseif gtype == QG_NOINCAP
					grade = (not p.pso.quest.incapacitations) and g[2] or 0
					playerval = p.pso.quest.incapacitations and 1 or 0
					total = $ + g[2]
					
					--print("INCAPACITATIONS: "..grade.."/"..g[2])
					
				elseif gtype == QG_DAMAGETAKEN

					local diffmult = 100
					
					if diff >= DIFF_VERYHARD
						diffmult = 1200
					elseif diff >= DIFF_HARD
						diffmult = 500
					end
				
					local full = g[2]
					local low = g[3]
					local mindamage = g[4]*diffmult/100
					local maxdamage = g[5]*diffmult/100
					
					if p.pso.quest.damagetaken < mindamage
						grade = full
					elseif p.pso.quest.damagetaken > maxdamage
						grade = low
					else
						local normal = FixedDiv((p.pso.quest.damagetaken - mindamage)*FRACUNIT, (maxdamage - mindamage)*FRACUNIT)
						--print(normal)
						-- and then we use black magic also known as more maths:
						grade = (normal*low + (FRACUNIT - normal)*full) /FRACUNIT
					end
					
					total = $ + full
					playerval = p.pso.quest.damagetaken
					
					--print("DAMAGETAKEN: "..grade.."/"..full)
				end
				totalgrade = $ + grade
				p.pso.quest.grades[i] = {playerval, grade}
			end

		/*grading = {
			{QG_TIME, full_point_time, 0_point_time, full_time, low_time},
			{QG_KILLS, max_points, max_kills, [MT_ type]},
			{QG_NOINCAP, points_for_no_incapacitation},
			{QG_DAMAGETAKEN, full_point_damagetaken, 0_point_damagetaken, low_damage_taken, high_damage_taken},
		}*/		
		end
		
		-- set our total rank:
		local percentage = FixedDiv(totalgrade<<FRACBITS, total<<FRACBITS)
		local gradelookup = {
			
			[RANK_S] = FRACUNIT,
			[RANK_A] = FRACUNIT*8/10,
			[RANK_B] = FRACUNIT*4/10,
			[RANK_C] = 0
		}
		
		for i = 1, #gradelookup
			if percentage >= gradelookup[i]
				p.pso.quest.ranking = i
				break
			end
		end
		
		p.pso.quest.totals = {totalgrade, total}
		p.pso.quest.typepoints = totalgrade	-- add difficulty scalar
		
		if p.pso.quest.ranking == RANK_S
			p.pso.quest.typepoints = $*2
		elseif p.pso.quest.ranking == RANK_A
			p.pso.quest.typepoints = $*3/2
		end
		
		if diff >= DIFF_HARD
			p.pso.quest.typepoints = $*5/2	-- x2.5 in hard
		end
		
		p.pso.typepoints[TYPE_HUNTER] = $+p.pso.quest.typepoints
		p.pso.typepoints[TYPE_RANGER] = $+p.pso.quest.typepoints
		p.pso.typepoints[TYPE_FORCE] = $+p.pso.quest.typepoints
		
		--print("GRADE: "..totalgrade.."/"..total)
	end
end

PSO.lib.clearQuest = function()
	
	if not server.questdata return end
	local q = PSO.lib.questInfo()
	
	server.questdata.complete = true
	
	if not q.noquestcompletemusic
		S_ChangeMusic("QCLEAR", true)
	end
	
	for p in players.iterate
		
		if p.deadtimer or p.pso.hp <= 0
			PSO.lib.revivePlayer(p, p, p.pso.maxhp)
		end
		
		-- disengage AIS on quest clear
		if p.pso.ais
			PSO.lib.playerDeactivateAIS(p, true)
		end
		
		calculateQuestScore(p)
		p.pso.quest.completetimer = 1
		PSO.lib.playerForceInputUp(p, IPT_CONFIRM)
		PSO.lib.rollPlayerShop(p)	-- reroll the player's shop
		
		-- if the quest has rewards for the current difficulty, award them now
		if q.reward
			local r = q.reward[PSO.lib.questDifficulty()]
			if r and r[1]
				local qty = r[2] or 1
				if not PSO.lib.playerGetItem(p, r[1], qty)	-- if we can't hold it, drop on the floor.
					
					local drop = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_PSO_DROP)
					drop.psoitem = r[1]
					drop.psoqty = qty
				end
			end
		end
	end
	
	-- despawn all enemies
	for m in mobjs.iterate() do
		if m and m.valid and m.flags & MF_ENEMY
			P_RemoveMobj(m)
		end
	end
	
	if q.unlocks
		for i = 1, #q.unlocks
			local unlock = q.unlocks[i]
			
			local mapn = unlock[1]
			local qstn = unlock[2]
			local diff = unlock[3] or server.questdata.difficulty
			local diffreq = unlock[4] or DIFF_NORMAL
			
			if PSO.lib.questDifficulty() < diffreq
				continue
			end
			
			if not PSO.lib.questUnlocked(mapn, qstn)
			or PSO.lib.questUnlocked(mapn, qstn) < diff
				PSO.lib.unlockQuest(mapn, qstn, diff)
			end
			
			-- always do that regardless.
			PSO.lib.localUnlockQuest(mapn, qstn, diff)
		end
	end
	
	S_StartSound(nil, sfx_rnkup)
end

PSO.lib.noPPNotif = function(p)
	
	S_StartSound(nil, sfx_mnot, p)
	
	if not p.pso.nopp
		p.pso.nopp = TICRATE
	end
end

-- sets a notification on the player's screen
PSO.lib.playerNotification = function(p, msg)
	S_StartSound(nil, sfx_mopen, p)
	p.pso.notification = {
		timer = TICRATE*4,
		text = msg,
	}
end

-- sends a notification to the side of the player's screen
PSO.lib.playerPushNotification = function(p, msg, nlines)
	
	S_StartSound(nil, sfx_radio, p)
	p.pso.slidenotifications = $ or {}
	
	p.pso.slidenotifications[#p.pso.slidenotifications+1] = {
		timer = TICRATE*4,
		slidetimer = TICRATE/3,
		text = msg,
		nlines = nlines or 1,
	}
	
end

-- prunes redundant data from items to make saves smaller
-- here's all we need to know from a weapon:
/*
	{"ref", grind, maxgrind, element, elementv, lv_req, base_pa_dmg, base_pa_cost, pa_dmg, pa_cost, [PAs], unidentified?, locked?}
*/

local function prune_item_data(t)
	for i = 1, #t
		local slot = t[i][1]
		if type(slot) ~= "string"
			
			if slot.type == IT_WEAPON
				--print("serializing weapon...")
				local new = {
					slot.type,
					slot.ref,
					slot.grind or 0,
					slot.maxgrind or PSO.info.MAXGRIND,
					slot.element or ELEM_NONE,
					slot.elementv or 0,
					slot.equiplevel or 0,
					slot.actions or {},
					slot.unidentified or 0,
					slot.locked or 0,
					
				}
				
				t[i] = {new}
			elseif slot.type == IT_ARMOR
				--print("serializing armour...")
				local new = {
					slot.type,
					slot.ref,
					slot.grind or 0,
					slot.maxgrind or PSO.info.MAXGRIND,
					slot.element or ELEM_NONE,
					slot.elementv or 0,
					slot.equiplevel or 0,
					slot.slots or {},
					slot.locked or 0,
				}
				
				t[i] = {new}
			
			elseif slot.type == IT_UNIT
				--print("serializing unit...")
				local new = {slot.type, slot.ref, slot.locked or 0}	-- yep that's all we don't need ANYTHING ELSE.
				
				t[i] = {new}
			end
			
		end
	end
end

local function read_item_data(t)
	for i = 1, #t
		local slot = t[i][1]
		if type(slot) ~= "string"
			
			--print("read slot "..i)
			local refslot = 2
			local ittype
			
			if slot[1] and type(slot[1]) == "string"
				refslot = 1
			else
				ittype = slot[1]
			end
			local shift = refslot-1
			
			if refslot
				
				local re
				if PSO.info.items[slot[refslot]]
					re = PSO.lib.makeItem(slot[refslot])
				else
					re = {ref=slot[refslot], rarity=1}
				end
				
				if not ittype	-- type wasn't saved...
					ittype = re.type
				end
				
				if ittype == IT_WEAPON
					--print("remaking weapon "..slot[1])
					
					re.type = ittype
					re.weapontype = $ or WPT_SWORD	-- we have to deal with dumb shit all the time, haha.
					re.grind = slot[2+shift]
					re.maxgrind = slot[3+shift]
					re.element = slot[4+shift]
					re.elementv = slot[5+shift]
					if slot[6+shift]
						re.equiplevel = slot[6+shift]
					end
					
					if #slot > 10	-- old weapon save
						shift = $ + (11-6) -1
					end
					
					re.pa_damage = $ or 100
					re.pa_cost = $ or 100	-- for modded items.
					-- because we used to save that garbage:
					--re.base_pa_damage = slot[7]
					--re.base_pa_cost = slot[8]
					--re.pa_damage = slot[9]
					--re.pa_cost = slot[10]
					re.actions = slot[7+shift]
					re.unidentified = slot[8+shift] and true or false
					re.locked = slot[9+shift] and true or false
					
					t[i] = {re, 1}	-- weapon is remade!
				elseif ittype == IT_ARMOR
					--print("remaking armour "..slot[1])
					re.type = ittype
					re.grind = slot[2+shift]
					re.maxgrind = slot[3+shift]
					re.element = slot[4+shift]
					re.elementv = slot[5+shift]
					if slot[6+shift]
						re.equiplevel = slot[6+shift]
					end
					re.slots = slot[7+shift]
					re.locked = slot[8+shift] and true or false
					
					t[i] = {re, 1}	-- armour is remade!
				
				elseif ittype == IT_UNIT
					re.type = ittype
					re.locked = slot[2+shift] and true or false
					t[i] = {re, 1}	-- yep that's also it LOL
				end
			end
		end
	end
	
	return t
end

-- jsonify the player's important variables...
PSO.lib.serializePlayer = function(p)
	
	if not p or not p.valid or not p.pso return end
	
	local d = p.pso
	local tabl = {
		
		-- whether the player has passed the first type selection check.
		type_initialized = d.type_initialized,
		
		-- core stats
		skin = skins[p.skin].name,	-- for some reason p.mo.skin is inaccessible outside of GS_LEVEL because this game sucks.
		colour = p.skincolor,
		type = d.type,
		level = d.level,
		exp = d.exp,
		playtime = d.playtime,
		
		-- money
		meseta = d.meseta,
		
		-- items
		inventory = PSO.lib.copyTable(d.inventory),
		storage = PSO.lib.copyTable(d.storage),	-- will not need any fix because you can't put shit with equipment in there
		
		-- type info
		typepoints = PSO.lib.copyTable(d.typepoints),
		weaponproficiencies = PSO.lib.copyTable(d.weaponproficiencies),
		typeskills = PSO.lib.copyTable(d.typeskills),
		--maxskillpoints = d.maxskillpoints,
		
		-- PAs
		photonarts = PSO.lib.compactPow2Table(d.photonarts),	-- (this drops the .techniques attribute)
		techniques = PSO.lib.copyTable(d.photonarts.techniques),	-- techs are a single value, we must separate it from PAs here
		-- otherwise we'll end up with a mixed table
		
		-- Palette
		palette = {}	--PSO.lib.copyTable(d.palette)	-- we don't actually need to save this, just the layout.
	}
	
	tabl.palette.layout = {{}, {}, {}}
	
	-- we can't serialize palette shit as-is so we're gonna do some hacking and save object # instead of references
	for i = 2, 3
		local pal = p.pso.palette.layout[i]
		for j = 1, PSO.info.MAXPALSIZE
			
			if pal[j]
				if i == 2	-- weapons
					tabl.palette.layout[i][j] = {0, 0}
					for k = 1, 2
						if pal[j][k]
							local num = PSO.lib.findPlayerItem(p, pal[j][k])
							tabl.palette.layout[i][j][k] = num or 0
						else
							--print("j: "..j.." k: "..k)
							tabl.palette.layout[i][j][k] = 0	-- save a 0 instead of nil.
							-- this avoids errors if you have, say, only a saber equipped in a slot.
						end
					end
					
				else
					
					local num = PSO.lib.findPlayerItem(p, pal[j])
					tabl.palette.layout[i][j] = num or 0
				end
			else
				if i == 2
					tabl.palette.layout[i][j] = {0, 0}
				else
					tabl.palette.layout[i][j] = 0
				end
			end
		end
	end
	
	-- also check for empty spaces in the item palette,......
	for i = 1, PSO.info.MAXPALSIZE
		local pal = p.pso.palette.layout[1]
		
		if not pal[i]
			tabl.palette.layout[1][i] = 0
		else
			tabl.palette.layout[1][i] = pal[i]
		end
	end	
	
	-- serialize units and weapon actions properly.
	-- there can be gap in the setup so we must account for them
	-- or the json will error.
	local tabls = {"inventory", "storage"}
	
	for tn = 1, #tabls
		local t = tabls[tn]
		for i = 1, #p.pso[t]
			local it = p.pso[t][i][1]
			
			if type(it) == "table"
			
				if it.type == IT_ARMOR
					local info = PSO.lib.itemInfo(it)
					if info.slots
						for j = 1, info.slots
							if it.slots[j]
								
								local slotnum = PSO.lib.findPlayerItem(p, it.slots[j])
								tabl[t][i][1].slots[j] = slotnum
							else
								it.slots[j] = 0
							end
						end
					end
				
				elseif it.type == IT_WEAPON
					
					local num_actions = (PSO.info.weaponHand[it.weapontype] == HAND_BOTH) and 6 or 3
					for j = 1, num_actions
						if not it.actions[j]
							tabl[t][i][1].actions[j] = 0
						end
					end
				end
			end
		end
	end
	
	prune_item_data(tabl.inventory)	-- prune unecessary weapon/armor data!!
	prune_item_data(tabl.storage)	-- prune unecessary weapon/armor data!!
	
	-- proficiencies is a sparse table, fix that
	for i = 1, #tabl.weaponproficiencies
		tabl.weaponproficiencies[i] = PSO.lib.compactPow2Table(tabl.weaponproficiencies[i])
	end
	
	-- serialize loadout stuff...
	/*
		we store item indexes for the weapons/armour/units.
		if there is no item ot be stored, store 0
		if the item we're trying to store doesn't exist, store as -1
		
		if it does, store positive numbers for items within the inventory
		if it is in storage, multiply the value by -1 then add -1.
		(storage values are negative, starting from -2, since -1 is our invalid item indicator)
	*/
	
	local function loadoutItemNum(item)
		local slot = PSO.lib.findPlayerItem(p, item)
		if not slot
			slot = PSO.lib.findPlayerStorageItem(p, item)
			if slot
				slot = $*(-1) -1
			else
				slot = -1	-- can't find the damn thing
			end
		end
		
		return slot
	end
	
	if p.pso.loadouts
		local loadouts = {}
		for i = 1, PSO.info.MAXLOADOUTS
			loadouts[i] = {}
			
			local currloadout = loadouts[i]
			
			if p.pso.loadouts[i]
				currloadout[1] = p.pso.loadouts[i][1]
				currloadout[2] = PSO.lib.copyTable(p.pso.loadouts[i][2])
				-- for palette obviously we have some more problems...
				
				currloadout[3] = {{}, {}, {}}
				for j = 1, PSO.info.MAXPALSIZE
					
					-- item palette
					local itpal = p.pso.loadouts[i][3][1]
					if itpal[j]
						currloadout[3][1][j] = itpal[j]
					else
						currloadout[3][1][j] = 0
					end
					
					-- weapon palette
					local wppal = p.pso.loadouts[i][3][2]
					if wppal[j]
						currloadout[3][2][j] = {}
						for k = 1, 2
							if wppal[j][k]
								-- try to find the item...
								currloadout[3][2][j][k] = loadoutItemNum(wppal[j][k])
							else
								currloadout[3][2][j][k] = 0	-- nothing
							end
						end
					else
						currloadout[3][2][j] = 0	-- no weapon
					end
					
					-- armour and units
					local apal = p.pso.loadouts[i][3][3]
					if apal[j] and #apal[j]
						local armour = apal[j][1]
						local slot = loadoutItemNum(armour)
						
						currloadout[3][3][j] = {slot}
						
						if slot ~= -1	-- it exists!!
							-- do the same shit for units
							local slots = PSO.lib.itemInfo(armour).slots or 0
							for k = 1, slots
								local unit = apal[j][k+1]
			
								if unit
									local uslot = loadoutItemNum(unit)
									currloadout[3][3][j][k+1] = uslot
								end
							end
						end
					else
						currloadout[3][3][j] = 0	-- jack shit!
					end
				end
			end
		end
		-- end my fucking life
		--print("ENCODE LOADOUTS")
		--print(json.encode(loadouts))
		tabl.loadouts = loadouts
	end
	
	return json.encode(tabl)
end

PSO.lib.deserializePlayer = function(p, data)
	
	if not data or not data.type return end	-- this is fine too, happens with empty files
	
	if not p or not p.valid or not p.pso return end
	
	local t = p.pso
	t.type_initialized = data.type_initialized
	
	if data.skin and skins[data.skin]
		R_SetPlayerSkin(p, data.skin)
	end
	
	if data.colour and data.colour < #skincolors and skincolors[data.colour].accessible
		p.skincolor = data.colour
		p.mo.color = p.skincolor
	end
	
	t.playtime = data.playtime or 0
	t.type = data.type
	/*t.hp = data.hp
	t.maxhp = data.maxhp
	t.pp = data.pp
	t.maxpp = data.maxpp*/
	t.exp = data.exp
	t.level = data.level
	/*t.satk = data.satk
	t.ratk = data.ratk
	t.tatk = data.tatk
	t.def = data.def
	t.ppregen = data.ppregen*/
	
	t.meseta = (data.meseta or 5000)
	
	t.typepoints = data.typepoints
	t.weaponproficiencies = data.weaponproficiencies
	for i = 1, #t.weaponproficiencies	-- make it sparse again
		t.weaponproficiencies[i] = PSO.lib.sparsePow2Table(t.weaponproficiencies[i])
	end
	
	-- check to make sure we have proficiencies in every weapon!
	local types = {TYPE_HUNTER, TYPE_RANGER, TYPE_FORCE}
	
	for i = 1, #types
		local prof = t.weaponproficiencies[types[i]]
		for j = 1, #PSO.info.weaponOrder
			local wep = PSO.info.weaponOrder[j]
			if prof[wep] == nil
				prof[wep] = PSO.info.typestats[types[i]].weapons[wep]
			end
		end
	end
	
	t.photonarts = PSO.lib.sparsePow2Table(data.photonarts)
	t.photonarts["techniques"] = data.techniques
	
	-- conversion of < 1.10 saves:
	if type(t.photonarts.techniques) == "number"
		print("Converting pre 1.10 save...")
		for k,v in pairs(t.photonarts)
			
			print(k)
			local list = PSO.info.weaponPhotonArts[k]
			local newval = {}
			
			if list
				for i = 1, #list
					
					if v & 2^(i-1)
						newval[#newval+1] = list[i]
						--print(list[i])
					end
				end
			end
			
			t.photonarts[k] = newval
		end
	end
	
	-- conversion of < 2.0 saves
	-- PAs:
	for k,v in pairs(PSO.info.weaponDefaultActions)
		if not t.photonarts[k]
			t.photonarts[k] = {}
			for i = 1, #v
				table.insert(t.photonarts[k], v[i])
			end
		end
	end
	
	-- proficiency
	-- set type default proficiency if necessary
	for i = 1, #t.weaponproficiencies
		local ref = PSO.info.typestats[i].weapons
		local targ = t.weaponproficiencies[i]
		
		for k,v in pairs(ref)
			if targ[k] == nil
				targ[k] = ref[k]
			end
		end
	end
	
	t.typepoints = data.typepoints
	t.typeskills = data.typeskills
	-- fix my skill points and then check if i can have all of them skills.
	t.maxskillpoints = 4 + (min(PSO.info.LEVELCAP, p.pso.level)/10)
	PSO.lib.playerValidateTpeSkills(p)
	
	t.storage = read_item_data(data.storage)	-- this is fine as is, armor have units unequipped when they're sent to storage
	
	-- now the annoying parts:
	local armours_to_check = {}
	t.inventory = read_item_data(data.inventory)
	-- fix inventory refs: units & weapon actions
	local tabls = {"inventory", "storage"}
	
	for tn = 1, #tabls
		
		local t = tabls[tn]
		
		for i = 1, #p.pso[t]
			local it = p.pso[t][i][1]
			
			if type(it) == "table"
			
				if it.type == IT_ARMOR
					local info = PSO.lib.itemInfo(it)
					if info.slots
						for j = 1, info.slots
							if it.slots[j]
								it.slots[j] = p.pso[t][$][1]
								it.slots[j].equipped = true
								it.slots[j].equippedarmours = $ and $+1 or 1
								-- the item is already pre-marked as equipped
							else
								it.slots[j] = nil
							end
						end
					end
					PSO.lib.calcItemStat(it)	-- fix stats up if we change em between patches
					it.timestamp = PSO.info.INVENTORYSIZE * (tn-1) + i	-- apply timestamp for sorting if it doesnt exist yet.
					armours_to_check[#armours_to_check+1] = it
				
				elseif it.type == IT_WEAPON
					local num_actions = (PSO.info.weaponHand[it.weapontype] == HAND_BOTH) and 4 or 2
					for j = 1, num_actions
						if not it.actions[j]
							it.actions[j] = nil
						end
					end
					PSO.lib.calcItemStat(it)	-- fix stats up if we change em between patches				
					it.timestamp = PSO.info.INVENTORYSIZE * (tn-1) + i
					
				
				elseif it.type == IT_UNIT
					PSO.lib.reloadUnit(it)
					it.timestamp = PSO.info.INVENTORYSIZE * (tn-1) + i	
				end
			end
		end
	end
	
	t.palette.layout = data.palette.layout
	-- same deal here
	for i = 2, 3
		local pal = p.pso.palette.layout[i]
		for j = 1, PSO.info.MAXPALSIZE
			
			if pal[j]
				if i == 2	-- weapons
					
					for k = 1, 2
						if pal[j][k]
							p.pso.palette.layout[i][j][k] = p.pso.inventory[$][1]
							p.pso.palette.layout[i][j][k].equipped = true
						end
					end
					
				else
					p.pso.palette.layout[i][j] = p.pso.inventory[$][1]
					p.pso.palette.layout[i][j].equipped = true
				end
			else
				p.pso.palette.layout[i][j] = nil
			end
		end
	end
	
	-- deserialize loadout stuff
	/*
		reminder: 
		item = 0: not set
		item = -1: item was invalid at the time of saving
		item > 0	item is in inventory
		item < -1	item is in storage (-2 = storage slot 1)
	*/
	
	local function fetchLoadoutItem(itn)
		--print("Try to fetch item "..itn)
		if not itn
			return
		elseif itn == -1
			return itn
		elseif itn > 0
			return p.pso.inventory[itn][1]
		elseif itn < 0
			return p.pso.storage[abs(itn)-1][1]
		end
	end
	
	if data.loadouts
		p.pso.loadouts = {}
		for i = 1, PSO.info.MAXLOADOUTS
			local loadout = data.loadouts[i]
			if loadout and #loadout
				
				p.pso.loadouts[i] = {}
				local ploadout = p.pso.loadouts[i]
				
				-- these are trivially easy 
				ploadout[1] = loadout[1]
				ploadout[2] = PSO.lib.copyTable(loadout[2])
				
				-- this is VERY MUCH NOT AAAAAA
				ploadout[3] = {{}, {}, {}}
				for j = 1, PSO.info.MAXPALSIZE
					
					-- item palette
					local itpal = loadout[3][1]
					if itpal[j]
						ploadout[3][1][j] = itpal[j]
					end
					
					-- weapon palette
					local wpal = loadout[3][2]
					if wpal[j]
						ploadout[3][2][j] = {}
						for k = 1, 2
							local w = wpal[j][k]
							if w
								ploadout[3][2][j][k] = fetchLoadoutItem(w)
							end
						end
					end
					
					-- armour/units
					local apal = loadout[3][3]
					ploadout[3][3][j] = {}	-- necessary or the game shits the bed
					if apal[j]
						local armourn = apal[j][1]
						local armour = fetchLoadoutItem(armourn)
						ploadout[3][3][j] = {armour}
						
						if armourn and armourn ~= -1
							local slots = PSO.lib.itemInfo(armour).slots or 0
							
							for k = 1, slots
								
								local unit = apal[j][k+1]
								if unit
									ploadout[3][3][j][k+1] = fetchLoadoutItem(unit)
								end
							end
						end
					end
				end
				
				-- the loadout is now complete
				PSO.lib.playerValidateLoadoutItems(p, p.pso.loadouts[i])
			end			
		end
	end
	
	PSO.lib.playerSortInventory(p)
	PSO.lib.playerFixUnits(p)
	PSO.lib.playerFixWeaponPalette(p)
	PSO.lib.playerSortStorage(p)
	PSO.lib.rollPlayerShop(p)
	
	-- this recalculates stats
	-- useful if we change stats during a patch etc
	PSO.lib.playerChangeType(p, p.pso.type, true)
	p.loadsuccess = true
	
	-- check for units equipped on our armours, mostly to know if they CAN be equipped...
	for i = 1, #armours_to_check
		PSO.lib.validateUnits(armours_to_check[i])
	end
	
	p.pso.forceupdateweapon = true
	
	-- ver2 stuff....
	/*if not p.pso.photonarts[WPT_WAND] or not #p.pso.photonarts[WPT_WAND]
		p.pso.photonarts[WPT_WAND] = {"wand_normal"}
	end*/
	
end

-- returns true if the player has a pb gauge they can fill.
-- returns what PB skill they have.
PSO.lib.playerHasPB = function(p)
	
	if p.pso.ais return end
	
	for i = 1, #p.pso.typeskills[p.pso.type]
		local sk = p.pso.typeskills[p.pso.type][i]
		local skinfo = PSO.info.typeSkills[sk]
		if skinfo and skinfo.photonblast
			return sk
		end
	end
end

-- returns the type of the PB being used right now
PSO.lib.usingPB = function(p)
	if p.pso.usepb
		return PSO.lib.playerHasPB(p)
	else
		return false
	end
end

-- uses the PB if the gauge is full
PSO.lib.playerUsePB = function(p)
	
	local mo = p.mo
	local pb = PSO.lib.playerHasPB(p)
	
	if not pb
	or p.pso.pb < PSO.info.MAXPB
		return false
	end
	
	--print("Using PB")
	
	p.pso.usepb = true
	
	-- Hu PB activation: full heal
	if pb == "hu_overload"
		PSO.lib.healObject(p.mo, p.mo, p.pso.maxhp)
		S_StartSound(p.mo, sfx_hulb, p)

		for i = 1, 2
			for j = 1, 16
				
				local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
				local a = (360/16)*j*ANG1
				d.angle = a + ANG1*90
				d.renderflags = $|RF_PAPERSPRITE
				d.colorized = true
				d.color = SKINCOLOR_RED
				
				if i > 1
					d.spritexscale = $*2
				end
				
				d.destscale = $*4
				
				P_InstaThrust(d, a, 32*FRACUNIT*i)
				d.momx = $ + mo.momx
				d.momy = $ + mo.momy
			end
		end
		
		local s = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_THOK)
		s.sprite = SPR_FORC
		s.frame = (leveltime%20 /2) | FF_TRANS50
		s.colorized = true
		s.color = SKINCOLOR_CRIMSON
		s.tics = -1
		s.fuse = TICRATE/3
		s.scalespeed = FRACUNIT/2
		s.destscale = FRACUNIT*8
		
	elseif pb == "hu_berserk"
		p.pso.injury = 75
		S_StartSound(p.mo, sfx_hulb, p)

		for i = 1, 2
			for j = 1, 16
				
				local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
				local a = (360/16)*j*ANG1
				d.angle = a + ANG1*90
				d.renderflags = $|RF_PAPERSPRITE
				d.colorized = true
				d.color = SKINCOLOR_RED
				
				if i > 1
					d.spritexscale = $*2
				end
				
				d.destscale = $*4
				
				P_InstaThrust(d, a, 32*FRACUNIT*i)
				d.momx = $ + mo.momx
				d.momy = $ + mo.momy
			end
		end
		
	-- Ra PB activation: restores PP
	elseif pb == "ra_overclock"
		p.pso.pp = p.pso.maxpp
		S_StartSound(p.mo, sfx_raoc, p)
	
	-- load blight round
	elseif pb == "ra_blightround"
		p.pso.blightround = true
		S_StartSound(p.mo, sfx_rabr, p)
	
	-- Fo PB activation: shifta, deband, resta
	elseif pb == "fo_photonflare"
		
		S_StartSound(p.mo, sfx_fopf, p)
		
		local techs = {"shifta", "deband", "resta"}
		for i = 1, #techs
			local tech = PSO.info.photonArts[techs[i]]
			
			local msl = PSO.lib.spawnMissile(mo, tech.object, p.drawangle, 0, tech.potency, DMT_TECHNIQUE, 0, 0, 0)
			if tech.sound
				S_StartSound(mo, tech.sound)
			end
			
			tech.spawnfunc(msl)
		end
		
	elseif pb == "fo_phantomcaster"
		S_StartSound(p.mo, sfx_fopf, p)
		
		for i = 1, 4
			local g = P_SpawnGhostMobj(p.mo)
			g.destscale = p.mo.scale*8 / i
			g.scalespeed = $*4/i
			g.fuse = 14
		end
		
	end
	
	return true
end

PSO.lib.playerActivateAIS = function(p, flight)
	
	local mo = p.mo
	
	p.pso.ais = true
	p.pso.aisflight = flight
	p.pso.aislockon = {}	-- targetting
	p.pso.aisreflect = nil	-- shield
	
	--p.mo.scale = $*4
	
	p.pso.aissavepb = p.pso.pb
	PSO.lib.playerCalcStats(p, p.pso.level)	-- level doesn't really matter in the context of AIS.
	p.pso.hp = p.pso.maxhp
	p.pso.pp = p.pso.maxpp
	
	-- spawn ais wings
	mo.wings = {}
	for i = 1, 4
		
		local w = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		w.tics = -1
		w.fuse = -1
		w.sprite = SPR_BFLW
		w.frame = A|FF_FULLBRIGHT|FF_PAPERSPRITE|FF_ADD
		w.flags = $|MF_NOCLIPHEIGHT|MF_NOCLIP
		w.flags2 = $|MF2_DONTDRAW	-- wings are invisible by default unless necessary
		
		mo.wings[#mo.wings+1] = w
	end
	
	PSO.lib.removeStatus(mo)
end

PSO.lib.playerDeactivateAIS = function(p, visual, skipexp)
	
	if not p.pso.ais return end	-- nothing to do here
	local mo = p.mo
	
	p.pso.ais = nil
	p.pso.aisflight = nil
	p.pso.aislockon = {}
	p.pso.aisreflect = nil
	p.pso.aistarget = nil

	
	PSO.lib.playerCalcStats(p, p.pso.level)
	p.pso.hp = p.pso.maxhp
	p.pso.pp = p.pso.maxpp
	p.pso.pb = p.pso.aissavepb or 0
	p.pso.aissavepb = nil
	
	if gamestate == GS_LEVEL
		if p.pso.aisexp and not skipexp and not PSO.lib.arenaQuestActive()
			PSO.lib.givePlayerExp(p, p.pso.aisexp)
			local n = PSO.lib.spawnDamageNumber(mo.x, mo.y, mo.z, p.pso.aisexp, 1)
			if p ~= consoleplayer
				n.flags2 = $|MF2_DONTDRAW
			end
			p.pso.aisexp = nil
		end
	end
	
	if gamestate == GS_LEVEL
		local vis = mo.aisvis
		if vis and vis.valid and visual
			local g = P_SpawnGhostMobj(vis)
			g.colorized = true
			g.color = SKINCOLOR_MASTER
			g.destscale = $*8
			g.scalespeed = p.mo.scale
			P_RemoveMobj(vis)
		end
		mo.aisvis = nil
		
		if mo.wings and visual
			for i = 1, #mo.wings
				
				local w = mo.wings[i]
				if w and w.valid
					w.fuse = 2

					local g = P_SpawnGhostMobj(w)
					g.destscale = $*8
					g.scalespeed = p.mo.scale
					
				end
			end
		end
		
		mo.wings = nil
		
		if visual
			mo.z = $ + mo.height/2
			mo.state = S_PLAY_ROLL
			mo.flags2 = $ & ~MF2_DONTDRAW
			if mo.followmobj and mo.followmobj.valid
				mo.followmobj.flags2 = $ & ~MF2_DONTDRAW
			end
			mo.momx = 0
			mo.momy = 0
			mo.momz = 0
			
			mo.height = mobjinfo[MT_PLAYER].height
			mo.radius = mobjinfo[MT_PLAYER].radius
			p.height = mo.height
		end
	end
	p.pso.forceupdateweapon = true
end

PSO.lib.playerStartAIS = function(p, flight)
	p.pso.aisusetimer = 1
	p.pso.aisuseflight = flight
end

PSO.lib.playerResetDPS = function(p)
	p.pso.dps_damage = 0
	p.pso.dps_time = 0
	p.pso.dps_display = 0
	p.pso.dps_resettimer = 0
end

freeslot("MT_HUNTERGALE", "S_HUNTERGALE")
mobjinfo[MT_HUNTERGALE] = {
	doomednum = -1,
	spawnstate = S_HUNTERGALE,
	spawnhealth = 1000,
	radius = 32<<FRACBITS,
	height = 32<<FRACBITS,
	flags = MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT,
}

function A_HunterGale(mo)
	--print("A")
	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end
	
	if not mo.set
		mo.dist = 16
		mo.scale = 1
		mo.destscale = mo.target.scale
		mo.scalespeed = FRACUNIT/15
		mo.fangle = P_RandomRange(0, 359)
		mo.flags2 = $ & ~MF2_DONTDRAW
		
		P_CreateFloorSpriteSlope(mo)
		mo.renderflags = $|RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD
		mo.floorspriteslope.o = {x=mo.target.x, y=mo.target.y, z=mo.target.z + mo.target.height/2}
		mo.floorspriteslope.xydirection = mo.angle
		mo.floorspriteslope.zangle = P_RandomRange(-20, 20)*ANG1
		
		mo.dir = P_RandomRange(0, 1) and 1 or -1
		
		mo.set = 1
	end
	
	local x = mo.target.x + mo.dist*cos(mo.fangle)
	local y = mo.target.y + mo.dist*sin(mo.fangle)
	P_MoveOrigin(mo, x, y, mo.target.z+mo.target.height/2)
	mo.rollangle = mo.fangle + ANG1*20
	
	mo.floorspriteslope.o = {x=mo.target.x, y=mo.target.y, z=mo.target.z + mo.target.height/2}
	
	if mo.fuse < 10
		mo.frame = $ & ~FF_TRANSMASK
		local trans = 5 + ((10-mo.fuse)/2)
		trans = max($, 5)
		trans = min($, 9)
		
		mo.frame = $|(trans<<FF_TRANSSHIFT)
	end
	
	mo.dist = $+2
	mo.fangle = $ + 20*ANG1*mo.dir
end

states[S_HUNTERGALE] = {SPR_HUWW, A|FF_ADD|FF_PAPERSPRITE|FF_FULLBRIGHT, 1, A_HunterGale, 0, 0, S_HUNTERGALE}

-- ranger's turret...
freeslot("MT_PSO_RANGERTURRET")
mobjinfo[MT_PSO_RANGERTURRET] = {
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 32<<FRACBITS,
	height = 32<<FRACBITS,
	flags = MF_NOCLIPTHING|MF_NOGRAVITY|MF_NOBLOCKMAP,
}

local stretches = {
	
	{FRACUNIT/8, FRACUNIT*8},
	{FRACUNIT/4, FRACUNIT*4},
	{FRACUNIT/2, FRACUNIT*2},
	{FRACUNIT, FRACUNIT},
}

addHook("MobjThinker", function(mo)
	
	mo.sprite = SPR_JETG
	mo.frame = (leveltime%4 /2)|FF_ADD
	mo.colorized = true
	mo.color = SKINCOLOR_CYAN
	
	mo.existtime = $ and $+1 or 1
	
	-- nice stretch when appearing/disappearing.
	local s = stretches[mo.existtime]
	
	if stretches[mo.fuse]
		s = stretches[mo.fuse]
	end
	
	if s
		mo.spritexscale = s[1]
		mo.spriteyscale = s[2]
	end
	
	if not mo.tracer or not mo.tracer.valid
		
		if mo.fuse > 6
			mo.fuse = 6
		end
		
		return
	end
	
	if mo.maxfiredelay
		
		mo.firedelay = $ and $+1 or 1

		-- check if our target player exists...
		-- and we have an enemy to fire at, too...
		local source = mo.tracer
		if not source or not source.valid
			return
		end
		
		if PSO.lib.passiveWeaponPotential(source.player, "remote guidance")
			mo.firedelay = $+2
		end
		
		if mo.firedelay > mo.maxfiredelay
		
			mo.firedelay = 0
			
			local target = mo.target
			if not target or not target.valid
				-- find a new one!
				local x1 = mo.x - FRACUNIT*1536
				local x2 = mo.x + FRACUNIT*1536
				local y1 = mo.y - FRACUNIT*1536
				local y2 = mo.y + FRACUNIT*1536
				
				local closestdist = FRACUNIT*1536
				local best
				
				searchBlockmap("objects", 
					function(ref, found)
						
						if found and found.valid and found.flags & MF_ENEMY
							local dist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
							if dist < closestdist
								best = found
								closestdist = dist
							end
						end
					end,
					
				mo, x1, x2, y1, y2)
				
				
				if best and best.valid
					mo.target = best
				end
			end
			
			if mo.target and mo.target.valid	-- second check here...
				-- fire the bullet!
				local potency = 20
				mo.angle = R_PointToAngle2(mo.x, mo.y, mo.target.x, mo.target.y)
				
				local msl = PSO.lib.spawnMissile(mo, MT_PSO_RIFLESHOT, mo.angle, 45*FRACUNIT, potency, DMT_RANGED, 0, 0, mo.scale*16, RT_NONE)
				msl.scale = $/3
				msl.color = mo.color
				msl.target = mo.tracer
							
				msl.pso.potency = potency	-- normals shouldn't benefit from PA%
												
				PSO.lib.aimMissile(msl, mo.target)			
			end
			
			
		end
	end
	
end, MT_PSO_RANGERTURRET)

-------------------------------------
-- ARENA QUEST STUFF
-------------------------------------


-- takes away anything "consumable" and returns a list of what was taken!
local function confiscateAllConsumables(p)

	local confiscated = {}
	
	local i = #p.pso.inventory
	while i
		local it = p.pso.inventory[i]
		local itinfo = PSO.lib.itemInfo(it[1])
		if itinfo.type == IT_CONSUMABLE
			--print(it[1])
			--print(i)
			confiscated[#confiscated+1] = {it[1], it[2]}
			PSO.lib.playerLoseItem(p, i, it[2])
		end
		
		i = $-1
	end
	
	return confiscated
end

-- sets up arena quest stuff for the player;
-- confiscates consumable items and backs them up.
PSO.lib.playerArenaQuestSetup = function(p)
	p.pso.arenaconfiscated = confiscateAllConsumables(p)
	p.pso.arenapoints = 2500
	p.pso.arenaexp = 0
	
	PSO.lib.rollPlayerShop(p, PSO.info.arenaShopList)
end

-- reverse process, give me back all my shit but take away the consumables I got during the quest!
PSO.lib.playerArenaQuestFinished = function(p)
	confiscateAllConsumables(p)
	
	if p.pso.arenaconfiscated
		for i = 1, #p.pso.arenaconfiscated
			--print("Give back "..p.pso.arenaconfiscated[i][1].." x"..p.pso.arenaconfiscated[i][2])
			PSO.lib.playerGetItem(p, p.pso.arenaconfiscated[i][1], p.pso.arenaconfiscated[i][2])
		end
	end

	-- give me my accumulated exp!
	if p.pso.arenaexp
		PSO.lib.givePlayerExp(p, p.pso.arenaexp)
		local n = PSO.lib.spawnDamageNumber(p.mo.x, p.mo.y, p.mo.z, p.pso.arenaexp, 1)
		if p ~= consoleplayer
			n.flags2 = $|MF2_DONTDRAW
		end
	end
	
	p.pso.arenapoints = nil
	p.pso.arenaexp = nil
	p.pso.arenaconfiscated = nil
	
	PSO.lib.rollPlayerShop(p)
end

-------------------------------------
-- ITEM LOADOUTS
-------------------------------------

-- creates a loadout based on the player's current equipment
/*
	What we need to generate a loadout:
	
	type,
	type skills,
	{item palette, weapon palette, armour palette*}
	
	(for each armour, we need to save the individual units!)
*/

PSO.lib.playerSaveLoadout = function(p)
	
	local itempalette = {}
	local weaponpalette = {}
	local armourpalette = {}
	
	local ap = p.pso.palette.layout[3]
	
	for i = 1, PSO.info.MAXPALSIZE
		
		armourpalette[i] = {}
		itempalette[i] = p.pso.palette.layout[1][i]
		
		-- weapon palette
		weaponpalette[i] = {}
		if p.pso.palette.layout[2][i]
			for j = 1, 2
				if p.pso.palette.layout[2][i][j]
					weaponpalette[i][j] = p.pso.palette.layout[2][i][j]
				end
			end
		end
		
		
		-- armour palette
		if ap[i]
			armourpalette[i] = {ap[i]}
			if PSO.lib.itemInfo(ap[i]).slots
				for j = 1, PSO.lib.itemInfo(ap[i]).slots
					armourpalette[i][j+1] = ap[i].slots[j] or 0
				end
			end
		end
	end
	
	local loadout = {
		
		p.pso.type,
		PSO.lib.copyTable(p.pso.typeskills[p.pso.type]),
		{
			itempalette,
			weaponpalette,
			armourpalette,
			--PSO.lib.copyTable(p.pso.palette.layout[3], true),
		},
	}
	
	return loadout
end

-- validates the player's type skills and removes them if something is wrong
PSO.lib.playerValidateTpeSkills = function(p)
	-- validate type skills
	local typs = {TYPE_HUNTER, TYPE_RANGER, TYPE_FORCE}
	for i = 1, #typs
		local skillpoints = p.pso.maxskillpoints
		local tt = typs[i]
		local j = #p.pso.typeskills[tt]
		while j
			local check = PSO.info.typeSkillsList[tt] or {}
			local found
			
			if check
				for k = 1, #check
					
					if p.pso.typeskills[tt][j] == check[k]
						found = true
						break
					end
				end
				
				if not found
					table.remove(p.pso.typeskills[tt], j)
					j = $-1
					continue
				end
			end
			
			local currskill = p.pso.typeskills[tt][j]
			
			-- check skill level
			local sk = PSO.info.typeSkills[currskill]
			
			if not sk
				print("\x82".."ANTI CHEAT: ".."\x80"..p.name..", Invalid skill: "..curskill..", removing.")
				table.remove(p.pso.typeskills[tt], j)
				j = $-1
				continue
			end
			skillpoints = $ - (sk.cost or 0)
			
			if skillpoints < 0	-- bruuuuh
				skillpoints = $ + sk.cost
				print("\x82".."ANTI CHEAT: ".."\x80"..p.name..", Too much SP used by "..sk.name..", removing.")
				table.remove(p.pso.typeskills[tt], j)
			elseif sk.level and p.pso.level < sk.level
				print("\x82".."ANTI CHEAT: ".."\x80"..p.name..", Player level too low for skill "..sk.name..", removing.")
				table.remove(p.pso.typeskills[tt], j)
			end
			j = $-1
		end
	end	
end

-- checks if we have every item in the loadout and sets invalid entries to -1
PSO.lib.playerValidateLoadoutItems = function(p, loadout)
	-- check for weapons
	if not loadout return end
	
	local pal = loadout[3]
	for i = 1, PSO.info.MAXPALSIZE
		local wpal = pal[2]
		local apal = pal[3]
		
		-- weapons
		if wpal[i]
			for j = 1, 2
				local w = wpal[i][j]
				if w
					if not PSO.lib.findPlayerItem(p, w) and not PSO.lib.findPlayerStorageItem(p, w) 
						wpal[i][j] = -1
					end	
				end
			end
		end
		
		-- armours
		if apal[i]
			local a = apal[i][1]
			
			if a 
				if not PSO.lib.findPlayerItem(p, a) and not PSO.lib.findPlayerStorageItem(p, a) 
					apal[i] = {-1}
				else	-- check units
					
					local slots = PSO.lib.itemInfo(a).slots or 0
					for j = 1, slots
						local unit = apal[i][j+1]
						
						if unit
							if not PSO.lib.findPlayerItem(p, unit) and not PSO.lib.findPlayerStorageItem(p, unit) 
								apal[i][j+1] = -1
							end
						end
					end
				end
			end
		end
	end	
end

-- sets a loadout onto the player
-- loadout is the loadout we're trying to use (duh)
-- sendstorage will send the currently equipped gear (weapons, armour, units) to storage if true.
PSO.lib.playerLoadLoadout = function(p, loadout, sendstorage)
	
	local success = true
	
	PSO.lib.playerChangeType(p, loadout[1], true)
	
	--set type skills
	p.pso.typeskills[p.pso.type] = PSO.lib.copyTable(loadout[2])
	PSO.lib.playerValidateTpeSkills(p)
	
	
	-- now set my palette stuff...
	local pal = loadout[3]
	p.pso.palette.layout[1] = PSO.lib.copyTable(pal[1])	-- this one is simple, we don't care if we actually have the items.
	--p.pso.palette.layout[2] = PSO.lib.copyTable(pal[2], true)	-- this one requires a bit more thought
	
	PSO.lib.dprint("====WEAPONS====")
	
	-- start with our weapons...
	for i = 1, PSO.info.MAXPALSIZE
		
		PSO.lib.dprint("Index "..i)
		
		if loadout[3][2][i]
			p.pso.palette.layout[2][i] = $ or {}	-- causes bugs otherwise!
		end
			
		if p.pso.palette.layout[2][i]
			for j = 1, 2
				PSO.lib.dprint("weaponslot "..j)
				if p.pso.palette.layout[2][i][j]
					
					-- unequip item
					p.pso.palette.layout[2][i][j].equipped = nil
					PSO.lib.dprint("Unequip weapon "..i..", "..j)
					
					-- try to send to storage (if storage is full, it will just stay in the inventory...)
					if sendstorage
						local itslot = PSO.lib.findPlayerItem(p, p.pso.palette.layout[2][i][j])
						PSO.lib.playerSendToStorage(p, itslot, 1)
						PSO.lib.dprint("Send weapon "..i..", "..j.." to storage")
					end
				end
				
				p.pso.palette.layout[2][i][j] = nil
				
				-- try to find the item we're supposed to replace this one with, IF there is any!
				
				if not loadout[3][2][i]
					PSO.lib.dprint("No weapon to replace at "..i)
					continue 
				end
				
				local replace = loadout[3][2][i][j]
				PSO.lib.dprint(replace)
				if replace and PSO.lib.canEquip(p, replace)	-- make sure the item can be equipped...
					-- search for the item in our inventory or storage
					if not PSO.lib.findPlayerItem(p, replace)	-- not in our inventory?
						PSO.lib.dprint("Replacement "..i..", "..j.." is in storage")
						local slot = PSO.lib.findPlayerStorageItem(p, replace)
							
						if not slot
							PSO.lib.dprint("Replacement "..i..", "..j.." could not be found")
							success = false
							continue	-- we cannot do anything here
						end
							
						-- if we have a slot, take it from our storage!
						if PSO.lib.playerRetrieveFromStorage(p, slot, 1)
							-- we succesfully got the item
							replace.equipped = true
							p.pso.palette.layout[2][i][j] = replace
							PSO.lib.dprint("Replacement "..i..", "..j.." succesfully equipped")
						else
							PSO.lib.dprint("Replacement "..i..", "..j.." could not be retrieved")
							success = false
							continue	-- the inventory is full? we can't do anything else.
						end
					else
						-- item is already in our inventory, we move.
						replace.equipped = true
						p.pso.palette.layout[2][i][j] = replace	
						PSO.lib.dprint("Replacement "..i..", "..j.." succesfully equipped")
					end
				else
					PSO.lib.dprint("Cannot equip / No weapon to replace at "..i..", "..j)
					if success
						success = replace ~= nil
					end
				end
			end
		end
	end
	
	PSO.lib.dprint("====ARMOUR====")
	
	-- and now armours AND units...
	for i = 1, PSO.info.MAXPALSIZE
		
		--if p.pso.palette.layout[3][i]

		if p.pso.palette.layout[3][i]
				
			local it = p.pso.palette.layout[3][i]
			-- unequip item and all of its units...
			it.equipped = nil
			PSO.lib.dprint("Unequip armour "..i)
				
			-- unequip units and send them to storage if requested and possible
			if PSO.lib.itemInfo(it).slots
				for j = 1, PSO.lib.itemInfo(it).slots
					PSO.lib.dprint("Unequip unit "..i..", "..j)
					local slot = it.slots[j]
						
					if slot
							
						slot.equippedarmours = $-1
						if not slot.equippedarmours
							slot.equipped = nil
							-- try to send the unit to storage
							if sendstorage
								local itslot = PSO.lib.findPlayerItem(p, slot)
								PSO.lib.playerSendToStorage(p, itslot, 1)		
								PSO.lib.dprint("Unit "..i..", "..j.." sent to storage")
							end
								
						end
					end
					it.slots[j] = nil
				end
			end
				
			-- try to send to storage (if storage is full, it will just stay in the inventory...)
			if sendstorage
				local itslot = PSO.lib.findPlayerItem(p, it)
				PSO.lib.playerSendToStorage(p, itslot, 1)
				PSO.lib.dprint("Armor "..i.." sent to storage")
			end
		end
			
		p.pso.palette.layout[3][i] = nil
			
		if not loadout[3][3][i]
			PSO.lib.dprint("No armour to replace at "..i)
			continue 
		end
			
		-- try to find the item we're supposed to replace this one with, IF there is any!
		local replace = loadout[3][3][i][1]
		if replace and PSO.lib.canEquip(p, replace)	-- make sure the armour can be put on...
			-- search for the item in our inventory or storage
			if not PSO.lib.findPlayerItem(p, replace)	-- not in our inventory?
				PSO.lib.dprint("Replacement "..i.." is in storage")
				local slot = PSO.lib.findPlayerStorageItem(p, replace)
							
				if not slot
					PSO.lib.dprint("Couldn't find replacement "..i)
					success = false
					continue	-- we cannot do anything here
				end
						
				-- if we have a slot, take it from our storage!
				if PSO.lib.playerRetrieveFromStorage(p, slot, 1)
					-- we succesfully got the item
					replace.equipped = true
					p.pso.palette.layout[3][i] = replace
					PSO.lib.dprint("Replacement "..i.." succesfully equipped")
				else
					PSO.lib.dprint("Could not withdraw replacement "..i)
					success = false
					continue	-- the inventory is full? we can't do anything else.
				end
			else
				-- item is already in our inventory, we move.
				PSO.lib.dprint("Replacement "..i.." succesfully equipped")
				replace.equipped = true
				p.pso.palette.layout[3][i] = replace					
			end
				
			-- if we're still here, then the item has been succesfully put in its place...
			-- however, we still need to get its units......
				
			for j = 2, #loadout[3][3][i]
				local unit = loadout[3][3][i][j]
					
				if unit
					PSO.lib.dprint("Fetching replacement unit "..i..", "..j)
					if not PSO.lib.findPlayerItem(p, unit)
						PSO.lib.dprint("Replacement "..i..", "..j.." is in storage")
						local slot = PSO.lib.findPlayerStorageItem(p, unit)
									
						if not slot
							PSO.lib.dprint("Couldn't find replacement unit "..i..", "..j)
							success = false
							continue	-- we cannot do anything here
						end
								
						-- if we have a slot, take it from our storage!
						if PSO.lib.playerRetrieveFromStorage(p, slot, 1)
							PSO.lib.dprint("Succesfully equipped replacement unit "..i..", "..j)
							-- we succesfully got the item
							unit.equipped = true
							unit.equippedarmours = $ and $+1 or 1
							replace.slots[j-1] = unit
						else
							PSO.lib.dprint("Couldn't withdraw replacement unit "..i..", "..j)
							success = false
							continue	-- the inventory is full? we can't do anything else.
						end							
							
					else	-- it's in our inventory
						PSO.lib.dprint("Succesfully equipped replacement unit "..i..", "..j)
						unit.equippedarmours = $ and $+1 or 1
						unit.equipped = true
						replace.slots[j-1] = unit
					end
					
				end
			end
		else
			if success
				success = PSO.lib.canEquip(replace)
			end
			PSO.lib.dprint("Cannot equip / No armour to replace at "..i)
		end
		--end
	end
	
	PSO.lib.playerCalcStats(p, p.pso.level)
	
	return success
end

-- loadout debug commands

COM_AddCommand("psod_saveloadout", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	local loadout = PSO.lib.playerSaveLoadout(p)
	
	local slot = tonumber(arg) or 1
	p.pso.loadouts = $ or {}
	p.pso.loadouts[slot] = loadout
	
	print("DBG: Loadout saved in slot "..slot)
	
end)

COM_AddCommand("psod_loadloadout", function(p, arg)	

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	local slot = tonumber(arg) or 1
	p.pso.loadouts = $ or {}
	
	print("DBG: Attempting to load loadout in slot "..slot)
	
	if p.pso.loadouts[slot]
		PSO.lib.playerLoadLoadout(p, p.pso.loadouts[slot], true)
	end
end)

-- commands

COM_AddCommand("psod_aisdebug", function(p, arg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	PSO.lib.playerActivateAIS(p, (arg and true or false))
	--PSO.lib.playerStartAIS(p, (arg and true or false))
end)

COM_AddCommand("psod_aisdismount", function(p)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	PSO.lib.playerDeactivateAIS(p, true)
end)

COM_AddCommand("psod_testnotif", function(p, msg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	PSO.lib.playerNotification(p, msg)
end)

COM_AddCommand("psod_testnotif2", function(p, msg)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	PSO.lib.playerPushNotification(p, msg)
end)

COM_AddCommand("psod_clearquest", function()

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	PSO.lib.clearQuest()
end)

COM_AddCommand("psod_getcard", function(p, n)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	n = tonumber($)
	server.keycards = $| 2^(n-1)
end)

COM_AddCommand("psod_setpb", function(p, arg)
	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	arg = tonumber($)
	p.pso.pb = min(PSO.info.MAXPB, arg)
	p.pso.pb = max(0, $)
	
end)

COM_AddCommand("psod_framedata", function(p, arg)
	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	p.pso.framedata = (not p.pso.framedata and true or false)
end)

COM_AddCommand("psod_realhp", function(p, arg)
	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	p.pso.showrealhp = (not p.pso.showrealhp and true or false)
end)


-- cheats for multiplayer
COM_AddCommand("psod_noclip", function(p, n)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	if p.pflags & PF_NOCLIP
		p.pflags = $ & ~PF_NOCLIP
	else
		p.pflags = $|PF_NOCLIP
	end
end)

COM_AddCommand("psod_god", function(p, n)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end

	if p.pflags & PF_GODMODE
		p.pflags = $ & ~PF_GODMODE
	else
		p.pflags = $|PF_GODMODE
	end
end)