-- Type Skill definitions

-- Worth noting that by the nature of type Skills their functionalities aren't implemented here but where they are relevant.
-- (eg: dodge roll extension goes in the dodge roll code where it checks if the player has the skill)

PSO.lib.addDebandToughness = function(p)
	if not p.pso.debandtoughness
		p.pso.debandtoughness = p.pso.maxhp/10
		p.pso.maxhp = $ + p.pso.debandtoughness
		p.pso.hp = $ + p.pso.debandtoughness
	end
end

PSO.lib.removeDebandToughness = function(p)
	if p.pso.debandtoughness
		p.pso.maxhp = $ - p.pso.debandtoughness
		p.pso.hp = $ - p.pso.debandtoughness
		p.pso.debandtoughness = nil
	end
end

PSO.info.MAXHUNTERGALE = 10000

PSO.info.ELEMBLOTRATE = 500	-- potency required for each stage of elemental blot
PSO.info.ELEMBLOTPB = {PSO.info.MAXPB/200, PSO.info.MAXPB/150, PSO.info.MAXPB/50}

PSO.info.typeSkills = {
	
	["all_automate"] = {
		name = "Auto-Mate Half",
		desc = "Automatically uses Mate items when your HP drops from above to below half.",
		cost = 1,
		level = 5,
		icon = "SK_MATE",
	},
	
	["all_hp1"] = {
		name = "HP Up",
		desc = "Increases Max HP by 50. Stacks with other Skills.",
		cost = 1,
		level = 10,
		icon = "SK_HP1",
		equip = function(p)
					p.pso.maxhp = $+50
				end,
		unequip = function(p)
					p.pso.maxhp = $-50
					p.pso.hp = min($, p.pso.maxhp)
				end,
	},

	["all_hp2"] = {
		name = "HP High Up",
		desc = "Increases Max HP by 75. Stacks with other Skills.",
		cost = 2,
		level = 50,
		icon = "SK_HP2",
		equip = function(p)
					p.pso.maxhp = $+75
				end,
		unequip = function(p)
					p.pso.maxhp = $-75
					p.pso.hp = min($, p.pso.maxhp)
				end,
	},
	
	["all_pp1"] = {
		name = "PP Up",
		desc = "Increases Max PP by 10. Stacks with other Skills.",
		cost = 1,
		level = 10,
		icon = "SK_PP1",
		equip = function(p)
					p.pso.maxpp = $+10
				end,
		unequip = function(p)
					p.pso.maxpp = $-10
					p.pso.pp = min($, p.pso.maxpp)
				end,		
	},

	["all_pp2"] = {
		name = "PP High Up",
		desc = "Increases Max PP by 20. Stacks with other Skills.",
		cost = 2,
		level = 50,
		icon = "SK_PP2",
		equip = function(p)
					p.pso.maxpp = $+20
				end,
		unequip = function(p)
					p.pso.maxpp = $-20
					p.pso.pp = min($, p.pso.maxpp)
				end,		
	},
	
	["all_dodgeextension"] = {
		name = "Dodge Master",
		desc = "Doubles the Dodge Roll's invulnerability period.",
		cost = 2,
		level = 1,
		icon = "SK_DGE_1",
	},

	["all_justreversal"] = {
		name = "Just Reversal",
		desc = "Enables the use of Dodge Rolls while in a knockback state at the cost of PP.",
		cost = 1,
		level = 40,
		icon = "SK_DGE_2",
	},
	
	-- Class specific Skills
	-- HUNTER
	["hu_ironwill"] = {
		name = "Iron Will",
		desc = "Survive a fatal attack with 1 HP once per quest.",
		cost = 2,
		level = 5,
		icon = "SK_HU_1",
	},

	["hu_hunterphysique"] = {
		name = "Hunter Physique",
		desc = "Reduces damage taken by 5% and ignores stagger effects.",
		cost = 1,
		level = 10,
		icon = "SK_HU_2",
	},

	["hu_hatefulaggression"] = {
		name = "War Brave",
		desc = "Increases aggro generation by 50%. Also increases damage dealt by 5% when attacking an enemy targetting you.",
		cost = 1,
		level = 60,
		icon = "SK_HU_3",
	},

	["hu_artsavenger"] = {
		name = "Arts Avenger",
		desc = "Using a Melee Photon Art or releasing a charged Photon Art right before taking damage negates the hit, increases damage by 33% and grants brief invulnerability. This state can also be stored for the next Photon Art by evading an attack with a Dodge Roll instead.",
		cost = 1,
		level = 15,
		icon = "SK_HU_4",
	},

	["hu_heavyarts"] = {
		name = "Heavy Arts",
		desc = "Gain knockback immunity to most attacks while performing Photon Arts.",
		cost = 2,
		level = 35,
		icon = "SK_HU_5",
	},

	["hu_reapingscythe"] = {
		name = "Reaping Scythe",
		desc = "Recover 10% HP when an enemy dies near you.",
		cost = 1,
		level = 60,
		icon = "SK_HU_6",
	},

	["hu_chainadrenaline"] = {
		name = "Chain Adrenaline",
		desc = "Gain Shifta and Deband on Chain detonation based on Chain and player counts, increasing PP recovery and recovering HP on Chain detonations.",
		cost = 1,
		level = 30,
		icon = "SK_HU_7",
	},

	["hu_overload"] = {
		name = "Stalwart Spirit",
		desc = "Photon Blast. Fully restores HP. Melee damage +50% and Damage resistance +30% for 20s.",
		cost = 1,
		level = 20,
		icon = "SK_HU_A1",
		photonblast = true,
	},

	["hu_whirlwind"] = {
		name = "Hunter Gale",
		desc = "Spawns a gale as you land Photon Art hits, dealing damage, recovering PP and building Chain. Duration refreshes and potency increases as more hits are landed.",
		cost = 2,
		level = 40,
		icon = "SK_HU_8",
	},

	["hu_swordbreaker"] = {
		name = "Swordbreaker",
		desc = "Reduces damage taken by 15% each time you get damaged, capping at 45%. Effect resets after 10s without taking damage.",
		cost = 1,
		level = 25,
		icon = "SK_HU_9",
	},
	
	["hu_berserk"] = {
		name = "Limit Break",
		desc = "Photon Blast. Reduces max HP by 75% and increases all damage dealt by 50%. Effect ends when lethal damage is sustained, retaining 1 HP instead.",
		cost = 3,
		level = 50,
		icon = "SK_HU_A2",
		photonblast = true,
	},	
	
	-- RANGER
	["ra_chaintrigger"] = {
		name = "Chain Trigger",
		desc = "Detonating a chain restores PP based on the Chain and player counts.",
		cost = 2,
		level = 15,
		icon = "SK_RA_1",
	},

	["ra_quickload"] = {
		name = "Quick Load",
		desc = "Reduces ranged Photon Art charge time when applicable.",
		cost = 1,
		level = 10,
		icon = "SK_RA_2",
	},	

	["ra_tacticalroll"] = {
		name = "Tactical Roll",
		desc = "Recover 35 PP when succesfully evading an attack with a dodge roll.",
		cost = 1,
		level = 35,
		icon = "SK_RA_3",
	},

	["ra_sharpshooter"] = {
		name = "Sharp Shooter",
		desc = "Increase damage by 10% when hitting enemies from far away. Your target marker will indicate when the bonus comes in effect.",
		cost = 1,
		level = 5,
		icon = "SK_RA_4",
	},

	["ra_zerorange"] = {
		name = "Zero Range",
		desc = "Increase damage by 10% when hitting enemies from close range. Your target marker will indicate when the bonus comes in effect.",
		cost = 1,
		level = 25,
		icon = "SK_RA_5",
	},

	["ra_jellenshot"] = {
		name = "Jellen Shot",
		desc = "Ranged normal attacks will apply a few seconds of Jellen to enemies, increasing chain duration on affected enemies.",
		cost = 1,
		level = 60,
		icon = "SK_RA_6",
	},

	["ra_zalureshot"] = {
		name = "Zalure Shot",
		desc = "Ranged normal attacks will apply a few seconds of Zalure to enemies, increasing Photon Blast gauge generation on hit.",
		cost = 1,
		level = 60,
		icon = "SK_RA_7",
	},

	["ra_overclock"] = {
		name = "Overclock",
		desc = "Photon Blast. Fully restores PP and increases Attack speed by 50% for 20s. Stacks additively with other Attack Speed increases.",
		cost = 2,
		level = 20,
		icon = "SK_RA_A1",
		photonblast = true,
	},

	["ra_chainreactor"] = {
		name = "Chain Reactor",
		desc = "Detonating a chain increases max PP to a maximum of 30 based on the Chain and player counts.",
		cost = 2,
		level = 30,
		icon = "SK_RA_8",
	},

	["ra_autoturret"] = {
		name = "Auto Turret",
		desc = "Detonating a chain summons a turret that keeps firing upon the target and builds chain. Duration varies based on Chain and player counts.",
		cost = 1,
		level = 40,
		icon = "SK_RA_9",
	},

	["ra_blightround"] = {
		name = "Blight Round",
		desc = "Photon Blast. Next ranged normal attack marks enemies hit for 20s. Chains build and last twice as long on marked enemies. 300% gauge build rate.",
		cost = 2,
		level = 50,
		icon = "SK_RA_A2",
		photonblast = true,
		photonblast_rate = 300,	-- 3x build rate
	},
	
	
	-- FORCE
	["fo_ppwellup"] = {
		name = "Cast PP Well Up",
		desc = "Allows natural PP recovery to occur at half the regular rate while using Techniques.",
		cost = 2,
		level = 15,
		icon = "SK_FO_1",
	},

	["fo_ppabsorb"] = {
		name = "Photon Absorb",
		desc = "When taking damage, recover PP proportionally to damage taken.",
		cost = 1,
		level = 10,
		icon = "SK_FO_2",
	},	
	
	["fo_supportbitch"] = {
		name = "Support Amplifier",
		desc = "Doubles the duration of Shifta & Deband.",
		cost = 1,
		level = 20,
		icon = "SK_FO_3",
	},

	["fo_photonscythe"] = {
		name = "Photon Scythe",
		desc = "Recover 10% PP when an enemy dies near you.",
		cost = 1,
		level = 60,
		icon = "SK_FO_4",
	},


	["fo_territoryburst"] = {
		name = "Territory Burst",
		desc = "Triples the range of support Techniques such as Shifta, Deband, Jellen, Zalure, Diverse and Resta.",
		cost = 1,
		level = 35,
		icon = "SK_FO_5",
	},

	["fo_debandtoughness"] = {
		name = "Deband Toughness",
		desc = "While Deband is active, increases your own Max HP by 10%.",
		cost = 1,
		level = 40,
		icon = "SK_FO_6",
		
		equip = function(p)
					if p.pso.deband
						PSO.lib.addDebandToughness(p)
					end
				end,
		
		unequip = function(p)
					PSO.lib.removeDebandToughness(p)
				end,
		
	},

	["fo_shiftacritical"] = {
		name = "Shifta Critical",
		desc = "While Shifta is active, increase Critical Hit rate by 15%.",
		cost = 2,
		level = 60,
		icon = "SK_FO_7",
	},

	["fo_photonflare"] = {
		name = "Photon Flare",
		desc = "Photon Blast. Casts Resta, Shifta & Deband, PP recovery speed +200%, Technique casting speed +20% for 20s.",
		cost = 1,
		level = 20,
		icon = "SK_FO_A1",
		photonblast = true,
	},	

	["fo_techartsalt"] = {
		name = "Tech Arts Alter",
		desc = "Technique casting speed +50% if the Technique's element differs from that of the last Technique cast.",
		cost = 2,
		level = 5,
		icon = "SK_FO_8",
	},

	["fo_elementalblot"] = {
		name = "Elemental Blot",
		desc = "Same-element Techniques build a sigil that detonates when using a Technique of a different element, granting Photon Blast charge based on the sigil's charge",
		cost = 2,
		level = 30,
		icon = "SK_FO_9",
	},

	["fo_phantomcaster"] = {
		name = "Phantom Caster",
		desc = "Photon Blast. Grants an invulnerability period when casting Techniques for 20s. On succesful dodge, the next 3 Techniques casted will be instant. 250% gauge build rate.",
		cost = 2,
		level = 50,
		icon = "SK_FO_A2",
		photonblast = true,
		photonblast_rate = 250,	-- 2x build speed
	},
	
}

-- level correction for beta tests
for k,v in pairs(PSO.info.typeSkills)
	if v and v.level and v.level > PSO.info.LEVELCAP
		v.level = PSO.info.LEVELCAP
	end
end

PSO.info.typeSkillsGrid = {
	
	[TYPE_HUNTER] = {
		
		"GENERAL",
		{"all_hp1",				"all_hp2",				"all_automate"},
		{"all_pp1",				"all_pp2"},
		{"all_dodgeextension",	"all_justreversal"},
		
		"TYPE-SPECIFIC SKILLS",
		{"hu_ironwill",			"hu_hunterphysique",	"hu_artsavenger"},
		{"hu_swordbreaker",		"hu_chainadrenaline",	"hu_heavyarts"},
		{"hu_whirlwind",		"hu_reapingscythe",		"hu_hatefulaggression"},
		
		"PHOTON BLASTS",
		{"hu_overload",			"hu_berserk"},
	
	},

	[TYPE_RANGER] = {
		
		"GENERAL",
		{"all_hp1",				"all_hp2",				"all_automate"},
		{"all_pp1",				"all_pp2"},
		{"all_dodgeextension",	"all_justreversal"},
		
		"TYPE-SPECIFIC SKILLS",

		{"ra_sharpshooter",		"ra_quickload",			"ra_chaintrigger"},
		{"ra_zerorange",		"ra_chainreactor",		"ra_tacticalroll"},
		{"ra_autoturret",		"ra_jellenshot",		"ra_zalureshot"},
		
		"PHOTON BLASTS",
		{"ra_overclock",		"ra_blightround"},
	
	},

	[TYPE_FORCE] = {
		
		"GENERAL",
		{"all_hp1",				"all_hp2",				"all_automate"},
		{"all_pp1",				"all_pp2"},
		{"all_dodgeextension",	"all_justreversal"},
		
		"TYPE-SPECIFIC SKILLS",
		{"fo_techartsalt",		"fo_ppabsorb",			"fo_ppwellup"},
		{"fo_supportbitch",		"fo_elementalblot",		"fo_territoryburst"},
		{"fo_debandtoughness",	"fo_photonscythe",		"fo_shiftacritical"},
		
		"PHOTON BLASTS",
		{"fo_photonflare",		"fo_phantomcaster"},
	
	},
	
}

PSO.info.typeSkillsList = {}

-- build this based on what we just had.
-- this list is used to know if a class is supposed to have a skill when we read data.
for i = 1, 3
	local classgrid = PSO.info.typeSkillsGrid[i]
	PSO.info.typeSkillsList[i] = {}
	
	for j = 1, #classgrid
		local row = classgrid[j]
		if type(row) == "table"
			for k = 1, #row
				local cell = row[k]
				if cell and PSO.info.typeSkills[cell]
					table.insert(PSO.info.typeSkillsList[i], cell)
				end
			end
		end
	end
end

/*
	[TYPE_HUNTER] = {
		"all_automate",
		"all_hp1",
		"all_hp2",
		"all_pp1",
		"all_pp2",
		"all_justreversal",
		--"all_elementalfocus",
		"all_dodgeextension",
		
		"hu_ironwill",
		"hu_hunterphysique",
		"hu_hatefulaggression",
		"hu_artsavenger",
		"hu_heavyarts",
		"hu_overload",
		"hu_reapingscythe",
		"hu_chainadrenaline",
		
	},

	[TYPE_RANGER] = {
		"all_automate",
		"all_hp1",
		"all_hp2",
		"all_pp1",
		"all_pp2",
		"all_justreversal",
		--"all_elementalfocus",
		"all_dodgeextension",	
		
		"ra_chaintrigger",
		"ra_quickload",		
		"ra_tacticalroll",
		"ra_sharpshooter",
		"ra_zerorange",
		"ra_overclock",
		"ra_jellenshot",
		"ra_zalureshot",
	},

	[TYPE_FORCE] = {
		"all_automate",
		"all_hp1",
		"all_hp2",
		"all_pp1",
		"all_pp2",
		"all_justreversal",
		--"all_elementalfocus",
		"all_dodgeextension",
		
		"fo_ppwellup",
		"fo_ppabsorb",		
		"fo_supportbitch",
		"fo_photonscythe",
		"fo_territoryburst",
		"fo_photonflare",
		"fo_debandtoughness",
		"fo_shiftacritical",
	},
}*/

-- iterates through all skills to reappply them.
-- used to ensure stuff like HP/PP up is properly kept between levelups and whatnot.
PSO.lib.reapplySkills = function(p)
	if p.pso.typeskills[p.pso.type]
		for i = 1, #p.pso.typeskills[p.pso.type]
			
			local sk = p.pso.typeskills[p.pso.type][i]
			if PSO.info.typeSkills[sk] and PSO.info.typeSkills[sk].equip
				PSO.info.typeSkills[sk].equip(p)
			end
		end
	end
end

PSO.lib.addSkill = function(p, sk)
	
	local dta = PSO.info.typeSkills[sk]
	local points = PSO.lib.skillCosts(p)
	
	assert(dta, "No valid data for skill "..tostring(sk))
	
	local cost = dta.cost or 0
	
	if points + cost > p.pso.maxskillpoints
		return false
	end
	
	-- unequip other photon blast if we have one...
	if dta.photonblast
		for i = 1, #p.pso.typeskills[p.pso.type]
			local sk2 = p.pso.typeskills[p.pso.type][i]
			local dta2 = PSO.info.typeSkills[sk2]
			if dta2.photonblast
				PSO.lib.removeSkill(p, sk2)
				break
			end
		end
	end
	
	p.pso.typeskills[p.pso.type][#p.pso.typeskills[p.pso.type]+1] = sk
	
	if PSO.info.typeSkills[sk].equip
		PSO.info.typeSkills[sk].equip(p)
	end
	
	
	return true
end

PSO.lib.hasSkill = function(p, sk)
	
	-- ais is considered not to have skills.
	if p.pso.ais
		return false
	end
	
	for i = 1, #p.pso.typeskills[p.pso.type]
		if p.pso.typeskills[p.pso.type][i] == sk
			return i
		end
	end
	return false
end

PSO.lib.removeSkill = function(p, sk)
	local index = PSO.lib.hasSkill(p, sk)
	if index
		table.remove(p.pso.typeskills[p.pso.type], index)
		
		if PSO.info.typeSkills[sk].unequip
			PSO.info.typeSkills[sk].unequip(p)
		end
		
	end
end

PSO.lib.skillCosts = function(p)
	local totalcost = 0
	for i = 1, #p.pso.typeskills[p.pso.type]
		local sk = PSO.info.typeSkills[p.pso.typeskills[p.pso.type][i]]
		if sk and sk.cost
			totalcost = $+sk.cost
		end
	end
	
	return totalcost
end