-- basic HUD

local display_dist = 3184

local function HU_login(v, p, c)
	if not p.psologin
		
		local w = 160
		local h = 32
		local x = 160 - (w/2)
		local y = 100 - (h/2)
		
		v.drawFill()
		v.drawFill(x, y, w, h, 253)
		v.drawString(160, 100, netgame and "Logging in..." or "Loading...", V_YELLOWMAP, "center")
		
		return
	end
end

--------------------------	
	-- low hp strawberry jam
--------------------------
local function HU_lowHPOverlay(v, p)
	
	if p.pso.lowhp
		
		local base_intensity = 3
		local max_screen = 1
		
		local maxhp = p.pso.maxhp - p.pso.maxhp*(p.pso.injury or 0)/100
		
		if p.pso.hp < maxhp*15/100
			base_intensity = 4
			max_screen = 2
		end
		
		v.fadeScreen(35, min(max_screen, p.pso.lowhp/8))
		
		local fadein = p.pso.lowhp/5
		local tflag = min(fadein, (base_intensity + (3*cos(leveltime*ANG1*4))/(FRACUNIT+2)))<<V_ALPHASHIFT
		
		v.drawScaled(0, 0, FRACUNIT/2, v.cachePatch("H_LOWHP2"), V_SNAPTOTOP|V_SNAPTOLEFT|V_ADD|tflag)
		v.drawScaled(320<<FRACBITS, 0, FRACUNIT/2, v.cachePatch("H_LOWHP2"), V_SNAPTORIGHT|V_SNAPTOTOP|V_ADD|V_FLIP|tflag)	
		v.drawScaled(0, 200<<FRACBITS, FRACUNIT/2, v.cachePatch("H_LOWHP"), V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_ADD|tflag)
		v.drawScaled(320<<FRACBITS, 200<<FRACBITS, FRACUNIT/2, v.cachePatch("H_LOWHP"), V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_ADD|V_FLIP|tflag)
	end
end


--------------------------
	-- open menu helper
--------------------------
local function HU_menuReminder(v, p)
	if gamemap == PSO.info.LOBBY_MAP
	and not (p.pso.menu and p.pso.menu.ref)
		local pp = PSO.lib.playerButtonIcon(p, IPT_MENU)
		v.draw(2, 1, v.cachePatch(pp), V_SNAPTOTOP|V_SNAPTOLEFT)
		v.drawString(20, 4, "OPEN MENU", V_YELLOWMAP|V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
		v.drawString(2, 14, "[F1] HELP & TIPS", V_YELLOWMAP|V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
	end
end

--------------------------	
	-- low pp indicator
-------------------------
local function HU_lowPP(v, p)
	if p.pso.nopp
		local pnum = (leveltime%10 / 5) +1
		
		if p.pso.nopp > TICRATE/3
		or leveltime & 1
			v.draw(160, 80, v.cachePatch("H_NOPP"..pnum))
		end
	end
end

--------------------------	
	-- target reticle:
-------------------------	
local function HU_targetReticle(v, p, c)
	
	-- blight round
	if server.blighted
		for i = 1, #server.blighted
			local mo = server.blighted[i]
			if mo and mo.valid
			and mo.pso.blighted and (mo.pso.blighted > TICRATE*5 or mo.pso.blighted & 1)
			and R_PointToDist2(p.mo.x, p.mo.y, mo.x, mo.y) < p.mo.scale*2048
				local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, mo.x, mo.y, mo.z + mo.height/2)
				v.drawScaled(cx, cy, FRACUNIT, v.cachePatch("H_BLIGHT"))
			end
		end
	end
	
	if p.pso.target and p.pso.target.valid
				
		-- the real reticle
		local ang = leveltime*4*ANG1
		local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, p.pso.target.x, p.pso.target.y, p.pso.target.z + p.pso.target.height/2)
		local sprite, flip, offc = v.getSpritePatch(SPR_AIMR, p.pso.lockon and B or A, 0, ang)

		-- optimal range for ranger shit
		local range = FixedHypot(FixedHypot(p.pso.target.x - p.mo.x, p.pso.target.y - p.mo.y), p.pso.target.z - p.mo.z)
		
		
		if p.pso.target.flags & MF_ENEMY	-- only on enemies.
			if range < PSO.info.CLOSERANGE*FRACUNIT
			and PSO.lib.hasSkill(p, "ra_zerorange")
			
			or range > PSO.info.LONGRANGE*FRACUNIT
			and PSO.lib.hasSkill(p, "ra_sharpshooter")
				
				v.drawScaled(cx, cy, FRACUNIT, v.cachePatch("H_CROSSH"))
			end
		end
		
		local offs = 0
		/*if offc	No longer necessary as of .14
			offs = 4*(flip and -1 or 1)
		end*/
		
		v.drawScaled(cx, cy+(offs*FRACUNIT), FRACUNIT, sprite)
	end
	
	-- AIS multi-target
	if p.pso.aislockon
		
		for i = 1, #p.pso.aislockon
			local target = p.pso.aislockon[i]
			
			if target and target.valid
				
				local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, target.x, target.y, target.z + target.height/2)
				v.drawScaled(cx, cy, FRACUNIT, v.cachePatch("H_CROSSH"), 0, v.getColormap(TC_RAINBOW, SKINCOLOR_MASTER))
			end
		end
	end
end

local function HU_elementalBlot(v, p, c)
	if p.pso.target and p.pso.target.valid and p.pso.target.pso
				
		local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, p.pso.target.x, p.pso.target.y, p.pso.target.z + p.pso.target.height/2)
		
		-- elemental blot
		local blot = p.pso.target.pso.blottrackers
		
		if blot and blot[#p]
			local level = min(4, blot[#p][1] / PSO.info.ELEMBLOTRATE)
			if level and level > 1
				v.drawScaled(cx, cy, FRACUNIT, v.cachePatch("H_BLOT"..(level-1)), V_10TRANS, v.getColormap(TC_DEFAULT, PSO.info.elementColours[blot[#p][2]]))
			end
		end
		
	end
end

local function HU_damageNumbers(v, p, c)
	
	if server.dmgnums
		for i = 1, #server.dmgnums
			
			local m = server.dmgnums[i]
			
			if not (m.flags2 & MF2_DONTDRAW) and R_PointToDist(m.x, m.y) < FRACUNIT*display_dist
				
				local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, m.x, m.y, m.z)
				
				if cx > 0 and cy > 0
					if m.tics > TICRATE/2 or m.tics & 1
						
						if m.extravalue1
							if not m.extravalue2
								if not m.angle
									v.drawString(cx/FRACUNIT, cy/FRACUNIT - 20, "\x82".."COOL!", 0, p.pso.ais and "small-center" or "center")
								elseif p.pso.level >= PSO.info.LEVELCAP
									v.drawString(cx/FRACUNIT, cy/FRACUNIT - 20, "\x87"..m.angle.." OEXP", 0, "center")
								else
									v.drawString(cx/FRACUNIT, cy/FRACUNIT - 20, "\x82"..m.angle.." EXP", 0, "center")
								end
							end
						else
							local scalefac = 3 --halfscale = true
							local transhalf = V_60TRANS
							if (m.source == p.mo or m.target == p.mo)
								if p.pso.ais
									transhalf = 0
								else
									scalefac = 1
									transhalf = 0
								end
							end
							
							local col = m.cusval and SKINCOLOR_MAGENTA or SKINCOLOR_WHITE
							
							if m.angle == 0
								col = (m.target and m.target.player) and SKINCOLOR_YELLOW or SKINCOLOR_GREY
							end	
							
							if m.extravalue2
								col = m.cusval and SKINCOLOR_SKY or SKINCOLOR_EMERALD
							elseif m.target and m.target.player
								col = SKINCOLOR_RED
							end
							
							if m.cvmem and leveltime%4 < 2
								col = SKINCOLOR_YELLOW
							end
							
							PSO.lib.HUD_drawDamageNumber(v, cx/FRACUNIT -5, cy/FRACUNIT, m.angle, transhalf, col, m.extravalue2, scalefac, m.cusval, m.cvmem)
							--v.drawNum(cx/FRACUNIT + (#tostring(m.angle) -1)*6, cy/FRACUNIT, m.angle)
						end
					end
				end
			end
		end
	end
end


--------------------------	
	-- chain numbers:
-------------------------
local function HU_chainCombo(v, p, c)
	if server.chaintrackers
		for i = 1, #server.chaintrackers
			
			local m = server.chaintrackers[i]
				
			if R_PointToDist(m.x, m.y) < FRACUNIT*display_dist and m.pso.chain and m.pso.chain > 1
				
				local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, m.x, m.y, m.z)

				local timer = m.pso.chainbreak and m.pso.chainbreak or m.pso.chaintime
				local maxtimer = m.pso.chainbreak and m.pso.maxchainbreak or m.pso.maxchaintime
				
				local col = m.pso.chainbreak and "\x82" or "\x88"
				local ccol = abs(timer-maxtimer) < 5 and "\x80" or col
				local col2 = m.pso.chainbreak and 73 or 130
				
				local str = ccol..string.format("%02d", m.pso.chain-1)..col.." CHAIN"
				local w = v.stringWidth(str)
				
				if cx > 0 and cy > 0
					
					local flag = p.pso.target ~= m and V_60TRANS or 0
					
					if timer > TICRATE or timer & 1
						v.drawString(cx/FRACUNIT, cy/FRACUNIT +4, str, flag, "center")
					end
					
					if not flag
						v.drawFill(cx/FRACUNIT - w/2, cy/FRACUNIT + 13, w, 2, 26)
					end
					
					v.drawFill(cx/FRACUNIT - w/2, cy/FRACUNIT + 13, timer*(w-1)/maxtimer, 1, col2|flag)
				end
			end
		end
	end
end

	
--------------------------	
	-- status bar:
-------------------------
PSO.lib.drawPlayerStatus = function(v, refp, p, sx, sy, small, flags)
	local x = sx -- p.pso.hidehudtimer*16
	if flags == nil
		flags = V_SNAPTOBOTTOM|V_SNAPTOLEFT
	end
	
	if refp and refp.pso
		x = $-refp.pso.hidehudtimer*16
	end
	
	local y = sy
	local corruption = p.pso.condition == COND_CORRUPTION

		
	local name = p.name
	if corruption
		local nlen = v.RandomRange(1, 16)
		name = ""
		for i = 1, nlen
			name = $..(string.char(v.RandomRange(47, 127)))
		end
	end
	
	if small	-- small bar
		
		v.drawFill(x-2, y, 88, 14, 253|flags)
		-- level & name
		local str = "\x82"
		
		if not p.pso.ais
			local lvnum = string.format("%03d", corruption and v.RandomRange(-255, 255) or p.pso.levelsynch or p.pso.level)
			local lvcolor = p.pso.levelsynch and "\x82" or p.pso.level >= PSO.info.LEVELCAP and "\x87" or "\x80"
			
			str = $.."Lv"..lvcolor+lvnum.."\x80 "
		else
			str = $.."A.I.S. \x80"
		end
		
		str = $..name
		v.drawString(x, y+2, str, flags|V_ALLOWLOWERCASE, "thin")
		
		-- hp bar
		local hp_col = {35, 34, 54, 73, 189, 98, 98, 98, 98, 98}
		local usecol = hp_col[(p.pso.hp*10/p.pso.maxhp) or 1] or 98
		
		v.drawFill(x, y+10, 84, 2, 26|flags)
		v.drawFill(x, y+10, corruption and v.RandomRange(0, 255) or (p.pso.hp*83/p.pso.maxhp), 1, usecol|flags)
		
	else	-- regular size bar for the local player
	
		for i = 1, 10
			v.drawFill(x-2, y-2 + (i-1), 48 + (i-1), 1, 253|flags)
		end
		
		v.drawFill(x-2, y+8, 132, 25, 253|flags)
		
		
		
		-- level and exp
		if not p.pso.ais
			local capped = p.pso.level >= PSO.info.LEVELCAP
			local synched = p.pso.levelsynch
			local lvnum = string.format("%03d", corruption and v.RandomRange(-255, 255) or synched and synched or p.pso.level)
			v.drawString(x, y-1, "\x82".."Lv", V_ALLOWLOWERCASE|flags)
			v.drawString(x + 19, y-1, lvnum, V_ALLOWLOWERCASE|flags|(synched and V_YELLOWMAP or capped and V_ORANGEMAP or 0))
			
			v.drawFill(x, y+8, 46, 2, 26|flags)
			
			if PSO.info.exptable[p.pso.level]
				v.drawFill(x, y+8, corruption and v.RandomRange(0, 128) or (p.pso.exp*45/PSO.info.exptable[p.pso.level]), 1, (capped and 54 or 73)|flags)
			end
		else
			v.drawString(x, y-1, "\x82".."A.I.S.", V_ALLOWLOWERCASE|flags)
		end
		
		v.drawString(x + 60, y-1, name, V_ALLOWLOWERCASE|flags)
		
		local hpnum = string.format("%03d", corruption and v.RandomRange(-999, 999) or p.pso.hp)
		local hpmnum = string.format("%03d", corruption and v.RandomRange(-999, 999) or p.pso.maxhp)
		local hpmcol = p.pso.debandtoughness and "\x84" or ""
		local ppnum = string.format("%03d", corruption and v.RandomRange(-999, 999) or p.pso.pp)
		local ppmnum = string.format("%03d", corruption and v.RandomRange(-999, 999) or p.pso.maxpp)
		local ppmcol = p.pso.chainreactor and "\x88" or ""
		
		v.drawString(x + 19, y+10, "\x82".."HP", V_ALLOWLOWERCASE|flags)
		v.drawString(x + 80, y+10, hpnum, V_ALLOWLOWERCASE|flags|(p.pso.hp <= p.pso.maxhp/3 and V_REDMAP or 0), "right")
		v.drawString(x + 81, y+10, "/", V_ALLOWLOWERCASE|flags|(p.pso.hp <= p.pso.maxhp/3 and V_REDMAP or 0))
		v.drawString(x + 90, y+10, hpmcol..hpmnum, V_ALLOWLOWERCASE|flags|(p.pso.hp <= p.pso.maxhp/3 and V_REDMAP or 0))
		
		local hp_col = {35, 34, 54, 73, 189, 98, 98, 98, 98, 98}
		--print(p.pso.hp*10/p.pso.maxhp)
		local usecol = hp_col[(p.pso.hp*10/p.pso.maxhp) or 1] or 98
		
		v.drawFill(x + 19, y+19, 110, 2, 26|flags)
		v.drawFill(x + 19, y+19, corruption and v.RandomRange(0, 255) or (p.pso.hp*109/p.pso.maxhp), 1, usecol|flags)
		
		-- injury:
		if p.pso.injury
			local maxhp = p.pso.maxhp - p.pso.maxhp*p.pso.injury/100
			local start = (maxhp*109/p.pso.maxhp)
			v.drawFill(x+19+start, y+19, 109-start, 1, 42|flags)
		end
		
		v.drawString(x + 19, y+21, "\x82".."PP", V_ALLOWLOWERCASE|flags)
		v.drawString(x + 80, y+21, ppnum, V_ALLOWLOWERCASE|flags, "right")
		v.drawString(x + 81, y+21, "/", V_ALLOWLOWERCASE|flags)
		v.drawString(x + 90, y+21, ppmcol..ppmnum, V_ALLOWLOWERCASE|flags)	
		--v.drawString(x + 60, y+21, ppnum.."/"..ppmnum, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT)

		local pp_col = {168, 166, 164, 162, 148, 132, 132, 132, 132, 132}
		--print(p.pso.hp*10/p.pso.maxhp)
		local ppusecol = pp_col[(p.pso.pp*10/p.pso.maxpp) or 1] or 132
		
		v.drawFill(x + 19, y+30, 110, 2, 26|flags)
		v.drawFill(x + 19, y+30, corruption and v.RandomRange(0, 255) or (p.pso.pp*109/p.pso.maxpp), 1, ppusecol|flags)
		
		-- photon blast gauge
		local pbx = x
		local pby = y+13
		
		if PSO.lib.playerHasPB(p) or corruption
			v.draw(pbx, pby, v.cachePatch("H_PB"..(corruption == false and p.pso.type or 0)), flags)
			
			-- pb gauge
			local gauge = corruption and v.RandomRange(0, 8) or (p.pso.pb * 8 / PSO.info.MAXPB)
			
			if gauge == 8 and leveltime%4 < 2
				gauge = 9
			end

			if p.pso.usepb and p.pso.pb
				gauge = min(8, $+1)
			end
			
			if gauge
			and (not p.pso.usepb or leveltime%4 < 2)
				v.draw(pbx, pby, v.cachePatch("H_PB_G"..gauge), flags)
			end

		else
			v.draw(pbx, pby, v.cachePatch("H_PB0"), flags)
		end
	end
	
	
	-- status conditions:
	local statusx = x + (small and 90 or 134)
	if p.pso.condition
		if (p.pso.conditiontimer > TICRATE*3 or p.pso.conditiontimer & 1)
			v.draw(statusx, y+(not small and 8 or 0), v.cachePatch("H_STATUS"..p.pso.condition), flags)
		end
		v.drawString(statusx+8, y+(not small and 8 or 0)+6, p.pso.conditiontimer/TICRATE, flags, "small-thin-right")
	end
	
	-- s/d j/z and other non-cleansable statuses
	local st = {"matesickness", "shifta", "deband", "jellen", "zalure", "rweakness", "rbrink"}
	local pp = {"H_MSICK", "H_STATUS8", "H_STATUS16", "H_STATUS32", "H_STATUS64", "H_REV1", "H_REV2"}
	local sx = statusx + (p.pso.condition and 12 or 0)
	
	for i = 1, #st
		if p.pso[st[i]]
			if (p.pso[st[i]] > TICRATE*10 or p.pso[st[i]] & 1)
				v.draw(sx, y+(not small and 8 or 0), v.cachePatch(pp[i]), flags)
			end
			v.drawString(sx+8, y+(not small and 8 or 0)+6, p.pso[st[i]]/TICRATE, flags, "small-thin-right")
			sx = $+ 12
		end
	end

	if p.pso.levelsynch
		v.draw(sx, y+(not small and 8 or 0), v.cachePatch("H_LSYNCH"), flags)
	end
	
	if corruption
		for i = 1, 3
			
			local cx = v.RandomRange(-30, 100) / (small and 2 or 1)
			local cy = v.RandomRange(-10, 50) / (small and 2 or 1)
			
			local cw = v.RandomRange(8, 128) / (small and 2 or 1)
			local ch = v.RandomRange(6, 24) / (small and 2 or 1)
			v.drawFill(x+cx,y+cy,cw,ch, 253|flags)
			
		end
	end
end


	
--------------------------	
	-- item palette:
-------------------------
local function HU_playerItemPalette(v, p)
	local px = 268 + p.pso.hidehudtimer*16 + p.pso.palette.open*32
	local py = 182
	
	-- indication for switching:
	local bt = v.cachePatch(PSO.lib.playerButtonIcon(p, IPT_PALETTE, flash))
	v.draw(px - 16, py+2, bt, V_SNAPTOBOTTOM|V_SNAPTORIGHT)
	
	-- weapon:
	local wdta = p.pso.palette.layout[2][p.pso.palette.weaponnum]
	local onehanded = PSO.lib.slotIsOneHanded(wdta)
	
	v.drawFill(px, py, 32, 16, 253|V_SNAPTOBOTTOM|V_SNAPTORIGHT)
	
	if p.pso.ais
		v.draw(px, py, v.cachePatch("H_AIS"), V_SNAPTOBOTTOM|V_SNAPTORIGHT, v.getColormap(TC_DEFAULT, SKINCOLOR_MASTER))
	elseif wdta
		for i = 1, 2
			local wp = wdta[i]
			if wp
				local col = PSO.info.elementColours[wp.element]
				local ico = PSO.info.weaponToIcon[wp.weapontype]
				local tflag = 0
				
				if onehanded
					
					if i == 1 and not p.pso.backpalette and wdta[2]
					or i == 2 and p.pso.backpalette and wdta[1]
						tflag = V_50TRANS
					end
				end
				
				v.draw(px, py, v.cachePatch(ico), V_SNAPTOBOTTOM|V_SNAPTORIGHT|tflag, v.getColormap(TC_DEFAULT, col))
			end
		end
	end
	
	if not p.pso.ais
		v.drawString(px+1, py+1, p.pso.palette.weaponnum, V_50TRANS|V_SNAPTOBOTTOM|V_SNAPTORIGHT, "thin")
	end
	
	-- armour:
	local arm = p.pso.palette.layout[3][p.pso.palette.armornum]
	v.drawFill(px + 33, py, 16, 16, 253|V_SNAPTOBOTTOM|V_SNAPTORIGHT)
	
	if p.pso.ais
		local col = SKINCOLOR_MASTER
		v.draw(px + 33, py, v.cachePatch("H_AISA"), V_SNAPTOBOTTOM|V_SNAPTORIGHT, v.getColormap(TC_DEFAULT, col))
	elseif arm
		local col = PSO.info.elementColours[arm.element]
		v.draw(px + 33, py, v.cachePatch("H_ARMR"), V_SNAPTOBOTTOM|V_SNAPTORIGHT, v.getColormap(TC_DEFAULT, col))
	end
end



--------------------------	
	-- action palette:
-------------------------
local function HU_playerActionPalette(v, p)
	local wp = PSO.lib.getPlayerHUDWeapon(p)
	local hand = HAND_BOTH
	local offset = 0

	local px = 226 + p.pso.hidehudtimer*16 + p.pso.palette.open*32
	local py = 167
	
	if wp
		hand = PSO.info.weaponHand[wp.weapontype]
	end


	-- background for PA display etc
	v.drawFill(px, py, 92, 14, 253|V_SNAPTORIGHT|V_SNAPTOBOTTOM)
	
	-- the thing for the back palette...
	local bpalb = PSO.lib.playerButtonIcon(p, IPT_SECONDPALETTE, p.pso.backpalette)
	v.drawFill(px + 65, py - 12, 27, 12, 253|V_SNAPTORIGHT|V_SNAPTOBOTTOM)
	v.draw(px + 67, py - 11, v.cachePatch("H_BPAL"), V_SNAPTORIGHT|V_SNAPTOBOTTOM)
	v.draw(px + 67 + 8, py - 12, v.cachePatch(bpalb), V_SNAPTORIGHT|V_SNAPTOBOTTOM)
	
	for i = 1, 11
		v.drawFill(px + 65 -i, py -12 +i, 1, (12-i), 253|V_SNAPTOBOTTOM|V_SNAPTORIGHT)
	end


	-- PB use indicator
	if PSO.lib.playerHasPB(p)
	and p.pso.pb >= PSO.info.MAXPB
		local icoback = PSO.lib.playerButtonIcon(p, IPT_SECONDPALETTE, PSO.lib.playerInputDown(p, IPT_SECONDPALETTE))
		local icopal = PSO.lib.playerButtonIcon(p, IPT_PALETTE, PSO.lib.playerInputDown(p, IPT_PALETTE))
		
		local pbx = px - 10
		
		v.draw(pbx, py - 15, v.cachePatch(icoback), V_SNAPTOBOTTOM|V_SNAPTORIGHT)
		v.drawString(pbx+15, py - 15 + 2, "+", V_SNAPTOBOTTOM|V_SNAPTORIGHT)
		v.draw(pbx + 21, py - 15, v.cachePatch(icopal), V_SNAPTOBOTTOM|V_SNAPTORIGHT)
		
		local yflag = leveltime%10 < 5 and V_YELLOWMAP or 0
		v.drawString(pbx+40, py - 15 + 3, "PB", V_SNAPTOBOTTOM|V_SNAPTORIGHT|yflag, "thin")
		
	end
	
	if wp and hand == HAND_BOTH
		offset = p.pso.backpalette and 3 or 0	-- if we hold swap palette, offset by 2
	end
	
	local butts = {IPT_ATTACK, IPT_ATTACK2, IPT_ATTACK3}	
	local cx = px+3
	
	local cy = py+1
		
	for i = 1 + offset, 3 + offset
			
		local flash = false
			
		if p.pso.actionnum == i
		and p.pso.actiontimer < TICRATE/2
		and p.pso.actiontimer%4 < 2
			flash = true
		end
			
		local bt =  PSO.lib.playerButtonIcon(p, butts[i - offset], flash)
			
		-- action name / ico:
		local dta
		
		if wp and wp.actions[i]
			dta = PSO.info.photonArts[wp.actions[i]]
		end
				
		if dta
			
			-- do we have blight rounds...?
			if p.pso.blightround
			and PSO.info.weaponDamageType[wp.weapontype] == DMT_RANGED
			and dta.normal
				
				--v.drawString(px - 46, cy+2, "Blight Round", V_ALLOWLOWERCASE|V_SNAPTORIGHT|V_SNAPTOBOTTOM|(leveltime%10 < 5 and V_YELLOWMAP or V_ORANGEMAP), "thin")
				if leveltime % 10 < 5
					v.draw(cx, cy, v.cachePatch("H_RSPCL"), V_SNAPTORIGHT|V_SNAPTOBOTTOM)
				end
				
			else
			
				local ico = v.cachePatch(dta.icon)
				--v.drawString(px - 46, cy+2, dta.name, V_ALLOWLOWERCASE|V_SNAPTORIGHT|V_SNAPTOBOTTOM, "thin")
				v.draw(cx, cy, ico, V_SNAPTORIGHT|V_SNAPTOBOTTOM)
			end
		elseif wp and wp.actions[i]
			v.drawString(cx, cy+2, "\x85".."?", V_ALLOWLOWERCASE|V_SNAPTORIGHT|V_SNAPTOBOTTOM, "thin")
		else
			v.drawString(cx, cy+2, "-", V_ALLOWLOWERCASE|V_SNAPTORIGHT|V_SNAPTOBOTTOM)
		end
			
		v.draw(cx+12, cy, v.cachePatch(bt), V_SNAPTORIGHT|V_SNAPTOBOTTOM)
		
		cx = $+30
	end
end

--------------------------	
	-- keycards:
-------------------------
local function HU_displayKeycards(v, p)
	if server.keycards
	and not (p.pso.menu and p.pso.menu.ref)
		local minimap_on = cv_useminimap.value and (gamemap ~= PSO.info.LOBBY_MAP or p.mo.y > PSO.info.LOBBY_YPOS*FRACUNIT)
		local tx = 320 - 64 - 2
		local ty = minimap_on and 68 or 2
		
		v.drawFill(tx, ty, 64, 21, 253|V_SNAPTOTOP|V_SNAPTORIGHT)
		v.drawString(tx+2, ty+2, "\x82".."KEYS", V_SNAPTOTOP|V_SNAPTORIGHT)
		
		for i = 1, 4
			if server.keycards & (2^(i-1))
				v.draw(tx+2, ty+11, v.cachePatch("H_KCARD"..i), V_SNAPTOTOP|V_SNAPTORIGHT)
				tx = $+9
			end
		end
	end	
end


--------------------------	
	-- target info:
-------------------------
local function HU_targetInfo(v, p)
	
	
	if p.pso.target and p.pso.target.valid and not p.pso.menu.ref
		local t = p.pso.target
		local corruption = p.pso.condition == COND_CORRUPTION
		local tx = 2
		local ty = 2
				
		if t.type == MT_PSO_DROP
					
			local it = t.psoitem
			if t.psoitems
				it = t.psoitems[#p+1]
			end
			
			-- meseta
			if it and type(it) == "number"
				v.drawFill(tx, ty, 140, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
				v.drawString(tx+2+17, ty+2, corruption and v.RandomRange(-9999999, 9999999) or PSO.lib.numCommas(it).." MESETA", V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)
				
				local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
				v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
				v.drawString(tx+2+20, ty+15, "PICK UP", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
				
			elseif it	-- normal item
				
				v.drawFill(tx, ty, 140, 33, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
				
				local qty = t.psoqty
				if not qty
					qty = 1
				end
				
				local info
				local id = true
				if type(it) == "string"
					info = PSO.info.items[it]
				else
					info = PSO.info.items[it.ref]
					if it.unidentified
						id = nil
					end
				end
				
				local name, rarity, ico, nopickup
				
				if info
					name = PSO.lib.getItemName(it)
					rarity = info.rarity
					ico = PSO.lib.getItemIcon(it)
				else
					name = tostring(it)
					rarity = 0
					nopickup = "Invalid item, please report this bug"
				end
				
				if corruption
					local nlen = v.RandomRange(1, 16)
					name = ""
					for i = 1, nlen
						name = $..(string.char(v.RandomRange(47, 127)))
					end	
					
					rarity = v.RandomRange(0, 15)
				end				
				
				if ico and not corruption
					v.draw(tx+2+7, ty+2, v.cachePatch(ico), V_SNAPTOTOP|V_SNAPTOLEFT)
				end
				v.drawString(tx+2+17, ty+2, name..(qty > 1 and " ("..qty..")" or ""), V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)
				PSO.lib.HUD_drawItemRarity(v, tx+2, ty+12, V_SNAPTOLEFT|V_SNAPTOTOP, id and rarity or 0)
				
				
				if info and info.carrylimit 
					local slot = PSO.lib.findPlayerItem(p, it)
					if slot
						local dta = p.pso.inventory[slot]
						if dta and dta[2] and dta[2] >= info.carrylimit
							nopickup = info.name.." STOCK FULL"
						end
					end
				elseif #p.pso.inventory >= PSO.info.INVENTORYSIZE
				and not (info and info.carrylimit)
					nopickup = "INVENTORY IS FULL"
				end
			
				if nopickup
					v.drawString(tx+2, ty+22, nopickup, V_REDMAP|V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
				else
					local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
					v.draw(tx+2, ty+19, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
					v.drawString(tx+2+18, ty+24, "PICK UP", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
				end
			end
		
		elseif t.type == MT_PSO_UNLOCKCARD

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "KEYCARD (ID: "..(t.id or "INVALID")..")"

			if corruption
				local nlen = v.RandomRange(1, 16)
				name = ""
				for i = 1, nlen
					name = $..(string.char(v.RandomRange(47, 127)))
				end				
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)

			local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
			v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
			v.drawString(tx+2+20, ty+15, "PICK UP", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")			
		
		elseif t.type == MT_PSO_CARDREADER

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "NEED ID: "..(t.id or "INVALID")

			if corruption
				local nlen = v.RandomRange(1, 16)
				name = ""
				for i = 1, nlen
					name = $..(string.char(v.RandomRange(47, 127)))
				end				
			end

			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)
			
			local keycards = server.keycards or 0
			
			if keycards & (t.id or 0)
				local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
						
				v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
				v.drawString(tx+2+20, ty+15, "USE KEYCARD", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
			else
				v.drawString(tx+2, ty+15, "KEYCARD NOT FOUND", V_REDMAP|V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
			end

		elseif t.type == MT_PSO_TELEPORTER
		or t.type == MT_PSO_BOSSTELEPORTER

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "TELEPORTER"
			if t.spawnpoint and t.spawnpoint.angle == 1
				name = "TO LOW ORBIT STATION"
			end

			if corruption
				local nlen = v.RandomRange(1, 16)
				name = ""
				for i = 1, nlen
					name = $..(string.char(v.RandomRange(47, 127)))
				end				
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)

			local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
			v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
			v.drawString(tx+2+20, ty+15, "TELEPORT", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")	

		elseif t.type == MT_PSO_JUMPPAD

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "Jump Pad"

			if corruption
				local nlen = v.RandomRange(1, 16)
				name = ""
				for i = 1, nlen
					name = $..(string.char(v.RandomRange(47, 127)))
				end				
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)

			local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
			v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
			v.drawString(tx+2+20, ty+15, "USE", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")	

		elseif t.type == MT_PSO_DATAPOD

			v.drawFill(tx, ty, 128, 26, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			local name = "Datapod"

			if corruption
				local nlen = v.RandomRange(1, 16)
				name = ""
				for i = 1, nlen
					name = $..(string.char(v.RandomRange(47, 127)))
				end				
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)

			local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					
			v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
			v.drawString(tx+2+20, ty+15, "INSPECT", V_SNAPTOTOP|V_SNAPTOLEFT, "thin")	
		
		elseif t.flags & MF_ENEMY	-- enemy
			
			local boxw = 130
			
			v.drawFill(tx, ty, boxw, 24, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
			
			if t.pso and t.pso.part
				if t.tracer
					t = t.tracer
				end
			end
			
			local name = tostring(t)
			if PSO.info.enemies[t.type]
				name = "Lv"..(string.format("%03d", t.pso.level) or "?").." "..PSO.info.enemies[t.type].name or $
			end
			
			if corruption
				local nlen = v.RandomRange(1, 16)
				name = ""
				for i = 1, nlen
					name = $..(string.char(v.RandomRange(47, 127)))
				end				
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")
			
			-- weakness?
			if t.pso.weakness
				local str = "WEAK:"
				v.drawString(tx+2, ty+11, "\x82"..str, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")
				
				local weak1 = t.pso.weakness
				local weak2 = t.pso.weakness2
				
				-- we don't actually want that behaviour.
				/*if weak1 and weak2
					local buff = max(weak1, weak2)
					weak1 = min(weak1, weak2)
					weak2 = buff
				end*/	
				
				local pp = v.cachePatch("H_ELEM"..(corruption and v.RandomRange(1, 6) or weak1))
				local sw = v.stringWidth(str, V_ALLOWLOWERCASE, "thin")
				
				v.draw(tx+4+sw, ty+11, pp, V_SNAPTOTOP|V_SNAPTOLEFT)
				
				if weak2
					local pp2 = v.cachePatch("H_ELEM"..(corruption and v.RandomRange(1, 6) or weak2))
					v.draw(tx+4+sw+10, ty+11, pp2, V_SNAPTOTOP|V_SNAPTOLEFT)
				end
			end
			
			-- attribute
			if t.pso.attribute
				v.drawString(tx+47, ty+11, "/ \x82TYPE: \x80"..PSO.info.attributeNames[t.pso.attribute], V_ALLOWLOWERCASE|V_SNAPTOTOP|V_SNAPTOLEFT, "thin")
			end
			
			-- targetted?
			if t.target == p.mo and not corruption
				v.draw(tx + boxw - 16, ty+1, v.cachePatch("H_HATE"), V_SNAPTOTOP|V_SNAPTOLEFT)
			end
			
			local info = t.pso
			if info
				if server.psodps	
					
					local str = "\x82".."BEGIN ATTACKING TO CALCULATE DPS"
					
					if server.psodamagetime
						local elapsed = (leveltime - server.psodamagetime)
						local dmg = server.psodamage*TICRATE / elapsed 
						local dpp = (not server.psodpspp) and "EXPAND PP TO CALCULATE" or (server.psodamage / server.psodpspp)
						local cps = server.psodpschain*100*TICRATE / elapsed
						local cpp = (not server.psodpspp) and "EXPAND PP TO CALCULATE" or server.psodpschain*100 / server.psodpspp
						
						local s = elapsed/TICRATE
						local c = tostring(G_TicsToCentiseconds(elapsed))
						
						str = "DPS: "..dmg.." ("..s.."."..(c:format("%02d")).."S)"
						str = $.."\n".."DPP: "..dpp.." ("..server.psodpspp.." PP USED)"
						str = $.."\n".."CPS: "..(type(cps) == "number" and ((cps/100).."."..string.format("%02d", cps%100)) or cps)
						str = $.."\n".."CPP: "..(type(cpp) == "number" and ((cpp/100).."."..string.format("%02d", cpp%100)) or cpp)
					end
					
					v.drawString(tx+2, ty+20, str, V_SNAPTOTOP|V_SNAPTOLEFT|V_RETURN8)
					
				else	-- normal rendering: hp bar
					v.drawFill(tx+2, ty+20, boxw-4, 2, 26|V_SNAPTOTOP|V_SNAPTOLEFT)
					v.drawFill(tx+2, ty+20, corruption and v.RandomRange(0, 255) or (info.hp*(boxw-5)/info.maxhp), 1, 98|V_SNAPTOTOP|V_SNAPTOLEFT)
					if not corruption
						local hppercent = info.hp*1000/info.maxhp
						local percentstart = hppercent/10
						local percentend = hppercent%10
						
						if p.pso.showrealhp
							v.drawString(tx+2 + (boxw-4), ty+17, info.hp, V_50TRANS|V_SNAPTOTOP|V_SNAPTOLEFT, "thin-right")
						else
							v.drawString(tx+2 + (boxw-4), ty+17, percentstart.."."..percentend.."%", V_50TRANS|V_SNAPTOTOP|V_SNAPTOLEFT, "thin-right")
						end
					end
				end
			end

			-- S/D J/Z
			local statusx = tx + boxw + 2
			if info.condition
				if (info.conditiontimer > TICRATE*3 or info.conditiontimer & 1)
					v.draw(statusx, ty+10, v.cachePatch("H_STATUS"..info.condition), V_SNAPTOTOP|V_SNAPTOLEFT)
				end
				v.drawString(statusx+8, ty+10+6, info.conditiontimer/TICRATE, V_SNAPTOTOP|V_SNAPTOLEFT, "small-thin-right")
				statusx = $+12	-- increment for s/d/j/z drawer next
			end
			
			local st = {"shifta", "deband", "jellen", "zalure"}
			local pp = {"H_STATUS8", "H_STATUS16", "H_STATUS32", "H_STATUS64"}
			local sx = statusx --tx + boxw + 2
			
			for i = 1, #st
				if info[st[i]]
					if (info[st[i]] > TICRATE*10 or info[st[i]] & 1)
						v.draw(sx, ty+10, v.cachePatch(pp[i]), V_SNAPTOTOP|V_SNAPTOLEFT)
					end
					v.drawString(sx+8, ty+10+6, info[st[i]]/TICRATE, V_SNAPTOTOP|V_SNAPTOLEFT, "small-thin-right")
					sx = $+ 12
				end
			end			
		else	-- other shootable objects
			v.drawFill(tx, ty, 128, 24, 253|V_SNAPTOTOP|V_SNAPTOLEFT)
		
			local name = tostring(t)
			if t.pso and t.pso.name
				name = t.pso.name
			end

			if corruption
				local nlen = v.RandomRange(1, 16)
				name = ""
				for i = 1, nlen
					name = $..(string.char(v.RandomRange(47, 127)))
				end				
			else
				if t.pso.subtext
					local bt =  PSO.lib.playerButtonIcon(p, IPT_INTERACT, true)
					v.draw(tx+2, ty+12, v.cachePatch(bt), V_SNAPTOTOP|V_SNAPTOLEFT)
					v.drawString(tx+2+20, ty+15, t.pso.subtext, V_SNAPTOTOP|V_SNAPTOLEFT, "thin")				
				end
			end
			
			v.drawString(tx+2, ty+2, name, V_SNAPTOTOP|V_SNAPTOLEFT|V_ALLOWLOWERCASE)		
		end

		if corruption
			for i = 1, 3
				
				local cx = v.RandomRange(-30, 100)
				local cy = v.RandomRange(-10, 50)
				
				local cw = v.RandomRange(8, 128)
				local ch = v.RandomRange(6, 24)
				v.drawFill(tx+cx,ty+cy,cw,ch, 253|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
				
			end
		end
		
	end
	
end

local function HU_showIncapacitations(v, p, c)
	for p2 in players.iterate do
			
		if p2 and p2.mo and p2.mo.valid
		and p2.pso and p2.pso.hp <= 0	-- blud is dead	
		and R_PointToDist(p2.mo.x, p2.mo.y) < FRACUNIT*display_dist
				
			local cx, cy = PSO.lib.HUD_screenCoords(v, p, c, p2.mo.x, p2.mo.y, p2.mo.z+p2.mo.height/3)
			cy = $ + 2*cos(leveltime*ANG1*4)
			
			v.drawScaled(cx, cy, FRACUNIT, v.cachePatch("H_INCAP"))
		end
	end
end

local function HU_springEvade(v, p, c)
	if p.pso.springbuffer
	and leveltime%10 < 5
	
		local pp = PSO.lib.playerButtonIcon(p, IPT_EVADE)
		v.draw(95, 138, v.cachePatch(pp))
		v.drawString(160, 140, "DOUBLE JUMP", V_YELLOWMAP, "center")
	end
end

local function playerHUD_base(v, p, c)
	
	
	HU_login(v, p, c)
	
	if netgame and p ~= server and not p.psologin
	or p.psologinfinish and p.psologinfinish > TICRATE-10
		PSO.lib.HU_dataTransfer(v, p, c)
		return
	end
	
	if not p.pso return end
	
	if not p.pso.type_initialized return end	-- only display the type selection menu
	

	HU_lowHPOverlay(v, p)
	
	-- for more convenient fades
	if p.mo and p.mo.fadecolor != nil and p.mo.fadestrength != nil
		v.fadeScreen(p.mo.fadecolor, p.mo.fadestrength)
	end

	if input.gameControlDown(GC_SCORES)
		return
	end	
	
	if p.pso.quest and p.pso.quest.completetimer return end	-- only show quest complete menu
	
	if p.pso.credits
		PSO.lib.drawCredits(v, 2*FRACUNIT, p.pso.credits*FRACUNIT/3, V_50TRANS)
	end
	
	HU_menuReminder(v, p)
	HU_lowPP(v, p)
	HU_showIncapacitations(v, p, c)
	HU_targetReticle(v, p, c)
	HU_damageNumbers(v, p, c)
	HU_elementalBlot(v, p, c)
	HU_chainCombo(v, p, c)
	PSO.lib.drawPlayerStatus(v, p, p, 4, 165)
	
	local maxplayers = 3
	local drawnplayers = 0
	for p2 in players.iterate	-- other players' status
		
		if p2 ~= p and p2.mo and p2.pso
			PSO.lib.drawPlayerStatus(v, p, p2, 4, 145 - 16*drawnplayers, true)	-- test
			drawnplayers = $+1
		
			if drawnplayers >= maxplayers break end
		end
	end	
	
	PSO.lib.runHook("playerHUD", v, p)
	
	HU_playerItemPalette(v, p)
	HU_playerActionPalette(v, p)
	HU_displayKeycards(v, p)
	HU_targetInfo(v, p)
	HU_springEvade(v, p, c)
	
	if p.psologinfinish
		v.fadeScreen(0, min(10, p.psologinfinish))
	end
end

PSO.lib.registerHudFunc(playerHUD_base)