-- Palette is separated because it's messier than other HUD elements...

local function playerQuickPaletteHUD(v, p, c)
	
	if not p.pso.quickpalette return end
	
	if p.pso.quickpalette > (PSO.info.QUICKPAL/4)
	or leveltime & 1
		
		local pal = p.pso.palette.layout[1]
		local rtime = (PSO.info.QUICKPAL - p.pso.quickpalette)
		local spacing = min(18, 6*rtime)
		local shift = -(#pal)/2
		
		for i = 1, #pal
			local x = 160 + shift*spacing + 1*((shift or 1)/abs(shift or 1))
			local y = 92

			v.drawFill(x, y, 16, 16, 253)
			v.drawString(x+8, y-9, "\x82"..i, 0, "center")
			
			local it = pal[i]
			
			if it
				local info = PSO.lib.itemInfo(it)
				local ico = info.paletteico
				local tflag = 0
				local amount = 0
				
				-- check how many of that item we have
				local index = PSO.lib.findPlayerItem(p, it)
				if not index
					tflag = V_60TRANS
				else
					amount = p.pso.inventory[index][2]
					
					if amount == 0
					or info.canuse and not info.canuse(p)
						tflag = V_60TRANS
					end
				end
				
				v.draw(x, y, v.cachePatch(ico), tflag)
				v.drawString(x + 15, y + 9, (amount == 0 and "\x85" or "")..tostring(amount), V_30TRANS, "thin-right")
			end
			
			shift = $+1
		end
	end
end

local function playerActionPaletteHUD(v, p, c)
	
	if not p.pso return end
	playerQuickPaletteHUD(v, p, c)
	
	local pal = p.pso.palette.layout
	
	local flags = V_SNAPTORIGHT|V_SNAPTOBOTTOM
	local x = 320 - p.pso.palette.open*10
	local y = 120
	
	local cursorx = 0
	local cursory = y + 16 + 16 + 2	-- 2 slots down
	
	local selectc = p.pso.palette.column	-- column selected
	
	if p.pso.palette.open
	and p.pso.palette.qlayout
				
		local pal = p.pso.palette.qlayout
		
		local sitem = p.pso.palette.qlayout[selectc][3][1]
		local sitem2
		
		local itqty = ""
		
		if selectc == 2
			-- weapons are weird...
			sitem2 = sitem[2]
			sitem = sitem[1]	-- for now just cheat and always take the first weapon.
		elseif selectc == 1	-- consumables, see if we can quantify it!
		
			if sitem
				local index = PSO.lib.findPlayerItem(p, sitem)
				local qty = 0
				if index
					qty = p.pso.inventory[index][2]
				end
				
				local info = PSO.lib.itemInfo(sitem)
				
				if info.carrylimit
					itqty = (qty == 0 and "\x85" or "").." ("..qty.."/"..info.carrylimit..")"
				end
			end
		end
		
		local itname = "--"
		local itname2
		
		if sitem
			itname = PSO.lib.getItemName(sitem)..itqty
		end
		if sitem2
			itname2 = PSO.lib.getItemName(sitem2)
		end
		
		
		local px = x
		
		for col = 1, #pal
			local py = y + (col == selectc and p.pso.palette.scroll or 0)
			local pw = col == 2 and 32 or 16
			local tflag = selectc != col and V_80TRANS or 0
			
			if selectc == col
				cursorx = px
			end
			
			for lin = 1, 5
				
				local nodraw
				if (selectc ~= col or not p.pso.palette.scroll) and (lin < 2 or lin > 4) or lin == 5	-- never draw line 5 due to how this works.
					nodraw = true
				end
				
				local offsy = 0	-- y offset (start drawing from)
				local offsh = 0 -- h offset (stop drawing after)
				if p.pso.palette.scroll and selectc == col
					if lin == 1
						offsy = 16-p.pso.palette.scroll
					elseif lin == 4
						offsh = p.pso.palette.scroll
					end
				end
				
				if not nodraw
					v.drawFill(px, py+offsy, pw, 16-offsy-offsh, 253|flags)
				
					-- draw the icon:
					--if not pal[col][lin] continue end
					local cell = pal[col][lin]
					local it
					
					if cell
						it = pal[col][lin][1]
					end	
					
					local ico
					local colour
					
					if it
						if type(it) == "string"
							ico = PSO.lib.itemInfo(it).paletteico
							local pp = v.cachePatch(ico or "what")
							--v.draw(px, py, , V_SNAPTOLEFT|tflag)
							
							-- do we have any of that item?
							local index = PSO.lib.findPlayerItem(p, it)
							local savetflag = tflag
							local qty = 0
							
							if col == selectc
								if index
								
									local info = PSO.lib.itemInfo(it)
								
									qty = p.pso.inventory[index][2]
									if info and info.canuse and not info.canuse(p)
										tflag = V_60TRANS
									end
								else
									tflag = V_60TRANS
								end
							end
							
							v.drawCropped(px*FRACUNIT, (py+offsy)*FRACUNIT, FRACUNIT, FRACUNIT, pp, flags|tflag, nil, 0, offsy*FRACUNIT, pp.width*FRACUNIT, (16-offsy-offsh)*FRACUNIT)
							if lin == 3 and col == selectc
								v.drawString(px + 15, py + 9, (qty == 0 and "\x85" or "")..tostring(qty), flags|V_30TRANS, "thin-right")
							end
							tflag = savetflag
							
						elseif type(it) == "table"
							
							if it.type == IT_ARMOR
								ico = "H_ARMR"
								local pp = v.cachePatch(ico or "what")
								colour = v.getColormap(TC_DEFAULT, PSO.info.elementColours[it.element])
								--v.draw(px, py, v.cachePatch(ico or "what"), V_SNAPTOLEFT|tflag, v.getColormap(TC_DEFAULT, col))
								v.drawCropped(px*FRACUNIT, (py+offsy)*FRACUNIT, FRACUNIT, FRACUNIT, pp, flags|tflag, colour, 0, offsy*FRACUNIT, pp.width*FRACUNIT, (16-offsy-offsh)*FRACUNIT)
								
							else	-- weapons...
								for i = 1, 2
									local t_it = it[i]
									if not t_it continue end

									local colour = v.getColormap(TC_DEFAULT, PSO.info.elementColours[t_it.element])
									local ico = PSO.info.weaponToIcon[t_it.weapontype]
									local pp = v.cachePatch(ico or "what")
											
									v.drawCropped(px*FRACUNIT, (py+offsy)*FRACUNIT, FRACUNIT, FRACUNIT, pp, flags|tflag, colour, 0, offsy*FRACUNIT, pp.width*FRACUNIT, (16-offsy-offsh)*FRACUNIT)
								end
								
								local pp = v.cachePatch("TNYFN0"..tostring(48 + cell[2]))
								v.drawCropped((px+1)*FRACUNIT, (py+offsy+1)*FRACUNIT, FRACUNIT, FRACUNIT, pp, flags|V_50TRANS, nil, 0, offsy*FRACUNIT, pp.width*FRACUNIT, (16-offsy-offsh)*FRACUNIT)
								
							end
						end
					end
				end
				
				py = $ + 16 + 1
			end
			px = $ + pw + 1
		end
		
		-- draw the button prompts
		local inputs = {IPT_ATTACK2, IPT_ATTACK, IPT_ATTACK3}
		local xpos = {x, x +16 +8, x +16 + 32}
		
		for i = 1, #inputs
			local flash = selectc == i and leveltime%4 < 2 and p.pso.palette.scroll
			local pp = v.cachePatch(PSO.lib.playerButtonIcon(p, inputs[i], flash))
			
			v.draw(xpos[i], y, pp, flags)
		end
		
		-- draw the cursor
		if leveltime%10 < 5
			local pp = selectc == 2 and "H_MMENUA" or "H_MMENU0"
			v.draw(cursorx, cursory, v.cachePatch(pp), flags)
		end
		
		if itname
			v.drawString(x + 17 + 33 + 17, 190, itname..(itname2 and " / "..itname2 or ""), flags|V_ALLOWLOWERCASE, "thin-right")
		end
	end
end
PSO.lib.registerHudFunc(playerActionPaletteHUD)