-- Type counter: Allows you to change your Type (class)

------------------------------------------------------
-- 	TYPE MAIN
------------------------------------------------------
local choices = {
		"Change Type",
		"Type Skills",
		"Type Proficiency",
		"Equipment Loadouts",
		"Salon",
		"Leave",
}


local function drawTypeMenu(v, p, c, choice, timer)
	local w = 160
	local h = 80	
	local x = 160-(w/2)
	local y = 100-(h/2)
	
	v.drawFill(x, y, w, h, 253)
	
	v.drawString(160, y+4, "TYPE COUNTER", V_YELLOWMAP, "center")
	
	local cy = y+16
	for i = 1, #choices
		v.drawString(160, cy, (i == choice and "\x82"..">" or "")..choices[i], 0, "center")
		cy = $+10
	end
end

PSO.info.menus["type"] = {
	
	hudfunc = function(v, p, c, choices, choice, timer)
		drawTypeMenu(v, p, c, choice, timer)
	end,
	
	runfunc = function(p)
		local input = PSO.lib.playerInputDown
		
		if input(p, IPT_CONFIRM) == 1
		and p.pso.menu.choice <= 5
			
			p.pso.shoptype = p.pso.menu.choice
			-- open the next menu:
			local nextmenu = {"type_change", "type_skill", "type_extend", "type_loadout", "salon"}
			local nmenu = nextmenu[p.pso.menu.choice]
			
			if nmenu
				PSO.lib.changeMenu(p, nmenu, true, true)
			else
				PSO.lib.closeMenu(p)
			end
			
			return true
			
		elseif input(p, IPT_BACK) == 1
		or (input(p, IPT_CONFIRM) and p.pso.menu.choice == 6)
			S_StartSound(nil, sfx_mback, p)
			PSO.lib.closeMenu(p)
		end
	end,
	
	choices = choices,
}

------------------------------------------------------
-- 	LOADOUTS
------------------------------------------------------

local loadoutchoices = {
	"EQUIP LOADOUT",
	"SAVE GEAR AS LOADOUT",
	"DELETE LOADOUT",
	"CANCEL",
	
}

local loadoutload = {
	"Send current gear to storage and equip",
	"Equip loadout",
	"Cancel",
}

local confirmation = {
	"EQUIP LOADOUT?",
	"SAVE LOADOUT?",
	"DELETE LOADOUT?",
}

local loadoutconfirm = {
	"PROCEED",
	"BACK",
}

local function drawLoadouts(v, p, c, choice, timer)
	
	local x1 = 2
	local w1 = 72
	local h1 = 120
	local y1 = 100 - h1/2
	
	v.drawFill(x1, y1, w1, h1, 253|V_SNAPTOLEFT)
	v.drawString(x1+2, y1+2, "LOADOUTS", V_YELLOWMAP|V_SNAPTOLEFT)
	
	local ly = y1 + 12
	for i = 1, PSO.info.MAXLOADOUTS+1
		v.drawString(x1+2, ly, ((choice == i) and "\x82>" or "")..p.pso.menu.choices[i], V_SNAPTOLEFT)
		ly = $+8
	end

	if p.pso.loadoutload
		local w2 = 190
		local h2 = 40
		local bx = 200
		local x2 = bx - w2/2
		local y2 = 100 - h2/2
		v.drawFill(x2, y2, w2, h2, 253)
		v.drawString(bx, y2+2, "EQUIP LOADOUT", V_YELLOWMAP, "center")
		
		for i = 1, #loadoutload
			v.drawString(bx, y2+12+(i-1)*8, ((p.pso.loadoutload == i) and "\x82>" or "")..loadoutload[i], flag, "thin-center")
		end
	
	elseif p.pso.loadoutconfirm
		local w2 = 120
		local h2 = 30
		local bx = 200
		local x2 = bx - w2/2
		local y2 = 100 - h2/2
		v.drawFill(x2, y2, w2, h2, 253)
		v.drawString(bx, y2+2, confirmation[p.pso.loadoutchoice] or "INVALID TEXT", V_YELLOWMAP, "center")
		
		for i = 1, #loadoutconfirm
			
			v.drawString(bx, y2+12+(i-1)*8, ((p.pso.loadoutconfirm == i) and "\x82>" or "")..loadoutconfirm[i], 0, "center")
		end
	
	elseif p.pso.loadoutchoice
		local w2 = 160
		local h2 = 60
		local bx = 200
		local x2 = bx - w2/2
		local y2 = 100 - h2/2
		v.drawFill(x2, y2, w2, h2, 253)
		v.drawString(bx, y2+2, "LOADOUT SETTINGS", V_YELLOWMAP, "center")
		
		for i = 1, #loadoutchoices
			local flag = 0
			if (i == 1 or i == 3)
			and not (p.pso.loadouts and p.pso.loadouts[choice])
				flag = V_50TRANS
			end
			
			v.drawString(bx, y2+12+(i-1)*8, ((p.pso.loadoutchoice == i) and "\x82>" or "")..loadoutchoices[i], flag, "center")
		end
	else	-- draw loadout
		
		local w2 = 230
		local h2 = 170
		local x2 = 320 - 2 - w2
		local y2 = 100 - h2/2
		
		v.drawFill(x2, y2, w2, h2, 253|V_SNAPTOLEFT)
		
		
		-- draw the loadout info if possible!
		if p.pso.loadouts
		and p.pso.loadouts[choice]
			local loadout = p.pso.loadouts[choice]
			local pal = loadout[3]
			local page = p.pso.loadoutpage
			
			local arr1 = leveltime%10 < 5 and "<" or "< "
			local arr2 = leveltime%10 < 5 and ">" or " >"
			
			v.drawString(x2 + w2/2, y2+2, arr1.."LOADOUT INFO ("..page.."/4)"..arr2, V_YELLOWMAP|V_SNAPTOLEFT, "center")
			
			local typ = PSO.info.typenames[loadout[1]] or "INVALID TYPE?"
			v.drawString(x2+2, y2+12, "\x82TYPE: \x80"..typ, V_SNAPTOLEFT)
			
			if page == 4		-- type skills
				v.drawString(x2+2, y2+22, "TYPE SKILLS:", V_YELLOWMAP|V_SNAPTOLEFT)
				
				if not #loadout[2]
					v.drawString(x2+2, y2+32, "No type skills.", V_SNAPTOLEFT|V_ALLOWLOWERCASE)
				else
					local sy = y2+32
					local sx = x2
					
					for i = 1, #loadout[2]
						local sk = PSO.info.typeSkills[loadout[2][i]]
						
						if sk
							v.draw(sx+2, sy, v.cachePatch(sk.icon), V_SNAPTOLEFT)
							v.drawString(sx+20, sy+3, sk.name, V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")
						end
						sy = $+18
						
						if i == 7
							sy = y2+32
							sx = x2 + w2*5/9 -16
						end
					end
				end
					
				
			elseif page == 2		-- armour/units
			or page == 3
			
				v.drawString(x2+2, y2+22, "ARMOURS & UNITS: ("..(page-1).."/2)", V_YELLOWMAP|V_SNAPTOLEFT)

				local wy = y2+22+10
				if pal and pal[2]
					
					local startd = (page == 2 and 1 or 4)
					local endd = (page == 2 and 3 or 6)
					
					for i = startd, endd
						
						local h3 = 10
						
						local apal = pal[3]
						if apal[1]
							local dy = wy
							
							local armour = apal[i][1]
							local str
							
							if armour == -1	-- invalid item
								str = "\x85".."ARMOUR NOT FOUND"
								v.draw(x2+14, dy+2, v.cachePatch("H_NSTAR0"), V_SNAPTOLEFT)
								v.drawString(x2+34, dy+2, str, V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")
								
							elseif armour
								local info = PSO.lib.itemInfo(armour)
								str = PSO.lib.getItemName(armour)
									
								v.draw(x2+14, dy+2, v.cachePatch("H_NSTAR"..info.rarity), V_SNAPTOLEFT)
								v.draw(x2+24, dy+2, v.cachePatch(PSO.lib.getItemIcon(armour)), V_SNAPTOLEFT)
								v.drawString(x2+34, dy+2, str, V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")	
								
								-- now shall we deal with units...
								local slots = info.slots or 0
								local ux = x2+34
								local uy = dy+2+12
								h3 = $+2
								
								for j = 1, 6	--slots
									if j < 4
										h3 = $+10
									end
									
									local unit = apal[i][j+1]
									
									if unit == -1	-- unit is invalid
										v.drawString(ux, uy, "\x85".."UNIT NOT FOUND", V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")
									elseif unit
										local info = PSO.lib.itemInfo(unit)
										v.draw(ux-20, uy, v.cachePatch("H_NSTAR"..info.rarity), V_SNAPTOLEFT)
										v.draw(ux-10, uy, v.cachePatch(PSO.lib.getItemIcon(unit)), V_SNAPTOLEFT)										
										v.drawString(ux, uy, info.name, V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")
									
									else
										if j > slots
											v.drawString(ux, uy, "LOCKED", V_SNAPTOLEFT|V_50TRANS, "thin")
										else
											v.drawString(ux, uy, "---", V_SNAPTOTOP, "thin")
										end
									end
									
									if j == 3
										ux = $ + w2*5/9 -16
										uy = dy+2+10
									else
										uy = $ +10
									end
									
								end
							else
								str = "---"
								v.drawString(x2+34, dy+2, str, V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")	
							end			
						else
							v.drawString(x2+32, wy+3, "---", V_SNAPTOLEFT, "thin")
						end
						
						v.drawString(x2+2, wy + h3/2 -4, i, V_YELLOWMAP|V_SNAPTOLEFT)
						v.drawFill(x2+11, wy, 1, h3, 0|V_SNAPTOLEFT)
						wy = $+h3+4
						
					end
				end				
				
				
			elseif page == 1	-- weapons!
				
				v.drawString(x2+2, y2+22, "WEAPONS:", V_YELLOWMAP|V_SNAPTOLEFT)
				
				local wy = y2+22+10
				if pal and pal[2]
					
					for i = 1, PSO.info.MAXPALSIZE
						
						local h3 = 10
						
						if pal[2][i]
							local dy = wy
							for j = 1, 2
								
								local w = pal[2][i][j]
								local str
								
								if w == -1	-- invalid item
									str = "\x85".."WEAPON NOT FOUND"
									v.draw(x2+14, dy+2, v.cachePatch("H_NSTAR0"), V_SNAPTOLEFT)
								elseif w
									local info = PSO.lib.itemInfo(w)
									str = PSO.lib.getItemName(w)
									
									v.draw(x2+14, dy+2, v.cachePatch("H_NSTAR"..info.rarity), V_SNAPTOLEFT)
									v.draw(x2+32, dy+2, v.cachePatch(PSO.lib.getItemIcon(w)), V_SNAPTOLEFT)
									
								else
									str = "---"
									if j == 2	-- no point continuing.
										break
									end
								end
								
								v.drawString(x2+42, dy+2, str, V_SNAPTOLEFT|V_ALLOWLOWERCASE, "thin")
								
								if j == 2
									h3 = $+8
								end
								
								dy = $+8
							end
						else
							v.drawString(x2+32, wy+3, "---", V_SNAPTOLEFT, "thin")
						end
						
						v.drawString(x2+2, wy + h3/2 -4, i, V_YELLOWMAP|V_SNAPTOLEFT)
						v.drawFill(x2+11, wy, 1, h3, 0|V_SNAPTOLEFT)
						wy = $+h3+2
						
					end
				end
			end
			
		else
			v.drawString(x2 + w2/2, y2+2, "NO LOADOUT SAVED", V_YELLOWMAP|V_SNAPTOLEFT, "center")
		end
	end
end

PSO.info.menus["type_loadout"] = {
	
	openfunc = function(p)

		
		p.pso.menu.choices = {}
		for i = 1, PSO.info.MAXLOADOUTS
			p.pso.menu.choices[i] = "SLOT "..i
			if p.pso.loadouts and p.pso.loadouts[i]
				-- update loadouts listing...
				PSO.lib.playerValidateLoadoutItems(p, p.pso.loadouts[i])
			end
		end
		p.pso.menu.choices[PSO.info.MAXLOADOUTS+1] = "BACK"
		
		p.pso.loadoutpage = 1	-- loadout information
		p.pso.maxloadoutpage = 4	-- weapons, armour (1), armour (2), typeskills
		
		p.pso.loadoutconfirm = nil
		p.pso.loadoutchoice = nil
		p.pso.loadoutload = nil
		
		p.pso.menu.minchoice = 1
		p.pso.menu.maxchoice = PSO.info.MAXLOADOUTS+1
		p.pso.hidehud = true
		
	end,
	
	closefunc = function(p)
		p.pso.hidehud = nil
	end,
	
	scroll = 13,
	
	hudfunc = function(v, p, c, choices, choice, timer)
		drawLoadouts(v, p, c, choice, timer)
	end,
	
	runfunc = function(p)
		
		local input = PSO.lib.playerInputDown
		
		-- load options
		if p.pso.loadoutload
			if input(p, "down") == 1
				p.pso.loadoutload = $+1
				S_StartSound(nil, sfx_mhvr, p)
				if p.pso.loadoutload > #loadoutload
					p.pso.loadoutload = 1
				end
			elseif input(p, "up") == 1
				p.pso.loadoutload = $-1
				S_StartSound(nil, sfx_mhvr, p)
				if p.pso.loadoutload < 1
					p.pso.loadoutload = #loadoutload
				end
			
			elseif input(p, IPT_CONFIRM) == 1
			and p.pso.loadoutload < 3	
				
				PSO.lib.playerLoadLoadout(p, p.pso.loadouts[p.pso.menu.choice], p.pso.loadoutload == 1)
				
				S_StartSound(nil, sfx_mconf, p)
				PSO.lib.closeMenu(p)
				
			elseif input(p, IPT_BACK) == 1
			or input(p, IPT_CONFIRM) == 1 and p.pso.loadoutload == 3
				p.pso.loadoutload = nil
				S_StartSound(nil, sfx_mback, p)
			end
			return true
		end
		
		-- yes/no
		if p.pso.loadoutconfirm
			if input(p, "down") == 1
			or input(p, "up") == 1
				p.pso.loadoutconfirm = p.pso.loadoutconfirm == 1 and 2 or 1
				S_StartSound(nil, sfx_mhvr, p)
			elseif input(p, IPT_CONFIRM) == 1 and p.pso.loadoutconfirm == 1
				-- yes
				S_StartSound(nil, sfx_mconf, p)
				
				if p.pso.loadoutchoice == 2	-- save
					local loadout = PSO.lib.playerSaveLoadout(p)
					
					local slot = p.pso.menu.choice
					p.pso.loadouts = $ or {}
					p.pso.loadouts[slot] = loadout
				
				elseif p.pso.loadoutchoice == 3	-- delete
					p.pso.loadouts[p.pso.menu.choice] = nil
				end
				
				p.pso.loadoutconfirm = nil
				p.pso.loadoutchoice = nil
				
			elseif input(p, IPT_CONFIRM) == 1 and p.pso.loadoutconfirm == 2
			or input(p, IPT_BACK) == 1
				S_StartSound(nil, sfx_mback, p)
				p.pso.loadoutconfirm = nil
			end
			
			return true
		end
		
		-- loadout choice? what will you do?
		-- load/overwrite/delete/cancel
		if p.pso.loadoutchoice
			
			if input(p, "down") == 1
				p.pso.loadoutchoice = $+1
				S_StartSound(nil, sfx_mhvr, p)
				if p.pso.loadoutchoice > #loadoutchoices
					p.pso.loadoutchoice = 1
				end
			elseif input(p, "up") == 1
				p.pso.loadoutchoice = $-1
				S_StartSound(nil, sfx_mhvr, p)
				if p.pso.loadoutchoice < 1
					p.pso.loadoutchoice = #loadoutchoices
				end
			elseif input(p, IPT_CONFIRM) ==  1
			and p.pso.loadoutchoice < #loadoutchoices
				
				-- first option is load, if there is no loadout here we obviously cannot do that
				-- ditto for delete, what are you deleting, nerd
				if (p.pso.loadoutchoice == 1 or p.pso.loadoutchoice == 3)
				and not (p.pso.loadouts and p.pso.loadouts[p.pso.menu.choice])
					S_StartSound(nil, sfx_mnot, p)
					return true
				end
				
				if p.pso.loadoutchoice == 1
					p.pso.loadoutload = 1
				else
					p.pso.loadoutconfirm = 1
				end
				S_StartSound(nil, sfx_mconf, p)
			elseif input(p, IPT_BACK) == 1
			or input(p, IPT_CONFIRM) == 1 and p.pso.loadoutchoice == #loadoutchoices
				p.pso.loadoutchoice = nil
				S_StartSound(nil, sfx_mback, p)
			end
			
			return true
		end
		
		if input(p, "left") == 1
			p.pso.loadoutpage = $-1
			if p.pso.loadoutpage < 1
				p.pso.loadoutpage = p.pso.maxloadoutpage
			end
			S_StartSound(nil, sfx_mhvr, p)
		elseif input(p, "right") == 1
			p.pso.loadoutpage = $+1
			if p.pso.loadoutpage > p.pso.maxloadoutpage
				p.pso.loadoutpage = 1
			end
			S_StartSound(nil, sfx_mhvr, p)
		end
		
		-- confirming...
		if input(p, IPT_CONFIRM) == 1
		and p.pso.menu.choice <= PSO.info.MAXLOADOUTS
			p.pso.loadoutchoice = 1
			S_StartSound(nil, sfx_mconf, p)
			
		elseif input(p, IPT_BACK) == 1
		or input(p, IPT_CONFIRM) == 1 and p.pso.menu.choice > PSO.info.MAXLOADOUTS
			PSO.lib.changeMenu(p, "type")
			S_StartSound(nil, sfx_mback, p)
			return true
		end
		
	end,
}

------------------------------------------------------
-- 	TYPE EXTENSION
------------------------------------------------------

local extend_costs = {100, 400, 800, 1500, 3000}
local init_proficiency_mult = {
	[0] = 200,
	[1] = 150,
	[2] = 100,
}

local function drawTypeExtend(v, p, c, choice, timer)
	local w = 170
	local h = 150	
	local x = 4	--160-(w/2)
	local y = 4
	local align = V_SNAPTOTOP|V_SNAPTOLEFT
	
	v.drawFill(x, y, w, h, 253|align)
	
	v.drawString(x+3, y+4, "TYPE PROFICIENCY", V_YELLOWMAP|align)
	
	v.drawString(x+3, y+12, PSO.info.typenames[p.pso.type], V_YELLOWMAP|align)
	v.drawString(x+w-3, y+12, "TP: "..p.pso.typepoints[p.pso.type], V_YELLOWMAP|align, "right")
	
	v.drawFill(x+4, y+22, w-8, 2, 27|align)
	v.drawFill(x+4, y+22, w-7, 1, align)
	
	local cy = y+32
	for i = p.pso.menu.scroll+1, min(#p.pso.menu.choices, p.pso.menu.scroll+6)
		
		local name
		local desc
		local ico
		
		if i == #p.pso.menu.choices
			name = "BACK"
		else
			local wp = p.pso.menu.choices[i]
			name = PSO.info.weaponnames[wp]
			
			ico = "H_WICO"..PSO.info.weaponnums[(wp or 1)]
			
			if i == choice
				desc = PSO.lib.wordWrap(v, PSO.info.weaponDesc[wp] or "", w-8, V_ALLOWLOWERCASE, "thin")
				ico = $.."B"
			end

			if desc
				v.drawString(x+4, 102, desc, align|V_ALLOWLOWERCASE, "thin")
			end	
			
			-- get prof level
			local prof = p.pso.weaponproficiencies[p.pso.type][wp]

			-- extend menu
			if p.pso.extendask
			and i == choice
				
				v.drawFill(x+w+2, y, 320 - (x+w+4), 32, 253|align)
				v.drawString(x+w+6, y, "SELECT ACTION", V_YELLOWMAP|align)
				
				local baseprof = PSO.info.typestats[p.pso.type].weapons[wp]
				local profmult = init_proficiency_mult[baseprof] or 100
				--print(prof)
				local acts = {"RANK +", "RANK -"}
				local tpcost = {(extend_costs[prof+1] or 0)*profmult/100, (extend_costs[prof] or 0)*profmult/100}
				
				for j = 1, #acts
					
					local act = acts[j]
					local cost = tpcost[j]
					local tflag = 0
					
					if cost
						act = $.." ("..(j == 2 and "+" or "-")..cost.."TP)"
					else
						tflag = V_50TRANS
					end
					
					v.drawString(x+w+6, y+12 +10*(j-1), (p.pso.extendask == j and "\x82"..">" or "")..act, align|tflag)
				end
			end
			
			-- draw weapon prof
			local px = x+w - 10
			while prof
				v.draw(px, cy, v.cachePatch("H_STAR"..prof), align)
				px = $-10
				prof = $-1
			end
			
		end
		
		if ico
			v.draw(x+12, cy, v.cachePatch(ico), align)
		end
		
		v.drawString(x+24, cy, (i == choice and "\x82" or "")..name, align)
		cy = $+10
	end

	v.drawFill(x+4, y+92, w-8, 2, 27|align)
	v.drawFill(x+4, y+92, w-7, 1, align)
end

PSO.info.menus["type_extend"] = {
	
	openfunc = function(p)
		local weaponslist = {}
		local num = 0
		
		/*while PSO.info.weaponnames[2^num]
			weaponslist[#weaponslist+1] = 2^num
			num = $+1
		end*/
		for i = 1, #PSO.info.weaponOrder
			weaponslist[i] = PSO.info.weaponOrder[i]
		end
		
		weaponslist[#weaponslist+1] = 0
		
		p.pso.menu.choices = weaponslist
		p.pso.menu.minchoice = 1
		p.pso.menu.maxchoice = #weaponslist
		
	end,
	
	scroll = 6,
	
	hudfunc = function(v, p, c, choices, choice, timer)
		drawTypeExtend(v, p, c, choice, timer)
	end,
	
	runfunc = function(p)
		local input = PSO.lib.playerInputDown
		
		if p.pso.extendask
			
			-- frame perfect stupidity.
			if p.pso.menu.choice == #p.pso.menu.choices
				p.pso.extendask = nil
				return
			end
			
			if input(p, "up") == 1 or input(p, "down") == 1
				p.pso.extendask = (p.pso.extendask == 1 and 2 or 1)
				S_StartSound(nil, sfx_mhvr, p)
			
			elseif input(p, IPT_BACK) == 1
				p.pso.extendask = nil
				S_StartSound(nil, sfx_mback, p)
			
			elseif input(p, IPT_CONFIRM) == 1
				
				local wp = p.pso.menu.choices[p.pso.menu.choice]
				local prof = p.pso.weaponproficiencies[p.pso.type][wp]
				local baseprof = PSO.info.typestats[p.pso.type].weapons[wp]
				local profmult = init_proficiency_mult[baseprof] or 100
				local tpcost = {(extend_costs[prof+1] or 0)*profmult/100, (extend_costs[prof] or 0)*profmult/100}
				
				
				if not tpcost[p.pso.extendask]
					S_StartSound(nil, sfx_mnot, p)
					return true
				end
				
				-- rank UP
				if p.pso.extendask == 1
					-- check points...
					local cost = tpcost[1]
					if p.pso.typepoints[p.pso.type] >= cost --or 1
						p.pso.typepoints[p.pso.type] = $-cost
						p.pso.weaponproficiencies[p.pso.type][wp] = $+1
						
						p.pso.extendask = nil
						S_StartSound(nil, sfx_mconf, p)
					else
						S_StartSound(nil, sfx_mnot, p)
					end
					
				else
					local cost = tpcost[2]
					p.pso.typepoints[p.pso.type] = $+cost
					p.pso.weaponproficiencies[p.pso.type][wp] = $-1

					p.pso.extendask = nil
					S_StartSound(nil, sfx_mconf, p)
					
					-- we have to check the player's palettes and unequip weapons that shouldn't be there anymore
					local pal = p.pso.palette.layout
					local wpal = pal[2]
					
					for i = 1, PSO.info.MAXPALSIZE
						local slot = wpal[i]
						
						if slot
							for j = 1, 2
								local wep = slot[j]
								
								if wep and not PSO.lib.canEquip(p, wep)
									--print("Unequipped "..i..", "..j)
									wep.equipped = nil
									slot[j] = nil
								end
							end
						end
					end
					
					p.pso.forceupdateweapon = true
				end
			end
			
			return true
		end
		
		if input(p, IPT_CONFIRM) == 1
		and p.pso.menu.choice < #p.pso.menu.choices 
				
			p.pso.extendask = 1
			S_StartSound(nil, sfx_mconf, p)

		elseif input(p, IPT_BACK) == 1
		or input(p, IPT_CONFIRM) == 1 and p.pso.menu.choice == #p.pso.menu.choices 
			PSO.lib.changeMenu(p, "type")
			S_StartSound(nil, sfx_mback, p)
			return true
		end
	end,
	
	choices = choices,
}

------------------------------------------------------
-- 	TYPE SKILLS
------------------------------------------------------


local function renderTypeSkill(v, x, y, flags, skill, locked, learned, selected)
	
	local skinfo = PSO.info.typeSkills[skill]
	flags = $ or 0
	
	local patch = learned and "SK_WINS" or "SK_WIN"
	
	if selected 
		patch = $.."2"
	end
	
	v.draw(x, y, v.cachePatch(patch), flags)
	
	if skinfo
		
		local name = locked and "LOCKED (Lv"..(skinfo.level or "UNDEFINED")..")" or skinfo.name
		local ico = (not locked) and skinfo.icon or nil
		
		if ico
			v.draw(x+3, y+3, v.cachePatch(ico), flags)
		end
		v.drawString(x+19, y+3, name, flags|V_ALLOWLOWERCASE|(locked and V_50TRANS or 0)|(selected and V_YELLOWMAP or 0), "thin")
		
		if locked
		
		else
			local cost = skinfo.cost or 0
			for i = 1, cost
				v.draw(x+19+(i-1)*8, y+11, v.cachePatch("H_SKON"), flags)
			end
		end
		
	end
end

local function drawTypeSkills(v, p, c, choice, timer)
	local w = 312
	local h = 190	
	local x = 4	--160-(w/2)
	local y = 4
	local align = V_SNAPTOTOP|V_SNAPTOLEFT
	
	v.drawFill(x, y, w, h, 253|align)
	
	v.drawString(x+3, y+4, "ADD TYPE SKILLS", V_YELLOWMAP|align)
	
	v.drawString(x+3, y+12, PSO.info.typenames[p.pso.type], V_YELLOWMAP|align)
	
	local sx = x+w -10
	local used = PSO.lib.skillCosts(p)
	for i = 1, p.pso.maxskillpoints
		v.draw(sx, y+11, i <= used and v.cachePatch("H_SKOFF") or v.cachePatch("H_SKON"), align)
		sx = $-8
	end
	
	v.drawFill(x+4, y+22, w-8, 2, 27|align)
	v.drawFill(x+4, y+22, w-7, 1, align)
	
	local desc
	
	local cy = y+32
	for i = p.pso.menu.scroll+1, #p.pso.skillslist
		
		local row = p.pso.skillslist[i]
		
		if type(row) == "table"	-- skills
			
			local sx = x+2
			for j = 1, #row
				
				local cell = row[j]
				if cell
					local skinfo = PSO.info.typeSkills[cell]
					local locked = p.pso.level < (skinfo.level or 0)
					local learned = PSO.lib.hasSkill(p, cell)
					local selected = (i == p.pso.skillsely) and (j == p.pso.skillselx)
					
					renderTypeSkill(v, sx, cy, align, cell, locked, learned, selected)
					
					if selected
						if locked
							desc = "Locked skill.\nUnlocks at Lv"..(skinfo.level or 0)
						else
							desc = PSO.lib.wordWrap(v, skinfo.desc or "", w-8, V_ALLOWLOWERCASE, "thin")
						end
					end
					
					sx = $ + 91+13
				end
			end
			cy = $+24
		else	-- title
			
			v.drawString(x+4, cy, row, V_YELLOWMAP|align)
			cy = $+12
		end
		
		-- dont draw any more than we have to.
		if cy >= 152
			break
		end
		
		/*if i == #p.pso.skillslist
			name = "BACK"
		else
			local sk = PSO.info.typeSkills[p.pso.skillslist[i]]
			
			if not sk.level or p.pso.level >= sk.level
				
				name = sk.name
				
				if sk.cost
					local costx = x+w-10
					for j = 1, sk.cost
						v.draw(costx, cy, v.cachePatch("H_SKON"), align)
						costx = $-10				
					end
				end
				
				if i == choice
					desc = PSO.lib.wordWrap(v, sk.desc or "", w-8, V_ALLOWLOWERCASE, "thin")
				end
				
				if PSO.lib.hasSkill(p, p.pso.skillslist[i])
					v.draw(x+4, cy, v.cachePatch("H_EQ"), align)
				end
			else
				name = "??? (LV"..sk.level..")"
				tflag = V_50TRANS
				if i == choice
					local str = "Locked skill. Reach level "..sk.level.." to unlock it."
					desc = PSO.lib.wordWrap(v, str or "", w-8, V_ALLOWLOWERCASE, "thin")
				end
			end	

			if desc
				v.drawString(x+4, 102, desc, align|V_ALLOWLOWERCASE, "thin")
			end
			
		end
		v.drawString(x+20, cy, (i == choice and "\x82"..">" or "")..name, align|tflag)
		cy = $+10*/
	end
	
	v.drawFill(x, y+152, w, h-152, 253|align)
	v.drawFill(x+4, y+152, w-8, 2, 27|align)
	v.drawFill(x+4, y+152, w-7, 1, align)
	
	if desc
		v.drawString(x+4, 160, desc, align|V_ALLOWLOWERCASE|V_RETURN8, "thin")
	end
	
end

PSO.info.menus["type_skill"] = {
	
	openfunc = function(p)
		
		p.pso.hidehud = true
		
		p.pso.skillconf = nil
		
		-- get a list of all the skills we can use at our level
		p.pso.skillslist = PSO.info.typeSkillsGrid[p.pso.type]
		p.pso.skillselx = 1
		p.pso.skillsely = 2
		
	end,
	
	scroll = 6,
	
	hudfunc = function(v, p, c, choices, choice, timer)
		--drawTypeMenu(v, p, c, choice, timer)
		drawTypeSkills(v, p, c, choice, timer)
	end,
	
	runfunc = function(p)
		local input = PSO.lib.playerInputDown
		local grid = PSO.info.typeSkillsGrid[p.pso.type]

		-- build scroll info
		local scroll = 4
		local scrollstart, scrollback, maxscroll
		
		if scroll and #grid > scroll
			scrollstart = scroll/2
			scrollback = #grid - scroll/2
			maxscroll = #grid - scroll
		end
		
		if input(p, "down") == 1
						
			while 1
				p.pso.skillsely = $+1
				
				if p.pso.skillsely > scrollstart+1
					p.pso.menu.scroll = min(maxscroll, $+1)
				end
				
				if p.pso.skillsely > #grid
					p.pso.skillsely = 1
					p.pso.menu.scroll = 0
				end
				
				-- check where we end up...
				local row = grid[p.pso.skillsely]
				if type(row) == "table"
					p.pso.skillselx = min($, #row)	-- clamp value.
					break
				end
			end
			
			S_StartSound(nil, sfx_mhvr, p)
			
		elseif input(p, "up") == 1
						
			while 1
				p.pso.skillsely = $-1
				
				if p.pso.skillsely < scrollback
				and (p.pso.skillsely > scrollstart+1 or p.pso.menu.scroll)
					p.pso.menu.scroll = max(0, $-1)
				end
				
				if p.pso.skillsely < 1
					p.pso.skillsely = #grid
					p.pso.menu.scroll = maxscroll
				end
				
				-- check where we end up...
				local row = grid[p.pso.skillsely]
				if type(row) == "table"
					p.pso.skillselx = min($, #row)	-- clamp value.
					break
				end
			end
			
			S_StartSound(nil, sfx_mhvr, p)
		
		elseif input(p, "right") == 1
			
			local row = grid[p.pso.skillsely]
			
			p.pso.skillselx = $+1
			if p.pso.skillselx > #row
				p.pso.skillselx = 1
			end
			
			S_StartSound(nil, sfx_mhvr, p)
			
		elseif input(p, "left") == 1
			
			local row = grid[p.pso.skillsely]
			
			p.pso.skillselx = $-1
			if p.pso.skillselx < 1
				p.pso.skillselx = #row
			end
			
			S_StartSound(nil, sfx_mhvr, p)
		elseif input(p, IPT_CONFIRM) == 1
			
			local skill = grid[p.pso.skillsely][p.pso.skillselx]
			local skinfo = PSO.info.typeSkills[skill]
			
			if skinfo.level and p.pso.level < skinfo.level
				S_StartSound(nil, sfx_mnot, p)
				return
			end
			
			if PSO.lib.hasSkill(p, skill)
				PSO.lib.removeSkill(p, skill)
			elseif not PSO.lib.addSkill(p, skill)
				S_StartSound(nil, sfx_mnot, p)
				return
			end
			
			S_StartSound(nil, sfx_mconf, p)

		elseif input(p, IPT_BACK) == 1
			PSO.lib.changeMenu(p, "type")
			S_StartSound(nil, sfx_mback, p)
			p.pso.hidehud = nil
		end
		
		return true
	end,
}

------------------------------------------------------
-- 	TYPE CHANGE
------------------------------------------------------


PSO.info.menus["type_change"] = {
	
	openfunc = function(p)
		p.pso.typeyn = nil
	end,
	
	hudfunc = function(v, p, c, choices, choice, timer)
		
		local xpos = 2
		local ypos = 12
		local w = 128
		local h = 60
		local ochoice = choice
		
		v.drawFill(xpos, ypos, w, h, 253|V_SNAPTOLEFT)
		
		local cy = ypos+2+10
		
		if p.pso.typeyn
			v.drawString(xpos+2, ypos+2, "CONFIRM CHOICE?", V_YELLOWMAP|V_SNAPTOLEFT)
			choices = {{"Yes"}, {"No"}}
			ochoice = p.pso.typeyn
		else
			v.drawString(xpos+2, ypos+2, "CHANGE TYPE", V_YELLOWMAP|V_SNAPTOLEFT)
		end
		
		for i = 1, #choices
			
			if i == p.pso.type
				v.draw(xpos+2, cy, v.cachePatch("H_EQ"), V_SNAPTOLEFT)
			end
			
			v.drawString(xpos+12, cy, ((i == ochoice) and "\x82>" or "")..choices[i][1], V_SNAPTOLEFT)
			cy = $+8
		end
		
		if PSO.info.typedescs[choice]
		
			-- description
			local nw = 320 - xpos - w - 2 - 2
			v.drawFill(xpos + w + 2, ypos, nw, h, 253|V_SNAPTOLEFT)
			
			if p.pso.typeyn
				local str = "Changing type will reset the weapon palette and unequip all currently equipped gear"
				local dstr = PSO.lib.wordWrap(v, str, nw-8)
				v.drawString(xpos + w + 4, ypos+2, "\x85".."NOTICE\x80\n"..dstr, V_SNAPTOLEFT|V_ALLOWLOWERCASE|V_RETURN8)
				
			else
				local str = PSO.lib.wordWrap(v, PSO.info.typedescs[choice], nw-8)
				v.drawString(xpos + w + 4, ypos+2, "\x82".."TYPE INFORMATION\x80\n"..str, V_SNAPTOLEFT|V_ALLOWLOWERCASE|V_RETURN8)
			end
			
			-- stats:
			local stats = PSO.lib.playerPreviewStats(p, choice, p.pso.level)
			
			v.drawFill(xpos, ypos + h + 2, w, h, 253|V_SNAPTOLEFT)
			v.drawString(xpos+2, ypos + h + 4, "TYPE STATS", V_YELLOWMAP|V_SNAPTOLEFT)
			local staty = ypos + h + 4 + 8
			
			local statrefs = {"maxhp", "maxpp", "satk", "ratk", "tatk", "def"}
			local statnames = {"HP", "PP", "S.ATK", "R.ATK", "T.ATK", "DEF"}
			
			for i = 1, #statrefs
				
				v.drawString(xpos+43, staty, "\x82"..statnames[i]..":", V_SNAPTOLEFT, "right")
				v.drawString(xpos+47, staty, stats[statrefs[i]], V_SNAPTOLEFT)
				staty = $+8
			end
			
			-- proficiency:
			local chart = p.pso.weaponproficiencies[choice]
			local i = 1
			
			v.drawFill(xpos + w + 2, ypos + h + 2, nw, h, 253|V_SNAPTOLEFT)
			v.drawString(xpos + w + 4, ypos + h + 4, "TYPE WEAPON PROFICIENCY", V_YELLOWMAP|V_SNAPTOLEFT)
			
			local wy = ypos + h + 4 + 8
			while chart[PSO.info.weaponOrder[i]] ~= nil
				local xoffs = ((i-1)/4)*48
				local ico = "H_WICO"..PSO.info.weaponnums[PSO.info.weaponOrder[i]]
				v.draw(xpos + w + 4 + 8 + xoffs, wy, v.cachePatch(ico), V_SNAPTOLEFT)
				
				local plevel = "H_STAR"..chart[PSO.info.weaponOrder[i]]
				v.draw(xpos + w + 4 + 18 + xoffs, wy, v.cachePatch(plevel), V_SNAPTOLEFT)
				
				wy = $ + 10
				i = $+1
				
				if (i-1)%4 == 0
					wy = ypos + h + 4 + 8
				end
				
			end
		end
	end,
	
	runfunc = function(p)
		local input = PSO.lib.playerInputDown
		
		if p.pso.typeyn
			
			if input(p, "down") == 1
			or input(p, "up") == 1
				p.pso.typeyn = p.pso.typeyn == 1 and 2 or 1
				S_StartSound(nil, sfx_mhvr, p)
			
			elseif input(p, IPT_BACK) == 1
			or input(p, IPT_CONFIRM) == 1 and p.pso.typeyn == 2
				p.pso.typeyn = nil
				S_StartSound(nil, sfx_mback, p)
			
			elseif input(p, IPT_CONFIRM) == 1
			and p.pso.typeyn == 1
				S_StartSound(nil, sfx_mconf, p)

				PSO.lib.playerChangeType(p, p.pso.menu.choice)
				p.pso.typeyn = nil
			end
			
			return true
		end
		
		if input(p, IPT_CONFIRM) == 1
		and p.pso.menu.choice < 4
			
			-- that's already your type dimwit
			if p.pso.menu.choice == p.pso.type
				S_StartSound(nil, sfx_mnot, p)
				return true
			end
			
			S_StartSound(nil, sfx_mconf, p)
			p.pso.typeyn = 1
			return true
			
		elseif input(p, IPT_BACK) == 1
		or (input(p, IPT_CONFIRM) and p.pso.menu.choice == 4)
			S_StartSound(nil, sfx_mback, p)
			PSO.lib.changeMenu(p, "type")
		end
	end,
	
	choices = {
		{"Hunter", TYPE_HUNTER},
		{"Ranger", TYPE_RANGER},
		{"Force", TYPE_FORCE},
		--{"Vanguard", TYPE_VANGUARD},
		{"Back", 0},
	},
}
