-- for late joiners, teleport em to the checkpoint or the leader, whichever is available
local function playerFirstTeleport(p)
	if server.questdata
	and gamemap ~= PSO.info.LOBBY_MAP
		
		local ref = server.keycheckpoint
		local ref2 = PSO.lib.getLeader()
		
		if ref and ref.valid
							
			local tempx = ref.x + 128*cos(ref.angle)
			local tempy = ref.y + 128*sin(ref.angle)
			local temp = P_SpawnMobj(tempx, tempy, ref.z, MT_THOK)
			temp.state = S_INVISIBLE
			temp.tics = TICRATE
			P_MoveOrigin(p.mo, temp.x, temp.y, temp.z)
			p.mo.angle = R_PointToAngle2(p.mo.x, p.mo.y, ref.x, ref.y)
			
		elseif ref2 and ref2.mo and ref2.mo.valid
			P_MoveOrigin(p.mo, ref2.mo.x, ref2.mo.y, ref2.mo.z)
		elseif server.lastspawnpoint	
			local pp = server.lastspawnpoint
			
			P_MoveOrigin(p.mo, pp[1], pp[2], pp[3])
			p.mo.angle = pp[4]
			
			if pp[5] & 2
				p.mo.flags2 = $|MF2_OBJECTFLIP
			end
		end	-- if we're out of options, well fuck you.
	end
end 

-- Initialize all PSO-related variables within the player...
local function playerInit(p)
	if not p.pso
		
		--COM_BufInsertText(p, "sessionanalog 0")	-- experimental
		
		local typ = TYPE_HUNTER
		local info = PSO.info.typestats[typ]
		
		local inputs = {}
		for k,v in pairs(PSO.info.inputDefaults)
			inputs[k] = v
		end
		
		
		p.pso = {
			
			playtime = 0,
			type = typ,
			
			-- monies
			meseta = 5000,
			
			-- level stats
			level = 1,
			exp = 0,
			
			-- stats
			maxhp = info.hp,
			hp = info.hp,
			
			maxpp = info.pp,
			pp = info.pp,
			
			satk = info.satk,
			ratk = info.ratk,
			tatk = info.tatk,
			def = info.def,
			
			pb = 0,
			
			-- pp regen
			ppregen = info.ppregen,
			ppregentimer = 0,	-- recover 1 pp when regentimer > regen
			
			-- action speed
			actionspd = 100,
			
			-- action variable
			backpalette = nil,	-- use actions 3, 4 instead of 1, 2
			action = nil,		-- action ref being used
			taction = nil,		-- tech action being used
			actionnum = 0,		-- action num being used (this is mostly for the HUD)
			actionbutton = nil,	-- which button did we use for that? keep track of that for holding.
			actionweapon = nil,	-- which weapon is performing the action?
			actiontimer = 0,	-- how long has it been performed for?
			actionheld = false,	-- stays ture until the button is released, cannot be set to true again 
			actionhold = 0,		-- are we still holding the action button? (doesn't matter for every anim)
			actionspeed = nil,	-- multiplier for movement speed: forces cmd forward/sidemove to lower values (0-100)
			actioncombo = 0,	-- for actions performed back to back
			
			pottimer = 0,		-- timer used for potentials. continuously counts up, resets when switching weapons.
			
			-- targetting:
			target = nil,		-- targetted enemy / item
			
			-- inventory / inputs / palette
			inventory = {},
			storage = {},	-- player storage
			shop = {},		-- shop (generates per player)
			
			-- inputs & preferences
			inputs = inputs,	-- used to remap controls on a player to player baisis
			buttons = {},		-- buttons being pressed by the player, in correlation to inputs
			toggle_lockon = false,		-- toggle lockon mode
			toggle_apalette = false,	-- toggle action palette mode
			toggle_bpalette = false,	-- toggle back palette mode
			pref_raredrop = 7,			-- show rare drop anim on 7*
			pref_screenshake = 10,		-- 100% screen shake
			
			-- weapon stuff:
			weaponpos = {
				{nil, 0, 0, 0, 0, 0, 0},	-- ref, x, y, z, angle, rollangle, slopeangle
				{nil, 0, 0, 0, 0, 0, 0},	-- ref, x, y, z, angle, rollangle, slopeangle (secondary weapon)
				-- secondary weapon is for dual wielding, but is also used for stuff like twin sabers and TMGs
			},
			
			-- palette stuff:
			palette = {
				
				layout = {{}, {}, {}},
				itemnum = 1,
				weaponnum = 1,
				aweaponnum = 1,	-- active weapon num
				armornum = 1,
				open = 0,	-- animation for opening the palette. goes up to TICRATE/4
				column = 0,	-- current column
				scroll = 0,
			},
			
			-- menu info:
			menu = {
				
				-- current opened menu index
				ref = nil,
				choice = 1,
				
				timer = 0,
				
				-- 0: up/down, 1: left/right
				inputtype = 0,			
			},
			
			-- used when opening menus.
			-- hidehudtimer will go up to TICRATE/2 when hidehud is true, and back down to 0 when it's false.
			hidehud = false,
			hidehudtimer = 0,
			
			-- saves cursor position for menus
			-- eg: menumemory["main"] = 2
			menumemory = {},
		}
		
		-- ready weapon proficiencies:
		p.pso.weaponproficiencies = {}
		local types = {TYPE_HUNTER, TYPE_RANGER, TYPE_FORCE}
		for i = 1, #types
			p.pso.weaponproficiencies[types[i]] = {}
			
			for k, v in pairs(PSO.info.typestats[types[i]].weapons)
				p.pso.weaponproficiencies[types[i]][k] = v
			end
		end
		
		-- ready class skills (well, the fact that you don't have any)
		p.pso.typeskills = {}
		p.pso.typepoints = {}
		p.pso.maxskillpoints = 4
		for i = 1, #types
			p.pso.typeskills[types[i]] = {}
			p.pso.typepoints[i] = 0
		end	
		
		-- ready base weapon photon arts etc
		-- we will always default to having 2 available action per weapons
		-- for most weapons this is normal + 1 PA
		-- for techs this is barta and foie
		p.pso.photonarts = {}
		for k,v in pairs(PSO.info.weaponDefaultActions)
			p.pso.photonarts[k] = {}
			for i = 1, #v
				table.insert(p.pso.photonarts[k], v[i])
			end
		end
		
		p.pso.photonarts.techniques = {"foie", "barta", "zonde", "diga"}
		
		PSO.lib.initPlayerType(p, TYPE_HUNTER)	-- by default
		
		-- generate sections for player inventory & storage
		PSO.lib.playerSortInventory(p)
		PSO.lib.playerSortStorage(p)
		PSO.lib.rollPlayerShop(p)
		
		-- preferences:
		--PSO.lib.playerLoadPrefs(p)
		
		-- TP into the quest
		playerFirstTeleport(p)
		--print(PSO.lib.serializePlayer(p))
	end
	
	-- fuck OFF with your stupid controls
	p.pflags = $ & ~PF_ANALOGMODE
	--local lmaogalog = CV_FindVar("sessionanalog")
	--CV_StealthSet(lmaogalog, 0)
end

PSO.lib.registerPlayerFunc(playerInit)

rawset(_G, "render_hack", 0)
/*hud.add(function(v)
	render_hack = 0	--v.renderer()
end)*/

/*local cv_fuck = CV_RegisterVar({
	name = "ang",
	defaultvalue = 0,
	PossibleValue = CV_Unsigned,
})*/


-- used specifically for lockon
addHook("PlayerCmd", function(p, cmd)
	if p.pso and p.pso.lockon and p.pso.lockon.valid and p.mo and p.mo.valid
		cmd.angleturn =  R_PointToAngle2(p.mo.x, p.mo.y, p.pso.lockon.x, p.pso.lockon.y)/FRACUNIT
		
		-- calc aiming with grade school maths yay
		if p.PSO_chasecam == 0
		and CV_FindVar("alwaysmlook").value
		
			local target = p.pso.lockon or p.pso.target
			local dist = FixedHypot(p.mo.x - target.x, p.mo.y - target.y)
			local zdiff = ((target.z+(target.height/2)) - (41*p.height/48)) -- middle of target mobj
			cmd.aiming = R_PointToAngle2(0, 0, dist, zdiff) / FRACUNIT
		else
			cmd.aiming = 0
		end
	end
end)

local function playerUpdateWeapon(p)
	local mo = p.mo
	if mo and mo.valid
		-- Update weapon position
		-- handles sheathe position outside of actions too.
		-- note: weapon position is relative from the player's
		
		if p.pso.palette.layout[2][p.pso.palette.aweaponnum]
			
			if not PSO.lib.slotHasWeapon(p.pso.palette.layout[2][p.pso.palette.aweaponnum])
			and p.pso.palette.aweaponnum ~= p.pso.palette.weaponnum
				p.pso.palette.aweaponnum = p.pso.palette.weaponnum
			end
			
			p.pso.pottimer = $+1
			
			for i = 1, 2	--#p.pso.palette.layout[2][p.pso.palette.aweaponnum]
				
				-- if the player switched weapon and is not doing an action, switch the weapon!
				if (p.pso.forceupdateweapon or p.pso.palette.weaponnum ~= p.pso.palette.aweaponnum)
				and not p.pso.action
					for j = 1, 2
						if p.pso.weaponpos[j][1] and  p.pso.weaponpos[j][1].valid
							P_RemoveMobj( p.pso.weaponpos[j][1])
							p.pso.weaponpos[j][1] = nil
						end
					end

					if i == 1	-- we only need to do that once.
						PSO.lib.runHook("switchingWeapons", p)
					end
					
					p.pso.palette.aweaponnum = p.pso.palette.weaponnum
					p.pso.forceupdateweapon = nil
					p.pso.pottimer = 0

					if i == 1	-- we only need to do that once.
						PSO.lib.runHook("switchedWeapons", p)
					end
					
				end
				
				local weapon = p.pso.palette.layout[2][p.pso.palette.aweaponnum][i]
				if p.pso.ais	-- hack
					weapon = PSO.lib.getPlayerActiveWeapon(p)
				end
				
				if weapon
					local info = PSO.lib.itemInfo(weapon)	--PSO.info.items[weapon.ref]
					local paperflag = (not info.nopaper) and FF_PAPERSPRITE or 0
					
					if not p.pso.weaponpos[i][1] or not p.pso.weaponpos[i][1].valid	-- how.
						
						local wp = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
						wp.tics = -1
						wp.fuse = -1
						
						wp.sprite = info.sprite
						wp.frame = info.frame | paperflag
						wp.color = info.color or SKINCOLOR_NONE
												
						p.pso.weaponpos[i][1] = wp
					end
					
					if not (p.pso.action and p.pso.actionslot == i)
						
						local sheatheinfo = info.sheatheinfo or PSO.info.defaultSheathePosition[weapon.weapontype]
						
						-- update positions to sheathe the weapon:
						local sheatheangle
						if sheatheinfo[6] ~= nil then sheatheangle = sheatheinfo[6] else sheatheangle = p.drawangle+ANG1*180 end
						
						local x = sheatheinfo[1]*cos(p.drawangle+sheatheangle)
						local y = sheatheinfo[2]*sin(p.drawangle+sheatheangle)
						local z = FixedDiv(mo.height*sheatheinfo[3]/100, mo.scale)
										
						if not p.pso.action
							mo.noactiontimer = $ and $+1 or 1
						else
							mo.noactiontimer = 0
						end
						
						if info.weapontype ~= WPT_TECHMAG or mo.noactiontimer > 4
							p.pso.weaponpos[i][2] = x
							p.pso.weaponpos[i][3] = y
							p.pso.weaponpos[i][4] = z
							p.pso.weaponpos[i][5] = sheatheinfo[4]
							p.pso.weaponpos[i][6] = sheatheinfo[5]
						end
						
						-- pslope_t data
						p.pso.weaponpos[i][7] = {
							use = false,
							o = {},
							xy = 0,
							za = 0,
						}
					end
					
					-- move the weapon to its destination position
					local tx = mo.x + FixedMul(p.pso.weaponpos[i][2], mo.scale)
					local ty = mo.y + FixedMul(p.pso.weaponpos[i][3], mo.scale)
					local tz = mo.z + FixedMul(p.pso.weaponpos[i][4], mo.scale)
					
					if info.weapontype == WPT_TECHMAG	-- in that case, it moves.	
						
						local wp = p.pso.weaponpos[i][1]
						
						if not p.pso.hitlag
							
							tz = $ + FixedMul(mo.scale*4, cos(leveltime*ANG1*4))
							wp.momx = (tx - wp.x)/3
							wp.momy = (ty - wp.y)/3
							wp.momz = (tz - wp.z)/3
							wp.angle = p.drawangle + p.pso.weaponpos[i][5]
						end
						wp.tics = 2
						
					else
						P_MoveOrigin(p.pso.weaponpos[i][1], tx, ty, tz)
						p.pso.weaponpos[i][1].angle = p.drawangle + p.pso.weaponpos[i][5]
						p.pso.weaponpos[i][1].tics = 2
					end
					
					-- scale stuff
					if info.spritescale
						p.pso.weaponpos[i][1].spritexscale = info.spritescale
						p.pso.weaponpos[i][1].spriteyscale = info.spritescale
					end

					-- lol renderer is broken
					if render_hack == "software"
					and R_PointToAngle(tx, ty, tz) - p.pso.weaponpos[i][1].angle > 0
					and p.pso.weaponpos[i][1].frame & FF_PAPERSPRITE
						p.pso.weaponpos[i][1].rollangle = -p.pso.weaponpos[i][6]	-- this is retarded
					else
						p.pso.weaponpos[i][1].rollangle = p.pso.weaponpos[i][6]
					end
					
					-- update splats:
					local function updateFrame(wp, prefix)
						
						prefix = $ or ""
						local framestart = info[prefix.."frame"]
						local frameend = info[prefix.."frameanim"] or framestart
						local framedelay = info[prefix.."framedelay"] or 2
								
						local del = (frameend - framestart)
						local ff = framestart + (leveltime/framedelay)%(del+1)
						
						-- hack to make the actual equipped weapon invisible when ais is in use
						if p.pso.ais
							wp.sprite = SPR_NULL
							wp.frame = A
						else
							wp.sprite = info.sprite
							wp.frame = $ & ~FF_FRAMEMASK
							wp.frame = $|ff
						end
					end	
					
					if p.pso.weaponpos[i][7]
						local dta = p.pso.weaponpos[i][7]
						local wep = p.pso.weaponpos[i][1]
						if dta.use
							
							if not wep.floorspriteslope
								-- ready pslope_t.
								-- this is useful for most striking weapons.
								
								P_CreateFloorSpriteSlope(wep)
							end

							if info.combatsplatframe ~= nil	-- sheathed weapon is never in splat form
								updateFrame(wep, "combatsplat")
								
							elseif info.splatframe ~= nil
								updateFrame(wep, "splat")
							
							elseif info.combatframe ~= nil
								updateFrame(wep, "combat")
							
							else
								updateFrame(wep, "")
							end
							
							wep.frame = $ & ~FF_PAPERSPRITE
							wep.flags2 = $ | MF2_SPLAT
							wep.renderflags = $|RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD
							wep.floorspriteslope.o = dta.o
							wep.floorspriteslope.xydirection = dta.xy
							wep.floorspriteslope.zangle = dta.za
							wep.angle = 0
							PSO.lib.fixSplatScale(wep)
						else
							if wep.floorspriteslope
								-- unset data
								P_RemoveFloorSpriteSlope(wep)
								wep.flags2 = $ & ~MF2_SPLAT
								wep.renderflags = $ & ~(RF_FLOORSPRITE|RF_SLOPESPLAT|RF_NOSPLATBILLBOARD|RF_NOSPLATROLLANGLE)								
								wep.spritexscale = FRACUNIT
								wep.spriteyscale = FRACUNIT
								wep.frame = $ | paperflag
							end
								
							if p.pso.action and info.combatframe ~= nil
								
								 updateFrame(wep, "combat")
									
							elseif info.frame ~= nil
								updateFrame(wep, "")
							end		
						end
					end
				
				end
			end	
		else	-- aweaponnum invalid
			
			if p.pso.palette.aweaponnum ~= p.pso.palette.weaponnum
				p.pso.palette.aweaponnum = p.pso.palette.weaponnum
			end
		end
	end
end

-- Make the player slow as dicks and put them in a wheelchair:
local function playerMovement(p)
	if p.mo and p.mo.valid
		local mo = p.mo
		
		if p.pso.dodgeroll
			p.normalspeed = 0
			p.thrustfactor = 0
		elseif not p.pso.menu.ref
			p.normalspeed = FRACUNIT*14
			p.thrustfactor = skins[p.mo.skin].thrustfactor
		end
		p.runspeed = FRACUNIT*24
		p.jumpfactor = 0
		p.charability = 0
		p.charability2 = 0
		
		
		if p.pso.target
		and p.pso.target.valid
		and not p.pso.lockon
		and p.pso.target.flags & MF_ENEMY	-- don't lock on to other shit lol
		and PSO.lib.playerInputDown(p, IPT_LOCKON) == 1
			p.pso.lockon = p.pso.target
			-- force the button to be released if we use toggle lockon
			if p.pso.toggle_lockon
				PSO.lib.playerForceInputUp(p, IPT_LOCKON)
			end
		end
		
		if p.pso.lockon
		and not p.pso.menu.ref
		and not P_PlayerInPain(p)
			
			if not p.pso.lockon.valid
			or (not p.pso.toggle_lockon and not PSO.lib.playerInputDown(p, IPT_LOCKON))
			or (p.pso.toggle_lockon and PSO.lib.playerInputDown(p, IPT_LOCKON) == 1)
			or p.pso.lockon ~= p.pso.target	-- if that somehow happens
				p.pflags = $ & ~PF_FORCESTRAFE
				p.pso.lockon = nil
			else
				p.mo.angle = R_PointToAngle2(mo.x, mo.y, p.pso.lockon.x, p.pso.lockon.y)
				p.drawangle = mo.angle
				p.pflags = $|PF_FORCESTRAFE
			end
		end
	end
end	

local function playerActionFreeze(p)
	return 
		p.powers[pw_nocontrol] 
		or p.pso.menu.ref 
		or p.pso.palette.open 
		or P_PlayerInPain(p) 
		or p.pso.teleport 
		or p.pso.jumppad 
		or p.pso.dodgeroll
		or p.pso.stagger
		or p.pso.blowdown
		or p.pso.knockback
		or p.pso.condition == COND_FREEZE
		--or p.pso.condition == COND_SHOCK
end

local function playerWeaponActions(p)
	local weapon = PSO.lib.getPlayerActiveWeapon(p)
	local loc_weapon = PSO.lib.getPlayerHUDWeapon(p)
	
	if p.mo and p.mo.valid
		local mo = p.mo
		if weapon
			if not p.pso.action and not playerActionFreeze(p) and not (p.pso.condition == COND_SHOCK)
			or p.pso.action and PSO.info.photonArts[p.pso.action]
			and PSO.info.photonArts[p.pso.action].combo 
			and p.pso.actiontimer >= PSO.info.photonArts[p.pso.action].combo	-- in that case, we can perform an action.
			and weapon == loc_weapon
			and not playerActionFreeze(p) and not (p.pso.condition == COND_SHOCK)
				
				local offset = 0
				local hand = PSO.info.weaponHand[weapon.weapontype]
				
				if hand == HAND_BOTH	-- 2-handed weapons can have 6 actions
					offset = p.pso.backpalette and 3 or 0	-- set to 3 if we're holding back palette
				end	
					
				local input = {IPT_ATTACK, IPT_ATTACK2, IPT_ATTACK3}
				
				for i = 1 + offset, 3 + offset
					if weapon.actions[i]
					
						local butt = PSO.lib.playerInputDown(p, input[i - offset])
						
						if butt	-- == 1
							
							local action = weapon.actions[i]
							local actiondta = PSO.info.photonArts[action]
							
							if not actiondta return end	-- shouldn't happen but in case it does
							
							if p.pso.action 
							and not (actiondta.combo and p.pso.actiontimer >= actiondta.combo and p.pso.actionnum == i)
								--print("Cannot perform again")
								continue
							elseif p.pso.action
								p.pso.actioncombo = $+1
								--print("Combo!")
							end
							
							local ppcost = actiondta.cost or 0
							if weapon.pa_cost and ppcost > 1 --and actiondta.normal
								ppcost = max(1, $*weapon.pa_cost/100)
							end
							
							-- hook for pp cost...
							local hookdta = {ppcost}
							PSO.lib.runHook("PPCost", p, hookdta, action, actiondta)
							ppcost = hookdta[1]	-- if this value was modified we will get it from here!
							
							local triggeravenger = p.pso.artsavenger and (ppcost and not actiondta.normal)
							
							if ppcost
							and p.pso.pp < ppcost
							and not triggeravenger
								--print("Insufficient PP")
								if butt == 1
									PSO.lib.noPPNotif(p)
								end
								continue
							else
								
								if not triggeravenger
									p.pso.pp = $ - ppcost
								end
								
								-- hunter arts avenger
								if triggeravenger
									p.pso.artsavenger_on = true
									p.pso.artsavenger = nil

									S_StartSoundAtVolume(p.mo, sfx_cburst, 127)
												
									for i = 1, 8
										local s = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_SUPERSPARK)
										s.destscale = 1
										s.momx = P_RandomRange(-32, 32)*FRACUNIT
										s.momy = P_RandomRange(-32, 32)*FRACUNIT
										s.momz = P_RandomRange(-32, 32)*FRACUNIT
										s.colorized = true
										s.color = SKINCOLOR_RED
									end
								end
							end
							
							--print("Using action "..i)
							
							if actiondta.technique
								--print("Technique used, using weapon cast action...")
								p.pso.action = PSO.info.castArt[weapon.weapontype]
								p.pso.taction = action
								actiondta = PSO.info.photonArts[action]	-- make sure to re-read that data
								
								if PSO.lib.usingPB(p) == "fo_phantomcaster"
									p.pso.phantomdodge = TICRATE
								end
								
							else
								p.pso.action = action
								p.pso.taction = nil
							end
							
							p.pso.actionnum = i
							p.pso.actionbutton = input[i - offset]
							p.pso.actionweapon = weapon
							p.pso.actionslot = (hand == HAND_RIGHT and 2 or 1)	-- know which weapon slot we're using.
							p.pso.actiontimer = 1
							p.pso.actionfor = 0
							p.pso.actionspeedtimer = 100
							p.pso.actionheld = true
							p.pso.actionhold = 1
							p.pso.cancancel = nil
							
							if actiondta.speed ~= nil
								p.pso.actionspeed = actiondta.speed
								if PSO.lib.usingPB(p) == "fo_phantomcaster"
									p.pso.actionspeed = 100	-- keep full speed.
								end
								
								-- remove any speed we have above that value...
								
								-- let's figure out what our new max speed should be...
								local ang = R_PointToAngle2(0, 0, p.mo.momx, p.mo.momy)
								local maxmomx = FixedMul(p.mo.scale, FixedMul(p.normalspeed, cos(ang)))
								local maxmomy = FixedMul(p.mo.scale, FixedMul(p.normalspeed, sin(ang)))
								
								if maxmomx > 0 and mo.momx > maxmomx
								or maxmomx < 0 and mo.momx < maxmomx
									mo.momx = maxmomx
								end
								
								if maxmomy > 0 and mo.momy > maxmomy
								or maxmomy < 0 and mo.momy < maxmomy
									mo.momy = maxmomy
								end
								
							else
								p.pso.actionspeed = 100
							end
							
							break
						end
					end
				end
			end
			
			if p.pso.action	-- menu OK: we need to let the player finish their action
				-- we're performing an action!
				local action = PSO.info.photonArts[p.pso.action]
				
				if action
					
					local actionspeed = p.pso.actionspd or 100
					-- ra overclock
					if PSO.lib.usingPB(p) == "ra_overclock"
						actionspeed = $+50
					end
					
					local dta = {actionspeed}
					PSO.lib.runHook("attackSpeed", p, dta)
					actionspeed = dta[1]
					
					--print(actionspeed)
					
					if action.use
	
						p.pso.actionfor = $+1
						if p.pso.framedata
							CONS_Printf(p, p.pso.actionfor)
						end
												
						-- heavy arts: grants super armor mid PA
						if PSO.lib.hasSkill(p, "hu_heavyarts")
							p.pso.superarmor = true
						end
						
						if p.pso.target and p.pso.target.valid and not p.pso.lockon and p.pso.target.flags & MF_SHOOTABLE
						and p.PSO_chasecam	-- don't autoaim in 1st person
							p.drawangle = R_PointToAngle2(p.mo.x, p.mo.y, p.pso.target.x, p.pso.target.y)
						end
						
						-- know how many frames to advance...
						local diff = (p.pso.actionspeedtimer/100) - p.pso.actiontimer
						local oldactiontimer = p.pso.actiontimer
						
						--print(p.pso.actiontimer.." / "..p.pso.actionspeedtimer.." ("..actionspeed..")")
						
						if diff >= 0
							for i = 1, 1+diff
								
								if not PSO.lib.runHook("playerActionOverride", p, p.pso.actiontimer, p.pso.actionheld, p.pso.actionhold - (i -1), p.pso.actioncombo)
									
									if action.use(p, p.pso.actiontimer, p.pso.actionheld, p.pso.actionhold - (i -1), p.pso.actioncombo)
									or PSO.lib.runHook("playerAction", p, p.pso.actiontimer, p.pso.actionheld, p.pso.actionhold - (i -1), p.pso.actioncombo)
										
										PSO.lib.runHook("playerActionFinish", p)
										PSO.lib.resetPlayerAction(p)
										p.pso.superarmor = nil
										return
									end	
								end
								
								-- in case resetPlayerAction is called in override or whatnot...
								if p.pso.actiontimer == nil
									return
								end
								
								-- in case we changed the animation timer mid anim
								-- also make sure to reflect that on the speed timer tracker
								if p.pso.actiontimer ~= oldactiontimer
									p.pso.actionspeedtimer = p.pso.actiontimer*100
								end
								
								p.pso.actiontimer = $ + 1
								oldactiontimer = $+1
								
								p.pso.actionhold = PSO.lib.playerInputDown(p, p.pso.actionbutton)*actionspeed
								if p.pso.actionheld and not p.pso.actionhold
									p.pso.actionheld = false
								end
								
							end
						end
						
						p.pso.actionspeedtimer = $ + actionspeed
						
					end
				end
			end
		end
	end
end

freeslot("sfx_targ")
local function playerTargetting(p)
	
	local range = 512
	local ang = 60
	
	local weapon = PSO.lib.getPlayerActiveWeapon(p)
	local mo = p.mo
	
	if not mo or not mo.valid return end
	if p.pso.menu.ref return end	-- useless
	
	local prevtarget = p.pso.target
	
	p.pso.target = nil	-- remove the target each frame
	
	-- blinded: can't lock on / interact
	if p.pso.condition == COND_BLIND
		return
	end
	
	-- some weapons have higher/lower range
	
	local info = PSO.lib.itemInfo(weapon)
	if weapon
		if weapon.range
			range = weapon.range
		elseif weapon.angle
			ang = weapon.angle
		end
	end
	
	-- range x2.5 in lockon
	if p.pso.lockon
		range = $*5/2
	end
	
	-- much longer range for flying AIS segments
	if p.pso.ais and p.pso.aisflight
		range = $*4
	end
	
	-- if we are locked on, only check to make sure the object is in range.
	-- if it isn't, the rest of the code will take care of it.
	if p.pso.lockon and p.pso.lockon.valid
	and R_PointToDist2(p.mo.x, p.mo.y, p.pso.lockon.x, p.pso.lockon.y) < range*FRACUNIT
	and not (p.pso.lockon.flags & MF_NOCLIPTHING)	-- untargettable
	and not (p.pso.lockon.flags & MF_NOBLOCKMAP)	-- ditto
		p.pso.target = p.pso.lockon
		return
	end
	
	-- use searchblockmap to find drops or cards in range.
	-- priority is always:
	-- enemy > card > others > drops
	
	local dropfound
	local cardfound
	local tpfound
	local npcfound
	
	-- yeah that should be far enough maybe probably perhaps
	local closest = FRACUNIT*10000
	local objectfound
	
	if not p.pso.action	-- if attacking, attacks take priority on the current target
	and not p.pso.lockon
		searchBlockmap("objects", function(ref, found)
			
			if found and found.valid and abs(ref.z - found.z) < ref.scale*128
				
				local dist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
				if dist < closest
					
					if (found.type == MT_PSO_UNLOCKCARD and found.extravalue1)		-- make sure it's visible.
					or (found.type == MT_PSO_CARDREADER and not found.extravalue2)	-- make sure it's not already on
						--cardfound = found
						objectfound = found
						closest = dist
					
					elseif found.type == MT_PSO_DROP
					and (not found.collected or not (found.collected & (2^(#p))))
					and (not found.psoitems or found.psoitems[#p+1])
					and not dropfound
					--and dist <= found.radius
						--dropfound = found
						objectfound = found
						closest = dist
					
					elseif found.type == MT_PSO_JUMPPAD
						--tpfound = found	-- hack
						objectfound = found
						closest = dist
					
					elseif found.type == MT_PSO_AIS_INTERACT
						objectfound = found.target
						closest = dist
					
					elseif (found.type == MT_PSO_TELEPORTER or found.type == MT_PSO_BOSSTELEPORTER)
					and not found.hold
						--tpfound = found
						objectfound = found
						closest = dist
					
					elseif found.type == MT_PSO_NPC_INTERACT
						--npcfound = found
						objectfound = found.target
						closest = dist
						
					elseif found.type == MT_PSO_DATAPOD
						objectfound = found
						closest = dist
					
					end
				elseif dist == closest	-- somehow the objects are at the exact same coordinates...
					
					-- prioritise cards and then drops
					if (found.type == MT_PSO_UNLOCKCARD and found.extravalue1)		-- make sure it's visible.
					or (found.type == MT_PSO_CARDREADER and not found.extravalue2)	-- make sure it's not already on					
						cardfound = found
					
					elseif found.type == MT_PSO_DROP
					and (not found.collected or not (found.collected & (2^(#p))))
					and (not found.psoitems or found.psoitems[#p+1])
					and not cardfound
					--and dist <= found.radius
						dropfound = found				
					end
				end
			end
		end, mo)
		
		-- priority 1 is card
		if cardfound
			
			if cardfound ~= prevtarget
				S_StartSound(nil, sfx_targ, p)
			end
			
			p.pso.target = cardfound
			return
		end
		
		-- then drops
		if dropfound

			if dropfound ~= prevtarget
				S_StartSound(nil, sfx_targ, p)
			end
		
			p.pso.target = dropfound
			return
		end
		
		/*-- then teleporters
		if tpfound
			p.pso.target = tpfound
			return
		end
		
		-- and then npcs.
		if npcfound
			p.pso.target = npcfound.target
		end*/
		
		-- if we make it here, get the closest object to us!
		if objectfound

			if objectfound ~= prevtarget
				S_StartSound(nil, sfx_targ, p)
			end
		
			p.pso.target = objectfound
			return
		end
	end
	
	local enemies = {}
	-- use searchblockmap to find enemies within our range...
	
	local cx = mo.x + range/2 * FixedMul(mo.scale, cos(mo.angle))
	local cy = mo.y + range/2 * FixedMul(mo.scale, sin(mo.angle))
	
	local x1 = cx - (range/2 *mo.scale)
	local x2 = cx + (range/2 *mo.scale)
	local y1 = cy - (range/2 *mo.scale)
	local y2 = cy + (range/2 *mo.scale)
	
	local function search(ref, found)
		
		if found and (found.flags & (MF_ENEMY|MF_SHOOTABLE)) and found.type ~= MT_PLAYER
		
			if not PSO.info.enemies[found.type] or (not PSO.info.enemies[found.type].notarget and found.pso and found.pso.hp)
				
				-- check angle
				local angle_diff = (p.cmd.angleturn*FRACUNIT) - R_PointToAngle2(mo.x, mo.y, found.x, found.y)
				
				if (PSO.lib.angToInt(angle_diff) > 180)
					angle_diff = InvAngle(angle_diff)
				end
							
				if (PSO.lib.angToInt(angle_diff) < ang)
				and (found == p.pso.lockon or P_CheckSight(mo, found))
					enemies[#enemies+1] = found
				end
			end
		end
	end
	
	searchBlockmap("objects", search, mo, x1, x2, y1, y2)
	
	-- sort enemies by distance / angle... give them some sort of score.
	local scores = {}
	for i = 1, #enemies
		
		scores[i] = {i, 0}
		scores[i][2] = $ + FixedDiv(R_PointToDist2(mo.x, mo.y, enemies[i].x, enemies[i].y), mo.scale) / FRACUNIT / 4	-- angle matters more
		scores[i][2] = $ + abs(mo.z - enemies[i].z) / FRACUNIT	-- z diff matters too
		scores[i][2] = $ + abs((mo.angle - R_PointToAngle2(mo.x, mo.y, enemies[i].x, enemies[i].y)) / ANG1)
		
		-- cheat and make it the best enemy if we're locked on to it.
		if enemies[i] == p.pso.lockon
			scores[i][2] = -FRACUNIT
		end
		
		if not (enemies[i].flags & MF_ENEMY)	-- another damageable thing? crate? It's very low priority.
			scores[i][2] = $*4
		end
		--print("enemy "..i.." score: "..scores[i])
	end
	
	table.sort(scores, function(a, b) return a[2] < b[2] end)
	
	if scores and scores[1]
		local final_target = enemies[scores[1][1]]
		if final_target and final_target.valid
		
			p.pso.target = final_target
			
			if p.pso.target ~= prevtarget
				S_StartSound(nil, sfx_targ, p)
			end
		end
	end
end

local function playerPPRegen(p)
	
	local mo = p.mo
	if not mo or not mo.valid return end
	
	local ppwellup
	local divider = 1
	
	if p.pso.action and p.pso.taction
	and PSO.lib.hasSkill(p, "fo_ppwellup")
		ppwellup = true
	end
	
	-- 1.7: allow PP regen while performing 0-cost actions
	local action
	if p.pso.action
	and PSO.info.photonArts[p.pso.action]
	and PSO.info.photonArts[p.pso.action].cost
	or p.pso.taction	-- techniques always count as a cost.
		action = true
	end
	
	-- photon flare has a 200% pp regen rate
	if PSO.lib.usingPB(p) == TYPE_FORCE
		divider = 3
	end
	
	if action
	and not ppwellup
	or p.pso.photondash
		p.pso.ppregentimer = 0
	else
		
		local regentarget = 1000*PSO.info.typestats[p.pso.type].ppregen
		local inc = p.pso.ppregen * divider / (ppwellup and 2 or 1)
		-- 20% increase w/ shifta
		if p.pso.shifta
			inc = $ + inc*PSO.info.SHIFTA/100
		elseif p.pso.jellen
			inc = $ - inc*PSO.info.JELLEN/100
		end
		
		p.pso.ppregentimer = $+inc
		
		while p.pso.ppregentimer >= regentarget
			p.pso.ppregentimer = $ - regentarget
			p.pso.pp = min(p.pso.maxpp, $+1)
		end
	end
	
	if not server.questdata
	or gamemap == PSO.info.LOBBY_MAP
	or p.pflags & PF_GODMODE
	or server.psodps	-- dps test
		
		if server.psodps
			server.psodpspp = $ + max(0, p.pso.maxpp-p.pso.pp)	-- check how much pp we ate last frame
		end
		
		p.pso.pp = p.pso.maxpp
		p.pso.hp = p.pso.maxhp
		p.pso.injury = nil
	end
	
end

local function playerHideHUD(p)
	if p.pso.hidehud
		p.pso.hidehudtimer = min(TICRATE/2, $+1)
	else
		p.pso.hidehudtimer = max(0, $-1)
	end
end

-- builds the quick palette
local function playerBuildQuickPalette(p)
	local vispal = {}
	local pal = p.pso.palette.layout
	-- build what's on the palette...
	local start = {p.pso.palette.itemnum, p.pso.palette.weaponnum, p.pso.palette.armornum}
	-- we need to go back by 2 indices (loop back to PALSIZE if we go up)
		
	for col = 1, #pal
			
		local cur = start[col]
		vispal[col] = {}
			
		vispal[col][3] = {pal[col][cur], cur}
		-- make it go back...
		local filled = 0
		local jackshit = 0
		
		while filled < 2 and jackshit < PSO.info.MAXPALSIZE
			cur = $-1
				
			if cur < 1
				cur = PSO.info.MAXPALSIZE
			end
			
			if pal[col][cur]
				filled = $+1
				jackshit = 0
				vispal[col][3-filled] = {pal[col][cur], cur}
			else
				jackshit = $+1
			end
		end
		
		-- and forward
		cur = start[col]
		filled = 0
		jackshit = 0
		
		while filled < 2 and jackshit < PSO.info.MAXPALSIZE
			cur = $+1
				
			if cur > PSO.info.MAXPALSIZE
				cur = 1
			end
			
			if pal[col][cur]
				filled = $+1
				jackshit = 0
				vispal[col][3+filled] = {pal[col][cur], cur}
			else
				jackshit = $+1
			end
		end
	end
	
	p.pso.palette.qlayout = vispal
end

local function playerHandlePaletteSwitch(p)
	
	--print(leveltime..": "..tostring(PSO.lib.playerInputDown(p, IPT_PALETTE)))
	
	if p.pso.actionrelease
	and PSO.lib.playerInputDown(p, IPT_PALETTE)
		return
	end
	
	p.pso.actionrelease = nil	-- not needed anymore
	
	if p.pso.toggle_apalette
		if PSO.lib.playerInputDown(p, IPT_PALETTE) == 1
			p.pso.openpal = not p.pso.openpal and true or false
		end
	else
		p.pso.openpal = PSO.lib.playerInputDown(p, IPT_PALETTE)
	end
	
	local openpal = p.pso.openpal
	--print(openpal)
	if p.pso.menu.ref
		openpal = 0
	elseif not openpal
	and not p.pso.quickpalette	-- using items...
		
		-- weapon quick switch
		for i = 1, min(#p.pso.buttons.mask, PSO.info.MAXPALSIZE)
			local input = p.pso.buttons.mask[i]
			
			if input == 1 and i ~= p.pso.palette.weaponnum
				if not p.pso.palette.layout[2][i]
					S_StartSound(nil, sfx_mnot, p)
					return
				end
				
				S_StartSound(nil, sfx_mhvr, p)
				
				p.pso.palette.weaponnum = i
				p.pso.backpalette = false
				return
			end
		end
	end
	
	-- okay, this is a hack to toggle photon blasts.
	
	if PSO.lib.playerHasPB(p)
	and p.pso.pb >= PSO.info.MAXPB
	and PSO.lib.playerInputDown(p, IPT_SECONDPALETTE) 
	and PSO.lib.playerInputDown(p, IPT_PALETTE) == 1	-- we just pressed
	and not p.pso.usepb
		--print("USE")
		PSO.lib.playerUsePB(p)
		p.pso.backpalette = false	-- for toggle
		p.pso.actionrelease = true	-- wait for this to be nil.
		
		return
	end
	
	p.pso.palette.scroll = $*3/4
	
	if openpal
		p.pso.palette.open = min($+1, TICRATE/4)
		
		if p.pso.palette.open == 1
			-- sets qlayout (not saved)
			p.pso.palette.column = 2	-- default to weapons
			playerBuildQuickPalette(p)
		end
		
		local inputs = {IPT_ATTACK2, IPT_ATTACK, IPT_ATTACK3}
		local vals = {"itemnum", "weaponnum", "armornum"}
		
		for i = 1, #inputs
			if PSO.lib.playerInputDown(p, inputs[i]) == 1
			and not p.pso.palette.scroll
				
				if p.pso.palette.column ~= i
					p.pso.palette.column = i
				else
					--print(p.pso.palette.itemnum)
					
					if p.pso.palette.qlayout[i][4] ~= nil
						p.pso.palette[vals[i]] = p.pso.palette.qlayout[i][4][2]	-- 4 is always the next one with how things are laid out.
					end
					
					--print(p.pso.palette.itemnum)
					--print("---")
					p.pso.palette.scroll = 16
					
					-- rebuild the palette layout
					playerBuildQuickPalette(p)
				end
			end
		end
		
		--print(p.pso.palette.open)
	else
		
		if p.pso.palette.open == (TICRATE/4 -1)
			if p.pso.palette.column == 1	-- consumables
				-- use em!
				local it = p.pso.palette.layout[1][p.pso.palette.itemnum]
				
				if it
					local itinfo = PSO.lib.itemInfo(it)
					
					-- either there's no check or it returns true
					if not itinfo.canuse or itinfo.canuse(p)
						local invslot = PSO.lib.findPlayerItem(p, it)
						-- if invslot is 0 then we don't have any of these.
						if not invslot
							PSO.lib.dprint("No "..it.." in inventory!")
						else
							-- if we made it here then we can use it!
							PSO.lib.playerLoseItem(p, invslot, 1)	-- use the item up
							
							if itinfo.use
								itinfo.use(p)
							end
							PSO.lib.dprint("Used "..it)
						end
					else
						PSO.lib.dprint("Cannot use item "..it)
					end
				end
			
			elseif p.pso.palette.column == 3	-- armours
				-- recalculate stats for units!
				PSO.lib.playerCalcStats(p, p.pso.level)
			end
		end
		
		p.pso.palette.open = max(0, $-1)
	end
end

-- pickup drops, keycards and interact w/ other items...
local function playerPickupDrops(p)
	
	-- note: the target can only be set if the drop can be SEEN by the player anyway
	-- there are sanity checks preventing getting it again regardless.
	if p.pso.target and p.pso.target.valid
	and PSO.lib.playerInputDown(p, IPT_INTERACT) == 1
	and not p.pso.action
	and not p.pso.menu.ref	
	and not p.pso.stagger
	and not p.pso.knockback
	and not p.pso.blowdown	-- this fixes a bunch of bugs, really.
	and not p.pso.quest.completetimer	-- also fixes a dumb bug
	and not p.pso.teleport	-- softlock central
		if p.pso.target.type == MT_PSO_DROP
			if PSO.lib.playerPickUpDrop(p, p.pso.target)
				p.pso.target = nil	-- remove that
			end		
			return true
		elseif p.pso.target.type == MT_PSO_UNLOCKCARD
			PSO.lib.playerGetKeyCard(p, p.pso.target)
			p.pso.target = nil	-- remove that
			return true
			
		elseif p.pso.target.type == MT_PSO_CARDREADER
			if PSO.lib.activateCardReader(p, p.pso.target)
				p.pso.target = nil
			end
			return true
		
		elseif p.pso.target.type == MT_PSO_JUMPPAD
		and not playerActionFreeze(p)
		and not p.pso.knockback
		and not p.pso.blowdown		
		and P_TryMove(p.mo, p.pso.target.x, p.pso.target.y, true)	-- check if something is in the way
			PSO.lib.playerUseJumpPad(p, p.pso.target)
			p.pso.target = nil
			return true
		
		elseif p.pso.target.type == MT_PSO_AIS
			PSO.lib.openMenu(p, "boardais")
			p.pso.aistarget = p.pso.target
			p.pso.target = nil
			return true
		
		elseif p.pso.target.type == MT_PSO_BOSSTELEPORTER
		and P_TryMove(p.mo, p.pso.target.x, p.pso.target.y, true)	-- check if something is in the way
			PSO.lib.openMenu(p, "bosstp")
			p.pso.teletarget = p.pso.target
			p.pso.target = nil
			return true
		
		elseif p.pso.target.type == MT_PSO_TELEPORTER
		and not playerActionFreeze(p)
		and P_TryMove(p.mo, p.pso.target.x, p.pso.target.y, true)	-- check if something is in the way
		and not p.pso.teleport
			if p.pso.target.target
				PSO.lib.softResetPlayer(p)
				p.pso.teleport = 1
				
				if p.mo.lastteleport and p.mo.lastteleport.valid
					p.pso.teleporttarget = p.mo.lastteleport
				else
					p.pso.teleporttarget = p.pso.target.target
				end
				p.mo.lastteleport = nil
				
				if p.pso.target.lobbytele 
					p.pso.teleporttype = 1
					p.mo.lastteleport = p.pso.target
				elseif p.pso.target.fromlobbytele
					
					if p.pso.usecheckpoint
						-- hack when respawning after dying to make walking back less of a slaugh
						p.pso.usecheckpoint = nil
						
						local ref = server.keycheckpoint
						if ref and ref.valid
							
							local tempx = ref.x + 128*cos(ref.angle)
							local tempy = ref.y + 128*sin(ref.angle)
							local temp = P_SpawnMobj(tempx, tempy, ref.z, MT_THOK)
							temp.state = S_INVISIBLE
							temp.tics = TICRATE
							
							p.pso.teleporttarget = temp
						end
					end
					
					p.pso.teleporttype = 2
				else
					p.pso.teleporttype = 0
				end
			end
			return true
		
		elseif p.pso.target.type == MT_PSO_NPC
			PSO.lib.NPCInteract(p, p.pso.target)
			return true
		
		elseif p.pso.target.type == MT_PSO_DATAPOD
			if p == consoleplayer
				local qinfo = PSO.lib.questInfo()
				local datapod = qinfo.datapods
				PSO.lib.spawnChatBox(p.pso.target, "DATAPOD", datapod[p.pso.target.extravalue2][4])
			end
			return true
		end
	end
end

local function playerHandleDamage(p)
	
	local mo = p.mo
	if p.pso.stagger
		mo.rollangle = 0

		
		if p.pso.stagger == 1
			local saveang = p.drawangle
			P_DoPlayerPain(p)
			p.drawangle = saveang
			P_InstaThrust(p.mo, 0, 0)
			if p.pso.ais
				p.mo.momz = 0
			else
				p.mo.momz = p.mo.scale*3
			end
			p.powers[pw_flashing] = TICRATE*5/2
			p.pso.stagger = nil
			return
		end
		
		if p.pso.ais and mo.aisvis and mo.aisvis.valid
			mo.aisvis.frame = I
			mo.z = max($, mo.floorz+FRACUNIT)
			mo.momz = 0
			P_InstaThrust(p.mo, 0, 0)
		end	
		
		p.pso.stagger = $+1
	end
	
	if p.pso.knockback
		mo.rollangle = 0
		
		p.powers[pw_flashing] = TICRATE/2
		p.normalspeed = 0
		
		if not P_IsObjectOnGround(mo)
			mo.momx = FixedMul(mo.momx, mo.friction)
			mo.momy = FixedMul(mo.momy, mo.friction)
			mo.state = S_PLAY_PAIN
		else
			local rad = mo.radius/FRACUNIT
			local d = P_SpawnMobj(mo.x + P_RandomRange(-rad, rad)*FRACUNIT, mo.y + P_RandomRange(-rad, rad)*FRACUNIT, mo.z, MT_SPINDUST)
			d.momz = P_RandomRange(1, 4)*mo.scale
			
			mo.state = S_PLAY_WALK
			mo.frame = A			
		end
		
		mo.momx = FixedMul(mo.momx, mo.friction)
		mo.momy = FixedMul(mo.momy, mo.friction)
		
		if FixedHypot(mo.momx, mo.momy) < mo.scale + (FixedHypot(p.cmomx, p.cmomy))
			p.pso.knockback = 0
		end
	end
	
	if p.pso.blowdown
	
		p.pso.blowdown = $+1
		mo.state = S_PLAY_PAIN
		p.powers[pw_flashing] = TICRATE
	
		if not P_IsObjectOnGround(mo)
			mo.blowdownmomz = mo.momz
			mo.rollangle = $ + ANG1*32
			mo.z = $ + mo.momz
		elseif p.pso.blowdown > 4
			
			if FixedHypot(mo.momx, mo.momy) < mo.scale*2
				p.pso.blowdown = nil
				mo.rollangle = 0
			else
			
				mo.momz = -(mo.blowdownmomz or 0)/3
				mo.momx = $/2
				mo.momy = $/2
				
				for i = 1, 8
					local d = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SPINDUST)
					d.angle = (360/8)*i*ANG1
					P_InstaThrust(d, d.angle, (mo.blowdownmomz or 0)*3)
					d.momz = -(mo.blowdownmomz or 0)/4
				end	
			end
		end
	end
end		

local sizes = {
	{FRACUNIT, FRACUNIT},
	{FRACUNIT/2, FRACUNIT*2},
	{FRACUNIT/4, FRACUNIT*4},
	{FRACUNIT/8, FRACUNIT*8},
}

local function playerHandleTeleport(p)
	if p.pso.teleport
		p.normalspeed = 0
		
		p.pso.teleport = $ + 1
		
		local index = p.pso.teleport
		if p.pso.teleport > 5
			index = 5-(p.pso.teleport-5)
		end
		
		if sizes[index]
			p.mo.spritexscale = sizes[index][1]
			p.mo.spriteyscale = sizes[index][2]
		end
		
		if p.pso.teleport == 5
			S_StartSound(nil, sfx_mixup, p)
			p.pflags = $|PF_NOCLIP	-- this helps the camera.
			
			-- special teleporter behaviour...
			p.mo.tpok = nil
			if p.pso.aistarget ~= nil
				P_LinedefExecute(4150, p.mo)	-- special case when activating ais that way
				PSO.lib.playerActivateAIS(p)
				p.pso.aistarget = nil
			end
			
			-- change sky/weather/music:
			if p.pso.teleporttype == 1			-- level -> lobby
				P_SetupLevelSky(PSO.lib.getLocalQuestProgress() >= 16 and 67 or 10, p)
				P_SwitchWeather(0, p)
				S_ChangeMusic("LOBBY", true, p)
				P_SetSkyboxMobj(nil, p)
				
			elseif p.pso.teleporttype == 2		-- lobby -> level
				local header = mapheaderinfo[gamemap]
				
				P_SetupLevelSky(header.skynum, p)
				P_SwitchWeather(header.weather, p)
				
				local dta = PSO.lib.questData()
				if dta
					local info = PSO.lib.questInfo()
					if info.questmusic
						S_ChangeMusic(info.questmusic, true, p)
					else
						S_ChangeMusic(header.musname, true, p)
					end
				else
					S_ChangeMusic(header.musname, true, p)
				end
				
				if PSO.lib.questDifficulty() >= DIFF_HARD
					if header.hardsky
						P_SetupLevelSky(header.hardsky, p)
					end
					
					if header.hardweather
						P_SwitchWeather(_G[header.hardweather] or 0, p)
					end	
					
					if header.hardmusic
						S_ChangeMusic(header.hardmusic, true, p)
					end
				end
				
				-- set skybox
				for m in mobjs.iterate()
					
					if m and m.valid and m.type == MT_SKYBOX
						P_SetSkyboxMobj(m, p)
						break
					end
				end
			end
			-- only teleport after, so linedef executors can be used to overwrite stuff
			P_SetOrigin(p.mo, p.pso.teleporttarget.x, p.pso.teleporttarget.y, p.pso.teleporttarget.z)
			
			PSO.lib.playerRoomAngleSet(p)	-- if we spawn inside a room of any sort, have us face whatever enemy is there.
		elseif p.pso.teleport == 8
			p.pflags = $ & ~PF_NOCLIP
			-- update pit coordinates
			p.pso.pitcoords = {p.mo.x, p.mo.y, p.mo.z, p.mo.angle}
			p.pso.bckuppitcoords = {p.mo.x, p.mo.y, p.mo.z, p.mo.angle}
			p.mo.lastteleportcoords = {p.mo.x, p.mo.y, p.mo.z, p.mo.angle}	-- for pits..............
		end
		
		if p.pso.teleport == 10
			p.pso.teleport = nil
		end
	end
end

local function playerBackPalette(p)
	
	if p.pso.toggle_bpalette
		
		if PSO.lib.playerInputDown(p, IPT_SECONDPALETTE) == 1
			p.pso.backpalette = not p.pso.backpalette and true or false
		end
	else
		p.pso.backpalette = PSO.lib.playerInputDown(p, IPT_SECONDPALETTE)
	end
end

local function playerHandleDeath(p)
	local mo = p.mo
	
	mo.state = S_PLAY_PAIN
	p.normalspeed = 0
	p.jumpfactor = 0
	
	p.pso.deadtimer = $ and $+1 or 1
	
	if p.pso.ais

		if leveltime%2 == 0
		and p.pso.deadtimer < TICRATE*3
			
			local boom = P_SpawnMobj(mo.x + P_RandomRange(-128, 128)*FRACUNIT, mo.y + P_RandomRange(-128, 128)*FRACUNIT, mo.z + P_RandomRange(0, 256)*FRACUNIT, MT_SONIC3KBOSSEXPLODE)
			boom.colorized = true
			boom.color = SKINCOLOR_EMERALD
			boom.renderflags = $|RF_FULLBRIGHT
			boom.tics = $ + P_RandomRange(0, 4)
			boom.scale = FRACUNIT*3 + P_RandomRange(-FRACUNIT/3, FRACUNIT/3)
			
			S_StartSound(mo, sfx_s3kb4)
		end
		
		local vis = mo.aisvis
		if vis and vis.valid
			
			if p.pso.deadtimer >= TICRATE*3
				local nt = p.pso.deadtimer - TICRATE*3
				
				local f = min(P, N + nt/3)
				vis.frame = f
				
			end
			
			if p.pso.deadtimer == TICRATE*5
				
				-- make variants of the AIS that get broken
				-- otherwise, just respawn
				
				local g = P_SpawnGhostMobj(vis)
				g.fuse = 6
				
				local flight = p.pso.aisflight
				p.powers[pw_flashing] = TICRATE*3 -1
				
				PSO.lib.playerDeactivateAIS(p, true, true)
				PSO.lib.playerStartAIS(p, flight)
				p.pso.deadtimer = nil
				return
				
			end
		end
		
	else
		if p.pso.deadtimer == TICRATE
			PSO.lib.openMenu(p, "skillissue")
		end
	end
end

-- mostly counts up the time if we're in the actual quest map lol
local function playerHandleQuestTrackers(p)
	
	if server.questdata
	and gamemap == server.questdata.map
	and p.pso.quest
		
		if p.pso.quest.timer_on
		and not (server.questdata and server.questdata.complete)
			p.pso.quest.time = $+1
		end
		
		if p.pso.quest.completetimer
			
			p.normalspeed = 0
			
			p.pso.quest.completetimer = $+1
			
			if PSO.lib.playerInputDown(p, IPT_CONFIRM) == 1
				
				if p.pso.quest.completetimer < TICRATE*3
					p.pso.quest.completetimer = TICRATE*3
				else
					p.pso.quest.completetimer = nil
					p.pso.quest.turnin = 1	-- can be turned in
				end
				
				local info = PSO.lib.questInfo()
				if info and info.playcredits
					p.pso.credits = 1	-- start credits after the quest complete screen if necessary
				end
				
				PSO.lib.playerForceInputUp(p, IPT_CONFIRM)
				PSO.lib.playerForceInputUp(p, IPT_EVADE)
			end
			
			if p.pso.quest.completetimer == TICRATE*3/2
				S_StartSound(nil, sfx_rnkup2, p)
			end
			
			if p.pso.quest.completetimer == TICRATE*3
				S_StartSound(nil, sfx_rank1 + (p.pso.quest.ranking -1), p)
			end
		end
	end
end
/*
addHook("PlayerThink", function(p)
	if p.pso and p.pso.jumppad
	and p.mo and p.mo.valid
	and p.mo.jumppad_next
		p.viewz = p.mo.jumppad_next[3]
	end
end)*/

local function playerJumpPad(p)
	if p.pso.jumppad
		
		local mo = p.mo
		
		--print("flags: "..(mo.flags & (MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING)))
					
		p.normalspeed = 0
		p.thrustfactor = 0
		mo.flags = $|MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT
		p.pflags = $|PF_NOCLIP

		p.pso.blowdown = nil
		p.pso.knockback = nil
		p.pso.stagger = nil	-- this should not happen but it can.
		
		p.pso.jumppadtimer = $+1
		if p.pso.jumppadtimer > TICRATE*10	-- this isn't normal

			P_MoveOrigin(p.mo, p.pso.jumppad.x, p.pso.jumppad.y, p.pso.jumppad.z)
			p.pso.jumppad = nil
			p.pso.jumppadtimer = nil
			p.thrustfactor = skins[p.mo.skin].thrustfactor
			mo.momx = 0
			mo.momy = 0
			mo.jumppad_next = nil

			mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
			p.pflags = $ & ~(PF_NOCLIP)
			
			return
		end
		
		-- emergency cancel
		if (mo.momz < 0 and mo.z <= p.pso.jumppad.z or not p.pso.hp)
			-- !?
			mo.z = p.pso.jumppad.z
			p.pso.jumppad = nil

			mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
			p.pflags = $ & ~(PF_NOCLIP)
			
			mo.momx = 0
			mo.momy = 0
			
			mo.jumppad_next = nil
			
			p.thrustfactor = skins[p.mo.skin].thrustfactor
			p.pso.jumppadtimer = nil
			return
		end

		if p.pso.jumppad.valid
			local dist = R_PointToDist2(mo.x, mo.y, p.pso.jumppad.x, p.pso.jumppad.y)
			local spd = FixedHypot(mo.momx, mo.momy)
			
			if dist < spd
			or p.pso.jumppadstop
				mo.momx = $*90/100
				mo.momy = $*90/100
				
				mo.flags = $ & ~(MF_NOCLIP|MF_NOCLIPTHING|MF_NOCLIPHEIGHT)
				p.pflags = $ & ~(PF_NOCLIP)
				
				p.pso.jumppadstop = true
				
				if FixedHypot(mo.momx, mo.momy) < mo.scale*2
				or P_IsObjectOnGround(mo)
				
					p.pso.jumppad = nil
					p.thrustfactor = skins[p.mo.skin].thrustfactor
					mo.momx = 0
					mo.momy = 0
					mo.jumppad_next = nil
					p.pso.jumppadtimer = nil
				end
			else
				-- noclip for players fucking sucks so we're gonna have to hack things.
				if mo.jumppad_next
					P_MoveOrigin(mo, mo.jumppad_next[1], mo.jumppad_next[2], mo.jumppad_next[3])
					--p.viewz = mo.jumppad_next[3]
				end
				
				if not P_IsObjectOnGround(mo)
					mo.jumppad_next = {
						mo.x + mo.momx,
						mo.y + mo.momy,
						mo.z + mo.momz,
					}
				else
					mo.jumppad_next = nil
				end

				if leveltime & 1

					if p.mo.aisvis and p.mo.aisvis.valid
						local g = P_SpawnGhostMobj(p.mo.aisvis)
						g.colorized = true				
						g.color = SKINCOLOR_MASTER
						g.flags = $|MF_NOCLIPHEIGHT|MF_NOCLIP
						P_MoveOrigin(g, mo.x, mo.y, mo.z)
					else
						local g = P_SpawnGhostMobj(p.mo)
						g.colorized = true
						g.flags = $|MF_NOCLIPHEIGHT|MF_NOCLIP
						P_MoveOrigin(g, mo.x, mo.y, mo.z)

						if g.tracer
							g.tracer.colorized = true
							g.tracer.tics = g.tics
							g.tracer.destscale = g.destscale
							g.tracer.scalespeed = g.scalespeed
							g.tracer.fuse = g.fuse
							g.tracer.momx = g.momx
							g.tracer.momy = g.momy
							g.tracer.flags = g.flags
							P_MoveOrigin(g.tracer, mo.x, mo.y, mo.z)
						end
						
					end
				end
					
				if mo.momz > -mo.scale*4
				and mo.momz < mo.scale*4
					mo.state = S_PLAY_ROLL
				elseif mo.momz < -mo.scale*4
					mo.state = S_PLAY_FALL
				end
				
			end
		end
	end
end

local function playerDodgeRoll(p, nododge)
	
	if p.pso.cancancel
		p.pso.cancancel = max(1, $-1)
	end
	
	if not p.pso.dodgeroll
	and PSO.lib.playerInputDown(p, IPT_EVADE) == 1
	and not (p.pso.quest and p.pso.quest.completetimer)

		if (P_IsObjectOnGround(p.mo) or p.pso.springbuffer)
		and not nododge
		and (not p.pso.action or p.pso.cancancel == 1)
		and p.normalspeed	-- menus etc
		--and not p.pso.jumppad
		--and not p.powers[pw_nocontrol]
		and not playerActionFreeze(p)
		
		-- reversal
		or PSO.lib.hasSkill(p, "all_justreversal")
		and (p.pso.stagger or p.pso.knockback or p.pso.blowdown)
			
			if p.pso.pp < PSO.info.dodgeroll_cost
				PSO.lib.noPPNotif(p)
				return
			end
			
			if p.pso.action
				PSO.lib.runHook("playerActionCancel", p)
			end
			
			PSO.lib.resetPlayerAction(p)	-- for cancelling
			
			p.pso.pp = $ - PSO.info.dodgeroll_cost
			
			p.pso.dodgeroll = TICRATE/3
			p.pso.stagger = 0
			p.pso.knockback = 0
			p.pso.blowdown = 0
			p.mo.rollangle = 0	-- make sure to reset
			
			if PSO.lib.hasSkill(p, "all_dodgeextension")
				p.pso.dodgeroll = $*3/2
			end
			
			S_StartSound(p.mo, p.pso.ais and sfx_beeyow or sfx_jump)
			
			local an = p.mo.angle
			-- angle depends on player input
			if FixedHypot(p.cmd.forwardmove*FRACUNIT, p.cmd.sidemove*FRACUNIT)
				an = $+R_PointToAngle2(0, 0, p.cmd.forwardmove*FRACUNIT, -p.cmd.sidemove*FRACUNIT)
			else
				an = p.drawangle
			end
				
			
			/*if not FixedHypot(p.mo.momx, p.mo.momy)
				an = p.mo.angle
			else
				an = R_PointToAngle2(0, 0, p.mo.momx, p.mo.momy)
			end*/
			
			P_InstaThrust(p.mo, an, p.mo.scale*(p.pso.ais and 50 or 20))
			if not p.pso.ais
				p.mo.momz = p.mo.scale*7
				
				p.mo.state = S_PLAY_ROLL
			end
		end
	end
	
	if p.pso.dodgeroll
		
		p.pso.dodgeroll = $-1
		if not p.pso.dodgeroll
			
			if PSO.lib.playerInputDown(p, IPT_EVADE) > 1
				p.pso.photondashbuffer = true
			end
			
			if not P_IsObjectOnGround(p.mo)
				p.mo.state = S_PLAY_FALL
			end
		else
			if leveltime & 1
				if p.mo.aisvis and p.mo.aisvis.valid
					local g = P_SpawnGhostMobj(p.mo.aisvis)
					g.colorized = true				
					g.color = SKINCOLOR_MASTER
				else
					local g = P_SpawnGhostMobj(p.mo)
					g.colorized = true
				end
			end	
		end
	elseif p.pso.photondashbuffer
		
		if not PSO.lib.playerInputDown(p, IPT_EVADE)
		or p.pso.action
		or p.pso.jumppad
		or not (p.cmd.forwardmove or p.cmd.sidemove)
		or not p.pso.pp
			p.pso.photondash = nil
			p.pso.photondashbuffer = nil
		else
			if P_IsObjectOnGround(p.mo)
			and p.pso.photondashbuffer
				p.pso.photondash = true
				--p.pso.photondashbuffer = nil
			end
			
			if p.pso.photondash
				p.normalspeed = FRACUNIT*(p.pso.ais and 38 or 26)
				p.runspeed = FRACUNIT*15
				
				if leveltime%6 == 0
					p.pso.pp = max(0, $-1)
				end
				
				if leveltime & 1
					if p.mo.aisvis and p.mo.aisvis.valid
						local g = P_SpawnGhostMobj(p.mo.aisvis)
						g.colorized = true				
						g.color = SKINCOLOR_MASTER
						g.tics = 2
					else
						local g = P_SpawnGhostMobj(p.mo)
						g.colorized = true
						g.tics = 2
						
						if g.tracer
							g.tracer.colorized = true
							g.tracer.tics = g.tics
						end
					end
				end
			end
		end
	end
end

local function noPPNotif(p)
	if p.pso.nopp
		p.pso.nopp = $-1
	end
end

local function playerTerrainDamage(p)
	local mo = p.mo
	
	-- sector damage
	if P_PlayerTouchingSectorSpecial(p, 1, 2)
	and P_IsObjectOnGround(mo)
		if leveltime%(TICRATE/2) == 0
		and not p.powers[pw_flashing]
			PSO.lib.damageObject(mo, nil, max(1, p.pso.maxhp/20))
		end
	end	
	
	-- space drowning...
	
	-- don't if we're in quest complete
	if p.powers[pw_spacetime] and server.questdata and server.questdata.complete
		p.powers[pw_spacetime] = TICRATE*20
		
	elseif p.powers[pw_spacetime] and p.powers[pw_spacetime] < 3
		p.powers[pw_spacetime] = 2
		
		if leveltime % (TICRATE) == 0
			PSO.lib.damageObject(mo, nil, max(1, p.pso.maxhp/11))
		end
		
	end
end

local function playerLowHP(p)
	local maxhp = p.pso.maxhp - p.pso.maxhp*(p.pso.injury or 0)/100
	if p.pso.hp <= maxhp*30/100
		p.pso.lowhp = $ and $+1 or 1
	else
		p.pso.lowhp = 0
	end
end

local max_slide_notifs = 3
local function playerHandleNotification(p)

	-- screen notifications
	if p.pso.notification
		p.pso.notification.timer = $-1
		
		if not p.pso.notification.timer
			p.pso.notification = nil
		end
	end
	
	-- push notifications
	if p.pso.slidenotifications
		
		for i = 1, #p.pso.slidenotifications
			
			local tt = p.pso.slidenotifications[i]
			
			if i > max_slide_notifs
				continue
			end
			
			if tt.slidetimer
				tt.slidetimer = $-1
				return
			end
			
			tt.timer = $-1
		end
		
		-- remove unwanted elements.
		local i = #p.pso.slidenotifications
		while i
			if p.pso.slidenotifications[i].timer <= 0
				table.remove(p.pso.slidenotifications, i)
			end
			i = $-1
		end
		
	end
end

local function playerSetQuest(p)
	
	local info = PSO.lib.questInfo()
	local timer_on_val = true
	if info and info.timer_off
		timer_on_val = false
	end
	
	p.pso.quest = {
		timer_on = timer_on_val,
		time = 0,
		damagetaken = 0,
		kills = {},	-- tracks kill per object type
		points = 0,	-- quest points (specific to some quests)
		incapacitations = 0
	}
end

local function playerAvengerVisuals(p)

		
	if PSO.lib.hasSkill(p, "hu_artsavenger")
	and p.pso.avengerguard
		
		if p.pso.avengerguard == 1
			local s = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_THOK)
			s.sprite = SPR_FORC
			s.frame = (leveltime%20 /2) | FF_TRANS50
			s.colorized = true
			s.color = SKINCOLOR_CRIMSON
			s.tics = -1
			s.fuse = 7
			s.scalespeed = FRACUNIT/2
			s.destscale = FRACUNIT*8		
		end
	
		if leveltime & 1
			p.mo.colorized = true
			p.mo.color = SKINCOLOR_CRIMSON
		else
			p.mo.colorized = false
			p.mo.color = p.skincolor
		end
	end
	
	if p.pso.avengerguard
		p.pso.avengerguard = $+1
		if p.pso.avengerguard > 10
			p.pso.avengerguard = 0
			p.mo.colorized = false
			p.mo.color = p.skincolor			
		end
	end
	
	if p.pso.artsavenger
	or p.pso.artsavenger_on
		
		if leveltime & 1
			local f = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_THOK)
			f.sprite = SPR_ELEC
			f.frame = P_RandomRange(0, 4)
			f.tics = 1
			f.colorized = true
			f.color = SKINCOLOR_RED
			f.scale = $*2/3
			
		end
	end
end	

local function playerHunterGale(p)
	local mo = p.mo
	if p.pso.huntergale	-- we can only have this IF we have the skill so this is a safe check
		
		
		local galedelay = 20
		local galehitboxdelay = TICRATE
		local galepotency = 20
		
		-- you're dead...
		if p.pso.hp <= 0
			p.pso.huntergale = 0
			return
		end
		
		if p.pso.huntergale >= PSO.info.MAXHUNTERGALE
			galedelay = 7
			galehitboxdelay = TICRATE/2
		end
		
		if p.pso.huntergale >= PSO.info.MAXHUNTERGALE/2
			if leveltime%galedelay == 0
				local gale = P_SpawnMobj(mo.x, mo.y, mo.z, MT_HUNTERGALE)
				gale.target = mo
				gale.fuse = TICRATE
				gale.flags2 = $|MF2_DONTDRAW
				gale.color = SKINCOLOR_CYAN
			end
			
			if leveltime%galehitboxdelay == 0
				local hb = PSO.lib.spawnHurtBox(mo.x, mo.y, mo.z-mo.scale*64, 255<<FRACBITS, 128<<FRACBITS, mo, galepotency, DMT_STRIKE, 3, RT_NONE)
				hb.pso.potency = galepotency	-- normals shouldn't benefit from PA%
				hb.pso.ppregen = 4		
			end
		end
		
		if p.pso.huntergaletimer
			p.pso.huntergaletimer = $-1
			if not p.pso.huntergaletimer
				p.pso.huntergale = 0
			end
		end
	end
end

freeslot("SPR_HUSB")
local function playerSwordbreaker(p)
	local mo = p.mo
	
	if p.pso.swordbreaker
	and (p.pso.swordbreakertimer > TICRATE*3 or leveltime&1)
		
		local a = leveltime*ANG1
		
		for i = 1, p.pso.swordbreaker
			local s = P_SpawnMobj(mo.x + 32*cos(a), mo.y + 32*sin(a), mo.z + mo.height/4, MT_THOK)
			s.tics = 1
			s.sprite = SPR_HUSB
			s.frame = A|FF_ADD|FF_PAPERSPRITE|FF_FULLBRIGHT
			s.angle = a+ANG1*90
			s.color = SKINCOLOR_CYAN
			
			a = $ + 360/3*ANG1
		end
	end
	
	if p.pso.swordbreakertimer
		p.pso.swordbreakertimer = $-1
		if not p.pso.swordbreakertimer
			p.pso.swordbreaker = nil
		end
	end
	
end

local function playerPhotonBlast(p)
	
	if p.pso.pb >= PSO.info.MAXPB
		if not p.pso.pbsound
			S_StartSound(p.mo, sfx_pblast, p)
			local rainbow = {SKINCOLOR_RED, SKINCOLOR_BLUE, SKINCOLOR_PINK, SKINCOLOR_GREEN, SKINCOLOR_YELLOW, SKINCOLOR_PURPLE}
			
			for i = 1, 16
				local spark = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z + p.mo.height/2, MT_SUPERSPARK)
				spark.colorized = true
				spark.color = rainbow[P_RandomRange(1, #rainbow)]
				spark.scale = $/3
				spark.momx = P_RandomRange(-16, 16)*p.mo.scale
				spark.momy = P_RandomRange(-16, 16)*p.mo.scale
				spark.momz = P_RandomRange(-16, 16)*p.mo.scale
			end
			
			p.pso.pbsound = true
		end
	else
		p.pso.pbsound = nil
	end
	
	if p.pso.usepb
		local drain = PSO.info.MAXPB / PSO.info.PBDURATION
		
		local usepb = PSO.lib.usingPB(p)
		
		if usepb ~= "hu_berserk"
		and not p.pso.blightround
			p.pso.pb = $ - drain
		end
		
		if usepb == "ra_blightround" and not p.pso.blightround
			p.pso.pb = $ - PSO.info.MAXPB/8
		end	
		
		-- visuals...
		if usepb == "hu_overload"

			if leveltime & 1
				local f = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_THOK)
				f.sprite = SPR_ELEC
				f.frame = P_RandomRange(0, 4)
				f.tics = 1
				f.colorized = true
				f.color = SKINCOLOR_RED
				f.scale = $*2/3
			
			end
			
			local s = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_THOK)
			s.sprite = SPR_FORC
			s.frame = (leveltime%20 /2) | FF_TRANS50
			s.colorized = true
			s.color = SKINCOLOR_CRIMSON
			s.tics = 1
			s.fuse = p.pso.pb+2
		
		elseif usepb == "hu_berserk"
			/*if leveltime & 1
				local f = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_THOK)
				f.sprite = SPR_ELEC
				f.frame = P_RandomRange(0, 4)
				f.tics = 1
				f.colorized = true
				f.color = SKINCOLOR_RED
				f.scale = $*2/3
			
			end*/
			
			if leveltime%5 == 0
				local f = P_SpawnMobjFromMobj(p.mo, P_RandomRange(-16, 16)*FRACUNIT, P_RandomRange(-16, 16)*FRACUNIT, 0, MT_THOK)
				f.sprite = SPR_BARD
				f.frame = A|FF_TRANS60
				f.tics = -1
				f.scale = $/3
				f.fuse = TICRATE*2/3
				f.destscale = p.mo.scale*4
				f.scalespeed = p.mo.scale/32
				f.colorized = true
				f.color = SKINCOLOR_RED
				f.momx = p.mo.momx
				f.momy = p.mo.momy
				f.momz = P_RandomRange(2, 3)*p.mo.scale
			end
		
		elseif usepb == "ra_overclock"
				
			local vars = {"raoc", "raoc2"}
			local zoffs = {p.mo.z + p.mo.height/5, p.mo.z + p.mo.height*4/5}
			local n = 24
			
			for i = 1, #vars
					
				local var = vars[i]
				local zpos = zoffs[i]
				
				if not p.mo[var] or not p.mo[var].valid
					p.mo[var] = P_SpawnMobj(p.mo.x, p.mo.y, zpos, MT_THOK)
					local oc = p.mo[var]
					
					oc.sprite = SPR_NULL
					
					oc.scale = 1
					oc.destscale = FRACUNIT
					oc.speed = -14
					
					oc.pellets = {}
					for j = 1, n
						oc.pellets[j] = P_SpawnMobj(oc.x, oc.y, oc.z, MT_THOK)
						oc.pellets[j].sprite = SPR_RAOC
						oc.pellets[j].frame = A|FF_ADD
						oc.pellets[j].renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD
						oc.pellets[j].fuse = PSO.info.PBDURATION +1
						
					end
					
				end
				
				local oc = p.mo[var]
				P_MoveOrigin(oc, p.mo.x, p.mo.y, zpos)
				
				for j = 1, n
					if oc.pellets[j] and oc.pellets[j].valid
						local dist = 32*oc.scale
						local a = oc.angle + (360/n)*j*ANG1
						local x = oc.x + FixedMul(dist, cos(a))
						local y = oc.y + FixedMul(dist, sin(a))
						
						P_MoveOrigin(oc.pellets[j], x, y, oc.z)
						oc.pellets[j].scale = oc.scale
						oc.pellets[j].rollangle = a+ANG1*90
						oc.pellets[j].tics = 2
					end
				end
				
				if oc.speed > 0
					oc.angle = $ + (oc.speed*(ANG1/4)*(i==1 and -1 or 1))
				end
				oc.tics = 2
				
				oc.speed = min(100, $ + 1)
			end
			
		elseif usepb == "fo_photonflare"
			
			if not p.mo.pbsigil or not p.mo.pbsigil.valid
				local s = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_THOK)
				s.sprite = SPR_SIGI
				s.frame = A|FF_ADD
				s.scale = 1
				s.destscale = FRACUNIT/2
				s.scalespeed = FRACUNIT/24
				s.renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD
				s.fuse = PSO.info.PBDURATION +10
				p.mo.pbsigil = s
			end
			
			local s = p.mo.pbsigil
			s.rollangle = $ + ANG1/2
			s.tics = 2
			
			P_MoveOrigin(p.mo.pbsigil, p.mo.x, p.mo.y, p.mo.z)
			
		elseif usepb == "fo_phantomcaster"
			if leveltime%5 == 0 and not p.pso.phantomdodge
				local g = P_SpawnGhostMobj(p.mo)
				g.destscale = p.mo.scale*2
				g.scalespeed = $/5
				g.fuse = 14
				g.momx = P_RandomRange(-1, 1)*FRACUNIT/8
				g.momy = P_RandomRange(-1, 1)*FRACUNIT/8

				if g.tracer
					g.tracer.colorized = true
					g.tracer.tics = g.tics
					g.tracer.destscale = g.destscale
					g.tracer.scalespeed = g.scalespeed
					g.tracer.fuse = g.fuse
					g.tracer.momx = g.momx
					g.tracer.momy = g.momy
				end
				
			end
		end
		
		if p.pso.pb <= 0
			p.pso.pb = 0
			p.pso.usepb = nil
			
			p.pso.injury = nil
		end
	end
	
	if p.pso.phantomdodge
		
		-- special dodge...
		p.pso.phantomdodge = $-1
		
		p.mo.frame = $ & ~FF_TRANSMASK
		p.mo.frame = $|FF_SUBTRACT|FF_TRANS30
		
		if p.pso.phantomdodge < 1
			p.mo.frame = $ & ~(FF_SUBTRACT|FF_TRANS30)
		end
	end
	
	if p.pso.phantomcast	-- free insta casts!!
		
		local ang = leveltime*ANG1*3
		local dist = 32
		
		p.mo.castballs = $ or {}
		
		for i = 1, p.pso.phantomcast
			
			local x = p.mo.x + dist*cos(ang)
			local y = p.mo.y + dist*sin(ang)
			
			local g = p.mo.castballs[i]
			if not g or not g.valid
				g = P_SpawnMobj(0, 0, 0, MT_THOK)
				p.mo.castballs[i] = g
				g.sprite = SPR_ARMA
				g.frame = FF_SUBTRACT|FF_TRANS30
				g.colorized = true
				g.color = SKINCOLOR_BLACK
				g.scale = $/3
				--g.renderflags = $|RF_SHADOWEFFECTS
			end
			
			g.tics = 3
			P_MoveOrigin(g, x, y, p.mo.z + FRACUNIT*16)
			
			ang = $ + 360/3*ANG1
		end
		
	end
end


local function playerAISSpawn_Ground(p)
	local t = p.pso.aisusetimer
	local mo = p.mo
	
	if p.pso.aisusetimer
	
	
		if t == 2
			S_StartSound(mo, sfx_becrsh)
			for i = 1, 16
				
				local s = P_SpawnMobj(mo.y, mo.y, mo.z, MT_SPINDUST)
				local a = 360/16 *i*ANG1
				
				P_InstaThrust(s, a, mo.scale*32)
			end
			
			mo.aisvis = P_SpawnMobj(mo.x, mo.y, mo.z + mo.scale*512, MT_THOK)
			local vis = mo.aisvis
			vis.tics = -1
			vis.sprite = SPR_BRAK
			vis.frame = X
			vis.angle = p.drawangle
			vis.momz = -mo.scale*32
			vis.scale = mo.scale*13/10
			
		elseif t == 6
			
			for i = 1, 12
				
				local sx = mo.x + P_RandomRange(-32, 32)*mo.scale
				local sy = mo.y + P_RandomRange(-32, 32)*mo.scale
				local s = P_SpawnMobj(sx, sy, mo.z, MT_SPINDUST)
				
				s.momz = P_RandomRange(3, 10)*FRACUNIT
				
			end
		elseif t == 18
		
			for i = 1, 16
				
				local s = P_SpawnMobj(mo.y, mo.y, mo.z, MT_SPINDUST)
				local a = 360/16 *i*ANG1
				
				P_InstaThrust(s, a, mo.scale*64)
			end
			
			mo.aisvis.frame = A
			mo.aisvis.momz = 0
			mo.aisvis.tics = 2
			p.pso.aisusetimer = nil
			mo.aisvis = nil
			
			PSO.lib.playerActivateAIS(p, p.pso.aisuseflight)
			S_StartSound(mo, sfx_befall)
			
			p.pso.aisuseflight = nil
			return
		end
		
		p.pso.aisusetimer = $+1
	end
end

local function playerAISVisuals(p)
	
	if p.pso.ais
		
		local mo = p.mo
		
		p.pso.shifta = 0
		p.pso.jellen = 0
		p.pso.deband = 0
		p.pso.zalure = 0
		
		mo.flags2 = $|MF2_DONTDRAW
		if mo.followmobj and mo.followmobj.valid
			mo.followmobj.flags2 = $|MF2_DONTDRAW
		end
		
		/*if not (p.pso.menu and p.pso.menu.ref)
			p.normalspeed = FRACUNIT*12
		end*/

		-- change radius/height acordingly
		mo.radius = FixedMul(mo.scale, mobjinfo[MT_PLAYER].radius *4)
		mo.height = FixedMul(mo.scale, mobjinfo[MT_PLAYER].height *4)		
		p.height = mo.height
		
		-- visual
		if not mo.aisvis or not mo.aisvis.valid
			mo.aisvis = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		end
		
		local vis = mo.aisvis
		vis.tics = -1
		vis.sprite = SPR_BRAK
		vis.frame = p.pso.stagger and I or (p.pso.aisflight and N or A)
		vis.scale = mo.scale*13/10
		vis.shadowscale = mo.scale*3/2
		
		-- put a fake point behind our player...
		local dist = 160	--CV_FindVar("cam_dist").value /FRACUNIT
		local height = 25	--CV_FindVar("cam_height").value /FRACUNIT
		-- my own cam values so that the result is consistent accross players.
		local a = mo.angle	--R_PointToAngle(mo.x, mo.y)
		
		local fcx = mo.x - (dist*7/2) *cos(a)
		local fcy = mo.y - (dist*7/2) *sin(a)
		local fakecam = P_SpawnMobj(fcx, fcy, mo.z + height*mo.scale*16, MT_THOK)
		fakecam.sprite = SPR_NULL
		fakecam.tics = 2
		
		if loc_camera and p == displayplayer
			loc_camera.momx = (fcx - loc_camera.x)/4
			loc_camera.momy = (fcy - loc_camera.y)/4
			loc_camera.momz = (fakecam.z - loc_camera.z)/4
			loc_camera.aiming = -ANG1*10
		end
		
		if p.pso.photondash or p.pso.dodgeroll
			vis.frame = N
		elseif mo.state == S_PLAY_WALK and not p.pso.aisflight
			local walkframe = B + (leveltime/8) %6
			vis.frame = walkframe
			
			if ((leveltime%48) == 24)
				S_StartSound(mo, sfx_bestep)
			elseif ((leveltime%48) == 0)
				S_StartSound(mo, sfx_bestp2)
			end
		end
		
		vis.angle = p.drawangle
		P_MoveOrigin(vis, mo.x, mo.y, mo.z)
		
		if (p.pso.dodgeroll or p.pso.photondash or p.pso.aisflight)
		and p.pso.hp
			vis.z = $ + mo.scale*16 + FixedMul(mo.scale, cos(leveltime*ANG1*20)*6)
		end
		
		-- handle wings
		local behind_dist = 80
		
		for i = 1, 2
			local rot = ANG1*20
			local dist = 120
			local ang = vis.angle + ANG1*90 + ANG1*180*(i-1)
			local zpos = vis.z + 160*mo.scale
			
			for j = 1, 2
				
				local wing = mo.wings[(i-1)*2 +j]
				local behindx = vis.x - behind_dist*cos(vis.angle)
				local behindy = vis.y - behind_dist*sin(vis.angle)
				
				local wx = behindx + dist*cos(ang)
				local wy = behindy + dist*sin(ang)
				P_MoveOrigin(wing, wx, wy, zpos)
				wing.angle = ang
				wing.rollangle = rot
				wing.spritexscale = vis.spritexscale
				wing.spriteyscale = vis.spriteyscale
				wing.flags2 = $ & ~MF2_DONTDRAW
				
				dist = $ + 80
				rot = $+ANG1*20
				zpos = $ + mo.scale*40
				
				if p.pso.dodgeroll
				or p.pso.photondash
					if leveltime & 1
						local g = P_SpawnGhostMobj(wing)
						if p.pso.photondash
							g.fuse = 2
						end

						if g.tracer
							g.tracer.colorized = true
							g.tracer.tics = g.tics
							g.tracer.destscale = g.destscale
							g.tracer.scalespeed = g.scalespeed
							g.tracer.fuse = g.fuse
							g.tracer.momx = g.momx
							g.tracer.momy = g.momy
						end
						
					end
				end
				
				if p.pso.aisflight
					if leveltime & 1
						local g = P_SpawnGhostMobj(wing)
						g.destscale = $*8
						g.momz = -mo.scale*8
						g.momx = P_RandomRange(-4, 4)*mo.scale
						g.momy = P_RandomRange(-4, 4)*mo.scale
					end
				end
			end
		end
		
	end
end

local function playerManageDPSCounter(p)
	
	if not (server.questdata and server.questdata.complete)
		if p.pso.dps_damage and not (p.pso.dps_resettimer and p.pso.dps_resettimer < TICRATE*10)
			p.pso.dps_time = $+1
			
			if p.pso.dps_time % (TICRATE/2) == 0
				
				p.pso.dps_display = p.pso.dps_damage*TICRATE / p.pso.dps_time
			end
		end
		
		if p.pso.dps_resettimer
			
			p.pso.dps_resettimer = $-1
			
			if p.pso.target and p.pso.target.valid and p.pso.target.flags & MF_ENEMY
				p.pso.dps_resettimer = TICRATE*15
			end
			
			if not p.pso.dps_resettimer
				PSO.lib.playerResetDPS(p)
			end
		end
	end
end

-- death pit handling
-- copypasted from srb2p lol
local function playerPitHandler(p)
	if not p.mo or not p.mo.valid return end
	
	p.pso.time_since_pit = $ and $+1 or 1
	
	-- makes it easier to reset things...
	if not p.pso.pitcoords
		p.pso.pitcoords = {p.mo.x, p.mo.y, p.mo.z, p.mo.angle}
		p.pso.backuppitcoords = {p.mo.x, p.mo.y, p.mo.z, p.mo.angle}
	end

	if not p.pso.pittime
	and leveltime% (TICRATE) == 0
	and P_IsObjectOnGround(p.mo)
	and not P_CheckDeathPitCollide(p.mo)
		-- every 5 seconds, save new positions
		if p.pso.pitcoords
			p.pso.backuppitcoords = {p.mo.x, p.mo.y, p.mo.z, p.mo.angle}
		end
		p.pso.pitcoords = {p.mo.x, p.mo.y, p.mo.z, p.mo.angle}
	end


	-- respawn handling
	if p.pso.pittime
		p.normalspeed = 0
		p.pso.pittime = $+1
		p.mo.flags2 = $ | MF2_DONTDRAW
		if p.pso.pittime > TICRATE
			
			--print("A")
			p.powers[pw_flashing] = TICRATE*2

			p.mo.state = S_PLAY_STND
			p.mo.rollangle = 0

			-- deal damage
			if p.pso.time_since_pit > TICRATE*3/2
				PSO.lib.damageObject(p.mo, nil, p.pso.maxhp/4)
			else -- something has gone wrong, use backup coords and do not penalize the player with more damage.]
				-- try to respawn at last "checkpoint"

				local ref = server.keycheckpoint
				if ref and ref.valid
							
					local tempx = ref.x + 128*cos(ref.angle)
					local tempy = ref.y + 128*sin(ref.angle)
					p.pso.pitcoords = {tempx, tempy, ref.z, ref.angle + ANG1*180}
				else	-- no checkpoint? do it at the player spawn, we've no choice lmao.
					
					--print(p.mo.spawnpoint)
					p.pso.pitcoords = {server.defaultspawnpoint[1], server.defaultspawnpoint[2], server.defaultspawnpoint[3], server.defaultspawnpoint[4]}
					
				end
			end
			
			if p.pso.pitcoords
				P_SetOrigin(p.mo, p.pso.pitcoords[1], p.pso.pitcoords[2], p.pso.pitcoords[3])
				p.mo.angle = p.pso.pitcoords[4]			
				PSO.lib.softResetPlayer(p)
			end

			p.pso.time_since_pit = 0
			
			if p.powers[pw_spacetime]
				p.powers[pw_spacetime] = TICRATE*11	-- reset this
			end

		end
	elseif not P_IsObjectOnGround(p.mo)
		p.pso.pitmomz = p.mo.momz
	end
	
	return p.pso.pittime
end

-- check if we hit a death pit etc
addHook("ShouldDamage", function(mo, inf, s, d, dtype)

	if (P_CheckDeathPitCollide(mo) or dtype == DMG_DEATHPIT)
		local p = mo.player
		if not p return false end
		if p.pflags & PF_GODMODE return false end

		if not p.pso.pittime
		and not p.pso.teleport	-- we're going to medbay!!
			local savecoords = p.pso.pitcoords
			PSO.lib.softResetPlayer(p)
			p.pso.pitcoords = savecoords
			p.pso.pittime = 1

			-- spawn a fake
			local fake = P_SpawnMobj(mo.x, mo.y, mo.z, MT_GHOST)	-- it has to be MT_GHOST, other mobjs just vanish on impact with pits
			fake.flags = MF_NOCLIPHEIGHT|MF_NOCLIP
			fake.skin = mo.skin
			fake.color = mo.color
			fake.state = S_PLAY_PAIN
			fake.tics = TICRATE*2
			fake.angle = mo.angle
			fake.momx = mo.momx
			fake.momy = mo.momy
			fake.momz = (p.pso.pitmomz) or 0
			fake.fuse = TICRATE*2

			--PLAY_nomove(mo.player)
			P_InstaThrust(mo, 0, 0)
			mo.state = S_PLAY_STND
			mo.tics = -1
			mo.sprite = SPR_NULL
			mo.frame = A

		end

		return false	-- don't get damaged
	end
end, MT_PLAYER)

local function playerSpringHandler(p)
	
	local mo = p.mo
	if mo.eflags & MFE_SPRUNG and not p.pso.springbuffer
		PSO.lib.resetPlayerAction(p)
		p.pso.springbuffer = true
	elseif p.pso.springbuffer
		local s = P_SpawnMobj(mo.x + P_RandomRange(-32, 32)*mo.scale, mo.y + P_RandomRange(-32, 32)*mo.scale, mo.z - mo.height/2, MT_SUPERSPARK)
		s.colorized = true
		s.color = SKINCOLOR_TEAL
	end
	
	if P_IsObjectOnGround(mo)
	or p.pso.action
		p.pso.springbuffer = nil
	end
end

-- handles injury status HP.
local function playerInjuryHandler(p)
	if p.pso.injury
		
		local maxhp = p.pso.maxhp - p.pso.maxhp*p.pso.injury/100
		p.pso.hp = min($, maxhp)
	end
end

-- hitlag: stops the player and then recovers momentum
-- cant just pause thinkers like in rr bc we just don't have access to that unfortunately 
local function playerHitlagHandler(p)
	
	if p.pso.hitlag
		
		if not p.pso.hitlagdata
			p.pso.hitlagdata = {
				momx = p.mo.momx,
				momy = p.mo.momy,
				momz = p.mo.momz,
				sprite = p.mo.sprite,
				sprite2 = p.mo.sprite2,
				frame = p.mo.frame,
				angle = p.mo.angle,
			}		
		end
		
		local dta  = p.pso.hitlagdata
		
		if p.pso.actiontimer
			p.pso.actionfor =$+1	-- hitlag counts for frame data too!!
		end
		
		p.mo.momx = 0
		p.mo.momy = 0
		p.mo.momz = 0
		p.mo.angle = dta.angle
		p.mo.sprite = dta.sprite
		p.mo.sprite2 = dta.sprite2
		p.mo.frame = dta.frame
		
		playerAISVisuals(p)
		
		p.pso.hitlag = $-1
		
		if p.pso.avengerhitlag
			for i = 1, P_RandomRange(1, 4)
				local a = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z +p.mo.height/4, MT_THOK)
				a.sprite = SPR_KLIT
				a.colorized = true
				a.color = SKINCOLOR_RED
				a.scale = p.mo.scale/2
				a.frame = P_RandomRange(0, P) |FF_FULLBRIGHT|TR_TRANS40|FF_ADD
				a.tics = 1
			end
			p.mo.colorized = true
		end
		
		if not p.pso.hitlag
			p.mo.momx = dta.momx
			p.mo.momy = dta.momy
			p.mo.momz = dta.momz
			p.pso.hitlagdata = nil
			p.pso.avengerhitlag = nil
			p.mo.colorized = false
			
			if P_IsObjectOnGround(p.mo)
				if FixedHypot(p.mo.momx, p.mo.momy)
					p.mo.state = S_PLAY_WALK
				else
					p.mo.state = S_PLAY_STND
				end
			end
		else
			return true
		end
	end
end

-- iframes
-- tick down and make the player flash white a little bit...
local function playerIFramesHandler(p)
	
	if p.pso.iframes
		
		if leveltime & 1
		and p.pso.iframes > 1
			p.mo.colorized = true
		else
			p.mo.colorized = false
		end
		
		p.pso.iframes = $-1
	end
end

-- displays barrier when hugging impassible lines.
addHook("MobjMoveBlocked", function(mo, thing, line)
	
	if not thing and line
	and mo.player
		mo.pso_blockedline = line
		mo.pso_blockedlinetimer = TICRATE/2
	end
end, MT_PLAYER)

local function playerBarrierDisplay(p)
	
	local barrier_range = 24
	local mo = p.mo
	
	if (p.pflags & PF_NOCLIP) return end
	
	if mo.pso_blockedline and mo.pso_blockedlinetimer
		
		if leveltime%16 == 0
			
			local l = mo.pso_blockedline
			
			--searchBlockmap("lines", function(mo, l)
			if l and l.valid
			
				if l.flags & ML_BLOCKMONSTERS
					local secs = {l.frontsector, l.backsector}
					local lineok
					
					for i = 1, #secs
						
						local sec = secs[i]
						for rover in sec.ffloors()
							
							if rover.flags & FF_BLOCKPLAYER
							and not (rover.flags & FF_BLOCKOTHERS)	-- blocks players but not other things.
								lineok = true
								break
							end
						end
					end
					
					if lineok
								
						local px, py = P_ClosestPointOnLine(mo.x, mo.y, l)
						--P_SpawnMobj(px, py, mo.z, MT_THOK)
								
						if R_PointToDist2(mo.x, mo.y, px, py) < barrier_range*FRACUNIT
									
							local a = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
									
							local base_spawnx = px + 8*cos(a-ANG1*90)
							local base_spawny = py + 8*sin(a-ANG1*90)
							--P_SpawnMobj(base_spawnx, base_spawny, mo.z, MT_THOK)
							--P_MoveOrigin(mo, base_spawnx, base_spawny, mo.z)
									
							local dist = P_RandomRange(-64, 64)
							local real_spawnx = base_spawnx + dist*cos(a)
							local real_spawny = base_spawny + dist*sin(a)
							local real_spawnz = mo.z + P_RandomRange(10, 64)*FRACUNIT
									
							local w = P_SpawnMobj(real_spawnx, real_spawny, real_spawnz, MT_PSO_LOBBYWINDOW)
							w.nopass = true
							w.color = SKINCOLOR_RED
							w.fuse = P_RandomRange(20, 50)
							w.angle = a
							w.scale = mo.scale*2/3
							
							for i = 1, 4
								local t = P_SpawnMobj(base_spawnx, base_spawny, mo.z + mo.height/2, MT_THOK)
								t.scale = 1
								t.destscale = (mo.scale/2) *i
								t.scalespeed = (mo.scale/8)*i
								t.fuse = 3 + 3*(i)
								t.tics = t.fuse
								t.sprite = SPR_NTHK
								t.frame = B|FF_PAPERSPRITE
								t.color = SKINCOLOR_WHITE
								t.angle = a
								t.momz = (FRACUNIT/2)*i
							end
							
							-- only display for local player
							if p ~= consoleplayer
								w.flags2 = $|MF2_DONTDRAW
							end
						end
					end
				end
			end	--, mo)
		end
		
		mo.pso_blockedlinetimer = $-1
		if not mo.pso_blockedlinetimer
			mo.pso_blockedline = nil
		end
	end
end

local function playerCheckLvSynch(p)
	
	local qdata = PSO.lib.questData()

	if netgame and p.pso.level > PSO.info.LEVELCAP	-- uuuhm
	and not p.pso.levelsynch
		p.pso.levelsynch = PSO.info.LEVELCAP
		PSO.lib.playerCalcStats(p, PSO.info.LEVELCAP)
	end
	
	if not qdata return end
	
	-- @todo: low level sync
	if qdata.levelsynch
	and gamemap ~= PSO.info.LOBBY_MAP
		if p.pso.level > qdata.levelsynch
			if not p.pso.levelsynch or p.pso.levelsynch > qdata.levelsynch
				p.pso.levelsynch = qdata.levelsynch
				p.pso.levelsynch_grind = qdata.levelsynch_grind
				--print(p.pso.levelsynch_grind)
				PSO.lib.playerCalcStats(p, qdata.levelsynch)
			end
		elseif p.pso.level < qdata.levelsynch_low
			if not p.pso.levelsynch
				p.pso.levelsynch = qdata.levelsynch_low
				p.pso.levelsynch_low = true
				PSO.lib.playerCalcStats(p, p.pso.levelsynch)
			end
		end
	elseif p.pso.levelsynch and p.pso.level <= PSO.info.LEVELCAP	-- cheaters!
		p.pso.levelsynch = nil
		p.pso.levelsynch_low = nil
		p.pso.levelsynch_grind = nil
		PSO.lib.playerCalcStats(p, p.pso.level)
	end
end

local function playerEnemyCollision(p)
	local mo = p.mo
	
	if p.pso.knockback or p.pso.blowdown return end	-- we don't handle enemy collision in these states.
	if mo.flags & (MF_NOCLIP|MF_NOCLIPTHING) return end	-- nuh
	
	local range = mo.radius*8
	
	local x1 = mo.x - (range/2 *mo.scale)
	local x2 = mo.x + (range/2 *mo.scale)
	local y1 = mo.y - (range/2 *mo.scale)
	local y2 = mo.y + (range/2 *mo.scale)
	
	local function search(ref, found)
		
		if not found or not found.valid return end				-- does not exist.
		if found.flags & (MF_NOCLIP|MF_NOCLIPTHING) or found.disablepushbox return end	-- cannot be collided with
		
		if found.flags & MF_ENEMY and PSO.info.enemies[found.type] and found.pso and found.pso.hp
		or found.type == MT_PSO_CRATE
		or found.type == MT_PSO_DARKFALZ_METEOR	-- uuuuugh.
			--print(found.flags & MF_NOCLIPTHING)
			-- check if we are colliding.
			-- remember that our collision bounds are squares
			if mo.x+mo.radius >= found.x-found.radius
			and mo.x-mo.radius <= found.x+found.radius
			and mo.y+mo.radius >= found.y-found.radius
			and mo.y-mo.radius <= found.y+found.radius
			and mo.z + mo.height >= found.z
			and mo.z <= found.z + found.height
				--print("A")
					
				-- calc overlap in both axes
				local overlapX = 0
				local overlapY = 0
					   
				if mo.x < found.x then
					overlapX = (mo.x + mo.radius) - (found.x - found.radius)
				else
					overlapX = (mo.x - mo.radius) - (found.x + found.radius)
				end
					   
				if mo.y < found.y then
					overlapY = (mo.y + mo.radius) - (found.y - found.radius)
				else
					overlapY = (mo.y - mo.radius) - (found.y + found.radius)
				end
					   
				-- get smallest adjustment...
				local newX, newY = mo.x, mo.y
				if abs(overlapX) < abs(overlapY) then
					-- Push out along the X-axis
					if overlapX > 0 then
						newX = mo.x - overlapX
					else
						newX = mo.x - overlapX
					end
				else
					-- Push out along the Y-axis
					if overlapY > 0 then
						newY = mo.y - overlapY
					else
						newY = mo.y - overlapY
					end
				end				
					
				-- try to see if we can move there
				if not P_TryMove(mo, newX, newY, true)
						
					-- push the enemy away instead
					-- calc overlap in both axes
					local overlapX = 0
					local overlapY = 0
							   
					if found.x < mo.x then
						overlapX = (found.x + found.radius) - (mo.x - mo.radius)
					else
						overlapX = (found.x - found.radius) - (mo.x + mo.radius)
					end
							   
					if found.y < mo.y then
						overlapY = (found.y + found.radius) - (mo.y - mo.radius)
					else
						overlapY = (found.y - found.radius) - (mo.y + mo.radius)
					end
							   
					-- get smallest adjustment...
					local newX, newY = found.x, found.y
					if abs(overlapX) < abs(overlapY) then
						-- Push out along the X-axis
						if overlapX > 0 then
							newX = found.x - overlapX
						else
							newX = found.x - overlapX
						end
					else
						-- Push out along the Y-axis
						if overlapY > 0 then
							newY = found.y - overlapY
						else
							newY = found.y - overlapY
						end
					end
					
					P_MoveOrigin(found, newX, newY, mo.z)
				
				-- movement completed normally
				else
					P_MoveOrigin(mo, newX, newY, mo.z)
				end
			end
		end
	end
	
	searchBlockmap("objects", search, mo, x1, x2, y1, y2)	
	
end

-- handles arena stuff
local function playerArenaThinker(p)
	if not p.psologin return end
	
	if PSO.lib.arenaQuestActive()
	and p.pso.arenapoints == nil	-- arena vars not initialized!
		PSO.lib.playerArenaQuestSetup(p)
		--print("DBG: arena quest start")
	elseif not PSO.lib.arenaQuestActive()
	and p.pso.arenapoints ~= nil	-- arena quest must be uninitialized
		PSO.lib.playerArenaQuestFinished(p)
		--print("DBG: arena quest end")
	end
end

-- quick item palette!
local function playerQuickItemPalette(p)
	if not p.psologin return end
	if p.pso.menu and p.pso.menu.ref
		p.pso.quickpalette = 0
		return
	end
	
			--for i = 1, min(#p.pso.buttons.mask, PSO.info.MAXPALSIZE)
	if p.pso.buttons.mask[7] == 1
		p.pso.quickpalette = PSO.info.QUICKPAL
		S_StartSound(nil, sfx_mopen, p)
	end
	
	if p.pso.quickpalette
		p.pso.quickpalette = $-1
		
		for i = 1, 6
			if p.pso.buttons.mask[i] == 1
				-- try to use an item in the palette
				
				p.pso.buttons.mask[i] = -FRACUNIT	-- shush.
				p.pso.quickpalette = 0
				
				local it = p.pso.palette.layout[1][i]
				
				if it
					local itinfo = PSO.lib.itemInfo(it)
					
					-- either there's no check or it returns true
					if not itinfo.canuse or itinfo.canuse(p)
						local invslot = PSO.lib.findPlayerItem(p, it)
						-- if invslot is 0 then we don't have any of these.
						if not invslot
							PSO.lib.dprint("No "..it.." in inventory!")
						else
							-- if we made it here then we can use it!
							PSO.lib.playerLoseItem(p, invslot, 1)	-- use the item up
							
							if itinfo.use
								itinfo.use(p)
							end
							PSO.lib.dprint("Used "..it)
						end
					else
						PSO.lib.dprint("Cannot use item "..it)
						S_StartSound(nil, sfx_mnot, p)
					end
				end
			end
		end
	end
end		

local function playerThinker(p)
	
	p.pso.playtime = $+1
	
	if not p.pso.quest
		playerSetQuest(p)
	end
	
	local hitlag = playerHitlagHandler(p)
	
	if not playerPitHandler(p)
	and not hitlag
		if not p.pso.deadtimer
			playerTargetting(p)
			local nododge = playerPickupDrops(p)
			playerMovement(p)
			playerSpringHandler(p)
			playerDodgeRoll(p, nododge)
			playerAISSpawn_Ground(p)
			playerAISVisuals(p)
			playerHandleDamage(p)
			
			if p.psologin
				playerBackPalette(p)
				playerWeaponActions(p)
				playerPPRegen(p)
				playerHandlePaletteSwitch(p)
				playerQuickItemPalette(p)
			end
			
			playerUpdateWeapon(p)
			playerTerrainDamage(p)
			PSO.lib.handleStatus(p.mo)
			playerIFramesHandler(p)
		else
			playerUpdateWeapon(p)
			playerAISVisuals(p)
			playerHandleDeath(p)
		end
	end
	
	-- do this to prevent HORRIBLE THINGS from happening
	if hitlag
		playerUpdateWeapon(p)
	end
	
	playerHandleQuestTrackers(p)
	playerHandleTeleport(p)
	playerJumpPad(p)
	playerHideHUD(p)
	noPPNotif(p)
	playerLowHP(p)
	playerHandleNotification(p)
	
	-- only display at normal size.
	if not p.pso.ais
		playerBarrierDisplay(p)
	end
	
	PSO.lib.handleEmergency(p)
	PSO.lib.playerHandleRareDropAnim(p)
	PSO.lib.playerHandleDropText(p)
	
	-- special stuff
	playerAvengerVisuals(p)
	playerHunterGale(p)
	playerSwordbreaker(p)
	playerPhotonBlast(p)
	playerManageDPSCounter(p)
	
	-- injury status. handled here to ensure nothing brings us above that threshold...
	playerInjuryHandler(p)
	
	-- level synch checker
	playerCheckLvSynch(p)
	
	-- The Arena!
	playerArenaThinker(p)
	
	-- make sure this doesn't stay on
	if not PSO.lib.playerHasPB(p)
		p.pso.pb = 0
	end
	
	p.lives = 1		-- only for singleplayer so that the retry command doesn't work
	p.dashmode = 0	-- dashmode is purely visual but we still don't want it.
	p.powers[pw_underwater] = 999999
	
	-- for cool visual!
	if p.psologinfinish
		p.psologinfinish = $-1
		
		if p.psologinfinish == 10
			S_SetInternalMusicVolume(100, p)
			if p == consoleplayer
				S_SetMusicPosition(0)	-- restart the music
				P_RestoreMusic(p)
			end
			S_StartSound(mo, sfx_telein)
			p.powers[pw_flashing] = TICRATE*3 -1
		end
	end
	
	-- eeeeh?
	if p.psologinfinish and p.psologinfinish <= 0
		p.psologinfinish = nil
	end
	
	-- unfuck the camera on spawn
	if leveltime == 2
		p.pflags = $|PF_NOCLIP
	elseif leveltime == 3
		p.pflags = $ & ~PF_NOCLIP
	end
	
	-- credits stuff
	if p.pso.credits
		p.pso.credits = $+1
	end
	
	PSO.lib.runHook("playerThink", p)
	
	playerEnemyCollision(p)
	
	-- singleplayer hack, kill tailsbot
	if not netgame
	and players[1]
	and players[1].valid
	and players[1].bot
		G_RemovePlayer(1)
	end
end

-- first person handler
-- courtesy of Spectrum
local chasecam = CV_FindVar("chasecam")

local function playerfunc(p)
	if p.PSO_chasecam == 0 and p.mo and p.mo.valid
		p.drawangle = p.mo.angle
	end
end

-- this needs to go before for weapon updating.
PSO.lib.registerPlayerFunc(playerfunc)
PSO.lib.registerPlayerFunc(playerThinker)

COM_AddCommand("pso_chasecamstatus", function(p, arg)
	local val = tonumber(arg)
	if val ~= 1 and val ~= 0
		CONS_Printf(p, "Don't use this command.")
		return
	end
	p.PSO_chasecam = val
end)

local function thinkfunc()
	if consoleplayer and consoleplayer.valid and chasecam
	and consoleplayer.PSO_chasecam ~= chasecam.value
		COM_BufInsertText(consoleplayer, "pso_chasecamstatus "..chasecam.value)
	end
end

PSO.lib.registerFrameFunc(thinkfunc)

-- player: reset quest specific trackers:
addHook("MapChange", function()
	for p in players.iterate do
		
		PSO.lib.resetPlayer(p)
		playerSetQuest(p)
	end
end)

-- player: reset potential specific stuff
-- this CANNOT be done in mapchange as player.mo cannot be accessed there and is used for some potentials.
addHook("MapLoad", do
	local firstp
	for p in players.iterate do
		
		firstp = p
		
		-- reset weapon potentials (we fake switching in and out)
		PSO.lib.runHook("switchingWeapons", p)
		PSO.lib.runHook("switchedWeapons", p)
		PSO.lib.closeMenu(p)
		p.pso.hidehud = nil
		
		local mo = p.mo
		-- ais related fixup
		if mo and mo.valid
			mo.height = mobjinfo[MT_PLAYER].height
			mo.radius = mobjinfo[MT_PLAYER].radius
			p.height = mo.height
		end
	end
	
	-- set default respawn point to that player's location.
	if firstp
		server.defaultspawnpoint = {firstp.mo.x, firstp.mo.y, firstp.mo.z, firstp.mo.angle}
	end
	
	-- uuuh this is also a very stupid fucking hack, but change the sky if we have dark falz unlocked.
	if gamemap == PSO.info.LOBBY_MAP and PSO.lib.getLocalQuestProgress() >= 16
		P_SetupLevelSky(67, consoleplayer)
	end
end)

-- if we have actionspeed
addHook("PlayerCmd", function(p, cmd)
	if p.pso and p.pso.actionspeed ~= nil and not p.pso.menu.ref	-- player controls are frozen during menu anyway
	
		local maxmove = 50 * p.pso.actionspeed / 100
	
		if cmd.sidemove < -maxmove
			cmd.sidemove = -maxmove
		elseif cmd.sidemove > maxmove
			cmd.sidemove = maxmove
		end
		
		if cmd.forwardmove < -maxmove
			cmd.forwardmove = -maxmove
		elseif cmd.forwardmove > maxmove
			cmd.forwardmove = maxmove
		end
	end
end)