local exportmode = false

-- these are used to know if we should overwrite data with the servers' etc when stuff is unlocked.
local local_quest_unlocks = {}
local local_quest_progress = 0

PSO.lib.getLocalUnlocks = function()
	return local_quest_unlocks
end

PSO.lib.setLocalUnlocks = function(t)
	local_quest_unlocks = t
end

-- this will be used for NPC dialogue in the future
PSO.lib.getLocalQuestProgress = function()
	return local_quest_progress
end

PSO.lib.setLocalQuestProgress = function(n)
	local_quest_progress = n
end

-- datapods
freeslot("MT_PSO_DATAPOD", "SPR_DTPD")

mobjinfo[MT_PSO_DATAPOD] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	flags = MF_NOGRAVITY,
	radius = 64<<FRACBITS,
	height = 64<<FRACBITS,
}

addHook("MobjThinker", function(mo)
	
	if not mo.extravalue1
		mo.sprite = SPR_DTPD
		mo.frame = A
		mo.flags2 = $ | MF2_SPLAT
		mo.renderflags = $|RF_FLOORSPRITE|RF_NOSPLATBILLBOARD		
		mo.extravalue1 = 1
		mo.tracer = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	end
	
	mo.tracer.sprite = SPR_DTPD
	mo.tracer.frame = B|FF_ADD
	mo.tracer.tics = 4
	if leveltime & 4 < 2
		mo.tracer.frame = $|FF_TRANS20
	else
		mo.tracer.frame = $|FF_TRANS60
	end
end, MT_PSO_DATAPOD)

PSO.info.questmaps = {}
-- list of maps that can be selected for quests

local questgrading = {
	"QG_TIME",	-- grade by time
	"QG_KILLS",	-- grade # of kills (of a specific enemy?)
	"QG_DAMAGETAKEN",	-- damage taken
	"QG_NOINCAP",		-- extra points for no incapacitation
	"QG_QUESTPOINTS",	-- TBI
}

for i = 1, #questgrading
	rawset(_G, questgrading[i], i)
end

PSO.info.quests = {
	
	{
		title = "Campaign Quests",
		tip = "Play through the story and unlock new quests.",
		--unlocked = true,
		quests = {},
	},
	
	{
		title = "Free Field",
		tip = "TBA in a later update.", --"Explore freely and cause PSE Bursts for EXP & drops.",
		--unlocked = false,
		quests = {},
	},
	
	{
		title = "Arena Quests",
		tip = "Face off against tough boss rematches.",
		--unlocked = false,
		quests = {},
	},
	
	{
		title = "Challenge Quests",
		tip = "TBA in a later update.", --"Start at Lv1 and make do with what you find.",
		--unlocked = false,
		quests = {},
	},
	
	{
		title = "???",	--"Battle Arena",
		tip = "TBA in a later update.",
		--unlocked = false,
		quests = {},
	},
}

/*
	quest table def:
	["name"] = {
		title = "",
		quests = {quest_t, quest_t...}
	}
	
*/

/*
	"quest_t" format:	(tentative type, doesn't actually exist)
	{
		>>COSMETIC STUFF
		name: 		quest name
		objective:	description of the objective
		desc: 		description
	
	
		>>ACTUAL GAMEPLAY RELEVANT STUFF
		map:		map number the quest takes place in
					(players will be warped to it as soon as everyone gets to the teleporter)
		
		layout = {}: 	object layout for the map.
						worth noting the object layout CANNOT be changed in the lobby.
		
		levels = {
			[DIFF_NORMAL] = x,
			[DIFF_HARD] = x,
			[...]
		}: enemy level per difficulty
		
		drops = {
			[DIFF_NORMAL] = {
			
			},
			
			[DIFF_HARD] = {
			
			},
			(...)
			
			special drop tables per difficulty (used for crates at the end of the mission)
		},
		
		
		grading = {
			{QG_TIME, full_point_time, 0_point_time, full_time, low_time},
			{QG_KILLS, max_points, max_kills, [MT_ type]},
			{QG_NOINCAP, points_for_no_incapacitation},
			{QG_DAMAGETAKEN, full_point_damagetaken, 0_point_damagetaken, low_damage_taken, high_damage_taken},
		}
		
		-- Grading will total all the points and then award them as Type points (with a multiplier depending on difficulty)
		-- Grading works like so based on a percentage:
		
		-- 0% - 49%: C Rank
		-- 50% - 79%: B Rank
		-- 80% - 99%: A Rank
		-- 100% S Rank
	}
*/


-- exports the object layout in the console so that it can be used for quests

PSO.lib.questExport = function()
	
	local fstr = "{\n"
	
	for mt in mapthings.iterate do
		if mt.y > PSO.info.LOBBY_YPOS
			local mtype = 0
			
			local mnum = mt.type
			if mt.mobj
				mtype = mt.mobj.type
				
				-- don't care.
				if mt.mobj.flags & MF_SCENERY
					continue
				end
				
				if PSO.info.mobjNameLookup[mtype]
					mtype = PSO.info.mobjNameLookup[mtype]	-- get mobj name.
				end
			end
			
			local options = 0
			local extrainfo = 0
			
			if udmf
				extrainfo = mt.args[5]
				if mt.args[6]
					options = $|1
				end
				
				if mt.args[7]
					options = $|4
				end
				
				if mt.args[8]
					options = $|8
				end
				
			else
				extrainfo = mt.extrainfo
				options = mt.options
			end
			
			
			fstr = $.."{"..mtype..", "..mt.x..", "..mt.y..", "..mt.angle..", "..options..", "..mt.z..", "..extrainfo..", "..mnum.."},\n"
		end
	end
	fstr = $.."}"
	
	print(fstr)
end

COM_AddCommand("qdump", function(p)
	if netgame return end	-- DO NOT.
	exportmode = (exportmode == false and true or false)
end)


-- accepts a quest and executes linedef 3000
-- opens the door to the teleporter and sets the exit params
PSO.lib.acceptQuest = function(section, qnum, map, difficulty, synch, special)
	
	-- this is okay actually, we're going to need this lol.
	if not special
		assert(gamemap == 1, "Attempting to accept a quest outside the Low Orbit Station?")
	end
	
	server.questdata = {
		
		map = map,
		section = section,
		index = qnum,
		difficulty = difficulty
	}
	
	local qinfo = PSO.lib.questInfo()
	
	if synch
		if qinfo.levelsynch
		and qinfo.levelsynch[difficulty]
			server.questdata.levelsynch_low = qinfo.levelsynch[difficulty][1]
			server.questdata.levelsynch = qinfo.levelsynch[difficulty][2]
			server.questdata.levelsynch_grind = qinfo.levelsynch[difficulty][3]
		end
	end
	
	if qinfo.arenalist
		server.questdata.arenaindex = 1
		server.questdata.arenatimer = 0
	end
		
	
	G_SetCustomExitVars(map, 1)
	
	if not special
		S_StartSound(nil, sfx_qacpt)
		P_LinedefExecute(3000)
	end
	
	for p in players.iterate
		if p.pso
			p.pso.ironwill = nil	-- not used
		end	
	end
end

PSO.lib.questData = function()
	if not server.questdata return nil end
	
	return server.questdata
end

PSO.lib.questInfo = function()
	local dta = PSO.lib.questData()
	if not dta return end
	
	
	return PSO.info.quests[dta.section].quests[dta.index]
end

PSO.lib.cancelQuest = function()
	server.questdata = nil
	G_SetCustomExitVars(1, 1)
	G_ExitLevel()
end

PSO.lib.goToQuest = function()
	assert(gamemap == 1, "Attempting to get to a quest from outside the Low Orbit Station?")
	assert(server.questdata, "No quest data")
	
	G_SetCustomExitVars(1, 1)
end

local function hardModeTextureSwaps()
	if PSO.lib.questDifficulty() >= DIFF_HARD
		
		local header = mapheaderinfo[gamemap]
		if header.hardsky
			P_SetupLevelSky(header.hardsky)
		end
		
		if header.hardweather
			P_SwitchWeather(_G[header.hardweather] or 0)
		end
		
		for s in sectors.iterate
			
			if s.tag == PSO.info.ENCORETAG
				--print("Found ref")
				for s2 in sectors.iterate
					
					if s2 ~= s1
					and s2.lines[0].v1.y > PSO.info.LOBBY_YPOS*FRACUNIT	-- don't affect lobby
						if s2.floorpic == s.floorpic
							s2.floorpic = s.ceilingpic
						end
						if s2.ceilingpic == s.floorpic
							s2.ceilingpic = s.ceilingpic
						end
					end
				end
			end
			
			if header.hardbrightness
			and s.lines[0].v1.y > PSO.info.LOBBY_YPOS*FRACUNIT
				s.lightlevel = max(1, $+header.hardbrightness)
			end
		end
		
		for l in lines.iterate
			if l.tag == PSO.info.ENCORETAG
				local s = l.frontside
				for s2 in sides.iterate
					
					if s2.line ~= s --and s2.v1
					--and s2.v1.y > PSO.info.LOBBY_YPOS*FRACUNIT	-- don't affect lobby
						local checks = {"toptexture", "midtexture", "bottomtexture"}
						for i = 1, #checks
							
							local c = checks[i]
							if s2[c] == s.toptexture
								s2[c] = s.midtexture
							end
						end
					end
				end
			end
		end
	end
end

-- call before processing objects (i.e: rooms)
PSO.lib.spawnQuestObjects = function(tp_only)
	
	if not server.questdata return end
	if gamemap ~= server.questdata.map return end
	
	local q = PSO.info.quests[server.questdata.section].quests[server.questdata.index]
	local objs = q.layout
	
	local lastspawnpoint
	local playersspawned = {}
	
	for i = 1, #objs
		-- format: {type, x, y, options, z, extrainfo}
		local o = objs[i]
		local mobjtype = o[1]
		local x = o[2]*FRACUNIT
		local y = o[3]*FRACUNIT
		local ang = o[4]
		local options = o[5]
		local z = o[6]*FRACUNIT
		local extrainfo = o[7]
		local mapthingnum = o[8]
		
		local sec = R_PointInSubsector(x, y).sector
			-- mapthing z is distance from floor and not absolute z height (or ceiling in antigrav)
		if options & 2
			z = sec.ceilingheight - z
		else
			z = $ + sec.floorheight
		end
		
		if mapthingnum <= 32	-- player start 1-32
			local p = players[mobjtype]
			
			if p and p.valid and p.mo and p.mo.valid
				P_MoveOrigin(p.mo, x, y, z)
				p.mo.angle = ang*ANG1
				if options & 2
					p.mo.flags2 = $|MF2_OBJECTFLIP
				end
			end
			playersspawned[mobjtype] = $ and $+1 or 1
			lastspawnpoint = o
			
		elseif not tp_only	-- regular object
		
			local mt = P_SpawnMobj(x, y, z, mobjtype)
			mt.options = options
			mt.extrainfo = extrainfo
			mt.mtangle = ang
			mt.angle = ang*ANG1
			mt.mtz = (z - mt.subsector.sector.floorheight) /FRACUNIT
			if options & 2
				mt.flags2 = $|MF2_OBJECTFLIP
			end			
		end
	end
	
	-- check for players that haven't been spawned
	if lastspawnpoint
		server.lastspawnpoint = {lastspawnpoint[2]*FRACUNIT, lastspawnpoint[3]*FRACUNIT, lastspawnpoint[6]*FRACUNIT, lastspawnpoint[4]*ANG1, lastspawnpoint[5]}
		for p in players.iterate
			
			if not playersspawned[#p] and p.mo and p.mo.valid
				
				local mobjtype = lastspawnpoint[1]
				local x = lastspawnpoint[2]*FRACUNIT
				local y = lastspawnpoint[3]*FRACUNIT
				local ang = lastspawnpoint[4]*ANG1
				local options = lastspawnpoint[5]
				local z = lastspawnpoint[6]*FRACUNIT
				
				local numspawns = playersspawned[mobjtype]
				
				-- make players spawn at different positions i guess
				local angs = {ang*ANG1 - ANG1*180, ang*ANG1 - ANG1*90, ang*ANG1 + ANG1*90}
				local shift_ang = angs[ ((numspawns-1) %3) +1 ] or 0
				local shift_dist = 32 + (numspawns/3)*32
				
				local sx = x + shift_dist*cos(shift_ang)
				local sy = y + shift_dist*sin(shift_ang)
				
				P_MoveOrigin(p.mo, sx, sy, z)
				p.mo.angle = ang*ANG1
				if options & 2
					p.mo.flags2 = $|MF2_OBJECTFLIP
				end
				
				playersspawned[mobjtype] = $ and $+1 or 1
				--lastspawnpoint = o
				
			end
		end
	end
	
	-- are there datapods?
	if q.datapods
		for i = 1, #q.datapods
			
			local dta = q.datapods[i]
			local x = dta[1]
			local y = dta[2]
			local z = dta[3]
			local dt = P_SpawnMobj(x*FRACUNIT, y*FRACUNIT, z*FRACUNIT, MT_PSO_DATAPOD)
			dt.extravalue2 = i
		end
	end
end

freeslot("sfx_ltime")
addHook("PreThinkFrame", function()
	
	/*if leveltime == 0
	and not server.questspawned
		PSO.lib.spawnQuestObjects(true)	-- setup player spawns
		server.questspawned = true
	end*/
	
	if leveltime == 0
	and server
	and not server.questspawned
		
		if exportmode
		and not netgame	-- BY NO MEANS
			PSO.lib.dprint("Exporting quest objects...")
			PSO.lib.questExport()
		end
		
		hardModeTextureSwaps()
		
		PSO.lib.dprint("Spawn quest objects...")
		PSO.lib.spawnQuestObjects()
		
		server.questspawned = true
	end
	
	if server		
		
		server.psounlocks = $ or {}
		
		-- unintialized?
		if server.uselevelsynch == nil
			server.uselevelsynch = netgame	-- by default, levelsynch is ON in netgames, OFF in singleplayer.
		end
		
		if server.questdata 
			
			if server.questdata.ap_delay
				server.questdata.ap_delay = $-1
			elseif server.questdata.ap_award
				local amount = min(server.questdata.ap_award, 100)
				for p in players.iterate
					if p.pso and p.pso.arenapoints ~= nil
						p.pso.arenapoints = $+amount
					end
				end	
				
				server.questdata.ap_award = $-amount
				if not server.questdata.ap_award
					S_StartSound(nil, sfx_chchng)
				else
					S_StartSound(nil, sfx_menu1)
				end
			end
			
			if server.questdata.arenatimer_on
				if server.questdata.arenatimer
					server.questdata.arenatimer = $-1
					
					if server.questdata.arenatimer <= TICRATE*20
					and server.questdata.arenatimer % TICRATE == 0
						S_StartSound(nil, sfx_ltime)
					end
					
					if not server.questdata.arenatimer
						server.questdata.failed = 1
						
						for p in players.iterate
							PSO.lib.playerNotification(p, "\x82".."TIME'S UP! QUEST FAILED...")
							p.powers[pw_nocontrol] = 32767
						end
						
					end
				end
			end
			
			if server.questdata.failed
				server.questdata.failed = $+1
				
				if server.questdata.failed == 2
					S_FadeOutStopMusic(500)
				end
				
				if server.questdata.failed == TICRATE
					S_StartSound(nil, sfx_qfail)
				end
				
				if server.questdata.failed == TICRATE*5
					
					G_SetCustomExitVars(1, 1)
					G_ExitLevel()

					for p in players.iterate
						if p and p.mo and p.mo.valid
							PSO.lib.rollPlayerShop(p)
							PSO.lib.healObject(p.mo, p.mo, p.pso.maxhp)
							p.pso.deadtimer = 0
						end
					end
					
				end
			end
		end
	end
end)

-- unset variable on map change
addHook("MapChange", function(mapnum)
	server.questspawned = nil
	
	-- change music if we're going to the map with music data
	local dta = PSO.lib.questData()
	if dta and dta.map == mapnum
		local info = PSO.lib.questInfo()
		if info.questmusic
			--S_ChangeMusic(info.questmusic, true)
			server.bufquestmusic = info.questmusic
		end
	end
end)

addHook("MusicChange", function(oldn, newn)
	
	if server and server.valid and server.bufquestmusic
		local s = server.bufquestmusic
		server.bufquestmusic = nil
		return s
	end
end)

-- in map 1, remove server quest data.
addHook("MapLoad", do
	if gamemap == 1
		server.questdata = nil
	end
	
	if server and server.unlockmsg
		
		for p in players.iterate
			PSO.lib.playerNotification(p, "\x82".."UNLOCKED: ".."\x80"..server.unlockmsg)
		end
		server.unlockmsg = nil
	end
	
end)

PSO.lib.questDifficulty = function()
	if not server.questdata
		return DIFF_NORMAL
	else
		return server.questdata.difficulty
	end
end

-- table that keeps track of unlocked maps & quests
PSO.lib.unlockQuest = function(category, quest, diff)

	if not server.psounlocks[category]
		server.psounlocks[category] = {}
	end
	
	-- SPECIFICALLY...
	if category == 1
	and quest == 1
		if diff == DIFF_HARD
			server.unlockmsg = "Hard Mode (All Quests)"
		end
	end
	
	if PSO.lib.maxDifficultyUnlocked() < DIFF_VERYHARD and diff == DIFF_VERYHARD
		server.unlockmsg = "Very Hard Mode (All Quests)"
	end
	
	if not server.unlockmsg
		server.unlockmsg = PSO.info.quests[category].quests[quest].name.." ("..PSO.info.diff_shortnames[diff]..")"
	end
	
	if not server.psounlocks[category][quest]
	or server.psounlocks[category][quest] < (diff or DIFF_NORMAL)
		if server.psounlocks[category][quest]
			server.psounlocks[category][quest] = max($, (diff or DIFF_NORMAL))
		else
			server.psounlocks[category][quest] = diff or DIFF_NORMAL
		end
	end
	
	local q = PSO.info.quests[category].quests[quest]
	if q.prog and q.prog > (server.psoprogress or 0)
		server.psoprogress = q.prog
		--print("Server prog is now "..q.prog)
	end
	
	PSO.lib.localUnlockQuest(category, quest, diff)
	--PSO.lib.server_saveData()
end

-- same but for my own save file... and without messages.
PSO.lib.localUnlockQuest = function(category, quest, diff)
	
	if not consoleplayer or not consoleplayer.valid	-- dedicated is fine
	or consoleplayer.psologin	-- make sure I have fully logged in before saving.
	
		local loc = PSO.lib.getLocalUnlocks()
		
		if not loc[category]
			loc[category] = {}
		end
		
		/*-- do nothing if we already have it, unless the difficulty is lower.
		if loc[category][quest] and loc[category][quest] < diff
			return
		end*/

		if not loc[category][quest]
			loc[category][quest] = diff or DIFF_NORMAL
		else
			loc[category][quest] = max($, diff or DIFF_NORMAL)
		end	
		
		-- NOW, if that quest happens to be a quest with "prog", unlock all the quests in the same category with less "prog".
		local q = PSO.info.quests[category].quests[quest]
		if q.prog and q.prog > (PSO.lib.getLocalQuestProgress() or 0)
			PSO.lib.setLocalQuestProgress(q.prog)
			--print("Local prog is now "..q.prog)
			for i = 2, q.prog
				if not loc[category][i]
					loc[category][i] = diff or DIFF_NORMAL
				else
					loc[category][i] = max($, diff or DIFF_NORMAL)
				end
			end
		end
		
		PSO.lib.setLocalUnlocks(loc)
		PSO.lib.server_saveData()	-- save it here locally instead.
	end
end

PSO.lib.questUnlocked = function(category, quest)
	--if not server.psounlocks[category] return false end
	
	if quest == nil
		-- most likely true if we have any quest in there
		if server.psounlocks[category] != nil
			return true
		end
		
		-- else, individually check for unlocked quests? there are quests that can also be marked as unlocked from the start.
		local c = PSO.info.quests[category]
		for i = 1, #c.quests
			local q = c.quests[i]
			if not q.locked
			or (server.psounlocks[category] and server.psounlocks[category][i])
				return true
			end
		end
	end
	
	return server.psounlocks[category] and server.psounlocks[category][quest]
end

-- synched prog
PSO.lib.serverProgressValue = function()
	local i = 2
	
	if not server.psounlocks or not server.psounlocks[1] return 0 end
	while 1
		if server.psounlocks[1][i]
			i = $+1
		else
			break
		end
	end
	
	return i-1
end

-- local prog
PSO.lib.localProgressValue = function()
	local i = 2
	
	local t = PSO.lib.getLocalUnlocks()
	if not t or not t[1] return 0 end
	
	while 1
		if t[1][i]
			i = $+1
		else
			break
		end
	end
	
	return i-1
end

-- a bit of a hack but because new difficulties are first unlocked on 1-1...
PSO.lib.maxDifficultyUnlocked = function()
	return PSO.lib.questUnlocked(1, 1) or DIFF_NORMAL
end

-- finds the best available drops from the quest loot pool
PSO.lib.findBestQuestDrops = function(quest, difficulty, itype)
	
	local bestitems = {}
	if not quest.drops return {} end
	local info = quest.drops[difficulty]
	
	if info
		for i = 1, #info
			local it = info[i]
			
			if type(it[1]) ~= "number"	-- not meseta lol
				local iteminfo = PSO.lib.itemInfo(it[1])				
				if iteminfo and iteminfo.rarity and (itype == nil or iteminfo.type == itype)
					bestitems[#bestitems+1] = it[1]
				end
			end
		end
	end
	
	-- now sort
	table.sort(bestitems, function(a, b)
							local info1 = PSO.lib.itemInfo(a)
							local info2 = PSO.lib.itemInfo(b)
							return info1.rarity > info2.rarity
						end)
						
	return bestitems
end

-- finds the best available drops from a map's crate loot pool
PSO.lib.findBestCrateDrops = function(mapnum, difficulty, itype)
	
	local bestitems = {}
	if not PSO.info.crateDropTables[mapnum] return {} end
	local info = PSO.info.crateDropTables[mapnum][difficulty]
	
	if info
		for i = 1, #info
			local it = info[i]
			
			if type(it[1]) ~= "number"	-- not meseta lol
				local iteminfo = PSO.lib.itemInfo(it[1])				
				if iteminfo and iteminfo.rarity and (itype == nil or iteminfo.type == itype)
					bestitems[#bestitems+1] = it[1]
				end
			end
		end
	end
	
	-- now sort
	table.sort(bestitems, function(a, b)
							local info1 = PSO.lib.itemInfo(a)
							local info2 = PSO.lib.itemInfo(b)
							return info1.rarity > info2.rarity
						end)
						
	return bestitems
end



-- returns 3 sorted lists of the best weapons, armors and units.
PSO.lib.bestQuestDrops = function(quest, difficulty)
	
	local bestweapons = {}
	local bestarmors = {}
	local bestunits = {}
	local bestmaterials = {}
	
	local types = {IT_WEAPON, IT_ARMOR, IT_UNIT, IT_MATERIAL}
	local lookup = {bestweapons, bestarmors, bestunits, bestmaterials}	-- refs to the table we can directly write to
	
	for i = 1, #types
		
		local best_crates = PSO.lib.findBestCrateDrops(quest.map, difficulty, types[i])
		local best_quest = PSO.lib.findBestQuestDrops(quest, difficulty, types[i])
		
		-- now for enemies
		-- this only works if we have a layout.
		
		local best_enemies = {}
		
		if quest.layout
			for j = 1, #quest.layout
				local tt = quest.layout[j][1]
				if PSO.info.enemies[tt]
					PSO.lib.mergeTablesNoDupes(best_enemies, PSO.lib.findBestEnemyDrops(tt, quest.levels[difficulty], types[i]))
				end
			end
		end
		
		-- supression target
		if quest.target
		and PSO.info.enemies[quest.target]
			PSO.lib.mergeTablesNoDupes(best_enemies, PSO.lib.findBestEnemyDrops(quest.target, quest.levels[difficulty], types[i]))
		end
		
		PSO.lib.mergeTablesNoDupes(lookup[i], best_crates)
		PSO.lib.mergeTablesNoDupes(lookup[i], best_quest)
		PSO.lib.mergeTablesNoDupes(lookup[i], best_enemies)
		
		table.sort(lookup[i], function(a, b)
								local info1 = PSO.lib.itemInfo(a)
								local info2 = PSO.lib.itemInfo(b)
								
								if info1.rarity ~= info2.rarity
									return info1.rarity > info2.rarity
								end
								
								if info1.type == IT_WEAPON
									if info1.atk ~= info2.atk
										return info1.atk > info2.atk
									else
										return info1.weapontype > info2.weapontype
									end
								
								elseif info1.type == IT_ARMOR
									if info1.def ~= info2.def
										return info1.def > info2.def
									end
								end
								
								-- default
								return info1.rarity > info2.rarity
							end)
							
		--print(best_enemies[1])
	end
	
	return lookup
end

-- returns true if we are currently in an arena quest.
-- this will not be true if we are in the lobby or if the quest is complete
PSO.lib.arenaQuestActive = function()
	local qinfo = PSO.lib.questInfo()
	return qinfo and qinfo.arena and gamemap != PSO.info.LOBBY_MAP and server.questdata and not server.questdata.complete
end

COM_AddCommand("psod_unlock", function(p, map, quest, diff)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	PSO.lib.unlockQuest(tonumber(map), tonumber(quest), tonumber(diff) or DIFF_NORMAL)
	PSO.lib.localUnlockQuest(tonumber(map), tonumber(quest), tonumber(diff) or DIFF_NORMAL)
	
end, COM_ADMIN)

COM_AddCommand("psod_enemylv", function(p, lv)

	if not server.psocheats
		print("Enable cheats with 'psod_debugcheats' to use this command")
		return 
	end
	
	server.psoenemylv = tonumber(lv or 0)
	
end, COM_ADMIN)